Type: Package
Package: DiagrammeR
Title: Graph/Network Visualization
Version: 1.0.12
Authors@R: c(
    person("Richard", "Iannone", , "riannone@me.com", role = c("aut", "cre"),
           comment = c(ORCID = "0000-0003-3925-190X")),
    person("Olivier", "Roy", , "olivierroy71@hotmail.com", role = "aut")
  )
Description: Build graph/network structures using functions for stepwise
    addition and deletion of nodes and edges. Work with data available in
    tables for bulk addition of nodes, edges, and associated metadata. Use
    graph selections and traversals to apply changes to specific nodes or
    edges. A wide selection of graph algorithms allow for the analysis of
    graphs. Visualize the graphs and take advantage of any aesthetic
    properties assigned to nodes and edges.
License: MIT + file LICENSE
URL: https://rich-iannone.github.io/DiagrammeR/,
        https://github.com/rich-iannone/DiagrammeR
BugReports: https://github.com/rich-iannone/DiagrammeR/issues
Depends: R (>= 4.1)
Imports: dplyr (>= 1.0.7), glue (>= 1.5.0), htmltools (>= 0.5.2),
        htmlwidgets (>= 1.5), igraph (>= 2.0.0), magrittr (>= 1.5),
        purrr (>= 0.3.4), RColorBrewer (>= 1.1-2), readr (>= 2.1.1),
        rlang (>= 1.1.0), cli, rstudioapi (>= 0.7), scales (>= 1.1),
        stringr (>= 1.4), tibble (>= 3.1), tidyr (>= 1.1), viridisLite
        (>= 0.4.2), visNetwork (>= 2.1.0)
Suggests: curl, DiagrammeRsvg, knitr, rmarkdown, rsvg, testthat (>=
        3.1.0), withr
Config/testthat/edition: 3
Encoding: UTF-8
LazyData: true
RoxygenNote: 7.3.3
NeedsCompilation: no
Packaged: 2026-04-27 13:46:53 UTC; riannone
Author: Richard Iannone [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-3925-190X>),
  Olivier Roy [aut]
Maintainer: Richard Iannone <riannone@me.com>
Repository: CRAN
Date/Publication: 2026-04-27 14:50:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(print,dgr_graph)
export("%>%")
export(DiagrammeR)
export(DiagrammeROutput)
export(add_balanced_tree)
export(add_cycle)
export(add_edge)
export(add_edge_clone)
export(add_edge_df)
export(add_edges_from_table)
export(add_edges_w_string)
export(add_forward_edges_ws)
export(add_full_graph)
export(add_global_graph_attrs)
export(add_gnm_graph)
export(add_gnp_graph)
export(add_graph_action)
export(add_graph_to_graph_series)
export(add_grid_2d)
export(add_grid_3d)
export(add_growing_graph)
export(add_islands_graph)
export(add_mathjax)
export(add_n_node_clones)
export(add_n_nodes)
export(add_n_nodes_ws)
export(add_node)
export(add_node_clones_ws)
export(add_node_df)
export(add_nodes_from_df_cols)
export(add_nodes_from_table)
export(add_pa_graph)
export(add_path)
export(add_prism)
export(add_reverse_edges_ws)
export(add_smallworld_graph)
export(add_star)
export(clear_selection)
export(colorize_edge_attrs)
export(colorize_node_attrs)
export(combine_edfs)
export(combine_graphs)
export(combine_ndfs)
export(copy_edge_attrs)
export(copy_node_attrs)
export(count_asymmetric_node_pairs)
export(count_automorphisms)
export(count_edges)
export(count_graphs_in_graph_series)
export(count_loop_edges)
export(count_mutual_node_pairs)
export(count_nodes)
export(count_s_connected_cmpts)
export(count_unconnected_node_pairs)
export(count_unconnected_nodes)
export(count_w_connected_cmpts)
export(create_edge_df)
export(create_graph)
export(create_graph_series)
export(create_node_df)
export(delete_cache)
export(delete_edge)
export(delete_edges_ws)
export(delete_global_graph_attrs)
export(delete_graph_actions)
export(delete_loop_edges_ws)
export(delete_node)
export(delete_nodes_ws)
export(deselect_edges)
export(deselect_nodes)
export(display_metagraph)
export(do_bfs)
export(do_dfs)
export(drop_edge_attrs)
export(drop_node_attrs)
export(edge_aes)
export(edge_data)
export(export_csv)
export(export_graph)
export(filter_graph_series)
export(from_adj_matrix)
export(from_igraph)
export(fully_connect_nodes_ws)
export(fully_disconnect_nodes_ws)
export(generate_dot)
export(get_adhesion)
export(get_agg_degree_in)
export(get_agg_degree_out)
export(get_agg_degree_total)
export(get_all_connected_nodes)
export(get_alpha_centrality)
export(get_articulation_points)
export(get_authority_centrality)
export(get_betweenness)
export(get_cache)
export(get_closeness)
export(get_closeness_vitality)
export(get_cmty_edge_btwns)
export(get_cmty_fast_greedy)
export(get_cmty_l_eigenvec)
export(get_cmty_louvain)
export(get_cmty_walktrap)
export(get_common_nbrs)
export(get_coreness)
export(get_degree_distribution)
export(get_degree_histogram)
export(get_degree_in)
export(get_degree_out)
export(get_degree_total)
export(get_dice_similarity)
export(get_eccentricity)
export(get_edge_attrs)
export(get_edge_attrs_ws)
export(get_edge_count_w_multiedge)
export(get_edge_df)
export(get_edge_df_ws)
export(get_edge_ids)
export(get_edge_info)
export(get_edges)
export(get_eigen_centrality)
export(get_girth)
export(get_global_graph_attr_info)
export(get_graph_actions)
export(get_graph_from_graph_series)
export(get_graph_info)
export(get_graph_log)
export(get_graph_name)
export(get_graph_series_info)
export(get_graph_time)
export(get_jaccard_similarity)
export(get_last_edges_created)
export(get_last_nodes_created)
export(get_leverage_centrality)
export(get_max_eccentricity)
export(get_mean_distance)
export(get_min_cut_between)
export(get_min_eccentricity)
export(get_multiedge_count)
export(get_nbrs)
export(get_node_attrs)
export(get_node_attrs_ws)
export(get_node_df)
export(get_node_df_ws)
export(get_node_ids)
export(get_node_info)
export(get_non_nbrs)
export(get_pagerank)
export(get_paths)
export(get_periphery)
export(get_predecessors)
export(get_radiality)
export(get_reciprocity)
export(get_s_connected_cmpts)
export(get_selection)
export(get_similar_nbrs)
export(get_successors)
export(get_w_connected_cmpts)
export(grViz)
export(grVizOutput)
export(import_graph)
export(invert_selection)
export(is_edge_loop)
export(is_edge_multiple)
export(is_edge_mutual)
export(is_edge_present)
export(is_graph_connected)
export(is_graph_dag)
export(is_graph_directed)
export(is_graph_empty)
export(is_graph_simple)
export(is_graph_undirected)
export(is_graph_weighted)
export(is_node_present)
export(is_property_graph)
export(join_edge_attrs)
export(join_node_attrs)
export(layout_nodes_w_string)
export(mermaid)
export(mutate_edge_attrs)
export(mutate_edge_attrs_ws)
export(mutate_node_attrs)
export(mutate_node_attrs_ws)
export(node_aes)
export(node_data)
export(nudge_node_positions_ws)
export(open_graph)
export(recode_edge_attrs)
export(recode_node_attrs)
export(remove_graph_from_graph_series)
export(rename_edge_attrs)
export(rename_node_attrs)
export(renderDiagrammeR)
export(renderGrViz)
export(render_graph)
export(render_graph_from_graph_series)
export(reorder_graph_actions)
export(replace_in_spec)
export(rescale_edge_attrs)
export(rescale_node_attrs)
export(rev_edge_dir)
export(rev_edge_dir_ws)
export(save_graph)
export(select_edges)
export(select_edges_by_edge_id)
export(select_edges_by_node_id)
export(select_last_edges_created)
export(select_last_nodes_created)
export(select_nodes)
export(select_nodes_by_degree)
export(select_nodes_by_id)
export(select_nodes_in_neighborhood)
export(set_cache)
export(set_edge_attr_to_display)
export(set_edge_attrs)
export(set_edge_attrs_ws)
export(set_graph_directed)
export(set_graph_name)
export(set_graph_time)
export(set_graph_undirected)
export(set_node_attr_to_display)
export(set_node_attr_w_fcn)
export(set_node_attrs)
export(set_node_attrs_ws)
export(set_node_position)
export(to_igraph)
export(transform_to_complement_graph)
export(transform_to_min_spanning_tree)
export(transform_to_subgraph_ws)
export(trav_both)
export(trav_both_edge)
export(trav_in)
export(trav_in_edge)
export(trav_in_node)
export(trav_in_until)
export(trav_out)
export(trav_out_edge)
export(trav_out_node)
export(trav_out_until)
export(trav_reverse_edge)
export(trigger_graph_actions)
export(visnetwork)
export(x11_hex)
import(RColorBrewer)
import(glue)
import(htmlwidgets)
import(rlang)
import(visNetwork)
importFrom(magrittr,"%>%")
importFrom(magrittr,not)

#' Deselect any selected edges in a graph
#'
#' @description
#'
#' Deselect edges in a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#' @param edges A vector of edge IDs that should be deselected.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with
#' # a single path
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 5)
#'
#' # Select edges with IDs `1`
#' # and `3`
#' graph <-
#'   graph %>%
#'   select_edges_by_edge_id(
#'     edges = c(1, 3))
#'
#' # Verify that an edge selection
#' # has been made
#' graph %>% get_selection()
#'
#' # Deselect edge `1`
#' graph <-
#'   graph %>%
#'   select_edges_by_edge_id(
#'     edges = c(1, 3)) %>%
#'   deselect_edges(edges = 1)
#'
#' # Verify that the edge selection
#' # has been made for edges `1` and
#' # `3` and that edge `1` has been
#' # deselected (leaving only `3`)
#' graph %>% get_selection()
#'
#' @export
deselect_edges <- function(
    graph,
    edges
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If no edge selection available, return graph
  if (!graph_contains_edge_selection(graph)) {
    return(graph)
  }

  # Extract the graph's edge selection
  graph$edge_selection <-
    graph$edge_selection %>%
    dplyr::filter(!(edge %in% edges))

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

test_that("selecting a node in a graph is possible", {

  # Create a simple graph
  nodes <-
    create_node_df(
      n = 4,
      type = "letter",
      label = TRUE,
      value = c(3.5, 2.6, 9.4, 2.7))

  edges <-
    create_edge_df(
      from = c(1, 2, 3),
      to = c(4, 3, 1),
      rel = "leading_to",
      width = c(1, 2, 3))

  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Select nodes `1` and `3`
  graph_a_c <-
    select_nodes(graph = graph, nodes = c(1, 3))

  # Expect that a node selection is available
  expect_gt(nrow(graph_a_c$node_selection), 0)

  # Expect that an edge selection is not available
  expect_equal(nrow(graph_a_c$edge_selection), 0)

  # Expect that nodes `1` and `3` are part of a selection
  # object in `nodes`
  expect_equal(
    graph_a_c$node_selection$node,
    c(1, 3))

  # Select nodes where `value` > 3
  graph_val_gt_3 <-
    graph %>%
    select_nodes(
      conditions = value > 3)

  # Expect that nodes `1` and `3` are part of a
  # selection object in `nodes`
  expect_equal(
    graph_val_gt_3$node_selection$node,
    c(1, 3))

  # Select nodes where `value` < 3
  graph_val_lt_3 <-
    graph %>%
    select_nodes(
      conditions = value < 3)

  # Expect that nodes `2` and `4` are part of a
  # selection object in `nodes`
  expect_equal(
    graph_val_lt_3$node_selection$node,
    c(2, 4))

  # Select nodes where `value` == 2.7
  graph_val_eq_2_7 <-
    graph %>%
    select_nodes(
      conditions = value == 2.7)

  # Expect that node `4` is part of a selection
  # object in `nodes`
  expect_equal(
    graph_val_eq_2_7$node_selection$node,
    4
  )

  # Select nodes where `value` != 2.7
  graph_val_neq_2_7 <-
    graph %>%
    select_nodes(
      conditions = value != 2.7)

  # Expect that nodes `1`, `2`, and `3` are part of
  # a selection object in `nodes`
  expect_equal(
    graph_val_neq_2_7$node_selection$node,
    c(1, 2, 3))

  # Select nodes where `type` is `letter`
  graph_val_letter <-
    graph %>%
    select_nodes(
      conditions = type == "let")

  # Expect that nodes `1`, `2`, `3`, and `4` are
  # part of a selection object in `nodes`
  # FIXME seems like this test is failing silently.
  # we have integer(0) as a result. (checked it was the case on main)
  expect_true(
    all(
      graph_val_letter$node_selection$node ==
        c(1, 2, 3, 4)))

  # Select nodes where `type` is `letter` and filter
  # to nodes `1` and `2`
  graph_val_letter_a_b <-
    graph %>%
    select_nodes(
      conditions = type == "let",
      nodes = 1:2)

  # Expect that nodes `1` and `2` are part of a
  # selection object in `nodes`
  # FIXME seems like this test is failing silently.
  # we have integer(0) as a result. (checked on main too)
  expect_true(
    all(
      graph_val_letter_a_b$node_selection$node == c(1, 2)))

  # Create a union of selections in a magrittr pipeline
  graph_sel_union_a_b_c_d <-
    graph %>%
    select_nodes(nodes = c(1, 2)) %>%
    select_nodes(nodes = c(3, 4))

  # Expect that all 4 nodes are part of a selection
  # object in `nodes`
  expect_equal(
    graph_sel_union_a_b_c_d$node_selection$node,
    c(1, 2, 3, 4))

  # Create a intersection of selections in a
  # magrittr pipeline
  graph_sel_intersect <-
    graph %>%
    select_nodes(nodes = c(1, 2, 3)) %>%
    select_nodes(
      nodes = c(2, 3, 4),
      set_op = "intersect")

  # Expect that nodes `2` and `3` are part of a
  # selection object in `nodes`
  expect_equal(
    graph_sel_intersect$node_selection$node,
    c(2, 3))

  # Create a selection that is a difference
  # of selections
  graph_sel_difference <-
    graph %>%
    select_nodes(nodes = 1:3) %>%
    select_nodes(
      nodes = 2:3,
      set_op = "difference")

  # Expect that node `1` is part of a selection
  # object in `nodes`
  expect_equal(
    graph_sel_difference$node_selection$node,
    1)

  # Expect an error if specifying a node that
  # doesn't exist
  expect_snapshot(error = TRUE, {
    select_nodes(
      graph = graph,
      nodes = 5)
    # FIXME Maybe this should warn?
    select_nodes(
      graph = graph,
      nodes = 4:5)
  })

 })

test_that("selecting an edge in a graph is possible", {

  # Create a simple graph
  nodes <-
    create_node_df(
      n = 4,
      type = "letter",
      label = TRUE,
      value = c(3.5, 2.6, 9.4, 2.7))

  edges <-
    create_edge_df(
      from = c(1, 2, 3),
      to = c(4, 3, 1),
      rel = "leading_to",
      width = c(1, 2, 3))

  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Select nodes `1`->`4` and `2`->`3`
  graph_ad_bc <-
    select_edges(
      graph = graph,
      from = c(1, 2),
      to = c(4, 3))

  # Expect that nodes `1` and `2` are part of a
  # selection object in `from`
  expect_setequal(
    graph_ad_bc$edge_selection$from, c(1, 2))

  # Expect that nodes `4` and `3` are part of a
  # selection object in `to`
  expect_equal(
    graph_ad_bc$edge_selection$to, c(4, 3))

  # Select edges where `width` > 2
  graph_width_gt_2 <-
    graph %>%
    select_edges(
      conditions = width > 2)

  # Expect that node `3` is part of a selection
  # object in 'edges/from'
  expect_equal(graph_width_gt_2$edge_selection$from, 3)

  # Expect that node `1` is part of a selection
  # object in 'edges/to'
  expect_equal(graph_width_gt_2$edge_selection$to, 1)

  # Select nodes where `width` < 3
  graph_width_lt_3 <-
    graph %>%
    select_edges(
      conditions = width < 3)

  # Expect that nodes `1` and `2` are part of a
  # selection object in 'edges/from'
  expect_setequal(
    graph_width_lt_3$edge_selection$from,
    c(1, 2))

  # Expect that nodes `4` and `3` are part of a
  # selection object in 'edges/to'
  expect_setequal(
    graph_width_lt_3$edge_selection$to,
    c(3, 4))

  # Select nodes where `width` == 2
  graph_width_eq_2 <-
    graph %>%
    select_edges(
      conditions = width == 2)

  # Expect that node `2` is part of a selection
  # object in 'edges/from'
  expect_equal(
    graph_width_eq_2$edge_selection$from, 2)

  # Expect that node `3` is part of a selection
  # object in 'edges/to'
  expect_equal(
    graph_width_eq_2$edge_selection$to, 3)

  # Select nodes where `width` != 2
  graph_width_neq_2 <-
    graph %>%
    select_edges(
      conditions = width != 2)

  # Expect that nodes `1` and `3` are part of a
  # selection object in 'edges/from'
  expect_equal(graph_width_neq_2$edge_selection$from, c(1, 3))

  # Expect that nodes `4` and `1` are part of a
  # selection object in 'edges/to'
  expect_equal(graph_width_neq_2$edge_selection$to, c(4, 1))

  # Select nodes where `rel` is `leading_to`
  graph_val_leading_to <-
    graph %>%
    select_edges(
      conditions = rel == "leading_to")

  # Expect that nodes `1`, `2`, and `3` are part of a
  # selection object in 'edges/from'
  expect_equal(
    graph_val_leading_to$edge_selection$from,
    c(1, 2, 3))

  # Expect that nodes `4`, `3`, and `1` are part of a
  # selection object in 'edges/to'
  expect_equal(
    graph_val_leading_to$edge_selection$to,
    c(4, 3, 1))

  # Create a union of selections in a magrittr pipeline
  graph_sel_union_ab_bc <-
    graph %>%
    select_edges(
      from = 1,
        to = 4) %>%
    select_edges(
      from = 2,
        to = 3)

  # Expect that nodes `1` and `2` are part of a
  # selection object in `edges/from`
  expect_equal(
    graph_sel_union_ab_bc$edge_selection$from,
    c(1, 2))

  # Expect that nodes `4` and `3` are part of a
  # selection object in `edges/to`
  expect_equal(
    graph_sel_union_ab_bc$edge_selection$to, c(4, 3))

  # Create a intersection of selections in a
  # magrittr pipeline
  graph_sel_intersect_bc <-
    graph %>%
    select_edges(
      from = c(1, 2),
        to = c(4, 3)) %>%
    select_edges(
      from = c(2, 3),
        to = c(3, 1),
      set_op = "intersect")

  # Expect that node `2` is part of a selection
  # object in `edges/from`
  expect_equal(
    graph_sel_intersect_bc$edge_selection$from, 2)

  # Expect that node `3` is part of a selection
  # object in `edges/to`
  expect_equal(
    graph_sel_intersect_bc$edge_selection$to, 3)

  # Create a selection that is a difference
  # of selections
  graph_sel_edge_difference_ad <-
    graph %>%
    select_edges(
      from = c(1, 2, 3),
        to = c(4, 3, 1)) %>%
    select_edges(
      from = c(2, 3),
        to = c(3, 1),
      set_op = "difference")

  # Expect that node `1` is part of a selection
  # object in `edges/from`
  expect_equal(
    graph_sel_edge_difference_ad$edge_selection$from, 1)

  # Expect that node `4` is part of a selection
  # object in `edges/to`
  expect_equal(
    graph_sel_edge_difference_ad$edge_selection$to, 4)

  # Select edges, specifying only the `from` node
  graph_from_a <-
    select_edges(
      graph,
      from = 1)

  # Expect that only the edge `1`->`4` is selected
  expect_equal(
    graph_from_a$edge_selection$from, 1)

  expect_equal(
    graph_from_a$edge_selection$to, 4)

  # Expect an error if when only specifying the
  # `from` node, that node isn't present in the graph
  expect_error(
    select_edges(
      graph,
      from = 7))

  # Select edges, specifying only the `to` node
  graph_to_c <-
    select_edges(
      graph,
      to = 3)

  # Expect that only the edge `2`->`3` is selected
  expect_equal(
    graph_to_c$edge_selection$from, 2)

  expect_equal(
    graph_to_c$edge_selection$to, 3)

  # Expect an error if when only specifying the
  # `to` node, that node isn't present in the graph
  expect_error(
    select_edges(
      graph,
      to = 7))

  # Expect an error if selecting edges from an
  # empty graph
  expect_error(
    select_edges(
      graph = create_graph(),
      from = 1,
      to = 3))

  # Expect an error if selecting edges from a graph
  # with nodes but no edges
  expect_error(
    select_edges(
      graph = create_graph(
        create_node_df(1))))

  # Expect an error if selecting edges from a
  # graph that doesn't contain edges
  expect_error(
    create_graph() %>%
      add_n_nodes(n = 2) %>%
      select_edges())
})

test_that("deselecting nodes in a graph is possible", {

  # Create a graph and select
  # all nodes in the graph
  graph <-
    create_graph() %>%
    add_path(n = 8) %>%
    select_nodes()

  # Deselect nodes `1` and `2`
  graph <-
    graph %>%
    deselect_nodes(nodes = c(1, 2))

  # Expect nodes `3`-`8` to be
  # in the active selection of nodes
  expect_equal(
    get_selection(graph = graph),
    3:8)

  # Expect that using `deselect_nodes()`
  # on a graph with no active selection
  # of nodes will just return the graph
  # (and the function won't appear in the
  # graph log)
  graph_2 <-
    create_graph() %>%
    add_path(n = 8) %>%
    deselect_nodes(nodes = 1:4)

  expect_equal(
    (graph_2 %>% get_graph_log())$function_used[
      nrow(graph_2 %>% get_graph_log())],
    "add_path")
})

test_that("deselecting edges in a graph is possible", {

  # Create a graph and select
  # all edges in the graph
  graph <-
    create_graph() %>%
    add_path(n = 8) %>%
    select_edges()

  # Deselect edges `1` and `2`
  graph <-
    graph %>%
    deselect_edges(edges = c(1, 2))

  # Expect edges `3`-`7` to be
  # in the active selection of edges
  expect_equal(
    get_selection(graph = graph),
    3:7)

  # Expect that using `deselect_edges()`
  # on a graph with no active selection
  # of eges will just return the graph
  # (and the function won't appear in the
  # graph log)
  graph_2 <-
    create_graph() %>%
    add_path(n = 8) %>%
    deselect_edges(edges = 1:4)

  expect_equal(
    (graph_2 %>% get_graph_log())$function_used[
      nrow(graph_2 %>% get_graph_log())],
    "add_path")
})

test_that("selecting edges via node IDs is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(n = 8)

  # Select all edges associated with nodes with
  # ID `3` and `4`
  graph <-
    select_edges_by_node_id(
      graph = graph,
      nodes = c(3, 4))

  # Expect that certain edges with be available
  # in the selection
  expect_in(
    graph$edge_selection$from,
    c(3, 4, 2))

  expect_in(
    graph$edge_selection$to,
    c(4, 5, 3))
})

test_that("selecting edges via edge IDs is possible", {

  # Create a graph with 5 nodes
  graph <-
    create_graph() %>%
    add_path(n = 5)

  # Create a graph selection by selecting edges
  # associated the edge IDs `1` and `2`
  graph <-
    graph %>%
    select_edges_by_edge_id(edges = 1:2)

  # Expect edges with IDs `1` and `2` in the
  # selection of edges
  expect_equal(
    get_selection(graph),
    c(1, 2))
})

test_that("selecting nodes in a neighborhood is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(n = 12)

  # Create a selection of nodes centered around
  # node `6` and including those nodes a depth of 2
  # edges away
  graph_sel_1_dist_2 <-
    select_nodes_in_neighborhood(
      graph = graph,
      node = 6,
      distance = 2)

  # Expect that specific nodes are part of a selection
  # object in `nodes`
  expect_equal(
    graph_sel_1_dist_2$node_selection$node,
    c(6, 7, 5, 8, 4))

  # Create a selection of nodes centered around node
  # `4` and including those nodes a distance of 4
  # nodes away
  graph_sel_1_sel_4_dist_2 <-
    select_nodes_in_neighborhood(
      graph = graph_sel_1_dist_2,
      node = 4,
      distance = 2)

  # Expect that specific nodes are part of a selection
  # object in `nodes`
  expect_equal(
    graph_sel_1_sel_4_dist_2$node_selection$node,
    c(6, 7, 5, 8, 4, 3, 2))

  # Create a selection of nodes centered around node
  # `1` and including those nodes a distance of
  # 3 nodes away
  graph_sel_1_sel_4_dist_2_sel_A_dist_3 <-
    select_nodes_in_neighborhood(
      graph = graph_sel_1_sel_4_dist_2,
      node = 1,
      distance = 3,
      set_op = "intersect")

  # Expect that specific nodes are part of a selection
  # object in `nodes`
  expect_equal(
    graph_sel_1_sel_4_dist_2_sel_A_dist_3$node_selection$node,
    c(4, 3, 2))

  # Create a selection of nodes centered around node
  # `7` and including those nodes a distance of 2
  # nodes away
  graph_sel_1_sel_4_dist_2_sel_A_dist_3 <-
    select_nodes_in_neighborhood(
      graph = graph_sel_1_sel_4_dist_2,
      node = 7,
      distance = 2,
      set_op = "difference")

  # Expect that specific nodes are part of a selection
  # object in `nodes`
  expect_equal(
    graph_sel_1_sel_4_dist_2_sel_A_dist_3$node_selection$node,
    c(4, 3, 2))
})

test_that("getting a selection is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(n = 12)

  # Select all nodes in graph and get selection
  graph_node_selection_1 <-
    graph %>%
    select_nodes() %>%
    get_selection()

  # Expect that specific nodes are returned
  expect_equal(
    graph_node_selection_1,
    c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12))

  # Select all edges in graph and get selection
  graph_edge_selection_1 <-
    graph %>%
    select_edges() %>%
    get_selection()

  # Expect that specific nodes are returned
  expect_equal(
    graph_edge_selection_1,
    c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11))
})

test_that("inverting a selection is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(n = 12)

  # Select nodes `1` and `2` in the graph
  graph_select_1_2 <-
    graph %>%
    select_nodes(nodes = 1) %>%
    select_nodes(nodes = 2)

  # Invert the selection so that every other node
  # is selected
  graph_select_1_2_inverted <-
    graph_select_1_2 %>%
    invert_selection()

  # Expect that all nodes except `1` and `2` are
  # in the selection
  expect_in(graph_select_1_2_inverted$node_selection$node,
          3:12)

  # Select edges `1`->`2` and `2`->`3` in the graph
  graph_select_edges_1_2__2_3 <-
    graph %>%
    select_edges(
      from = 1,
        to = 2) %>%
    select_edges(
      from = 2,
        to = 3)

  # Invert the selection so that every other edge
  # is selected
  graph_select_edges_1_2__2_3_inverted <-
    graph_select_edges_1_2__2_3 %>%
    invert_selection()

  # Expect that every other edge is now in
  # the selection
  expect_equal(
    graph_select_edges_1_2__2_3_inverted$edge_selection$from,
      c(3, 4, 5, 6, 7, 8, 9, 10, 11))

  expect_equal(
    graph_select_edges_1_2__2_3_inverted$edge_selection$to,
    c(4, 5, 6, 7, 8, 9, 10, 11, 12)
  )

  # Expect an error if inverting selection that
  # doesn't exist
  expect_error(
    invert_selection(graph))
})

test_that("getting/clearing a selection is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
        to = 2)

  # Select all nodes in the graph
  graph_select_all_nodes <-
    graph %>%
    select_nodes()

  # Get the selection and expect both nodes to
  # be present
  expect_contains(get_selection(graph_select_all_nodes), c(1, 2))

  # Clear the selection
  graph_select_all_nodes_cleared <-
    clear_selection(graph_select_all_nodes)

  # Expect no selection to be present in the graph
  expect_equal(nrow(graph_select_all_nodes_cleared$node_selection), 0)

  # Expect an `NA` value returned when getting a
  # selection that is not present
  expect_true(
    is.na(get_selection(graph_select_all_nodes_cleared)))
})

test_that("selecting nodes by node degree is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 35,
      m = 125,
      set_seed = 23)

  # Expect specific nodes when making a selection
  # of nodes which have a total degree (indegree
  # + outdegree) of exactly 9
  expect_equal(
    graph %>%
      select_nodes_by_degree("deg == 9") %>%
      get_selection(),
    c(5, 10, 26, 31))

  # Expect specific nodes when making a selection
  # of nodes which have a total degree greater
  # than or equal to 9
  expect_equal(
    graph %>%
      select_nodes_by_degree("deg >= 9") %>%
      get_selection(),
    c(1, 2, 4, 5, 10, 12, 18, 25, 26, 31))
})

# Graph validation

test_that("Using an invalid graph triggers an error.", {

  # Create an empty graph
  graph <- create_graph()

  # Make the graph invalid
  graph$nodes_df <- NULL
  invalid_graph <- graph

  # Create a valid graph
  valid_graph <- create_graph()

  # Expect errors with all functions that perform
  # graph validation
  expect_snapshot(error = TRUE, {
    add_balanced_tree(invalid_graph)
    add_cycle(invalid_graph)
    add_edge(invalid_graph)
    add_edge_clone(invalid_graph)
    add_edge_df(invalid_graph)
    add_edges_from_table(invalid_graph)
    add_edges_w_string(invalid_graph)
    add_forward_edges_ws(invalid_graph)
    add_full_graph(invalid_graph)
    add_global_graph_attrs(invalid_graph)
    add_gnm_graph(invalid_graph)
    add_gnp_graph(invalid_graph)
    add_graph_action(invalid_graph)
    # graph is the second argument here.
    add_graph_to_graph_series(graph_series = invalid_graph)
    add_grid_2d(invalid_graph)
    add_grid_3d(invalid_graph)
    add_growing_graph(invalid_graph)
    add_islands_graph(invalid_graph)
    add_n_node_clones(invalid_graph)
    add_n_nodes(invalid_graph)
    add_n_nodes_ws(invalid_graph)
    add_node(invalid_graph)
    add_node_clones_ws(invalid_graph)
    add_node_df(invalid_graph)
    add_nodes_from_df_cols(invalid_graph)
    add_nodes_from_table(invalid_graph)
    add_pa_graph(invalid_graph)
    add_path(invalid_graph)
    add_prism(invalid_graph)
    add_smallworld_graph(invalid_graph)
    add_star(invalid_graph)
    add_reverse_edges_ws(invalid_graph)
    clear_selection(invalid_graph)
    colorize_edge_attrs(invalid_graph)
    colorize_node_attrs(invalid_graph)
    copy_edge_attrs(invalid_graph)
    copy_node_attrs(invalid_graph)
    count_asymmetric_node_pairs(invalid_graph)
    count_automorphisms(invalid_graph)
    count_edges(invalid_graph)
    count_loop_edges(invalid_graph)
    count_mutual_node_pairs(invalid_graph)
    count_nodes(invalid_graph)
    count_s_connected_cmpts(invalid_graph)
    count_unconnected_node_pairs(invalid_graph)
    count_unconnected_nodes(invalid_graph)
    count_w_connected_cmpts(invalid_graph)
    create_graph_series(invalid_graph)
    delete_edge(invalid_graph)
    delete_edges_ws(invalid_graph)
    delete_global_graph_attrs(invalid_graph)
    delete_graph_actions(invalid_graph)
    delete_loop_edges_ws(invalid_graph)
    deselect_edges(invalid_graph)
    deselect_nodes(invalid_graph)
    delete_node(invalid_graph)
    delete_nodes_ws(invalid_graph)
    display_metagraph(invalid_graph)
    do_bfs(invalid_graph)
    do_dfs(invalid_graph)
    drop_edge_attrs(invalid_graph)
    drop_node_attrs(invalid_graph)
    export_csv(invalid_graph)
    export_graph(invalid_graph)
    fully_connect_nodes_ws(invalid_graph)
    fully_disconnect_nodes_ws(invalid_graph)
    generate_dot(invalid_graph)
    get_adhesion(invalid_graph)
    get_agg_degree_in(invalid_graph)
    get_agg_degree_out(invalid_graph)
    get_agg_degree_total(invalid_graph)
    get_all_connected_nodes(invalid_graph)
    get_alpha_centrality(invalid_graph)
    get_articulation_points(invalid_graph)
    get_authority_centrality(invalid_graph)
    get_betweenness(invalid_graph)
    get_cache(invalid_graph)
    get_closeness(invalid_graph)
    get_closeness_vitality(invalid_graph)
    get_cmty_edge_btwns(invalid_graph)
    get_cmty_fast_greedy(invalid_graph)
    get_cmty_l_eigenvec(invalid_graph)
    get_cmty_louvain(invalid_graph)
    get_cmty_walktrap(invalid_graph)
    get_common_nbrs(invalid_graph)
    get_coreness(invalid_graph)
    get_degree_distribution(invalid_graph)
    get_degree_histogram(invalid_graph)
    get_degree_in(invalid_graph)
    get_degree_out(invalid_graph)
    get_degree_total(invalid_graph)
    get_dice_similarity(invalid_graph)
    get_eccentricity(invalid_graph)
    get_edge_attrs(invalid_graph)
    get_edge_attrs_ws(invalid_graph)
    get_edge_count_w_multiedge(invalid_graph)
    get_edge_df(invalid_graph)
    get_edge_df_ws(invalid_graph)
    get_edge_info(invalid_graph)
    get_eigen_centrality(invalid_graph)
    get_girth(invalid_graph)
    get_global_graph_attr_info(invalid_graph)
    get_graph_actions(invalid_graph)
    get_graph_log(invalid_graph)
    get_graph_name(invalid_graph)
    get_graph_time(invalid_graph)
    get_jaccard_similarity(invalid_graph)
    get_last_edges_created(invalid_graph)
    get_last_nodes_created(invalid_graph)
    get_leverage_centrality(invalid_graph)
    get_max_eccentricity(invalid_graph)
    get_mean_distance(invalid_graph)
    get_min_cut_between(invalid_graph)
    get_min_eccentricity(invalid_graph)
    get_multiedge_count(invalid_graph)
    get_nbrs(invalid_graph)
    get_node_attrs(invalid_graph)
    get_node_attrs_ws(invalid_graph)
    get_node_df(invalid_graph)
    get_node_df_ws(invalid_graph)
    get_node_ids(invalid_graph)
    get_node_info(invalid_graph)
    get_non_nbrs(invalid_graph)
    get_pagerank(invalid_graph)
    get_paths(invalid_graph)
    get_periphery(invalid_graph)
    get_predecessors(invalid_graph)
    get_radiality(invalid_graph)
    get_reciprocity(invalid_graph)
    get_s_connected_cmpts(invalid_graph)
    get_selection(invalid_graph)
    get_similar_nbrs(invalid_graph)
    get_successors(invalid_graph)
    get_w_connected_cmpts(invalid_graph)
    get_graph_info(invalid_graph)
    invert_selection(invalid_graph)
    is_edge_loop(invalid_graph)
    is_edge_multiple(invalid_graph)
    is_edge_mutual(invalid_graph)
    is_edge_present(invalid_graph)
    is_graph_connected(invalid_graph)
    is_graph_dag(invalid_graph)
    is_graph_directed(invalid_graph)
    is_graph_empty(invalid_graph)
    is_graph_simple(invalid_graph)
    is_graph_undirected(invalid_graph)
    is_graph_weighted(invalid_graph)
    is_node_present(invalid_graph)
    is_property_graph(invalid_graph)
    join_edge_attrs(invalid_graph)
    join_node_attrs(invalid_graph)
    layout_nodes_w_string(invalid_graph)
    mutate_edge_attrs(invalid_graph)
    mutate_edge_attrs_ws(invalid_graph)
    mutate_node_attrs(invalid_graph)
    mutate_node_attrs_ws(invalid_graph)
    nudge_node_positions_ws(invalid_graph)
    recode_edge_attrs(invalid_graph)
    recode_node_attrs(invalid_graph)
    rename_edge_attrs(invalid_graph)
    rename_node_attrs(invalid_graph)
    render_graph(invalid_graph)
    reorder_graph_actions(invalid_graph)
    rescale_edge_attrs(invalid_graph)
    rescale_node_attrs(invalid_graph)
    rev_edge_dir(invalid_graph)
    rev_edge_dir_ws(invalid_graph)
    select_edges(invalid_graph)
    select_edges_by_edge_id(invalid_graph)
    select_edges_by_node_id(invalid_graph)
    select_last_edges_created(invalid_graph)
    select_last_nodes_created(invalid_graph)
    select_nodes(invalid_graph)
    select_nodes_by_degree(invalid_graph)
    select_nodes_by_id(invalid_graph)
    select_nodes_in_neighborhood(invalid_graph)
    set_cache(invalid_graph)
    set_edge_attr_to_display(invalid_graph)
    set_edge_attrs(invalid_graph)
    set_edge_attrs_ws(invalid_graph)
    set_graph_directed(invalid_graph)
    set_graph_name(invalid_graph)
    set_graph_time(invalid_graph)
    set_graph_undirected(invalid_graph)
    set_node_attr_to_display(invalid_graph)
    set_node_attr_w_fcn(invalid_graph)
    set_node_attrs(invalid_graph)
    set_node_attrs_ws(invalid_graph)
    set_node_position(invalid_graph)
    to_igraph(invalid_graph)
    transform_to_complement_graph(invalid_graph)
    transform_to_min_spanning_tree(invalid_graph)
    transform_to_subgraph_ws(invalid_graph)
    trav_both(invalid_graph)
    trav_both_edge(invalid_graph)
    trav_in(invalid_graph)
    trav_in_edge(invalid_graph)
    trav_in_node(invalid_graph)
    trav_in_until(invalid_graph)
    trav_reverse_edge(invalid_graph)
    trav_out(invalid_graph)
    trav_out_edge(invalid_graph)
    trav_out_node(invalid_graph)
    trav_out_until(invalid_graph)
    trav_reverse_edge(invalid_graph)
    trigger_graph_actions(invalid_graph)
  })




  expect_error(combine_graphs(x = invalid_graph, y = valid_graph))
  expect_error(combine_graphs(x = valid_graph, y = invalid_graph))
})

test_that("Using an empty graph with trigger an error", {

  # Create an empty graph
  empty_graph <- create_graph()

  # Expect errors with all functions that check
  # for an empty graph
  expect_snapshot(error = TRUE, {
    add_edge(empty_graph)
    add_edge_df(empty_graph)
  })

  expect_error(add_edges_from_table(empty_graph))
  expect_error(add_edges_w_string(empty_graph))
  expect_error(add_n_node_clones(empty_graph))
  expect_error(add_n_nodes_ws(empty_graph))
  expect_error(add_node_clones_ws(empty_graph))
  expect_error(delete_edge(empty_graph))
  expect_error(delete_edges_ws(empty_graph))
  expect_error(delete_loop_edges_ws(empty_graph))
  expect_error(delete_node(empty_graph))
  expect_error(delete_nodes_ws(empty_graph))
  expect_error(do_bfs(empty_graph))
  expect_error(do_dfs(empty_graph))
  expect_error(fully_connect_nodes_ws(empty_graph))
  expect_error(fully_disconnect_nodes_ws(empty_graph))
  expect_error(get_degree_distribution(empty_graph))
  expect_error(get_degree_histogram(empty_graph))
  expect_error(get_eccentricity(empty_graph))
  expect_error(get_last_nodes_created(empty_graph))
  expect_error(mutate_node_attrs_ws(empty_graph))
  expect_error(nudge_node_positions_ws(empty_graph))
  expect_error(recode_node_attrs(empty_graph))
  expect_error(rename_node_attrs(empty_graph))
  expect_error(rescale_node_attrs(empty_graph))
  expect_error(select_edges(empty_graph))
  expect_error(select_last_nodes_created(empty_graph))
  expect_error(select_nodes(empty_graph))
  expect_error(select_nodes_by_degree(empty_graph))
  expect_error(select_nodes_by_id(empty_graph))
  expect_error(select_nodes_in_neighborhood(empty_graph))
  expect_error(set_node_attr_to_display(empty_graph))
  expect_error(set_node_attrs(empty_graph))
  expect_error(set_node_attrs_ws(empty_graph))
  expect_error(set_node_position(empty_graph))
  expect_error(transform_to_complement_graph(empty_graph))
  expect_error(trav_both(empty_graph))
  expect_error(trav_both_edge(empty_graph))
  expect_error(trav_in(empty_graph))
  expect_error(trav_in_edge(empty_graph))
  expect_error(trav_in_node(empty_graph))
  expect_error(trav_in_until(empty_graph))
  expect_error(trav_out(empty_graph))
  expect_error(trav_out_edge(empty_graph))
  expect_error(trav_out_node(empty_graph))
  expect_error(trav_out_until(empty_graph))
})

test_that("Using a graph with no edges with trigger an error", {

  # Create a graph with nodes but no edges
  graph_no_edges <-
    create_graph() %>%
    add_n_nodes(5)

  # Expect errors with all functions that check
  # for an empty graph
  expect_snapshot(error = TRUE, {
    add_edge_clone(graph_no_edges)
    add_forward_edges_ws(graph_no_edges)
    add_reverse_edges_ws(graph_no_edges)
    delete_edge(graph_no_edges)
    delete_edges_ws(graph_no_edges)
    get_edge_count_w_multiedge(graph_no_edges)
    get_last_edges_created(graph_no_edges)
    get_multiedge_count(graph_no_edges)
    is_edge_loop(graph_no_edges)
    is_edge_multiple(graph_no_edges)
    is_edge_mutual(graph_no_edges)
    mutate_edge_attrs_ws(graph_no_edges)
    recode_edge_attrs(graph_no_edges)
    rename_edge_attrs(graph_no_edges)
    rescale_edge_attrs(graph_no_edges)
    rev_edge_dir(graph_no_edges)
    rev_edge_dir_ws(graph_no_edges)
    select_edges(graph_no_edges)
    select_edges_by_edge_id(graph_no_edges)
    select_edges_by_node_id(graph_no_edges)
    select_last_edges_created(graph_no_edges)
    set_edge_attr_to_display(graph_no_edges)
    set_edge_attrs_ws(graph_no_edges)

    trav_both(graph_no_edges)
    trav_both_edge(graph_no_edges)
    trav_in(graph_no_edges)
    trav_in_edge(graph_no_edges)
    trav_in_node(graph_no_edges)
    trav_in_until(graph_no_edges)
    trav_out(graph_no_edges)
    trav_out_edge(graph_no_edges)
    trav_out_node(graph_no_edges)
    trav_out_until(graph_no_edges)
    trav_reverse_edge(graph_no_edges)
  })

})

#' Get a minimum spanning tree subgraph
#'
#' @description
#'
#' Get a minimum spanning tree subgraph for a connected graph of class
#' `dgr_graph`.
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Obtain Jaccard similarity
#' # values for each pair of
#' # nodes as a square matrix
#' j_sim_matrix <-
#'   graph |>
#'     get_jaccard_similarity()
#'
#' # Create a weighted, undirected
#' # graph from the resultant matrix
#' # (effectively treating that
#' # matrix as an adjacency matrix)
#' graph <-
#'   j_sim_matrix |>
#'   from_adj_matrix(weighted = TRUE)
#'
#' # The graph in this case is a fully connected graph
#' # with loops, where jaccard similarity values are
#' # assigned as edge weights (edge attribute `weight`);
#' # The minimum spanning tree for this graph is the
#' # connected subgraph where the edges retained have
#' # the lowest similarity values possible
#' min_spanning_tree_graph <-
#'   graph |>
#'   transform_to_min_spanning_tree() |>
#'   copy_edge_attrs(
#'     edge_attr_from = weight,
#'     edge_attr_to = label) |>
#'   set_edge_attrs(
#'     edge_attr = fontname,
#'     values = "Helvetica") |>
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "gray85") |>
#'   rescale_edge_attrs(
#'     edge_attr_from = weight,
#'     to_lower_bound = 0.5,
#'     to_upper_bound = 4.0,
#'       edge_attr_to = penwidth)
#'
#' @export
transform_to_min_spanning_tree <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Transform the graph to an igraph object
  igraph <- to_igraph(graph)

  # Get the minimum spanning tree
  igraph_mst <- igraph::mst(igraph)

  # Generate the graph object from an igraph graph
  from_igraph(igraph_mst)
}

# Obtaining the minimum spanning tree of a graph

test_that("the MST algorithm is functional", {

  # Create a random graph and obtain Jaccard
  # similarity values for each pair of nodes
  # as a square matrix
  j_sim_matrix <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23) %>%
    get_jaccard_similarity()

  # Create a weighted, undirected graph from the
  # resultant matrix (effectively treating that
  # matrix as an adjacency matrix)
  graph <-
    j_sim_matrix %>%
    from_adj_matrix(weighted = TRUE)

  # The graph in this case is a fully connected graph
  # with loops, where jaccard similarity values are
  # assigned as edge weights (edge attribute `weight`);
  # The minimum spanning tree for this graph is the
  # connected subgraph where the edges retained have
  # the lowest similarity values possible
  min_spanning_tree_graph <-
    graph %>%
    transform_to_min_spanning_tree()

  # Expect no loops in `min_spanning_tree_graph`
  expect_true(
    all(get_node_info(min_spanning_tree_graph)$loops == 0))

  # Expect that all nodes are connected
  expect_true(
    is_graph_connected(min_spanning_tree_graph))
})

#' Set the node attribute values to be rendered
#'
#' @description
#'
#' Set a node attribute type to display as node text when calling the
#' [render_graph()] function. This allows for display of different types of node
#' attribute values on a per-node basis. Without setting the `display`
#' attribute, rendering a graph will default to printing text from the `label`
#' attribute on nodes. Setting the `display` node attribute with this function
#' for the first time (i.e., the `display` column doesn't exist in the graph's
#' internal node data frame) will insert the `attr` value for all nodes
#' specified in `nodes` and a default value (`default`) for all remaining nodes.
#'
#' @inheritParams render_graph
#' @param attr The name of the attribute from which label text for the node will
#'   be obtained. If set to `NULL`, then `NA` values will be assigned to the
#'   `display` column for the chosen nodes.
#' @param nodes A length vector containing one or several node ID values (as
#'   integers) for which node attributes are set for display in the rendered
#'   graph. If `NULL`, all nodes from the graph are assigned the `display` value
#'   given as `attr`.
#' @param default The name of an attribute to set for all other graph nodes not
#'   included in `nodes`. This value only gets used if the `display` node
#'   attribute is not in the graph's internal node data frame.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 4,
#'     m = 4,
#'     set_seed = 23) %>%
#'   set_node_attrs(
#'     node_attr = value,
#'     values = c(2.5, 8.2, 4.2, 2.4))
#'
#' # For node ID values of `1`,
#' # `2`, and `3`, choose to display
#' # the node `value` attribute (for
#' # the other nodes, display nothing)
#' graph <-
#'   graph %>%
#'   set_node_attr_to_display(
#'     nodes = 1:3,
#'     attr = value,
#'     default = NA)
#'
#' # Show the graph's node data frame; the
#' # `display` node attribute will show for
#' # each row, which node attribute value to
#' # display when the graph is rendered
#' graph %>% get_node_df()
#'
#' # This function can be called multiple
#' # times on a graph; after the first time
#' # (i.e., creation of the `display`
#' # attribute), the `default` value won't
#' # be used
#' graph %>%
#'   set_node_attr_to_display(
#'     nodes = 4,
#'     attr = label) %>%
#'   set_node_attr_to_display(
#'     nodes = c(1, 3),
#'     attr = id) %>%
#'   get_node_df()
#'
#' @family node creation and removal
#'
#' @export
set_node_attr_to_display <- function(
    graph,
    attr = NULL,
    nodes = NULL,
    default = "label"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get the requested `attr`
  attr <-
    rlang::enquo(attr) %>% rlang::get_expr() %>% as.character()

  # If nothing provided for `attr`, set
  # it as NULL
  if (length(attr) == 0) {
    attr <- NULL
  }

  # Get the graph's node data frame as an object
  ndf <- graph$nodes_df

  # If `nodes` is NULL, assume that all nodes to
  # be assigned a `display` value
  if (is.null(nodes)) {
    nodes <- get_node_ids(graph)
  }

  # Stop function if any of the node ID values
  # provided in `nodes` do not exist in the graph
  if (!any(nodes %in% ndf$id)) {

    cli::cli_abort(
      "One or more node ID values in `nodes` are not present in the graph.")
  }

  # Stop function if the node attribute supplied as
  # `attr` does not exist in the ndf
  if (!is.null(attr) && !(attr %in% colnames(ndf))) {

    cli::cli_abort(
      "The node attribute given in `attr` is not in the graph's ndf.")
  }

  # If the `display` node attribute doesn't exist,
  # create that column and fill with the default value
  if (!("display" %in% colnames(ndf))) {
    ndf$display <- as.character(default)
  }

  # Create a tibble with the node ID values and the
  # requested node attribute to display
  if (is.null(attr)) {
    attr_to_display <-
      dplyr::tibble(
        id = as.integer(nodes),
        display = "is_na")

  } else {
    attr_to_display <-
      dplyr::tibble(
        id = as.integer(nodes),
        display = as.character(attr))

  }

  # Join the `attr_to_display` table with the `ndf`
  ndf <-
    ndf %>%
    dplyr::left_join(attr_to_display, by = "id")

  # Get the column numbers for the `.x`
  # and `.y` columns
  x_col <- grep("\\.x$", colnames(ndf))
  y_col <- grep("\\.y$", colnames(ndf))

  # Coalesce the 2 generated columns and create a
  # single-column data frame
  if (is.null(attr)) {
    display_col <-
      dplyr::coalesce(ndf[, y_col], ndf[, x_col])

    display_col <- dplyr::na_if(display_col, "is_na") %>%
      as.data.frame(stringsAsFactors = FALSE)

  } else {
    display_col <-
      dplyr::coalesce(ndf[, y_col], ndf[, x_col]) %>%
      as.data.frame(stringsAsFactors = FALSE)
  }

  # Rename the column
  colnames(display_col)[1] <- "display"

  # Remove column numbers that end with ".x" or ".y"
  ndf <- ndf[-grep("\\.x$", colnames(ndf))]
  ndf <- ndf[-grep("\\.y$", colnames(ndf))]

  # Bind the `display_col` df to the `ndf` df and
  # modify the ordering of the columns
  ndf <-
    dplyr::bind_cols(ndf, display_col) %>%
    dplyr::relocate("id", "type", "label", "display")

  # Replace the graph's node data frame with `ndf`
  graph$nodes_df <- ndf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

# Copying attribute values as part of a traversal

test_that("copying values with `trav_out_edge()` works", {

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(data = c(1.0, 2.0, 3.0)),
      edge_data = edge_data(data = c(1.5, 2.5))
    )

  # Select node ID `1`, traverse to
  # outward edges and copy the `data`
  # data from nodes to edges traversed to
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(copy_attrs_from = data)

  # Expect that there are 5 columns in the edf
  expect_equal(ncol(graph$edges_df), 5)

  # Expect that the `data` column is
  # the last column in edf
  expect_equal(colnames(graph$edges_df)[length(colnames(graph$edges_df))], "data")

  # Expect certain values in the
  # `value` column of the edf
  expect_equal(graph$edges_df$data, c(1.0, 2.5))

  #
  # Test where there are no common column names
  #

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(node_value = c(1.0, 2.0, 3.0)),
      edge_data = edge_data(edge_value = c(1.5, 2.5))
    )

  # Select node ID `1`, traverse to
  # outward edges, copy value from
  # ndf to edf
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(copy_attrs_from = node_value)

  # Expect that there are 6 columns in the edf
  expect_equal(ncol(graph$edges_df), 6)

  # Expect that the `edge_value`
  # column in the last column in
  # the edge data frame
  expect_equal(colnames(graph$edges_df)[length(colnames(graph$edges_df))], "edge_value")

  # Expect certain values in the `node_value` column of the
  # edge data frame
  expect_equal(graph$edges_df$node_value, c(1, NA))
})

test_that("copying values with `trav_in_edge()` works", {

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(data = c(15, 20, 30)),
      edge_data = edge_data(data = c(1.5, 2.5))
    ) %>%
    set_node_attr_to_display(attr = "data") %>%
    set_edge_attr_to_display(attr = "data")

  # Select node ID `3`, traverse to
  # inward edges and copy the `data`
  # data from nodes to edges traversed to
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_in_edge(copy_attrs_from = data)

  # Expect that there are 6 columns in the edf
  expect_equal(ncol(graph$edges_df), 6)

  # Expect that the `data` column is
  # the last column in edf
  expect_equal(colnames(graph$edges_df)[length(colnames(graph$edges_df))], "data")

  # Expect certain values in the
  # `value` column of the edf
  expect_equal(graph$edges_df$data, c(1.5, 30))

  #
  # Test where there are no common column names
  #

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(node_value = c(15, 20, 30)),
      edge_data = edge_data(edge_value = c(1.5, 2.5))
    ) %>%
    set_node_attr_to_display(attr = "node_value") %>%
    set_edge_attr_to_display(attr = "edge_value")

  # Select node ID `3`, traverse to
  # outward edges, copy value from
  # ndf to edf
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_in_edge(copy_attrs_from = node_value)

  # Expect that there are 7 columns in the edf
  expect_equal(ncol(graph$edges_df), 7)

  # Expect that the `edge_value`
  # column in the last column in
  # the edge data frame
  expect_equal(colnames(graph$edges_df)[length(colnames(graph$edges_df))], "edge_value")

  # Expect certain values in the `node_value` column of the
  # edge data frame
  expect_equal(graph$edges_df$node_value, c(NA, 30))
})

test_that("copying values with `trav_both_edge()` works", {

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(data = c(1.0, 2.0, 3.0)),
      edge_data = edge_data(data = c(1.5, 2.5))
    )

  # Select node ID `2`, traverse to
  # inward/outward edges and copy the `data`
  # data from nodes to edges traversed to
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 2) %>%
    trav_both_edge(copy_attrs_from = data)

  # Expect that there are 5 columns in the edf
  expect_equal(ncol(graph$edges_df), 5)

  # Expect that the `data` column is
  # the last column in edf
  expect_equal(colnames(graph$edges_df)[length(colnames(graph$edges_df))], "data")

  # Expect certain values in the
  # `value` column of the edf
  expect_equal(graph$edges_df$data, c(3.5, 4.5))

  #
  # Test where there are no common column names
  #

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(node_value = c(1.0, 2.0, 3.0)),
      edge_data = edge_data(edge_value = c(1.5, 2.5))
    )

  # Select node ID `2`, traverse to
  # inward/outward edges, copy value from
  # ndf to edf
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 2) %>%
    trav_both_edge(copy_attrs_from = node_value)

  # Expect that there are 6 columns in the edf
  expect_equal(ncol(graph$edges_df), 6)

  # Expect that the `edge_value`
  # column in the last column in
  # the edge data frame
  expect_equal(colnames(graph$edges_df)[length(colnames(graph$edges_df))], "edge_value")

  # Expect certain values in the `node_value` column of the
  # edge data frame
  expect_equal(graph$edges_df$node_value, c(2, 2))
})

test_that("copying values with `trav_both()` works", {

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(data = c(1.0, 2.0, 3.0)),
      edge_data = edge_data(data = c(1.5, 2.5))
    )

  # Select node ID `2`, traverse to
  # inward/outward nodes and copy the `data`
  # data from nodes to edges traversed to
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 2) %>%
    trav_both(copy_attrs_from = data)

  # Expect certain values in the `node_value` column of the
  # edge data frame
  expect_equal(graph$nodes_df$data, c(2, 2, 2))
})

test_that("copying values with `trav_in()` works", {

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(data = c(1.0, 2.0, 3.0)),
      edge_data = edge_data(data = c(1.5, 2.5))
    )

  # Select node ID `3`, traverse to
  # inward/outward edges and copy the `data`
  # data from nodes to edges traversed to
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_in(copy_attrs_from = data)

  # Expect that there are 4 columns in the edf
  expect_equal(ncol(graph$nodes_df), 4)

  # Expect that the `data` column is
  # the last column in ndf
  expect_equal(colnames(graph$nodes_df)[length(colnames(graph$nodes_df))], "data")

  # Expect certain values in the
  # `value` column of the edf
  expect_equal(graph$nodes_df$data, c(1, 3, 3))
})

test_that("copying values with `trav_out()` works", {

  # Create a graph with a path
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      node_data = node_data(data = c(1.0, 2.0, 3.0)),
      edge_data = edge_data(data = c(1.5, 2.5))
    )

  # Select node ID `1`, traverse to
  # outward nodes and copy the `data`
  # data to nodes traversed to
  graph <-
    graph %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out(copy_attrs_from = data)

  # Expect that there are 4 columns in the ndf
  expect_equal(ncol(graph$nodes_df), 4)

  # Expect that the `data` column is
  # the last column in ndf
  expect_equal(colnames(graph$nodes_df)[length(colnames(graph$nodes_df))], "data")

  # Expect certain values in the
  # `value` column of the edf
  expect_equal(graph$nodes_df$data, c(1, 1, 3))
})

test_that("copying values with `trav_out_node()` works", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    select_nodes() %>%
    add_n_nodes_ws(n = 2, direction = "from") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 2) %>%
    set_node_attrs_ws(node_attr = value, value = 8) %>%
    clear_selection() %>%
    select_edges_by_edge_id(edges = 1) %>%
    set_edge_attrs_ws(edge_attr = value, value = 5) %>%
    clear_selection() %>%
    select_edges_by_edge_id(edges = 2) %>%
    set_edge_attrs_ws(edge_attr = value, value = 5) %>%
    clear_selection() %>%
    select_edges()

  # Expect that specific values will be copied
  # from selected nodes and passed to adjacent nodes
  expect_equal(
    graph %>%
      trav_out_node(copy_attrs_from = value, agg = "sum") %>%
      get_node_attrs(node_attr = value) %>%
      as.numeric(),
    c(10, 8, NA)
  )
})

test_that("copying values with `trav_in_node()` works", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    select_nodes() %>%
    add_n_nodes_ws(n = 2, direction = "to") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 2) %>%
    set_node_attrs_ws(node_attr = value, value = 8) %>%
    clear_selection() %>%
    select_edges_by_edge_id(edges = 1) %>%
    set_edge_attrs_ws(edge_attr = value, value = 5) %>%
    clear_selection() %>%
    select_edges_by_edge_id(edges = 2) %>%
    set_edge_attrs_ws(edge_attr = value, value = 5) %>%
    clear_selection() %>%
    select_edges()

  # Expect that specific values will be copied
  # from selected nodes and passed to adjacent nodes
  expect_equal(
    graph %>%
      trav_in_node(copy_attrs_from = value, agg = "sum") %>%
      get_node_attrs(node_attr = value) %>%
      as.numeric(),
    c(10, 8, NA)
  )
})

test_that("node attributes can be chosen for display", {

  # Create a graph with a path
  # and a `value` node attribute
  graph <-
    create_graph() %>%
    add_path(
      n = 5,
      node_data = node_data(
        value = 1:5))

  # For node ID values of `1`
  # to `3`, choose to display the
  # node `value` attribute (for
  # the other nodes, display nothing)
  graph <-
    graph %>%
    set_node_attr_to_display(
      nodes = 1:3,
      attr = value,
      default = NA)

  # Expect `value` for `display`
  # node attribute in first 3 rows,
  # then, 2 NA values
  expect_equal(
    graph$nodes_df$display,
    c("value", "value", "value", NA, NA))

  # Call function several times
  # to set the `display` node
  # attribute
  graph <-
    graph %>%
    set_node_attr_to_display(
      nodes = 4,
      attr = label) %>%
    set_node_attr_to_display(
      nodes = c(1, 5),
      attr = id)

  # Expect set values for the
  # `display` node attribute
  expect_equal(
    graph$nodes_df$display,
    c("id", "value", "value", "label", "id"))
})

test_that("edge attributes can be chosen for display", {

  # Create a graph with a path
  # and a `value` edge attribute
  graph <-
    create_graph() %>%
    add_path(
      n = 5,
      edge_data = edge_data(
        value = 1:4))

  # For edge ID values of `1` to
  # `3`, choose to display the
  # edge `value` attribute (for
  # the other nodes, display nothing)
  graph <-
    graph %>%
    set_edge_attr_to_display(
      edges = 1:3,
      attr = value,
      default = NA)

  # Expect `value` for `display` edge
  # attribute in the first 3 rows,
  # then, an NA value
  expect_equal(
    graph$edges_df$display,
    c("value", "value", "value", NA))

  # Call function again times to
  # set the `display` edge attribute
  graph <-
    graph %>%
    set_edge_attr_to_display(
      edges = c(1, 3),
      attr = id)

  # Expect set values for the
  # `display` edge attribute
  expect_equal(
    graph$edges_df$display,
    c("id", "value", "id", NA))
})

test_that("No warning is produced.", {
  # this seems not to pass all the time.
  skip_on_cran()
  # before changing render_graph circa line 177, there was a warn
  withr::local_seed(10)
  expect_no_warning(create_graph() %>%
                      add_balanced_tree(
                        k = 2, h = 3) %>%
                      render_graph(layout = "tree"))

  # example_graph from README
  example_graph <-
    create_graph() %>%
    add_pa_graph(
      n = 50, m = 1,
      set_seed = 23
    ) %>%
    add_gnp_graph(
      n = 50, p = 1 / 100,
      set_seed = 23
    ) %>%
    join_node_attrs(df = get_betweenness(.)) %>%
    join_node_attrs(df = get_degree_total(.)) %>%
    colorize_node_attrs(
      node_attr_from = total_degree,
      node_attr_to = fillcolor,
      palette = "Greens",
      alpha = 90
    ) %>%
    rescale_node_attrs(
      node_attr_from = betweenness,
      to_lower_bound = 0.5,
      to_upper_bound = 1.0,
      node_attr_to = height
    ) %>%
    select_nodes_by_id(nodes = get_articulation_points(.)) %>%
    set_node_attrs_ws(node_attr = peripheries, value = 2) %>%
    set_node_attrs_ws(node_attr = penwidth, value = 3) %>%
    clear_selection() %>%
    set_node_attr_to_display(attr = NULL)

  expect_no_warning(render_graph(example_graph, layout = "nicely"))
  expect_no_warning(render_graph(example_graph, layout = "circle"))
  expect_no_warning(render_graph(example_graph, layout = "kk"))
  expect_no_warning(render_graph(example_graph, layout = "tree"))

  expect_no_warning(render_graph(example_graph, layout = "fr"))

})

test_that("render_graph errors on incorrect graph and layout input", {
  g <- create_graph() %>%
    add_balanced_tree(
      k = 2, h = 3)

  expect_error(
    render_graph(g, layout = "xx")
  )
  expect_error(
    render_graph(g, output = "xx")
  )
})

# Displaying the graph's metagraph")

test_that("the display of the metagraph works", {

  # Create a randomized property
  # graph with 1000 nodes and 1350 edges
  property_graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 100,
      m = 135,
      set_seed = 23) %>%
    select_nodes_by_degree(
      expressions = "deg >= 3") %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "a") %>%
    clear_selection() %>%
    select_nodes_by_degree(
      expressions = "deg < 3") %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "b") %>%
    clear_selection() %>%
    select_nodes_by_degree(
      expressions = "deg == 0") %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "c") %>%
    set_node_attr_to_display(
      attr = type) %>%
    select_edges_by_node_id(
      nodes =
        get_node_ids(.) %>%
        sample(
          size = 0.15 * length(.) %>%
            floor())) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "r_1") %>%
    invert_selection() %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "r_2") %>%
    clear_selection() %>%
    copy_edge_attrs(
      edge_attr_from = rel,
      edge_attr_to = label)

  # Get the metagraph object
  metagraph_object <- display_metagraph(property_graph)

  # Expect that the object has the classes of `grViz` and `htmlwidget`
  expect_s3_class(
    metagraph_object, c("grViz", "htmlwidget"))
})

#' Get graph name
#'
#' @description
#'
#' Get the name of a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#'
#' @return A single-length character vector with the assigned graph name. If a
#'   graph name has not been set, NA is returned.
#'
#' @examples
#' # Create an empty graph
#' graph <- create_graph()
#'
#' # Provide the new graph with a name
#' graph <-
#'   set_graph_name(
#'     graph,
#'     name = "the_name")
#'
#' # Get the graph's name
#' graph |> get_graph_name()
#'
#' @export
get_graph_name <- function(graph) {
  check_graph_valid(graph)
  graph$graph_info$graph_name
}

test_that("Setting a graph name can be done", {

  # Create an empty graph
  graph <- create_graph()

  # Set a graph name
  graph_name <-
    graph %>%
    set_graph_name(name = "test_that_name")

  # Expect that the name was set
  expect_equal(
    graph_name$graph_info$graph_name, "test_that_name")

  # Add 2 nodes and an edge and then
  # select a node
  graph_1 <-
    graph %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    select_nodes_by_id(nodes = 1)

  # Set a graph name
  graph_name_1 <-
    set_graph_name(
      graph = graph_1,
      name = "test_that_name_again")

  # Expect that the name was set
  expect_equal(
    graph_name_1$graph_info$graph_name, "test_that_name_again")
})

test_that("Setting a time for the graph can be done", {

  # Create an empty graph
  graph <- create_graph()

  # Provide the new graph with a timestamp (`tz` not supplied so
  # `GMT` is used as the time zone)
  graph_1 <-
    set_graph_time(
      graph = graph,
      time = "2015-10-25 15:23:00")

  # Expect that the time value was passed in properly
  expect_equal(
    graph_1$graph_info$graph_time,
    as.POSIXct("2015-10-25 15:23:00"),
    ignore_attr = TRUE
    )

  # Expect that the time zone is set to `GMT`
  expect_equal(
    graph_1$graph_info$graph_tz, "GMT")

  # Update tz when a timestamp is already present
  graph_2 <-
    set_graph_time(
      graph = graph_1,
      tz = "America/Los_Angeles")

  # Expect that the time zone has been changed
  expect_equal(
    graph_2$graph_info$graph_tz, "America/Los_Angeles")

  # Expect an error when setting a time zone that is
  # not in `OlsonNames()`
  expect_error(
    set_graph_time(
      graph = graph_2,
      tz = "Moon/Moon"))

  # Create a graph with a node and a node selection
  graph_selection <-
    create_graph() %>%
    add_node() %>%
    select_nodes() %>%
    set_graph_time("2015-10-25 15:23:00")
})

test_that("Getting the graph name is possible", {

  # Create a new graph and set a graph name
  graph <-
    create_graph() %>%
    set_graph_name(name = "test_graph")

  # Verify that the graph name returned is a
  # character vector
  expect_type(
    get_graph_name(graph), "character")

  # Expect that the returned vector has a length of 1
  expect_length(
    get_graph_name(graph), 1)

  # Expect that the graph name that was set is returned
  expect_equal(
    get_graph_name(graph), "test_graph")
})

test_that("Getting the graph time is possible", {

  # Create a graph with a time
  graph <-
    create_graph() %>%
    set_graph_time(
      time = "2015-10-25 15:23:00")

  # Expect a graph time as POSIXct
  expect_s3_class(
    get_graph_time(graph), "POSIXct")

  # Expect that the returned vector has a length of 1
  expect_length(get_graph_time(graph), 1)

  # Create a graph without a time set
  graph <- create_graph()

  # Verify that a POSIXct time is returned
  expect_s3_class(
    get_graph_time(graph), "POSIXct")
})

test_that("Getting global graph attrs is possible", {

  # Create an empty graph with no global graph
  # parameters
  graph <- create_graph(attr_theme = NULL)

  # Expect an NA value if getting global graph
  # attributes where there are none set
  expect_equal(
    graph %>%
      get_global_graph_attr_info() %>%
      nrow(), 0)

  # Add 3 global graph attrs
  graph <-
    graph %>%
    add_global_graph_attrs(
      attr = c("overlap", "color", "penwidth"),
      value = c("true", "red", "5"),
      attr_type = c("graph", "node", "edge"))

  # Get a table with the attributes
  # using `get_global_graph_attr_info()`
  global_graph_attrs <-
    graph %>%
    get_global_graph_attr_info()

  # Expect that the returned table is
  # equivalent to the table stored in
  # the graph object
  expect_equal(
    graph$global_attrs %>% dplyr::as_tibble(),
    global_graph_attrs)
})

test_that("Adding global graph attrs is possible", {

  # Create an empty graph with the default global
  # graph attributes
  graph <- create_graph()

  # Add 2 global graph attrs
  graph_add_2 <-
    graph %>%
    add_global_graph_attrs(
      attr = c("overlap", "penwidth"),
      value = c("true", "5"),
      attr_type = c("graph", "edge"))

  # Expect that the new graph object has 2 more
  # global graph attributes than the original
  expect_equal(
    nrow(graph_add_2$global_attrs) -
      nrow(graph$global_attrs), 2)

  # Expect that the new graph attributes are
  # the last 2 rows in the `global_attrs` df
  expect_equal(
    tail(graph_add_2$global_attrs, 2)[, 1],
    c("overlap", "penwidth"))

  expect_equal(
    tail(graph_add_2$global_attrs, 2)[, 2],
    c("true", "5"))

  expect_equal(
    tail(graph_add_2$global_attrs, 2)[, 3],
    c("graph", "edge"))

  # Add 1 global graph attribute
  graph_add_1 <-
    graph %>%
    add_global_graph_attrs(
      attr = "overlap",
      value = TRUE,
      attr_type = "graph")

  # Expect that the new graph object has 1 more
  # global graph attribute than the original
  expect_equal(
    nrow(graph_add_1$global_attrs) -
      nrow(graph$global_attrs), 1)

  # Expect that the new graph attribute is
  # in the last row in the `global_attrs` df
  expect_equal(
    tail(graph_add_1$global_attrs, 1)[, 1],
    "overlap")

  expect_equal(
    tail(graph_add_1$global_attrs, 1)[, 2],
    "true")

  expect_equal(
    tail(graph_add_1$global_attrs, 1)[, 3],
    "graph")
})

test_that("Deleting global graph attrs is possible", {

  # Create an empty graph with the default global
  # graph attributes
  graph <- create_graph()

  # Remove a single global graph attr
  graph_del_1 <-
    graph %>%
    delete_global_graph_attrs(
      attr = "layout",
      attr_type = "graph")

  # Expect that the new graph object has 1 less
  # global graph attribute than the original
  expect_equal(
    nrow(graph$global_attrs) -
      nrow(graph_del_1$global_attrs), 1)

  # Remove 2 global graph attributes
  graph_del_2 <-
    graph %>%
    delete_global_graph_attrs(
      attr = c("layout", "outputorder"),
      attr_type = c("graph", "graph"))

  # Expect that the new graph object has 2 less
  # global graph attributes than the original
  expect_equal(
    nrow(graph$global_attrs) -
      nrow(graph_del_2$global_attrs), 2)

  # Expect an error if deleting with an invalid
  # `attr_type` (using `nodes` instead of `node`)
  expect_error(
    graph %>%
      delete_global_graph_attrs(
        attr = "layout",
        attr_type = "nodes"))
})

#' Traverse inward node-by-node until stopping conditions are met
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, move along inward edges from one or
#' more nodes present in a selection to other connected nodes, replacing the
#' current nodes in the selection with those nodes traversed to until reaching
#' nodes that satisfy one or more conditions.
#'
#' This traversal function makes use of an active selection of nodes. After the
#' traversal, depending on the traversal conditions, there will either be a
#' selection of nodes or no selection at all.
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param conditions An option to use a stopping condition for the traversal. If
#'   the condition is met during the traversal (i.e., the node(s) traversed to
#'   match the condition), then those traversals will terminate at those nodes.
#'   Otherwise, traversals with continue and terminate when the number of steps
#'   provided in `max_steps` is reached.
#' @param max_steps The maximum number of `trav_in()` steps (i.e., node-to-node
#'   traversals in the inward direction) to allow before stopping.
#' @param exclude_unmatched If `TRUE` (the default value) then any nodes not
#'   satisfying the conditions provided in `conditions` that are in the ending
#'   selection are excluded.
#' @param add_to_selection If `TRUE` then every node traversed will be part of
#'   the final selection of nodes. If `FALSE` (the default value) then only the
#'   nodes finally traversed to will be part of the final node selection.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a path graph and add
#' # values of 1 to 10 across the
#' # nodes from beginning to end;
#' # select the last path node
#' graph <-
#'   create_graph() %>%
#'   add_path(
#'     n = 10,
#'     node_data = node_data(
#'       value = 1:10)) %>%
#'   select_nodes_by_id(
#'     nodes = 10)
#'
#' # Traverse inward, node-by-node
#' # until stopping at a node where
#' # the `value` attribute is 1
#' graph <-
#'   graph %>%
#'   trav_in_until(
#'     conditions =
#'       value == 1)
#'
#' # Get the graph's node selection
#' graph %>% get_selection()
#'
#' # Create two cycles in a graph and
#' # add values of 1 to 6 to the
#' # first cycle, and values 7 to
#' # 12 in the second; select nodes
#' # `6` and `12`
#' graph <-
#'   create_graph() %>%
#'   add_cycle(
#'     n = 6,
#'     node_data = node_data(
#'       value = 1:6)) %>%
#'   add_cycle(
#'     n = 6,
#'     node_data = node_data(
#'       value = 7:12)) %>%
#'   select_nodes_by_id(
#'     nodes = c(6, 12))
#'
#' # Traverse inward, node-by-node
#' # from `6` and `12` until stopping
#' # at the first nodes where the
#' # `value` attribute is 1, 2, or 10;
#' # specify that we should only
#' # keep the finally traversed to
#' # nodes that satisfy the conditions
#' graph <-
#'   graph %>%
#'   trav_in_until(
#'     conditions =
#'       value %in% c(1, 2, 10),
#'     exclude_unmatched = TRUE)
#'
#' # Get the graph's node selection
#' graph %>% get_selection()
#'
#' @export
trav_in_until <- function(
    graph,
    conditions,
    max_steps = 30,
    exclude_unmatched = TRUE,
    add_to_selection = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(
    graph,
    c("Any traversal requires an active selection.",
      "This type of traversal requires a selection of nodes."))

  # Initialize the node stack and
  # the step count
  node_stack <- vector(mode = "integer")
  step <- 0

  starting_nodes <-
    suppressMessages(
      graph %>%
        get_selection())

  # Determine which nodes satisfy the
  # conditions provided
  all_nodes_conditions_met <-
    graph %>%
    get_node_ids(conditions = {{ conditions }})

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  if (exclude_unmatched && all(is.na(all_nodes_conditions_met))) {

    # Clear the active selection
    graph <-
      suppressMessages(
        graph %>%
          clear_selection())

    # Remove action from graph log
    graph$graph_log <-
      graph$graph_log[-nrow(graph$graph_log), ]

    # Update the `graph_log` df with an action
    graph$graph_log <-
      add_action_to_log(
        graph_log = graph$graph_log,
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df))

    # Perform graph actions, if any are available
    if (nrow(graph$graph_actions) > 0) {
      graph <-
        trigger_graph_actions(graph)
    }

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }

    return(graph)
  }

  repeat {

    # Perform traversal
    graph <- graph %>% trav_in()

    # Remove action from graph log
    graph$graph_log <-
      graph$graph_log[-nrow(graph$graph_log), ]

    # If any nodes are `all_nodes_conditions_met` nodes
    # deselect that node and save the node in a stack
    if (any(suppressMessages(graph %>% get_selection()) %in%
            all_nodes_conditions_met)) {

      node_stack <-
        c(node_stack,
          intersect(
            suppressMessages(graph %>% get_selection()),
            all_nodes_conditions_met))

      # Remove the node from the active selection
      graph <-
        graph %>% deselect_nodes(nodes = node_stack)

      # Remove action from graph log
      graph$graph_log <-
        graph$graph_log[-nrow(graph$graph_log), ]
    }

    if (all(is.na(suppressMessages(get_selection(graph))))) break

    step <- step + 1

    if (step == max_steps) break
  }

  if (length(node_stack > 0)) {

    if (add_to_selection) {

      if (exclude_unmatched) {
        node_stack <-
          intersect(node_stack, all_nodes_conditions_met)
      }

      path_nodes <-
        node_stack %>%
        purrr::map(
          .f = function(x) {
            graph %>%
              to_igraph() %>%
              igraph::all_simple_paths(
                from = x,
                to = starting_nodes,
                mode = "out") %>%
              unlist() %>%
              as.integer()}) %>%
        unlist() %>%
        unique()

      graph <-
        graph %>%
        select_nodes_by_id(unique(path_nodes))

    } else {

      graph <-
        graph %>%
        select_nodes_by_id(unique(node_stack))
    }

    # Remove action from graph log
    graph$graph_log <-
      graph$graph_log[-nrow(graph$graph_log), ]

  } else if (length(node_stack) < 1) {

    if (exclude_unmatched &&
        !all(is.na(suppressMessages(get_selection(graph))))) {

      new_selection <- suppressMessages(get_selection(graph))

      graph <-
        suppressMessages(
          graph %>%
            clear_selection() %>%
            select_nodes_by_id(
              intersect(new_selection, all_nodes_conditions_met)))

      # Remove action from graph log
      graph$graph_log <-
        graph$graph_log[-nrow(graph$graph_log), ]

      # Remove action from graph log
      graph$graph_log <-
        graph$graph_log[-nrow(graph$graph_log), ]
    }
  }

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      graph %>%
      trigger_graph_actions()
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

test_that("selective traversals with `trav_in_until()` are possible", {

  # Create a path graph and add
  # values of 1 to 10 across the
  # nodes from beginning to end;
  # select the last path node
  graph <-
    create_graph() %>%
    add_path(
      n = 10,
      node_data = node_data(
        value = 1:10)) %>%
    select_nodes_by_id(
      nodes = 10)

  # Traverse outward, node-by-node
  # until stopping at a node where
  # the `value` attribute is 1
  graph <-
    graph %>%
    trav_in_until(
      conditions =
        value == 1)

  # Expect that the graph's node
  # selection is node `1` (which
  # has a node attr `value` of 1)
  expect_equal(
    graph %>%
      get_selection(), 1)

  # Create two cycles in a graph and
  # add values of 1 to 6 to the
  # first cycle, and values 7 to
  # 12 in the second; select nodes
  # `6` and `12`
  graph <-
    create_graph() %>%
    add_cycle(
      n = 6,
      node_data = node_data(
        value = 1:6)) %>%
    add_cycle(
      n = 6,
      node_data = node_data(
        value = 7:12)) %>%
    select_nodes_by_id(
      nodes = c(6, 12))

  # Traverse inward, node-by-node
  # from `6` and `12` until stopping
  # at the first nodes where the
  # `value` attribute is 1, 2, or 10;
  # specify that we should only
  # keep the finally traversed to
  # nodes that satisfy the conditions
  graph <-
    graph %>%
    trav_in_until(
      conditions =
        value %in% c(1, 2, 10),
      exclude_unmatched = TRUE)

  # Expect that the graph's node
  # selection contains nodes
  # `2` and `10`
  expect_equal(
    graph %>%
      get_selection(), c(2, 10))

  # Expect an error if there isn't
  # an active node selection
  expect_snapshot(error = TRUE,
    create_graph() %>%
      add_path(
        n = 10,
        node_data = node_data(
          value = 1:10)) %>%
      trav_in_until(
        conditions =
          value == 1))
})

#' Traverse from one or more selected nodes onto adjacent, inward edges
#'
#' @description
#'
#' From a graph object of class `dgr_graph` move to incoming edges from a
#' selection of one or more selected nodes, thereby creating a selection of
#' edges. An optional filter by edge attribute can limit the set of edges
#' traversed to.
#'
#' This traversal function makes use of an active selection of nodes. After the
#' traversal, depending on the traversal conditions, there will either be a
#' selection of edges or no selection at all.
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param conditions An option to use filtering conditions for the traversal.
#' @param copy_attrs_from Providing a node attribute name will copy those node
#'   attribute values to the traversed edges. If the edge attribute already
#'   exists, the values will be merged to the traversed edges; otherwise, a new
#'   edge attribute will be created.
#' @param copy_attrs_as If a node attribute name is provided in
#'   `copy_attrs_from`, this option will allow the copied attribute values to be
#'   written under a different edge attribute name. If the attribute name
#'   provided in `copy_attrs_as` does not exist in the graph's edf, the new edge
#'   attribute will be created with the chosen name.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(
#'     n = 2,
#'     type = "a",
#'     label = c("asd", "iekd")) |>
#'   add_n_nodes(
#'     n = 3,
#'     type = "b",
#'     label = c("idj", "edl", "ohd")) |>
#'   add_edges_w_string(
#'     edges = "1->2 1->3 2->4 2->5 3->5",
#'     rel = c(NA, "A", "B", "C", "D"))
#'
#' # Create a data frame with node ID
#' # values representing the graph edges
#' # (with `from` and `to` columns), and,
#' # a set of numeric values
#' df <-
#'   data.frame(
#'     from = c(1, 1, 2, 2, 3),
#'     to = c(2, 3, 4, 5, 5),
#'     values = round(rnorm(5, 5), 2))
#'
#' # Join the data frame to the graph's
#' # internal edge data frame (edf)
#' graph <-
#'   graph |>
#'   join_edge_attrs(df = df)
#'
#' # Show the graph's internal edge data frame
#' graph |> get_edge_df()
#'
#' # Perform a simple traversal from
#' # nodes to inbound edges with no
#' # conditions on the nodes
#' # traversed to
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_in_edge() |>
#'   get_selection()
#'
#' # Traverse from node `2` to any
#' # inbound edges, filtering to
#' # those edges that have NA values
#' # for the `rel` edge attribute
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_in_edge(
#'     conditions = is.na(rel)) |>
#'   get_selection()
#'
#' # Traverse from node `2` to any
#' # inbound edges, filtering to those
#' # edges that do not have NA values
#' # for the `rel` edge attribute
#' # (since there are no allowed
#' # traversals, the selection of node
#' # `2` is retained)
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_in_edge(
#'     conditions = !is.na(rel)) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any
#' # inbound edges, filtering to those
#' # edges that have numeric values
#' # greater than `5.5` for the `rel`
#' # edge attribute
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in_edge(
#'     conditions = values > 5.5) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any
#' # inbound edges, filtering to those
#' # edges that have values equal to
#' # `D` for the `rel` edge attribute
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in_edge(
#'     conditions = rel == "D") |>
#'   get_selection()
#'
#' # Traverse from node `5` to any
#' # inbound edges, filtering to those
#' # edges that have values in the set
#' # `C` and `D` for the `rel` edge
#' # attribute
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in_edge(
#'     conditions = rel %in% c("C", "D")) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any
#' # inbound edges, and use multiple
#' # conditions for the traversal
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in_edge(
#'     conditions =
#'       rel %in% c("C", "D") &
#'       values > 5.5) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any
#' # inbound edges, and use multiple
#' # conditions with a single-length
#' # vector
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in_edge(
#'     conditions =
#'       rel %in% c("D", "E") |
#'       values > 5.5) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any
#' # inbound edges, and use a regular
#' # expression as a filtering condition
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in_edge(
#'     conditions = grepl("C|D", rel)) |>
#'   get_selection()
#'
#' # Show the graph's internal ndf
#' graph |> get_node_df()
#'
#' # Show the graph's internal edf
#' graph |> get_edge_df()
#'
#' # Perform a traversal from all
#' # nodes to their incoming edges and,
#' # while doing so, copy the `label`
#' # node attribute to any of the nodes'
#' # incoming edges
#' graph <-
#'   graph |>
#'   select_nodes() |>
#'   trav_in_edge(
#'     copy_attrs_from = label)
#'
#' # Show the graph's internal edge
#' # data frame after this change
#' graph |> get_edge_df()
#'
#' @export
trav_in_edge <- function(
    graph,
    conditions = NULL,
    copy_attrs_from = NULL,
    copy_attrs_as = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(
    graph,
    c("Any traversal requires an active selection.",
      "This type of traversal requires a selection of nodes."))

  # Get the requested `copy_attrs_from`
  copy_attrs_from <-
    rlang::enquo(copy_attrs_from) |> rlang::get_expr() |> as.character()

  # Get the requested `copy_attrs_as`
  copy_attrs_as <-
    rlang::enquo(copy_attrs_as) |> rlang::get_expr() |> as.character()

  if (length(copy_attrs_from) == 0) {
    copy_attrs_from <- NULL
  }

  if (length(copy_attrs_as) == 0) {
    copy_attrs_as <- NULL
  }

  if (!is.null(copy_attrs_as) && !is.null(copy_attrs_from)) {
    if (copy_attrs_as == copy_attrs_from) {
      copy_attrs_as <- NULL
    }
  }

  # Get the selection of nodes as the starting
  # nodes for the traversal
  starting_nodes <- graph$node_selection$node

  # Get the graph's node data frame
  ndf <- graph$nodes_df

  # Get the graph's edge data frame
  edf <- graph$edges_df

  # Find all edges that lead toward the
  # starting nodes and remove edges that
  # are loops
  valid_edges <-
    edf |>
    dplyr::filter(to %in% starting_nodes) |>
    dplyr::filter(to != from)

  # If no rows returned, then there are no
  # valid traversals, so return the same graph
  # object without modifying the selection
  if (nrow(valid_edges) == 0) {
    return(graph)
  }

  # If traversal conditions are provided then
  # pass in those conditions and filter the
  # data frame of `valid_edges`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    valid_edges <- dplyr::filter(.data = valid_edges, {{ conditions }})
  }

  # If no rows returned, then there are no
  # valid traversals, so return the same graph
  # object without modifying the selection
  if (nrow(valid_edges) == 0) {
    return(graph)
  }

  # If the option is taken to copy node attribute
  # values to the traversed edges, perform the join
  # operation
  if (!is.null(copy_attrs_from)) {

    ndf_2 <-
      ndf |>
      dplyr::filter(id %in% starting_nodes) |>
      dplyr::select("id", !!enquo(copy_attrs_from))

    if (!is.null(copy_attrs_as)) {

      if (copy_attrs_as %in% c("id", "from", "to")) {
        cli::cli_abort("Copied attributes should not overwrite either of the `id`, `from`, or `to` edge attributes.")
      }

      colnames(ndf_2)[2] <- copy_attrs_from <- copy_attrs_as
    }

    # If node attribute doesn't exist in the edf,
    # perform a full join
    if (!(copy_attrs_from %in% colnames(edf))) {

      edges <-
        ndf_2 |>
        dplyr::right_join(edf, c("id" = "to")) |>
        dplyr::rename(to = "id") |>
        dplyr::rename(id = "id.y") |>
        dplyr::relocate("id", "from", "to", "rel") |>
        dplyr::arrange(id)
    }

    # If node attribute exists as a column in the edf
    if (copy_attrs_from %in% colnames(edf)) {

      edges <-
        ndf_2 |>
        dplyr::right_join(edf, by = c("id" = "to")) |>
        dplyr::rename(to = "id", id = "id.y")

      # Get column numbers that end with ".x" or ".y"
      split_var_x_col <-
        grep("\\.x$", colnames(edges))

      split_var_y_col <-
        grep("\\.y$", colnames(edges))

      # Selectively merge in values to the existing
      # edge attribute column
      for (i in seq_len(nrow(edges))) {
        if (!is.na(edges[i, split_var_x_col])) {
          edges[i, split_var_y_col] <- edges[i, split_var_x_col]
        }
      }

      # Rename the ".y" column
      colnames(edges)[split_var_y_col] <- copy_attrs_from

      # Drop the ".x" column
      edges <- edges[-split_var_x_col]

      # Reorder columns
      edges <-
        edges |>
        dplyr::relocate("id", "from", "to", "rel") |>
        dplyr::arrange(id)
    }

    # Update the graph's internal node data frame
    graph$edges_df <- edges
  }

  # Add the edge information to the active selection
  # of edges in `graph$edge_selection`
  graph$edge_selection <-
    replace_graph_edge_selection(
      graph = graph,
      edge_id = valid_edges$id,
      from_node = valid_edges$from,
      to_node = valid_edges$to)

  # Replace `graph$node_selection` with an empty df
  graph$node_selection <- create_empty_nsdf()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

# Traversals through nodes and edges in a graph object

test_that("simple traversals are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4)

  # Starting at node `1`, traverse to node `4`, storing
  # the traversed location as a selection in the graph
  # object
  graph <-
    graph %>%
    select_nodes(nodes = 1) %>%
    trav_out() %>%
    trav_out() %>%
    trav_out()

  # Expect that node `4` is the current selection
  expect_equal(
    get_selection(graph), 4)

  # Traverse back to node `1` from node `4`
  graph <-
    graph %>%
    trav_in() %>%
    trav_in() %>%
    trav_in()

  # Expect that node `1` is the current selection
  expect_equal(
    get_selection(graph), 1)

  # Traverse from node `1` to `2`, then,
  # traverse to nodes in both directions
  graph <-
    graph %>%
    trav_out() %>%
    trav_both()

  # Expect that nodes `1` and `3` are in the
  # current selection
  expect_in(get_selection(graph), c(1, 3))

  # Traverse by moving from nodes onto edges, then,
  # onto nodes; from `1` to `4`
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes(nodes = 1) %>%
    trav_out_edge() %>%
    trav_in_node() %>%
    trav_out_edge() %>%
    trav_in_node() %>%
    trav_out_edge() %>%
    trav_in_node()

  # Expect that node `4` is the current selection
  expect_equal(
    get_selection(graph), 4)

  # Traverse back to node `1` from node `4`,
  # using the same types of traversals
  graph <-
    graph %>%
    trav_in_edge() %>%
    trav_out_node() %>%
    trav_in_edge() %>%
    trav_out_node() %>%
    trav_in_edge() %>%
    trav_out_node()

  # Expect that node `1` is the current selection
  expect_equal(
    get_selection(graph), 1)

  # Modify the graph so that it contains a branch
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes(nodes = 3) %>%
    add_n_nodes_ws(
      n = 1,
      direction = "from") %>%
    clear_selection() %>%
    select_nodes(nodes = 2) %>%
    add_n_nodes_ws(
      n = 2,
      direction = "from") %>%
    clear_selection()

  # Traverse nodes from `1` until
  # traversal can no longer occur
  graph <-
    graph %>%
    select_nodes(nodes = 1) %>%
    trav_out()

  # Expect that node `2` is the current selection
  expect_equal(
    get_selection(graph), 2)

  # Continue traversal outward by node
  graph <-
    graph %>%
    trav_out()

  # Expect that nodes `3`, `6`, and `7` are
  # in the current selection
  expect_equal(
    get_selection(graph), c(3, 6, 7))

  # Continue traversal outward by node
  graph <-
    graph %>%
    trav_out()

  # Expect that nodes `4` and `5` are in the
  # current selection
  expect_equal(
    get_selection(graph), c(4, 5))

  # Continue traversal outward, even though at end
  graph <-
    graph %>%
    trav_out()

  # Expect that nodes `4` and `5` are still in the
  # current selection
  expect_equal(
    get_selection(graph), c(4, 5))



  expect_snapshot(error = TRUE, {
    # Expect an error if attempting to perform a node
    # traversal without any selection of nodes
    graph <-
      graph %>%
      clear_selection()

    trav_in(graph)
    trav_out(graph)
    trav_both(graph)
    trav_in_node(graph)
    trav_out_node(graph)
    trav_in_edge(graph)
    trav_out_edge(graph)
  })
})

test_that("selective traversals with `trav_out()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = 5) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at node `1`, traverse to
  # node `2` with a match expression (==)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out(
      conditions = data_value == 10)

  # Expect that node `2` is the
  # current selection
  expect_equal(
    get_selection(graph), 2)

  # Starting at node `1`, traverse to
  # node `2`, using a different match
  # expression (<)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out(
      conditions = data_value < 15)

  # Expect that node `2` is the
  # current selection
  expect_equal(
    get_selection(graph), 2)

  # Starting at node `1`, traverse to
  # node `2`, using a different match expression (>)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out(
      conditions = data_value > 5)

  # Expect that node `2` is the
  # current selection
  expect_equal(
    get_selection(graph), 2)

  # Starting at node `1`, traverse to
  # node `2`, using a different match expression (!=)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out(
      conditions = data_value != 5)

  # Expect that node `2` is the
  # current selection
  expect_equal(
    get_selection(graph), 2)

  # Starting at node `1`, attempt to
  # traverse to node `2` using a match
  # expression that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out(
      conditions = data_value != 10)

  # Expect that node `1` is the
  # current selection (since no traversal
  # had occurred)
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `3`, traverse to
  # node `4` using a match on a character field
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_out(
      conditions = shape == "square")

  # Expect that node `4` is the
  # current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at node `3`, attempt to
  # traverse to node `4` using a match
  # expression that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_out(
      conditions = shape == "triangle")

  # Expect that node `3` is the current selection
  expect_equal(
    get_selection(graph), 3)
})

test_that("selective traversals with `trav_in()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = 5) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at node `4`, traverse to node `3` with a
  # match expression (==)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in(
      conditions = data_value == 10)

  # Expect that node `3` is the current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, traverse to node `3`, using a
  # different match expression (<)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in(
      conditions = data_value < 15)

  # Expect that node `3` is the
  # current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, traverse
  # to node `3`, using a different
  # match expression (>)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in(
      conditions = data_value > 5)

  # Expect that node `3` is the
  # current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, traverse
  # to node `3`, using a
  # different match expression (!=)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in(
      conditions = data_value != 5)

  # Expect that node `3` is the
  # current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, attempt
  # to traverse to node `3` using a
  # match expression that won't
  # yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in(
      conditions = data_value != 10)

  # Expect that node `4` is the
  # current selection (since no
  # traversal had occurred)
  expect_equal(
    get_selection(graph), 4)

  # Starting at node `2`, traverse
  # to node `1` using a match on a
  # character field
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 2) %>%
    trav_in(
      conditions = shape == "triangle")

  # Expect that node `1` is the
  # current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `2`, attempt
  # to traverse to node `1` using a
  # match expression that won't
  # yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 2) %>%
    trav_in(
      conditions = shape == "square")

  # Expect that node `2` is the
  # current selection (since no
  # traversal had occurred)
  expect_equal(
    get_selection(graph), 2)
})

test_that("selective traversals with `trav_out_edge()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = 5) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at node `1`, traverse to edge between
  # nodes `1` and `2` with a match expression (==)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(
      conditions = data_value == 5)

  # Expect that the edge `1` -> `2`
  # is the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `1`, traverse to
  # the edge between nodes `1` and `2`
  # with a match expression (<)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(
      conditions = data_value < 10)

  # Expect that the edge `1` -> `2` is
  # the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `1`, traverse to
  # the edge between nodes `1` and `2`
  # with a match expression (>)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(
      conditions = data_value > 2)

  # Expect that the edge `1` -> `2`
  # is the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `1`, traverse to
  # the edge between nodes `1` and `2`
  # with a match expression (!=)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(
      conditions = data_value != 1)

  # Expect that the edge `1` -> `2`
  # is the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `1`, attempt to
  # traverse to edge between nodes `1`
  # and `2` using a match expression
  # that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(
      conditions = data_value != 5)

  # Expect that node `1` is the
  # current selection (since no traversal
  # had occurred)
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `1`, traverse to
  # the edge between nodes `1` and `2`
  # using a match on a character field
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(
      conditions = rel == "related_to")

  # Expect that the edge `1` -> `2`
  # is the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `1`, attempt to
  # traverse to the edge between nodes
  # `1` and `2` using a match expression
  # that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    trav_out_edge(
      conditions = rel == "belongs_with")

  # Expect that node `1` is the
  # current selection (since no
  # traversal had occurred)
  expect_equal(
    get_selection(graph), 1)
})

test_that("selective traversals with `trav_in_edge()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = 5) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at node `4`, traverse to
  # the edge between nodes `3` and `4`
  # with a match expression (==)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in_edge(
      conditions = data_value == 5)

  # Expect that the edge `3` -> `4`
  # is the current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, traverse to
  # the edge between nodes `3` and `4`
  # with a match expression (<)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in_edge(
      conditions = data_value < 10)

  # Expect that the edge `3` -> `4`
  # is the current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, traverse to
  # the edge between nodes `3` and `4`
  # with a match expression (>)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in_edge(
      conditions = data_value > 2)

  # Expect that the edge `3` -> `4`
  # is the current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, traverse to
  # the edge between nodes `3` and `4`
  # with a match expression (!=)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in_edge(
      conditions = data_value != 1)

  # Expect that the edge `3` -> `4`
  # is the current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, attempt to
  # traverse to the edge between nodes
  # `3` and `4` using a match expression
  # that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in_edge(
      conditions = data_value != 5)

  # Expect that node `4` is the
  # current selection (since no traversal
  # had occurred)
  expect_equal(
    get_selection(graph), 4)

  # Starting at node `4`, traverse to
  # the edge between nodes `3` and `4`
  # using a match on a character field
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in_edge(
      conditions = rel == "related_to")

  # Expect that the edge `3` -> `4`
  # is the current selection
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, attempt to
  # traverse to edge between nodes `3`
  # and `4` using a match expression
  # that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    trav_in_edge(
      conditions = rel == "belongs_with")

  # Expect that node `4` is the
  # current selection (since no traversal
  # had occurred)
  expect_equal(
    get_selection(graph), 4)
})

test_that("selective traversals with `trav_in_node()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = 5) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at edge `3` -> `4`,
  # traverse to node `4` with a
  # match expression (==)
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 3,
      to = 4) %>%
    trav_in_node(
      conditions = data_value == 5)

  # Expect that the node `4` is
  # the current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at edge `3` -> `4`,
  # traverse to node `4` with a match
  # expression (<)
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 3,
      to = 4) %>%
    trav_in_node(
      conditions = data_value < 10)

  # Expect that the node `4` is
  # the current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at edge `3` -> `4`,
  # traverse to node `4` with a
  # match expression (>)
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 3,
      to = 4) %>%
    trav_in_node(
      conditions = data_value > 1)

  # Expect that the node `4` is
  # the current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at edge `3` -> `4`,
  # traverse to node `4` with a
  # match expression (!=)
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 3,
      to = 4) %>%
    trav_in_node(
      conditions = data_value != 1)

  # Expect that the node `4`
  # is the current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at edge `3` -> `4`,
  # attempt to traverse to node `4`
  # using a match expression that won't
  # yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 3,
      to = 4) %>%
    trav_in_node(
      conditions = data_value != 5)

  # Expect that the edge `3` -> `4`
  # is the current selection (since
  # no traversal had occurred)
  expect_equal(
    get_selection(graph), 3)

  # Starting at node `4`, traverse to
  # edge between nodes `3` and `4` using
  # a match on a character field
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 3,
      to = 4) %>%
    trav_in_node(
      conditions = shape == "square")

  # Expect that the node `4` is
  # the current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at edge `3` -> `4`,
  # attempt to traverse to node `4`
  # using a match on a character field
  # that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 3,
      to = 4) %>%
    trav_in_node(
      conditions = shape == "triangle")

  # Expect that the edge `3` -> `4`
  # is the current selection (since
  # no traversal had occurred)
  expect_equal(
    get_selection(graph), 3)
})

test_that("selective traversals with `trav_out_node()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = 5) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at edge `1` -> `2`,
  # traverse to node `1` with a
  # match expression (==)
  graph <-
    graph %>%
    select_edges(
      from = 1,
      to = 2) %>%
    trav_out_node(
      conditions = data_value == 5)

  # Expect that the node `1` is
  # the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at edge `1` -> `2`,
  # traverse to node `1` with a
  # match expression (<)
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 1,
      to = 2) %>%
    trav_out_node(
      conditions = data_value < 10)

  # Expect that the node `1` is
  # the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at edge `1` -> `2`,
  # traverse to node `1` with a
  # match expression (>)
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 1,
      to = 2) %>%
    trav_out_node(
      conditions = data_value > 1)

  # Expect that the node `1` is
  # the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at edge `1` -> `2`,
  # traverse to node `1` with a
  # match expression (!=)
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 1,
      to = 2) %>%
    trav_out_node(
      conditions = data_value != 1)

  # Expect that the node `1` is
  # the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at edge `1` -> `2`,
  # attempt to traverse to node `1`
  # using a match expression that
  # won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 1,
      to = 2) %>%
    trav_out_node(
      conditions = data_value != 5)

  # Expect that the node `1`
  # is the current selection (since
  # no traversal had occurred)
  expect_equal(
    get_selection(graph), 1)

  # Starting at edge `1` -> `2`,
  # traverse to node `1` using a
  # match on a character field
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 1,
      to = 2) %>%
    trav_out_node(
      conditions = shape == "triangle")

  # Expect that the node `1`
  # is the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at edge `1` -> `2`,
  # attempt to traverse to node `1`
  # using a match on a character field
  # that won't yield a match
  graph <-
    graph %>%
    clear_selection() %>%
    select_edges(
      from = 1,
      to = 2) %>%
    trav_out_node(
      conditions = shape == "circle")

  # Expect that the edge `1` -> `2`
  # is the current selection (since no
  # traversal had occurred)
  expect_equal(
    get_selection(graph), 1)
})

test_that("selective traversals with `trav_both()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = 5) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at node `3`, traverse
  # to nodes `2` and `4` with a
  # match expression (==)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both(
      conditions = data_value == 10)

  # Expect that node `2` is in
  #@ the current selection
  expect_equal(
    get_selection(graph), 2)

  # Starting at node `3`, traverse
  # to nodes `2` and `4` with a
  # match expression (<)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both(
      conditions = data_value < 15)

  # Expect that nodes `2` and `4`
  # are in the current selection
  expect_equal(
    get_selection(graph), c(2, 4))

  # Starting at node `3`, traverse
  # to nodes `2` and `4` with a
  # match expression (>)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both(
      conditions = data_value > 5)

  # Expect that node `2` is in
  # the current selection
  expect_equal(
    get_selection(graph), 2)

  # Starting at node `3`, traverse
  # to nodes `2` and `4` with a
  # match expression (!=)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both(
      conditions = data_value != 5)

  # Expect that node `2` is in
  # the current selection
  expect_equal(
    get_selection(graph), 2)

  # Starting at node `3`, attempt
  # to traverse to nodes `2` and `4`
  # with a match expression that won't
  # yield a match in one direction
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both(
      conditions = data_value != 10)

  # Expect that the node `4` is
  # the current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at node `3`, traverse to
  # nodes `2` and `4` with a character
  # match expression
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both(
      conditions = type == "circle")

  # Expect that nodes `2` and `4`
  # are in the current selection
  expect_equal(
    get_selection(graph), c(2, 4))

  # Starting at node `3`, attempt to
  # traverse to nodes `2` and `4` with
  # a character match expression that
  # won't yield a match in one direction
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both(
      conditions = shape == "square")

  # Expect that the node `4` is
  # the current selection
  expect_equal(
    get_selection(graph), 4)

  # Starting at node `2`, traverse
  # to nodes `1` and `3` with a
  # character match expression
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 2) %>%
    trav_both(
      conditions = shape == "triangle")

  # Expect that the node `1` is
  # the current selection
  expect_equal(
    get_selection(graph), 1)

  # Starting at node `2`, traverse to
  # node `2` with a character match expression
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 2) %>%
    trav_both(
      conditions = shape == "circle")

  # Expect that the node `2` is
  # the current selection
  expect_equal(
    get_selection(graph), 2)
})

test_that("selective traversals with `trav_both_edge()` are possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3) %>%
    add_edge(
      from = 3,
      to = 4) %>%
    select_nodes() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "circle") %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = c(2, 3)) %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 10) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 4) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "square") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = shape,
      value = "triangle") %>%
    set_node_attrs_ws(
      node_attr = data_value,
      value = 5) %>%
    clear_selection() %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = data_value,
      value = c(1, 2, 3)) %>%
    set_edge_attrs_ws(
      edge_attr = rel,
      value = "related_to") %>%
    clear_selection()

  # Starting at node `3`, traverse
  # to edges `3` -> `4` and `2` -> `3`
  # with a match expression (==)
  graph <-
    graph %>%
    clear_selection() %>%
    select_nodes_by_id(nodes = 3) %>%
    trav_both_edge(
      conditions = data_value == 2)

  # Expect that edge `2` is in
  # the current selection
  expect_equal(
    get_selection(graph), 2)

  # Expect an error if the graph
  # has no selection of nodes
  expect_error(
    graph %>%
      clear_selection() %>%
      trav_both_edge(
        conditions = data_value == 2))

  # Return the same graph if a
  # traversal with conditions does
  # not result in a traversal
  expect_equal(
    graph %>%
      clear_selection() %>%
      select_nodes_by_id(nodes = 3) %>%
      trav_both_edge(
        conditions = data_value == 6) %>%
      get_selection(),
    3)
})

test_that("traversing to reverse edges with an edge selection is possible", {

  # Create a node data frame (ndf)
  ndf <-
    create_node_df(
      n = 4,
      type = "basic",
      label = TRUE)

  # Create an edge data frame (edf)
  edf <-
    create_edge_df(
      from = c(1, 4, 2, 3, 3),
      to = c(4, 1, 3, 2, 1))

  # Create a graph with the ndf and edf
  graph <-
    create_graph(
      nodes_df = ndf,
      edges_df = edf)

  # Explicitly select the edges `1`->`4`
  # and `2`->`3`
  graph <-
    graph %>%
    select_edges(
      from = 1,
      to = 4) %>%
    select_edges(
      from = 2,
      to = 3)

  # Add to the selection the reverse edges
  graph_add_selection <-
    graph %>%
    trav_reverse_edge(
      add_to_selection = TRUE)

  # Expect certain edges to be available
  # in the selection
  expect_equal(
    get_selection(graph_add_selection),
    c(1, 2, 3, 4))

  # Flip the selection to the reverse edge
  graph_flip_selection <-
    graph %>%
    trav_reverse_edge(
      add_to_selection = FALSE)

  # Expect certain edges to be available
  # in the selection
  expect_equal(
    get_selection(graph_flip_selection),
    c(2, 4))

  # Expect an error if there is no active edge
  # selection in the graph
  expect_error(
    graph %>%
      clear_selection() %>%
      trav_reverse_edge(
        add_to_selection = TRUE))
})

test_that("selective traversals with `trav_out_until()` are possible", {

  # Create a path graph and add
  # values of 1 to 10 across the
  # nodes from beginning to end;
  # select the first path node
  graph <-
    create_graph() %>%
    add_path(
      n = 10,
      node_data = node_data(
        value = 1:10)) %>%
    select_nodes_by_id(
      nodes = 1)

  # Traverse outward, node-by-node
  # until stopping at a node where
  # the `value` attribute is 8
  graph <-
    graph %>%
    trav_out_until(
      conditions =
        value == 8)

  # Expect that the graph's node
  # selection is node `8` (which
  # has a node attr `value` of 8)
  expect_equal(
    graph %>%
      get_selection(), 8)

  # Create two cycles in a new graph
  # and add values of 1 to 6 to the
  # first cycle, and values 7 to
  # 12 in the second; select nodes
  # `1` and `7`
  graph <-
    create_graph() %>%
    add_cycle(
      n = 6,
      node_data = node_data(
        value = 1:6)) %>%
    add_cycle(
      n = 6,
      node_data = node_data(
        value = 7:12)) %>%
    select_nodes_by_id(
      nodes = c(1, 7))

  # Traverse outward, node-by-node
  # from `1` and `7` until stopping
  # at the first nodes where the
  # `value` attribute is 5, 6, or 15;
  # specify that we should only
  # keep the finally traversed to
  # nodes that satisfy the conditions
  graph <-
    graph %>%
    trav_out_until(
      conditions =
        value %in% c(5, 6, 9),
      exclude_unmatched = TRUE)

  # Expect that the graph's node
  # selection contains nodes
  # `5` and `9`
  expect_equal(
    graph %>%
      get_selection(), c(5, 9))

  # Expect an error if there isn't
  # an active node selection
  expect_error(
    create_graph() %>%
      add_path(
        n = 10,
        node_data = node_data(
          value = 1:10)) %>%
      trav_out_until(
        conditions =
          value == 8))
})

#' Get the number of asymmetrically-connected node pairs
#'
#' @description
#'
#' Get the number of asymmetrically-connected node pairs. This works for
#' directed graphs.
#'
#' @inheritParams render_graph
#'
#' @return A single numeric value representing the number of
#'   asymmetrically-connected node pairs.
#'
#' @examples
#' # Create a cycle graph
#' graph <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' # Get a count of asymmetrically-
#' # connected node pairs
#' graph %>%
#'   count_asymmetric_node_pairs()
#'
#' # Create a full graph and then
#' # count the asymmetrically-
#' # connected node pairs
#' create_graph() %>%
#'   add_full_graph(n = 10) %>%
#'   count_asymmetric_node_pairs()
#'
#' @export
count_asymmetric_node_pairs <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the number of asymmetrically-connected node pairs
  # in the graph
  unname(unlist(igraph::dyad_census(ig_graph)["asym"]))
}

# Get graph metrics

test_that("Getting a degree histogram is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get the degree histogram for total degree
  degree_hist_all <-
    get_degree_histogram(
      graph = graph,
      mode = "all")

  # Expect that `degree_hist_all` inherits from `data.frame`
  expect_s3_class(
    degree_hist_all, "data.frame")

  # Expect certain column names for the `degree_hist_all` object
  expect_identical(
    colnames(degree_hist_all),
    c("degree", "total_degree_hist"))

  # Get the degree histogram for in-degree
  degree_hist_in <-
    get_degree_histogram(
      graph = graph,
      mode = "in")

  # Expect that `degree_hist_in` inherits from `data.frame`
  expect_s3_class(
    degree_hist_in, "data.frame")

  # Expect certain column names for the `degree_hist_in` object
  expect_identical(
    colnames(degree_hist_in),
    c("degree", "indegree_hist"))

  # Get the degree histogram for out-degree
  degree_hist_out <-
    get_degree_histogram(
      graph = graph,
      mode = "out")

  # Expect that `degree_hist_out` inherits from `data.frame`
  expect_s3_class(
    degree_hist_out, "data.frame")

  # Expect certain column names for the `degree_hist_out` object
  expect_identical(
    colnames(degree_hist_out),
    c("degree", "outdegree_hist"))
})

test_that("Getting degree distribution is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get the degree distribution for total degree
  degree_dist_all <-
    get_degree_distribution(
      graph = graph,
      mode = "all")

  # Expect that `degree_dist_all` inherits from `data.frame`
  expect_s3_class(
    degree_dist_all, "data.frame")

  # Expect certain column names for the `degree_dist_all` object
  expect_identical(
    colnames(degree_dist_all),
    c("degree", "total_degree_dist"))

  # Get the degree distribution for in-degree
  degree_dist_in <-
    get_degree_distribution(
      graph = graph,
      mode = "in")

  # Expect that `degree_dist_in` inherits from `data.frame`
  expect_s3_class(
    degree_dist_in, "data.frame")

  # Expect certain column names for the `degree_dist_in` object
  expect_identical(
    colnames(degree_dist_in),
    c("degree", "indegree_dist"))

  # Get the degree distribution for out-degree
  degree_dist_out <-
    get_degree_distribution(
      graph = graph,
      mode = "out")

  # Expect that `degree_dist_out` inherits from `data.frame`
  expect_s3_class(
    degree_dist_out, "data.frame")

  # Expect certain column names for the `degree_dist_out` object
  expect_identical(
    colnames(degree_dist_out),
    c("degree", "outdegree_dist"))
})

test_that("Getting maximum eccentricity is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  max_eccen <- get_max_eccentricity(graph)

  # Expect that `max_eccen` is a numeric vector
  expect_type(max_eccen, "double")

  # Expect that `max_eccen` is of length 1
  expect_length(max_eccen, 1)

  # Expect that `max_eccen` has the value 5
  expect_equal(max_eccen, 5)

  # Expect NA if there aren't any nodes
  # in the graph
  expect_equal(
    get_max_eccentricity(
      graph = create_graph()),
    NA_real_)
})

test_that("Getting minimum eccentricity is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 23,
      set_seed = 23)

  min_eccen <-
    get_min_eccentricity(
      graph = graph,
      direction = "all")

  # Expect that `min_eccen` is a numeric vector
  expect_type(min_eccen, "double")

  # Expect that `min_eccen` is of length 1
  expect_length(min_eccen, 1)

  # Expect that `min_eccen` has the value 2
  expect_equal(min_eccen, 2)

  # Expect certain values with different
  # values provided for `direction`
  expect_equal(
    get_min_eccentricity(
      graph = create_graph() %>% add_cycle(n = 6),
      direction = "out"),
    5)

  expect_equal(
    get_min_eccentricity(
      graph = create_graph() %>% add_cycle(n = 6),
      direction = "in"),
    5)

  # Expect an error if `direction` is not valid
  expect_error(
    get_min_eccentricity(
      graph = graph,
      direction = "away"))

  # Expect NA if there aren't any nodes
  # in the graph
  expect_equal(
    get_min_eccentricity(
      graph = create_graph()),
    NA_real_
  )
})

test_that("Getting graph eccentricity is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 23,
      set_seed = 23)

  graph_eccen <- get_eccentricity(graph)

  # Expect that `graph_eccen` is a data frame
  expect_s3_class(
    graph_eccen, "data.frame")

  # Expect that `graph_eccen` has 10 rows
  # this case
  expect_equal(
    nrow(graph_eccen), 10)

  # Expect certain column names for the `graph_eccen` object
  expect_named(
    graph_eccen, c("id", "eccentricity"))

  # Expect that `eccentricity` column is numeric
  expect_type(
    graph_eccen$eccentricity, "double")

  # Expect that `id` column has integer values
  expect_type(
    graph_eccen$id, "integer")
})

test_that("Getting graph periphery is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 23,
      set_seed = 23)

  graph_periphery <- get_periphery(graph)

  # Expect that `graph_periphery` is an integer vector
  expect_type(
    graph_periphery, "integer")

  # Expect that `graph_periphery` has length 3 in
  # this case
  expect_length(
    graph_periphery, 3)

  # Expect certain values for the
  # `graph_periphery` object
  expect_equal(
    graph_periphery, c(3, 7, 10))
})

test_that("Getting graph info is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 23,
      set_seed = 23)

  # Add a graph name
  graph <-
    set_graph_name(
      graph = graph,
      name = "random")

  # Add a graph time
  graph <-
    set_graph_time(
      graph = graph,
      time = "2015-10-25 15:23:00")

  # Use the `get_graph_info()` function to create a
  # data frame with graph metrics
  graph_i <-
    graph %>%
    get_graph_info()

  # Expect that `graph_i` is a data frame
  expect_s3_class(
    graph_i, "data.frame")

  # Expect that `graph_i` has 9 columns
  expect_equal(
    ncol(graph_i), 9)

  # Expect that `graph_i` has 1 row
  expect_equal(
    nrow(graph_i), 1)

  # Expect that `graph_i` has certain column names
  expect_named(
    graph_i,
    c("name", "n", "e", "dens", "mn_deg",
      "mx_deg", "avg_deg", "time", "tz"))
})

test_that("Checking whether the graph is connected is possible", {

  # Create a random graph
  graph_connected <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 25,
      set_seed = 23)

  # Test that the graph is indeed connected
  expect_true(
    is_graph_connected(graph_connected))

  graph_not_connected <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 8,
      set_seed = 23)

  # Test that the graph is indeed connected
  expect_false(
    is_graph_connected(graph_not_connected))
})

test_that("Getting graph adhesion is possible", {

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a full graph
  graph_full <-
    create_graph() %>%
    add_full_graph(n = 8)

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific adhesion values for
  # the two non-empty graphs
  expect_equal(
    get_adhesion(graph_cycle), 1)

  expect_equal(
    get_adhesion(graph_full), 7)

  # Expect NA for the empty graph
  expect_equal(
    get_adhesion(graph_empty), NA_real_)
})

test_that("Getting graph girth is possible", {

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a full graph
  graph_full <-
    create_graph() %>%
    add_full_graph(n = 8)

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific girth values for
  # the two non-empty graphs
  expect_equal(
    get_girth(graph_cycle), 5)

  expect_equal(
    get_girth(graph_full), 3)

  # Expect NA for the empty graph
  expect_equal(
    get_girth(graph_empty), NA_real_)
})

test_that("Getting the mean distance for a graph is possible", {

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a full graph
  graph_full <-
    create_graph() %>%
    add_full_graph(n = 8)

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific mean distance values for
  # the two non-empty graphs
  expect_equal(
    get_mean_distance(graph_cycle), 2.5)

  expect_equal(
    get_mean_distance(graph_full), 1)

  # Expect NA for the empty graph
  expect_equal(
    get_mean_distance(graph_empty), NA_real_)
})

test_that("Getting the reciprocity for a graph is possible", {

  # Define a graph where 2 edge definitions
  # have pairs of reciprocal edges
  graph <-
   create_graph() %>%
   add_cycle(n = 3) %>%
   add_node(
     from = 1,
       to = 1) %>%
   add_node(
     from = 1,
       to = 1)

  # Expect that the reciprocity
  # will be the ratio of reciprocating
  # edges (4) to the total number
  # of graph edges (7)
  expect_equal(
    get_reciprocity(graph), 4 / 7)

  # Expect that a graph with no
  # edges will return NA
  expect_true(
    is.na(get_reciprocity(
      create_graph() %>%
        add_n_nodes(n = 5))))
})

test_that("Getting the minimum cut between nodes is possible", {

  # Set a seed
  suppressWarnings(RNGversion("3.5.0"))
  set.seed(23)

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a cycle graph with a
  # `capacity` edge attribute
  graph_capacity <-
    create_graph() %>%
    add_cycle(n = 5) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = "capacity",
      value =
        rnorm(
          n = count_edges(graph = .),
          mean = 5,
          sd = 1)) %>%
    clear_selection()

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific minimum cut values
  # with different pairs of nodes from
  # the two non-empty graphs
  expect_equal(
    get_min_cut_between(
      graph = graph_cycle,
      from = 1,
      to = 2),
    1)

  expect_equal(
    get_min_cut_between(
      graph = graph_capacity,
      from = 1,
      to = 2),
    tolerance = 0.02,
    expected = 4.479822)

  # Expect NA for the empty graph
  expect_equal(
    get_min_cut_between(graph_empty), NA_real_)
})

test_that("Getting a count of graph automorphisms is possible", {

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a full graph
  graph_full <-
    create_graph() %>%
    add_full_graph(n = 8)

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific automorphism counts
  # for the two non-empty graphs
  expect_equal(
    count_automorphisms(graph_cycle), 10)

  expect_equal(
    count_automorphisms(graph_full), 40320)

  # Expect NA for the empty graph
  expect_equal(
    count_automorphisms(graph_empty), NA_real_)
})

test_that("Getting a count of asymmetric node pairs is possible", {

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a full graph
  graph_full <-
    create_graph() %>%
    add_full_graph(n = 8)

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific automorphism counts
  # for the two non-empty graphs
  expect_equal(
    count_asymmetric_node_pairs(graph_cycle), 5)

  expect_equal(
    count_asymmetric_node_pairs(graph_full), 0)

  # Expect NA for the empty graph
  expect_equal(
    count_asymmetric_node_pairs(graph_empty), NA_real_)
})

test_that("Getting a count of mutual node pairs is possible", {

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a full graph
  graph_full <-
    create_graph() %>%
    add_full_graph(n = 8)

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific automorphism counts
  # for the two non-empty graphs
  expect_equal(
    count_mutual_node_pairs(graph_cycle), 0)

  expect_equal(
    count_mutual_node_pairs(graph_full), 28)

  # Expect NA for the empty graph
  expect_equal(
    count_mutual_node_pairs(graph_empty), NA_real_)
})

test_that("Getting a count of unconnected node pairs is possible", {

  # Create a cycle graph
  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  # Create a full graph
  graph_full <-
    create_graph() %>%
    add_full_graph(n = 8)

  # Create an empty graph
  graph_empty <-
    create_graph()

  # Expect specific automorphism counts
  # for the two non-empty graphs
  expect_equal(
    count_unconnected_node_pairs(graph_cycle), 5)

  expect_equal(
    count_unconnected_node_pairs(graph_full), 0)

  # Expect NA for the empty graph
  expect_equal(
    count_unconnected_node_pairs(graph_empty), NA_real_)
})

#' Traverse to any reverse edges
#'
#' @description
#'
#' From an active selection of edges in a graph object of class `dgr_graph`,
#' traverse to any available reverse edges between the nodes common to the
#' selected edges. For instance, if an active selection has the edge `1->2` but
#' there is also an (not selected) edge `2->1`, then this function can either
#' switch to the selection of `2->1`, or, incorporate both those edges into the
#' active selection of edges.
#'
#' This traversal function makes use of an active selection of edges. After the
#' traversal, depending on the traversal conditions, there will either be a
#' selection of edges or no selection at all.
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#' @param add_to_selection An option to either add the reverse edges to the
#'   active selection of edges (`TRUE`) or switch the active selection entirely
#'   to those reverse edges (`FALSE`, the default case).
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = "basic",
#'     label = TRUE)
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 4, 2, 3, 3),
#'     to =   c(4, 1, 3, 2, 1))
#'
#' # Create a graph with the
#' # ndf and edf
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Explicitly select the edges
#' # `1`->`4` and `2`->`3`
#' graph <-
#'   graph %>%
#'   select_edges(
#'     from = 1,
#'       to = 4) %>%
#'   select_edges(
#'     from = 2,
#'       to = 3)
#'
#' # Get the inital edge selection
#' graph %>% get_selection()
#'
#' # Traverse to the reverse edges
#' # (edges `2`: `4`->`1` and
#' # `4`:`3`->`2`)
#' graph <-
#'   graph %>%
#'   trav_reverse_edge()
#'
#' # Get the current selection of edges
#' graph %>% get_selection()
#'
#' @export
trav_reverse_edge <- function(
    graph,
    add_to_selection = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid edge selection
  check_graph_contains_edge_selection(graph)

  # Extract the selection of edges
  edges_from <- graph$edge_selection$from
  edges_to <- graph$edge_selection$to

  # Extract the graph's internal edf
  edf <- graph$edges_df

  # Get the available reverse edges
  reverse_edges_df <- data.frame(to = edges_from, from = edges_to)
  reverse_edges <- edf %>% dplyr::inner_join(reverse_edges_df, by = c("from", "to"))

  # Add the reverse edges to the existing,
  # selected edges
  if (add_to_selection) {
    edges_df <- data.frame(to = edges_to, from = edges_from)
    edges <-
      edf %>%
      dplyr::inner_join(edges_df, by = c("from", "to")) %>%
      dplyr::bind_rows(reverse_edges)
  } else {
    edges <- reverse_edges
  }

  # Modify `edges` to create a correct esdf
  edges <-
    edges %>%
    dplyr::select(edge = "id", "from", "to") %>%
    dplyr::arrange(edge)

  # Add the edge ID values to the active selection
  # of edges in `graph$edge_selection`
  graph$edge_selection <- edges

  # Replace `graph$node_selection` with an empty df
  graph$node_selection <- create_empty_nsdf()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Add one or several unconnected nodes to the graph
#'
#' @description
#'
#' Add `n` new nodes to a graph object of class `dgr_graph`. Optionally, set
#' node `type` values for the new nodes.
#'
#' @inheritParams node_edge_aes_data
#' @inheritParams render_graph
#' @param n The number of new nodes to add to the graph.
#' @param type An optional character vector that provides group identifiers for
#'   the nodes to be added.
#' @param label An optional character object that describes the nodes to be
#'   added.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph and
#' # add 5 nodes; these nodes
#' # will be assigned ID values
#' # from `1` to `5`
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(n = 5)
#'
#' # Get the graph's node IDs
#' graph %>% get_node_ids()
#'
#' @family node creation and removal
#'
#' @export
add_n_nodes <- function(
    graph,
    n,
    type = NULL,
    label = NULL,
    node_aes = NULL,
    node_data = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If `type` or `label` is NULL. Then NA
  type <- type %||% NA_character_
  label <- label %||% NA_character_

  # Collect node aesthetic attributes
  if (!is.null(node_aes)) {

    node_aes_tbl <- dplyr::as_tibble(node_aes)

    if (nrow(node_aes_tbl) < n) {

      node_aes$index__ <- seq_len(n)

      node_aes_tbl <-
        dplyr::as_tibble(node_aes) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_aes_tbl)) {
      node_aes_tbl$id <- NULL
    }
  }

  # Collect node data attributes
  if (!is.null(node_data)) {

    node_data_tbl <- dplyr::as_tibble(node_data)

    if (nrow(node_data_tbl) < n) {

      node_data$index__ <- seq_len(n)

      node_data_tbl <-
        dplyr::as_tibble(node_data) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_data_tbl)) {
      node_data_tbl$id <- NULL
    }
  }

  # Create a ndf of the correct length
  new_nodes <-
    create_node_df(
      n = n,
      type = type,
      label = label)

  # Add node aesthetics if available
  if (exists("node_aes_tbl")) {

    new_nodes <-
      new_nodes %>%
      dplyr::bind_cols(node_aes_tbl)
  }

  # Add node data if available
  if (exists("node_data_tbl")) {

    new_nodes <-
      new_nodes %>%
      dplyr::bind_cols(node_data_tbl)
  }

  new_nodes[, 1] <- new_nodes[, 1] + graph$last_node

  graph$nodes_df <-
    dplyr::bind_rows(graph$nodes_df, new_nodes)

  # Update the `last_node` counter
  graph$last_node <- graph$last_node + n

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = n)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

# Setting and getting node attributes

test_that("setting node attributes is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(n = 8)

  # Set attribute for node `1`
  graph_set_a <-
    set_node_attrs(
      graph = graph,
      nodes = 1,
      node_attr = value,
      values = 5)

  # Expect that node `1` has node attr set for `value`
  expect_equal(
    graph_set_a$nodes_df[
      which(graph_set_a$nodes_df$id == 1), 4], 5)

  # Set attribute for node `1` with a different value
  graph_set_a <-
    set_node_attrs(
      graph = graph,
      nodes = 1,
      node_attr = value,
      values = 8)

  # Expect that node `1` has node attr set for `value`
  expect_equal(
    graph_set_a$nodes_df[
      which(graph_set_a$nodes_df$id == 1), 4], 8)

  # Select node `1`
  graph_select_a <-
    select_nodes(
      graph = graph,
      nodes = 1)

  # Set attribute for selected node `1`
  graph_select_a <-
    set_node_attrs_ws(
      graph = graph_select_a,
      node_attr = value,
      value = 5)

  # Expect that node `1` has node attr set for `value`
  expect_equal(
    graph_select_a$nodes_df[
      which(graph_select_a$nodes_df$id == 1), 4], 5)

  # Set attribute for all nodes
  graph_set_all <-
    set_node_attrs(
      graph = graph,
      node_attr = value,
      values = 5)

  # Expect that all nodes have the attribute set
  expect_in(graph_set_all$nodes_df$value, 5)

  # Select node `1` and apply a node attribute
  # using that node selection
  graph_node_selection <-
    graph %>%
    select_nodes(nodes = 1) %>%
    set_node_attrs_ws(
      node_attr = value,
      value = 5)

  # Expect that node `1` has node attr set for `value`
  expect_equal(
    graph_node_selection$nodes_df[
      which(graph_node_selection$nodes_df[, 1] == 1), 4], 5)

  # Expect an error if the length of `value` is greater than 1
  expect_error(
    set_node_attrs(
      graph = graph,
      nodes = 1,
      node_attr = value,
      values = c(1, 2)))
})

test_that("setting edge attributes is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(n = 8)

  # Set edge attribute for edge `1`->`2`
  graph_set_a_1 <-
    set_edge_attrs(
      graph = graph,
      from = 1,
      to = 2,
      edge_attr = value,
      values = 5)

  # Expect that edge `1`->`2` has edge attr set for `value`
  expect_equal(
    graph_set_a_1$edges_df[
      which(graph_set_a_1$edges_df$from == 1 &
              graph_set_a_1$edges_df$to == 2), 5], 5)

  # Set attribute for named edge `1`->`2` with a different value
  graph_set_a_1 <-
    set_edge_attrs(
      graph = graph_set_a_1,
      from = 1,
      to = 2,
      edge_attr = value,
      values = 8)

  # Expect that edge `1`->`2` has edge attr set for `value`
  expect_equal(
    graph_set_a_1$edges_df[
      which(graph_set_a_1$edges_df$from == 1 &
              graph_set_a_1$edges_df$to == 2), 5], 8)

  # Select edge `1`->`2`
  graph_select_a_1 <-
    select_edges(
      graph = graph,
      from = 1,
      to = 2)

  # Set attribute for selected edge `1`->`2`
  graph_select_a_1 <-
    set_edge_attrs_ws(
      graph = graph_select_a_1,
      edge_attr = value,
      value = 5)

  # Expect that edge `1`->`2` has edge attr set for `value`
  expect_equal(
    graph_select_a_1$edges_df[
      which(graph_select_a_1$edges_df$from == 1 &
              graph_select_a_1$edges_df$to == 2), 5], 5)

  # Set attribute for all edges
  graph_set_all <-
    set_edge_attrs(
      graph = graph,
      edge_attr = value,
      values = 5)

  # Expect that all edges have the attribute set
  expect_in(graph_set_all$edges_df$value, 5)

  # Select edge `1`->`2` and apply an edge attribute using that
  # edge selection
  graph_edge_selection <-
    graph %>%
    select_edges(
      from = 1,
      to = 2) %>%
    set_edge_attrs_ws(
      edge_attr = value,
      value = 5)

  # Expect that edge `1`->`2` has edge attr set for `value`
  expect_equal(
    graph_edge_selection$edges_df[
      which(graph_edge_selection$edges_df$from == 1 &
              graph_edge_selection$edges_df$to == 2), 5], 5)
})

test_that("Getting node attributes is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 4,
      m = 4,
      node_data = node_data(
        value = 1:4),
      set_seed = 23)

  # Get node attributes for all nodes with the
  # `value` attribute
  all_nodes <-
    get_node_attrs(
      graph = graph,
      node_attr = value)

  # Expect a vector of integer values (length 4)
  expect_type(all_nodes, "integer")
  expect_length(all_nodes, 4)
  # Expect certain names to be in the vector

  expect_named(all_nodes, as.character(1:4))

  # Expect certain values to be in the vector
  expect_equal(
    all_nodes[[1]], 1)

  expect_equal(
    all_nodes[[2]], 2)

  expect_equal(
    all_nodes[[3]], 3)

  expect_equal(
    all_nodes[[4]], 4)

  # Get node attributes for nodes `1` and `3`
  nodes_1_3 <-
    get_node_attrs(
      graph = graph,
      node_attr = value,
      nodes = c(1, 3))

  # Expect the vector to have length 2 (named 1, 3)
  expect_length(nodes_1_3, 2)
  expect_named(nodes_1_3, c("1", "3"))

  # Expect certain values to be in the vector
  expect_equal(
    nodes_1_3[[1]], 1)

  expect_equal(
    nodes_1_3[[2]], 3)

  # Expect an error if referencing `id`
  expect_error(
    get_node_attrs(
      graph = graph,
      node_attr = id))
})

test_that("Getting edge attributes is possible", {

  # Create a simple graph where edges have an edge
  # attribute named `value`
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    {
      edges <-
        create_edge_df(
          from = c(1, 2, 1, 4),
            to = c(2, 3, 4, 3),
           rel = "rel")
      add_edge_df(
        graph = .,
        edge_df = edges)
    } %>%
    set_edge_attrs(
      edge_attr = value,
      values = 1.6,
      from = 1,
        to = 2) %>%
    set_edge_attrs(
      edge_attr = value,
      values = 4.3,
      from = 1,
        to = 4) %>%
    set_edge_attrs(
      edge_attr = value,
      values = 2.9,
      from = 2,
        to = 3) %>%
    set_edge_attrs(
      edge_attr = value,
      values = 8.4,
      from = 4,
        to = 3)

  # Get node attributes for all nodes with the
  # `value` attribute
  all_edges <-
    get_edge_attrs(
      graph = graph,
      edge_attr = value)

  # Expect a numeric vector
  expect_type(
    all_edges, "double")

  # Expect the vector to have length 4
  expect_length(all_edges, 4)

  # Expect certain names to be in the vector
  expect_contains(
    names(all_edges),
    c("1->2", "2->3", "1->4", "4->3")
    )

  # Expect certain values to be in the vector
  expect_equal(
    all_edges[[1]], 1.6)

  expect_equal(
    all_edges[[2]], 2.9)

  expect_equal(
    all_edges[[3]], 4.3)

  expect_equal(
    all_edges[[4]], 8.4)

  # Get only edge attribute values for specified
  # edges using the `from` and `to` arguments
  some_edges <-
    get_edge_attrs(
      graph = graph,
      edge_attr = value,
      from = c(1, 2),
        to = c(2, 3))

  # Expect the vector to have length 2
  expect_length(some_edges, 2)

  # Expect certain names to be in the vector
  expect_contains(
    names(some_edges),
    c("1->2", "2->3")
  )

  # Expect certain values to be in the vector
  expect_equal(
    some_edges[[1]], 1.6)

  expect_equal(
    some_edges[[2]], 2.9)

  expect_snapshot(error = TRUE, {
    # Expect an error if referencing `from`
    get_edge_attrs(graph = graph, edge_attr = from)
    # Expect an error if referencing `to`
    get_edge_attrs(graph = graph, edge_attr = to)
    # Expect an error if unequal vector lengths
    # provided for `from` and `to`
    get_edge_attrs(
      graph = graph,
      edge_attr = value,
      from = c(1, 2),
      to = 2)
  })
})

test_that("Getting node attributes with a selection is possible", {

  # Create a random graph and
  # select nodes `1` and `3`
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 4, m = 4,
      node_data = node_data(value = 1:4),
      set_seed = 23
    ) %>%
    select_nodes_by_id(nodes = c(1, 3))

  # Get node attributes for all
  # selected nodes with the
  # `value` attribute
  nodes_1_3 <-
    get_node_attrs_ws(
      graph = graph,
      node_attr = value)

  # Expect a vector of integer values
  expect_type(nodes_1_3, "integer")

  # Expect the vector to have length 2
  expect_length(nodes_1_3, 2)

  # Expect certain names to be in the vector
  expect_named(
    nodes_1_3,
    c("1", "3")
  )

  # Expect certain values to be in the vector
  expect_equal(
    nodes_1_3[[1]], 1)

  expect_equal(
    nodes_1_3[[2]], 3)

  # Expect an error if there is no
  # node selection
  expect_error(
    get_node_attrs_ws(
      graph = graph %>%
        clear_selection(),
      node_attr = value))

  # Expect an error if referencing
  # `id` column
  expect_error(
    get_node_attrs_ws(
      graph = graph,
      node_attr = id))
})

test_that("Getting edge attributes with a selection is possible", {

  # Create a simple graph where
  # edges have an edge attribute
  # named `value`; select the edges
  # defined as `1`->`3` and `2`->`3`
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    {
      edges <-
        create_edge_df(
          from = c(1, 2, 1, 4),
          to = c(2, 3, 4, 3),
          rel = "rel")
      add_edge_df(
        graph = .,
        edge_df = edges)
    } %>%
    set_edge_attrs(
      edge_attr = value,
      values = 1.6,
      from = 1,
        to = 2) %>%
    set_edge_attrs(
      edge_attr = value,
      values = 4.3,
      from = 1,
        to = 4) %>%
    set_edge_attrs(
      edge_attr = value,
      values = 2.9,
      from = 2,
        to = 3) %>%
    set_edge_attrs(
      edge_attr = value,
      values = 8.4,
      from = 4,
        to = 3) %>%
    select_edges(
      from = c(1, 2),
      to = c(2, 3))

  # Get edge attributes for all
  # selected edge with the
  # `value` attribute
  edge_attr_values <-
    graph %>%
    get_edge_attrs_ws(
      edge_attr = value)

  # Expect a numeric vector
  expect_type(
    edge_attr_values, "double")

  # Expect the vector to have length 2
  expect_length(edge_attr_values, 2)

  # Expect certain names to be in the vector
  # expect_contains == expect_true(all(expected %in% object))
  expect_named(
    edge_attr_values,
    c("1->2", "2->3"),
    ignore.order = TRUE
    )

  # Expect certain values to be in the vector
  expect_equal(
    edge_attr_values[[1]], 1.6)

  expect_equal(
    edge_attr_values[[2]], 2.9)

  # Expect an error if there is no
  # edge selection
  expect_error(
    get_edge_attrs_ws(
      graph = graph %>%
        clear_selection(),
      edge_attr = value))
})

# Adding nodes and/or edges to an existing graph object

test_that("adding a node to a graph is possible", {

  # Create a graph with two nodes
  graph <-
    create_graph() %>%
    add_n_nodes(n = 2)

  # Expect that names in this graph object match a
  # prescribed set of names
  expect_named(
    graph,
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  # Expect a graph object of class `dgr_graph`
  expect_s3_class(
    graph, "dgr_graph")

  # Expect that the `global_attrs` components is not NULL
  expect_false(is.null(graph$global_attrs))

  # Expect that the `nodes_df` and  `edges_df`  component are a data frame
  expect_s3_class(graph$nodes_df, "data.frame")
  expect_s3_class(graph$edges_df, "data.frame")

  # Expect that the graph is a directed graph
  expect_true(graph$directed)

  # Expect that the `nodes_df` data frame has 3 columns
  expect_equal(
    ncol(graph$nodes_df), 3)

  # Expect that the `nodes_df` data frame has 2 rows
  expect_equal(
    nrow(graph$nodes_df), 2)

  # Add a node with attributes to the graph
  graph_3 <-
    add_node(
      graph = graph,
      type = "fresh")

  # Expect that there will be 3 nodes in the graph
  expect_equal(
    count_nodes(graph = graph_3), 3)

  # Expect that the "type" value will be present for
  # the node in the new graph
  expect_equal(
    get_node_attrs(
      graph = graph_3,
      node_attr = type,
      node = 3) %>% unname(),
    "fresh")

  # Expect that the other nodes in the graph will
  # still have unassigned `type` values
  expect_true(
    is.na(get_node_attrs(
      graph = graph_3,
      node_attr = type,
      node = 1) %>% unname()))

  expect_true(
    is.na(get_node_attrs(
      graph = graph_3,
      node_attr = type,
      node = 2) %>% unname()))

  # Create an empty graph
  graph_unlabeled <- create_graph()

  # Add a single, unlabeled node to the graph
  graph_unlabeled <- add_node(graph_unlabeled)

  # Expect that the graph will have one unlabeled node
  expect_true(
    is.na(graph_unlabeled$nodes_df$label))

  # Add a node to the graph that is joined from another
  graph_from <-
    add_node(
      graph = graph_3,
      from = 3)

  # Expect that `edges_df` is not NULL
  expect_false(
    is.null(graph_from$edges_df))

  # Expect that the new node is available in the graph
  expect_true(
    is_node_present(
      graph = graph_from,
      node = 4))

  # Expect that the edge from `3` to `4` is in the graph
  expect_true(
    is_edge_present(
      graph = graph_from,
      from = 3,
      to = 4))

  # Expect that the node label is NA for node `4`
  expect_true(
    is.na(graph_from$nodes_df$label[4]))

  # Expect that for node `4`, the `type` is not set
  # since the default value for `type` is NULL
  expect_true(
    is.na(get_node_attrs(
      graph = graph_from,
      node_attr = type,
      node = 4) %>% unname()))

  # Add a node to the graph that is joined to another
  graph_to <-
    add_node(
      graph = graph_3,
      to = 3)

  # Expect that `edges_df` is not NULL
  expect_false(
    is.null(graph_to$edges_df))

  # Expect that the new node is available in the graph
  expect_true(
    is_node_present(
      graph = graph_to,
      node = 4))

  # Expect that the edge from `4` to `3` is in the graph
  expect_true(
    is_edge_present(
      graph = graph_to,
      from = 4,
      to = 3))

  # Expect that the node label for all nodes is NA
  expect_true(
    all(is.na(graph_to$nodes_df$label)))

  # Expect that for node `4`, the `type` is not set
  # since the default value for `type` is NULL
  expect_true(
    is.na(get_node_attrs(
      graph = graph_to,
      node_attr = type,
      node = 4) %>% unname()))

  # Add a node to the graph that is joined from
  # another and to another
  graph_to_from <-
    add_node(
      graph = graph_3,
      from = 1,
      to = 2)

  # Expect that `edges_df` is not NULL
  expect_false(
    is.null(graph_to_from$edges_df))

  # Expect that the new node is available in the graph
  expect_true(
    is_node_present(
      graph = graph_to_from,
      node = 4))

  # Expect that the edge from `1` to `4` is in the graph
  expect_true(
    is_edge_present(
      graph = graph_to_from,
      from = 1,
      to = 4))

  # Expect that the edge from `4` to `2` is in the graph
  expect_true(
    is_edge_present(
      graph = graph_to_from,
      from = 4,
      to = 2))

  # Expect that for node `4`, the `type` is not set
  # since the default value for `type` is NULL
  expect_true(
    is.na(get_node_attrs(
      graph = graph_to_from,
      node_attr = type,
      node = 4) %>% unname()))

  # Create a graph with one node
  graph <-
    create_graph() %>%
    add_node()

  # Add another node, connecting with only a value
  # provided as `from` but where the referenced node
  # is not in the graph
  expect_error(
    add_node(
      graph = graph,
      from = 3))
})

test_that("adding an edge to a graph is possible", {

  # Create a graph with two nodes
  graph <-
    create_graph() %>%
    add_n_nodes(n = 2)

  # Add an edge
  graph <-
    add_edge(
      graph = graph,
      from = 1,
      to = 2,
      rel = "to_get")

  # Expect that names in this graph object match a
  # prescribed set of names
  expect_named(
    graph,
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  # Expect a graph object of class `dgr_graph`
  expect_s3_class(
    graph, "dgr_graph")

  # Expect that the `global_attrs` components is not NULL
  expect_false(
    is.null(graph$global_attrs))

  # Expect that the `nodes_df` component is a data frame
  expect_s3_class(
    graph$nodes_df, "data.frame")

  # Expect that the `edges_df` component is a data frame
  expect_s3_class(
    graph$edges_df, "data.frame")

  # Expect that the graph is a directed graph
  expect_true(graph$directed)

  # Expect that the `nodes_df` data frame has 1 row and 4 columns
  expect_equal(dim(graph$edges_df),  c(1, 4))

  # Expect an error when attempting to add more than 1 edge
  expect_error(
    add_edge(graph, from = c(1, 2), to = 2))

  # Expect an error when calling `add_edge()` on an empty graph
  expect_error(
    add_edge(graph = create_graph()))

  # Create a graph with 4 nodes and labels for each
  graph_2 <-
    create_graph() %>%
    add_node(label = "one") %>%
    add_node(label = "two") %>%
    add_node(label = "three") %>%
    add_node(label = "four")

  # Add two edges to the graph using the node
  # `label` values
  graph_2 <-
    graph_2 %>%
    add_edge(
      from = "three",
      to = "four",
      rel = "L") %>%
    add_edge(
      from = "four",
      to = "one",
      rel = "L")

  # Expect that 2 edges are available in the graph
  expect_equal(
    count_edges(graph = graph_2), 2)

  # Expect that the edges in the graph are:
  # `3->4` and `4->1`
  expect_contains(
    get_edges(graph_2),
    c("3->4", "4->1")
  )
})

test_that("adding several nodes to a graph at once is possible", {

  # Create a graph with 10 nodes
  graph <-
    create_graph() %>%
    add_n_nodes(n = 10)

  # Expect that 10 nodes were added to the empty graph
  expect_equal(
    count_nodes(graph = graph), 10)

  # Expect monotonically-increasing node ID values
  # from 1 to 10
  expect_equal(
    get_node_ids(graph),
    c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))

  # Expect that no `type` values have been set
  expect_equal(
    get_node_df(graph)[, 2],
    rep(NA_character_, 10))

  # Expect that no `label` values have been set
  expect_equal(
    get_node_df(graph)[, 3],
    rep(NA_character_, 10))

  # Create a graph with 10 nodes of a specified type
  graph <- create_graph()

  graph <-
    add_n_nodes(
      graph = graph,
      n = 10,
      type = "test_node")

  # Expect that 10 nodes were added to the empty graph
  expect_equal(
    count_nodes(graph = graph), 10)

  # Expect monotonically-increasing node ID values
  # from `1` to `10`
  expect_equal(
    get_node_ids(graph),
    seq(1, 10))

  # Expect that a `type` value have been set
  # for all nodes
  expect_equal(
    get_node_df(graph)[, 2],
    rep("test_node", 10))

  # Expect that `label` values have not been set
  expect_equal(
    get_node_df(graph)[, 3],
    rep(NA_character_, 10))
})

test_that("adding several nodes to a graph at once (with extra attrs) is possible", {

  # Create a graph with 10 nodes
  graph <-
    create_graph() %>%
    add_n_nodes(n = 10)

  # Expect that 10 nodes were added to the empty graph
  expect_equal(
    count_nodes(graph = graph), 10)

  # Expect monotonically-increasing node ID values
  # from 1 to 10
  expect_equal(
    get_node_ids(graph),
    c(1, 2, 3, 4, 5, 6, 7, 8, 9, 10))

  # Expect that no `type` values have been set
  expect_equal(
    get_node_df(graph)[, 2],
    rep(NA_character_, 10))

  # Expect that no `label` values have been set
  expect_equal(
    get_node_df(graph)[, 3],
    rep(NA_character_, 10))
})

test_that("adding several nodes from a selected node is possible", {

  # Create a graph with 10 nodes
  graph <-
    create_graph() %>%
    add_n_nodes(n = 10)

  # Select the node with ID of `5`
  graph <-
    select_nodes(
      graph = graph,
      nodes = 5)

  # Add 10 nodes as successors to the selected node
  graph <-
    add_n_nodes_ws(
      graph = graph,
      n = 10,
      direction = "from")

  # Expect a total of 20 nodes in the graph
  expect_equal(
    count_nodes(graph = graph), 20)

  # Expect monotonically-increasing node ID
  # values from `1` to `20`
  expect_equal(
    get_node_ids(graph), seq(1, 20))

  # Expect a total of 10 edges in the graph
  expect_equal(
    count_edges(graph = graph), 10)

  # Expect that node IDs where edges are `from` belong
  # to the node with ID of `5`
  expect_equal(
    graph$edges_df$from, rep(5, 10))

  # Expect that node IDs where edges are `to` increase
  # from `11` to `20`
  expect_equal(
    graph$edges_df$to, seq(11, 20))

  # Expect that the edge relationship has not been set
  # for any of the edges
  expect_equal(
    graph$edges_df$rel,
    rep(NA_character_, 10))

  # Create another empty graph
  graph <- create_graph()

  # Add 10 nodes to the empty graph
  graph <-
    add_n_nodes(
      graph = graph,
      n = 10)

  # Select the node with ID of `5`
  graph <-
    select_nodes(
      graph = graph,
      nodes = 5)

  # Add 10 nodes as successors to the selected node,
  # this time with a node `type` and edge `rel` set for
  # each of the new nodes and edges
  graph <-
    add_n_nodes_ws(
      graph = graph,
      n = 10,
      direction = "from",
      type = "new",
      rel = "related")

  # Expect that all edges have a `rel` set
  expect_in(
    graph$edges_df$rel,
    "related"
  )

  # Expect that all new nodes have a `type` set
  expect_in(
    get_node_df(graph)[11:20, 2],
    "new"
  )
})

test_that("adding several nodes to a selected node is possible", {

  # Create a graph with 10 nodes
  graph <-
    create_graph() %>%
    add_n_nodes(n = 10)

  # Select the node with ID of `5`
  graph <-
    select_nodes(
      graph = graph,
      nodes = 5)

  # Add 10 nodes as predecessors to the selected node
  graph <-
    add_n_nodes_ws(
      graph = graph,
      n = 10,
      direction = "to")

  # Expect a total of 20 nodes in the graph
  expect_equal(
    count_nodes(graph = graph), 20)

  # Expect monotonically-increasing node ID values
  # from `1` to `20`
  expect_equal(
    get_node_ids(graph),
    seq(1, 20))

  # Expect a total of 10 edges in the graph
  expect_equal(
    count_edges(graph = graph), 10)

  # Expect that node IDs where edges are `to` belong
  # to the node with ID of `5`
  expect_equal(
    graph$edges_df$to, rep(5, 10))

  # Expect that node IDs where edges are `from`
  # increase from `11` to `20`
  expect_equal(
    graph$edges_df$from, seq(11, 20))

  # Expect that the edge relationship has not been set
  # for any of the edges
  expect_equal(
    graph$edges_df$rel,
    rep(NA_character_, 10))

  # Create another empty graph
  graph <- create_graph()

  # Add 10 nodes to the empty graph
  graph <-
    add_n_nodes(
      graph = graph,
      n = 10)

  # Select the node with ID of `5`
  graph <-
    select_nodes(
      graph = graph,
      nodes = 5)

  # Add 10 nodes as predecessors to the selected node,
  # this time with a node `type` and edge `rel` set for
  # each of the new nodes and edges
  graph <-
    add_n_nodes_ws(
      graph = graph,
      n = 10,
      direction = "to",
      type = "new",
      rel = "related")

  # Expect that all edges have a `rel` set
  expect_in(
    graph$edges_df$rel, "related")

  # Expect that all new nodes have a `type` set
  expect_in(
    graph$nodes_df$type[11:20], "new")
})

test_that("adding several edges with a string is possible", {

  # Create a graph with 10 nodes
  graph <-
    create_graph() %>%
    add_n_nodes(n = 10)

  # Add edges via a string
  graph <-
    add_edges_w_string(
      graph = graph,
      edges = "1->2 1->3 2->4 2->5 3->6 3->7 4->8 4->9 5->10")

  # Expect a total of 10 nodes in the graph
  expect_equal(
    count_nodes(graph = graph), 10)

  # Expect monotonically-increasing node ID values
  # from 1 to 10
  expect_equal(
    get_node_ids(graph), seq(1, 10))

  # Expect a total of 9 edges in the graph
  expect_equal(
    count_edges(graph = graph), 9)

  # Expect that the edge relationship has not been set
  # for any of the edges
  expect_equal(
    graph$edges_df$rel,
    rep(NA_character_, 9))

  # Create another empty graph
  graph <- create_graph()

  # Add 10 nodes to the empty graph
  graph <-
    add_n_nodes(
      graph = graph,
      n = 10)

  # Add edges via a string and add the rel value
  # `connected_to` to all new edges
  graph <-
    add_edges_w_string(
      graph = graph,
      edges = "1->2 1->3 2->4 2->5 3->6 3->7 4->8 4->9 5->10",
      rel = "connected_to")

  # Expect that the edge relationship `connected_to` is
  # set for all edges
  expect_equal(
    graph$edges_df$rel,
    rep("connected_to", 9))

  # Create an empty graph, set as `undirected`
  graph <- create_graph(directed = FALSE)

  # Add 10 nodes to the empty graph
  graph <-
    add_n_nodes(
      graph = graph,
      n = 10)

  # Add edges via a string and add the rel value
  # `connected_to` to all new edges
  graph <-
    add_edges_w_string(
      graph = graph,
      edges = "1--2 1--3 2--4 2--5 3--6 3--7 4--8 4--9 5--10",
      rel = "connected_to")

  # Expect a total of 9 edges in the graph
  expect_equal(
    count_edges(graph = graph), 9)

  # Create a graph with 4 nodes and 4 distinct labels
  graph <-
    create_graph() %>%
    add_node(label = "one") %>%
    add_node(label = "two") %>%
    add_node(label = "three") %>%
    add_node(label = "four")

  # Add edges between nodes using a
  # character string with node label values
  # and setting `use_labels = TRUE`
  graph_node_label <-
    graph %>%
    add_edges_w_string(
      edges = "one->two one->three
               two->four two->three",
      use_labels = TRUE)

  # Expect a total of 4 edges in the graph
  expect_equal(
    count_edges(graph = graph_node_label), 4)

  # Expect certain edges to be in the graph
  expect_equal(
    get_edges(graph_node_label),
    c("1->2", "1->3", "2->4", "2->3"))
})

test_that("adding node clones is possible", {

  # Create a graph with a path of
  # nodes having non-NA `label`, `type`,
  # and `value` node attributes
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      label = c("d", "g", "r"),
      type = c("a", "b", "c"))

  # Create 3 clones of node `1`
  # and assign new node label
  # values
  graph_1 <-
    graph %>%
    add_n_node_clones(
      n = 3,
      node = 1,
      label = c("x", "y", "z"))

  # Get the graph's node data frame
  ndf_1 <-
    graph_1 %>%
    get_node_df()

  # Expect that the last three nodes
  # will be have the same `type` value
  # as node `1`
  expect_equal(
    unique(ndf_1[c(1, 4:6), 2]), "a")

  # Expect that the new nodes will
  # have `label` values that were
  # specified in the function call
  expect_equal(
    ndf_1[4:6, 3], c("x", "y", "z"))

  # Create 3 clones of node `1`
  # and don't assign `label` values
  # (have them as NA values)
  graph_2 <-
    graph %>%
    add_n_node_clones(
      n = 3,
      node = 1,
      label = NULL)

  # Get the graph's node data frame
  ndf_2 <-
    graph_2 %>%
    get_node_df()

  # Expect that the last three nodes
  # will be have the same `type` value
  # as node `1`
  expect_equal(
    unique(ndf_2[c(1, 4:6), 2]), "a")

  # Expect that the new nodes will
  # have NA `label` values
  expect_equal(
    ndf_2[4:6, 3],
    rep(NA_character_, 3))
})

test_that("adding node clones with a selection is possible", {

  # Create a graph with a path of
  # nodes; supply `label`, `type`,
  # and `value` node attributes,
  # and select the created nodes
  graph <-
    create_graph() %>%
    add_path(
      n = 3,
      label = c("d", "g", "r"),
      type = c("a", "b", "c")) %>%
    select_last_nodes_created()

  # Create clones of all nodes
  # in the selection
  graph_1 <-
    graph %>%
    add_node_clones_ws()

  # Expect 6 nodes in the graph
  expect_equal(
    count_nodes(graph_1), 6)

  # Expect nodes with ID values
  # from 1 to 6
  expect_equal(
    get_node_df(graph_1)[, 1],
    1:6)

  # Expect cloned values for the
  # `type` attribute for the
  # last 3 nodes in the graph
  expect_equal(
    get_node_df(graph_1)[4:6, 2],
    c("a", "b", "c"))

  # Expect NA values for the
  # `label` attributes for the
  # last 3 nodes in the graph
  expect_true(
    all(is.na(get_node_df(graph_1)[4:6, 3]))
    )

  # Create clones of all nodes
  # in the selection but assign
  # new node label values
  graph_2 <-
    graph %>%
    add_node_clones_ws(
      label = c("a", "b", "v"))

  # Expect 6 nodes in the graph
  expect_equal(
    count_nodes(graph_2), 6)

  # Expect nodes with ID values
  # from 1 to 6
  expect_equal(
    get_node_df(graph_2)[, 1], 1:6)

  # Expect cloned values for the
  # `type` attribute for the
  # last 3 nodes in the graph
  expect_equal(
    get_node_df(graph_2)[4:6, 2],
    c("a", "b", "c"))

  # Expect specific values for the
  # `label` attributes for the
  # last 3 nodes in the graph
  expect_equal(
    get_node_df(graph_2)[4:6, 3],
    c("a", "b", "v"))

  # Create clones of all nodes
  # in the selection and
  # add edges from the orginal
  # nodes to the cloned nodes
  graph_3 <-
    graph %>%
    add_node_clones_ws(
      add_edges = TRUE,
      direction = "to")

  # Expect 6 nodes in the graph
  expect_equal(
    count_nodes(graph_3), 6)

  # Expect nodes with ID values
  # from 1 to 6
  expect_equal(
    get_node_df(graph_3)[, 1], 1:6)

  # Expect cloned values for the
  # `type` attribute for the
  # last 3 nodes in the graph
  expect_equal(
    get_node_df(graph_3)[4:6, 2],
    c("a", "b", "c"))

  # Expect NA values for the
  # `label` attributes for the
  # last 3 nodes in the graph
  expect_true(
    all(is.na(get_node_df(graph_3))[4:6, 3])
    )

  # Expect 5 edges in the graph
  expect_equal(
    count_edges(graph = graph_3), 5)

  # Expect edges with ID values
  # from 1 to 5
  expect_equal(
    get_edge_df(graph_3)[, 1],
    1:5)

  # Expect that specific edges have
  # been created
  expect_equal(
    get_edges(graph_3)[3:5],
    c("1->4", "2->5", "3->6")
    )

  # Expect NA values for the
  # `rel` attributes for the
  # last 3 edges in the graph
  expect_true(
    all(is.na(get_edge_df(graph_3)[3:5, 4]))
    )

  # Create clones of all nodes
  # in the selection and
  # add edges to the orginal
  # nodes from the cloned nodes
  graph_4 <-
    graph %>%
    add_node_clones_ws(
      add_edges = TRUE,
      direction = "from")

  # Expect 6 nodes in the graph
  expect_equal(
    count_nodes(graph_4), 6)

  # Expect nodes with ID values
  # from 1 to 6
  expect_equal(
       get_node_df(graph_4)[, 1], 1:6)

  # Expect cloned values for the
  # `type` attribute for the
  # last 3 nodes in the graph
  expect_equal(
    get_node_df(graph_4)[4:6, 2],
    c("a", "b", "c"))

  # Expect NA values for the
  # `label` attributes for the
  # last 3 nodes in the graph
  expect_true(
    all(is.na(get_node_df(graph_4))[4:6, 3]))

  # Expect 5 edges in the graph
  expect_equal(
    count_edges(graph = graph_4), 5)

  # Expect edges with ID values
  # from 1 to 5
  expect_equal(
    get_edge_df(graph_4)[, 1], 1:5)

  # Expect that specific edges have
  # been created
  expect_equal(
    get_edges(graph_4)[3:5],
    c("4->1", "5->2", "6->3"))

  # Expect NA values for the
  # `rel` attributes for the
  # last 3 edges in the graph
  expect_true(
    all(is.na(get_edge_df(graph_4)[3:5, 4]))
    )
})

test_that("adding edge clones is possible", {

  # Create a graph with a path of
  # 2 nodes; supply a common `rel`
  # edge attribute for all edges
  # in this path and then add a
  # `color` edge attribute
  graph <-
    create_graph() %>%
    add_path(
      n = 2,
      rel = "a") %>%
    select_last_edges_created() %>%
    set_edge_attrs(
      edge_attr = color,
      values = "steelblue") %>%
    clear_selection() %>%
    add_node()

  # Create an edge clone (new edge
  # with attributes from edge `1`)
  graph_1 <-
    graph %>%
    add_edge_clone(
      edge = 1,
      from = 3,
        to = 1)

  # Get the graph's edge data frame
  edf_1 <-
    graph_1 %>%
    get_edge_df()

  # Expect 2 edges in the edf
  expect_equal(
    nrow(edf_1), 2)

  # Expect that the new edge has
  # definition `3` -> `1`
  expect_equal(
    c(edf_1[2, 2], edf_1[2, 3]),
    c(3, 1))

  # Expect that the new edge has
  # edge ID `2`
  expect_equal(
    edf_1[2, 1], 2)

  # Expect that both edges will be have
  # the same `rel` values
  expect_in(edf_1[, 4], "a")

  # Expect that both edges will be have
  # the same `color` values
  expect_in(edf_1[, 5], "steelblue")
})

# Adding forward and reverse edges through a selection

test_that("forward edges can be added given a selection of edges", {
  # Create an empty graph, add 2 nodes to it,
  # and create the edge `1->2`
  graph <-
    create_graph() %>%
    add_n_nodes(
      n = 2,
      type = "type_a",
      label = c("a_1", "a_2")) %>%
    add_edge(
      from = 1,
      to = 2,
      rel = "a")

  # Select the edge and create 2 additional edges
  # with the same definition (`1->2`) but with
  # different `rel` values (`b` and `c`)
  graph <-
    graph %>%
    select_edges() %>%
    add_forward_edges_ws("b") %>%
    add_forward_edges_ws("c") %>%
    clear_selection()

  # Expect that there are 3 edges in the graph
  expect_equal(
    count_edges(graph = graph), 3)

  # Expect edge `id` values of `1`, `2`, and `3`
  expect_equal(
    graph$edges_df$id,
    c(1, 2, 3))

  # Expect the edges have the same definition
  # of `1->2`
  expect_in(
    get_edges(graph),
    "1->2")

  # Expect `rel` values of `a`, `b`, and `c`
  expect_equal(
    graph$edges_df$rel,
    c("a", "b", "c"))
})

test_that("reverse edges can be added given a selection of edges", {

  # Create an empty graph, add 2 nodes to it,
  # and create the edge `1->2`
  graph <-
    create_graph() %>%
    add_n_nodes(
      n = 2,
      type = "type_a",
      label = c("a_1", "a_2")) %>%
    add_edge(
      from = 1,
      to = 2,
      rel = "a")

  # Select the edge and create 2 additional edges
  # with the oppossite definition (`2->1`) and with
  # different `rel` values (`b` and `c`)
  graph <-
    graph %>%
    select_edges() %>%
    add_reverse_edges_ws("b") %>%
    add_reverse_edges_ws("c") %>%
    clear_selection()

  # Expect that there are 3 edges in the graph
  expect_equal(
    count_edges(graph = graph), 3)

  # Expect edge `id` values of `1`, `2`, and `3`
  expect_equal(
    graph$edges_df$id, c(1, 2, 3))

  # Expect the last 2 edges have the definition
  # of `2->1` whereas the original edge is `1->2`
  expect_equal(
    get_edges(graph),
    c("1->2", "2->1", "2->1"))

  # Expect `rel` values of `a`, `b`, and `c`
  expect_equal(
    graph$edges_df$rel,
    c("a", "b", "c"))

  # Expect an error if the graph does not
  # have an active selection of edges
  expect_snapshot(error = TRUE,
    create_graph() %>%
      add_n_nodes(
        n = 2,
        type = "type_a",
        label = c("a_1", "a_2")) %>%
      add_edge(
        from = 1,
        to = 2,
        rel = "a") %>%
      add_reverse_edges_ws("b"))
})

# Defining node and edge aesthetics and data attributes")

test_that("creating node data attribute values is possible", {

  # Create a node data list object
  # using the `node_data()` function
  node_data <-
    node_data(
      values_1 = c(1, 2, 4, 5),
      values_2 = c("one", "two", "four", "five"))

  # Expect that a list object has
  # been generated
  expect_type(
    node_data, "list")

  # Expect specific names in the object
  expect_named(
    node_data,
    c("values_1", "values_2"))

  # Expect specific values in each
  # of the list components
  expect_equal(
    node_data[[1]],
    c(1, 2, 4, 5))

  expect_equal(
    node_data[[2]],
    c("one", "two", "four", "five"))

  # Expect the function to return
  # an error when supplying data
  # attributes that are also node
  # aesthetic attributes
  expect_error(
    node_data(
      penwidth = c(1, 2, 4, 5)))

  # Expect an error if using the
  # attributes named `x` or `y`
  expect_error(
    node_data(
      x = c(1, 2, 4, 5)))

  expect_error(
    node_data(
      y = c(1, 2, 4, 5)))
})

test_that("creating edge data attribute values is possible", {

  # Create a edge data list object
  # using the `edge_data()` function
  edge_data <-
    edge_data(
      values_1 = c(1, 2, 4, 5),
      values_2 = c("one", "two", "four", "five"))

  # Expect that a list object has
  # been generated
  expect_type(
    edge_data, "list")

  # Expect specific names in the object
  expect_named(
    edge_data,
    c("values_1", "values_2"))

  # Expect specific values in each
  # of the list components
  expect_equal(
    edge_data[[1]],
    c(1, 2, 4, 5))

  expect_equal(
    edge_data[[2]],
    c("one", "two", "four", "five"))

  # Expect the function to return
  # an error when supplying data
  # attributes that are also edge
  # aesthetic attributes
  expect_error(
    edge_data(
      penwidth = c(1, 2, 4, 5)))
})

test_that("creating node aesthetic attribute values is possible", {

  # Create a node aesthetic list object
  # using the `node_aes()` function
  node_aes <-
    node_aes(
      shape = c("circle", "rectangle"),
      style = c("filled", "filled"),
      penwidth = c(2.3, 5.3),
      color = c("red", "#FFFFFF"),
      fillcolor = c("orange", "#000000"),
      fontname = c("Helvetica", "Helvetica"),
      fontsize = c(12, 10.5),
      fontcolor = c("gray25", "#000000"),
      peripheries = c(1, 2),
      height = c(6.2, 5),
      width = c(5.5, 6.0),
      x = c(2.1, 5.3),
      y = c(2, 4.7),
      group = c(1, 1),
      tooltip = c("tooltip_1", "tooltip_2"),
      xlabel = c("xlabel_1", "xlabel_2"),
      URL = c("http://www.website.net", "http:://www.internet.org"),
      sides = c(4, 5),
      orientation = c(0, 45.5),
      skew = c(0.4, -0.8),
      distortion = c(-0.7, 0.3),
      gradientangle = c(1.2, 45.0),
      fixedsize = c(TRUE, FALSE),
      labelloc = c("T", "C"),
      margin = c("0.5,0.01", "0.1,0.05"))

  # Expect that a list object has
  # been generated
  expect_type(
    node_aes, "list")

  # Expect specific names in the object
  expect_named(
    node_aes,
    c("shape", "style", "penwidth", "color",
      "fillcolor", "fontname", "fontsize",
      "fontcolor", "peripheries", "height",
      "width", "x", "y", "group", "tooltip",
      "xlabel", "URL", "sides", "orientation",
      "skew", "distortion", "gradientangle",
      "fixedsize", "labelloc", "margin"))

  # Expect specific values in each
  # of the list components
  expect_equal(node_aes$shape, c("circle", "rectangle"))
  expect_equal(node_aes$style, c("filled", "filled"))
  expect_equal(node_aes$penwidth, c(2.3, 5.3))
  expect_equal(node_aes$color, c("red", "#FFFFFF"))
  expect_equal(node_aes$fillcolor, c("orange", "#000000"))
  expect_equal(node_aes$fontname, c("Helvetica", "Helvetica"))
  expect_equal(node_aes$fontsize, c(12, 10.5))
  expect_equal(node_aes$fontcolor, c("gray25", "#000000"))
  expect_equal(node_aes$peripheries, c(1, 2))
  expect_equal(node_aes$height, c(6.2, 5))
  expect_equal(node_aes$width, c(5.5, 6.0))
  expect_equal(node_aes$x, c(2.1, 5.3))
  expect_equal(node_aes$y, c(2, 4.7))
  expect_equal(node_aes$group, c(1, 1))
  expect_equal(node_aes$tooltip, c("tooltip_1", "tooltip_2"))
  expect_equal(node_aes$xlabel, c("xlabel_1", "xlabel_2"))
  expect_equal(node_aes$URL, c("http://www.website.net", "http:://www.internet.org"))
  expect_equal(node_aes$sides, c(4, 5))
  expect_equal(node_aes$orientation, c(0, 45.5))
  expect_equal(node_aes$skew, c(0.4, -0.8))
  expect_equal(node_aes$distortion, c(-0.7, 0.3))
  expect_equal(node_aes$gradientangle, c(1.2, 45.0))
  expect_equal(node_aes$fixedsize, c(TRUE, FALSE))
  expect_equal(node_aes$labelloc, c("T", "C"))
  expect_equal(node_aes$margin, c("0.5,0.01", "0.1,0.05"))

  # Expect the function to return
  # an error when supplying any
  # argument name that is not an
  # aesthetic attribute
  expect_error(
    node_aes(
      value_1 = c(1, 2, 4, 5)))
})

test_that("creating edge aesthetic attribute values is possible", {

  # Create an edge aesthetic list object
  # using the `edge_aes()` function
  edge_aes <-
    edge_aes(
      style = c("bold", "dashed"),
      penwidth = c(1.0, 1.5),
      color = c("red", "#FFFFFF"),
      arrowsize = c(0.5, 1.0),
      arrowhead = c("vee", "tee"),
      arrowtail = c("normal", "dot"),
      fontname = c("Helvetica", "Helvetica"),
      fontsize = c(12, 10.5),
      fontcolor = c("gray25", "#000000"),
      len = c(1.0, 0.5),
      tooltip = c("tooltip_1", "tooltip_2"),
      URL = c("http://www.website.net", "http:://www.internet.org"),
      label = c("label_1", "label_2"),
      labelfontname = c("Helvetica", "Helvetica"),
      labelfontsize = c(12, 10.5),
      labelfontcolor = c("gray25", "#000000"),
      labeltooltip = c("tooltip_1", "tooltip_2"),
      labelURL = c("http://www.website.net", "http:://www.internet.org"),
      edgetooltip = c("tooltip_1", "tooltip_2"),
      edgeURL = c("http://www.website.net", "http:://www.internet.org"),
      dir = c("forward", "back"),
      headtooltip = c("tooltip_1", "tooltip_2"),
      headURL = c("http://www.website.net", "http:://www.internet.org"),
      headclip = c(TRUE, FALSE),
      headlabel = c("label_1", "label_2"),
      headport = c("n", "ne"),
      tailtooltip = c("tooltip_1", "tooltip_2"),
      tailURL = c("http://www.website.net", "http:://www.internet.org"),
      tailclip = c(TRUE, FALSE),
      taillabel = c("label_1", "label_2"),
      tailport = c("s", "nw"),
      decorate = c(FALSE, TRUE))

  # Expect that a list object has
  # been generated
  expect_type(edge_aes, "list")

  # Expect specific values in each
  # of the list components
  expect_equal(edge_aes$style, c("bold", "dashed"))
  expect_equal(edge_aes$penwidth, c(1.0, 1.5))
  expect_equal(edge_aes$color, c("red", "#FFFFFF"))
  expect_equal(edge_aes$arrowsize, c(0.5, 1.0))
  expect_equal(edge_aes$arrowhead, c("vee", "tee"))
  expect_equal(edge_aes$arrowtail, c("normal", "dot"))
  expect_equal(edge_aes$fontname, c("Helvetica", "Helvetica"))
  expect_equal(edge_aes$fontsize, c(12, 10.5))
  expect_equal(edge_aes$fontcolor, c("gray25", "#000000"))
  expect_equal(edge_aes$len, c(1.0, 0.5))
  expect_equal(edge_aes$tooltip, c("tooltip_1", "tooltip_2"))
  expect_equal(edge_aes$URL, c("http://www.website.net", "http:://www.internet.org"))
  expect_equal(edge_aes$label, c("label_1", "label_2"))
  expect_equal(edge_aes$labelfontname, c("Helvetica", "Helvetica"))
  expect_equal(edge_aes$labelfontsize, c(12, 10.5))
  expect_equal(edge_aes$labelfontcolor, c("gray25", "#000000"))
  expect_equal(edge_aes$labeltooltip, c("tooltip_1", "tooltip_2"))
  expect_equal(edge_aes$labelURL, c("http://www.website.net", "http:://www.internet.org"))
  expect_equal(edge_aes$edgetooltip, c("tooltip_1", "tooltip_2"))
  expect_equal(edge_aes$edgeURL, c("http://www.website.net", "http:://www.internet.org"))
  expect_equal(edge_aes$headtooltip, c("tooltip_1", "tooltip_2"))
  expect_equal(edge_aes$headURL, c("http://www.website.net", "http:://www.internet.org"))
  expect_equal(edge_aes$headclip, c(TRUE, FALSE))
  expect_equal(edge_aes$headlabel, c("label_1", "label_2"))
  expect_equal(edge_aes$headport, c("n", "ne"))
  expect_equal(edge_aes$tailtooltip, c("tooltip_1", "tooltip_2"))
  expect_equal(edge_aes$tailURL, c("http://www.website.net", "http:://www.internet.org"))
  expect_equal(edge_aes$tailclip, c(TRUE, FALSE))
  expect_equal(edge_aes$taillabel, c("label_1", "label_2"))
  expect_equal(edge_aes$tailport, c("s", "nw"))
  expect_equal(edge_aes$dir, c("forward", "back"))
  expect_equal(edge_aes$decorate, c(FALSE, TRUE))

  # Expect the function to return
  # an error when supplying any
  # argument name that is not an
  # aesthetic attribute
  expect_error(
    edge_aes(
      value_1 = c(1, 2, 4, 5)))
})

test_that("binding data and aesthetic attribute values is always possible", {

  node_data_ <-
    node_data(
      node_value = 1)

  edge_data_ <-
    edge_data(
      edge_value = 1)

  node_aes_ <-
    node_aes(
      fillcolor = "steelblue")

  edge_aes_ <-
    edge_aes(
      color = "red")

  graph_w_attrs_1 <-
    create_graph() %>%
    add_balanced_tree(
      k = 2, h = 2,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_2 <-
    create_graph() %>%
    add_cycle(
      n = 4,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_3 <-
    create_graph() %>%
    add_path(
      n = 4,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_4 <-
    create_graph() %>%
    add_prism(
      n = 4,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_5 <-
    create_graph() %>%
    add_star(
      n = 4,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_6 <-
    create_graph() %>%
    add_full_graph(
      n = 4,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_7 <-
    create_graph() %>%
    add_grid_2d(
      x = 4, y = 4,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_8 <-
    create_graph() %>%
    add_grid_3d(
      x = 4, y = 4, z = 4,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_9 <-
    create_graph() %>%
    add_gnm_graph(
      n = 20, m = 20,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_10 <-
    create_graph() %>%
    add_gnp_graph(
      n = 20, p = 0.05,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_11 <-
    create_graph() %>%
    add_growing_graph(
      n = 100, m = 1,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_12 <-
    create_graph() %>%
    add_islands_graph(
      n_islands = 4, island_size = 10, p = 0.5, edges_between = 1,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_13 <-
    create_graph() %>%
    add_pa_graph(
      n = 50, m = 1,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_14 <-
    create_graph() %>%
    add_smallworld_graph(
      dimension = 1, size = 50, neighborhood = 1, p = 0.05,
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_15 <-
    create_graph() %>%
    add_node(
      node_aes = node_aes_,
      node_data = node_data_)

  graph_w_attrs_16 <-
    create_graph() %>%
    add_n_nodes(
      n = 2,
      node_aes = node_aes_,
      node_data = node_data_)

  graph_w_attrs_17 <-
    create_graph() %>%
    add_node() %>%
    select_nodes() %>%
    add_n_nodes_ws(
      n = 1,
      direction = "from",
      node_aes = node_aes_, edge_aes = edge_aes_,
      node_data = node_data_, edge_data = edge_data_)

  graph_w_attrs_18 <-
    create_graph() %>%
    add_n_nodes(n = 2) %>%
    add_edge(
      from = 1, to = 2,
      edge_aes = edge_aes_,
      edge_data = edge_data_)

  graph_w_attrs_19 <-
    create_graph() %>%
    add_path(n = 2) %>%
    select_edges() %>%
    add_reverse_edges_ws(
      edge_aes = edge_aes_,
      edge_data = edge_data_)

  expect_contains(names(graph_w_attrs_1$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_2$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_3$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_4$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_5$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_6$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_7$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_8$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_9$nodes_df),  c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_10$nodes_df), c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_11$nodes_df), c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_12$nodes_df), c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_13$nodes_df), c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_14$nodes_df), c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_15$nodes_df), c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_16$nodes_df), c("fillcolor", "node_value"))
  expect_contains(names(graph_w_attrs_17$nodes_df), c("fillcolor", "node_value"))

  expect_contains(names(graph_w_attrs_1$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_2$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_3$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_4$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_5$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_6$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_7$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_8$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_9$edges_df),  c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_10$edges_df), c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_11$edges_df), c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_12$edges_df), c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_13$edges_df), c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_14$edges_df), c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_17$edges_df), c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_18$edges_df), c("color", "edge_value"))
  expect_contains(names(graph_w_attrs_19$edges_df), c("color", "edge_value"))
})

test_that("Graph backups for `add_balanced_tree()` works", {

  #
  # Backup from `add_balanced_tree()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_balanced_tree(
      k = 2, h = 2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_cycle()` works", {

  #
  # Backup from `add_cycle()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_cycle(n = 6)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_edge()` works", {

  #
  # Backup from `add_edge()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_n_nodes(
      n = 2) %>%
    add_edge(
      from = 1,
      to = 2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_edge_clone()` works", {

  #
  # Backup from `add_edge_clone()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_n_nodes(
      n = 3) %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge_clone(
      edge = 1,
      from = 2,
      to = 3)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_edge_clone()` works", {

  #
  # Backup from `add_edge_clone()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_n_nodes(
      n = 3) %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge_clone(
      edge = 1,
      from = 2,
      to = 3)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_edge_df()` works", {

  #
  # Backup from `add_edge_df()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  ndf <-
    create_node_df(
      n = 4,
      type = "letter",
      color = c(
        "red", "green",
        "grey", "blue"),
      value = c(
        3.5, 2.6, 9.4, 2.7))

  edf <-
    create_edge_df(
      from = c(1, 2, 3),
      to = c(4, 3, 1),
      rel = "leading_to")

  create_graph(
    nodes_df = ndf,
    write_backups = TRUE) %>%
    add_edge_df(
      edge_df = edf)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_edge_df()` works", {

  #
  # Backup from `add_edges_from_table()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_nodes_from_table(
      table = currencies) %>%
    add_edges_from_table(
      table = usd_exchange_rates,
      from_col = from_currency,
      to_col = to_currency,
      from_to_map = iso_4217_code)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_edges_w_string()` works", {

  #
  # Backup from `add_edges_w_string()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_n_nodes(n = 4) %>%
    add_edges_w_string(
      edges = "1->2 1->3 2->4 2->3")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_forward_edges_ws()` works", {

  #
  # Backup from `add_forward_edges_ws()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_n_nodes(
      n = 2,
      type = "type_a",
      label = c("a_1", "a_2")) %>%
    add_edge(
      from = 1, to = 2, rel = "a") %>%
    select_edges() %>%
    add_forward_edges_ws(rel = "b")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_full_graph()` works", {

  #
  # Backup from `add_full_graph()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_full_graph(n = 5)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_gnm_graph()` works", {

  #
  # Backup from `add_gnm_graph()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnm_graph(n = 100, m = 120)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_gnp_graph()` works", {

  #
  # Backup from `add_gnp_graph()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnp_graph(n = 100, p = 0.05)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_node_clones_ws()` works", {

  #
  # Backup from `add_node_clones_ws()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnm_graph(
      n = 10, m = 22) %>%
    select_nodes() %>%
    add_node_clones_ws()

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_n_node_clones()` works", {

  #
  # Backup from `add_n_node_clones()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnm_graph(
      n = 10, m = 22) %>%
    add_n_node_clones(
      n = 2,
      node = 1)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_global_graph_attrs()` works", {

  #
  # Backup from `add_global_graph_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_global_graph_attrs(
      attr = "penwidth",
      value = 12,
      attr_type = "node")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `delete_global_graph_attrs()` works", {

  #
  # Backup from `delete_global_graph_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    delete_global_graph_attrs(
      attr = "outputorder",
      attr_type = "graph")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `add_graph_action()` works", {

  #
  # Backup from `add_graph_action()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnm_graph(
      n = 10, m = 22) %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_betweenness",
      column_name = "btwns",
      action_name = "get_btwns")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `delete_graph_actions()` works", {

  #
  # Backup from `delete_graph_actions()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnm_graph(
      n = 10, m = 22) %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_betweenness",
      column_name = "btwns",
      action_name = "get_btwns") %>%
    delete_graph_actions(
      actions = "get_btwns")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `reorder_graph_actions()` works", {

  #
  # Backup from `reorder_graph_actions()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnm_graph(
      n = 10, m = 22) %>%
    add_graph_action(
      fcn = "rescale_node_attrs",
      node_attr_from = "pagerank",
      node_attr_to = "width",
      action_name = "pgrnk_to_width") %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_pagerank",
      column_name = "pagerank",
      action_name = "get_pagerank") %>%
    reorder_graph_actions(
      indices = c(2, 1))

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `transform_to_complement_graph()` works", {

  #
  # Backup from `transform_to_complement_graph()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_cycle(n = 4) %>%
    transform_to_complement_graph()

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `copy_node_attrs()` works", {

  #
  # Backup from `copy_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      node_aes = node_aes(
        color = "blue")) %>%
    copy_node_attrs(
      node_attr_from = color,
      node_attr_to = color_2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `copy_edge_attrs()` works", {

  #
  # Backup from `copy_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      edge_aes = edge_aes(
        color = "blue")) %>%
    copy_edge_attrs(
      edge_attr_from = color,
      edge_attr_to = color_2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `colorize_node_attrs()` works", {

  #
  # Backup from `colorize_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  graph <-
    create_graph(write_backups = TRUE) %>%
    add_path(
      n = 8,
      node_data = node_data(
        weight = c(
          8.2, 3.7, 6.3, 9.2,
          1.6, 2.5, 7.2, 5.4))) %>%
    colorize_node_attrs(
      node_attr_from = weight,
      node_attr_to = fillcolor,
      palette = "Greens",
      cut_points = c(1, 3, 5, 7, 9),
      alpha = 90)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `colorize_edge_attrs()` works", {

  #
  # Backup from `colorize_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  graph <-
    create_graph(write_backups = TRUE) %>%
    add_path(
      n = 8,
      edge_data = edge_data(
        weight = c(
          8.2, 3.7, 6.3, 9.2,
          1.6, 2.5, 7.2))) %>%
    colorize_edge_attrs(
      edge_attr_from = weight,
      edge_attr_to = color,
      palette = "Greens",
      cut_points = c(1, 3, 5, 7, 9),
      alpha = 90)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `drop_node_attrs()` works", {

  #
  # Backup from `drop_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      node_aes = node_aes(
        color = "blue")) %>%
    drop_node_attrs(
      node_attr = color)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `drop_edge_attrs()` works", {

  #
  # Backup from `drop_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      edge_aes = edge_aes(
        color = "blue")) %>%
    drop_edge_attrs(
      edge_attr = color)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `join_node_attrs()` works", {

  #
  # Backup from `join_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  df <-
    data.frame(
          id = c(1, 2, 3, 4, 5),
      values = c(5.5, 2.3, 6.3, 2.1, 8.7))

  create_graph(write_backups = TRUE) %>%
    add_path(n = 5) %>%
    join_node_attrs(df = df)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `join_edge_attrs()` works", {

  #
  # Backup from `join_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  df <-
    data.frame(
      from = c(1, 2, 3, 4),
      to = c(2, 3, 4, 5),
      values = c(5.5, 2.3, 6.3, 2.1))

  create_graph(write_backups = TRUE) %>%
    add_path(n = 5) %>%
    join_edge_attrs(df = df)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `mutate_node_attrs()` works", {

  #
  # Backup from `mutate_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      node_data = node_data(
        weight = c(
          8.2, 3.7, 6.3, 9.2))) %>%
    mutate_node_attrs(
      half_weight = weight / 2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `mutate_edge_attrs()` works", {

  #
  # Backup from `mutate_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      edge_data = edge_data(
        weight = c(
          8.2, 3.7, 6.3))) %>%
    mutate_edge_attrs(
      half_weight = weight / 2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `mutate_node_attrs_ws()` works", {

  #
  # Backup from `mutate_node_attrs_ws()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      node_data = node_data(
        weight = c(
          8.2, 3.7, 6.3, 9.2))) %>%
    select_nodes_by_id(nodes = c(1, 2)) %>%
    mutate_node_attrs_ws(
      half_weight = weight / 2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `mutate_edge_attrs_ws()` works", {

  #
  # Backup from `mutate_edge_attrs_ws()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      edge_data = edge_data(
        weight = c(
          8.2, 3.7, 6.3))) %>%
    select_edges_by_edge_id(edges = c(1, 2)) %>%
    mutate_edge_attrs_ws(
      half_weight = weight / 2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `rename_node_attrs()` works", {

  #
  # Backup from `rename_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      node_data = node_data(
        weight = c(
          8.2, 3.7, 6.3, 9.2))) %>%
    rename_node_attrs(
      node_attr_from = weight,
      node_attr_to = weight_2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `rename_edge_attrs()` works", {

  #
  # Backup from `rename_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      edge_data = edge_data(
        weight = c(
          8.2, 3.7, 6.3))) %>%
    rename_edge_attrs(
      edge_attr_from = weight,
      edge_attr_to = weight_2)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `recode_node_attrs()` works", {

  #
  # Backup from `recode_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_gnm_graph(
      n = 5,
      m = 10,
      set_seed = 23) %>%
    set_node_attrs(
      node_attr = shape,
      values =
        c("circle", "hexagon",
          "rectangle", "rectangle",
          "circle")) %>%
    recode_node_attrs(
      node_attr_from = shape,
      "circle -> square",
      "rectangle -> triangle")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `recode_edge_attrs()` works", {

  #
  # Backup from `recode_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 6,
      edge_data = edge_data(
        values = c(
          "circle", "hexagon",
          "rectangle", "rectangle",
          "circle"))) %>%
    recode_edge_attrs(
      edge_attr_from = values,
      "circle -> square",
      "rectangle -> triangle")

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `rescale_node_attrs()` works", {

  #
  # Backup from `rescale_node_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 4,
      node_data = node_data(
        weight = c(
          8.2, 3.7, 6.3, 9.2))) %>%
    rescale_node_attrs(
      node_attr_from = weight,
      to_lower_bound = 0,
      to_upper_bound = 1)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `rescale_edge_attrs()` works", {

  #
  # Backup from `rescale_edge_attrs()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(
      n = 5,
      edge_data = edge_data(
        weight = c(
          8.2, 3.7, 6.3, 9.2))) %>%
    rescale_edge_attrs(
      edge_attr_from = weight,
      to_lower_bound = 0,
      to_upper_bound = 1)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `rev_edge_dir()` works", {

  #
  # Backup from `rev_edge_dir()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(n = 2) %>%
    rev_edge_dir()

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `rev_edge_dir_ws()` works", {

  #
  # Backup from `rev_edge_dir_ws()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_path(n = 3) %>%
    select_edges_by_edge_id(edges = 1) %>%
    rev_edge_dir_ws()

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `set_node_position()` works", {

  #
  # Backup from `set_node_position()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_node() %>%
    set_node_position(
      node = 1, x = 1, y = 1)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `nudge_node_positions_ws()` works", {

  #
  # Backup from `nudge_node_positions_ws()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_node() %>%
    set_node_position(
      node = 1, x = 1, y = 1) %>%
    select_nodes() %>%
    nudge_node_positions_ws(
      dx = 2, dy = 0)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

test_that("Graph backups for `nudge_node_positions_ws()` works", {

  #
  # Backup from `nudge_node_positions_ws()`
  #

  main_wd <- getwd()
  random_dir <- paste(sample(letters[1:10], 10), collapse = "")
  dir.create(random_dir)
  setwd(random_dir)

  create_graph(write_backups = TRUE) %>%
    add_node() %>%
    set_node_position(
      node = 1, x = 1, y = 1) %>%
    select_nodes() %>%
    nudge_node_positions_ws(
      dx = 2, dy = 0)

  expect_length(list.files(), 1)

  setwd(main_wd)
  unlink(random_dir, recursive = TRUE)
})

# Joining data frames to graph dfs")

test_that("joining a data frame to an ndf is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 5) %>%
    add_edges_w_string(
      edges = "1->2 1->3 2->4 2->5 3->5")

  # Create a data frame with node ID values and a
  # set of numeric values
  suppressWarnings(RNGversion("3.5.0"))
  set.seed(25)

  # Create a data frame from which `values` will
  # be a join column
  df <-
    data.frame(
      id = 1:6,
      values = round(rnorm(6, 5), 2))

  # Perform the join operation
  graph <-
    graph %>%
    join_node_attrs(df = df)

  # Expect that 4 columns exists in the graph's
  # node data frame
  expect_equal(
    ncol(graph$nodes_df), 4)

  # Expect a column named `values` in the graph's
  # node data frame
  expect_contains(colnames(graph$nodes_df), "values")

  # Expect that the `values` column in the graph's
  # node data frame is equivalent to the first 5 values
  # in the `values` column of `df`
  expect_equal(
    graph$nodes_df$values, df$values[1:5])

  # Expect an error if `by_graph` specified but not
  # `by_df`
  expect_error(
    create_graph() %>%
    add_n_nodes(n = 5) %>%
    add_edges_w_string(
      edges = "1->2 1->3 2->4 2->5 3->5") %>%
    join_node_attrs(
      df = df,
      by_graph = "nodes"))

  # Expect an error if `by_df` specified but not
  # `by_graph`
  expect_error(
    create_graph() %>%
      add_n_nodes(n = 5) %>%
      add_edges_w_string(
        edges = "1->2 1->3 2->4 2->5 3->5") %>%
      join_node_attrs(
        df = df,
        by_df = "node"))
})

test_that("joining a data frame to an edf is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 5) %>%
    add_edges_w_string(
      edges = "1->2 1->3 2->4 2->5 3->5")

  # Create a data frame with node ID values and a
  # set of numeric values
  suppressWarnings(RNGversion("3.5.0"))
  set.seed(25)

  # Create a data frame from which `values` will
  # be a join column
  df <-
    data.frame(
      from = c(1, 1, 2, 2, 3),
      to = c(2, 3, 4, 5, 5),
      values = rnorm(5, 5))

  # Perform the join operation
  graph <-
    graph %>%
    join_edge_attrs(df = df)

  # Expect that 5 columns exists in the graph's
  # edge data frame
  expect_equal(
    ncol(graph$edges_df), 5)

  # Expect a column named `values` in the graph's
  # edge data frame
  expect_contains(colnames(graph$edges_df), "values")

  # Expect that the `values` column in the graph's
  # node data frame is equivalent to the first 5 values
  # in the `values` column of `df`
  expect_equal(
    graph$edges_df$values, df$values[1:5])

  # Expect an error if `by_graph` specified but not
  # `by_df`
  expect_error(
    create_graph() %>%
      add_n_nodes(n = 5) %>%
      add_edges_w_string(edges = "1->2 1->3 2->4 2->5 3->5") %>%
      join_edge_attrs(
        df = df,
        by_graph = c("from", "to")))

  # Expect an error if `by_df` specified but not
  # `by_graph`
  expect_error(
    create_graph() %>%
      add_n_nodes(n = 5) %>%
      add_edges_w_string(
        edges = "1->2 1->3 2->4 2->5 3->5") %>%
      join_edge_attrs(
        df = df,
        by_df = c("from", "to")))
})

# Deletion of nodes and edges from an existing graph

test_that("nodes and edges can be deleted from a graph", {

  # Create an empty graph and
  # add four nodes
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node()

  # Add edges to the graph
  graph <-
    graph %>%
    add_edge(
      from = 1,
      to = 2,
      rel = "to_get") %>%
    add_edge(
      from = 1,
      to = 3,
      rel = "received_from") %>%
    add_edge(
      from = 3,
      to = 4,
      rel = "received_from")

  # Get the graph's nodes
  graph_nodes <- get_node_ids(graph)

  # Get the graph's edges
  graph_edges <-
    get_edges(
      graph = graph,
      return_type = "list")

  # Remove a node (removing a node removes its edges)
  graph <-
    delete_node(
      graph = graph,
      node = 4)

  # Get the graph's nodes after calling `delete_node()`
  graph_nodes_delete_node <- get_node_ids(graph)

  # Get the graph's edges after calling `delete_node()`
  graph_edges_delete_node <-
    get_edges(
      graph = graph,
      return_type = "list")

  # Expect that the number of nodes will be decreased
  expect_lt(
    length(graph_nodes_delete_node),
    length(graph_nodes))

  # Expect that the number of edges will be decreased
  expect_lt(
    length(graph_edges_delete_node[[1]]),
    length(graph_edges[[1]]))

  # Expect that node ID `4` will not be present in
  # the revised graph
  expect_false(4 %in% graph_nodes_delete_node)

  expect_false(4 %in% graph_edges_delete_node[[1]])

  expect_false(4 %in% graph_edges_delete_node[[2]])

  # Remove an edge (removing an edge retains nodes)
  graph <-
    delete_edge(
      graph = graph,
      from = 1,
      to = 3)

  # Get the graph's edges after calling `delete_node()`
  graph_edges_delete_edge <-
    get_edges(
      graph = graph,
      return_type = "list")

  # Expect that edge between nodes `1` and `3` will
  # not be present
  expect_false(
    is_edge_present(
      graph = graph,
      from = 1,
      to = 3))

  # Expect that the nodes involved in the edge
  # deletion are retained
  expect_true(
    is_node_present(graph, node = 1))

  expect_true(
    is_node_present(graph, node = 3))
})

test_that("errors can occur when deleting nodes or edges with certain input values", {

  # Create an empty graph and
  # add four nodes
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node()

  # Add edges to the graph
  graph <-
    graph %>%
    add_edge(
      from = 1,
      to = 2,
      rel = "to_get") %>%
    add_edge(
      from = 1,
      to = 3,
      rel = "received_from") %>%
    add_edge(
      from = 3,
      to = 4,
      rel = "received_from")

  # Expect an error the node specified is not a
  # single value
  expect_error(
    delete_node(graph, node = c(1, 2)))

  # Expect an error if the node specified is not
  # present in the graph
  expect_error(
    delete_node(graph, node = 5))

  # Expect an error if either node specified is not
  # a single value
  expect_error(
    delete_edge(graph, from = c(1, 2), to = 3))

  # Expect an error if both nodes specified are not
  # present in the graph
  expect_error(
    delete_edge(graph, from = 1, to = 5))
})

test_that("nodes and edges can be deleted from a graph via a selection", {

  # Create an empty graph and
  # add four nodes
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node()

  # Add edges to the graph
  graph <-
    graph %>%
    add_edge(
      from = 1,
      to = 2,
      rel = "to_get") %>%
    add_edge(
      from = 1,
      to = 3,
      rel = "received_from") %>%
    add_edge(
      from = 3,
      to = 4,
      rel = "received_from")

  # Select 2 nodes
  graph <-
    select_nodes(
      graph = graph,
      nodes = c(3, 4))

  # Delete the nodes specified in the selection
  graph_node_deletion <- delete_nodes_ws(graph)

  # Expect a node count of 2
  expect_equal(
    count_nodes(graph = graph_node_deletion), 2)

  # Expect nodes `1` and `2` to be present
  expect_true(
    all(c(1, 2) %in% get_node_ids(graph_node_deletion)))

  # Expect an error if trying to delete a node where
  # there is no node selection
  expect_error(
    graph %>%
      clear_selection() %>%
      delete_nodes_ws())

  # Expect a node count of 4
  expect_equal(
    count_nodes(graph = graph), 4)

  # Expect nodes `1`, `2`, `3`, and `4` to be present
  expect_true(
    all(1:4 %in% get_node_ids(graph)))

  # Select 2 edges
  graph <-
    select_edges(
      graph = graph,
      from = 1,
      to = 2)

  graph <-
    select_edges(
      graph = graph,
      from = 1,
      to = 3)

  # Delete the edges specified in the selection
  graph_edge_deletion <- delete_edges_ws(graph)

  # Expect an edge count of 1
  expect_equal(
    count_edges(graph = graph_edge_deletion), 1)

  # Expect edge `3`->`4` to be present
  expect_true(
    "3->4" %in% get_edges(
      graph = graph_edge_deletion,
      return_type = "vector"))

  # Expect an error if trying to delete an edge where
  # there is no edge selection
  expect_error(
    graph %>%
      clear_selection() %>%
      delete_edges_ws())

  # Expect an edge count of 3
  expect_equal(
    count_edges(graph = graph), 3)

  # Expect edges `1`->`2`, `1`->`3`, and `3`->`4`
  # to be present
  expect_true(
    all(
      c("1->2", "1->3", "3->4") %in% get_edges(
        graph = graph,
        return_type = "vector")))
})

test_that("edges can be deleted from a graph using node label values", {

  # Create a directed graph with 2
  # labeled nodes and an edge
  graph_labeled_nodes <-
    create_graph() %>%
    add_n_nodes(
      n = 2,
      label = c("one", "two")) %>%
    add_edge(
      from = "one",
      to = "two")

  # Delete the single graph edge by
  # specifying the node label names
  graph_deleted_edge <-
    graph_labeled_nodes %>%
    delete_edge(
      from = "one",
      to = "two")

  # Expect a node count of 2
  expect_equal(
    count_nodes(graph = graph_deleted_edge), 2)

  # Expect an edge count of 0
  expect_equal(
    count_edges(graph = graph_deleted_edge), 0)

  # Expect an error when specifying a node
  # label that does not exist
  expect_snapshot(error = TRUE, {
    graph_labeled_nodes %>%
      delete_edge(
        from = "zero",
        to = "two")

    graph_labeled_nodes %>%
      delete_edge(
        from = "one",
        to = "three")
  })

  # Create a directed graph with 3
  # labelled nodes (with indistinct labels)
  # and 1 edge
  graph_labeled_nodes_not_distinct_labels_1 <-
    create_graph() %>%
    add_path(
      n = 2,
      label = c("one", "two")) %>%
    add_node(
      label = "one")

  # Expect an error when trying to delete
  # an edge using non-distinct node labels
  expect_error(
    graph_labeled_nodes_not_distinct_labels_1 %>%
      delete_edge(
        from = "one",
        to = "two"))

  # Create another directed graph with 3
  # labeled nodes (with indistinct labels)
  # and 1 edge
  graph_labeled_nodes_not_distinct_labels_2 <-
    create_graph() %>%
    add_path(
      n = 2,
      label = c("one", "two")) %>%
    add_node(
      label = "two")

  # Expect an error when trying to delete
  # an edge using non-distinct node labels
  expect_error(
    graph_labeled_nodes_not_distinct_labels_2 %>%
      delete_edge(
        from = "one",
        to = "two"))
})

# Getting counts of nodes and edges in graph objects

test_that("getting a node count (with `count_nodes()`) is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_path(n = 5)

  # Obtain a total count of nodes
  n_nodes <-
    count_nodes(graph)

  # Expect that the `total_count_of_nodes` object
  # is a named vector
  expect_type(
    n_nodes, "integer")

  expect_equal(
    n_nodes, 5)
})

test_that("getting an edge count (with `count_edges()`) is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_path(n = 5)

  # Obtain a total count of edges
  n_edges <-
    count_edges(graph)

  # Expect that the `n_edges` object is
  # an integer
  expect_equal(n_edges, 4L)
})

test_that("getting a node/edge count for an empty graph is possible", {

  # Create an empty graph
  empty_graph <- create_graph()

  # Expect that a node count of an empty graph
  # will return `0`
  expect_equal(
    count_nodes(
      graph = empty_graph), 0)

  # Expect that an edge count of an empty graph
  # will return `0`
  expect_equal(
    count_edges(
      graph = empty_graph), 0)
})

test_that("counting the number of edges with the same definition is possible", {

  # Create a node data frame (ndf)
  ndf <-
    create_node_df(
      n = 5,
      label = TRUE)

  # Create an edge data frame (edf)
  edf <-
    create_edge_df(
      from = c(1, 4, 4, 3, 5, 1, 3, 4),
        to = c(4, 1, 1, 2, 2, 2, 2, 1))

  # Create a graph with the ndf and edf
  graph <-
    create_graph(
      nodes_df = ndf,
      edges_df = edf)

  # Get the total number of edge
  # definitions (e.g., `4` -> `1`) where
  # there are multiple edges (i.e.,
  # distinct edges with separate edge
  # ID values)
  edges_where_multiple_occur <-
    get_edge_count_w_multiedge(graph = graph)

  # Expect that there are 2 such
  # edge definitions
  expect_equal(
    edges_where_multiple_occur, 2)
})

test_that("counting the number of multiple edges between a specific node pair is possible", {

  # Create a node data frame (ndf)
  ndf <-
    create_node_df(
      n = 5,
      label = TRUE)

  # Create an edge data frame (edf)
  edf <-
    create_edge_df(
      from = c(1, 4, 4, 3, 5, 1, 3, 4),
      to = c(4, 1, 1, 2, 2, 2, 2, 1))

  # Create a graph with the ndf and edf
  graph <-
    create_graph(
      nodes_df = ndf,
      edges_df = edf)

  # Get the total number of multiple
  # edges (those edges that share an
  # edge definition) in the graph
  global_multiple_edges_count <-
    get_multiedge_count(graph = graph)

  # Expect that there are 3 edges that
  # serve as multiple edges
  expect_equal(
    global_multiple_edges_count, 3)
})

test_that("counting the number of strongly connected components is possible", {

  # Create a graph with several
  # graph islands
  graph <-
     create_graph() %>%
     add_islands_graph(
       n_islands = 4,
       island_size = 1,
       p = 1 / 555,
       edges_between = 0)

  # Expect that the count of strongly
  # connected components is 4
  expect_equal(
    graph %>% count_s_connected_cmpts(), 4)

  # Expect that an empty graph will
  # return NA
  expect_true(
    is.na(count_s_connected_cmpts(graph = create_graph())))
})

test_that("counting the number of weakly connected components is possible", {

  # Create a graph with 2 cycles
  graph <-
    create_graph() %>%
    add_cycle(n = 5) %>%
    add_cycle(n = 5)

  # Expect that the count of weakly
  # connected components is 2
  expect_equal(
    graph %>% count_w_connected_cmpts(), 2)

  # Expect that an empty graph will
  # return NA
  expect_true(
    is.na(count_w_connected_cmpts(graph = create_graph())))
})

test_that("counting the number of loop edges is possible", {

  # Create a full graph that
  # includes loops
  graph <-
    create_graph(
      directed = FALSE) %>%
    add_full_graph(
      n = 3,
      keep_loops = TRUE)

  # Expect that the number of
  # loop edges is 3
  expect_equal(
    graph %>% count_loop_edges(), 3)

  # Expect that an empty graph will
  # return 0
  expect_equal(
    count_loop_edges(graph = create_graph()), 0)
})

test_that("counting the unconnected graph nodes is possible", {

  # Create a graph with a
  # path of nodes and 3
  # unconnected nodes
  graph <-
    create_graph() %>%
    add_path(n = 3) %>%
    add_n_nodes(n = 3)

  # Expect that the count of
  # unconnected nodes is 3
  expect_equal(
    graph %>% count_unconnected_nodes(), 3)

  # Expect that an empty graph will
  # return 0
  expect_equal(
    count_unconnected_nodes(graph = create_graph()), 0)

  # Expect that a graph without unconnected
  # nodes will return 0
  expect_equal(
    create_graph() %>%
      add_path(n = 4) %>%
      count_unconnected_nodes(), 0)
})

# Perform graph object transformations

test_that("Converting to igraph is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Create an igraph object from a
  # DiagrammeR graph object
  igraph_graph <-
    to_igraph(graph)

  # Expect that the new object is an igraph object
  expect_s3_class(
    igraph_graph, "igraph")
})

test_that("Changing to undirected mode is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_path(n = 5)

  undirected_graph <-
    set_graph_undirected(graph)

  # Expect the the graph is directed
  expect_true(graph$directed)
})

test_that("Changing to directed mode is possible", {

  # Create a simple graph
  graph <-
    create_graph(
      directed = FALSE) %>%
    add_path(n = 5)

  directed_graph <-
    set_graph_directed(graph)

  # Expect the the graph is undirected
  expect_false(
    graph$directed)
})

test_that("Reversing the graph edges is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get a vector of `from` nodes
  from_nodes <- graph$edges_df$from

  # Get a vector of `to` nodes
  to_nodes <- graph$edges_df$to

  # Change the edge direction for all edges in
  # the graph
  graph <- rev_edge_dir(graph)

  # Expect that the `from` nodes in the new graph
  # object are identical to the `to` nodes before
  # reversing the edge direction
  expect_equal(
    graph$edges_df$from, to_nodes)

  # Expect that the `to` nodes in the new graph
  # object are identical to the `from` nodes before
  # reversing the edge direction
  expect_equal(
    graph$edges_df$to, from_nodes)

  # Expect an error if reversing edge direction
  # in an undirected graph
  expect_error(
    graph <-
      create_graph(
        directed = FALSE) %>%
      add_gnm_graph(
        n = 10,
        m = 22,
        set_seed = 23) %>%
      rev_edge_dir())
})

test_that("Reversing edges in a selection is possible", {

  # Create a graph with a directed tree
  graph <-
    create_graph() %>%
    add_balanced_tree(
      k = 2, h = 2)

  # Select all edges associated with
  # nodes `1` and `2` (this selects
  # 4 graph edges)
  graph <-
    select_edges_by_node_id(
      graph = graph,
      nodes = 1:2)

  # Reverse the edge directions of
  # the selected edges
  graph_reversed_edges <-
    graph %>%
    rev_edge_dir_ws()

  # Expect certain edge definitions
  # in the transformed graph
  expect_identical(
    graph_reversed_edges %>%
      get_edges(),
    c("2->1", "3->1", "4->2",
      "5->2", "3->6", "3->7"))

  # Expect an error if there is no
  # selection of edges
  expect_error(
    create_graph() %>%
      add_balanced_tree(
        k = 2, h = 2) %>%
      rev_edge_dir_ws())

  # Expect an error if the graph
  # is undirected
  expect_snapshot(error = TRUE,
    create_graph(
      directed = FALSE) %>%
      add_balanced_tree(
        k = 2, h = 2) %>%
      select_edges_by_node_id(
        nodes = 1:2) %>%
      rev_edge_dir_ws())
})

test_that("Creating a complement graph is possible", {

  # Create a simple graph with a single cycle
  graph <-
    create_graph() %>%
    add_cycle(n = 4)

  # Create the complement of the graph
  graph_c <-
    transform_to_complement_graph(graph)

  # Expect 8 edges in the complement graph
  expect_equal(
    graph_c %>%
      get_edge_df() %>%
      nrow(), 8
  )

  # Expect that there are no loops in the
  # complement graph
  expect_equal(
    graph_c %>%
      get_edge_df() %>%
      dplyr::filter(from == to) %>%
      nrow(), 0
  )

  # Create the complement of the original graph
  # with loops created
  graph_cl <-
    transform_to_complement_graph(
      graph = graph,
      loops = TRUE
    )

  # Expect 12 edges in this complement graph
  expect_equal(
    graph_cl %>%
      get_edge_df() %>%
      nrow(), 12
  )
})

test_that("Fully connecting selected nodes is possible", {

  # Create a directed graph with a path
  # of 3 nodes and two isolated nodes
  graph_directed <-
    create_graph() %>%
    add_path(n = 3) %>%
    add_n_nodes(n = 2)

  # Select a node in the path
  # of nodes (node `3`) and
  # the two isolated nodes (`4`
  # and `5`); then, and fully
  # connect these nodes together
  graph_directed <-
    graph_directed %>%
    select_nodes_by_id(
      nodes = 3:5) %>%
    fully_connect_nodes_ws()

  # Expect 8 edges in the new graph
  expect_equal(
    graph_directed %>%
      count_edges(), 8)

  # Expect certain edges in the graph
  expect_identical(
    graph_directed %>%
      get_edges(),
    c("1->2", "2->3", "3->4", "3->5",
      "4->5", "4->3", "5->3", "5->4"))

  # Create an undirected graph with a path
  # of 3 nodes and two isolated nodes
  graph_undirected <-
    create_graph(
      directed = FALSE) %>%
    add_path(n = 3) %>%
    add_n_nodes(n = 2)

  # Select a node in the path
  # of nodes (node `3`) and
  # the two isolated nodes (`4`
  # and `5`); then, and fully
  # connect these nodes together
  graph_undirected <-
    graph_undirected %>%
    select_nodes_by_id(
      nodes = 3:5) %>%
    fully_connect_nodes_ws()

  # Expect 8 edges in the new graph
  expect_equal(
    graph_undirected %>%
      count_edges(), 5)

  # Expect certain edges in the graph
  expect_identical(
    graph_undirected %>%
      get_edges(),
    c("1->2", "2->3", "3->4",
      "3->5", "4->5"))

  # Expect an error if there is no valid
  # selection of node
  expect_error(
    create_graph() %>%
      add_n_nodes(n = 2) %>%
      fully_connect_nodes_ws())
})

test_that("Fully disconnecting selected nodes is possible", {

  # Create a directed graph with a path
  # of 3 nodes and two isolated nodes
  graph_directed <-
    create_graph() %>%
    add_path(n = 3)

  # Select all nodes and fully
  # disconnect these nodes
  graph_directed <-
    graph_directed %>%
    select_nodes() %>%
    fully_disconnect_nodes_ws()

  # Expect no edges in the new graph
  expect_equal(
    graph_directed %>%
      count_edges(), 0)

  # Expect all nodes to be retained
  expect_equal(
    graph_directed %>%
      count_nodes(), 3)

  # Create an undirected graph with
  # a path of 3 nodes
  graph_undirected <-
    create_graph(
      directed = FALSE) %>%
    add_path(n = 3)

  # Select all nodes and fully
  # disconnect these nodes
  graph_undirected <-
    graph_undirected %>%
    select_nodes() %>%
    fully_disconnect_nodes_ws()

  # Expect no edges in the new graph
  expect_equal(
    graph_undirected %>%
      count_edges(), 0)

  # Expect an error if there is no valid
  # selection of node
  expect_snapshot(error = TRUE,
    create_graph() %>%
      add_path(n = 2) %>%
      fully_disconnect_nodes_ws())
})

test_that("Removing loop edges via a selection is possible", {

  # Create an undirected, full graph
  # of 5 nodes with loops retained
  graph <-
    create_graph(
      directed = FALSE) %>%
    add_full_graph(
      n = 5,
      keep_loops = TRUE)

  # Select nodes `3` and `4`
  # and remove the loop edges
  # associated with those nodes
  graph_loops_removed <-
    graph %>%
    select_nodes_by_id(
      nodes = c(3, 4)) %>%
    delete_loop_edges_ws()

  # Expect that there are 3 loops
  # remaining in the graph
  expect_equal(
    graph_loops_removed %>%
      get_edge_df() %>%
      dplyr::filter(from == to) %>%
      nrow(), 3
  )

  # Expect an error if there is
  # isn't a valid node selection
  expect_error(
    delete_loop_edges_ws(graph)
  )
})

# Print graph summary

get_printed_output <- function(graph, line = NULL) {

  captured_output <-
    capture_output(
      graph %>% print()) %>%
    stringr::str_split(pattern = "\n") %>%
    unlist()

  if (!is.null(line)) {
    captured_output <- captured_output[line]
  }

  captured_output
}

test_that("Printing a summary of an empty graph works", {

  graph <- create_graph()

  expect_equal(
    graph %>% get_printed_output(1),
    "DiagrammeR Graph // no nodes")

  expect_equal(
    graph %>% get_printed_output(2),
    "  -- empty graph (mode: directed)")

  expect_equal(
    create_graph(directed = FALSE) %>% get_printed_output(2),
    "  -- empty graph (mode: undirected)")

  expect_equal(
    graph %>% get_printed_output(3), "")

  expect_equal(
    graph %>%
      get_printed_output(4) %>% substr(1, 45) %>% stringr::str_trim(),
    "NODES / type: <unused> / label: <unused>")

  expect_equal(
    graph %>% get_printed_output(5) %>% substr(1, 45) %>% stringr::str_trim(),
    "-- no additional node attributes")

  expect_equal(
    graph %>%
      get_printed_output(6) %>% substr(1, 45) %>% stringr::str_trim(),
    "EDGES / rel: <unused>")

  expect_equal(
    graph %>%
      get_printed_output(7) %>% substr(1, 45) %>% stringr::str_trim(),
    "-- no additional edge attributes")

  expect_equal(
    graph %>%
      get_printed_output(8) %>% substr(1, 45) %>% stringr::str_trim(),
    "SELECTION / <none>")

  expect_equal(
    graph %>%
      get_printed_output(9) %>% substr(1, 45) %>% stringr::str_trim(),
    "CACHE / <none>")

  expect_equal(
    graph %>%
      get_printed_output(10) %>% substr(1, 30) %>% stringr::str_trim(),
    "GLOBAL ATTRS / 17 are set")

  expect_equal(
    graph %>%
      get_printed_output(11) %>% substr(1, 45) %>% stringr::str_trim(),
    "GRAPH ACTIONS / <none>")

  expect_equal(
    graph %>%
      get_printed_output(12) %>% substr(1, 45) %>% stringr::str_trim(),
    "GRAPH LOG / create_graph()")
})

test_that("Printing a summary of node types works", {

  graph_type_complete <-
    create_graph() %>%
    add_path(
      n = 4,
      type = "a")

  graph_type_incomplete <-
    create_graph() %>%
    add_path(
      n = 4,
      type = c("a", "a", NA, NA))

  expect_equal(
    graph_type_complete %>% get_printed_output(4) %>% stringr::str_trim(),
    "NODES / type: 1 vals - complete / label: 4 vals - complete & unique")

  expect_equal(
    graph_type_incomplete %>% get_printed_output(4) %>% stringr::str_trim(),
    "NODES / type: 1 val / label: 4 vals - complete & unique")
})

test_that("Printing a summary of node labels works", {

  graph_label_complete <-
    create_graph() %>%
    add_path(
      n = 4,
      label = c(1, 2, 3, 4))

  graph_label_incomplete <-
    create_graph() %>%
    add_path(
      n = 4,
      label = c(1, 2, NA, NA))

  expect_equal(
    graph_label_complete %>%
      get_printed_output(4) %>% substr(1, 60) %>% stringr::str_trim(),
    "NODES / type: <unused> / label: 4 vals - complete & unique")

  expect_equal(
    graph_label_incomplete %>%
      get_printed_output(4) %>% substr(1, 45) %>% stringr::str_trim(),
    "NODES / type: <unused> / label: 2 vals")
})

test_that("Printing a summary of extra node attrs works", {

  graph_node_attr_1 <-
    create_graph() %>%
    add_path(
      n = 4,
      node_data = node_data(
        value_1 = c(1, 2, 3, 4)))

  graph_node_attr_2 <-
    create_graph() %>%
    add_path(
      n = 4,
      node_data = node_data(
        value_1 = c(1, 2, 3, 4),
        value_2 = c(5, 6, 7, 9)))

  graph_node_attr_4 <-
    create_graph() %>%
    add_path(
      n = 4,
      node_data = node_data(
        value_1 = c(1, 2, 3, 4),
        value_2 = c(5, 6, 7, 9),
          ctg_1 = c("a", "b", "c", "d"),
          ctg_2 = c("e", "f", "g", "h")))

  expect_equal(
    graph_node_attr_1 %>%
      get_printed_output(5) %>% stringr::str_trim(),
    "-- 1 additional node attribute (value_1)")

  expect_equal(
    graph_node_attr_2 %>%
      get_printed_output(5) %>% stringr::str_trim(),
    "-- 2 additional node attributes (value_1, value_2)")

  expect_equal(
    graph_node_attr_4 %>%
      get_printed_output(5) %>% stringr::str_trim(),
    "-- 4 additional node attributes (value_1, value_2, ctg_1 + 1 more)")
})

test_that("Printing a summary of edge rel values works", {

  graph_rel_complete <-
    create_graph() %>%
    add_path(
      n = 4,
      label = c(1, 2, 3, 4),
      rel = c("a", "a", "b"))

  graph_rel_incomplete <-
    create_graph() %>%
    add_path(
      n = 4,
      label = c(1, 2, 3, 4),
      rel = c("a", "b", NA))

  expect_equal(
    graph_rel_complete %>%
      get_printed_output(6) %>% substr(1, 45) %>% stringr::str_trim(),
    "EDGES / rel: 2 vals - complete")

  expect_equal(
    graph_rel_incomplete %>%
      get_printed_output(4) %>% substr(1, 60) %>% stringr::str_trim(),
    "NODES / type: <unused> / label: 4 vals - complete & unique")
})

test_that("Printing a summary of extra edge attrs works", {

  graph_edge_attr_1 <-
    create_graph() %>%
    add_path(
      n = 5,
      edge_data = edge_data(
        value_1 = c(1, 2, 3, 4)))

  graph_edge_attr_2 <-
    create_graph() %>%
    add_path(
      n = 5,
      edge_data = edge_data(
        value_1 = c(1, 2, 3, 4),
        value_2 = c(5, 6, 7, 9)))

  graph_edge_attr_4 <-
    create_graph() %>%
    add_path(
      n = 5,
      edge_data = edge_data(
        value_1 = c(1, 2, 3, 4),
        value_2 = c(5, 6, 7, 9),
        ctg_1 = c("a", "b", "c", "d"),
        ctg_2 = c("e", "f", "g", "h")))

  expect_equal(
    graph_edge_attr_1 %>%
      get_printed_output(7) %>% stringr::str_trim(),
    "-- 1 additional edge attribute (value_1)")

  expect_equal(
    graph_edge_attr_2 %>%
      get_printed_output(7) %>% stringr::str_trim(),
    "-- 2 additional edge attributes (value_1, value_2)")

  expect_equal(
    graph_edge_attr_4 %>%
      get_printed_output(7) %>% stringr::str_trim(),
    "-- 4 additional edge attributes (value_1, value_2, ctg_1 + 1 more)")
})

test_that("Describing if a graph is weighted works", {

  graph_weighted <-
    create_graph() %>%
    add_path(
      n = 5,
      edge_data = edge_data(
        weight = c(1, 2, 3, 4)))

  graph_unweighted <-
    create_graph() %>%
    add_path(n = 5)

  expect_match(
    graph_weighted %>% get_printed_output(2),
    "weighted")

  expect_no_match(
    graph_unweighted %>% get_printed_output(2),
    "weighted")
})

test_that("Describing if a graph is a DAG works", {

  graph_dag <-
    create_graph() %>%
    add_path(n = 5)

  graph_cycle <-
    create_graph() %>%
    add_cycle(n = 5)

  expect_match(
    graph_dag %>% get_printed_output(2),
    "DAG")

  expect_no_match(
    graph_cycle %>% get_printed_output(2),
    "DAG")
})

test_that("Describing if a graph is a property graph works", {

  graph_pg <-
    create_graph() %>%
    add_path(
      n = 4,
      type = "a",
      label = 1:4,
      rel = c("a", "a", "b", "b"))

  graph_not_pg <-
    create_graph() %>%
    add_path(
      n = 4,
      label = 1:4,
      rel = c("a", "a", "b", "b"))

  expect_match(
    graph_pg %>% get_printed_output(2),
    "property graph")

  expect_no_match(
    graph_not_pg %>% get_printed_output(2),
    "property graph")
})

test_that("Describing if a graph is a simple graph works", {

  graph_simple <-
    create_graph() %>%
    add_path(
      n = 4)

  graph_not_simple <-
    create_graph() %>%
    add_path(
      n = 4) %>%
    add_edges_w_string(
      edges = "1->2")

  expect_match(
    graph_simple %>% get_printed_output(2),
    "simple")

  expect_no_match(
    graph_not_simple %>% get_printed_output(2),
    "simple")
})

test_that("Describing if a graph is a connected or not works", {

  graph_connected <-
    create_graph() %>%
    add_path(
      n = 4)

  graph_not_connected <-
    create_graph() %>%
    add_n_nodes(n = 4)

  expect_match(
    graph_connected %>% get_printed_output(2),
    " connected ")

  expect_match(
    graph_not_connected %>% get_printed_output(2),
    " disconnected ")
})

test_that("The number of reported nodes is correct", {

  graph_no_nodes <-
    create_graph()

  graph_1_node <-
    create_graph() %>%
    add_node()

  graph_4_nodes <-
    create_graph() %>%
    add_n_nodes(n = 4)

  expect_match(
    graph_no_nodes %>% get_printed_output(1),
    " no nodes")

  expect_match(
    graph_1_node %>% get_printed_output(1),
    " 1 node")

  expect_match(
    graph_4_nodes %>% get_printed_output(1),
    " 4 nodes")
})

test_that("The number of reported edges is correct", {

  graph_no_edges <-
    create_graph() %>%
    add_n_nodes(n = 4)

  graph_1_edge <-
    create_graph() %>%
    add_path(n = 2)

  graph_3_edges <-
    create_graph() %>%
    add_path(n = 4)

  expect_no_match(
    graph_no_edges %>% get_printed_output(1),
    " edge")

  expect_match(
    graph_1_edge %>% get_printed_output(1),
    " 1 edge")

  expect_match(
    graph_3_edges %>% get_printed_output(1),
    " 3 edges")

  expect_match(
    graph_3_edges %>% get_printed_output(1),
    "DiagrammeR Graph // 4 nodes / 3 edges")
})

test_that("Printing a summary line for node/edge selections works", {

  graph_all_nodes_selected <-
    create_graph() %>%
    add_path(n = 4) %>%
    select_nodes()

  graph_2_nodes_selected <-
    create_graph() %>%
    add_path(n = 4) %>%
    select_nodes_by_id(nodes = c(1, 2))

  graph_all_edges_selected <-
    create_graph() %>%
    add_path(n = 4) %>%
    select_edges()

  graph_2_edges_selected <-
    create_graph() %>%
    add_path(n = 4) %>%
    select_edges_by_edge_id(edges = c(1, 2))

  expect_equal(
    graph_all_nodes_selected %>%
      get_printed_output(8) %>% substr(1, 45) %>% stringr::str_trim(),
    "SELECTION / 4 nodes selected (ALL)")

  expect_equal(
    graph_2_nodes_selected %>%
      get_printed_output(8) %>% substr(1, 45) %>% stringr::str_trim(),
    "SELECTION / 2 nodes selected")

  expect_equal(
    graph_all_edges_selected %>%
      get_printed_output(8) %>% substr(1, 45) %>% stringr::str_trim(),
    "SELECTION / 3 edges selected (ALL)")

  expect_equal(
    graph_2_edges_selected %>%
      get_printed_output(8) %>% substr(1, 45) %>% stringr::str_trim(),
    "SELECTION / 2 edges selected")
})

test_that("Printing a summary line for graph caches works", {

  graph_no_cache <-
    create_graph() %>%
    add_path(n = 4)

  graph_w_1_cache <-
    create_graph() %>%
    add_path(n = 4) %>%
    set_cache(
      name = "cache",
      to_cache = get_node_ids(.))

  graph_w_2_caches <-
    create_graph() %>%
    add_path(n = 4) %>%
    set_cache(
      name = "cache_1",
      to_cache = get_node_ids(.)) %>%
    set_cache(
      name = "cache_2",
      to_cache = get_node_ids(.) * 2)

  expect_equal(
    graph_no_cache %>%
      get_printed_output(9) %>% substr(1, 45) %>% stringr::str_trim(),
    "CACHE / <none>")

  expect_equal(
    graph_w_1_cache %>%
      get_printed_output(9) %>% substr(1, 45) %>% stringr::str_trim(),
    "CACHE / 1 cache")

  expect_equal(
    graph_w_2_caches %>%
      get_printed_output(9) %>% substr(1, 45) %>% stringr::str_trim(),
    "CACHE / 2 caches")
})

test_that("Printing a summary line for global graph attributes works", {

  graph_default_global_attrs <-
    create_graph()

  graph_no_global_attrs <-
    create_graph(
      attr_theme = NULL)

  expect_equal(
    graph_default_global_attrs %>%
      get_printed_output(10) %>% substr(1, 30) %>% stringr::str_trim(),
    "GLOBAL ATTRS / 17 are set")

  expect_equal(
    graph_no_global_attrs %>%
      get_printed_output(10) %>% substr(1, 30) %>% stringr::str_trim(),
    "GLOBAL ATTRS / <none>")
})

test_that("Printing a summary line for graph actions works", {

  graph_no_actions <-
    create_graph()

  graph_w_1_action <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23) %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_betweenness",
      column_name = "btwns",
      action_name = "get_btwns")

  graph_w_2_actions <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23) %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_pagerank",
      column_name = "pagerank",
      action_name = "get_pagerank") %>%
    add_graph_action(
      fcn = "rescale_node_attrs",
      node_attr_from = "pagerank",
      node_attr_to = "width",
      action_name = "pagerank_to_width")

  expect_equal(
    graph_no_actions %>%
      get_printed_output(11) %>% substr(1, 30) %>% stringr::str_trim(),
    "GRAPH ACTIONS / <none>")

  expect_equal(
    graph_w_1_action %>%
      get_printed_output(11) %>% substr(1, 30) %>% stringr::str_trim(),
    "GRAPH ACTIONS / 1 is set")

  expect_equal(
    graph_w_2_actions %>%
      get_printed_output(11) %>% substr(1, 30) %>% stringr::str_trim(),
    "GRAPH ACTIONS / 2 are set")
})

# Is this thing what we think it is?")

test_that("Detecting a directed or undirected graph is possible", {

  # Create a directed graph
  graph_directed <-
    create_graph(
      directed = TRUE) %>%
    add_path(n = 4)

  # Expect that the graph is directed
  expect_true(
    is_graph_directed(graph_directed))

  # Expect that the graph is not undirected
  expect_false(
    is_graph_undirected(graph_directed))

  # Create an undirected graph
  graph_undirected <-
    create_graph(
      directed = FALSE) %>%
    add_path(n = 4)

  # Expect that the graph is undirected
  expect_true(
    is_graph_undirected(graph_undirected))

  # Expect that the graph is not directed
  expect_false(
    is_graph_directed(graph_undirected))
})

test_that("Detecting the presence of a node is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(
      n = 4,
      label = c(
        "one", "two", "three", "four"))

  # Expect that node `1` is present
  expect_true(
    is_node_present(
      graph = graph,
      node = 1))

  # Expect that node `5` is not present
  expect_false(
    is_node_present(
      graph = graph,
      node = 5))

  # Expect that node with label
  # `one` is present
  expect_true(
    is_node_present(
      graph = graph,
      node = "one"))

  # Expect that node with label
  # `five` is not present
  expect_false(
    is_node_present(
      graph = graph,
      node = "five"))

  # Expect an error if multiple
  # values provided for `node`
  expect_error(
    is_node_present(
      graph = graph,
      node = c(1, 2)))
})

test_that("Detecting the presence of an edge is possible", {

  # Create a graph
  graph <-
    create_graph() %>%
    add_path(
      n = 4,
      label = c(
        "one", "two", "three", "four"))

  # Expect that edge with ID `3`
  # is present
  expect_true(
    is_edge_present(
      graph = graph,
      edge = 3))

  # Expect that the edge with the
  # definition `1` -> `2` is present
  expect_true(
    is_edge_present(
      graph = graph,
      from = 1,
      to = 2))

  # Expect that edge with ID `5`
  # is not present
  expect_false(
    is_edge_present(
      graph = graph,
      edge = 5))

  # Expect that the edge with the
  # definition `4` -> `5` is not present
  expect_false(
    is_edge_present(
      graph = graph,
      from = 4,
      to = 5))

  # Expect that the edge,
  # defined by its labels as
  # `two` -> `three`, exists in
  # the graph
  expect_true(
    is_edge_present(
      graph = graph,
      from = "two",
      to = "three"))

  # Expect an error if values
  # not fully provided for either
  # `from` & `to` or for `edge`
  expect_error(
    is_edge_present(
      graph = graph,
      from = 1))

  expect_error(
    is_edge_present(
      graph = graph,
      to = 2))

  expect_error(
    is_edge_present(
      graph = graph))

  # Expect an error if multiple
  # values provided for `edge`
  expect_error(
    is_edge_present(
      graph = graph,
      edge = c(1, 2)))

  # Expect an error if a non-numeric
  # value provided for `edge`
  expect_error(
    is_edge_present(
      graph = graph,
      edge = "one"))

  # Expect an error if a multiple
  # values provided for either
  # `from` or `to`
  expect_error(
    is_edge_present(
      graph = graph,
      from = 1,
      to = c(2, 3)))

  expect_error(
    is_edge_present(
      graph = graph,
      from = c(1, 2),
      to = 3))

  # Expect an error if not providing
  # an edge ID or an incomplete set
  # of nodes that define an edge
  expect_error(
    is_edge_present(
      graph = graph,
      from = 2))

  # Expect `FALSE` if defining
  # an edge with character labels
  # but both labels do not
  # correspond to any node `label`
  expect_false(
    is_edge_present(
      graph = graph,
      from = "five",
      to = "six"))

  # Expect `FALSE` if defining
  # an edge with character labels
  # but the `from` label does not
  # correspond to any node `label`
  expect_false(
    is_edge_present(
      graph = graph,
      from = "five",
      to = "four"))

  # Expect `FALSE` if defining
  # an edge with character labels
  # but the `to` label does not
  # correspond to any node `label`
  expect_false(
    is_edge_present(
      graph = graph,
      from = "four",
      to = "five"))

  # Create an undirected graph
  graph_undirected <-
    create_graph(
      directed = FALSE) %>%
    add_path(
      n = 4,
      label = c(
        "one", "two", "three", "four"))

  # Expect that edge with ID `3`
  # is present
  expect_true(
    is_edge_present(
      graph = graph_undirected,
      edge = 3))

  # Expect that the edge with the
  # definition `1` -> `2` is present
  expect_true(
    is_edge_present(
      graph = graph_undirected,
      from = 1,
      to = 2))

  # Expect that edge with ID `5`
  # is not present
  expect_false(
    is_edge_present(
      graph = graph_undirected,
      edge = 5))

  # Expect that the edge with the
  # definition `4` -> `5` is not present
  expect_false(
    is_edge_present(
      graph = graph_undirected,
      from = 4,
      to = 5))

  # Expect that the edge,
  # defined by its labels as
  # `two` -> `three`, exists in
  # the graph
  expect_true(
    is_edge_present(
      graph = graph_undirected,
      from = "two",
      to = "three"))
})

test_that("Detecting edge loops is possible", {

  # Create a graph that has multiple
  # loop edges
  graph <-
    create_graph() %>%
    add_path(n = 4) %>%
    add_edge(
      from = 1,
      to = 1) %>%
    add_edge(
      from = 3,
      to = 3)

  # Expect that edge `4` is a loop edge
  expect_true(
    graph %>%
      is_edge_loop(edge = 4))

  # Expect that edge `2` is not a loop edge
  expect_false(
    graph %>%
      is_edge_loop(edge = 2))
})

test_that("Detecting an edge being part of a multiple edge is possible", {

  # Create a graph that has multiple
  # edges across some node pairs
  graph <-
    create_graph() %>%
    add_path(n = 4) %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 3,
      to = 4)

  # Expect that edge `1` is a multiple edge
  expect_true(
    graph %>%
      is_edge_multiple(edge = 1))

  # Expect that edge `2` is not a multiple edge
  expect_false(
    graph %>%
      is_edge_multiple(edge = 2))
})

test_that("Detecting an edge being part of a mutual edge is possible", {

  # Create a graph that has mutual
  # edges across some node pairs
  graph <-
    create_graph() %>%
    add_path(n = 4) %>%
    add_edge(
      from = 4,
      to = 3) %>%
    add_edge(
      from = 2,
      to = 1)

  # Expect that edge `1` has a mutual edge
  expect_true(
    graph %>%
      is_edge_mutual(edge = 1))

  # Expect that edge `2` does not have
  # a mutual edge
  expect_false(
    graph %>%
      is_edge_mutual(edge = 2))
})

test_that("Identifying the graph as a DAG is possible", {

  # Create a graph with 2 nodes (with `type`
  # values) and a single edge (with a `rel`)
  non_dag <-
    create_graph() %>%
    add_cycle(n = 5) %>%
    add_n_nodes(n = 2)

  # Expect that this graph not be classified
  # as a directed acyclic graph (DAG)
  expect_false(
    is_graph_dag(non_dag))

  # Create a graph with a balanced tree and
  # several isolated nodes; this is a DAG
  a_dag <-
    create_graph() %>%
    add_balanced_tree(k = 3, h = 2) %>%
    add_n_nodes(n = 2)

  # Expect that this graph is classified
  # as a directed acyclic graph (DAG)
  expect_true(is_graph_dag(a_dag))

  # If a graph contains no nodes, then
  # that graph is not a DAG
  expect_false(is_graph_dag(create_graph()))
})

test_that("Detecting whether a graph is simple is possible", {

  # Create a graph with 2 cycles
  graph <-
    create_graph() %>%
    add_cycle(n = 4) %>%
    add_cycle(n = 3)

  # Expect that the graph is simple
  expect_true(
    is_graph_simple(graph))

  # Create a graph that has multiple
  # loop edges
  graph <-
    create_graph() %>%
    add_path(n = 4) %>%
    add_edge(
      from = 1,
      to = 1) %>%
    add_edge(
      from = 3,
      to = 3)

  # Expect that the graph is not simple
  expect_false(
    is_graph_simple(graph))

  # Create a graph that has multiple
  # edges across some node pairs
  graph <-
    create_graph() %>%
    add_path(n = 4) %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 3,
      to = 4)

  # Expect that the graph is not simple
  expect_false(
    is_graph_simple(graph))
})

test_that("Identifying the graph as weighted is possible", {

  # Create a graph where the edges
  # have a `weight` attribute
  graph_weighted <-
    create_graph() %>%
    add_cycle(n = 5) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = "weight",
      value = c(3, 5, 2, 9, 6)) %>%
    clear_selection()

  # Expect that this graph is classified
  # as a weighted graph
  expect_true(
    is_graph_weighted(graph_weighted))

  # Create a graph where the edges
  # have a `weight` attribute, but
  # there are some NA values in the
  # `weight` column
  graph_weights_incomplete <-
    create_graph() %>%
    add_cycle(n = 5) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = "weight",
      value = c(3, 5, NA_real_, 9, NA_real_)) %>%
    clear_selection()

  # Expect that this graph won't be
  # classified as a weighted graph since
  # we need a complete set of values
  # assigned to the edges
  expect_false(
    is_graph_weighted(graph_weights_incomplete))

  # Create a graph where the edges
  # have a `weight` attribute, but
  # there the values contained therein
  # are of the `character` class
  graph_weights_chr <-
    create_graph() %>%
    add_cycle(n = 5) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = "weight",
      value = as.character(c(3, 5, 2, 9, 6))) %>%
    clear_selection()

  # Expect that this graph won't be
  # classified as a weighted graph since
  # we the `weight` values to be either
  # as `numeric` or as `integer`
  expect_false(
    is_graph_weighted(graph_weights_chr))

  # Expect that an empty graph will
  # return FALSE
  expect_false(
    is_graph_weighted(create_graph()))

  # Expect that a graph with no edges
  # will return FALSE
  expect_false(
    is_graph_weighted(
      create_graph() %>%
        add_n_nodes(n = 5)))

  # Create a graph that does not have
  # a weight attribute
  graph <-
    create_graph() %>%
    add_path(n = 4) %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 3,
      to = 4)

  # Expect that a graph without a
  # weight attribute will return FALSE
  expect_false(
    is_graph_weighted(graph))
})

test_that("Identifying the graph as a property graph is possible", {

  # Create a graph with 2 nodes (with `type`
  # values) and a single edge (with a `rel`)
  simple_property_graph <-
    create_graph() %>%
    add_node(
      type = "a",
      label = "first") %>%
    add_node(
      type = "b",
      label = "second") %>%
    add_edge(
      from = "first",
      to = "second",
      rel = "rel_1")

  # Expect that this graph be classified
  # as a property graph
  expect_true(
    is_property_graph(simple_property_graph))

  # Create an analogous graph that is not
  # a property graph because it lacks a value
  # for `rel`
  not_a_property_graph <-
    create_graph() %>%
    add_node(
      type = "a",
      label = "first") %>%
    add_node(
      type = "b",
      label = "second") %>%
    add_edge(
      from = "first",
      to = "second")

  # Expect that the `is_property_graph()`
  # function returns FALSE for this graph
  expect_false(
    is_property_graph(not_a_property_graph))

  # Expect that an empty graph will
  # return FALSE
  expect_false(
    is_property_graph(create_graph()))
})

# Getting node IDs in traversal paths

test_that("getting lists of node IDs using `get_paths()` is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 8) %>%
    add_edge(from = 1, to = 2) %>%
    add_edge(from = 1, to = 3) %>%
    add_edge(from = 3, to = 4) %>%
    add_edge(from = 3, to = 5) %>%
    add_edge(from = 4, to = 6) %>%
    add_edge(from = 2, to = 7) %>%
    add_edge(from = 7, to = 5) %>%
    add_edge(from = 4, to = 8)

  # Get a list of all paths outward from node `1`
  paths_from_1 <-
    get_paths(
      graph = graph,
      from = 1)

  # Expect a list of 4 components
  expect_length(
    paths_from_1, 4)

  # Expect specific node IDs in each list
  expect_true(
    all(paths_from_1[[1]] == c(1, 3, 5)))

  expect_true(
    all(paths_from_1[[2]] == c(1, 2, 7, 5)))

  expect_true(
    all(paths_from_1[[3]] == c(1, 3, 4, 6)))

  expect_true(
    all(paths_from_1[[4]] == c(1, 3, 4, 8)))

  # Get a list of all paths leading to node `6`
  paths_to_6 <-
    get_paths(
      graph = graph,
      to = 6)

  # Expect a list of 1 component
  expect_length(
    paths_to_6, 1)

  # Expect specific node IDs in the list
  expect_true(
    all(paths_to_6[[1]] == c(1, 3, 4, 6)))

  #' # Get a list of all paths from `1` to `5`
  paths_from_1_to_5 <-
    get_paths(
      graph = graph,
      from = 1,
      to = 5)

  # Expect a list of 2 components
  expect_length(
    paths_from_1_to_5, 2)

  # Expect specific node IDs in each list
  expect_true(
    all(paths_from_1_to_5[[1]] == c(1, 3, 5)))

  expect_true(
    all(paths_from_1_to_5[[2]] == c(1, 2, 7, 5)))

  # Get a list of all paths from `1` up to a distance
  # of 2 node traversals
  paths_from_1_dist_2 <-
    get_paths(
      graph = graph,
      from = 1,
      distance = 2)

  # Expect a list of 3 components
  expect_length(
    paths_from_1_dist_2, 3)

  # Expect specific node IDs in each list
  expect_true(
    all(paths_from_1_dist_2[[1]] == c(1, 3, 5)))

  expect_true(
    all(paths_from_1_dist_2[[2]] == c(1, 2, 7)))

  expect_true(
    all(paths_from_1_dist_2[[3]] == c(1, 3, 4)))

  # Get a list of the shortest paths from `1` to `5`
  paths_from_1_to_5_shortest <-
    get_paths(
      graph = graph,
      from = 1,
      to = 5,
      shortest_path = TRUE)

  # Expect a list of 1 component
  expect_length(
    paths_from_1_to_5_shortest, 1)

  # Expect specific node IDs for the one list
  expect_true(
    all(paths_from_1_to_5_shortest[[1]] == c(1, 3, 5)))

  # Get a list of the longest paths from `1` to `5`
  paths_from_1_to_5_longest <-
    get_paths(
      graph = graph,
      from = 1,
      to = 5,
      longest_path = TRUE)

  # Expect a list of 1 component
  expect_length(
    paths_from_1_to_5_longest, 1)

  # Expect specific node IDs for the one list
  expect_true(
    all(paths_from_1_to_5_longest[[1]] == c(1, 2, 7, 5)))
})

# Copying, dropping, mutating, renaming, and recoding attrs

test_that("Copying node attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    set_node_attrs(
      node_attr = value,
      values = 1)

  # Make a copy the `value` node attribute as
  # the `value_2` node attribute
  graph <-
    graph %>%
    copy_node_attrs(
      node_attr_from = value,
      node_attr_to = value_2)

  # Verify that the new column has been made
  expect_contains(colnames(graph$nodes_df), "value_2")

  # Expect that the values in `value` and `value_2`
  # are identical
  expect_identical(
    graph$nodes_df$value,
    graph$nodes_df$value_2)

  # Expect an error if `node_attr_from` and
  # `node_attr_to` are identical
  expect_error(
    graph %>%
    copy_node_attrs(
      node_attr_from = value_2,
      node_attr_to = value_2))

  # Expect an error if `node_attr_to` is
  # `nodes` or `node`
  expect_error(
    graph %>%
      copy_node_attrs(
        node_attr_from = value_2,
        node_attr_to = nodes))

  expect_error(
    graph %>%
      copy_node_attrs(
        node_attr_from = value_2,
        node_attr_to = node))

  # Expect an error if `node_attr_from` is not one
  # of the graph's columns
  expect_error(
    graph %>%
    copy_node_attrs(
      node_attr_from = values,
      node_attr_to = value_3))
})

test_that("Copying edge attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    add_edges_w_string(
      edges = "1->3 2->4 1->4 3->2") %>%
    set_edge_attrs(
      edge_attr = value,
      values = 1)

  # Make a copy the `value` node attribute as
  # the `value_2` node attribute
  graph <-
    graph %>%
    copy_edge_attrs(
      edge_attr_from = value,
      edge_attr_to = value_2)

  # Verify that the new column has been made
  expect_contains(colnames(graph$edges_df), "value_2")

  # Expect that the values in `value` and `value_2`
  # are identical
  expect_identical(
    graph$edges_df$value,
    graph$edges_df$value_2)

  # Expect an error if `edge_attr_from` and
  # `edge_attr_to` are identical
  expect_error(
    graph %>%
    copy_edge_attrs(
      edge_attr_from = value_2,
      edge_attr_to = value_2))

  # Expect an error if `edge_attr_to` is
  # `from` or `to`
  expect_error(
    graph %>%
    copy_edge_attrs(
      edge_attr_from = value_2,
      edge_attr_to = from))

  expect_error(
    graph %>%
    copy_edge_attrs(
      edge_attr_from = value_2,
      edge_attr_to = to))

  # Expect an error if `edge_attr_from` is not one
  # of the graph's columns
  expect_error(
    graph %>%
    copy_edge_attrs(
      edge_attr_from = values,
      edge_attr_to = value_3))
})

test_that("Dropping node attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    set_node_attrs(
      node_attr = value,
      values = 1)

  # Expect an error if a column to drop does not exist
  expect_error(
    graph %>%
    drop_node_attrs(
      node_attr = value_2))

  # Drop the `value` node attribute
  graph <-
    graph %>%
    drop_node_attrs(node_attr = value)

  # Verify that the `value` column is not present
  expect_false(
    "value" %in% colnames(graph$nodes_df))

  # Expect an error if `node_attr` is any of
  # `nodes`, `node`, `type`, or `label`
  expect_error(
    drop_node_attrs(graph, node_attr = nodes))

  expect_error(
    drop_node_attrs(graph, node_attr = node))

  expect_error(
    drop_node_attrs(graph, node_attr = type))

  expect_error(
    drop_node_attrs(graph, node_attr = label))
})

test_that("Dropping edge attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    add_edges_w_string(
      edges = "1->3 2->4 1->4 3->2") %>%
    set_edge_attrs(
      edge_attr = value,
      values = 1)

  # Expect an error if a column to drop does not exist
  expect_error(
    graph %>%
    drop_edge_attrs(edge_attr = value_2))

  # Drop the `value` edge attribute
  graph <-
    graph %>%
    drop_edge_attrs(edge_attr = value)

  # Verify that the `value` column is not present
  expect_false(
    "value" %in% colnames(graph$edges_df))

  # Expect an error if `edge_attr` is any of
  # `from`, `to`, or `rel`
  expect_error(
    drop_edge_attrs(graph, edge_attr = from))

  expect_error(
    drop_edge_attrs(graph, edge_attr = to))

  expect_error(
    drop_edge_attrs(graph, edge_attr = rel))
})

test_that("Renaming node attributes is possible", {

  # Create a random graph with extra node attrs
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23) %>%
    set_node_attrs(
      node_attr = shape,
      values = "circle")

  # Rename the `shape` attribute as `color`
  graph <-
    graph %>%
    rename_node_attrs(
      node_attr_from = shape,
      node_attr_to = color)

  # Expect that the `color` node attr is in the
  # graph's node data frame
  expect_contains(colnames(graph$nodes_df), "color")

  # Expect that the `shape` node attr is not in the
  # graph's node data frame
  expect_false(
    "shape" %in% colnames(graph$nodes_df))

  # Expect an error if trying to rename a node
  # attribute that doesn't exist in the graph
  expect_error(
    graph %>%
    rename_node_attrs(
      node_attr_from = perhipheries,
      node_attr_to = shape))

  # Expect an error if `node_attr_from` and
  # `node_attr_to` are identical
  expect_error(
    graph %>%
    rename_node_attrs(
      node_attr_from = color,
      node_attr_to = color))

  # Expect an error if `node_attr_to` is `id`
  expect_error(
    graph %>%
    rename_node_attrs(
      node_attr_from = color,
      node_attr_to = id))

  # Expect an error if `node_attr_to` is any of
  # the existing node attrs
  expect_error(
    graph %>%
    rename_node_attrs(
      node_attr_from = value,
      node_attr_to = color))

  # Expect an error if `node_attr_from` is any of
  # `id`, `type`, or `label`
  expect_error(
    graph %>%
    rename_node_attrs(
      node_attr_from = id,
      node_attr_to = node_id))

  expect_error(
    graph %>%
    rename_node_attrs(
      node_attr_from = type,
      node_attr_to = type_2))

  expect_error(
    graph %>%
    rename_node_attrs(
      node_attr_from = label,
      node_attr_to = label_2))
})

test_that("Renaming edge attributes is possible", {

  # Create a random graph with extra edge attrs
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 5,
      m = 10,
      set_seed = 23) %>%
    set_edge_attrs(
      edge_attr = penwidth,
      values = 5) %>%
    set_edge_attrs(
      edge_attr = arrowhead,
      values = "dot")

  # Rename the `penwidth` attribute as `width`
  graph <-
    graph %>%
    rename_edge_attrs(
      edge_attr_from = penwidth,
      edge_attr_to = width)

  # Expect that the `width` node attr is in the
  # graph's edge data frame
  expect_contains(colnames(graph$edges_df), "width")

  # Expect that the `penwidth` node attr is not in the
  # graph's node data frame
  expect_false(
    "penwidth" %in% colnames(graph$edges_df))

  # Expect an error if trying to rename an edge
  # attribute that doesn't exist in the graph
  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = perhipheries,
      edge_attr_to = shape))

  # Expect an error if `edge_attr_from` and
  # `edge_attr_to` are identical
  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = width,
      edge_attr_to = width))

  # Expect an error if `node_attr_to` is `from` or `to`
  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = width,
      edge_attr_to = from))

  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = width,
      edge_attr_to = to))

  # Expect an error if `edge_attr_to` is any of
  # the existing edge attrs
  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = width,
      edge_attr_to = arrowhead))

  # Expect an error if `edge_attr_from` is any of
  # `from`, `to`, or `rel`
  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = from,
      edge_attr_to = from_2))

  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = to,
      edge_attr_to = to_2))

  expect_error(
    graph %>%
    rename_edge_attrs(
      edge_attr_from = rel,
      edge_attr_to = rel_2))
})

test_that("Mutating node attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    set_node_attrs(
      node_attr = value,
      values = 1) %>%
    set_node_attrs(
      node_attr = shape,
      values = "square")

  # Mutate the `value` node attribute
  graph <-
    graph %>%
    mutate_node_attrs(
      value = value * 2)

  # Expect each value in `value` to be 2
  expect_equal(
    graph$nodes_df$value,
    c(2, 2))

  # Create node attribute `value_2`
  graph <-
    graph %>%
    mutate_node_attrs(
      value_2 = value * 2)

  # Verify that the new column has been made
  expect_contains(colnames(graph$nodes_df), "value_2")

  # Expect each value in `value_2` to be 4
  expect_equal(
    graph$nodes_df$value_2,
    c(4, 4))

  # Expect an error if `node_attr_to` is `id`
  expect_error(
    mutate_node_attrs(
      id = value * 2))

  # Overwrite `value_2` with new values
  graph <-
    graph %>%
    mutate_node_attrs(
      value_2 = value * 4)

  # Expect each value in `value_2` to be 8
  expect_equal(
    graph$nodes_df$value_2,
    c(8, 8))
})

test_that("Mutating edge attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    add_edges_w_string(
      edges = "1->3 2->4 1->4 3->2") %>%
    set_edge_attrs(
      edge_attr = value,
      values = 1) %>%
    set_edge_attrs(
      edge_attr = color,
      values = "red")

  # Mutate the `value` edge attribute
  graph <-
    graph %>%
    mutate_edge_attrs(
      value = value * 2)

  # Expect each value in `value` to be 2
  expect_equal(
    graph$edges_df$value,
    c(2, 2, 2, 2))

  # Create node attribute `value_2`
  graph <-
    graph %>%
    mutate_edge_attrs(
      value_2 = value * 2)

  # Verify that the new column has been made
  expect_contains(colnames(graph$edges_df), "value_2")

  # Expect each value in `value_2` to be 4
  expect_equal(
    graph$edges_df$value_2,
    c(4, 4, 4, 4))

  # Expect an error if `edge_attr_to`
  # is `id`, `from`, or `to`
  expect_error(
    mutate_edge_attrs(
      id = value * 2))

  expect_error(
    mutate_edge_attrs(
      from = value * 2))

  expect_error(
    mutate_edge_attrs(
      to = value * 2))

  # Overwrite `value_2` with new values
  graph <-
    graph %>%
    mutate_edge_attrs(
      value_2 = value * 4)

  # Expect each value in `value_2` to be 8
  expect_equal(
    graph$edges_df$value_2,
    c(8, 8, 8, 8))
})

test_that("Mutating node attributes with a selection is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    set_node_attrs(
      node_attr = value,
      values = 1) %>%
    set_node_attrs(
      node_attr = shape,
      values = "square") %>%
    select_nodes_by_id(nodes = 1)

  # Mutate the `value` node attribute
  # for the selected node
  graph <-
    graph %>%
    mutate_node_attrs_ws(
      value = value * 2)

  # Expect specific values in `value`
  expect_equal(
    graph$nodes_df$value,
    c(2, 1))
})

test_that("Mutating edge attributes with a selection is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    add_edges_w_string(
      edges = "1->3 2->4 1->4 3->2") %>%
    set_edge_attrs(
      edge_attr = value,
      values = 1) %>%
    set_edge_attrs(
      edge_attr = color,
      values = "red") %>%
    select_edges_by_edge_id(edges = 1)

  # Mutate the `value` edge attribute
  # for the selected edge
  graph <-
    graph %>%
    mutate_edge_attrs_ws(
      value = value * 2)

  # Expect specific values in `value`
  expect_equal(
    graph$edges_df$value,
    c(2, 1, 1, 1))
})

test_that("Recoding node attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    set_node_attrs(
      node_attr = value,
      values = 1) %>%
    set_node_attrs(
      node_attr = shape,
      values = "square")

  # Expect an error if `node_attr_from`
  # is not one of the graph's columns
  expect_error(
    recode_node_attrs(
      graph, "value_2", "1->2"))

  # Recode the `value` node attribute
  graph <-
    recode_node_attrs(
      graph, "value", "1->2")

  # Expect each value in `value` to be 2
  expect_equal(
    graph$nodes_df$value,
    c(2, 2))

  # Recode the `value` node attribute again but
  # also copy recoded values into `value_2`
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = value,
      "2->4",
      node_attr_to = value_2)

  # Verify that the new column has been made
  expect_contains(colnames(graph$nodes_df), "value_2")

  # Expect each value in `value_2` to be 4
  expect_equal(
    graph$nodes_df$value_2,
    c(4, 4))

  # Expect an error if `node_attr_to` is `id` or `node`
  expect_error(
    graph %>%
    recode_node_attrs(
      node_attr_from = value,
      "2->4",
      node_attr_to = id))

  expect_error(
    graph %>%
    recode_node_attrs(
      node_attr_from = value,
      "2->4",
      node_attr_to = nodes))

  # Recode the `value` node attribute and
  # copy and overwrite recoded values into `value_2`
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = value,
      "2->8",
      node_attr_to = value_2)

  # Expect each value in `value_2` to be 8
  expect_equal(
    graph$nodes_df$value_2,
    c(8, 8))

  # Create another graph but use different node
  # attribute values
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    set_node_attrs(
      node_attr = value,
      values = c(1, 1, 2, 3)) %>%
    set_node_attrs(
      node_attr = shape,
      values = c("square", "square",
                 "circle", "triangle"))

  # Recode the `value` node attribute, using the
  # `otherwise` argument
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = value,
      "1->2",
      otherwise = 4)

  # Expect values in `value` to be 2, 2, 4, 4
  expect_equal(
    graph$nodes_df$value,
    c(2, 2, 4, 4))

  # Recode the `value` node attribute again but
  # also copy recoded values into `value_2`
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = value,
      "2->4",
      otherwise = 8,
      node_attr_to = value_2)

  # Expect values in `value_2` to be 4, 4, 8, 8
  expect_equal(
    graph$nodes_df$value_2,
    c(4, 4, 8, 8))

  # Recode the `value` node attribute and
  # copy and overwrite recoded values into `value_2`
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = value,
      "2->8",
      otherwise = 16,
      node_attr_to = value_2)

  # Expect values in `value_2` to be 8, 8, 16, 16
  expect_equal(
    graph$nodes_df$value_2,
    c(8, 8, 16, 16))

  # Recode the `shape` node attribute
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = shape,
      "square -> rectangle")

  # Expect certain values in `shape`
  expect_equal(
    graph$nodes_df$shape,
    c("rectangle", "rectangle",
      "circle", "triangle"))

  # Recode the `shape` node attribute using the
  # `otherwise` argument
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = shape,
      "rectangle -> circle",
      "circle -> square",
      otherwise = "diamond")

  # Expect certain values in `shape`
  expect_equal(
    graph$nodes_df$shape,
    c("circle", "circle",
      "square", "diamond"))

  # Recode the `shape` node attribute and copy into
  # results to `shape_2`
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = shape,
      "circle -> rectangle",
      "square -> circle",
      node_attr_to = shape_2)

  # Expect certain values in `shape_2`
  expect_equal(
    graph$nodes_df$shape_2,
    c("rectangle", "rectangle",
      "circle", "diamond"))

  # Recode the `shape` node attribute and overwrite
  # results in `shape_2`
  graph <-
    graph %>%
    recode_node_attrs(
      node_attr_from = shape,
      "circle -> rectangle",
      "square -> circle",
      otherwise = "triangle",
      node_attr_to = shape_2)

  # Expect certain values in `shape_2`
  expect_equal(
    graph$nodes_df$shape_2,
    c("rectangle", "rectangle",
      "circle", "triangle"))
})

test_that("Recoding edge attributes is possible", {

  # Create a starting graph
  graph <-
    create_graph() %>%
    add_n_nodes(n = 4) %>%
    add_edges_w_string(
      edges = "1->3 2->4 1->4 3->2") %>%
    set_edge_attrs(
      edge_attr = value,
      values = c(1, 1, 2, 3)) %>%
    set_edge_attrs(
      edge_attr = color,
      values = c("red", "red",
                 "blue", "black"))

  # Expect an error if `edge_attr_from` is not one
  # of the graph's columns
  expect_error(
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value_2,
      "1->2"))

  # Recode the `value` edge attribute
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value,
      "1->2")

  # Expect values in `value` to be 2, 2, 2, 3
  expect_equal(
    graph$edges_df$value,
    c(2, 2, 2, 3))

  # Recode the `value` edge attribute again but
  # also copy recoded values into `value_2`
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value,
      "2->4",
      edge_attr_to = value_2)

  # Verify that the new column has been made
  expect_contains(colnames(graph$edges_df), "value_2")

  # Expect certain values in `value_2`
  expect_equal(
    graph$edges_df$value_2,
    c(4, 4, 4, 3))

  # Expect an error if `edge_attr_to` is `from` or `to`
  expect_error(
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value,
      "2->4",
      edge_attr_to = from))

  expect_error(
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value,
      "2->4",
      edge_attr_to = to))

  # Recode the `value` edge attribute and
  # copy and overwrite recoded values into `value_2`
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value,
      "2->8",
      edge_attr_to = value_2)

  # Expect certain values in `value_2`
  expect_equal(
    graph$edges_df$value_2,
    c(8, 8, 8, 3))

  # Recode the `value` edge attribute, using the
  # `otherwise` argument
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value,
      "2->4",
      otherwise = 8)

  # Expect certain values in `value`
  expect_equal(
    graph$edges_df$value,
    c(4, 4, 4, 8))

  # Recode the `value` edge attribute again but
  # also copy recoded values into `value_3`
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = value,
      "4->8",
      otherwise = 16,
      edge_attr_to = value_3)

  # Expect certain values in `value_3`
  expect_equal(
    graph$edges_df$value_3,
    c(8, 8, 8, 16))

  # Recode the `value` edge attribute and
  # copy and overwrite recoded values into `value_3`
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = "value",
      "4->12",
      otherwise = 16,
      edge_attr_to = value_3)

  # Expect certain values in `value_3`
  expect_equal(
    graph$edges_df$value_3,
    c(12, 12, 12, 16))

  # Recode the `color` edge attribute
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = color,
      "red->gray")

  # Expect certain values in `color`
  expect_equal(
    graph$edges_df$color,
    c("gray", "gray",
      "blue", "black"))

  # Recode the `color` edge attribute using the
  # `otherwise` argument
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = color,
      "gray -> blue",
      "blue -> green",
      otherwise = "yellow")

  # Expect certain values in `color`
  expect_equal(
    graph$edges_df$color,
    c("blue", "blue",
      "green", "yellow"))

  # Recode the `color` edge attribute and copy into
  # results to `color_2`
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = color,
      "blue -> yellow",
      "green -> blue",
      edge_attr_to = color_2)

  # Expect certain values in `color_2`
  expect_equal(
    graph$edges_df$color_2,
    c("yellow", "yellow",
      "blue", "yellow"))

  # Recode the `color` node attribute and overwrite
  # results in `color_2`
  graph <-
    graph %>%
    recode_edge_attrs(
      edge_attr_from = color,
      "blue -> green",
      otherwise = "black",
      edge_attr_to = color_2)

  # Expect certain values in `color_2`
  expect_equal(
    graph$edges_df$color_2,
    c("green", "green",
      "black", "black"))
})

#' Add a clone of an existing edge to the graph
#'
#' @description
#'
#' Add a new edge to a graph object of class `dgr_graph` which is a clone of an
#' edge already in the graph. All edge attributes are preserved.
#'
#' @inheritParams render_graph
#' @param edge An edge ID corresponding to the graph edge to be cloned.
#' @param from The outgoing node from which the edge is connected.
#' @param to The incoming nodes to which each edge is connected.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with a path of
#' # 2 nodes; supply a common `rel`
#' # edge attribute for all edges
#' # in this path and then add a
#' # `color` edge attribute
#' graph <-
#'   create_graph() %>%
#'   add_path(
#'     n = 2,
#'     rel = "a") %>%
#'   select_last_edges_created() %>%
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "steelblue") %>%
#'   clear_selection()
#'
#' # Display the graph's internal
#' # edge data frame
#' graph %>% get_edge_df()
#'
#' # Create a new node (will have
#' # node ID of `3`) and then
#' # create an edge between it and
#' # node `1` while reusing the edge
#' # attributes of edge `1` -> `2`
#' # (edge ID `1`)
#' graph_2 <-
#'   graph %>%
#'   add_node() %>%
#'   add_edge_clone(
#'     edge = 1,
#'     from = 3,
#'       to = 1)
#'
#' # Display the graph's internal
#' # edge data frame
#' graph_2 %>% get_edge_df()
#'
#' # The same change can be performed
#' # with some helper functions in the
#' # `add_edge_clone()` function call
#' graph_3 <-
#'   graph %>%
#'     add_node() %>%
#'     add_edge_clone(
#'       edge = get_last_edges_created(.),
#'       from = get_last_nodes_created(.),
#'       to = 1)
#'
#' # Display the graph's internal
#' # edge data frame
#' graph_3 %>% get_edge_df()
#'
#' @family edge creation and removal
#' @export
add_edge_clone <- function(
    graph,
    edge,
    from,
    to
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Stop function if edge is not a single numerical value
  check_number_decimal(edge)

  # Stop function the edge ID does not correspond
  # to an edge in the graph
  if (!(edge %in% graph$edges_df$id)) {

    cli::cli_abort(
      "The value provided in `edge` does not correspond to an edge in the graph.")
  }

  # Get the value for the latest `version_id` for
  # graph (in the `graph_log`)
  current_graph_log_version_id <-
    max(graph$graph_log$version_id)

  # Get the number of columns in the graph's
  # internal edge data frame
  n_col_edf <-
    graph %>%
    get_edge_df() %>%
    ncol()

  # Extract all of the edge attributes
  # (`rel` and additional edge attrs)
  edge_attr_vals <-
    graph %>%
    get_edge_df() %>%
    dplyr::filter(id == edge) %>%
    dplyr::select(4:dplyr::all_of(n_col_edf))

  # Create the requested edge
  graph <-
    graph %>%
    add_edge(
      from = from,
      to = to)

  # Create an edge selection for the
  # new edge in the graph
  graph <-
    select_last_edges_created(graph)

  # Iteratively set edge attribute values for
  # the new edges in the graph
  for (i in seq_len(ncol(edge_attr_vals))) {

    graph$edges_df[
      nrow(graph$edges_df),
      which(colnames(graph$edges_df) == colnames(edge_attr_vals)[i])] <-
      edge_attr_vals[[i]]
  }

  # Clear the graph's active selection
  graph <-
    suppressMessages(
      clear_selection(graph))

  # Remove extra items from the `graph_log`
  graph$graph_log <-
    graph$graph_log %>%
    dplyr::filter(version_id <= current_graph_log_version_id)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = 1)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Reverse the direction of all edges in a graph
#'
#' @description
#'
#' Using a directed graph as input, reverse the direction of all edges in that
#' graph.
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with a
#' # directed tree
#' graph <-
#'   create_graph() |>
#'   add_balanced_tree(
#'     k = 2, h = 2)
#'
#' # Inspect the graph's edges
#' graph |> get_edges()
#'
#' # Reverse the edge directions
#' # such that edges are directed
#' # toward the root of the tree
#' graph <-
#'   graph |>
#'   rev_edge_dir()
#'
#' # Inspect the graph's edges
#' # after their reversal
#' graph |> get_edges()
#'
#' @family edge creation and removal
#'
#' @export
rev_edge_dir <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # If graph is undirected, stop function
  if (!graph$directed) {

    cli::cli_abort("The input graph must be a directed graph.")
  }

  # Get the graph nodes in the `from` and `to` columns
  # of the edf
  from <- get_edges(graph, return_type = "df")[, 1]
  to <- get_edges(graph, return_type = "df")[, 2]

  # Extract the graph's edge data frame
  edges <- get_edge_df(graph)

  # Replace the contents of the `from` and `to` columns
  edges$from <- to
  edges$to <- from

  # Modify the graph object
  graph$edges_df <- edges

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Delete all loop edges associated with a selection of nodes
#'
#' @description
#'
#' With a selection of nodes in a graph, remove any associated loop edges.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an undirected, full graph
#' # of 5 nodes with loops retained
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_full_graph(
#'     n = 5,
#'     keep_loops = TRUE)
#'
#' # Select nodes `3` and `4`
#' # and remove the loop edges
#' # associated with those nodes
#' graph <-
#'   graph |>
#'   select_nodes_by_id(
#'     nodes = 3:4) |>
#'   delete_loop_edges_ws()
#'
#' # Count the number of loop
#' # edges remaining in the graph
#' graph |> count_loop_edges()
#'
#' @family edge creation and removal
#'
#' @export
delete_loop_edges_ws <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, "So, there cannot be edges to delete.")

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # Get the graph's edf
  edf <- graph$edges_df

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Filter edf such that any loop edges
  # associated with the selected nodes
  # are removed
  selected_nodes <- suppressMessages(get_selection(graph))

  edges_to_remove <-
    selected_nodes |>
    purrr::map_df(
      .f = function(x) {
        edf |>
          dplyr::filter(
            (from == x &
               to == x))
      })

  edf_replacement <-
    dplyr::setdiff(edf, edges_to_remove)

  # Update the graph's edf
  graph$edges_df <- edf_replacement

  # Scavenge any invalid, linked data frames
  graph <-
    graph |>
    remove_linked_dfs()

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph |> count_edges()

  # Get the number of edges added to
  # the graph
  edges_removed <- edges_graph_2 - edges_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_removed)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Add one or several clones of an existing node to the graph
#'
#' @description
#'
#' Add `n` new nodes to a graph object of class `dgr_graph` which are clones of
#' a node already in the graph. All node attributes are preserved except for the
#' node `label` attribute (to maintain the uniqueness of non-`NA` node label
#' values). A vector of node `label` can be provided to bind new labels to the
#' cloned nodes.
#'
#' @inheritParams render_graph
#' @param n The number of node clones to add to the graph.
#' @param node A node ID corresponding to the graph node to be cloned.
#' @param label An optional vector of node label values. The vector length
#'   should correspond to the value set for `n`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with a path of
#' # nodes; supply `label`, `type`,
#' # and `value` node attributes
#' graph <-
#'   create_graph() |>
#'   add_path(
#'     n = 3,
#'     label = c("d", "g", "r"),
#'     type = c("a", "b", "c"))
#'
#' # Display the graph's internal
#' # node data frame
#' graph |> get_node_df()
#'
#' # Create 3 clones of node `1`
#' # but assign new node label
#' # values (leaving `label` as
#' # NULL yields NA values)
#' graph <-
#'   graph |>
#'   add_n_node_clones(
#'     n = 3,
#'     node = 1,
#'     label = c("x", "y", "z"))
#'
#' # Display the graph's internal
#' # node data frame: nodes `4`,
#' # `5`, and `6` are clones of `1`
#' graph |> get_node_df()
#'
#' @family node creation and removal
#'
#' @export
add_n_node_clones <- function(
    graph,
    n,
    node,
    label = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, extra_msg = "So, clones of nodes cannot be added.")

  # Stop function if node is not a single numerical value
  check_number_decimal(node)

  # Stop function the node ID does not correspond
  # to a node in the graph
  if (!(node %in% graph$nodes_df$id)) {

    cli::cli_abort(c(
      "The value of `node` ({node}) must correspond to a node in the graph.",
      "Any of {.or {unique(graph$nodes_df$id)}} is acceptable.")
      )
  }

  # Stop function if vector provided for label but it
  # is not of length `n`
  if (!is.null(label)) {

    # a check_length should exist soon in rlang
    # https://github.com/r-lib/rlang/issues/1618
    check_number_whole(length(label), min = n, max = n)
  }

  # Get the value for the latest `version_id` for
  # graph (in the `graph_log`)
  current_graph_log_version_id <-
    max(graph$graph_log$version_id)

  # Get the number of columns in the graph's
  # internal node data frame
  n_col_ndf <-
    graph |>
    get_node_df() |>
    ncol()

  # Extract all of the node attributes
  # (`type` and additional node attrs)
  if (n_col_ndf >= 4) {

    node_attr_vals <-
      graph |>
      get_node_df() |>
      dplyr::filter(id == node) |>
      dplyr::select("type", 4:dplyr::all_of(n_col_ndf))
  }

  # Create one or more clones of
  # the selected node in the graph
  group_id <- as.character(
    get_node_attrs(graph = graph, node_attr = type, nodes = node))

  graph <-
    graph |>
    add_n_nodes(
      n = n,
      type = group_id,
      label = label)

  # Obtain the node ID values for
  # the new nodes
  new_node_ids <-
    suppressMessages(
      graph |>
        select_last_nodes_created() |>
        get_selection())

  # Create a node selection for the
  # new nodes in the graph
  graph <-
    graph |>
    select_nodes_by_id(
      nodes = new_node_ids)

  # Iteratively set node attribute values for
  # the new nodes in the graph

  if (exists("node_attr_vals")) {

    for (i in seq_len(ncol(node_attr_vals))) {
      for (j in seq_along(new_node_ids)) {

        graph$nodes_df[
          which(graph$nodes_df[, 1] == new_node_ids[j]),
          which(colnames(graph$nodes_df) == colnames(node_attr_vals)[i])] <-
          node_attr_vals[[i]]
      }
    }
  }

  # Clear the graph's active selection
  graph <-
    suppressMessages(
      clear_selection(graph))

  # Remove extra items from the `graph_log`
  graph$graph_log <-
    graph$graph_log |>
    dplyr::filter(version_id <= current_graph_log_version_id)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = n)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Fully connect all nodes in a selection of nodes
#'
#' @description
#'
#' With a selection of nodes in a graph, add any remaining edges required to
#' fully connect this group of edges to each other.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph and
#' # then add a path of 3 nodes
#' # and two isolated nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 3) %>%
#'   add_n_nodes(n = 2)
#'
#' # Select a node in the path
#' # of nodes (node `3`) and
#' # the two isolated nodes (`4`
#' # and `5`); then, and fully
#' # connect these nodes together
#' graph <-
#'   graph %>%
#'   select_nodes_by_id(
#'     nodes = 3:5) %>%
#'   fully_connect_nodes_ws()
#'
#' # Get the graph's edge data frame
#' graph %>% get_edge_df()
#'
#' # Create an undirected, empty
#' # graph; add a path of 3 nodes
#' # and two isolated nodes
#' graph <-
#'   create_graph(
#'     directed = FALSE) %>%
#'   add_path(n = 3) %>%
#'   add_n_nodes(n = 2)
#'
#' # Select a node in the path
#' # of nodes (node `3`) and
#' # the two isolated nodes (`4`
#' # and `5`); then, and fully
#' # connect these nodes together
#' graph <-
#'   graph %>%
#'   select_nodes_by_id(
#'     nodes = 3:5) %>%
#'   fully_connect_nodes_ws()
#'
#' # Get the graph's edge data
#' # frame; in the undirected
#' # case, reverse edges aren't
#' # added
#' graph %>% get_edge_df()
#'
#' @export
fully_connect_nodes_ws <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, "So, there are no nodes to connect.")

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # Get the number of edges in the graph
  edges_graph_1 <- graph %>% count_edges()

  # Get the graph's edf
  edf <- graph$edges_df

  # Get the combination of edges
  edge_candidates <-
    utils::combn(suppressMessages(get_selection(graph = graph)), 2) %>%
    t() %>%
    as.data.frame() %>%
    dplyr::rename(from = "V1", to = "V2")

  # Determine the complete set of edges
  # to add to the graph

  if (is_graph_directed(graph)) {

    edges_to_add <-
      dplyr::setdiff(
        dplyr::bind_rows(
          edge_candidates,
          edge_candidates %>%
            dplyr::select(to, from) %>%
            dplyr::rename(from = "to", to = "from")),
        edf %>%
          dplyr::select("from", "to"))

  } else {

    edges_to_add <-
      dplyr::setdiff(
        edge_candidates,
        edf %>%
          dplyr::select("from", "to"))
  }

  # Add new edges to the graph for every edge
  # in the `edges_to_add` df
  if (nrow(edges_to_add) > 0) {
    for (i in seq_len(nrow(edges_to_add))) {

      # Create a graph edge
      graph <-
        add_edge(
          graph = graph,
          from = edges_to_add[i, 1],
          to = edges_to_add[i, 2])

      # Redact the signing of the action to the log
      graph$graph_log <-
        graph$graph_log[-nrow(graph$graph_log), ]
    }
  }

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get the count of multiple edges
#'
#' @description
#'
#' Get a count of the number of multiple edges in the graph. Included in the
#' count is the number of separate edges that share the same edge definition
#' (i.e., same pair of nodes) across the entire graph. So, for example, if there
#' are 2 edge definitions in the graph that involve 6 separate edge IDs, the
#' count will be `4`.
#'
#' @inheritParams render_graph
#'
#' @return A vector with a single, numerical value.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 5,
#'     label = TRUE)
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 4, 4, 3, 5, 1, 3, 4),
#'       to = c(4, 1, 1, 2, 2, 2, 2, 1))
#'
#' # Create a graph with the ndf and edf
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Get the total number of multiple
#' # edges (those edges that share an
#' # edge definition) in the graph
#' graph |> get_multiedge_count()
#'
#' @export
get_multiedge_count <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Check for the number of multiple edges
  # regardless of which definitions these
  # edges have
  multiedge_count <-
    (graph$edges_df |>
       nrow()) -
    (graph$edges_df |>
       dplyr::distinct(from, to) |>
       nrow())

  multiedge_count
}

#' Reorder the stored series of graph actions
#'
#' @description
#'
#' Reorder the graph actions stored in the graph through the use of the
#' [add_graph_action()] function. These actions are be invoked in a specified
#' order via the [trigger_graph_actions()] function.
#'
#' @inheritParams render_graph
#' @param indices A numeric vector that provides the new ordering of graph
#'   actions. This vector can be the same length as the number of graph actions,
#'   or, of shorter length. In the latter case, the ordering places the given
#'   items first and the remaining actions will follow.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 4,
#'     m = 4,
#'     set_seed = 23)
#'
#' # Add three graph actions to the
#' # graph
#' graph <-
#'   graph %>%
#'   add_graph_action(
#'     fcn = "rescale_node_attrs",
#'     node_attr_from = "pagerank",
#'     node_attr_to = "width",
#'     action_name = "pgrnk_to_width") %>%
#'   add_graph_action(
#'     fcn = "set_node_attr_w_fcn",
#'     node_attr_fcn = "get_pagerank",
#'     column_name = "pagerank",
#'     action_name = "get_pagerank") %>%
#'   add_graph_action(
#'     fcn = "colorize_node_attrs",
#'     node_attr_from = "width",
#'     node_attr_to = "fillcolor",
#'     action_name = "pgrnk_fillcolor")
#'
#' # View the graph actions for the graph
#' # object by using the function called
#' # `get_graph_actions()`
#' graph %>% get_graph_actions()
#'
#' # We note that the order isn't
#' # correct and that the `get_pagerank`
#' # action should be the 1st action
#' # and `pgrnk_to_width` should go
#' # in 2nd place; to fix this, use the
#' # function `reorder_graph_actions()`
#' # and specify the reordering with a
#' # numeric vector
#' graph <-
#'   graph %>%
#'   reorder_graph_actions(
#'     indices = c(2, 1, 3))
#'
#' # View the graph actions for the graph
#' # object once again to verify that
#' # we have the desired order of actions
#' graph %>% get_graph_actions()
#'
#' @export
reorder_graph_actions <- function(
    graph,
    indices
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Determine whether there any
  # available graph actions
  if (nrow(graph$graph_actions) == 0) {

    cli::cli_abort(
      "There are no graph actions to reorder.")
  }

  # Get the `action_index` values
  # available in `graph$graph_actions`
  available_indices <-
    graph %>%
    get_graph_actions() %>%
    dplyr::pull("action_index")

  # Verify that the provided values
  # do not refer to an `action_index`
  # that does not exist
  if (!any(indices %in% available_indices)) {

    cli::cli_abort(
      "One or more provided indices do not exist in the graph.")
  }

  remaining_indices <-
    which(!(available_indices %in% indices))

  revised_indices <-
    c(indices, remaining_indices)

  # Extract the graph actions table from
  # the graph
  graph_actions_tbl <-
    get_graph_actions(graph)

  # Get a revised data frame with graph actions
  # in the requested order
  revised_graph_actions <-
    graph_actions_tbl[revised_indices, ] %>%
    dplyr::mutate(action_index = dplyr::row_number())

  # Replace `graph$graph_actions` with the
  # revised version
  graph$graph_actions <- revised_graph_actions

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

# Using graph actions")

test_that("actions can be added to a graph object", {

  # Create an empty graph object
  graph <- create_graph()

  # Add a graph action that sets a node
  # attr column with a function; the
  # main function `set_node_attr_w_fcn()`
  # uses the `get_betweenness()` function
  # to provide betweenness values in the
  # `btwns` column
  graph <-
    graph %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_betweenness",
      column_name = "btwns",
      action_name = "get_btwns")

  # Expect a `data.frame` object with in
  # `graph$graph_actions`
  expect_s3_class(
    graph$graph_actions, "data.frame")

  # Extract `graph$graph_actions` to a
  # separate object
  graph_actions <-
    graph$graph_actions

  # Expect a single row in the data frame
  expect_equal(
    nrow(graph_actions), 1)

  # Expect three columns in the data frame
  expect_equal(
    ncol(graph_actions), 3)

  # Expect certain column names in the
  # data frame object
  expect_equal(
    colnames(graph_actions),
    c("action_index", "action_name", "expression"))

  # Expect the `action_index` to be 1
  expect_equal(
    graph_actions$action_index, 1)

  # Expect the `action_name` to be `get_btwns`
  expect_equal(
    graph_actions$action_name, "get_btwns")

  # Expect the action in the data frame to
  # be correctly generated
  expect_equal(
    graph_actions$expression,
    "set_node_attr_w_fcn(graph = graph, node_attr_fcn = 'get_betweenness', column_name = 'btwns')")
})

test_that("actions can be deleted from a graph object", {

  # Create an empty graph object
  graph <- create_graph()

  # Expect an error when trying to
  # delete graph actions and none exist
  expect_error(
    graph %>%
      delete_graph_actions(
        actions = 1))

  # Add three graph actions to the
  # graph
  graph <-
    graph %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_pagerank",
      column_name = "pagerank",
      action_name = "get_pagerank") %>%
    add_graph_action(
      fcn = "rescale_node_attrs",
      node_attr_from = "pagerank",
      node_attr_to = "width",
      action_name = "pagerank_to_width") %>%
    add_graph_action(
      fcn = "colorize_node_attrs",
      node_attr_from = "width",
      node_attr_to = "fillcolor",
      action_name = "pagerank_fillcolor")

  n_graph_actions_before_deletion <-
    nrow(graph$graph_actions)

  # Delete two of the graph actions
  graph_delete_2 <-
    graph %>%
    delete_graph_actions(
      actions = c(2, 3))

  # Expect that one graph action remains
  # Expect a single row in the data frame
  expect_equal(
    nrow(graph_delete_2$graph_actions),
    n_graph_actions_before_deletion - 2)

  # Expect that the first graph action
  # remains in the graph
  # Expect the `action_index` to be 1
  expect_equal(
    graph_delete_2$graph_actions$action_index, 1)

  # Expect the `action_name` to be `get_btwns`
  expect_equal(
    graph_delete_2$graph_actions$action_name, "get_pagerank")

  # Expect the action in the data frame to
  # be correctly generated
  expect_equal(
    graph_delete_2$graph_actions$expression,
    "set_node_attr_w_fcn(graph = graph, node_attr_fcn = 'get_pagerank', column_name = 'pagerank')")

  # Delete two of the graph actions by
  # their `action_name` values
  graph_delete_2_by_name <-
    graph %>%
    delete_graph_actions(
      actions = c("pagerank_to_width", "pagerank_fillcolor"))

  # Expect that one graph action remains
  # Expect a single row in the data frame
  expect_equal(
    nrow(graph_delete_2_by_name$graph_actions),
    n_graph_actions_before_deletion - 2)

  # Expect that the first graph action
  # remains in the graph
  # Expect the `action_index` to be 1
  expect_equal(
    graph_delete_2_by_name$graph_actions$action_index, 1)

  # Expect the `action_name` to be `get_btwns`
  expect_equal(
    graph_delete_2_by_name$graph_actions$action_name, "get_pagerank")

  # Expect the action in the data frame to
  # be correctly generated
  expect_equal(
    graph_delete_2_by_name$graph_actions$expression,
    "set_node_attr_w_fcn(graph = graph, node_attr_fcn = 'get_pagerank', column_name = 'pagerank')")
})

test_that("actions within a graph object can be reordered", {

  # Create an empty graph object
  graph <- create_graph()

  # Add three graph actions to the
  # graph
  graph <-
    graph %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_pagerank",
      column_name = "pagerank",
      action_name = "get_pagerank") %>%
    add_graph_action(
      fcn = "rescale_node_attrs",
      node_attr_from = "pagerank",
      node_attr_to = "width",
      action_name = "pagerank_to_width") %>%
    add_graph_action(
      fcn = "colorize_node_attrs",
      node_attr_from = "width",
      node_attr_to = "fillcolor",
      action_name = "pagerank_fillcolor")

  # Get the names of the graph actions
  # before the reordering occurs
  names_of_graph_actions_before_reordering <-
    graph$graph_actions$action_name

  # Reorder the graph actions so that `2`,
  # precedes `3`, which precedes `1`
  graph <-
    graph %>%
    reorder_graph_actions(
      indices = c(2, 3, 1))

  # Expect three graph actions in the
  # graph object
  expect_equal(
    nrow(graph$graph_actions), 3)

  # Get the names of the graph actions
  # before the reordering occurs
  names_of_graph_actions_after_reordering <-
    graph$graph_actions$action_name

  # Expect that the graph action names
  # appear in the order according to the
  # vector provided as `indices`
  expect_equal(
    names_of_graph_actions_after_reordering,
    names_of_graph_actions_before_reordering[c(2, 3, 1)])
})

test_that("graph actions can be triggered to modify the graph", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 5,
      m = 10,
      set_seed = 23)

  # Add three graph actions to:
  #  - add PageRank values
  #  - rescale PageRank values
  #  - create a `fillcolor` attr
  # ...then, manually trigger the
  # actions to perform evaluation
  graph <-
    graph %>%
    add_graph_action(
      fcn = "set_node_attr_w_fcn",
      node_attr_fcn = "get_pagerank",
      column_name = "pagerank",
      action_name = "get_pagerank") %>%
    add_graph_action(
      fcn = "rescale_node_attrs",
      node_attr_from = "pagerank",
      node_attr_to = "width",
      action_name = "pgrnk_to_width") %>%
    add_graph_action(
      fcn = "colorize_node_attrs",
      node_attr_from = "width",
      node_attr_to = "fillcolor",
      action_name = "pgrnk_fillcolor") %>%
    trigger_graph_actions()

  # Expect certain columns to be available
  # in the graph's internal node data frame
  expect_named(
    graph$nodes_df,
    c("id", "type", "label",
      "pagerank", "width", "fillcolor"))

  # Expect the `pagerank` column to have
  # numeric values less than 1
  expect_type(
    graph$nodes_df$pagerank, "double")

  expect_true(
    all(graph$nodes_df$pagerank <= 1))

  # Expect the `width` column to have
  # numeric values less than 1
  expect_type(
    graph$nodes_df$width, "double")

  expect_true(
    all(graph$nodes_df$width <= 1))

  # Expect the `fillcolor` column to have
  # character values with color codes
  expect_type(
    graph$nodes_df$fillcolor, "character")

  expect_match(
    graph$nodes_df$fillcolor,
    "#[A-F0-9]*"
  )

  # Expect a warning if using the
  # `trigger_graph_actions()` function
  # when there are no graph actions
  expect_message(
    create_graph() %>%
      trigger_graph_actions())
})

#' Get count of all unconnected nodes
#'
#' From a graph object of class `dgr_graph`, get a count of nodes in the graph
#' that are not connected to any other node.
#'
#' @inheritParams render_graph
#'
#' @return A numeric vector of single length.
#'
#' @examples
#' # Create a graph with a
#' # path of nodes and 3
#' # unconnected nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 3) %>%
#'   add_n_nodes(n = 3)
#'
#' # Get a count of all nodes
#' # in the graph
#' graph %>% count_nodes()
#'
#' # Get a count of all
#' # unconnected nodes in the
#' # graph
#' graph %>%
#'   count_unconnected_nodes()
#'
#' @export
count_unconnected_nodes <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If graph is empty, return 0
  if (is_graph_empty(graph)) {
    return(0L)
  }

  # Get tbl with all nodes that are part
  # of edges
  nodes_in_edf <-
    dplyr::bind_rows(
      graph$edges_df %>% dplyr::select(node_id = "from"),
      graph$edges_df %>% dplyr::select(node_id = "to")
      ) %>%
    dplyr::distinct()

  # Get tbl with all nodes that are
  # in the node data frame
  nodes_in_ndf <-
    graph$nodes_df %>%
    dplyr::select(node_id = "id")

  # Get nodes not in edge definitions
  nodes_not_in_edf <-
    dplyr::setdiff(
      nodes_in_ndf,
      nodes_in_edf) %>%
    dplyr::pull("node_id")

  if (length(nodes_in_edf > 0)) {
    return(length(nodes_not_in_edf))
  } else {
    return(0L)
  }
}

#' Add new edges with identical definitions as with a selection of edges
#'
#' @description
#'
#' Add edges in the same direction of one or more edges available as an edge
#' selection in a graph object of class `dgr_graph`. New graph edges have the
#' same edge definitions as those in the selection except with new edge ID
#' values. There is also the option to assign a common `rel` grouping to the
#' newly created edges. Upon addition of the edges, the edge selection will be
#' retained for further selection or traversal operations.
#'
#' This function makes use of an active selection of edges (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#' @param rel An optional string to apply a `rel` attribute to all newly created
#'   edges.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph, add 2 nodes
#' # to it, and create the edge `1->2`
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(
#'     n = 2,
#'     type = "type_a",
#'     label = c("a_1", "a_2")) %>%
#'   add_edge(
#'     from = 1, to = 2, rel = "a")
#'
#' # Get the graph's edges
#' graph %>% get_edge_ids()
#'
#' # Select the edge and create 2
#' # additional edges with the same
#' # definition (`1->2`) but with
#' # different `rel` values (`b` and `c`)
#' graph <-
#'   graph %>%
#'   select_edges() %>%
#'   add_forward_edges_ws(rel = "b") %>%
#'   add_forward_edges_ws(rel = "c") %>%
#'   clear_selection()
#'
#' # Get the graph's edge data frame
#' graph %>% get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
add_forward_edges_ws <- function(
    graph,
    rel = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid edge selection
  check_graph_contains_edge_selection(graph)

  # If no value(s) provided for `rel`, set to NA
  rel <- rel %||% NA_character_

  # Get a vector of edges available in the
  # graph's selection
  edges_in_selection <-
    graph$edge_selection %>%
    dplyr::select("from", "to")

  # Get the number of edges in the graph
  edges_graph_1 <- count_edges(graph)

  # Add new edges to the graph for every edge
  # in the graph's active selection
  for (i in seq_len(nrow(edges_in_selection))) {

    # Create a graph edge
    graph <-
      add_edge(
        graph = graph,
        from = edges_in_selection[i, 1],
        to = edges_in_selection[i, 2],
        rel = rel)

    # Redact the signing of the action to the log
    graph$graph_log <-
      graph$graph_log[-nrow(graph$graph_log), ]
  }

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Move layout positions of a selection of nodes
#'
#' @description
#'
#' With an active selection of nodes, move the position in either the `x` or `y`
#' directions, or both. Nodes in the selection that do not have position
#' information (i.e., `NA` values for the `x` or `y` node attributes) will be
#' ignored.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param dx A single numeric value specifying the amount that selected nodes
#'   (with non-`NA` values for the `x` and `y` attributes) will be moved in the
#'   x direction. A positive value will move nodes right, negative left.
#' @param dy A single numeric value specifying the amount that selected nodes
#'   (with non-`NA` values for the `x` and `y` attributes) will be moved in the
#'   y direction. A positive value will move nodes up, negative down.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a simple graph with 4 nodes
#' graph <-
#'   create_graph() %>%
#'   add_node(
#'     type = "a",
#'     label = "one") %>%
#'   add_node(
#'     type = "a",
#'     label = "two") %>%
#'   add_node(
#'     type = "b",
#'     label = "three") %>%
#'   add_node(
#'     type = "b",
#'     label = "four")
#'
#' # Add position information to each of
#' # the graph's nodes
#' graph <-
#'   graph %>%
#'   set_node_position(
#'     node = 1, x = 1, y = 1) %>%
#'   set_node_position(
#'     node = 2, x = 2, y = 2) %>%
#'   set_node_position(
#'     node = 3, x = 3, y = 3) %>%
#'   set_node_position(
#'     node = 4, x = 4, y = 4)
#'
#' # Select all of the graph's nodes using the
#' # `select_nodes()` function (and only
#' # specifying the graph object)
#' graph <- select_nodes(graph)
#'
#' # Move the selected nodes (all the nodes,
#' # in this case) 5 units to the right
#' graph <-
#'   graph %>%
#'   nudge_node_positions_ws(
#'     dx = 5, dy = 0)
#'
#' # View the graph's node data frame
#' graph %>% get_node_df()
#'
#' # Now select nodes that have `type == "b"`
#' # and move them in the `y` direction 2 units
#' # (the graph still has an active selection
#' # and so it must be cleared first)
#' graph <-
#'   graph %>%
#'   clear_selection() %>%
#'   select_nodes(
#'     conditions = type == "b") %>%
#'   nudge_node_positions_ws(
#'     dx = 0, dy = 2)
#'
#' # View the graph's node data frame
#' graph %>% get_node_df()
#'
#' @export
nudge_node_positions_ws <- function(
    graph,
    dx,
    dy
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # Get the graph's node data frame as an object
  ndf <- graph$nodes_df

  # If both the `x` and `y` attributes do not exist,
  # stop the function
  if (!("x" %in% colnames(ndf)) ||
      !("y" %in% colnames(ndf))) {

    cli::cli_abort(
      "There are no `x` and `y` attribute values to modify.")
  }

  # Get the current selection of nodes
  nodes <- graph$node_selection$node

  # Determine which of the nodes selected have position
  # information set (i.e., not NA)
  ndf_filtered <-
    ndf %>%
    dplyr::filter(id %in% nodes, !is.na(x), !is.na(y))

  # If there are nodes to move, replace the `nodes`
  # vector with those node ID values; otherwise,
  # stop function
  if (nrow(ndf_filtered) == 0) {

    cli::cli_abort(
      "There are no nodes can be moved to different `x` or `y` locations.")

  } else {
    nodes <- ndf_filtered$id
  }

  # Use `case_when` statements to selectively perform
  # a vectorized `if` statement across all nodes for
  # the `x` and `y` node attribute
  ndf_new <-
    ndf %>%
    dplyr::mutate(x = dplyr::case_when(
      id %in% as.integer(nodes) ~ x + dx,
      !(id %in% as.integer(nodes)) ~ x)) %>%
    dplyr::mutate(y = dplyr::case_when(
      id %in% nodes ~ y + dy,
      !(id %in% as.integer(nodes)) ~ y))

  # Replace the graph's node data frame with `ndf_new`
  graph$nodes_df <- ndf_new

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

test_that("setting a single node's position is possible", {

  # Create a simple graph with 4 nodes
  graph <-
    create_graph() %>%
    add_node(label = "one") %>%
    add_node(label = "two") %>%
    add_node(label = "three") %>%
    add_node(label = "four")

  # Add position information to each of
  # the graph's nodes
  graph_1 <-
    graph %>%
    set_node_position(
      node = 1, x = 1, y = 1) %>%
    set_node_position(
      node = 2, x = 2, y = 2) %>%
    set_node_position(
      node = 3, x = 3, y = 3) %>%
    set_node_position(
      node = 4, x = 4, y = 4)

  # Expect that the `x` and `y` node
  # attributes columns are available
  expect_contains(colnames(graph_1$nodes_df), c("x", "y"))

  # Expect specific values in the `x` column
  expect_equal(
    graph_1$nodes_df$x, c(1, 2, 3, 4))

  # Expect specific values in the `y` column
  expect_equal(
    graph_1$nodes_df$y, c(1, 2, 3, 4))

  # Add position information to each of the
  # graph's nodes using node `label` values
  graph_2 <-
    graph %>%
    set_node_position(
      node = "one",
      x = 1, y = 1,
      use_labels = TRUE) %>%
    set_node_position(
      node = "two",
      x = 2, y = 2,
      use_labels = TRUE) %>%
    set_node_position(
      node = "three",
      x = 3, y = 3,
      use_labels = TRUE) %>%
    set_node_position(
      node = "four",
      x = 4, y = 4,
      use_labels = TRUE)

  # Expect that both `x` and `y` node
  # attributes columns are available
  expect_contains(colnames(graph_2$nodes_df), c("x", "y"))

  # Expect specific values in the `x` column
  expect_equal(
    graph_2$nodes_df$x, c(1, 2, 3, 4))

  # Expect specific values in the `y` column
  expect_equal(
    graph_2$nodes_df$y, c(1, 2, 3, 4))
})

test_that("altering several nodes' positions is possible", {

  # Create a simple graph with 4 nodes
  graph <-
    create_graph() %>%
    add_node(type = "a", label = "one") %>%
    add_node(type = "a", label = "two") %>%
    add_node(type = "b", label = "three") %>%
    add_node(type = "b", label = "four")

  # Add position information to each of
  # the graph's nodes
  graph_1 <-
    graph %>%
    set_node_position(
      node = 1, x = 1, y = 1) %>%
    set_node_position(
      node = 2, x = 2, y = 2) %>%
    set_node_position(
      node = 3, x = 3, y = 3) %>%
    set_node_position(
      node = 4, x = 4, y = 4)

  # Select all of the graph's nodes using the
  # `select_nodes()` function (and only specifying
  # the graph object)
  graph_1 <- select_nodes(graph_1)

  # Move the selected nodes (all the nodes,
  # in this case) 5 units to the right
  graph_1 <-
    graph_1 %>%
    nudge_node_positions_ws(
      dx = 5, dy = 0)

  # Expect that the `x` values have specific values
  expect_equal(
    graph_1$nodes_df$x, c(6, 7, 8, 9))

  # Expect that the `y` values have specific values
  expect_equal(
    graph_1$nodes_df$y, c(1, 2, 3, 4))

  # Now select nodes that have `type == "b"`
  # and move them in the `y` direction 2 units
  graph_2 <-
    graph %>%
    select_nodes(
      conditions = type == "b") %>%
    set_node_position(
      node = 1, x = 1, y = 1) %>%
    set_node_position(
      node = 2, x = 2, y = 2) %>%
    set_node_position(
      node = 3, x = 3, y = 3) %>%
    set_node_position(
      node = 4, x = 4, y = 4) %>%
    nudge_node_positions_ws(
      dx = 0, dy = 2)

  # Expect that the `x` values have specific values
  expect_equal(
    graph_2$nodes_df$x, c(1, 2, 3, 4))

  # Expect that the `y` values have specific values
  expect_equal(
    graph_2$nodes_df$y, c(1, 2, 5, 6))
})

test_that("setting positions with a text string is possible", {

  # Create a graph with unique labels and
  # several node `type` groupings
  graph <-
    create_graph() %>%
    add_node(type = "a", label = "a") %>%
    add_node(type = "a", label = "b") %>%
    add_node(type = "b", label = "c") %>%
    add_node(type = "b", label = "d") %>%
    add_node(type = "b", label = "e") %>%
    add_node(type = "c", label = "f") %>%
    add_node(type = "c", label = "g")

  # Define a 'layout' for groups of nodes
  # using a text string
  layout <-
    "
1--------
1--------
---222---
--------3
--------3
"

  # Use the `layout` along with what nodes
  # the numbers correspond to in the graph
  # with the `nodes` named vectors
  graph <-
    graph %>%
    layout_nodes_w_string(
      layout = layout,
      nodes = c("1" = "type:a",
                "2" = "type:b",
                "3" = "type:c"),
      sort = c("1" = "label:asc",
               "2" = "label:desc",
               "3" = "label:desc"))

  # Expect that both the `x` and `y` node
  # attributes columns are available
  expect_contains(colnames(graph$nodes_df), c("x", "y"))

  # Expect specific values in the `x` column
  expect_equal(
    graph$nodes_df$x, c(0, 0, 5, 4, 3, 8, 8))

  # Expect specific values in the `y` column
  expect_equal(
    graph$nodes_df$y, c(8, 6, 4, 4, 4, 0, 2))
})

#' Get indegree values for all nodes
#'
#' @description
#'
#' Get the indegree values for all nodes in a graph.
#'
#' @inheritParams render_graph
#' @param normalized set as `FALSE` (the default), the indegree will be
#'   provided for each of the nodes (as a count of edges to each node). When set
#'   as `TRUE`, then the result for each node will be divided by the total
#'   number of nodes in the graph minus 1.
#'
#' @return a data frame with indegree values for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the indegree values for
#' # all nodes in the graph
#' graph %>%
#'   get_degree_in()
#'
#' # Add the indegree values
#' # to the graph as a node
#' # attribute
#' graph <-
#'   graph %>%
#'   join_node_attrs(
#'     df = get_degree_in(.))
#'
#' # Display the graph's
#' # node data frame
#' graph %>% get_node_df()
#'
#' @export
get_degree_in <- function(
    graph,
    normalized = FALSE
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the indegree values for each of the
  # graph's nodes
  if (normalized) {
    indegree_values <-
      igraph::degree(
        ig_graph,
        mode = "in",
        normalized = TRUE)
  }

  if (!normalized) {
    indegree_values <-
      igraph::degree(
        ig_graph,
        mode = "in",
        normalized = FALSE)
  }

  # Create df with indegree scores
  data.frame(
    id = indegree_values %>%
      names() %>%
      as.integer(),
    indegree = indegree_values,
    stringsAsFactors = FALSE)

}

# Getting degree information for all nodes in a graph

test_that("a degree data frame can be generated", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Get the total degree values for all nodes
  # in the graph
  total_degree_df <- get_degree_total(graph)

  # Expect that the output is a data frame
  expect_s3_class(
    total_degree_df, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(total_degree_df), 2)

  # Expect 10 rows in the data frame
  expect_equal(
    nrow(total_degree_df), 10)

  # Expect certain column names for the df
  expect_identical(
    colnames(total_degree_df),
    c("id", "total_degree"))

  # Expect certain values in the 2nd column
  expect_equal(
    total_degree_df$total_degree,
    c(1, 1, 5, 3, 3, 1, 0, 2, 1, 3))

  # Get the total degree values for all nodes
  # in the graph as normalized values
  total_degree_df_norm <-
    get_degree_total(
      graph = graph,
      normalized = TRUE)

  # Expect that the output is a data frame
  expect_s3_class(
    total_degree_df_norm, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(total_degree_df_norm), 2)

  # Expect 10 rows in the data frame
  expect_equal(
    nrow(total_degree_df_norm), 10)

  # Expect certain column names for the df
  expect_identical(
    colnames(total_degree_df_norm),
    c("id", "total_degree"))

  # Expect certain values in the 2nd column
  expect_equal(
    total_degree_df_norm$total_degree,
    c(0.1111111, 0.1111111, 0.5555556, 0.3333333, 0.3333333,
      0.1111111, 0.0000000, 0.2222222, 0.1111111, 0.3333333),
    tolerance = 0.002)

  # Get the in-degree values for all nodes
  # in the graph
  in_degree_df <- get_degree_in(graph)

  # Expect that the output is a data frame
  expect_s3_class(
    in_degree_df, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(in_degree_df), 2)

  # Expect 10 rows in the data frame
  expect_equal(
    nrow(in_degree_df), 10)

  # Expect certain column names for the df
  expect_identical(
    colnames(in_degree_df),
    c("id", "indegree"))

  # Expect certain values in the 2nd column
  expect_equal(
    in_degree_df$indegree,
    c(1, 0, 2, 1, 0, 1, 0, 2, 1, 2))

  # Get the total degree values for all nodes
  # in the graph as normalized values
  in_degree_df_norm <-
    get_degree_in(
      graph = graph,
      normalized = TRUE)

  # Expect that the output is a data frame
  expect_s3_class(
    in_degree_df_norm, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(in_degree_df_norm), 2)

  # Expect 10 rows in the data frame
  expect_equal(
    nrow(in_degree_df_norm), 10)

  # Expect certain column names for the df
  expect_identical(
    colnames(in_degree_df_norm),
    c("id", "indegree"))

  # Expect certain values in the 2nd column
  expect_equal(
    in_degree_df_norm$indegree,
    c(0.1111111, 0.0000000, 0.2222222, 0.1111111, 0.0000000,
      0.1111111, 0.0000000, 0.2222222, 0.1111111, 0.2222222),
    tolerance = 0.002)

  # Get the out-degree values for all nodes
  # in the graph
  out_degree_df <- get_degree_out(graph)

  # Expect that the output is a data frame
  expect_s3_class(
    out_degree_df, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(out_degree_df), 2)

  # Expect 10 rows in the data frame
  expect_equal(
    nrow(out_degree_df), 10)

  # Expect certain column names for the df
  expect_identical(
    colnames(out_degree_df),
    c("id", "outdegree"))

  # Expect certain values in the 2nd column
  expect_equal(
    out_degree_df$outdegree,
    c(0, 1, 3, 2, 3, 0, 0, 0, 0, 1))

  # Get the total degree values for all nodes
  # in the graph as normalized values
  out_degree_df_norm <-
    get_degree_out(
      graph = graph,
      normalized = TRUE)

  # Expect that the output is a data frame
  expect_s3_class(
    out_degree_df_norm, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(out_degree_df_norm), 2)

  # Expect 10 rows in the data frame
  expect_equal(
    nrow(out_degree_df_norm), 10)

  # Expect certain column names for the df
  expect_identical(
    colnames(out_degree_df_norm),
    c("id", "outdegree"))

  # Expect certain values in the 2nd column
  expect_equal(
    out_degree_df_norm$outdegree,
    c(0.0000000, 0.1111111, 0.3333333, 0.2222222, 0.3333333,
      0.0000000, 0.0000000, 0.0000000, 0.0000000, 0.1111111),
    tolerance = 0.002)
})

#' Display a property graph's underlying model
#'
#' @description
#'
#' With a graph object of class `dgr_graph` that is also a property graph (i.e.,
#' all nodes have an assigned `type` value and all edges have an assigned `rel`
#' value), display its metagraph in the RStudio Viewer. This representation
#' provides all combinations of edges of different `rel` values to all nodes
#' with distinct `type` values, including any edges to nodes of the same `type`
#' (shown as loops). The precondition of the graph being a property graph can be
#' verified by using the [is_property_graph()] function.
#'
#' @param graph A graph object of class `dgr_graph`. This graph must fulfill the
#'   condition of being a property graph, otherwise the function yields an
#'   error.
#'
#' @examples
#' # Create a randomized property
#' # graph with 1000 nodes and 1350 edges
#' property_graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 1000,
#'     m = 1350,
#'     set_seed = 23) |>
#'   select_nodes_by_degree(
#'     expressions = "deg >= 3") |>
#'   set_node_attrs_ws(
#'     node_attr = type,
#'     value = "a") |>
#'   clear_selection() |>
#'   select_nodes_by_degree(
#'     expressions = "deg < 3") |>
#'   set_node_attrs_ws(
#'     node_attr = type,
#'     value = "b") |>
#'   clear_selection() |>
#'   select_nodes_by_degree(
#'     expressions = "deg == 0") |>
#'   set_node_attrs_ws(
#'     node_attr = type,
#'     value = "c") |>
#'   set_node_attr_to_display(
#'     attr = type)
#'
#' # Select a random subset of edges
#' # for setting edge attributes
#' node_ids <- get_node_ids(property_graph)
#' sampled_nodes <- sample(
#'   node_ids,
#'   size = floor(0.15 * length(node_ids)))
#'
#' property_graph <-
#'   property_graph |>
#'   select_edges_by_node_id(
#'     nodes = sampled_nodes) |>
#'   set_edge_attrs_ws(
#'     edge_attr = rel,
#'     value = "r_1") |>
#'   invert_selection() |>
#'   set_edge_attrs_ws(
#'     edge_attr = rel,
#'     value = "r_2") |>
#'   clear_selection() |>
#'   copy_edge_attrs(
#'     edge_attr_from = rel,
#'     edge_attr_to = label) |>
#'   add_global_graph_attrs(
#'     attr = "fontname",
#'     value = "Helvetica",
#'     attr_type = "edge") |>
#'   add_global_graph_attrs(
#'     attr = "fontcolor",
#'     value = "gray50",
#'     attr_type = "edge") |>
#'   add_global_graph_attrs(
#'     attr = "fontsize",
#'     value = 10,
#'     attr_type = "edge")
#'
#' # Display this graph's
#' # metagraph, or, the underlying
#' # graph model for a property graph
#' # display_metagraph(property_graph)
#'
#' @export
display_metagraph <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph object is a property graph
  check_property_graph(graph)

  # Get a distinct list of node `type` values
  unique_node_list <-
    graph$nodes_df |>
    dplyr::distinct(type)

  # Get a distinct list of edges between types
  unique_edge_list <-
    graph$edges_df |>
    dplyr::inner_join(
      graph$nodes_df |>
        dplyr::select(from = "id", from_type = "type"),
      by = "from") |>
    dplyr::inner_join(
      graph$nodes_df |>
        dplyr::select(to = "id", to_type = "type"),
      by = "to") |>
    dplyr::distinct(rel, from_type, to_type)

  # Create the initial metagraph
  metagraph <-
    create_graph() |>
    add_nodes_from_df_cols(
      unique_node_list,
      columns = "type") |>
    add_edges_from_table(
      unique_edge_list,
      from_to_map = label,
      from_col = from_type,
      to_col = to_type,
      rel_col = rel)

  # Copy the `label` values to the `type` attribute
  metagraph$nodes_df$type <- metagraph$nodes_df$label

  # Apply coloring and other aesthetics to nodes and edges
  metagraph <-
    metagraph |>
    colorize_node_attrs(
      node_attr_from = "type",
      node_attr_to = "fillcolor") |>
    copy_edge_attrs(
      edge_attr_from = "rel",
      edge_attr_to = "label") |>
    add_global_graph_attrs(
      attr = "fontname",
      value = "Helvetica",
      attr_type = "edge") |>
    add_global_graph_attrs(
      attr = "fontcolor",
      value = "gray50",
      attr_type = "edge") |>
    add_global_graph_attrs(
      attr = "fontsize",
      value = 10,
      attr_type = "edge") |>
    colorize_edge_attrs(
      edge_attr_from = "rel",
      edge_attr_to = "color") |>
    add_global_graph_attrs(
      attr = "fontsize",
      value = 6,
      attr_type = "edge") |>
    add_global_graph_attrs(
      attr = "len",
      value = 3.5,
      attr_type = "edge") |>
    add_global_graph_attrs(
      attr = "layout",
      value = "dot",
      attr_type = "graph")

  # Render the `metagraph` object
  render_graph(metagraph)
}

#' Get community membership using the Walktrap method
#'
#' @description
#'
#' With the Walktrap community finding algorithm, obtain the group membership
#' values for each of the nodes in the graph.
#'
#' @inheritParams render_graph
#' @param steps the number of steps to take for each of the random walks.
#'
#' @return A data frame with group membership assignments for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the group membership
#' # values for all nodes in the
#' # graph through the Walktrap
#' # community finding algorithm
#' graph %>%
#'   get_cmty_walktrap()
#'
#' # Add the group membership
#' # values to the graph as a
#' # node attribute
#' graph <-
#'   graph %>%
#'   join_node_attrs(
#'     df = get_cmty_walktrap(.))
#'
#' # Display the graph's
#' # node data frame
#' graph %>% get_node_df()
#'
#' @export
get_cmty_walktrap <- function(
    graph,
    steps = 4
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the community object using the
  # `cluster_walktrap()` function
  cmty_walktrap_obj <-
    igraph::cluster_walktrap(ig_graph, steps = steps)

  # Create df with node memberships
  data.frame(
    id = igraph::membership(cmty_walktrap_obj) %>% names() %>% as.integer(),
    walktrap_group = as.vector(igraph::membership(cmty_walktrap_obj)),
    stringsAsFactors = FALSE)
}

# Community detection algorithms

test_that("the edge betweeness algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Get the edge betweenness for the graph
  edge_betweeness <- get_cmty_edge_btwns(graph)

  # Expect that a `data.frame` object is returned
  expect_s3_class(
    edge_betweeness, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(
      edge_betweeness), 2)

  # Expect specific column names in this data frame
  expect_equal(
    colnames(edge_betweeness),
    c("id", "edge_btwns_group"))

  # Expect as many rows in the data frame as there
  # are nodes in the graph
  expect_equal(
    nrow(edge_betweeness),
    count_nodes(graph = graph))

  # For this analysis expect 5 different groups
  expect_equal(
    sort(unique(edge_betweeness$edge_btwns_group)),
    c(1, 2, 3, 4, 5, 6))

  # Expect the first column to be integer
  expect_type(
    edge_betweeness$id, "integer")

  # Expect the second column to be numeric
  expect_type(
    edge_betweeness$edge_btwns_group, "double")
})

test_that("the fast-greedy algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 15,
      m = 10,
      set_seed = 23)

  # Get the edge betweenness for the graph
  fast_greedy <- get_cmty_fast_greedy(graph)

  # Expect that a `data.frame` object is returned
  expect_s3_class(
    fast_greedy, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(fast_greedy), 2)

  # Expect specific column names in this data frame
  expect_equal(
    colnames(fast_greedy),
    c("id", "f_g_group"))

  # Expect as many rows in the data frame as there
  # are nodes in the graph
  expect_equal(
    nrow(fast_greedy), count_nodes(graph = graph))

  # For this analysis expect two different groups
  # identified with labels `1` and `2`
  expect_equal(
    sort(unique(fast_greedy$f_g_group)),
    c(1, 2, 3, 4, 5, 6))

  # Expect the first column to be integer
  expect_type(
    fast_greedy$id, "integer")

  # Expect the second column to be numeric
  expect_type(
    fast_greedy$f_g_group, "double")
})

test_that("the leading eigenvector algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Get the edge betweenness for the graph
  l_eigenvec <- get_cmty_l_eigenvec(graph)

  # Expect that a `data.frame` object is returned
  expect_s3_class(
    l_eigenvec, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(l_eigenvec), 2)

  # Expect specific column names in this data frame
  expect_equal(
    colnames(l_eigenvec),
    c("id", "l_eigenvec_group"))

  # Expect as many rows in the data frame as there
  # are nodes in the graph
  expect_equal(
    nrow(l_eigenvec), count_nodes(graph = graph))

  # For this analysis expect two different groups
  # identified with labels `1` and `2`
  expect_equal(
    sort(unique(l_eigenvec$l_eigenvec_group)),
    c(1, 2, 3, 4))

  # Expect the first column to be integer
  expect_type(l_eigenvec$id, "integer")

  # Expect the second column to be numeric
  expect_type(l_eigenvec$l_eigenvec_group, "double")
})

test_that("the Louvain algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Get the edge betweenness for the graph
  louvain <- get_cmty_louvain(graph)

  # Expect that a `data.frame` object is returned
  expect_s3_class(
    louvain, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(louvain), 2)

  # Expect specific column names in this data frame
  expect_equal(
    colnames(louvain),
    c("id", "louvain_group"))

  # Expect as many rows in the data frame as there
  # are nodes in the graph
  expect_equal(
    nrow(louvain), count_nodes(graph = graph))

  # For this analysis expect two different groups
  # identified with labels `1` and `2`
  expect_equal(
    sort(unique(louvain$louvain_group)),
    c(1, 2, 3, 4))

  # Expect the first column to be integer
  expect_type(
    louvain$id, "integer")

  # Expect the second column to be numeric
  expect_type(
    louvain$louvain_group, "double")
})

test_that("the walktrap algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Get the edge betweenness for the graph
  walktrap <- get_cmty_walktrap(graph)

  # Expect that a `data.frame` object is returned
  expect_s3_class(
    walktrap, "data.frame")

  # Expect 2 columns in the data frame
  expect_equal(
    ncol(walktrap), 2)

  # Expect specific column names in this data frame
  expect_named(
    walktrap,
    c("id", "walktrap_group"))

  # Expect as many rows in the data frame as there
  # are nodes in the graph
  expect_equal(
    nrow(walktrap), count_nodes(graph = graph))

  # For this analysis expect two different groups
  # identified with labels `1` and `2`
  expect_equal(
    sort(unique(walktrap$walktrap_group)),
    c(1, 2, 3, 4))

  # Expect the first column to be integer
  expect_type(
    walktrap$id, "integer")

  # Expect the second column to be numeric
  expect_type(
    walktrap$walktrap_group, "double")
})

# Colorizing based on node and edge attrs

test_that("Adding color based on node attributes is possible", {

  # Create a random graph of
  # 50 nodes and 85 edges
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 50,
      m = 85,
      node_data = node_data(
        value = 1:50),
      set_seed = 23)

  # Find group membership values for all nodes
  # in the graph through the Walktrap community
  # finding algorithm and join those group values
  # to the graph's internal node data frame (ndf)
  # with the `join_node_attrs()` function
  graph <-
    graph %>%
    join_node_attrs(
      get_cmty_walktrap(.))

  # Use the `colorize_node_attrs()` function
  # to set different `fillcolor` values
  graph <-
    graph %>%
    colorize_node_attrs(
      node_attr_from = walktrap_group,
      node_attr_to = fillcolor)

  # Expect that the `fillcolor` column has
  # been created in the node data frame
  expect_contains(colnames(graph$nodes_df), "fillcolor")

  # Expect that there are as many different
  # colors in the `fillcolor` column as there
  # are walktrap communities
  expect_equal(
    length(unique(graph$nodes_df$walktrap_group)),
    length(unique(graph$nodes_df$fillcolor)))

  # Expect that each value in the `fillcolor`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$nodes_df$fillcolor,
    "#[0-9A-F]{6}")

  # Use the `colorize_node_attrs()` function
  # to set different `color` values with an
  # alpha value set to `90`
  graph <-
    graph %>%
    colorize_node_attrs(
      node_attr_from = walktrap_group,
      node_attr_to = color,
      alpha = 90) %>%
    set_node_attrs(
      node_attr = fontcolor,
      values = "white")

  # Expect that the `color` column has
  # been created in the node data frame
  expect_contains(colnames(graph$nodes_df), "color")

  # Expect that there are as many different
  # colors in the `color` column as there
  # are walktrap communities
  expect_equal(
    length(unique(graph$nodes_df$walktrap_group)),
    length(unique(graph$nodes_df$color)))

  # Expect that each value in the `color` column
  # is a properly-formed hexadecimal color code
  # with alpha value as suffix
  expect_match(
    graph$nodes_df$color,
    "#[0-9A-F]{6}[0-9]{2}")

  # Create a random graph of
  # 10 nodes and 22 edges
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23)

  # Bucketize values in `value` using `cut_points`
  # and assign colors to each of the bucketed ranges
  # (for values not part of any bucket, a gray color
  # is assigned by default)
  graph <-
    graph %>%
    colorize_node_attrs(
      node_attr_from = value,
      node_attr_to = fillcolor,
      cut_points = c(1, 3, 5, 7, 9))

  # Expect that the `fillcolor` column has
  # been created in the node data frame
  expect_contains(colnames(graph$nodes_df), "fillcolor")

  # Expect that there are 5 colors in the
  # `fillcolor` column
  expect_length(unique(graph$nodes_df$fillcolor), 5)

  # Expect that each value in the `fillcolor`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$nodes_df$fillcolor,
    "#[0-9A-F]{6}")

  # Bucketize values as before but use an alpha
  # value of `90`
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23) %>%
    colorize_node_attrs(
      node_attr_from = value,
      node_attr_to = fillcolor,
      cut_points = c(1, 3, 5, 7, 9),
      alpha = 90)

  # Expect that there are 5 colors in the
  # `fillcolor` column
  expect_length(unique(graph$nodes_df$fillcolor), 5)

  # Expect that each value in the `color` column
  # is a properly-formed hexadecimal color code
  # with alpha value as suffix
  expect_match(
    graph$nodes_df$fillcolor,
    "#[0-9A-F]{6}[0-9]{2}")

  # Bucketize values as before but use an alpha
  # value of `100`
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23) %>%
    colorize_node_attrs(
      node_attr_from = value,
      node_attr_to = fillcolor,
      cut_points = c(1, 3, 5, 7, 9),
      alpha = 100)

  # Expect that there are 5 colors in the
  # `fillcolor` column
  expect_length(unique(graph$nodes_df$fillcolor), 5)

  # Expect that each value in the `fillcolor`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$nodes_df$fillcolor,
    "#[0-9A-F]{6}")

  # Bucketize as before but color according to hexadecimal vector input
  # Includes one valid hex color not in x11_hex()
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23) %>%
    colorize_node_attrs(
      node_attr_from = value,
      node_attr_to = fillcolor,
      cut_points = c(1, 3, 5, 7, 9),
      palette = c("#64ccc9", "#8b3e2f", "#00eeee", "#556b2f", "#9932cc",
                  "#698b69", "#cd2626", "#4a4a4a"))

  # Expect that there are 5 colors in the
  # `fillcolor` column
  expect_length(unique(graph$nodes_df$fillcolor), 5)

  # Expect that each value in the `fillcolor`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$nodes_df$fillcolor,
    "#[0-9A-F]{6}")

  # Expect error when given mix of valid and invalid hexadecimal colors
  expect_error(
    create_graph() %>%
      add_gnm_graph(
        n = 10,
        m = 22,
        node_data = node_data(
          value = 1:10),
        set_seed = 23) %>%
      colorize_node_attrs(
        node_attr_from = value,
        node_attr_to = fillcolor,
        cut_points = c(1, 3, 5, 7, 9),
        palette = c("458b00l15", "foo", "#00eeee", "bar", "orange")))

})

test_that("Adding color based on edge attributes is possible", {

  # Create a random graph of
  # 10 nodes and 10 edges;
  # add the `weight` and `rel`
  # edge attrs
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      edge_data = edge_data(
        weight = rnorm(10, 5, 2)),
      set_seed = 23) %>%
    set_edge_attrs(
      edge_attr = "rel",
      values = c("A", "A", "B", "B", "D",
                 "A", "B", "C", "D", "A"))

  # Use the `colorize_edge_attrs()` function
  # to set different `color` values
  graph <-
    graph %>%
    colorize_edge_attrs(
      edge_attr_from = rel,
      edge_attr_to = color) %>%
    colorize_edge_attrs(
      edge_attr_from = rel,
      edge_attr_to = fontcolor,
      alpha = 90)

  # Expect that the `color` and `fontcolor`
  # columns have been created in the edf
  expect_contains(
    colnames(graph$edges_df),
    c("color", "fontcolor")
  )

  # Expect that there are as many different
  # colors in the `color` and `fontcolor`
  # columns as there are distinct `rel` values
  expect_equal(
    length(unique(graph$edges_df$rel)),
    length(unique(graph$edges_df$color)))

  expect_equal(
    length(unique(graph$edges_df$rel)),
    length(unique(graph$edges_df$fontcolor)))

  # Expect that each value in the `color`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$edges_df$color,
    "#[0-9A-F]{6}")

  # Expect that each value in the `fontcolor`
  # column is a properly-formed hexadecimal
  # color code with alpha value as suffix
  expect_match(
    graph$edges_df$fontcolor,
    "#[0-9A-F]{6}[0-9]{2}")

  # Bucketize values in `weight` using `cut_points`
  # and assign colors to each of the bucketed ranges
  # (for values not part of any bucket, a gray color
  # is assigned by default)
  graph <-
    graph %>%
    colorize_edge_attrs(
      edge_attr_from = weight,
      edge_attr_to = labelfontcolor,
      cut_points = c(0, 2, 4, 6, 8, 10))

  # Expect that the `labelfontcolor` column has
  # been created in the edge data frame
  expect_contains(
    colnames(graph$edges_df),
    "labelfontcolor"
    )

  # Expect that there are 4 colors in the
  # `labelfontcolor` column
  expect_length(unique(graph$edges_df$labelfontcolor), 4)

  # Expect that each value in the `labelfontcolor`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$edges_df$labelfontcolor,
    "#[0-9A-F]{6}")

  # Create a random graph of
  # 10 nodes and 10 edges;
  # add the `weight` and `rel`
  # edge attrs
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      edge_data = edge_data(
        weight = rnorm(10, 5, 2)),
      set_seed = 23) %>%
    set_edge_attrs(
      edge_attr = "rel",
      values = c("A", "A", "B", "B", "D",
                 "A", "B", "C", "D", "A"))

  # Bucketize values as before but use an alpha
  # value of `90`
  graph <-
    graph %>%
    colorize_edge_attrs(
      edge_attr_from = weight,
      edge_attr_to = labelfontcolor,
      cut_points = c(0, 2, 4, 6, 8, 10),
      alpha = 90)

  # Expect that there are 4 colors in the
  # `labelfontcolor` column
  expect_length(unique(graph$edges_df$labelfontcolor), 4)

  # Expect that each value in the `labelfontcolor`
  # column is a properly-formed hexadecimal color
  # code with alpha value as suffix
  expect_match(
    graph$edges_df$labelfontcolor,
    "#[0-9A-F]{6}[0-9]{2}")

  # Create a random graph of 10 nodes and 10 edges;
  # add the `weight` and `rel` edge attrs
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      edge_data = edge_data(
        weight = 1:10),
      set_seed = 23) %>%
    set_edge_attrs(
      edge_attr = rel,
      values = c("A", "A", "B", "B", "D",
                 "A", "B", "C", "D", "A"))

  # Bucketize values as before but use an alpha
  # value of `100`
  graph <-
    graph %>%
    colorize_edge_attrs(
      edge_attr_from = weight,
      edge_attr_to = labelfontcolor,
      cut_points = c(0, 2, 4, 6, 8, 10),
      alpha = 100)

  # Expect that there are 6 colors in the
  # `labelfontcolor` column
  expect_length(
    unique(graph$edges_df$labelfontcolor),
    6)

  # Expect that each value in the `labelfontcolor`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$edges_df$labelfontcolor,
    "#[0-9A-F]{6}")

  # Color edges according to hexadecimal vector input
  # Includes one valid hex color not in x11_hex()
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      edge_data = edge_data(
        weight = rnorm(10, 5, 2)),
      set_seed = 23) %>%
    set_edge_attrs(
      edge_attr = "rel",
      values = c("A", "A", "B", "B", "D",
                 "A", "B", "C", "D", "A")) %>%
    colorize_edge_attrs(
      edge_attr_from = rel,
      edge_attr_to = color,
      palette = c("#64ccc9", "#8b3e2f", "#00eeee", "#556b2f", "#9932cc",
                  "#698b69", "#cd2626", "#4a4a4a"))

  # Expect that each value in the `color`
  # column is a properly-formed hexadecimal color
  # code
  expect_match(
    graph$edges_df$color,
    "#[0-9A-F]{6}")

  # Expect error message if given invalid hexadecimal color
  expect_snapshot(error = TRUE,
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      edge_data = edge_data(
        weight = rnorm(10, 5, 2)),
      set_seed = 23) %>%
    set_edge_attrs(
      edge_attr = "rel",
      values = c("A", "A", "B", "B", "D",
                 "A", "B", "C", "D", "A")) %>%
    colorize_edge_attrs(
      edge_attr_from = rel,
      edge_attr_to = color,
      palette = c("#458b00l15", "foo", "#00eeee", "bar", "orange")))
})

#' Set node attribute values with a graph function
#'
#' @description
#'
#' From a graph object of class `dgr_graph` or a node data frame, set node
#' attribute properties for all nodes in the graph using one of several
#' whole-graph functions.
#'
#' @inheritParams render_graph
#' @param node_attr_fcn The name of the function to use for creating a column of
#'   node attribute values. Valid functions are: [get_alpha_centrality()],
#'   [get_authority_centrality()], [get_betweenness()], [get_closeness()],
#'   [get_cmty_edge_btwns()], [get_cmty_fast_greedy()], [get_cmty_l_eigenvec()],
#'   [get_cmty_louvain()], [get_cmty_walktrap()], [get_degree_distribution()],
#'   [get_degree_histogram()], [get_degree_in()], [get_degree_out()],
#'   [get_degree_total()], [get_eccentricity()], [get_eigen_centrality()],
#'   [get_pagerank()], [get_s_connected_cmpts()], and [get_w_connected_cmpts()].
#' @param ... Arguments and values to pass to the named function in
#'   `node_attr_fcn`, if necessary.
#' @param column_name An option to supply a column name for the new node
#'   attribute column. If `NULL` then the column name supplied by the function
#'   will used along with a `__A` suffix.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 22,
#'     set_seed = 23) |>
#'   set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = 10,
#'       mean = 5,
#'       sd = 1) |> round(1))
#'
#' # Get the betweenness values for
#' # each of the graph's nodes as a
#' # node attribute
#' graph_1 <-
#'   graph |>
#'   set_node_attr_w_fcn(
#'     node_attr_fcn = "get_betweenness")
#'
#' # Inspect the graph's internal
#' # node data frame
#' graph_1 |> get_node_df()
#'
#' # If a specified function takes argument
#' # values, these can be supplied as well
#' graph_2 <-
#'   graph |>
#'   set_node_attr_w_fcn(
#'     node_attr_fcn = "get_alpha_centrality",
#'     alpha = 2,
#'     exo = 2)
#'
#' # Inspect the graph's internal
#' # node data frame
#' graph_2 |> get_node_df()
#'
#' # The new column name can be provided
#' graph_3 <-
#'   graph |>
#'   set_node_attr_w_fcn(
#'     node_attr_fcn = "get_pagerank",
#'     column_name = "pagerank")
#'
#' # Inspect the graph's internal
#' # node data frame
#' graph_3 |> get_node_df()
#'
#' # If `graph_3` is modified by
#' # adding a new node then the column
#' # `pagerank` will have stale data; we
#' # can run the function again and re-use
#' # the existing column name to provide
#' # updated values
#' graph_3 <-
#'   graph_3 |>
#'   add_node(
#'     from = 1,
#'     to = 3) |>
#'   set_node_attr_w_fcn(
#'     node_attr_fcn = "get_pagerank",
#'     column_name = "pagerank")
#'
#' # Inspect the graph's internal
#' # node data frame
#' graph_3 |> get_node_df()
#'
#' @family node creation and removal
#'
#' @export
set_node_attr_w_fcn <- function(
    graph,
    node_attr_fcn,
    ...,
    column_name = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  value_per_node_fcn_names <-
    names(value_per_node_functions())

  if (!any(value_per_node_fcn_names %in% node_attr_fcn)) {

    cli::cli_abort(
      "`node_attr_fcn` must be a function that produces values for every graph node.")
  }

  # Collect extra vectors of arguments and values as `extras`
  extras <- list(...)

  if (length(extras) > 0) {

  nodes_df <-
    graph$nodes_df |>
    dplyr::inner_join(
      eval(
        parse(
          text = paste0(
            node_attr_fcn,
            "(graph, ",
            paste(names(extras),
                  "=",
                  extras,
                  collapse =  ", "),
            ")"))) |>
        dplyr::mutate(id = as.integer(id)),
      by = "id")

  } else {

    nodes_df <-
      graph$nodes_df |>
      dplyr::inner_join(
        eval(
          parse(
            text = paste0(
              node_attr_fcn,
              "(graph)"))) |>
          dplyr::mutate(id = as.integer(id)),
        by = "id")
  }

  if (!is.null(column_name)) {

    # If a new column name is specified in
    # `column_name`, use that for the new column
    colnames(nodes_df)[length(colnames(nodes_df))] <-
      column_name
  } else {

    # Add the `_A` tag to the column name normally
    # supplied by the function
    colnames(nodes_df)[length(colnames(nodes_df))] <-
      paste0(colnames(nodes_df)[length(colnames(nodes_df))], "__A")
  }

  # Determine if there is a column with the same name;
  # if there is, replace its contents with that of the
  # new column
  if (colnames(nodes_df)[length(colnames(nodes_df))] %in%
      colnames(nodes_df)[seq_len(length(colnames(nodes_df)) - 1)]) {

    col_no_matching <-
      which(colnames(nodes_df)[seq_len(length(colnames(nodes_df)) - 1)] ==
              colnames(nodes_df)[length(colnames(nodes_df))])

    # Move contents of new column to matching column
    nodes_df[, col_no_matching] <-
      nodes_df[, length(colnames(nodes_df))]

    # Remove the last column from the ndf
    nodes_df[, length(colnames(nodes_df))] <- NULL
  }

  if (paste0(colnames(nodes_df)[length(colnames(nodes_df))], ".x") %in%
      colnames(nodes_df)[1:(length(colnames(nodes_df)) - 1)]) {

    col_no_matching <-
      which(colnames(nodes_df)[seq_len(length(colnames(nodes_df)) - 1)] ==
              paste0(colnames(nodes_df)[length(colnames(nodes_df))], ".x"))

    # Get the column name of the new column
    new_col_name <- colnames(nodes_df)[length(colnames(nodes_df))]

    # Move contents of new column to matching column
    nodes_df[, col_no_matching] <-
      nodes_df[, length(colnames(nodes_df))]

    # Remove the last column from the ndf
    nodes_df[, length(colnames(nodes_df))] <- NULL

    # Rename the refreshed column
    colnames(nodes_df)[col_no_matching] <- new_col_name
  }

  # Replace the graph's ndf with the
  # revised version
  graph$nodes_df <- nodes_df

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get closeness centrality values
#'
#' @description
#'
#' Get the closeness centrality values for all nodes in a graph.
#'
#' @inheritParams render_graph
#' @param direction using `all` (the default), the search will ignore edge
#'   direction while traversing through the graph. With `out`, measurements
#'   of paths will be from a node whereas with `in`, measurements of paths
#'   will be to a node.
#'
#' @return A data frame with closeness values for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 12,
#'     set_seed = 23)
#'
#' # Get closeness values for all nodes
#' # in the graph
#' graph |> get_closeness()
#'
#' # Add the closeness values to
#' # the graph as a node attribute
#' graph <-
#'   graph |>
#'   join_node_attrs(
#'     df = get_closeness(graph))
#'
#' # Display the graph's node data frame
#' graph |> get_node_df()
#'
#' @export
get_closeness <- function(
    graph,
    direction = "all"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Ensure that values provided for the
  # `direction` argument are from the
  # valid options
  rlang::arg_match0(direction, c("all", "in", "out"))

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the betweenness scores for each of the
  # graph's nodes
  if (direction == "all") {
    closeness_values <-
      igraph::closeness(ig_graph, mode = "all")
  }

  if (direction == "out") {
    closeness_values <-
      igraph::closeness(ig_graph, mode = "out")
  }

  if (direction == "in") {
    closeness_values <-
      igraph::closeness(ig_graph, mode = "in")
  }

  # Create df with betweenness scores
  data.frame(
    id = closeness_values |>
      names() |>
      as.integer(),
    closeness = closeness_values,
    stringsAsFactors = FALSE)
}

# Get node calculations

test_that("Getting betweenness is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  betweenness_vals <- get_betweenness(graph)

  # Expect a data frame as output
  expect_s3_class(
    betweenness_vals, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(betweenness_vals), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(betweenness_vals), 10)

  # Expect node ID values in the first column
  expect_identical(
    betweenness_vals[, 1],
    1:10)

  # Expect numerical values in the
  # second column
  expect_type(
    betweenness_vals[, 2],
    "double")
})

test_that("Getting closeness is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 20,
      set_seed = 23)

  # Get closness values with `direction = all`
  closeness_vals_all <-
    get_closeness(
      graph = graph,
      direction = "all")

  # Expect a data frame as output
  expect_s3_class(
    closeness_vals_all, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(closeness_vals_all), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(closeness_vals_all), 10)

  # Expect node ID values in the first column
  expect_identical(
    closeness_vals_all[, 1],
    1:10)
})

test_that("Getting coreness values is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Get coreness values in all directions
  coreness_vals_all <- get_coreness(graph)

  # Expect a data frame as output
  expect_s3_class(
    coreness_vals_all, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(coreness_vals_all), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(coreness_vals_all), 10)

  # Expect node ID values in the first column
  expect_identical(
    coreness_vals_all[, 1],
    1:10)

  # Expect numeric values in the second column
  expect_type(
    coreness_vals_all[, 2],
    "double")

  # Get coreness values in the `in` direction
  coreness_vals_in <-
    get_coreness(
      graph = graph,
      direction = "in")

  # Expect a data frame as output
  expect_s3_class(
    coreness_vals_in, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(coreness_vals_in), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(coreness_vals_in), 10)

  # Expect node ID values in the first column
  expect_identical(
    coreness_vals_in[, 1],
    1:10)

  # Expect numeric values in the second column
  expect_type(
    coreness_vals_in[, 2],
    "double")

  # Get coreness values in the `out` direction
  coreness_vals_out <-
    get_coreness(
      graph = graph,
      direction = "out")

  # Expect a data frame as output
  expect_s3_class(
    coreness_vals_out, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(coreness_vals_out), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(coreness_vals_out), 10)

  # Expect node ID values in the first column
  expect_identical(
    coreness_vals_out[, 1],
    1:10)

  # Expect numeric values in the second column
  expect_type(
    coreness_vals_out[, 2],
    "double")

  # Expect an error if value for `direction`
  # is not any of `all`, `in`, or `out`
  expect_error(
    get_coreness(
      graph = graph,
      direction = "away"))
})

test_that("Getting closeness vitality is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Get closness vitality values
  closeness_vitality_vals <- get_closeness_vitality(graph)

  # Expect a data frame as output
  expect_s3_class(
    closeness_vitality_vals, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(closeness_vitality_vals), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(closeness_vitality_vals), 10)

  # Expect node ID values in the first column
  expect_identical(
    closeness_vitality_vals[, 1],
    1:10)

  # Expect numeric values in the second column
  expect_type(
    closeness_vitality_vals[, 2],
    "double")
})

test_that("Getting alpha centrality is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 15,
      m = 15,
      set_seed = 23)

  alpha_central_vals <-
    get_alpha_centrality(graph)

  # Expect a data frame as output
  expect_s3_class(
    alpha_central_vals, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(alpha_central_vals), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(alpha_central_vals), 15)

  # Expect node ID values in the first column
  expect_identical(
    alpha_central_vals[, 1],
    1:15)

  # Expect numerical values in the
  # second column
  expect_type(
    alpha_central_vals[, 2],
    "double")
})

test_that("Getting leverage centrality is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  leverage_central_vals <-
    get_leverage_centrality(graph)

  # Expect a data frame as output
  expect_s3_class(
    leverage_central_vals, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(leverage_central_vals), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(leverage_central_vals), 10)

  # Expect node ID values in the first column
  expect_identical(
    leverage_central_vals[, 1],
    1:10)

  # Expect numerical values in the
  # second column
  expect_type(
    leverage_central_vals[, 2],
    "double")
})

test_that("Getting authority centrality is possible", {

  # Create a random graph
  graph_1 <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Create a random graph
  graph_2 <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = weight,
      value = rnorm(n = 22, mean = 3, sd = 0.5)) %>%
    clear_selection()

  # Create a random graph with non-numeric weights
  graph_3 <-
    create_graph() %>%
    add_gnm_graph(
      n = 5,
      m = 6,
      set_seed = 23) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = weight,
      value = as.character(1:6)) %>%
    clear_selection()

  # Expect an error when the specified
  # `weights_attr` does not exist in the graph
  expect_error(
    graph_2 %>%
      get_authority_centrality(
        weights_attr = "weight_2"))

  # Expect an error when the specified
  # `weights_attr` does not exist in the graph
  expect_error(
    graph_2 %>%
      get_authority_centrality(
        weights_attr = "weight_2"))

  # Expect an error when the specified
  # `weights_attr` is not numeric
  expect_error(
    graph_3 %>%
      get_authority_centrality(
        weights_attr = "weight"))

  # Get values from a graph
  # without weights
  auth_central_vals <-
    get_authority_centrality(
      graph = graph_1)

  # Get values from a graph
  # with weights
  auth_central_vals_weight_1 <-
    get_authority_centrality(
      graph = graph_2,
      weights_attr = "weight")

  # Get values from a graph
  # with weights (w/o specifying the column)
  auth_central_vals_weight_2 <-
    get_authority_centrality(graph = graph_2)

  # Expect a data frame as output for all
  expect_s3_class(
    auth_central_vals, "data.frame")

  expect_s3_class(
    auth_central_vals_weight_1, "data.frame")

  expect_s3_class(
    auth_central_vals_weight_2, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(auth_central_vals), 2)

  expect_equal(
    ncol(auth_central_vals_weight_1), 2)

  expect_equal(
    ncol(auth_central_vals_weight_2), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(auth_central_vals), 10)

  expect_equal(
    nrow(auth_central_vals_weight_1), 10)

  expect_equal(
    nrow(auth_central_vals_weight_2), 10)

  # Expect node ID values in the first column
  expect_identical(
    auth_central_vals[, 1],
    1:10)

  expect_identical(
    auth_central_vals_weight_1[, 1],
    1:10)

  expect_identical(
    auth_central_vals_weight_1[, 1],
    1:10)

  # Expect numerical values in the
  # second column
  expect_type(
    auth_central_vals[, 2],
    "double")

  expect_type(
    auth_central_vals_weight_1[, 2],
    "double")

  expect_type(
    auth_central_vals_weight_2[, 2],
    "double")
})

test_that("Getting eigenvector centrality is possible", {

  suppressWarnings(RNGversion("3.5.0"))
  set.seed(23)

  # Create a random graph
  graph_1 <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Create a random graph
  graph_2 <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = weight,
      value = rnorm(n = 22, mean = 3, sd = 0.5)) %>%
    clear_selection()

  # Create a random graph with non-numeric weights
  graph_3 <-
    create_graph() %>%
    add_gnm_graph(
      n = 5,
      m = 6,
      set_seed = 23) %>%
    select_edges() %>%
    set_edge_attrs_ws(
      edge_attr = weight,
      value = as.character(1:6)) %>%
    clear_selection()

  # Expect an error when the specified
  # `weights_attr` does not exist in the graph
  expect_error(
    graph_2 %>%
    get_eigen_centrality(
      weights_attr = "weight_2"))

  # Expect an error when the specified
  # `weights_attr` does not exist in the graph
  expect_error(
    graph_2 %>%
      get_eigen_centrality(
        weights_attr = "weight_2"))

  # Expect an error when the specified
  # `weights_attr` is not numeric
  expect_error(
    graph_3 %>%
      get_eigen_centrality(
        weights_attr = "weight"))

  # Get values from a graph
  # without weights
  eigen_central_vals <-
    get_eigen_centrality(
      graph = graph_1)

  # Get values from a graph
  # with weights
  eigen_central_vals_weight_1 <-
    get_eigen_centrality(
      graph = graph_2,
      weights_attr = "weight")

  # Get values from a graph
  # with weights (w/o specifying the column)
  eigen_central_vals_weight_2 <-
    get_eigen_centrality(graph = graph_2)

  # Expect a data frame as output for all
  expect_s3_class(
    eigen_central_vals, "data.frame")

  expect_s3_class(
    eigen_central_vals_weight_1, "data.frame")

  expect_s3_class(
    eigen_central_vals_weight_2, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(eigen_central_vals), 2)

  expect_equal(
    ncol(eigen_central_vals_weight_1), 2)

  expect_equal(
    ncol(eigen_central_vals_weight_2), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(eigen_central_vals), 10)

  expect_equal(
    nrow(eigen_central_vals_weight_1), 10)

  expect_equal(
    nrow(eigen_central_vals_weight_2), 10)

  # Expect node ID values in the first column
  expect_identical(
    eigen_central_vals[, 1],
    1:10)

  expect_identical(
    eigen_central_vals_weight_1[, 1],
    1:10)

  expect_identical(
    eigen_central_vals_weight_1[, 1],
    1:10)

  # Expect numerical values in the
  # second column
  expect_type(
    eigen_central_vals[, 2],
    "double")

  expect_type(
    eigen_central_vals_weight_1[, 2],
    "double")

  expect_type(
    eigen_central_vals_weight_2[, 2],
    "double")
})

test_that("Getting radiality values is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get radiality values for all nodes in the graph
  radiality_vals <- get_radiality(graph)

  # Expect a data frame as output
  expect_s3_class(
    radiality_vals, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(radiality_vals), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(radiality_vals), 10)

  # Expect node ID values in the first column
  expect_identical(
    radiality_vals[, 1],
    as.integer(1:10))

  # Expect numerical values in the
  # second column
  expect_type(
    radiality_vals[, 2],
    "double")

  # Expect certain radiality values depending
  # on the `direction` parameter
  expect_gte(
    get_radiality(
      graph = graph,
      direction = "all")[, 2] %>% min(),
    2.1111)

  expect_lte(
    get_radiality(
      graph = graph,
      direction = "all")[, 2] %>% max(),
    3.1111)

  expect_gte(
    get_radiality(
      graph = graph,
      direction = "in")[, 2] %>% min(),
    3.4444)

  expect_lte(
    get_radiality(
      graph = graph,
      direction = "in")[, 2] %>% max(),
    4.6667)

  expect_gte(
    get_radiality(
      graph = graph,
      direction = "out")[, 2] %>% min(),
    0.6667)

  expect_lte(
    get_radiality(
      graph = graph,
      direction = "out")[, 2] %>% max(),
    5.2222)

  # Expect an error if using a `direction`
  # value that isn't valid
  expect_error(
    get_radiality(
      graph = graph,
      direction = "away"))
})

test_that("Getting PageRank values is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get pagerank values for all nodes in the graph
  pagerank_vals <- get_pagerank(graph)

  # Expect a data frame as output
  expect_s3_class(
    pagerank_vals, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(pagerank_vals), 2)

  # Expect 10 rows in the df
  expect_equal(
    nrow(pagerank_vals), 10)

  # Expect node ID values in the first column
  expect_identical(
    pagerank_vals[, 1],
    1:10)

  # Expect numerical values in the
  # second column
  expect_type(
    pagerank_vals[, 2],
    "double")
})

test_that("Getting articulation points is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get articulation points for the graph
  articulation_points <-
    get_articulation_points(graph)

  # Expect an integer vector as output
  expect_type(
    articulation_points, "integer")

  # Expect 2 values in the vector
  expect_length(articulation_points, 2)
})

test_that("Getting weakly connected components is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get connected components for the graph
  connected_components <-
    get_w_connected_cmpts(graph)

  # Expect a data frame as output
  expect_s3_class(
    connected_components, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(connected_components), 2)

  # Expect 30 rows in the df
  expect_equal(
    nrow(connected_components), 10)

  # Expect numerical values in the
  # second column
  expect_type(
    connected_components[, 2],
    "double")
})

test_that("Getting strongly connected components is possible", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 5,
      m = 10,
      set_seed = 23)

  # Get connected components for the graph
  s_connected_components <-
    get_s_connected_cmpts(graph)

  # Expect a data frame as output
  expect_s3_class(
    s_connected_components, "data.frame")

  # Expect 2 columns in the df
  expect_equal(
    ncol(s_connected_components), 2)

  # Expect 5 rows in the df
  expect_equal(
    nrow(s_connected_components), 5)

  # Expect numerical values in the
  # second column
  expect_type(
    s_connected_components[, 2],
    "double")
})

# Caching attributes

test_that("Setting a cache is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23)

  # Get the closeness values (as a data frame) for
  # all nodes from `1` to `10`
  closeness_df <-
    get_closeness(graph)

  # Set the values in the `closeness_df` column
  # `closeness` in the graph's cache
  graph_cache <-
    graph %>%
    set_cache(
      name = "closeness_vector",
      to_cache = closeness_df,
      col = "closeness")

  # Expect that the column `closeness` in the df
  # is equivalent to the values in the cache
  expect_equal(
    closeness_df$closeness,
    graph_cache$cache$closeness_vector)

  # Expect an error if providing a data frame
  # and not specifying a column to extract as a vector
  # or specifying a column that doesn't exist
  expect_snapshot(error = TRUE, {
    set_cache(
      graph = graph,
      name = "closeness_df_2",
      to_cache = closeness_df)

    set_cache(
      graph = graph,
      name = "closeness_df_3",
      to_cache = closeness_df,
      col = "nonexistent")
  })

  # Get the closeness values (as a vector)
  closeness_vec <- closeness_df$closeness

  # Set the values from the `closeness_vec` vector
  # in the graph's cache
  graph_cache_from_vec <-
    graph %>%
    set_cache(
      name = "closeness_vector",
      to_cache = closeness_vec)

  # Expect that the cache (originating from a vector)
  # is equivalent to the `closeness_vec` vector
  expect_equal(
    closeness_vec,
    graph_cache_from_vec$cache$closeness_vector)

  # Set the values in the `closeness_df`
  # column `closeness` in the graph's cache
  # but don't provide a name for the
  # cached object
  graph_cache_no_name <-
    graph %>%
    set_cache(
      to_cache = closeness_df,
      col = "closeness")

  # Expect that an unnamed cache
  # object with no other caches
  # available will have the name `1`
  expect_equal(
    closeness_df$closeness,
    graph_cache_no_name$cache$`1`)

  # Set again the values in the `closeness_df`
  # column `closeness` in the graph's cache
  # and don't provide a name for this
  # second cached object
  graph_cache_no_name_2 <-
    graph_cache_no_name %>%
    set_cache(
      to_cache = closeness_df,
      col = "closeness")

  # Expect two vectors in [graph]$cache
  expect_length(
    graph_cache_no_name_2$cache, 2)
})

test_that("Getting a cache is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23)

  # Get the closeness values (as a data frame) for
  # all nodes from `1` to `10`
  closeness_df <-
    get_closeness(graph)

  # Set the values in the `closeness_df` column
  # `closeness` in the graph's cache
  graph_cache <-
    graph %>%
    set_cache(
      name = "closeness_vector",
      to_cache = closeness_df,
      col = "closeness")

  # Get the graph's cached values
  cached <-
    get_cache(
      graph = graph_cache,
      name = "closeness_vector")

  # Expect a numeric vector of length 10
  expect_length(cached, 10)
  expect_type(cached, "double")


  # Expect that the cached values in `cached` are
  # equivalent to those in `graph$cache`
  expect_equal(
    cached, graph_cache$cache$closeness_vector)

  # Get the last cached vector
  cached_last <-
    get_cache(
      graph = graph_cache)

  # Expect a numeric vector of length 10
  expect_length(cached_last, 10)
  expect_type(cached_last, "double")

  # Create a new graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23)

  # Expect NA when trying to obtain a cache that is
  # not present
  expect_true(is.na(get_cache(graph)))
})

#' Use the depth-first search (dfs) algorithm
#'
#' @description
#'
#' With a chosen or random node serving as the starting point, perform a
#' depth-first search of the whole graph and return the node ID values visited.
#' The dfs algorithm differs from breadth-first search (bfs) in that dfs will
#' follow tree branches as far as possible until terminating at leaf node (bfs
#' traverses branches one level at a time).
#'
#' @inheritParams render_graph
#' @param node An optional node ID value to specify a single starting point for
#'   the dfs. If not provided, a random node from the graph will be chosen.
#' @param direction Using `all` (the default), the bfs will ignore edge
#'   direction while traversing through the graph. With `out` and `in`,
#'   traversals between adjacent nodes will respect the edge direction.
#'
#' @return A vector containing node ID values for nodes visited during the
#'   depth-first search. The order of the node IDs corresponds to the order
#'   visited.
#'
#' @examples
#' # Create a graph containing
#' # two balanced trees
#' graph <-
#'   create_graph() |>
#'   add_balanced_tree(
#'     k = 2, h = 2) |>
#'   add_balanced_tree(
#'   k = 3, h = 2)
#'
#' # Perform a depth-first
#' # search of the graph,
#' # beginning at the root
#' # node `1` (the default
#' # `direction = "all"`
#' # doesn't take edge
#' # direction into account)
#' graph |>
#'   do_dfs(node = 1)
#'
#' # If not specifying a
#' # starting node, the function
#' # will begin the search
#' # from a random node
#' graph |>
#'   do_dfs()
#'
#' # It's also possible to
#' # perform dfs while taking
#' # into account edge direction;
#' # using `direction = "in"`
#' # causes the dfs routine to
#' # visit nodes along inward edges
#' graph |>
#'   do_dfs(
#'     node = 1,
#'     direction = "in")
#'
#' # Using `direction = "out"`
#' # results in the dfs moving
#' # along solely outward edges
#' graph |>
#'   do_dfs(
#'     node = 1,
#'     direction = "out")
#'
#' @export
do_dfs <- function(
    graph,
    node = NULL,
    direction = "all"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: direction is in the allowed values.
  rlang::arg_match0(direction, c("all", "out", "in"))

  # If no node provided, choose a random node
  if (is.null(node)) {
    node <- sample(graph$nodes_df[, 1], 1)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Perform the depth-first search algorithm in
  # the direction requested
  if (direction == "all") {

    dfs_result <-
      igraph::dfs(
        graph = ig_graph,
        root = node,
        mode = "all")

  } else if (direction == "out") {

    dfs_result <-
      igraph::dfs(
        graph = ig_graph,
        root = node,
        mode = "out")

  } else if (direction == "in") {

    dfs_result <-
      igraph::dfs(
        graph = ig_graph,
        root = node,
        mode = "in")

  }

  # Get the nodes visited during the dfs
  as.integer(as.numeric(dfs_result$order))
}

# Performing the dfs and bfs algorithms

test_that("the dfs algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Perform a depth-first search of the graph,
  # beginning at the root node `1` (the default
  # `direction = "all"` doesn't take edge
  # direction into account)
  dfs_all <-
    graph %>%
    do_dfs(node = 1)

  # Expect that an integer vector is returned
  expect_type(dfs_all, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    dfs_all,
    count_nodes(graph = graph))

  # Expect that certain values are returned
  expect_equal(
    dfs_all,
    c(1, 5, 4, 8, 3, 2, 6, 10, 9, 7))

  # If not specifying a starting node, the function
  # will begin the search from a random node
  dfs_all_no_start <-
    graph %>%
    do_dfs()

  # Expect that an integer vector is returned
  expect_type(dfs_all_no_start, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    dfs_all_no_start,
    count_nodes(graph = graph))

  # Using `direction = "in"` to cause the dfs
  # routine to visit nodes along inward edges
  dfs_in <-
    graph %>%
    do_dfs(
      node = 1,
      direction = "in")

  # Expect that an integer vector is returned
  expect_type(dfs_in, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    dfs_in,
    count_nodes(graph = graph))

  # Expect that certain values are returned
  expect_equal(
    dfs_in,
    c(1, 5, 2, 3, 10, 4, 6, 7, 8, 9))

  # Using `direction = "out"` results in the dfs
  # moving along solely outward edges
  dfs_out <-
    graph %>%
    do_dfs(
      node = 1,
      direction = "out")

  # Expect that an integer vector is returned
  expect_type(dfs_out, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    dfs_out,
    count_nodes(graph = graph))

  # Expect that certain values are returned
  expect_equal(
    dfs_out,
    c(1, 2, 3, 6, 8, 10, 4, 9, 5, 7))

  # Expect an error if performing dfs without
  # a node data frame in the graph
  expect_error(
    create_graph() %>%
      do_dfs())

  # Expect an error if an invalid `direction`
  # value provided
  expect_error(
    graph %>%
      do_dfs(direction = "any"))
})

test_that("the bfs algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  # Perform a breadth-first search of the graph,
  # beginning at the root node `1` (the default
  # `direction = "all"` doesn't take edge
  # direction into account)
  bfs_all <-
    graph %>%
    do_bfs(node = 1)

  # Expect that an integer vector is returned
  expect_type(bfs_all, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    bfs_all,
    count_nodes(graph = graph))

  # Expect that certain values are returned
  expect_equal(
    bfs_all,
    c(1, 5, 4, 10, 8, 9, 3, 2, 6, 7))

  # If not specifying a starting node, the function
  # will begin the search from a random node
  bfs_all_no_start <-
    graph %>%
    do_bfs()

  # Expect that an integer vector is returned
  expect_type(bfs_all_no_start, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    bfs_all_no_start,
    count_nodes(graph = graph))

  # Using `direction = "in"` to cause the bfs
  # routine to visit nodes along inward edges
  bfs_in <-
    graph %>%
    do_bfs(
      node = 1,
      direction = "in")

  # Expect that an integer vector is returned
  expect_type(bfs_in, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    bfs_in,
    count_nodes(graph = graph))

  # Expect that certain values are returned
  expect_equal(
    bfs_in,
    c(1, 5, 2, 3, 10, 4, 6, 7, 8, 9))

  # Using `direction = "out"` results in the bfs
  # moving along solely outward edges
  bfs_out <-
    graph %>%
    do_bfs(
      node = 1,
      direction = "out")

  # Expect that an integer vector is returned
  expect_type(bfs_out, "integer")

  # Expect the length of the vector to be the
  # same as the number of nodes in the graph
  expect_length(
    bfs_out,
    count_nodes(graph = graph))

  # Expect that certain values are returned
  expect_equal(
    bfs_out,
    c(1, 2, 3, 6, 8, 10, 4, 9, 5, 7))

  # Expect an error if performing bfs without
  # a node data frame in the graph
  expect_error(
    create_graph() %>%
      do_bfs())

  # Expect an error if an invalid `direction`
  # value provided
  expect_error(
    graph %>%
      do_bfs(direction = "any"))
})

#' Add graph object to a graph series object
#'
#' @description
#'
#' Add a graph object to an extant graph series object for storage of multiple
#' graphs across a sequential or temporal one-dimensional array.
#'
#' @param graph_series A graph series object to which the graph object will be
#'   added.
#' @param graph A graph object to add to the graph series object.
#'
#' @return A graph series object of type `dgr_graph_1D`.
#'
#' @examples
#' # Create three graphs
#' graph_1 <-
#'   create_graph() %>%
#'   add_path(n = 4)
#'
#' graph_2 <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' graph_3 <-
#'   create_graph() %>%
#'   add_star(n = 6)
#'
#' # Create an empty graph series
#' # and add the graphs
#' series <-
#'   create_graph_series() %>%
#'   add_graph_to_graph_series(
#'     graph = graph_1) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_2) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_3)
#'
#' # Count the number of graphs
#' # in the graph series
#' series %>%
#'   count_graphs_in_graph_series()
#'
#' @export
add_graph_to_graph_series <- function(
    graph_series,
    graph
) {

  # Validation: Graph object is valid
  rlang::check_required(graph)
  check_graph_valid(graph)


  # Get the series type
  series_type <- graph_series$series_type

  # Stop function if graph series type is not valid
  rlang::arg_match0(series_type, c("sequential", "temporal"), arg_nm = "graph series")

  # Add graph to series
  graph_series$graphs[[length(graph_series$graphs) + 1]] <- graph

  graph_series
}

test_that("graph series information can be obtained", {

  # Create an empty graph series
  empty_series <- create_graph_series()

  empty_series_info <-
    empty_series %>%
    get_graph_series_info()

  # Expect that `empty_series_info` is a
  # data frame object
  expect_s3_class(
    empty_series_info, "data.frame")

  # Expect that the data frame has no rows
  expect_equal(
    nrow(empty_series_info), 0)

  # Expect that the data frame has 7 columns
  expect_equal(
    ncol(empty_series_info), 7)

  # Expect specific column classes in the data frame
  expect_type(
    empty_series_info[, 1], "integer")

  expect_type(
    empty_series_info[, 2], "character")

  expect_s3_class(
    empty_series_info[, 3], "POSIXct")

  expect_type(
    empty_series_info[, 4], "character")

  expect_type(
    empty_series_info[, 5], "integer")

  expect_type(
    empty_series_info[, 6], "integer")

  expect_type(
    empty_series_info[, 7], "logical")

  # Create 3 graph objects
  graph_1 <-
    create_graph() %>%
    add_node(type = 1) %>%
    add_node(type = 2) %>%
    add_node(type = 3) %>%
    add_edge(
      from = 1,
      to = 3) %>%
    add_edge(
      from = 1,
      to = 2) %>%
    add_edge(
      from = 2,
      to = 3)

  graph_2 <-
    graph_1 %>%
    add_node(type = 4) %>%
    add_edge(
      from = 4,
      to = 3)

  graph_3 <-
    graph_2 %>%
    add_node(type = 5) %>%
    add_edge(
      from = 5,
      to = 2)

  # Create an empty graph series
  series <- create_graph_series(series_type = "sequential")

  # Add graphs to the graph series
  series <-
    series %>%
    add_graph_to_graph_series(
      graph = graph_1) %>%
    add_graph_to_graph_series(
      graph = graph_2) %>%
    add_graph_to_graph_series(
      graph = graph_3)

  # Get information on the graphs in the series
  info_on_series <-
    series %>%
    get_graph_series_info()

  # Expect that `info_on_series` is a data frame object
  expect_s3_class(
    info_on_series, "data.frame")

  # Expect that the data frame has 3 rows
  expect_equal(
    nrow(info_on_series), 3)

  # Expect that the data frame has 7 columns
  expect_equal(
    ncol(info_on_series), 7)

  # Expect specific column classes in the data frame
  expect_type(
    info_on_series[, 1], "integer")

  expect_type(
    info_on_series[, 2], "character")

  expect_s3_class(
    info_on_series[, 3], "POSIXct")

  expect_type(
    info_on_series[, 4], "character")

  expect_type(
    info_on_series[, 5], "integer")

  expect_type(
    info_on_series[, 6], "integer")

  expect_type(
    info_on_series[, 7], "logical")

  # Expect that the values in the `graph`
  # column are sequential
  expect_equal(
    info_on_series[1, 1], 1)

  expect_equal(
    info_on_series[2, 1], 2)

  expect_equal(
    info_on_series[3, 1], 3)

  # Create a temporal graph series and add
  # a graph with name and time information
  graph_series_temporal_type <-
    create_graph_series(series_type = "temporal")

  graph <-
    create_graph(
      graph_name = "graph_no_tz_provided") %>%
    set_graph_time(
      time = "2015-03-25 03:00",
      tz = "GMT")

  graph_series_temporal_type <-
    graph_series_temporal_type %>%
    add_graph_to_graph_series(
      graph = graph)

  info_on_series_temporal <-
    graph_series_temporal_type %>%
    get_graph_series_info()

  # Expect that `info_on_series_temporal` is
  # a data frame object
  expect_s3_class(
    info_on_series_temporal, "data.frame")

  # Expect that the data frame has 1 row
  expect_equal(
    nrow(info_on_series_temporal), 1)

  # Expect that the data frame has 7 columns
  expect_equal(
    ncol(info_on_series_temporal), 7)

  # Expect specific column classes in the data frame
  expect_type(
    info_on_series_temporal[, 1], "integer")

  expect_type(
    info_on_series_temporal[, 2], "character")

  expect_s3_class(
    info_on_series_temporal[, 3], "POSIXct")

  expect_type(
    info_on_series_temporal[, 4], "character")

  expect_type(
    info_on_series_temporal[, 5], "integer")

  expect_type(
    info_on_series_temporal[, 6], "integer")

  expect_type(
    info_on_series_temporal[, 7], "logical")

  # Expect that the `name`, `date_time`,
  # and `tz` columns are populated with information
  expect_equal(
    info_on_series_temporal[, 2], "graph_no_tz_provided")

  expect_false(
    is.na(info_on_series_temporal[, 3]))

  expect_equal(
    info_on_series_temporal[, 4], "GMT")
})

# Creating and manipulating graph series objects")

test_that("creating an empty series is possible", {

  # Create an empty graph series of the `sequential` type
  series <-
    create_graph_series(series_type = "sequential")

  # Expect an object of class `dgr_graph_1D`
  expect_s3_class(
    series, "dgr_graph_1D")

  # Expect that the series type is `sequential`
  expect_equal(
    series$series_type, "sequential")

  # Expect that several series components are `NULL`
  expect_null(series$graphs)
  expect_null(series$series_name)
  expect_null(series$series_scripts)

  # Expect that the series is empty
  expect_equal(
    count_graphs_in_graph_series(series), 0)

  # Create an empty graph series of the `temporal` type
  series_temporal <-
    create_graph_series(series_type = "temporal")

  # Expect that this series is empty
  expect_equal(
    count_graphs_in_graph_series(series_temporal), 0)

  # Expect that several series components are `NULL`
  expect_null(series_temporal$graphs)
  expect_null(series_temporal$series_name)
  expect_null(series_temporal$series_scripts)

  # Expect an object of class `dgr_graph_1D`
  expect_s3_class(
    series_temporal, "dgr_graph_1D")

  # Expect that the series type is `sequential`
  expect_equal(
    series_temporal$series_type, "temporal")
})

test_that("adding graphs to a series is also possible", {

  # Create an empty graph series
  series <-
    create_graph_series(series_type = "sequential")

  # Create a set of graphs for a graph series
  graph_1 <-
    create_graph() %>%
    add_path(n = 4)

  graph_2 <-
    create_graph() %>%
    add_cycle(n = 5)

  graph_3 <-
    create_graph() %>%
    add_star(n = 6)

  # Add graphs to the graph series
  series <-
    series %>%
    add_graph_to_graph_series(
      graph = graph_1) %>%
    add_graph_to_graph_series(
      graph = graph_2) %>%
    add_graph_to_graph_series(
      graph = graph_3)

  # Expect an object of class `dgr_graph_1D`
  expect_s3_class(
    series, "dgr_graph_1D")

  # Expect that the series type is `sequential`
  expect_equal(
    series$series_type, "sequential")

  # Expect that the `graphs` component is not `NULL`
  expect_false(is.null(series$graphs))

  # Expect that several series components are `NULL`
  expect_null(series$series_name)
  expect_null(series$series_scripts)

  # Expect that the series has a graph count of 3
  expect_equal(
    count_graphs_in_graph_series(series), 3L)

  expect_length(series$graphs, 3)

  # Expect that the graphs within the graph series object
  # are indeed graph objects
  expect_s3_class(
    series$graphs[[1]], "dgr_graph")

  expect_s3_class(
    series$graphs[[2]], "dgr_graph")

  expect_s3_class(
    series$graphs[[3]], "dgr_graph")

  # Expect that the graphs within the graph series object
  # are the same as those outside the series
  expect_equal(
    graph_1, series$graphs[[1]])

  expect_equal(
    graph_2, series$graphs[[2]])

  expect_equal(
    graph_3, series$graphs[[3]])

  # Create a series with a graph
  series_w_graph <-
    create_graph_series(
      graph = graph_1,
      series_type = "sequential")

  # Expect an object of class `dgr_graph_1D`
  expect_s3_class(
    series_w_graph, "dgr_graph_1D")

  # Expect that the series type is `sequential`
  expect_equal(
    series_w_graph$series_type, "sequential")

  # Expect that the 'graphs' component is not `NULL`
  expect_false(is.null(series_w_graph$graphs))

  # Expect that the series has a graph count of 1
  expect_equal(
    count_graphs_in_graph_series(series_w_graph), 1)

  # Expect that several series components are `NULL`
  expect_null(series_w_graph$series_name)
  expect_null(series_w_graph$series_scripts)

  # Expect that the graphs within the graph series object
  # are indeed graph objects
  expect_s3_class(
    series_w_graph$graphs[[1]], "dgr_graph")

  # Expect an error when adding something other
  # than a graph object to a graph series
  expect_error(
    add_graph_to_graph_series(
      graph_series = series_w_graph,
      graph = series_w_graph))

  # Expect an error if graph series type is not valid
  graph_series_invalid_type <-
    create_graph_series(series_type = "circular")

  expect_error(
    add_graph_to_graph_series(
      graph_series = graph_series_invalid_type,
      graph = graph_1))
})

test_that("removing graphs from a series is possible", {

  # Create an empty graph series
  series <-
    create_graph_series(
      series_type = "sequential")

  # Create a set of graphs for a graph series
  graph_1 <-
    create_graph() %>%
    add_path(n = 4)

  graph_2 <-
    create_graph() %>%
    add_cycle(n = 5)

  graph_3 <-
    create_graph() %>%
    add_star(n = 6)

  # Add graphs to the graph series
  series <-
    series %>%
    add_graph_to_graph_series(
      graph = graph_1) %>%
    add_graph_to_graph_series(
      graph = graph_2) %>%
    add_graph_to_graph_series(
      graph = graph_3)

  # Expect that the series has a graph count of 3
  expect_equal(
    count_graphs_in_graph_series(series), 3)

  expect_length(series$graphs, 3)

  # Remove the last graph from the series
  series_2 <-
    remove_graph_from_graph_series(graph_series = series)

  # Expect that the graph count is now 2
  expect_equal(
    count_graphs_in_graph_series(series_2), 2)

  # Remove the first graph from the series
  series_removed_1 <-
    remove_graph_from_graph_series(
      graph_series = series,
      index = 1)

  # Expect that the first created graph is
  # not in the series
  expect_false(
    count_nodes(graph = series_removed_1$graphs[[1]]) == count_nodes(graph = graph_1))

  # Expect that `graph_1` in the series is
  # equivalent to the `graph_2` object
  expect_equal(
    count_nodes(graph = series_removed_1$graphs[[1]]),
    count_nodes(graph = graph_2))

  # Remove the first graph from the series using
  # the `first` character vector
  series_removed_2 <-
    remove_graph_from_graph_series(
      graph_series = series,
      index = "first")

  # Expect that the first created graph is
  # not in the series
  expect_false(
    count_nodes(graph = series_removed_2$graphs[[1]]) == count_nodes(graph = graph_1))

  # Expect that graph 1 in the series is
  # equivalent to the `graph_2` object
  expect_true(
    count_nodes(graph = series_removed_2$graphs[[1]]) == count_nodes(graph = graph_2))
})

test_that("subsetting graphs from a temporal series is possible", {

  # Create a set of graphs for a graph series
  graph_time_1 <-
    create_graph(
      graph_name = "graph_with_time_1") %>%
    set_graph_time(
      time = "2015-03-25 03:00",
      tz = "GMT") %>%
    add_path(n = 4)

  graph_time_2 <-
    create_graph(
      graph_name = "graph_with_time_2") %>%
    set_graph_time(
      time = "2015-03-26 03:00",
      tz = "GMT") %>%
    add_cycle(n = 5)

  graph_time_3 <-
    create_graph(
      graph_name = "graph_with_time_3") %>%
    set_graph_time(
      time = "2015-03-27 15:00",
      tz = "GMT") %>%
    add_star(n = 6)

  # Create an empty graph series
  series_temporal <-
    create_graph_series(
      series_type = "temporal")

  # Add graphs to the graph series
  series_temporal <-
    series_temporal %>%
    add_graph_to_graph_series(
      graph = graph_time_1) %>%
    add_graph_to_graph_series(
      graph = graph_time_2) %>%
    add_graph_to_graph_series(
      graph = graph_time_3)

  # Expect a graph count of 3
  expect_equal(
    count_graphs_in_graph_series(series_temporal), 3)

  # Subset graph series by sequence
  series_sequence_subset <-
    filter_graph_series(
      graph_series = series_temporal,
      by = "number",
      values = 2)

  # Expect a single graph in the series
  expect_equal(
    count_graphs_in_graph_series(series_sequence_subset), 1)

  # Expect that this subset graph is the same
  # as `graph_time_2`
  expect_equal(
    count_nodes(graph = series_sequence_subset$graphs[[1]]),
    count_nodes(graph = graph_time_2))

  # Subset graph series by date-time
  series_time_subset <-
    filter_graph_series(
      graph_series = series_temporal,
      by = "time",
      values = c("2015-03-25 12:00",
                 "2015-03-26 12:00"),
      tz = "GMT")

  # Expect 2 graphs in the series
  expect_equal(
    count_graphs_in_graph_series(series_time_subset), 2)
})

test_that("Getting a graph from a series is possible", {

  # Create a set of graphs for a graph series
  graph_1 <-
    create_graph() %>%
    add_path(n = 4)

  graph_2 <-
    create_graph() %>%
    add_cycle(n = 5)

  graph_3 <-
    create_graph() %>%
    add_star(n = 6)

  # Create an empty graph series and add
  # the graphs
  series <-
    create_graph_series() %>%
    add_graph_to_graph_series(
      graph = graph_1) %>%
    add_graph_to_graph_series(
      graph = graph_2) %>%
    add_graph_to_graph_series(
      graph = graph_3)

  # Get the second graph in the series
  extracted_graph <-
    get_graph_from_graph_series(
      graph_series = series,
      graph_no = 2)

  # Expect the equivalent object among
  # `extracted_graph` and `graph_2`
  expect_equal(extracted_graph, graph_2)

  # Expect an error if extracting a graph from
  # an empty graph series
  expect_snapshot(error = TRUE,
    create_graph_series() %>%
      get_graph_from_graph_series(
        graph_no = 1))

  # Expect an error if the index for the graph to
  # be extracted is out of range
  expect_error(
    series %>%
      get_graph_from_graph_series(
        graph_no = 4))
})

test_that("rendering a graph is indeed possible", {

  # Create a node data frame
  ndf <-
    create_node_df(
      n = 26,
      type = "letter",
      shape = sample(c("circle", "rectangle"),
                     length(1:26),
                     replace = TRUE),
      fillcolor = sample(c("steelblue", "gray80",
                           "#FFC0CB", "green",
                           "azure", NA),
                         length(1:26),
                         replace = TRUE))

  # Create an edge data frame
  edf <-
    create_edge_df(
      from = sample(1:26, replace = TRUE),
      to = sample(1:26, replace = TRUE),
      relationship = "letter_to_letter")

  # Create the graph object using the node and
  # edge data frames
  graph <-
    create_graph(
      nodes_df = ndf,
      edges_df = edf)

  # Render the graph object and create a
  # `grViz`/`htmlwidget` object
  rendered_graph <- render_graph(graph)

  # Expect that the `rendered_graph` object inherits
  # from `grViz` & `htmlwidget`
  expect_s3_class(
    rendered_graph, c("grViz", "htmlwidget"))
})

test_that("rendering a graph from a series is also possible", {

  # Create a set of graphs for a graph series
  graph_1 <-
    create_graph() %>%
    add_path(n = 4)

  graph_2 <-
    create_graph() %>%
    add_cycle(n = 5)

  graph_3 <-
    create_graph() %>%
    add_star(n = 6)

  # Create an empty graph series
  series <-
    create_graph_series(series_type = "sequential")

  # Add graphs to the graph series
  series <-
    series %>%
    add_graph_to_graph_series(
      graph = graph_1) %>%
    add_graph_to_graph_series(
      graph = graph_2) %>%
    add_graph_to_graph_series(
      graph = graph_3)

  # View the second graph from the graph series in
  # the RStudio Viewer
  series_graph_2 <-
    render_graph_from_graph_series(
      graph_series = series,
      graph_no = 2)

  # Expect error when rendering graph 4 from the
  # series (which doesn't exist)
  expect_error(
    render_graph_from_graph_series(
      graph_series = series,
      graph_no = 4))

  # Expect that each of the graphs is different
  expect_true(
    render_graph_from_graph_series(
      graph_series = series,
      graph_no = 1)$x$diagram !=
      render_graph_from_graph_series(
        graph_series = series,
        graph_no = 2)$x$diagram)

  expect_true(
    render_graph_from_graph_series(
      graph_series = series,
      graph_no = 2)$x$diagram !=
      render_graph_from_graph_series(
        graph_series = series,
        graph_no = 3)$x$diagram)

  expect_true(
    render_graph_from_graph_series(
      graph_series = series,
      graph_no = 1)$x$diagram !=
      render_graph_from_graph_series(
        graph_series = series,
        graph_no = 3)$x$diagram)

  # Create an empty graph series
  empty_series <-
    create_graph_series(series_type = "sequential")

  # Expect an error if there are no graphs
  # in the series
  expect_error(
    render_graph_from_graph_series(
      graph_series = empty_series))
})

#' Get a vector of node ID values
#'
#' @description
#'
#' Obtain a vector of node ID values from a graph object. An optional filter by
#' node attribute can limit the set of node ID values returned.
#'
#' @inheritParams render_graph
#' @param conditions An option to use filtering conditions for the retrieval of
#'   nodes.
#'
#' @return A vector of node ID values.
#'
#' @examples
#' # Create a node data
#' # frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = "letter",
#'     color = c(
#'       "red", "green",
#'       "blue", "blue"),
#'     value = c(
#'       3.5, 2.6, 9.4, 2.7))
#'
#' # Create a graph using
#' # the ndf
#' graph <-
#'   create_graph(
#'     nodes_df = ndf)
#'
#' # Get a vector of all nodes in a graph
#' graph |> get_node_ids()
#'
#' # Get a vector of node ID values using a
#' # numeric comparison (i.e., all nodes with
#' # `value` attribute greater than 3)
#' graph |>
#'   get_node_ids(
#'     conditions = value > 3)
#'
#' # Get a vector of node ID values using
#' # a match pattern (i.e., all nodes with
#' # `color` attribute of `green`)
#' graph |>
#'   get_node_ids(
#'     conditions = color == "green")
#'
#' # Use multiple conditions to return nodes
#' # with the desired attribute values
#' graph |>
#'   get_node_ids(
#'     conditions =
#'       color == "blue" &
#'       value > 5)
#'
#' @export
get_node_ids <- function(
    graph,
    conditions = NULL
) {

  if (is_graph_empty(graph)) {
    return(NA)
  }

  nodes_df <- graph$nodes_df


  # If conditions are provided then
  # pass in those conditions and filter the
  # data frame of `nodes_df`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    nodes_df <- dplyr::filter(.data = nodes_df, {{ conditions }})
  }

  # If no nodes remain then return NA
  if (nrow(nodes_df) == 0) {
    return(NA)
  }

  nodes_df |> dplyr::pull("id")
}

# Creating a subgraph

test_that("a subgraph can be created and such an object is correct", {

  # Create a simple graph
  nodes <-
    create_node_df(
      n = 8,
      value = c(3.5, 2.6, 9.4, 2.7,
                5.2, 2.1, 4.8, 8.5))

  edges <-
    create_edge_df(
      from = c(1, 2, 3, 7, 5,
               5, 8, 6, 1, 3),
      to = c(4, 3, 1, 3, 8,
             2, 4, 5, 6, 4),
      value = c(1.3, 2.8, 3.5, 7.1,
                0.2, 2.1, 4.8, 8.5,
                3.5, 7.2))

  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Create a selection of nodes, stored within the
  # graph object
  graph_ns <-
    graph %>%
    select_nodes(conditions = value > 3)

  # Create a subgraph based on the selection
  subgraph_ns <-
    transform_to_subgraph_ws(graph_ns)

  # Expect that only those nodes with a value >3 are in the subgraph
  expect_setequal(
    get_node_ids(subgraph_ns),
    c(1, 3, 5, 7, 8))

  # Expect only certain edges to be present in the subgraph
  expect_setequal(
    get_edges(
      graph = subgraph_ns,
      return_type = "vector"),
      c("3->1", "7->3", "5->8"))

  # Create a selection of edges, stored within the
  # graph object
  graph_es <-
    graph %>%
    select_edges(conditions = value > 4)

  # Create a subgraph based on the selection
  subgraph_es <-
    transform_to_subgraph_ws(graph_es)

  # Expect that only those edges with a value >4 are in the subgraph
  expect_contains(get_edge_ids(subgraph_es), c(4, 7, 8, 10))

  # Expect only certain edges to be present in the subgraph
  expect_contains(
    get_edges(
      graph = subgraph_es,
      return_type = "vector"),
    c("7->3", "8->4", "6->5", "3->4"))

  # Expect an error when attempting to create a subgraph with
  # a graph without an active selection
  graph <- clear_selection(graph)

  expect_snapshot(error = TRUE,
    transform_to_subgraph_ws(graph))
})

# Add graphs to graphs

test_that("Adding a balanced tree is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Add a balanced tree
  graph <-
    add_balanced_tree(
      graph = graph,
      k = 3,
      h = 3,
      type = "a",
      rel = "z")

  # Expect 40 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 40)

  # Expect 39 edges to have been created
  expect_equal(
    count_edges(graph = graph), 39)

  # Expect node ID values from 1 to 40
  expect_identical(
    get_node_ids(graph), 1:40)

  # Expect label values from 1 to 40
  expect_identical(
    graph$nodes_df$label,
    as.character(1:40))

  # Expect type values to all be `a`
  expect_in(graph$nodes_df$type, "a")

  # Expect rel values to all be `z`
  expect_in(graph$edges_df$rel, "z")

  # Expect an error if h <2 or k <2
  expect_snapshot(error = TRUE, {
    # k < 2
    add_balanced_tree(
      graph = graph,
      k = 1,
      h = 3,
      type = "a",
      rel = "z")
    # h < 2
    add_balanced_tree(
      graph = graph,
      k = 3,
      h = 1,
      type = "a",
      rel = "z")
  }
  )

  # Add another balanced tree to the graph
  graph <-
    add_balanced_tree(
      graph = graph,
      k = 2,
      h = 2,
      type = "b",
      rel = "y")

  # Expect that 47 nodes are now in the graph
  expect_equal(count_nodes(graph = graph), 47)

  # Expect 45 edges are now in the graph
  expect_equal(count_edges(graph = graph), 45)

  # Expect node ID values from 1 to 47
  expect_identical(
    get_node_ids(graph), 1:47)

  # Expect label values from 1 to 47
  expect_equal(
    graph$nodes_df$label,
    as.character(1:47))

  # Expect type values to be either `a` or `b`
  expect_in(graph$nodes_df$type, c("a", "b"))

  # Expect rel values to be either `y` or `z`
  expect_in(graph$edges_df$rel, c("y", "z"))
})

test_that("Adding a cycle is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Add a cycle
  graph <-
    add_cycle(
      graph = graph,
      n = 3,
      type = "a",
      rel = "z")

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 3 edges to have been created
  expect_equal(
    count_edges(graph = graph), 3)

  # Expect node ID values from 1 to 3
  expect_identical(
    get_node_ids(graph), 1:3)

  # Expect label values from 1 to 3
  expect_identical(
    graph$nodes_df$label,
    as.character(1:3))

  # Expect type values to all be `a`
  expect_equal(
    unique(graph$nodes_df$type), "a")

  # Expect rel values to all be `z`
  expect_equal(
    unique(graph$edges_df$rel), "z")

  # Expect an error if n is <3
  expect_error(
    add_cycle(
      graph = graph,
      n = 2,
      type = "a",
      rel = "z"))

  # Add another cycle to the graph
  graph <-
    add_cycle(
      graph = graph,
      n = 3,
      type = "b",
      rel = "y")

  # Expect that 6 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph), 6)

  # Expect 6 edges are now in the graph
  expect_equal(
    count_edges(graph = graph), 6)

  # Expect node ID values from 1 to 6
  expect_identical(
    get_node_ids(graph), 1:6)

  # Expect label values from 1 to 6
  expect_identical(
    graph$nodes_df$label,
    as.character(1:6))

  # Expect type values to be either `a` or `b`
  expect_in(graph$nodes_df$type, c("a", "b"))

  # Expect rel values to be either `y` or `z`
  expect_in(graph$edges_df$rel, c("y", "z"))

  # Create a graph with a cycle that
  # has different types of node and edge
  # attributes included
  graph_2 <-
    create_graph() %>%
    add_cycle(
      n = 6,
      label = c("one", "two",
                "three", "four",
                "five", "six"),
      type = c("a", "a",
               "b", "b",
               "c", "c"),
      rel = c("cycle_a", "cycle_a",
              "cycle_b", "cycle_b",
              "cycle_c", "cycle_c"))

  # Get the graph's node data frame
  ndf <- get_node_df(graph_2)

  # Get the graph's edge data frame
  edf <- get_edge_df(graph_2)

  # Expect that 6 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph_2), 6)

  # Expect 6 edges are now in the graph
  expect_equal(
    count_edges(graph = graph_2), 6)

  # Expect node ID values from 1 to 6
  expect_identical(
    ndf$id, 1:6)

  # Expect specific node `label` values
  expect_identical(
    ndf$label,
    c("one", "two",
      "three", "four",
      "five", "six"))

  # Expect specific node `type` values
  expect_identical(
    ndf$type,
    c("a", "a", "b",
      "b", "c", "c"))

  # Expect specific edge `rel` values
  expect_identical(
    edf$rel,
    c("cycle_a", "cycle_a",
      "cycle_b", "cycle_b",
      "cycle_c", "cycle_c"))
})

test_that("Adding a path is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Add a path
  graph <-
    add_path(
      graph = graph,
      n = 3,
      type = "a",
      rel = "z")

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 2 edges to have been created
  expect_equal(
    count_edges(graph = graph), 2)

  # Expect node ID values from 1 to 3
  expect_identical(
    get_node_ids(graph), 1:3)

  # Expect label values from 1 to 3
  expect_identical(
    graph$nodes_df$label,
    as.character(1:3))

  # Expect type values to all be `a`
  expect_equal(
    unique(graph$nodes_df$type), "a")

  # Expect rel values to all be `z`
  expect_equal(
    unique(graph$edges_df$rel), "z")

  # Expect an error if n is <2
  expect_snapshot(error = TRUE,
    add_path(
      graph = graph,
      n = 1,
      type = "a",
      rel = "z"))

  # Add another path to the graph
  graph <-
    add_path(
      graph = graph,
      n = 3,
      type = "b",
      rel = "y")

  # Expect that 6 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph), 6)

  # Expect 6 edges are now in the graph
  expect_equal(
    count_edges(graph = graph), 4)

  # Expect node ID values from 1 to 6
  expect_identical(
    get_node_ids(graph), 1:6)

  # Expect label values from 1 to 6
  expect_identical(
    graph$nodes_df$label,
    as.character(1:6))

  # Expect type values to be either `a` or `b`
  expect_setequal(
    graph$nodes_df$type,
    c("a", "b"))

  # Expect rel values to be either `y` or `z`
  expect_setequal(
    graph$edges_df$rel,
    c("y", "z"))

  # Create a graph with a path that
  # has different types of node and edge
  # attributes included
  graph_2 <-
    create_graph() %>%
    add_path(
      n = 6,
      label = c("one", "two",
                "three", "four",
                "five", "six"),
      type = c("a", "a",
               "a", "b",
               "b", "b"),
      rel = c("path_a", "path_b",
              "path_c", "path_d",
              "path_e"))

  # Get the graph's node data frame
  ndf <- get_node_df(graph_2)

  # Get the graph's edge data frame
  edf <- get_edge_df(graph_2)

  # Expect that 6 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph_2), 6)

  # Expect 5 edges are now in the graph
  expect_equal(
    count_edges(graph = graph_2), 5)

  # Expect node ID values from 1 to 6
  expect_identical(
    ndf$id, 1:6)

  # Expect specific node `label` values
  expect_identical(
    ndf$label,
    c("one", "two",
      "three", "four",
      "five", "six"))

  # Expect specific node `type` values
  expect_identical(
    ndf$type,
    c("a", "a", "a",
      "b", "b", "b"))

  # Expect specific edge `rel` values
  expect_identical(
    edf$rel,
    c("path_a", "path_b",
      "path_c", "path_d",
      "path_e"))
})

test_that("Adding a prism is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Add a prism
  graph <-
    add_prism(
      graph = graph,
      n = 3,
      type = "a",
      rel = "z")

  # Expect 6 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 6)

  # Expect 9 edges to have been created
  expect_equal(
    count_edges(graph = graph), 9)

  # Expect node ID values from 1 to 6
  expect_identical(
    get_node_ids(graph), 1:6)

  # Expect label values from 1 to 6
  expect_identical(
    graph$nodes_df$label,
    as.character(1:6))

  # Expect type values to all be `a`
  expect_equal(
    unique(graph$nodes_df$type), "a")

  # Expect rel values to all be `z`
  expect_equal(
    unique(graph$edges_df$rel), "z")

  # Expect an error if n is <3
  expect_error(
    add_prism(
      graph = graph,
      n = 2,
      type = "a",
      rel = "z"))

  # Add another prism to the graph
  graph <-
    add_prism(
      graph = graph,
      n = 3,
      type = "b",
      rel = "y")

  # Expect that 12 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph), 12)

  # Expect 18 edges are now in the graph
  expect_equal(
    count_edges(graph = graph), 18)

  # Expect label values from 1 to 12
  expect_identical(
    graph$nodes_df$label,
    as.character(1:12))

  # Expect type values to be either `a` or `b`
  expect_identical(
    unique(graph$nodes_df$type),
    c("a", "b"))

  # Expect rel values to be either `a` or `b`
  expect_identical(
    unique(graph$edges_df$rel),
    c("z", "y"))

  # Create a graph with a prism that
  # has different types of node and edge
  # attributes included
  graph_2 <-
    create_graph() %>%
    add_prism(
      n = 3,
      label = c("one", "two",
                "three", "four",
                "five", "six"),
      type = c("a", "a",
               "a", "b",
               "b", "b"),
      rel = c("prism_a", "prism_a",
              "prism_a", "prism_b",
              "prism_b", "prism_b",
              "prism_c", "prism_c",
              "prism_c"))

  # Get the graph's node data frame
  ndf <- get_node_df(graph_2)

  # Get the graph's edge data frame
  edf <- get_edge_df(graph_2)

  # Expect that 6 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph_2), 6)

  # Expect 9 edges are now in the graph
  expect_equal(
    count_edges(graph = graph_2), 9)

  # Expect node ID values from 1 to 6
  expect_identical(
    ndf$id, 1:6)

  # Expect specific node `label` values
  expect_identical(
    ndf$label,
    c("one", "two",
      "three", "four",
      "five", "six"))

  # Expect specific node `type` values
  expect_identical(
    ndf$type,
    c("a", "a", "a",
      "b", "b", "b"))

  # Expect specific edge `rel` values
  expect_identical(
    edf$rel,
    c("prism_a", "prism_a",
      "prism_a", "prism_b",
      "prism_b", "prism_b",
      "prism_c", "prism_c",
      "prism_c"))
})

test_that("Adding a star is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Add a star
  graph <-
    add_star(
      graph = graph,
      n = 4,
      type = "a",
      rel = "z")

  # Expect 4 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 4)

  # Expect 3 edges to have been created
  expect_equal(
    count_edges(graph = graph), 3)

  # Expect node ID values from 1 to 4
  expect_identical(
    get_node_ids(graph), 1:4)

  # Expect label values from 1 to 4
  expect_identical(
    graph$nodes_df$label,
    as.character(1:4))

  # Expect type values to all be `a`
  expect_equal(
    unique(graph$nodes_df$type), "a")

  # Expect rel values to all be `z`
  expect_equal(
    unique(graph$edges_df$rel), "z")

  # Expect an error if n is <4
  expect_error(
    add_star(
      graph = graph,
      n = 3,
      type = "a",
      rel = "z"))

  # Add another star to the graph
  graph <-
    add_star(
      graph = graph,
      n = 4,
      type = "b",
      rel = "y")

  # Expect that 8 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph), 8)

  # Expect 6 edges are now in the graph
  expect_equal(
    count_edges(graph = graph), 6)

  # Expect node ID values from 1 to 8
  expect_identical(
    get_node_ids(graph), 1:8)

  # Expect label values from 1 to 8
  expect_identical(
    graph$nodes_df$label,
    as.character(1:8))

  # Expect type values to be either `a` or `b`
  expect_identical(
    unique(graph$nodes_df$type),
    c("a", "b"))

  # Expect rel values to be either `z` or `y`
  expect_identical(
    unique(graph$edges_df$rel),
    c("z", "y"))

  # Create a graph with a star that
  # has different types of node and edge
  # attributes included
  graph_2 <-
    create_graph() %>%
    add_star(
      n = 6,
      label = c("one", "two",
                "three", "four",
                "five", "six"),
      type = c("a", "a",
               "b", "b",
               "c", "c"),
      rel = c("star_a", "star_b",
              "star_c", "star_d",
              "star_e"))

  # Get the graph's node data frame
  ndf <- graph_2 %>% get_node_df()

  # Get the graph's edge data frame
  edf <- graph_2 %>% get_edge_df()

  # Expect that 6 nodes are now in the graph
  expect_equal(
    count_nodes(graph = graph_2), 6)

  # Expect 5 edges are now in the graph
  expect_equal(
    count_edges(graph = graph_2), 5)

  # Expect node ID values from 1 to 6
  expect_identical(
    ndf$id, 1:6)

  # Expect specific node `label` values
  expect_identical(
    ndf$label,
    c("one", "two",
      "three", "four",
      "five", "six"))

  # Expect specific node `type` values
  expect_identical(
    ndf$type,
    c("a", "a", "b",
      "b", "c", "c"))

  # Expect specific edge `rel` values
  expect_identical(
    edf$rel,
    c("star_a", "star_b",
      "star_c", "star_d",
      "star_e"))
})

test_that("Adding a 2D grid is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Add a 2D grid
  graph <-
    add_grid_2d(
      graph = graph,
      x = 3,
      y = 3,
      type = "a",
      rel = "z")

  # Expect 9 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 9)

  # Expect 12 edges to have been created
  expect_equal(
    count_edges(graph = graph), 12)

  # Expect node ID values from 1 to 9
  expect_identical(
    get_node_ids(graph = graph), 1:9)

  # Expect label values from 1 to 9
  expect_identical(
    graph$nodes_df$label,
    as.character(1:9))

  # Expect type values to all be `a`
  expect_equal(
    unique(graph$nodes_df$type), "a")

  # Expect rel values to all be `z`
  expect_equal(
    unique(graph$edges_df$rel), "z")

  # Expect an error if x is <2
  expect_error(
    add_grid_2d(
      graph = graph,
      x = 1,
      y = 3,
      type = "a",
      rel = "z"))

  # Expect an error if y is <2
  expect_error(
    add_grid_2d(
      graph = graph,
      x = 3,
      y = 1,
      type = "a",
      rel = "z"))

  # Add another 2D grid to the graph
  graph <-
    add_grid_2d(
      graph = graph,
      x = 2,
      y = 2,
      type = "b",
      rel = "y")

  # Expect that 13 nodes are now in the graph
  expect_equal(
    count_nodes(graph), 13)

  # Expect 16 edges are now in the graph
  expect_equal(
    count_edges(graph), 16)

  # Expect label values from 1 to 13
  expect_identical(
    graph$nodes_df$label,
    as.character(1:13))

  # Expect type values to be either `a` or `b`
  expect_identical(
    unique(graph$nodes_df$type),
    c("a", "b"))

  # Expect rel values to be either `z` or `y`
  expect_identical(
    unique(graph$edges_df$rel),
    c("z", "y"))

  # Create a graph with a 2D grid that
  # has different types of node and edge
  # attributes included
  graph_2 <-
    create_graph() %>%
    add_grid_2d(
      x = 2,
      y = 2,
      label = c(
        "one", "two",
        "three", "four"),
      type = c(
        "a", "a",
        "b", "b"),
      rel = c(
        "grid_a", "grid_a",
        "grid_b", "grid_b"))

  # Get the graph's node data frame
  ndf <- get_node_df(graph_2)

  # Get the graph's edge data frame
  edf <- get_edge_df(graph_2)

  # Expect that 4 nodes are now in the graph
  expect_equal(
    count_nodes(graph_2), 4)

  # Expect 4 edges are now in the graph
  expect_equal(
    count_edges(graph_2), 4)

  # Expect node ID values from 1 to 6
  expect_identical(
    ndf$id, 1:4)

  # Expect specific node `label` values
  expect_identical(
    ndf$label,
    c("one", "two", "three", "four"))

  # Expect specific node `type` values
  expect_identical(
    ndf$type,
    c("a", "a", "b", "b"))

  # Expect specific edge `rel` values
  expect_identical(
    edf$rel,
    c("grid_a", "grid_a", "grid_b", "grid_b"))
})

test_that("Adding a 3D grid is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Add a 3D grid
  graph <-
    add_grid_3d(
      graph = graph,
      x = 2,
      y = 2,
      z = 2,
      type = "a",
      rel = "z")

  # Expect 8 nodes to have been created
  expect_equal(
    count_nodes(graph), 8)

  # Expect 12 edges to have been created
  expect_equal(
    count_edges(graph), 12)

  # Expect node ID values from 1 to 8
  expect_identical(
    get_node_ids(graph), 1:8)

  # Expect label values from 1 to 8
  expect_identical(
    graph$nodes_df$label,
    as.character(1:8))

  # Expect type values to all be `a`
  expect_equal(
    unique(graph$nodes_df$type), "a")

  # Expect rel values to all be `z`
  expect_equal(
    unique(graph$edges_df$rel), "z")

  # Expect an error if x is <2
  expect_error(
    add_grid_3d(
      graph = graph,
      x = 1,
      y = 3,
      z = 3,
      type = "a",
      rel = "z"))

  # Expect an error if y is <2
  expect_error(
    add_grid_3d(
      graph = graph,
      x = 3,
      y = 1,
      z = 3,
      type = "a",
      rel = "z"))

  # Expect an error if z is <2
  expect_error(
    add_grid_3d(
      graph = graph,
      x = 3,
      y = 3,
      z = 1,
      type = "a",
      rel = "z"))

  # Add another 3D grid to the graph
  graph <-
    add_grid_3d(
      graph = graph,
      x = 2,
      y = 2,
      z = 2,
      type = "b",
      rel = "y")

  # Expect that 13 nodes are now in the graph
  expect_equal(
    count_nodes(graph), 16)

  # Expect 24 edges are now in the graph
  expect_equal(
    count_edges(graph), 24)

  # Expect label values from 1 to 16
  expect_identical(
    graph$nodes_df$label,
    as.character(1:16))

  # Expect type values to be either `a` or `b`
  expect_identical(
    unique(graph$nodes_df$type),
    c("a", "b"))

  # Expect rel values to be either `z` or `y`
  expect_identical(
    unique(graph$edges_df$rel),
    c("z", "y"))

  # Create a graph with a 2D grid that
  # has different types of node and edge
  # attributes included
  graph_2 <-
    create_graph() %>%
    add_grid_3d(
      x = 2,
      y = 2,
      z = 2,
      label = c(
        "one", "two",
        "three", "four",
        "five", "six",
        "seven", "eight"),
      type = c(
        "a", "a", "a", "a",
        "b", "b", "b", "b"),
      rel = c(
        "grid_a", "grid_a", "grid_a",
        "grid_a", "grid_a", "grid_a",
        "grid_b", "grid_b", "grid_b",
        "grid_b", "grid_b", "grid_b"))

  # Get the graph's node data frame
  ndf <- get_node_df(graph_2)

  # Get the graph's edge data frame
  edf <- get_edge_df(graph_2)

  # Expect that 8 nodes are now in the graph
  expect_equal(
    count_nodes(graph_2), 8)

  # Expect 12 edges are now in the graph
  expect_equal(
    count_edges(graph_2), 12)

  # Expect node ID values from 1 to 6
  expect_identical(
    ndf$id, 1:8)

  # Expect specific node `label` values
  expect_identical(
    ndf$label,
    c("one", "two", "three", "four",
      "five", "six", "seven", "eight"))

  # Expect specific node `type` values
  expect_identical(
    ndf$type,
    c("a", "a", "a", "a",
      "b", "b", "b", "b"))

  # Expect specific edge `rel` values
  expect_identical(
    edf$rel,
    c("grid_a", "grid_a", "grid_a",
      "grid_a", "grid_a", "grid_a",
      "grid_b", "grid_b", "grid_b",
      "grid_b", "grid_b", "grid_b"))
})

test_that("Adding a full graph is possible", {

  # Create a graph and add a full graph
  # with 3 nodes to it; keep loops
  graph <-
    create_graph() %>%
    add_full_graph(
      n = 3,
      keep_loops = TRUE)

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 9 edges to have been created
  expect_equal(
    count_edges(graph = graph), 9)

  # Expect that there are 3 loops in the graph
  expect_length(
    which(graph$edges_df$from == graph$edges_df$to),
    3)

  # Expect node ID values from 1 to 3
  expect_identical(
    get_node_ids(graph), 1:3)

  # Expect type values to be character NA values
  expect_equal(
    unique(graph$nodes_df$type),
    NA_character_)

  # Create a graph and add a full graph
  # with 3 nodes to it; discard loops
  graph <-
    create_graph() %>%
    add_full_graph(
      n = 3,
      keep_loops = FALSE)

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 6 edges to have been created
  expect_equal(
    count_edges(graph = graph), 6)

  # Expect that there are no loops in the graph
  expect_length(
    which(graph$edges_df$from == graph$edges_df$to),
    0)

  # Create a graph, add a full graph
  # (with no loops), set values for the
  # node `label`, node `type`, and edge `rel`
  graph <-
    create_graph() %>%
    add_full_graph(
      n = 3,
      type = "connected",
      label = c("1st", "2nd", "3rd"),
      rel = "connected_to")

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 6 edges to have been created
  expect_equal(
    count_edges(graph = graph), 6)

  # Expect that there are no loops in the graph
  expect_length(
    which(graph$edges_df$from == graph$edges_df$to),
    0)

  # Expect that the `type` node attr
  # has the value `connected` for all
  # nodes created
  expect_equal(
    unique(get_node_df(graph)[, 2]),
    "connected")

  # Expect that the `label` values
  # assigned to the nodes are in the
  # graph's internal node data frame
  expect_equal(
    graph$nodes_df$label,
    c("1st", "2nd", "3rd"))

  # Expect that the `label` edge attr
  # has the value `connected_to` for
  # all edges created
  expect_equal(
    unique(graph$edges_df$rel),
    "connected_to")

  # Create a fully-connected and directed
  # graph with 3 nodes, and, where a matrix
  # provides edge weights; first, create a
  # matrix to be used for edge weights
  # matrix (with row names to be used as
  # node labels)
  suppressWarnings(RNGversion("3.5.0"))
  set.seed(23)

  edge_wt_matrix_rownames <-
    rnorm(n = 100, mean = 5, sd = 2) %>%
    sample(size = 9, replace = FALSE) %>%
    round(digits = 2) %>%
    matrix(
      ncol = 3,
      nrow = 3,
      dimnames = list(c("a", "b", "c")))

  # Create a fully-connected graph without
  # loops and use the matrix to provide
  # values for the edge `weight` attribute
  graph <-
    create_graph() %>%
    add_full_graph(
      n = 3,
      type = "weighted",
      label = TRUE,
      rel = "related_to",
      edge_wt_matrix = edge_wt_matrix_rownames,
      keep_loops = FALSE)

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 6 edges to have been created
  expect_equal(
    count_edges(graph = graph), 6)

  # Expect that there are no loops in the graph
  expect_length(
    which(graph$edges_df$from == graph$edges_df$to),
    0)

  # Expect that the `type` node attr
  # has the value `connected` for all
  # nodes created
  expect_equal(
    unique(get_node_df(graph)[, 2]),
    "weighted")

  # Expect that the `label` values
  # assigned to the nodes are the same
  # as the matrix rownames
  expect_equal(
    graph$nodes_df$label,
    rownames(edge_wt_matrix_rownames))

  # Expect that the `rel` edge attr
  # has the value `connected_to` for
  # all edges created
  expect_equal(
    unique(graph$edges_df$rel),
    "related_to")

  # Expect certain values for the
  # edge weight attribute
  expect_equal(
    graph$edges_df$weight,
    c(3.3, 5.02, 4.13,
      6.49, 6.03, 5.55))

  # Create a fully-connected but undirected
  # graph without loops and use the matrix
  # to provide values for the edge `weight`
  # attribute; in this case, the lower
  # triangle of the matrix will be used
  graph <-
    create_graph(directed = FALSE) %>%
    add_full_graph(
      n = 3,
      type = "weighted",
      label = TRUE,
      rel = "related_to",
      edge_wt_matrix = edge_wt_matrix_rownames,
      keep_loops = FALSE)

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 3 edges to have been created
  expect_equal(
    count_edges(graph = graph), 3)

  # Expect that there are no loops in the graph
  expect_length(
    which(graph$edges_df$from ==  graph$edges_df$to),
    0)

  # Expect that the `type` node attr
  # has the value `connected` for all
  # nodes created
  expect_equal(
    unique(graph$nodes_df$type),
    "weighted")

  # Expect that the `label` values
  # assigned to the nodes are the same
  # as the matrix rownames
  expect_equal(
    graph$nodes_df$label,
    rownames(edge_wt_matrix_rownames))

  # Expect that the `rel` edge attr
  # has the value `connected_to` for
  # all edges created
  expect_equal(
    unique(graph$edges_df$rel),
    "related_to")

  # Expect certain values for the
  # edge weight attribute
  expect_equal(
    graph$edges_df$weight,
    c(3.3, 5.02, 6.49))

  # Create a fully-connected graph with
  # loop preserved and use the matrix to
  # provide values for the edge `weight`
  # attribute
  graph <-
    create_graph() %>%
    add_full_graph(
      n = 3,
      type = "weighted",
      label = TRUE,
      rel = "related_to",
      edge_wt_matrix = edge_wt_matrix_rownames,
      keep_loops = TRUE)

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 9 edges to have been created
  expect_equal(
    count_edges(graph = graph), 9)

  # Expect that there are 3 loops in the graph
  expect_length(
    which(graph$edges_df$from == graph$edges_df$to),
    3)

  # Expect that the `type` node attr
  # has the value `connected` for all
  # nodes created
  expect_equal(
    unique(graph$nodes_df$type),
    "weighted")

  # Expect that the `label` values
  # assigned to the nodes are the same
  # as the matrix rownames
  expect_equal(
    graph$nodes_df$label,
    rownames(edge_wt_matrix_rownames))

  # Expect that the `rel` edge attr
  # has the value `connected_to` for
  # all edges created
  expect_equal(
    unique(graph$edges_df$rel),
    "related_to")

  # Expect certain values for the
  # edge weight attribute
  expect_equal(
    graph$edges_df$weight,
    c(8.66, 3.3, 5.02, 4.13, 6.83,
      6.49, 6.03, 5.55, 3.8))

  # Create a fully-connected and undirected
  # graph with loops preserved; use the
  # matrix to provide values for the edge
  # `weight` attribute
  graph <-
    create_graph(directed = FALSE) %>%
    add_full_graph(
      n = 3,
      type = "weighted",
      label = TRUE,
      rel = "related_to",
      edge_wt_matrix = edge_wt_matrix_rownames,
      keep_loops = TRUE)

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 6 edges to have been created
  expect_equal(
    count_edges(graph = graph), 6)

  # Expect that there are 3 loops in the graph
  expect_length(
    which(graph$edges_df$from == graph$edges_df$to),
    3)

  # Expect that the `type` node attr
  # has the value `connected` for all
  # nodes created
  expect_equal(
    unique(graph$nodes_df$type),
    "weighted")

  # Expect that the `label` values
  # assigned to the nodes are the same
  # as the matrix rownames
  expect_equal(
    graph$nodes_df$label,
    rownames(edge_wt_matrix_rownames))

  # Expect that the `rel` edge attr
  # has the value `connected_to` for
  # all edges created
  expect_equal(
    unique(graph$edges_df$rel),
    "related_to")

  # Expect certain values for the
  # edge weight attribute
  expect_equal(
    graph$edges_df$weight,
    c(8.66, 3.3, 5.02, 6.83, 6.49, 3.8))

  # Create a fully-connected and undirected
  # graph with loops preserved; use the
  # matrix to provide values for the edge
  # `weight` attribute, however, do not add
  # labels to this graph
  graph <-
    create_graph(directed = FALSE) %>%
    add_full_graph(
      n = 3,
      type = "weighted",
      label = NULL,
      rel = "related_to",
      edge_wt_matrix = edge_wt_matrix_rownames,
      keep_loops = TRUE)

  # Expect 3 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 3)

  # Expect 6 edges to have been created
  expect_equal(
    count_edges(graph = graph), 6)

  # Expect that there are 3 loops in the graph
  expect_length(
    which(graph$edges_df$from == graph$edges_df$to),
    3)

  # Expect that the `type` node attr
  # has the value `connected` for all
  # nodes created
  expect_equal(
    unique(graph$nodes_df$type),
    "weighted")

  # Expect that the `label` values are
  # character NA values
  expect_equal(
    graph$nodes_df$label,
    rep(NA_character_, 3))

  # Expect that the `rel` edge attr
  # has the value `connected_to` for
  # all edges created
  expect_equal(
    unique(graph$edges_df$rel),
    "related_to")

  # Expect certain values for the
  # edge weight attribute
  expect_equal(
    graph$edges_df$weight,
    c(8.66, 3.3, 5.02, 6.83, 6.49, 3.8))

  # Create a graph with a cycle then
  # and add a full graph
  # with 3 nodes to it; discard loops
  graph <-
    create_graph() %>%
    add_cycle(
      n = 5,
      type = "cycle") %>%
    add_full_graph(
      n = 8,
      keep_loops = FALSE)

  # Expect 13 nodes to have been created
  expect_equal(
    count_nodes(graph = graph), 13)

  # Expect 6 edges to have been created
  expect_equal(
    count_edges(graph = graph), 61)
})

test_that("Adding a G(n, m) Erdos-Renyi graph is possible", {

  # Create an undirected GNM
  # graph with 100 nodes and
  # 120 edges
  gnm_graph <-
    create_graph(
      directed = FALSE) %>%
    add_gnm_graph(
      n = 100,
      m = 120)

  # Expect 100 nodes in the graph
  expect_equal(
    gnm_graph %>%
      count_nodes(),
    100)

  # Expect 120 edges in the graph
  expect_equal(
    gnm_graph %>%
      count_edges(),
    120)

  # Create 2 graphs made with the same
  # `n`, `m`, and `set_seed` values
  graph_1 <-
    create_graph() %>%
    add_gnm_graph(
      n = 100,
      m = 120,
      set_seed = 23)

  graph_2 <-
    create_graph() %>%
    add_gnm_graph(
      n = 100,
      m = 120,
      set_seed = 23)

  # Expect that these graphs will
  # have the same node and edge definitions
  expect_identical(
    graph_1$nodes_df,
    graph_2$nodes_df)

  expect_identical(
    graph_1$edges_df,
    graph_2$edges_df)

  # Create a directed graph, add
  # a cycle of 100 nodes, and then
  # add a GNM graph with 100 nodes
  # and 120 edges
  gnm_graph_added <-
    create_graph() %>%
    add_cycle(
      n = 100,
      type = "cycle") %>%
    add_gnm_graph(
      n = 100,
      m = 120) %>%
    select_last_nodes_created() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "GNM") %>%
    clear_selection()

  # Expect that the first 100 nodes
  # belong to the `cycle` type
  expect_equal(
    gnm_graph_added$nodes_df$type[1:100] %>%
      unique(),
    "cycle")

  # Expect that the 200 nodes have a
  # node ID sequence from 1 to 200
  expect_identical(
    gnm_graph_added$nodes_df$id,
    1:200)

  # Expect an error if the value for
  # `n` is too small (< 1)
  expect_error(
    create_graph() %>%
    add_gnm_graph(n = 0, m = 5))
})

test_that("Adding a G(n, p) Erdos-Renyi graph is possible", {

  # Create an undirected GNM
  # graph with 100 nodes and
  # a probability of 0.05 for
  # creating edges
  gnp_graph <-
    create_graph(
      directed = FALSE) %>%
    add_gnp_graph(
      n = 100,
      p = 0.05,
      set_seed = 23)

  # Expect 100 nodes in the graph
  expect_equal(
    gnp_graph %>%
      count_nodes(),
    100)

  # Expect 233 edges in the graph
  expect_equal(
    gnp_graph %>%
      count_edges(),
    233)

  # Create 2 graphs made with the same
  # `n`, `p`, and `set_seed` values
  graph_1 <-
    create_graph() %>%
    add_gnp_graph(
      n = 100,
      p = 0.06,
      set_seed = 23)

  graph_2 <-
    create_graph() %>%
    add_gnp_graph(
      n = 100,
      p = 0.06,
      set_seed = 23)

  # Expect that these graphs will
  # have the same node and edge definitions
  expect_identical(
    graph_1$nodes_df,
    graph_2$nodes_df)

  expect_identical(
    graph_1$edges_df,
    graph_2$edges_df)

  # Create a directed graph, add
  # a cycle of 100 nodes, and then
  # add a GNP graph with 100 nodes
  gnp_graph_added <-
    create_graph() %>%
    add_cycle(
      n = 100,
      type = "cycle") %>%
    add_gnp_graph(
      n = 100,
      p = 0.06) %>%
    select_last_nodes_created() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "GNP") %>%
    clear_selection()

  # Expect that the first 100 nodes
  # belong to the `cycle` type
  expect_equal(
    gnp_graph_added$nodes_df$type[1:100] %>%
      unique(),
    "cycle")

  # Expect that the 200 nodes have a
  # node ID sequence from 1 to 200
  expect_identical(
    gnp_graph_added$nodes_df$id,
    1:200)

  # Expect an error if the value for
  # `n` is too small (< 1)
  expect_error(
    create_graph() %>%
      add_gnp_graph(n = 0, p = 0.05))
})

test_that("Adding a growing graph is possible", {

  # Create a random, growing
  # citation graph with 100
  # nodes, adding an edge after
  # each node addition
  growing_graph <-
    create_graph() %>%
    add_growing_graph(
      n = 100,
      m = 1,
      citation = TRUE,
      set_seed = 23)

  # Expect 100 nodes in the graph
  expect_equal(
    growing_graph %>%
      count_nodes(),
    100)

  # Expect 99 edges in the graph
  expect_equal(
    growing_graph %>%
      count_edges(),
    99)

  # Create 2 graphs made with the same
  # `n`, `m`, and `set_seed` values
  graph_1 <-
    create_graph() %>%
    add_growing_graph(
      n = 100,
      m = 1,
      citation = TRUE,
      set_seed = 23)

  graph_2 <-
    create_graph() %>%
    add_growing_graph(
      n = 100,
      m = 1,
      citation = TRUE,
      set_seed = 23)

  # Expect that these graphs will
  # have the same node and edge definitions
  expect_identical(
    graph_1$nodes_df,
    graph_2$nodes_df)

  expect_identical(
    graph_1$edges_df,
    graph_2$edges_df)

  # Create a directed graph, add
  # a cycle of 100 nodes, and then
  # add a growing graph with 100 nodes
  growing_graph_added <-
    create_graph() %>%
    add_cycle(
      n = 100,
      type = "cycle") %>%
    add_growing_graph(
      n = 100,
      m = 1) %>%
    select_last_nodes_created() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "growing") %>%
    clear_selection()

  # Expect that the first 100 nodes
  # belong to the `cycle` type
  expect_equal(
    growing_graph_added$nodes_df$type[1:100] %>%
      unique(),
    "cycle")

  # Expect that the 200 nodes have a
  # node ID sequence from 1 to 200
  expect_identical(
    growing_graph_added$nodes_df$id,
    1:200)

  # Expect an error if the value for
  # `n` is too small (< 1)
  expect_error(
    create_graph() %>%
      add_growing_graph(n = 0, m = 1))
})

test_that("Adding an islands graph is possible", {

  # Create an undirected islands
  # graph with standard parameters
  islands_graph <-
    create_graph(
      directed = FALSE) %>%
    add_islands_graph(
      n_islands = 4,
      island_size = 10,
      p = 0.5,
      edges_between = 1,
      set_seed = 23)

  # Expect 40 nodes in the graph
  expect_equal(
    islands_graph %>%
      count_nodes(),
    40)

  # Expect more than 90 edges in the graph
  # with igraph < 2.0.0 there are 98, igraph 2.0.0 107
  expect_gt(
    count_edges(islands_graph),
    90)

  # Create 2 graphs made with the same
  # parameters and `set_seed` values
  graph_1 <-
    create_graph() %>%
    add_islands_graph(
      n_islands = 5,
      island_size = 10,
      p = 0.5,
      edges_between = 1,
      set_seed = 23)

  graph_2 <-
    create_graph() %>%
    add_islands_graph(
      n_islands = 5,
      island_size = 10,
      p = 0.5,
      edges_between = 1,
      set_seed = 23)

  # Expect that these graphs will
  # have the same node and edge definitions
  expect_identical(
    graph_1$nodes_df,
    graph_2$nodes_df)

  expect_identical(
    graph_1$edges_df,
    graph_2$edges_df)

  # Create a directed graph, add
  # a cycle of 100 nodes, and then
  # add an islands graph with 100 nodes
  islands_graph_added <-
    create_graph() %>%
    add_cycle(
      n = 100,
      type = "cycle") %>%
    add_islands_graph(
      n_islands = 10,
      island_size = 10,
      p = 0.5,
      edges_between = 1) %>%
    select_last_nodes_created() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "islands") %>%
    clear_selection()

  # Expect that the first 100 nodes
  # belong to the `cycle` type
  expect_equal(
    islands_graph_added$nodes_df$type[1:100] %>%
      unique(),
    "cycle")

  # Expect that the 200 nodes have a
  # node ID sequence from 1 to 200
  expect_identical(
    islands_graph_added$nodes_df$id,
    1:200)
})

test_that("Adding a preferential attachment graph is possible", {

  # Create an undirected PA
  # graph with 100 nodes, creating
  # an edge for every node created
  pa_graph <-
    create_graph(
      directed = FALSE) %>%
    add_pa_graph(
      n = 100,
      m = 1)

  # Expect 100 nodes in the graph
  expect_equal(
    pa_graph %>%
      count_nodes(),
    100)

  # Expect 99 edges in the graph
  expect_equal(
    pa_graph %>%
      count_edges(),
    99)

  # Create 2 graphs made with the same
  # `n`, `m`, and `set_seed` values
  graph_1 <-
    create_graph() %>%
    add_pa_graph(
      n = 200,
      m = 1,
      set_seed = 23)

  graph_2 <-
    create_graph() %>%
    add_pa_graph(
      n = 200,
      p = 1,
      set_seed = 23)

  # Expect that these graphs will
  # have the same node and edge definitions
  expect_identical(
    graph_1$nodes_df,
    graph_2$nodes_df)

  expect_identical(
    graph_1$edges_df,
    graph_2$edges_df)

  # Create a directed graph, add
  # a cycle of 100 nodes, and then
  # add a PA graph with 100 nodes
  pa_graph_added <-
    create_graph() %>%
    add_cycle(
      n = 100,
      type = "cycle") %>%
    add_pa_graph(
      n = 100,
      m = 1) %>%
    select_last_nodes_created() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "PA") %>%
    clear_selection()

  # Expect that the first 100 nodes
  # belong to the `cycle` type
  expect_equal(
    pa_graph_added$nodes_df$type[1:100] %>%
      unique(),
    "cycle")

  # Expect that the 200 nodes have a
  # node ID sequence from 1 to 200
  expect_identical(
    pa_graph_added$nodes_df$id,
    1:200)


  expect_snapshot(error = TRUE, {
    # Expect an error if the value for
    # `n` is too small (< 1)
    create_graph() %>%
      add_pa_graph(n = 0, m = 1)

    # Expect an error if the value for
    # `algo` is not a valid value
    create_graph() %>%
      add_pa_graph(n = 0, m = 1, algo = "plumtree")
    })

})

test_that("Adding a small world graph is possible", {

  # Create an undirected small world
  # graph with standard parameters
  smallworld_graph <-
    create_graph(
      directed = FALSE) %>%
    add_smallworld_graph(
      dimension = 1,
      size = 50,
      neighborhood = 1,
      p = 0.05,
      set_seed = 23)

  # Expect 50 nodes in the graph
  expect_equal(
    smallworld_graph %>%
      count_nodes(),
    50)

  # Expect 50 edges in the graph
  expect_equal(
    smallworld_graph %>%
      count_edges(),
    50)

  # Create 2 graphs made with the same
  # parameters and `set_seed` values
  graph_1 <-
    create_graph() %>%
    add_smallworld_graph(
      dimension = 1,
      size = 50,
      neighborhood = 1,
      p = 0.05,
      set_seed = 23)

  graph_2 <-
    create_graph() %>%
    add_smallworld_graph(
      dimension = 1,
      size = 50,
      neighborhood = 1,
      p = 0.05,
      set_seed = 23)

  # Expect that these graphs will
  # have the same node and edge definitions
  expect_identical(
    graph_1$nodes_df,
    graph_2$nodes_df)

  expect_identical(
    graph_1$edges_df,
    graph_2$edges_df)

  # Create a directed graph, add
  # a cycle of 100 nodes, and then
  # add a small world graph
  smallworld_graph_added <-
    create_graph() %>%
    add_cycle(
      n = 100,
      type = "cycle") %>%
    add_smallworld_graph(
      dimension = 1,
      size = 50,
      neighborhood = 1,
      p = 0.05) %>%
    select_last_nodes_created() %>%
    set_node_attrs_ws(
      node_attr = type,
      value = "smallworld") %>%
    clear_selection()

  # Expect that the first 100 nodes
  # belong to the `cycle` type
  expect_equal(
    smallworld_graph_added$nodes_df$type[1:100] %>%
      unique(),
    "cycle")

  # Expect that the 150 nodes have a
  # node ID sequence from 1 to 150
  expect_identical(
    smallworld_graph_added$nodes_df$id,
    1:150)
})

# Getting node IDs from the entire graph or within edges")

test_that("getting node IDs from various objects is possible", {

  suppressWarnings(RNGversion("3.5.0"))
  set.seed(23)

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 26,
      label = TRUE,
      type = c(
        rep("a_to_g", 7),
        rep("h_to_p", 9),
        rep("q_to_x", 8),
        rep("y_and_z", 2)))

  # Create an edge data frame
  edges <-
    create_edge_df(
      from = sample(1:26, replace = TRUE),
      to = sample(1:26, replace = TRUE),
      label = "edge",
      rel = "letter_to_letter")

  # Create the graph object using
  # the node and edge data frames
  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Get the graph's node ID values
  gotten_nodes <- get_node_ids(graph)

  # Expect an `integer` vector object
  expect_type(
    gotten_nodes, "integer")

  # Expect that the integer vector object
  # has no names
  expect_named(gotten_nodes, NULL)

  # Expect a vector that is sequence
  # from `1` to `26`
  expect_equal(
    gotten_nodes,
    1:26
  )

  # Expect that `get_node_ids()` when
  # used on an empty graph returns NA
  expect_true(
    create_graph() %>%
      get_node_ids() %>%
      is.na())

  # Expect that `get_node_ids()` when
  # used on graph where the conditions
  # supplied yields no nodes returns NA
  expect_true(
    create_graph() %>%
      add_path(n = 2) %>%
      get_node_ids(
        conditions = id > 2) %>%
      is.na())

  # Expect that the number of nodes
  # obtained from the entire graph will
  # be greater than the nodes associated
  # with edges (since there will be free
  # nodes with no edges)
  expect_gt(
    length(get_node_ids(graph)),
    length(unique(c(graph$edges_df$from,
                    graph$edges_df$to))))

  # Get the node df from the graph using `get_node_df()`
  node_df_from_graph <- get_node_df(graph)

  # Expect that using `get_node_df()` on a graph with
  # no nodes will return an empty data frame
  expect_equal(
    nrow(get_node_df(create_graph())), 0)

  # Get the graph's node ID values with
  # a condition attached
  gotten_nodes_w_condition <-
    graph %>%
    get_node_ids(
      conditions =
        type == "h_to_p")

  # Expect an `integer` vector object
  expect_type(
    gotten_nodes_w_condition, "integer")

  # Expect that the integer vector object
  # has no names
  expect_named(gotten_nodes_w_condition, NULL)

  # Expect a vector that is sequence from `8` to `16`
  expect_equal(
    gotten_nodes_w_condition,
    8:16
  )
})

test_that("getting node IDs associated within a graph's edges is possible", {

  # Set a seed to make results reproducible
  suppressWarnings(RNGversion("3.5.0"))
  set.seed(23)

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 26,
      label = TRUE,
      type = c(rep("a_to_g", 7),
               rep("h_to_p", 9),
               rep("q_to_x", 8),
               rep("y_and_z", 2)))

  # Create an edge data frame
  edges <-
    create_edge_df(
      from = sample(1:26, replace = TRUE),
      to = sample(1:26, replace = TRUE),
      rel = "letter_to_letter",
      label = "edge")

  # Create the graph object using
  # the node and edge data frames
  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Get the `outgoing` and `incoming`
  # node ID values in a list object
  gotten_edges_list <-
    get_edges(graph, return_type = "list")

  # Expect a list object
  expect_type(
    gotten_edges_list, "list")

  # Expect that the list is of length 2
  expect_length(gotten_edges_list, 2)

  # Expect integer vectors of length 26
  # in `gotten_edges_list`
  expect_length(gotten_edges_list[[1]], 26)

  expect_type(
    gotten_edges_list[[1]], "integer")

  expect_length(
    gotten_edges_list[[2]], 26)

  expect_type(
    gotten_edges_list[[2]], "integer")

  # Get the `outgoing` and `incoming` node ID values
  # in a data frame object
  gotten_edges_df <-
    get_edges(
      graph = graph,
      return_type = "df")

  # Expect a data frame object
  expect_s3_class(
    gotten_edges_df, "data.frame")

  # Expect that the data frame has 2 columns
  expect_equal(
    ncol(gotten_edges_df), 2)

  # Expect columns of class `integer` and that there
  # are 26 rows in `gotten_edges_df`
  expect_type(
    gotten_edges_df[, 1], "integer")

  expect_type(
    gotten_edges_df[, 2], "integer")

  expect_equal(
    nrow(gotten_edges_df), 26)

  # Get the `outgoing` and `incoming` node ID values
  # in a vector object
  gotten_edges_vector <-
    get_edges(
      graph = graph,
      return_type = "vector")

  # Expect a vector object of class `character` of length 26
  expect_type(gotten_edges_vector, "character")
  expect_length(gotten_edges_vector, 26)

  # Expect that the '->' substring is in
  # each vector component
  expect_match(
    gotten_edges_vector,
    "->"
  )

  # Get the edge df from the graph using `get_edge_df()`
  edge_df_from_graph <- get_edge_df(graph)

  # Expect that using `get_edge_df()` on a graph
  # with no edges will return an empty data frame

  edge_df <- get_edge_df(create_graph(nodes_df = create_node_df(1)))
  expect_equal(nrow(edge_df), 0)
})

test_that("getting edge information from a graph with no edges is possible ", {

  # Create a node data frame with 2 nodes
  ndf <- create_node_df(n = 2)

  # Create a graph with 2 nodes but no edges
  graph_no_edges <-
    create_graph(nodes_df = ndf)

  # Get edges from an edgeless graph returned as a vector
  edges_vector_from_graph_no_edges <-
    get_edges(
      graph = graph_no_edges,
      return_type = "vector")

  # Expect a vector object of class `logical`
  expect_type(
    edges_vector_from_graph_no_edges, "logical")

  # Expect that an NA is returned
  expect_true(
    is.na(edges_vector_from_graph_no_edges))

  # Get edges from an edgeless graph returned as a list
  edges_list_from_graph_no_edges <-
    get_edges(
      graph = graph_no_edges,
      return_type = "list")

  # Expect that an NA is returned
  expect_true(
    is.na(edges_list_from_graph_no_edges))

  # Get edges from an edgeless graph returned as a data frame
  edges_df_from_graph_no_edges <-
    get_edges(
      graph = graph_no_edges,
      return_type = "df")

  # Expect that an NA is returned
  expect_true(
    is.na(edges_df_from_graph_no_edges))
})

test_that("getting an ndf based on a selection of nodes is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_path(n = 5)

  # Select 2 nodes from the graph
  # and view a subset of the
  # node data frame that only contains
  # that selection of nodes
  ndf_subset <-
    graph %>%
    select_nodes_by_id(
      nodes = c(1, 2)) %>%
    get_node_df_ws()

  # Expect the same column names
  # and order of column names compared
  # to output generated by `get_node_df()`
  expect_identical(
    ndf_subset %>% colnames(),
    graph %>% get_node_df() %>% colnames())

  # Expect that there are 2 nodes
  # in `ndf_subset`
  expect_equal(
    nrow(ndf_subset), 2)

  # Expect that just nodes `1` and
  # `2` are in `ndf_subset`
  expect_equal(
    ndf_subset$id, c(1, 2))

  # Expect an error if there is no
  # valid node selection when calling
  # `get_node_df_ws()`
  expect_error(
    graph %>% get_node_df_ws())
})

test_that("getting an edf based on a selection of edges is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_path(n = 5)

  # Select 2 edges from the graph
  # and view a subset of the
  # edge data frame that only contains
  # that selection of edges
  edf_subset <-
    graph %>%
    select_edges_by_edge_id(
      edges = c(1, 2)) %>%
    get_edge_df_ws()

  # Expect the same column names
  # and order of column names compared
  # to output generated by `get_edge_df()`
  expect_identical(
    edf_subset %>% colnames(),
    graph %>% get_edge_df() %>% colnames())

  # Expect that there are 2 edges
  # in `edf_subset`
  expect_equal(
    nrow(edf_subset), 2)

  # Expect that just edges `1` and
  # `2` are in `edf_subset`
  expect_equal(
    edf_subset$id, c(1, 2))

  # Expect an error if there is no
  # valid edge selection when calling
  # `get_edge_df_ws()`
  expect_error(
    graph %>% get_edge_df_ws())
})

test_that("getting connected nodes is possible", {

  # Create a randomized graph with
  # many weakly connected components
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 10,
      set_seed = 23)

  connect_node_1 <-
    get_all_connected_nodes(
      graph = graph,
      node = 1)

  # Expect certain node IDs to be returned
  expect_equal(
    connect_node_1,
    c(2, 3, 4, 5, 6, 8, 9, 10))

  # Expect that the node ID provided won't be
  # returned in the set of node ID values
  expect_false(
    1 %in% connect_node_1)

  # Expect an NA value if there are no connected
  # nodes to the provided node
  expect_true(
    is.na(get_all_connected_nodes(graph, 7)))

  # Expect an error if providing a node ID that
  # doesn't exist in the graph
  expect_error(
    get_all_connected_nodes(graph, 35))
})

test_that("getting edge IDs from graph objects is possible", {

  # Create a node data frame (ndf)
  ndf <-
    create_node_df(
      n = 4,
      type = "letter",
      color = c("red", "green", "grey", "blue"),
      value = c(3.5, 2.6, 9.4, 2.7))

  # Create an edge data frame (edf)
  edf <-
    create_edge_df(
      from = c(1, 2, 3),
      to = c(4, 3, 1),
      rel = "leading_to",
      color = c("pink", "blue", "blue"),
      value = c(3.9, 2.5, 7.3))

  # Create a graph
  graph <-
    create_graph(
      nodes_df = ndf,
      edges_df = edf)

  # Expect a vector of all edge IDs in the graph
  expect_equal(
    get_edge_ids(graph), c(1, 2, 3))

  # Using a numeric comparison (i.e., all edges
  # with `value` attribute greater than 3),
  # expect edges `1` and `3`
  expect_equal(
    get_edge_ids(
      graph,
      conditions = value > 3), c(1, 3))

  # Using an equality for a character object,
  # (i.e., all nodes with `color` attribute of
  # `pink`), expect edge `1`
  expect_equal(
    get_edge_ids(
      graph = graph,
      conditions = color == "pink"),
    1)

  # Expect that multiple conditions will work
  # to return edges with the desired attribute
  # values (in this case, edge `3`)
  expect_equal(
    get_edge_ids(
      graph,
      conditions =
        color == "blue" &
        value > 5),
    3)

  # Expect NA if no edges are matched after
  # providing unmatched conditions
  expect_true(
    is.na(
      get_edge_ids(
        graph,
        conditions = color == "red")))

  # Expect NA if there are no edges in the graph
  expect_true(
    is.na(
      get_edge_ids(
        create_graph())))
})

#' Delete an edge from an existing graph object
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, delete an existing edge by
#' specifying either: (1) a pair of node IDs corresponding to the edge (keeping
#' into consideration the direction of the edge in a directed graph), or (2) an
#' edge ID.
#'
#' @inheritParams render_graph
#' @param from a node ID from which the edge to be removed is outgoing. If an
#'   edge ID is provided to `id`, then this argument is ignored. There is
#'   the option to use a node `label` value here (and this must
#'   correspondingly also be done for the `to` argument) for defining node
#'   connections. Note that this is only possible if all nodes have distinct
#'   `label` values set and none exist as an empty string.
#' @param to a node ID to which the edge to be removed is incoming. If an edge
#'   ID is provided to `id`, then this argument is ignored. There is the
#'   option to use a node `label` value here (and this must correspondingly
#'   also be for the `from` argument) for defining node connections. Note
#'   that this is only possible if all nodes have distinct `label` values
#'   set and none exist as an empty string.
#' @param id an edge ID of the edge to be removed.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 2 nodes
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(n = 2)
#'
#' # Add an edge
#' graph <-
#'   graph |>
#'   add_edge(
#'     from = 1,
#'     to = 2)
#'
#' # Delete the edge
#' graph <-
#'   graph |>
#'   delete_edge(
#'     from = 1,
#'     to = 2)
#'
#' # Get the count of edges in the graph
#' graph |> count_edges()
#'
#' # Create an undirected graph with
#' # 2 nodes and an edge
#' graph_undirected <-
#'   create_graph(directed = FALSE) |>
#'   add_n_nodes(n = 2) |>
#'   add_edge(
#'     from = 1,
#'     to = 2)
#'
#' # Delete the edge; the order of node ID
#' # values provided in `from` and `to`
#' # don't matter for the undirected case
#' graph_undirected |>
#'   delete_edge(
#'     from = 2,
#'     to = 1) |>
#'   count_edges()
#'
#' # The undirected graph has a single
#' # edge with ID `1`; it can be
#' # deleted by specifying `id`
#' graph_undirected |>
#'   delete_edge(id = 1) |>
#'   count_edges()
#'
#' # Create a directed graph with 2
#' # labeled nodes and an edge
#' graph_labeled_nodes <-
#'   create_graph() |>
#'   add_n_nodes(
#'     n = 2,
#'     label = c("one", "two")) |>
#'   add_edge(
#'     from = "one",
#'     to = "two")
#'
#' # Delete the edge using the node
#' # labels in `from` and `to`; this
#' # is analogous to creating the
#' # edge using node labels
#' graph_labeled_nodes |>
#'   delete_edge(
#'     from = "one",
#'     to = "two") |>
#'   count_edges()
#'
#' @family edge creation and removal
#'
#' @export
delete_edge <- function(
    graph,
    from = NULL,
    to = NULL,
    id = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, "So, there cannot be edges to delete.")

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # If a value is supplied for `id`, determine which
  # node ID values the edge ID references
  if (!is.null(id)) {

    # Stop function if the edge ID value supplied is
    # not in the graph's edf
    if (!(id %in% graph$edges_df$id)) {
      cli::cli_abort(c(
        "The edge specified is not present in the graph.",
        "The edge can be any of {.val {unique(graph$edges_df$id)}}."
      ))
    }

    # Get the node ID values for the edge ID
    from_id <- graph$edges_df[which(graph$edges_df$id == id)[1], 2]
    to_id <- graph$edges_df[which(graph$edges_df$id == id)[1], 3]

  } else if (is.null(id)) {

    # Verify that each of the values for `from` and
    # `to` are given as single values
    from_is_single_value <- length(from) == 1

    to_is_single_value <- length(to) == 1

    # Stop function if either node is not a single value
    if (!from_is_single_value || !to_is_single_value) {

      cli::cli_abort(
        "Single-length vectors for `from` and `to` should be specified.")
    }

    # If `from` and `to` values provided as character
    # values, assume that these values refer to node
    # `label` attr values
    if (is.character(from) && is.character(to)) {

      # Stop function if the label for `from` exists in the graph
      if (!from %in% graph$nodes_df$label) {

        cli::cli_abort(
          "The value provided in `from` does not exist as a node `label` value.")
      }

      # Stop function if the label for `from` is not distinct in the graph
      if (graph$nodes_df |>
          dplyr::select("label") |>
          dplyr::filter(label == from) |>
          nrow() > 1) {

        cli::cli_abort(
          "The node `label` provided in `from` is not distinct in the graph.")
      }

      # Stop function if the label for `to` exists in the graph
      if (!to %in% graph$nodes_df$label) {

        cli::cli_abort(
          "The value provided in `to` does not exist as a node `label` value.")
      }

      # Stop function if the label for `to` is not distinct in the graph
      if (graph$nodes_df |>
          dplyr::select("label") |>
          dplyr::filter(label == to) |>
          nrow() > 1) {

        cli::cli_abort(
          "The node `label` provided in `to` is not distinct in the graph.")
      }

      # Use the `translate_to_node_id()` helper function to map
      # node `label` values to node `id` values
      from_to_node_id <-
        translate_to_node_id(
          graph = graph,
          from = from,
          to = to)

      from_id <- from_to_node_id$from
      to_id <- from_to_node_id$to

    } else {

      from_id <- from
      to_id <- to
    }
  }

  # Determine whether the pair of nodes provided
  # are in the graph
  nodes_available_in_graph <-
      all(c(from_id, to_id) %in% get_node_ids(graph))

  # Stop function if both nodes not present in graph
  if (!nodes_available_in_graph) {

    cli::cli_abort(
      "The nodes specified are not both present in the graph.")
  }

  # Extract the graph's edge data frame
  edf <- graph$edges_df

  # Edge removal case for directed graphs
  if (is_graph_directed(graph)) {

    # Stop function if the edge provided is not
    # in the edge data frame
    if (edf |>
        dplyr::filter(from == from_id, to == to_id) |>
        nrow() == 0) {

      cli::cli_abort(
        "The edge provided is not in the graph.")
    }

    # Filter out relevant rows from `edf`
    edf <-
      edf |>
      dplyr::filter(!(from == from_id & to == to_id))

    # Reset the row names in the edf
    row.names(edf) <- NULL

    # Update the graph's edge data frame
    graph$edges_df <- edf
  }

  # Edge removal case for undirected graphs
  if (!is_graph_directed(graph)) {

    # Stop function if the edge provided is not
    # in the edge data frame
    if (edf |>
        dplyr::filter((from == from_id & to == to_id) |
                      (from == to_id & to == from_id)) |>
        nrow() == 0) {

      cli::cli_abort(
        "The edge provided is not in the graph.")
    }

    # Filter out relevant rows from `edf`
    edf <-
      edf |>
      dplyr::filter(!((from == from_id & to == to_id) |
                        (from == to_id & to == from_id)))

    row.names(edf) <- NULL

    # Update the graph's edge data frame
    graph$edges_df <- edf
  }

  # Scavenge any invalid, linked data frames
  graph <-
    remove_linked_dfs(graph)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = -1L)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

# Getting or selection of last-created nodes or edges in a graph")

test_that("getting or selecting the last edges created is possible", {

  # Create a graph, add a cycle, and then
  # add a tree
  graph <-
    create_graph() %>%
    add_cycle(
      n = 3,
      type = "cycle",
      rel = "a") %>%
    add_balanced_tree(
      k = 2,
      h = 2,
      type = "tree",
      rel = "b")

  # Select the last edges created (all edges
  # from the tree)
  graph_e <-
    graph %>%
    select_last_edges_created()

  # Expect that the selection available in
  # the graph is an edge selection
  expect_gt(
    nrow(graph_e$edge_selection), 0)

  expect_equal(
    nrow(graph_e$node_selection), 0)

  # Expect that the edges selected are
  # those that have `rel == 'b'` as an
  # edge attribute (since that attribute
  # belongs to the balanced tree structure,
  # which was created last
  expect_equal(
    get_edge_df(graph_e)[which(get_edge_df(graph_e)$rel == "b"), 1],
    get_selection(graph_e))

  # Get the last edges created directly
  # with `get_last_edges_created()`
  last_edges_created <-
    graph %>%
    get_last_edges_created()

  # Expect the same edge ID values as
  # those that have `rel == 'b'` as an
  # edge attribute
  expect_identical(
    get_edge_df(graph_e)[which(get_edge_df(graph_e)$rel == "b"), 1],
    last_edges_created)

  # Delete an edge from the graph
  graph_edge_deleted <-
    graph %>%
    delete_edge(id = 5)

  # Expect an error when attempting to
  # get the last edges created after
  # having just deleted an edge
  expect_error(
    graph_edge_deleted %>%
      get_last_edges_created())

  # Expect an error when attempting to
  # select the last edges created after
  # having just deleted an edge
  expect_error(
    graph_edge_deleted %>%
      select_last_edges_created())
})

test_that("getting or selecting the last nodes created is possible", {

  # Create a graph, add a cycle, and then
  # add a tree
  graph <-
    create_graph() %>%
    add_cycle(
      n = 3,
      type = "cycle",
      rel = "a") %>%
    add_balanced_tree(
      k = 2,
      h = 2,
      type = "tree",
      rel = "b")

  # Select the last nodes created (all nodes
  # from the tree)
  graph_n <-
    graph %>%
    select_last_nodes_created()

  # Expect that the selection available in
  # the graph is a node selection
  expect_gt(
    nrow(graph_n$node_selection), 0)

  expect_equal(
    nrow(graph_n$edge_selection), 0)

  # Expect that the nodes selected are
  # those that have `type == 'tree'` as a
  # node attribute (since that attribute
  # belongs to the balanced tree structure,
  # which was created last
  expect_identical(
    get_node_df(graph_n)[which(get_node_df(graph_n)$type == "tree"), 1],
    get_selection(graph_n))

  # Get the last nodes created directly
  # with `get_last_nodes_created()`
  last_nodes_created <-
    graph %>%
    get_last_nodes_created()

  # Expect the same node ID values as
  # those that have `type == 'tree'` as a
  # node attribute
  expect_identical(
    get_node_df(graph_n)[which(get_node_df(graph_n)$type == "tree"), 1],
    last_nodes_created)

  # Delete a node from the graph
  graph_node_deleted <-
    graph %>%
    delete_node(node = 10)

  # Expect an error when attempting to
  # get or select the last nodes created after
  # having just deleted a node
  expect_snapshot(error = TRUE, {
    get_last_nodes_created(graph_node_deleted)
    select_last_nodes_created(graph_node_deleted)
  })
})

#' Get closeness vitality
#'
#' @description
#'
#' Get the closeness vitality values for all nodes in the graph.
#'
#' @inheritParams render_graph
#'
#' @return A data frame with closeness vitality values for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 12,
#'     set_seed = 23)
#'
#' # Get closeness vitality values
#' # for all nodes in the graph
#' graph %>% get_closeness_vitality()
#'
#' # Add the closeness vitality
#' # values to the graph as a
#' # node attribute
#' graph <-
#'   graph %>%
#'   join_node_attrs(
#'     df = get_closeness_vitality(.))
#'
#' # Display the graph's
#' # node data frame
#' graph %>% get_node_df()
#'
#' @export
get_closeness_vitality <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the sum of distances between all pairs
  # of nodes in the graph
  sum_distances <- sum(igraph::distances(ig_graph))

  # Calculated closeness vitality values for
  # all nodes in the graph
  closeness_vitality_values <-
    purrr::map(
      seq_len(nrow(graph$nodes_df)),
      function(x) {
        distances <- igraph::distances(igraph::delete_vertices(ig_graph, x))
        sum_distances - sum(distances[!is.infinite(distances)])
      }) %>% unlist()

  # Create df with closeness vitality values
  data.frame(
    id = graph$nodes_df$id,
    closeness_vitality = closeness_vitality_values,
    stringsAsFactors = FALSE)
}

#' Get the PageRank values for all nodes
#'
#' @description
#'
#' Get the PageRank values for all nodes in the graph.
#'
#' @inheritParams render_graph
#' @param directed If `TRUE` (the default) then directed paths will be
#'   considered for directed graphs. This is ignored for undirected graphs.
#' @param damping The damping factor. The default value is set to `0.85`.
#'
#' @return A data frame with PageRank values for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the PageRank scores
#' # for all nodes in the graph
#' graph %>%
#'   get_pagerank()
#'
#' # Colorize nodes according to their
#' # PageRank scores
#' graph <-
#'   graph %>%
#'   join_node_attrs(
#'     df = get_pagerank(graph = .)) %>%
#'   colorize_node_attrs(
#'     node_attr_from = pagerank,
#'     node_attr_to = fillcolor,
#'     palette = "RdYlGn")
#'
#' @export
get_pagerank <- function(
    graph,
    directed = TRUE,
    damping = 0.85
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the PageRank values for each of the
  # graph's nodes
  pagerank_values <-
    igraph::page_rank(
      graph = ig_graph,
      directed = directed,
      damping = damping)$vector

  # Create df with the PageRank values
  data.frame(
    id = names(pagerank_values) %>% as.integer(),
    pagerank = round(pagerank_values, 4),
    stringsAsFactors = FALSE)
}

#nocov start

utils::globalVariables(
  c(
    ".",
    "action_index",
    "action_name",
    "attr",
    "attr_type",
    "deg",
    "degree",
    "df_id",
    "df_id__",
    "display",
    "duration",
    "edge",
    "edges",
    "edge_from_to",
    "e_id",
    "fa_uri",
    "fillcolor",
    "from",
    "from.y",
    "from_label",
    "from_to",
    "from_type",
    "function_used",
    "hex",
    "id",
    "id_external",
    "id.x",
    "id.y",
    "id__",
    "indeg",
    "indegree",
    "indegree_dist",
    "indegree_hist",
    "index",
    "index__",
    "image",
    "label",
    "label_col",
    "n",
    "new_fillcolor",
    "new_node_id",
    "node",
    "nodes",
    "node_edge__",
    "node_id",
    "outdeg",
    "outdegree",
    "outdegree_dist",
    "outdegree_hist",
    "rel",
    "step_created_edges",
    "step_created_nodes",
    "step_deleted_edges",
    "step_deleted_nodes",
    "step_init_with_edges",
    "step_init_with_nodes",
    "string",
    "time_modified",
    "to",
    "to.y",
    "to_label",
    "to_type",
    "total_degree",
    "total_degree_dist",
    "total_degree_hist",
    "type",
    "V1",
    "V2",
    "value",
    "value.x",
    "value_x",
    "value.y",
    "version_id",
    "x",
    "y"
    )
  )

#nocov end

#' Read a graph or graph series from disk
#'
#' @description
#'
#' Load a graph or a graph series object from disk.
#'
#' @param file The filename for the graph or graph series. Optionally, this may
#'   contain a path to the file.
#' @examples
#' # Create an undirected GNP
#' # graph with 100 nodes using
#' # a probability value of 0.05
#' gnp_graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnp_graph(
#'     n = 100,
#'     p = 0.05
#'   )
#'
#' # Save the graph to disk; use
#' # the file name `gnp_graph.dgr`
#' # save_graph(
#' #   x = gnp_graph,
#' #   file = "gnp_graph"
#' # )
#'
#' # To read the graph file from
#' # disk, use `open_graph()`
#' # gnp_graph_2 <-
#' #   open_graph(
#' #     file = "gnp_graph.dgr"
#' #   )
#'
#' @export
open_graph <- function(file) {

  check_string(file)

  # Read the graph or graph series
  x <- readRDS(file = file)

  if (!rlang::inherits_any(x, c("dgr_graph", "dgr_graph_1D"))) {

    cli::cli_abort(c(
      "The object must be a graph or graph series.",
      i = "Make sure that {.file {file}} contains a graph or graph series."))
  }

  x
}

#' Determine whether a specified edge is present
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, determine whether an edge (defined
#' by a pair of node IDs or node label values) is present.
#'
#' @inheritParams render_graph
#' @param edge An edge ID value to test for presence in the graph. If a single,
#'   numeric value is provided then values for `from` or `to` needn't be
#'   supplied.
#' @param from A node ID from which the edge is outgoing, or, the label
#'   associated with the node. For an undirected graph, the value in `from` can
#'   be interchangeable with that in `to`.
#' @param to A node ID to which the edge is incoming, or, the label associated
#'   with the node. For an undirected graph, the value in `to` can be
#'   interchangeable with that in `from`.
#'
#' @return A logical value.
#'
#' @examples
#' # Create a simple graph with
#' # a path of four nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(
#'     n = 4,
#'     type = "path",
#'     label = c("one", "two",
#'               "three", "four"))
#'
#' # Find out if edge ID `3`
#' # is present in the graph
#' graph %>%
#'   is_edge_present(edge = 3)
#'
#' # Determine if there are any edges
#' # with the definition `1` -> `2`
#' graph %>%
#'   is_edge_present(
#'     from = 1,
#'     to = 2)
#'
#' # Determine if there are any edges
#' # with the definition `4` -> `5`
#' graph %>%
#'   is_edge_present(
#'     from = 4,
#'     to = 5)
#'
#' # Determine whether an edge,
#' # defined by its labels as
#' # `two` -> `three`, exists in
#' # the graph
#' graph %>%
#'   is_edge_present(
#'     from = "two",
#'     to = "three")
#'
#' # Set the graph as undirected
#' # and determine whether an
#' # edge between nodes with labels
#' # `three` and `two` exists
#' graph %>%
#'   set_graph_undirected() %>%
#'   is_edge_present(
#'     from = "three",
#'     to = "two")
#'
#' @export
is_edge_present <- function(
    graph,
    edge = NULL,
    from = NULL,
    to = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  use_edge <- !is.null(edge)

  use_from_to <- !is.null(from) && !is.null(to)

  if (use_edge && use_from_to) {
    use_from_to <- FALSE
  }

  if (!use_edge && !use_from_to) {
    cli::cli_abort(
      "Either provide an edge ID or a pair of nodes to test for edge presence.")
  }

  if (use_edge) {

    # Verify that what is provided for `edge`
    # is a numeric value of length 1
    check_number_decimal(edge)

    edge_is_in_graph <- edge %in% graph$edges_df$id

    return(edge_is_in_graph)
  }

  if (use_from_to) {

    # Verify that each of the values for `from` and
    # `to` are given as a single value
    from_is_single_value <- length(from) == 1
    to_is_single_value   <- length(to) == 1

    # Stop function if either node is not a single value
    if (!from_is_single_value || !to_is_single_value) {
      cli::cli_abort(
        "Only single nodes for `from` and `to` should be specified.")
    }

    if (inherits(from, "character") && inherits(to, "character")) {

      # Determine whether the pair of
      # labels provided are in the graph
      nodes_available_in_graph <-
        all(c(from, to) %in% graph$nodes_df$label)

      # Return FALSE if both nodes not
      # present in graph
      if (!nodes_available_in_graph) {
        return(FALSE)
      }

      # Use the `translate_to_node_id()` helper function to map
      # node `label` values to node `id` values
      from_to_node_id <-
        translate_to_node_id(
          graph = graph,
          from = from,
          to = to)

      from <- from_to_node_id$from
      to <- from_to_node_id$to

      # Determine whether a matching edge is
      # available in the graph
      if (graph$directed) {

        edge_is_in_graph <-
            any(graph$edges_df$from == from &
                  graph$edges_df$to == to)

      } else if (!graph$directed) {

        edge_is_in_graph <-
            any(graph$edges_df$from == from &
                  graph$edges_df$to == to) |
              any(graph$edges_df$from == to &
                    graph$edges_df$to == from)
      }

      return(edge_is_in_graph)
    }

    if (inherits(from, "numeric") && inherits(to, "numeric")) {

      # Determine whether the pair of
      # labels provided are in the graph
      if (from_is_single_value && to_is_single_value) {
        nodes_available_in_graph <-
          all(c(from, to) %in% get_node_ids(graph))
      }

      # Return FALSE if both nodes not
      # present in graph
      if (!nodes_available_in_graph) {
        return(FALSE)
      }

      # Determine whether a matching edge is
      # available in the graph
      if (graph$directed) {

        edge_is_in_graph <-
            any(graph$edges_df$from == from &
                  graph$edges_df$to == to)

      } else if (!graph$directed) {

        edge_is_in_graph <-
            any(graph$edges_df$from == from &
                  graph$edges_df$to == to) |
              any(graph$edges_df$from == to &
                    graph$edges_df$to == from)
      }

      return(edge_is_in_graph)
    }
  }
}

#' Get the maximum graph eccentricity
#'
#' @description
#'
#' Get the diameter of a graph, which is the largest eccentricity in the graph.
#' The graph eccentricity of a node is its shortest path from the farthest other
#' node in the graph.
#'
#' @inheritParams render_graph
#'
#' @return A single numeric value representing the maximum eccentricity of the
#'   graph.
#'
#' @examples
#' # Create a cycle graph
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' # Determine the graph's maximum
#' # eccentricity
#' graph |>
#'   get_max_eccentricity()
#'
#' # Create a full graph and then
#' # get the maximum eccentricity
#' # value for that
#' create_graph() |>
#'   add_full_graph(n = 10) |>
#'   get_max_eccentricity()
#'
#' @export
get_max_eccentricity <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the maximum eccentricity of the graph
  igraph::diameter(ig_graph)
}

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @import rlang
#' @import htmlwidgets
#' @import RColorBrewer
#' @import glue
#' @importFrom magrittr not
#' @import visNetwork
## usethis namespace: end
NULL

#' Add a multiple of new nodes with edges to or from one or more selected nodes
#'
#' @description
#'
#' Add `n` new nodes to or from one or more nodes available as a selection in a
#' graph object of class `dgr_graph`. New graph edges will all move either from
#' the nodes in the selection toward the newly created nodes (with the option
#' `direction = "from"`), or to the selected nodes already in the graph (using
#' `direction = "to"`). Optionally, set node `type` and edge `rel` values for
#' all the new nodes and edges created, respectively.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams node_edge_aes_data
#' @inheritParams render_graph
#' @param n The number of new nodes to attach as successor nodes to the nodes in
#'   the selection.
#' @param direction Using `from` will create new edges from existing nodes to
#'   the new nodes. The `to` option will create new edges directed toward the
#'   existing nodes.
#' @param type An optional character vector that provides group identifiers for
#'   the nodes to be added.
#' @param label An optional character object that describes the nodes to be
#'   added.
#' @param rel An optional string to apply a `rel` attribute to all newly created
#'   edges.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph, add a node to it, select
#' # that node, and then add 5 more nodes to the graph
#' # with edges from the original node to all of the
#' # new nodes
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(n = 1) %>%
#'   select_last_nodes_created() %>%
#'   add_n_nodes_ws(
#'     n = 5,
#'     direction = "from")
#'
#' # Get the graph's nodes
#' graph %>% get_node_ids()
#'
#' # Get the graph's edges
#' graph %>% get_edges()
#'
#' # Create an empty graph, add a node to it, select
#' # that node, and then add 5 more nodes to the graph
#' # with edges toward the original node from all of
#' # the new nodes
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(n = 1) %>%
#'   select_last_nodes_created() %>%
#'   add_n_nodes_ws(
#'     n = 5,
#'     direction = "to")
#'
#' # Get the graph's nodes
#' graph %>% get_node_ids()
#'
#' # Get the graph's edges
#' graph %>% get_edges()
#'
#' @family node creation and removal
#'
#' @export
add_n_nodes_ws <- function(
    graph,
    n,
    direction = NULL,
    type = NULL,
    label = NULL,
    rel = NULL,
    node_aes = NULL,
    edge_aes = NULL,
    node_data = NULL,
    edge_data = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, extra_msg = "Existing nodes are required.")

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # If the graph is directed and there is no value
  # given for the `direction` argument, stop function
  if (is_graph_directed(graph) && is.null(direction)) {

    cli::cli_abort("A `direction` edge must be provided for a directed graph.")
  }

  # If the graph is undirected, set the direction
  # to `to`
  if (!is_graph_directed(graph)) {
    direction <- "to"
  }

  # If `type`, `label`, or `rel` is NULL, convert to NA
  type  <- type %||% NA_character_
  label <- label %||% NA_character_
  rel   <-   rel %||% NA_character_

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph %>% count_nodes()

  # Get the number of edges in the graph
  edges_graph_1 <- graph %>% count_edges()

  # Get a vector of nodes available in the
  # graph's selection
  nodes_in_selection <- graph$node_selection$node

  # Case where nodes are added with edges to the
  # selected nodes
  if (direction == "to") {

    for (i in seq_along(nodes_in_selection)) {

      graph <-
        add_n_nodes(
          graph = graph,
          n = n,
          type = type,
          label = label)

      graph$graph_log <-
        graph$graph_log[-nrow(graph$graph_log), ]

      for (j in (max(graph$nodes_df$id) - n + 1):max(graph$nodes_df$id)) {

        graph <-
          add_edge(
            graph = graph,
            from = j,
            to = nodes_in_selection[i],
            rel = rel)

        graph$graph_log <-
          graph$graph_log[-nrow(graph$graph_log), ]
      }
    }
  }

  # Case where nodes are added with edges from the
  # selected nodes
  if (direction == "from") {

    for (i in seq_along(nodes_in_selection)) {

      graph <-
        add_n_nodes(
          graph = graph,
          n = n,
          type = type,
          label = label)

      graph$graph_log <-
        graph$graph_log[-nrow(graph$graph_log), ]

      for (j in (max(graph$nodes_df$id) - n + 1):max(graph$nodes_df$id)) {

        graph <-
          add_edge(
            graph = graph,
            from = nodes_in_selection[i],
            to = j,
            rel = rel)

        graph$graph_log <-
          graph$graph_log[-nrow(graph$graph_log), ]
      }
    }
  }

  # Modify the graph object
  graph$directed <- is_graph_directed(graph)

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph %>% count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Get the edge ID values for the
  # last edges created
  new_edge_id <-
    graph %>%
    get_edge_df() %>%
    utils::tail(edges_added) %>%
    dplyr::pull("id")

  # Get the node ID values for the
  # last nodes created
  new_node_id <-
    graph %>%
    get_node_df() %>%
    utils::tail(nodes_added) %>%
    dplyr::pull("id")

  # Collect node aesthetic attributes
  if (!is.null(node_aes)) {

    node_aes_tbl <- dplyr::as_tibble(node_aes)

    if (nrow(node_aes_tbl) < nodes_added) {

      node_aes$index__ <- seq_len(nodes_added)

      node_aes_tbl <-
        dplyr::as_tibble(node_aes) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_aes_tbl)) {
      node_aes_tbl$id <- NULL
    }
  }

  # Collect node data attributes
  if (!is.null(node_data)) {

    node_data_tbl <- dplyr::as_tibble(node_data)

    if (nrow(node_data_tbl) < nodes_added) {

      node_data$index__ <- seq_len(nodes_added)

      node_data_tbl <-
        dplyr::as_tibble(node_data) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_data_tbl)) {
      node_data_tbl$id <- NULL
    }
  }

  # Collect edge aesthetic attributes
  if (!is.null(edge_aes)) {

    edge_aes_tbl <- dplyr::as_tibble(edge_aes)

    if (nrow(edge_aes_tbl) < edges_added) {

      edge_aes$index__ <- seq_len(edges_added)

      edge_aes_tbl <-
        dplyr::as_tibble(edge_aes) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_aes_tbl)) {
      edge_aes_tbl$id <- NULL
    }
  }

  # Collect edge data attributes
  if (!is.null(edge_data)) {

    edge_data_tbl <- dplyr::as_tibble(edge_data)

    if (nrow(edge_data_tbl) < edges_added) {

      edge_data$index__ <- seq_len(edges_added)

      edge_data_tbl <-
        dplyr::as_tibble(edge_data) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_data_tbl)) {
      edge_data_tbl$id <- NULL
    }
  }

  # Add node aesthetics if available
  if (exists("node_aes_tbl")) {

    node_aes_tbl <-
      node_aes_tbl %>%
      dplyr::mutate(id = new_node_id, .before = 0)

    columns_to_select <-
      c("id", base::setdiff(colnames(graph$nodes_df), colnames(node_aes_tbl)))

    graph$nodes_df <-
      dplyr::bind_rows(
        graph$nodes_df %>%
          dplyr::filter(!(id %in% new_node_id)),
        graph$nodes_df %>%
          dplyr::filter(id %in% new_node_id) %>%
          dplyr::select(dplyr::all_of(columns_to_select)) %>%
          dplyr::left_join(node_aes_tbl, by = "id"))
  }

  # Add node data if available
  if (exists("node_data_tbl")) {

    node_data_tbl <-
      node_data_tbl %>%
      dplyr::mutate(id = new_node_id, .before = 0)

    columns_to_select <-
      c("id", base::setdiff(colnames(graph$nodes_df), colnames(node_data_tbl)))

    graph$nodes_df <-
      dplyr::bind_rows(
        graph$nodes_df %>%
          dplyr::filter(!(id %in% new_node_id)),
        graph$nodes_df %>%
          dplyr::filter(id %in% new_node_id) %>%
          dplyr::select(dplyr::all_of(columns_to_select)) %>%
          dplyr::left_join(node_data_tbl, by = "id"))
  }

  # Add edge aesthetics if available
  if (exists("edge_aes_tbl")) {

    edge_aes_tbl <-
      edge_aes_tbl %>%
      dplyr::mutate(id = new_edge_id, .before = 0)

    columns_to_select <-
      c("id", base::setdiff(colnames(graph$edges_df), colnames(edge_aes_tbl)))

    graph$edges_df <-
      dplyr::bind_rows(
        graph$edges_df %>%
          dplyr::filter(!(id %in% new_edge_id)),
        graph$edges_df %>%
          dplyr::filter(id %in% new_edge_id) %>%
          dplyr::select(dplyr::all_of(columns_to_select)) %>%
          dplyr::left_join(edge_aes_tbl, by = "id"))
  }

  # Add edge data if available
  if (exists("edge_data_tbl")) {

    edge_data_tbl <-
      edge_data_tbl %>%
      dplyr::mutate(id = new_edge_id, .before = 0)

    columns_to_select <-
      c("id", base::setdiff(colnames(graph$edges_df), colnames(edge_data_tbl)))

    graph$edges_df <-
      dplyr::bind_rows(
        graph$edges_df %>%
          dplyr::filter(!(id %in% new_edge_id)),
        graph$edges_df %>%
          dplyr::filter(id %in% new_edge_id) %>%
          dplyr::select(dplyr::all_of(columns_to_select)) %>%
          dplyr::left_join(edge_data_tbl, by = "id"))
  }

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_added,
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get neighboring nodes based on node attribute similarity
#'
#' @description
#'
#' With a graph a single node serving as the starting point, get those nodes in
#' a potential neighborhood of nodes (adjacent to the starting node) that have a
#' common or similar (within threshold values) node attribute to the starting
#' node.
#'
#' @inheritParams render_graph
#' @param node A single-length vector containing a node ID value.
#' @param node_attr The name of the node attribute to use to compare with
#'   adjacent nodes.
#' @param tol_abs If the values contained in the node attribute `node_attr` are
#'   numeric, one can optionally supply a numeric vector of length 2 that
#'   provides a lower and upper numeric bound as criteria for neighboring node
#'   similarity to the starting node.
#' @param tol_pct If the values contained in the node attribute `node_attr` are
#'   numeric, one can optionally supply a numeric vector of length 2 that
#'   specifies lower and upper bounds as negative and positive percentage
#'   changes to the value of the starting node. These bounds serve as criteria
#'   for neighboring node similarity to the starting node.
#'
#' @return A vector of node ID values.
#'
#' @examples
#' # Getting similar neighbors can
#' # be done through numerical comparisons;
#' # start by creating a random, directed
#' # graph with 18 nodes and 22 edges
#' # using the `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 18,
#'     m = 25,
#'     set_seed = 23) %>%
#'   set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = count_nodes(.),
#'       mean = 5,
#'       sd = 1) %>% round(0))
#'
#' # Starting with node `10`, we
#' # can test whether any nodes
#' # adjacent and beyond are
#' # numerically equivalent in `value`
#' graph %>%
#'   get_similar_nbrs(
#'     node = 10,
#'     node_attr = value)
#'
#' # We can also set a tolerance
#' # for ascribing similarly by using
#' # either the `tol_abs` or `tol_pct`
#' # arguments (the first applies absolute
#' # lower and upper bounds from the
#' # value in the starting node and the
#' # latter uses a percentage difference
#' # to do the same); try setting `tol_abs`
#' # with a fairly large range to
#' # determine if several nodes can be
#' # selected
#' graph %>%
#'   get_similar_nbrs(
#'     node = 10,
#'     node_attr = value,
#'     tol_abs = c(1, 1))
#'
#' # That resulted in a fairly large
#' # set of 4 neigboring nodes; for
#' # sake of example, setting the range
#' # to be very large will effectively
#' # return all nodes in the graph
#' # except for the starting node
#' graph %>%
#'   get_similar_nbrs(
#'     node = 10,
#'     node_attr = value,
#'     tol_abs = c(10, 10)) %>%
#'     length()
#'
#' @export
get_similar_nbrs <- function(
    graph,
    node,
    node_attr,
    tol_abs = NULL,
    tol_pct = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

    # Get the requested `node_attr`
  node_attr <-
    rlang::enquo(node_attr) %>% rlang::get_expr() %>% as.character()

  # Get value to match on
  match <-
    get_node_df(graph)[
      which(get_node_df(graph)[, 1] ==
              node),
      which(colnames(get_node_df(graph)) ==
              node_attr)]

  # Create an empty list object
  nodes <- list()

  # Extract all `node_attr` values to test for their
  # type
  attr_values <-
    get_node_df(graph)[
      , which(colnames(get_node_df(graph)) ==
                node_attr)]

  # Determine whether `node_attr` values are numeric
  node_attr_numeric <-
    suppressWarnings(
      !anyNA(as.numeric(attr_values)))

  if (node_attr_numeric) {

    # Get the set of all nodes in graph that
    # satisfy one or more conditions
    graph_nodes_with_attr <-
      graph$nodes_df[
        which(
          graph$nodes_df[, which(
            colnames(graph$nodes_df) ==
              node_attr)] %in% match), 1]
  }

  if (node_attr_numeric) {

    match <- as.numeric(match)

    if (!is.null(tol_abs)) {
      match_range <-
        c(match - tol_abs[1], match + tol_abs[2])
    }

    if (!is.null(tol_pct)) {
      match_range <-
        c(match - match * tol_pct[1]/100,
          match + match * tol_pct[2]/100)
    }

    if (is.null(tol_abs) && is.null(tol_pct)) {
      match_range <- c(match, match)
    }

    # Get the set of all nodes in graph that
    # satisfy one or more conditions
    graph_nodes_with_attr <-
      graph$nodes_df[
        intersect(
          which(
            as.numeric(graph$nodes_df[, which(
              colnames(graph$nodes_df) ==
                node_attr)]) >= match_range[1]),
          which(
            as.numeric(graph$nodes_df[, which(
              colnames(graph$nodes_df) ==
                node_attr)]) <= match_range[2])), 1]
  }

  # place starting node in the neighbourhood vector
  neighborhood <- node

  # Initialize `i`
  i <- 1

  repeat {

    # From the starting node get all adjacent nodes
    # that are not in the `neighborhood` vector
    neighborhood <-
      unique(
        c(neighborhood,
          intersect(
            unique(c(
              get_edges(
                graph,
                return_type = "df")[
                  which(get_edges(
                    graph,
                    return_type = "df")[, 1] %in%
                      neighborhood), 2],
              get_edges(
                graph,
                return_type = "df")[
                  which(get_edges(
                    graph,
                    return_type = "df")[, 2] %in%
                      neighborhood), 1])),
            graph_nodes_with_attr)))

    # Place revised neighborhood nodes in `nodes` list
    nodes[[i]] <- neighborhood

    # Break if current iteration yields no change in
    # the `nodes` list
    if (i > 1) {
      if (identical(nodes[[i]], nodes[[i - 1]])) break
    }
    i <- i + 1
  }

  # Get the final set of nodes that satisfy similarity
  # and adjacency conditions
  matching_nodes <-
    base::setdiff(nodes[length(nodes)][[1]], node)

  # If there are no matching nodes return `NA`
  if (length(matching_nodes) == 0) {
    return(NA)
  }

  sort(matching_nodes)

}

# Determining neighbors and non-neighbors of nodes")

test_that("Getting all neighbors of one or more nodes is possible", {

  # Create a random, directed graph with 18 nodes
  # and 22 edges
  random_graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 18,
      m = 22,
      set_seed = 23)

  # Find all neighbor nodes for node `5`
  all_nbrs_5 <-
    get_nbrs(
      graph = random_graph,
      nodes = 5)

  # Expect certain nodes as neighbors
  expect_identical(
    all_nbrs_5, 6)

  # Find all neighbor nodes for nodes `5`, `7`,
  # and `15`
  all_nbrs_5_7_15 <-
    get_nbrs(
      graph = random_graph,
      nodes = c(5, 7, 15))

  # Expect certain nodes as neighbors
  expect_identical(
    all_nbrs_5_7_15,
    c(3, 4, 6, 7, 12, 15))

  # Expect an NA value if there are no neighbors
  expect_true(
    is.na(
      get_nbrs(
        graph = random_graph,
        nodes = 17)))
})

test_that("Getting non-neighbors of a node is possible", {

  # Create a random, directed graph
  # with 18 nodes and 22 edges
  random_graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 18,
      m = 22,
      set_seed = 23)

  # Find all non-neighbors of node `5`
  non_nbrs_5 <-
    get_non_nbrs(
      graph = random_graph,
      node = 5)

  # Expect certain nodes that
  # are non-neighbors of node `5`
  expect_equal(
    non_nbrs_5,
    c(1, 2, 3, 4, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18))

  # Expect that NA is returned when
  # there are no non-neighbors
  expect_true(
    create_graph() %>%
      add_path(n = 3) %>%
      get_non_nbrs(node = 2) %>%
      is.na())
})

test_that("Getting common neighbors of 2 or more nodes is possible", {

  # Create a random, directed graph with 18 nodes
  # and 22 edges
  random_graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 18,
      m = 22,
      set_seed = 23)

  # Expect NA when finding all
  # common neighbor nodes for nodes
  # `5` and `7` (there are no
  # common neighbors)
  expect_true(
    is.na(
      get_common_nbrs(
        graph = random_graph,
        nodes = c(5, 7))))

  # Expect a common neighbor
  # node of `4` for nodes
  # `1` and  `7`
  expect_equal(
    get_common_nbrs(
      graph = random_graph,
      nodes = c(1, 7)), 4)
})

test_that("Getting similar neighbors of a node is possible", {

  # Create a random, directed graph with 18 nodes
  # and 22 edges
  random_graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 18,
      m = 22,
      set_seed = 23) %>%
    set_node_attrs(
      node_attr = value,
      values = rnorm(count_nodes(graph = .), 5, 2))

  # Expect NA when searching any nodes adjacent to
  # node `2` and beyond since the immediately adjacent
  # are not numerically equivalent in `value`
  expect_true(
    is.na(
      get_similar_nbrs(
        graph = random_graph,
        node = 2,
        node_attr = value)))

  # Expect several nodes to be
  # matched when using an absolute
  # tolerance range
  expect_identical(
    get_similar_nbrs(
      graph = random_graph,
      node = 8,
      node_attr = value,
      tol_abs = c(2, 2)),
    c(1, 10))

  # Expect all connected nodes
  # to be matched, except the
  # starting node, when using a
  # very high absolute tolerance
  # range
  expect_equal(
    get_similar_nbrs(
      graph = random_graph,
      node = 8,
      node_attr = value,
      tol_abs = c(12, 12)),
    setdiff(1:18, c(8, 17)))

  # Expect certain nodes to be
  # matched, when using tolerance
  # specified as a low and high percentage
  expect_equal(
    get_similar_nbrs(
      graph = random_graph,
      node = 3,
      node_attr = value,
      tol_pct = c(50, 50)),
    c(1, 2, 4, 5, 6, 8, 9, 10, 11, 13, 14, 15, 18))
})

#' Is the graph a connected graph?
#'
#' @description
#'
#' Determines whether a graph is a connected graph.
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function; this
#' # graph is not connected
#' create_graph() %>%
#'   add_gnm_graph(
#'     n = 15,
#'     m = 10,
#'     set_seed = 23) %>%
#'   is_graph_connected()
#'
#' # Create another random graph;
#' # this graph is connected
#' create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23) %>%
#'   is_graph_connected()
#'
#' @export
is_graph_connected <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  wc_components <- get_w_connected_cmpts(graph)

  if (length(unique(wc_components$wc_component)) > 1) {
    return(FALSE)
  } else {
    return(TRUE)
  }
}

#' Create an edge data frame
#'
#' @description
#'
#' Combine several vectors for edges and their attributes into a data frame,
#' which can be combined with other similarly-generated data frames, or, added
#' to a graph object. An edge data frame, or edf, has at least the following
#' columns:
#'
#' - `id` (of type `integer`)
#'
#' - `from` (of type `integer`)
#'
#' - `to` (of type `integer`)
#'
#' - `rel` (of type `character`)
#'
#' An arbitrary number of additional columns containing aesthetic or data
#' attributes can be part of the edf, so long as they follow the aforementioned
#' columns. Some examples are included in [edge_aes()]
#'
#' @param from A vector of node ID values from which edges are outbound. The
#'   vector length must equal that of the `to` vector.
#' @param to A vector of node ID values to which edges are incoming. The vector
#'   length must equal that of the `from` vector.
#' @param rel An optional `rel` label for each edge.
#' @param ... One or more vectors for associated edge attributes. Can be some of [edge_aes()]
#'
#' @return An edge data frame (edf).
#'
#' @examples
#' # Create a simple edge data frame (edf) and
#' # view the results
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "a")
#'
#' # Display the edge data frame
#' edf
#'
#' # Create an edf with additional edge
#' # attributes (where their classes will
#' # be inferred from the input vectors)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "a",
#'     length = c(50, 100, 250),
#'     color = "green",
#'     width = c(1, 5, 2))
#'
#' # Display the edge data frame
#' edf
#'
#' @family edge creation and removal
#' @export
create_edge_df <- function(
    from,
    to,
    rel = NULL,
    ...
) {

  # Stop function if vector lengths for `from` and
  # `to` are not equal
  stopifnot(length(from) == length(to))

  # Get the number of edges to be created
  n <- length(from)

  # Ensure that `from` and `to` vector are integers
  from <- as.integer(from)
  to <- as.integer(to)

  # if `rel` is NULL, create character vector with
  # NA values; class as character otherwise
  if (is.null(rel)) {
    rel <- rep(NA_character_, length(from))
  } else {
    rel <- as.character(rel)
  }

  if (!is.null(rel)) {
    # Expand vectors with single values to fill to
    # number of edges
    if (length(rel) == 1) {
      rel <- rep(rel, length(from))
    }

    # Expand vectors with `length` > `1` and
    # `length` < `length(from)`
    if (length(rel) > 1 &&
        length(rel) < length(from)) {
      rel <-
        c(rel,
          rep(NA_character_,
              (length(from) - length(rel))))
    }

    # Trim vectors with number of values exceeding
    # the number of edges
    if (length(rel) > length(from)) {
      rel <- rel[seq_along(from)]
    }
  }

  # Collect extra vectors of data as `extras`
  extras <- list(...)

  if (length(extras) > 0) {
    for (i in seq_along(extras)) {

      # Expand vectors with single values to fill to
      # the number of edges
      if (length(extras[[i]]) == 1) {
        extras[[i]] <- rep(extras[[i]], length(from))
      }

      # Expand vectors with `length` > `1` and
      # `length` < `length(from)`
      if (length(extras[[i]]) > 1 &&
          length(extras[[i]]) < length(from)) {
        extras[[i]] <-
          c(extras[[i]],
            rep(NA_character_,
                (length(from) -
                   length(extras[[i]]))))
      }

      # Trim vectors with number of values exceeding
      # the number of edges
      if (length(extras[[i]]) > length(from)) {
        extras[[i]] <- extras[[i]][seq_along(from)]
      }
    }

    # Create a data frame from the `extras` list
    extras <-
      as.data.frame(
        extras, stringsAsFactors = FALSE)
  }

  if (inherits(extras, "data.frame")) {
    edges_df <-
      dplyr::bind_cols(
        data.frame(
          id = seq_len(n),
          from = from,
          to = to,
          rel = rel,
          stringsAsFactors = FALSE),
        extras)
  } else {
    edges_df <-
      data.frame(
        id = seq_len(n),
        from = from,
        to = to,
        rel = rel,
        stringsAsFactors = FALSE)
  }

  edges_df
}

# Getting info about a node's predecessors and successors

test_that("getting a node's predecessors/successors is possible", {

  suppressWarnings(RNGversion("3.5.0"))
  set.seed(26)

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 26,
      label = TRUE,
      type = c(rep("a_to_g", 7),
               rep("h_to_p", 9),
               rep("q_to_x", 8),
               rep("y_and_z", 2)))

  # Create an edge data frame
  edges <-
    create_edge_df(
      from = sample(1:26, replace = FALSE),
      to = sample(1:26, replace = FALSE),
      label = "edge",
      rel = "letter_to_letter")

  # Create the graph object using the node and
  # edge data frames
  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Tests for `get_predecessors()`
  expect_type(
    get_predecessors(
      graph = graph,
      node = 26),
    "integer")

  expect_equal(
    get_predecessors(
      graph = graph,
      node = 26), 1)

  # Tests for `get_successors()`
  expect_type(
    get_successors(
      graph = graph,
      node = 1),
    "integer")

  expect_equal(
    get_successors(
      graph = graph,
      node = 1), 26)
})

# Getting info about nodes and edges

test_that("getting info about a graph's nodes is possible", {

  suppressWarnings(RNGversion("3.5.0"))
  set.seed(26)

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 26,
      label = LETTERS,
      type = c(rep("a_to_g", 7),
               rep("h_to_p", 9),
               rep("q_to_x", 8),
               rep("y_and_z", 2)))

  # Create an edge data frame
  edges <-
    create_edge_df(
      from = sample(1:26, replace = TRUE),
      to = sample(1:26, replace = TRUE),
      label = "edge",
      rel = "letter_to_letter")

  # Create the graph object using the node and
  # edge data frames
  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Get information on the graph's nodes
  info_nodes <- get_node_info(graph)

  # Expect a data frame object
  expect_s3_class(
    info_nodes, "data.frame")

  # Expect that the data frame has 7 columns
  expect_equal(
    ncol(info_nodes), 7)

  # Expect that the data frame has 26 rows
  expect_equal(
    nrow(info_nodes), 26)

  # Expect that certain columns will be classed
  # as `integer` or `character`
  expect_type(
    info_nodes$id, "integer")

  expect_type(
    info_nodes$label, "character")

  expect_type(
    info_nodes$type, "character")

  # Expect that certain columns will be classed
  # as `numeric` double
  expect_type(
    info_nodes$deg, "double")

  expect_type(
    info_nodes$indeg, "double")

  expect_type(
    info_nodes$outdeg, "double")

  expect_type(
    info_nodes$loops, "double")

  # Create a graph with 4 nodes (with type
  # information) but no edges
  graph <-
    create_graph() %>%
    add_node(type = "free") %>%
    add_node(type = "free") %>%
    add_node(type = "free") %>%
    add_node(type = "free")

  # Get information on nodes that have no edges
  info_nodes_no_edges <- get_node_info(graph)

  # Expect a data frame object
  expect_s3_class(
    info_nodes_no_edges, "data.frame")

  # Expect that the data frame has 7 columns
  expect_equal(
    ncol(info_nodes_no_edges), 7)

  # Expect that the data frame has 4 rows
  expect_equal(
    nrow(info_nodes_no_edges), 4)

  # Expect that the `deg`, `indeg`, `outdeg`,
  # and `loops` columns have show 0
  expect_equal(
    info_nodes_no_edges$deg %>%
      unique(), 0)

  expect_equal(
    info_nodes_no_edges$indeg %>%
      unique(), 0)

  expect_equal(
    info_nodes_no_edges$outdeg %>%
      unique(), 0)

  expect_equal(
    info_nodes_no_edges$loops %>%
      unique(), 0)

  # Create an empty graph
  graph <- create_graph()

  # Get information on nodes from the empty graph
  info_nodes_empty_graph <- get_node_info(graph)

  # Expect a NULL value
  expect_null(
    info_nodes_empty_graph)
})

test_that("getting info about a graph's edges is possible", {

  suppressWarnings(RNGversion("3.5.0"))
  set.seed(26)

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 26,
      label = LETTERS,
      type = c(rep("a_to_g", 7),
               rep("h_to_p", 9),
               rep("q_to_x", 8),
               rep("y_and_z", 2)))

  # Create an edge data frame
  edges <-
    create_edge_df(
      from = sample(1:26, replace = TRUE),
      to = sample(1:26, replace = TRUE),
      label = "edge",
      rel = "letter_to_letter")

  # Create the graph object using the node and
  # edge data frames
  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Get information on the graph's edges
  info_edges <- get_edge_info(graph)

  # Expect a data frame object
  expect_s3_class(
    info_edges, "data.frame")

  # Expect that the data frame has 4 columns
  expect_equal(
    ncol(info_edges), 4)

  # Expect that the data frame has 26 rows
  expect_equal(
    nrow(info_edges), 26)

  # Expect that columns will be classed
  # as either as `integer` or `character`
  expect_type(info_edges$from, "integer")
  expect_type(info_edges$to, "integer")
  expect_type(info_edges$rel, "character")

  # Create a graph with 4 nodes but no edges
  graph <-
    create_graph() %>%
    add_node() %>%
    add_node() %>%
    add_node() %>%
    add_node()

  # Get information on a graph that has no edges
  info_graph_no_edges <- get_edge_info(graph)

  # Expect an NA value
  expect_true(is.na(info_graph_no_edges))
})

# Generate dot syntax from graph objects

# helper functions
node <- function(id, label = NA) {
  if (all(is.na(label))) {
    paste0("'", id, "'")
  } else {
    paste0("'", id, "' \\[label = '", label, "'\\]")
  }
}
edge <- function(from, to) paste0("'", from, "'->'", to, "'")
attrib_block <- "\\[[[:alnum:]'=.,[:space:]]*\\]"

expect_dot <- function(graph, pattern) {
  expect_match(
    object = generate_dot(graph),
    regexp = paste0(pattern, collapse = "[[:space:]]*"))
}

test_that("Simple graph translates into specific form", {

  # Create simple graph
  nodes <- create_node_df(n = 10)

  edges <-
    create_edge_df(
      from = 1:9,
      to = 2:10)

  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  expect_dot(
    graph,
    c("^digraph", "\\{",
      "graph", attrib_block,
      "node", attrib_block,
      "edge", attrib_block,
      node(1:10),
      edge(1:9, 2:10),
      "\\}$")
    )
})

# test_that("Graph with clustered nodes create subgraph", {
#
#   # Create node clusters
#
#   nodes <-
#     create_node_df(
#       n = 6,
#       cluster = c(NA, 'foo', 'foo', 'bar', NA, 'bar'))
#
#   edges <-
#     create_edge_df(
#       from = 1:5,
#       to = 2:6)
#
#   graph <-
#     create_graph(
#       nodes_df = nodes,
#       edges_df = edges)
#
#   expect_dot(
#     graph,
#     c("^digraph", "\\{",
#       "graph", attrib_block,
#       "node", attrib_block,
#       "edge", attrib_block,
#       node(c(1,5)),
#       "subgraph cluster2\\{\nlabel='bar'",
#       node(c(4,6)), "\\}",
#       "subgraph cluster3\\{\nlabel='foo'",
#       node(c(2,3)), "\\}",
#       edge(1:5, 2:6),
#       "\\}$")
#     )
#
#   # Example from @Enchufa2
#
#   nodes = data.frame(id = c(1,2,3),
#                      label = c("a","b","c"),
#                      cluster = c(NA,"data","data"))
#   edges = data.frame(from = c(1,2),
#                      to = c(3,3))
#
#   graph <- create_graph() %>%
#     add_nodes_from_table(nodes, label_col = label) %>%
#     add_edges_from_table(edges, from, to, id_external) %>%
#     add_global_graph_attrs("layout", "dot", "graph")
#
#   expect_dot(
#     graph,
#     c("^digraph", "\\{",
#       "graph", attrib_block,
#       "node", attrib_block,
#       "edge", attrib_block,
#       node(1, 'a'),
#       "subgraph cluster2\\{\nlabel='data'",
#       node(c(2,3), c('b', 'c')), "\\}",
#       edge(c(1, 2), c(3, 3)),
#       "\\}$")
#     )
# })

# Combining two graph objects into a single graph object

test_that("graphs can be combined", {

  # Create two simple graphs
  nodes_1 <- create_node_df(n = 10)

  edges_1 <-
    create_edge_df(
      from = 1:9,
      to = 2:10)

  graph_1 <-
    create_graph(
      nodes_df = nodes_1,
      edges_df = edges_1)

  nodes_2 <- create_node_df(n = 10)

  edges_2 <-
    create_edge_df(
      from = 1:9,
      to = 2:10)

  graph_2 <-
    create_graph(
      nodes_df = nodes_2,
      edges_df = edges_2)

  # Combine the two graphs
  combined_graph_1 <-
    combine_graphs(graph_1, graph_2)

  # Expect a graph object of class `dgr_graph`
  expect_s3_class(
    combined_graph_1, "dgr_graph")

  # Expect that names in graph object match a
  # prescribed set of names
  expect_in(
    names(combined_graph_1),
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  # Expect that the `global_attrs` graph component
  # is not NULL
  expect_false(is.null(combined_graph_1$global_attrs))

  # Expect that the `is_graph_directed()` function
  # will return TRUE
  expect_true(is_graph_directed(combined_graph_1))

  # Expect that the `nodes_df` df has 3 columns
  expect_equal(
    ncol(combined_graph_1$nodes_df), 3)

  # Expect that the `nodes_df` df has 20 rows
  expect_equal(
    nrow(combined_graph_1$nodes_df), 20)

  # Expect that the `edges_df` df has 4 columns
  expect_equal(
    ncol(combined_graph_1$edges_df), 4)

  # Expect that the `edges_df` df has 18 rows
  expect_equal(
    nrow(combined_graph_1$edges_df), 18)
})

# Creating and combining edge data frames

test_that("a correct edge data frame is generated", {

  # Create `edges_1` edge data frame
  edges_1 <-
    create_edge_df(
      from = c(1, 1, 2, 3),
      to = c(2, 4, 4, 1),
      rel = "requires",
      color = "green",
      data = c(2.7, 8.9, 2.6, 0.6))

  # Create `edges_2` edge data frame
  edges_2 <-
    create_edge_df(
      from = c(5, 7, 8, 8),
      to = c(7, 8, 6, 5),
      rel = "receives",
      arrowhead = "dot",
      color = "red")

  # Expect that data frames are generated
  expect_s3_class(edges_1, "data.frame")

  expect_s3_class(edges_2, "data.frame")

  # Expect that each of the edfs has 4 rows
  expect_equal(
    nrow(edges_1), 4)

  expect_equal(
    nrow(edges_2), 4)

  # Expect that each of the edfs has 6 columns
  expect_equal(
    ncol(edges_1), 6)

  expect_equal(
    ncol(edges_2), 6)

  # Create an edge data frame using a vector with
  # length > 1 and length < length(from | to)
  edges_var_1 <-
    create_edge_df(
      from = c(1, 1, 2, 3),
      to = c(2, 4, 4, 1),
      rel = "requires",
      color = c("green", "green"),
      data = c(2.7, 8.9, 2.6, 0.6))

  # Expect that a data frame is generated
  expect_s3_class(
    edges_var_1, "data.frame")

  # Expect that the data frame has 4 rows
  expect_equal(
    nrow(edges_var_1), 4)

  # Expect that the `color` attribute is only
  # written twice and not repeated down
  expect_equal(
    edges_var_1$color, c("green", "green", NA, NA))

  # Create an edge data frame using a vector with
  # length > length(from | to)
  edges_var_2 <-
    create_edge_df(
      from = c(1, 1, 2, 3),
      to = c(2, 4, 4, 1),
      rel = "requires",
      color = c("green", "green",
                "green", "green",
                "green", "green"),
      data = c(2.7, 8.9, 2.6, 0.6))

  # Expect that a data frame is generated
  expect_s3_class(
    edges_var_2, "data.frame")

  # Expect that the data frame has 4 rows
  expect_equal(
    nrow(edges_var_2), 4)

  # Expect that the 'color' attribute is repeated down
  expect_equal(
    edges_var_2$color,
    c("green", "green", "green", "green"))
})

test_that("two edge data frames can be successfully combined", {

  # Create `edges_1` edge data frame
  edges_1 <-
    create_edge_df(
      from = c(1, 1, 2, 3),
      to = c(2, 4, 4, 1),
      rel = "requires",
      color = "green",
      data = c(2.7, 8.9, 2.6, 0.6))

  # Create `edges_2` edge data frame
  edges_2 <-
    create_edge_df(
      from = c(5, 7, 8, 8),
      to = c(7, 8, 6, 5),
      rel = "receives",
      arrowhead = "dot",
      color = "red")

  # Combine the 2 edge data frames
  all_edges <-
    combine_edfs(edges_1, edges_2)

  # Expect that a data frame is generated
  expect_s3_class(
    all_edges, "data.frame")

  # Expect that the combined edf has 8 rows
  expect_equal(
    nrow(all_edges), 8)

  # Expect that the combined edf has 7 columns
  expect_equal(
    ncol(all_edges), 7)
})

test_that("three edge data frames can be successfully combined", {

  # Create `edges_1` edge data frame
  edges_1 <-
    create_edge_df(
      from = c(1, 1),
      to = c(2, 4),
      rel = "requires",
      color = "green",
      data = c(2.7, 8.9))

  # Create `edges_2` edge data frame
  edges_2 <-
    create_edge_df(
      from = c(2, 3),
      to = c(4, 1),
      rel = "requires",
      color = "green",
      data = c(2.6, 0.6))

  # Create `edges_3` edge data frame
  edges_3 <-
    create_edge_df(
      from = c(5, 7, 8, 8),
      to = c(7, 8, 6, 5),
      rel = "receives",
      arrowhead = "dot",
      color = "red")

  # Combine the 3 edge data frames
  all_edges <-
    combine_edfs(
      edges_1, edges_2, edges_3)

  # Expect that a data frame is generated
  expect_s3_class(
    all_edges, "data.frame")

  # Expect that the combined edf has 8 rows
  expect_equal(
    nrow(all_edges), 8)

  # Expect that the combined edf has 7 columns
  expect_equal(
    ncol(all_edges), 7)
})

# Adding node and/or edge data frames to an existing graph object

test_that("adding a node df to a graph is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 4,
      type = "letter",
      color = c("red", "green", "grey", "blue"),
      value = c(3.5, 2.6, 9.4, 2.7))

  # Add the node data frame to the graph object to
  # create a graph with nodes
  graph_2 <-
    add_node_df(
      graph = graph,
      node_df = nodes)

  # Create another node data frame
  nodes_2 <-
    create_node_df(
      n = 4,
      type = "letter",
      color = c("white", "brown", "aqua", "pink"),
      value = c(1.6, 6.4, 0.8, 4.2))

  # Add the second node data frame to the graph object
  # to add more nodes with attributes to the graph
  graph_3 <-
    add_node_df(
      graph = graph_2,
      node_df = nodes_2)

  # Expect that names in these graph objects match a
  # prescribed set of names
  expect_named(
    graph_2,
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  expect_named(
    graph_3,
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  # Expect graph objects of class `dgr_graph`
  expect_s3_class(graph_2, "dgr_graph")
  expect_s3_class(graph_3, "dgr_graph")

  # Expect that the `nodes_df` component is a data frame
  expect_s3_class(graph_2$nodes_df, "data.frame")
  expect_s3_class(graph_3$nodes_df, "data.frame")

  # Expect that the `nodes_df` data frame has 4 and 8 rows
  expect_equal(nrow(graph_2$nodes_df), 4)
  expect_equal(nrow(graph_3$nodes_df), 8)

  # Expect that the `nodes_df` data frame has 5 columns
  expect_equal(ncol(graph_2$nodes_df), 5)
  expect_equal(ncol(graph_3$nodes_df), 5)
})

test_that("adding an edge df to a graph is possible", {

  # Create an empty graph
  graph <- create_graph()

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 4,
      type = "letter",
      color = c("red", "green", "grey", "blue"),
      value = c(3.5, 2.6, 9.4, 2.7))

  # Add the node data frame to the graph object to create a
  # graph with nodes
  graph_2 <-
    add_node_df(
      graph = graph,
      node_df = nodes)

  # Create another node data frame
  nodes_2 <-
    create_node_df(
      n = 4,
      type = "letter",
      color = c("white", "brown", "aqua", "pink"),
      value = c(1.6, 6.4, 0.8, 4.2))

  # Add the second node data frame to the graph object to
  # add more nodes with attributes to the graph
  graph_3 <-
    add_node_df(
      graph = graph_2,
      node_df = nodes_2)

  # Create an edge data frame
  edges <-
    create_edge_df(
      from = c(1, 2, 3),
      to = c(4, 3, 1),
      rel = "leading_to")

  # Adding an edge df to an empty graph results in an error
  expect_error(
    add_edge_df(
      graph = create_graph(),
      edge_df = edges))

  # Add the edge data frame to the graph
  graph_3 <-
    add_edge_df(
      graph = graph_3,
      edge_df = edges)

  # Expect a graph object of class `dgr_graph`
  expect_s3_class(
    graph_3, "dgr_graph")

  # Expect that the `edges_df` component is a data frame
  expect_s3_class(
    graph_3$edges_df, "data.frame")

  # Expect that the `edges_df` data frame has 3 rows
  expect_equal(
    nrow(graph_3$edges_df), 3)

  # Expect that the `edges_df` data frame has 4 columns
  expect_equal(
    ncol(graph_3$edges_df), 4)

  # Add another edge to a graph that already has some
  # edges defined
  graph_3 <-
    add_edge_df(
      graph = graph_3,
      edge_df = create_edge_df(from = 2, to = 4))

  # Expect that the `edges_df` data frame has 4 rows
  expect_equal(
    nrow(graph_3$edges_df), 4)
})

# Creating a graph object

test_that("an empty graph object can be created and such an object is correct", {

  # Create an empty graph
  graph <- create_graph()

  # Expect that names in graph object match a
  # prescribed set of names
  expect_named(
    graph,
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  # Expect a graph object of class `dgr_graph`
  expect_s3_class(graph, "dgr_graph")

  # Expect that the `nodes_df` and  `edges_df` components are
  # a data frame
  expect_s3_class(graph$nodes_df, "data.frame")
  expect_s3_class(graph$edges_df, "data.frame")

  # Expect that the use of `is_graph_empty()` function
  # will result in TRUE
  expect_true(is_graph_empty(graph))

  # Expect that the `global_attrs` component is not NULL
  expect_false(is.null(graph$global_attrs))

  # Expect that the empty graph is directed by default
  expect_true(graph$directed)

  # Expect that the `is_graph_directed()` function
  # will return TRUE
  expect_true(is_graph_directed(graph))
})

test_that("a graph object with nodes can be created correctly", {

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 4,
      type = "lower",
      style = "filled",
      color = "aqua",
      shape = c("circle", "circle",
                "rectangle", "rectangle"),
      data = c(3.5, 2.6, 9.4, 2.7))

  # Create the graph object using the node data frame
  graph <- create_graph(nodes_df = nodes)

  # Expect that names in this graph object match a
  # prescribed set of names
  expect_in(
    names(graph),
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  # Expect a graph object of class `dgr_graph`
  expect_s3_class(
    graph, "dgr_graph")

  # Expect that several of the graph components
  # are not NULL
  expect_false(is.null(graph$nodes_df))

  expect_false(is.null(graph$global_attrs))

  # Expect that the `nodes_df` and `edges_df` component is
  # a data frame
  expect_s3_class(graph$nodes_df, "data.frame")
  expect_s3_class(graph$edges_df, "data.frame")

  # Expect that the graph is a directed graph
  expect_true(graph$directed)

  # Expect that the `nodes_df` df has 7 columns
  expect_equal(
    ncol(graph$nodes_df), 7)

  # Expect that the `nodes_df` df has 4 rows
  expect_equal(
    nrow(graph$nodes_df), 4)
})

test_that("a graph object with nodes and edges can be created correctly", {

  # Create a node data frame
  nodes <-
    create_node_df(
      n = 4,
      type = "lower",
      style = "filled",
      color = "aqua",
      shape = c("circle", "circle",
                "rectangle", "rectangle"),
      data = c(3.5, 2.6, 9.4, 2.7))

  # Create an edge data frame
  edges <-
    create_edge_df(
      from = c(1, 2, 3),
      to = c(4, 3, 1),
      rel = "leading_to")

  # Create the graph object using the node and
  # edge data frames
  graph <-
    create_graph(
      nodes_df = nodes,
      edges_df = edges)

  # Expect that names in this graph object match a
  # prescribed set of names
  expect_in(
    names(graph),
    c(
      "graph_info",
      "nodes_df",
      "edges_df",
      "global_attrs",
      "directed",
      "last_node",
      "last_edge",
      "node_selection",
      "edge_selection",
      "cache",
      "graph_actions",
      "graph_log"
    )
  )

  # Expect a graph object of class `dgr_graph`
  expect_s3_class(
    graph, "dgr_graph")

  # Expect that the `global_attrs` component is not NULL
  expect_false(is.null(graph$global_attrs))

  # Expect that the `nodes_df` component is
  # a data frame
  expect_s3_class(
    graph$nodes_df, "data.frame")

  # Expect that the `edges_df` component is
  # a data frame
  expect_s3_class(
    graph$edges_df, "data.frame")

  # Expect that the graph is a directed graph
  expect_true(graph$directed)

  # Expect that the `nodes_df` df has 7 columns
  expect_equal(
    ncol(graph$nodes_df), 7)

  # Expect that the `nodes_df` df has 4 rows
  expect_equal(
    nrow(graph$nodes_df), 4)

  # Expect that the `edges_df` df has 4 columns
  expect_equal(
    ncol(graph$edges_df), 4)

  # Expect that the `edges_df` df has 3 rows
  expect_equal(
    nrow(graph$edges_df), 3)
})

#' Get node eccentricities
#'
#' @description
#'
#' Get a data frame with node eccentricity values.
#'
#' @inheritParams render_graph
#' @param mode the mode with which the shortest paths to or from the given
#'   vertices should be calculated for directed graphs. If `out` (the
#'   default) then the shortest paths from the node, if `in` then only
#'   shortest paths to each node are considered. If `all` is used, then the
#'   corresponding undirected graph will be used and edge directions will be
#'   ignored. For undirected graphs, this argument is ignored.
#'
#' @return A data frame containing eccentricity values by node ID value.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the eccentricity values for
#' # all nodes in the graph
#' graph |> get_eccentricity()
#'
#' @export
get_eccentricity <- function(
    graph,
    mode = "out"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the eccentricity with the given mode
  eccentricity <-
    igraph::eccentricity(
      graph = ig_graph,
      mode = mode)

  # Create a data frame with node ID values
  # and eccentrity values
  data.frame(
    id = eccentricity |>
      names() |>
      as.integer(),
    eccentricity = eccentricity,
    stringsAsFactors = FALSE)
}

#' Get all nodes connected to a specified node
#'
#' @description
#'
#' With a single node serving as the starting point get all nodes connected
#' (i.e., reachable with a traversable path) to that node.
#'
#' @inheritParams render_graph
#' @param node a single-length vector containing a node ID value.
#'
#' @return A vector of node ID values.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function; it
#' # has an unconnected node (`6`)
#' graph_1 <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 20,
#'     m = 32,
#'     set_seed = 23)
#'
#' # There won't be any connected
#' # nodes to `6` so when specifying
#' # this node with `get_all_connected_nodes()`
#' # we get NA back
#' graph_1 %>%
#'   get_all_connected_nodes(
#'     node = 6)
#'
#' # Any other node in `graph_1` will
#' # provide a vector of all the nodes
#' # other than `6`
#' graph_1 %>%
#'   get_all_connected_nodes(
#'     node = 1)
#'
#' # The following graph has two
#' # clusters of nodes (i.e., the
#' # graph has two connected components)
#' graph_2 <-
#'   create_graph() %>%
#'   add_path(n = 6) %>%
#'   add_path(n = 4)
#'
#' # In `graph_2`, node `1` is in
#' # the larger of the two
#' # connected components
#' graph_2 %>%
#'   get_all_connected_nodes(
#'     node = 1)
#'
#' # Also in `graph_2`, node `8`
#' # is in the smaller of the two
#' # connected components
#' graph_2 %>%
#'   get_all_connected_nodes(
#'     node = 8)
#'
#' @export
get_all_connected_nodes <- function(
    graph,
    node
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Verify that the node ID provided is in the graph
  if (!(node %in% get_node_ids(graph))) {

    cli::cli_abort(
      "The node ID must be in the graph.")
  }

  # Get a data frame of the weakly-connected
  # components for the entire graph
  wc_components_df <- get_w_connected_cmpts(graph)

  # Get single-row data frame with the node's
  # `id` and `wc_component` grouping
  wc_component <-
    wc_components_df %>%
    dplyr::filter(id == node)

  # Extract the `wc_component` grouping as
  # a single-length vector
  wc_component_val <- wc_component$wc_component

  # Filter the table of weakly-connected
  # components for the nodes in the `wc_component`
  # grouping (and removing the original node as well)
  connected_nodes <-
    wc_components_df %>%
    dplyr::filter(id != node) %>%
    dplyr::filter(wc_component == wc_component_val)

  # If the resulting df has rows, get the node ID
  # values from each of the rows; otherwise, return NA
  if (nrow(connected_nodes) > 0) {
    connected_nodes <-
      connected_nodes$id
  } else {
    connected_nodes <- NA
  }

  connected_nodes
}

#' Select nodes in the graph based on their degree values
#'
#' @description
#'
#' Using a graph object of class `dgr_graph`, create a selection of nodes that
#' have certain degree values.
#'
#' @inheritParams render_graph
#' @param expressions One or more expressions for filtering nodes by degree
#'   values. Use a combination of degree type (`deg` for total degree, `indeg`
#'   for in-degree, and `outdeg` for out-degree) with a comparison operator and
#'   values for comparison (e.g., use `"deg >= 2"` to select nodes with a degree
#'   greater than or equal to 2).
#' @param set_op The set operation to perform upon consecutive selections of
#'   graph nodes. This can either be as a `union` (the default), as an
#'   intersection of selections with `intersect`, or, as a `difference` on the
#'   previous selection, if it exists.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using
#' # the `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 35, m = 125,
#'     set_seed = 23)
#'
#' # Report which nodes have a
#' # total degree (in-degree +
#' # out-degree) of exactly 9
#' graph |>
#'   select_nodes_by_degree(
#'     expressions = "deg == 9") |>
#'   get_selection()
#'
#' # Report which nodes have a
#' # total degree greater than or
#' # equal to 9
#' graph |>
#'   select_nodes_by_degree(
#'     expressions = "deg >= 9") |>
#'   get_selection()
#'
#' # Combine two calls of
#' # `select_nodes_by_degree()` to
#' # get those nodes with total
#' # degree less than 3 and total
#' # degree greater than 10 (by
#' # default, those `select...()`
#' # functions will `union` the
#' # sets of nodes selected)
#' graph |>
#'   select_nodes_by_degree(
#'     expressions = "deg < 3") |>
#'   select_nodes_by_degree(
#'     expressions = "deg > 10") |>
#'   get_selection()
#'
#' # Combine two calls of
#' # `select_nodes_by_degree()` to
#' # get those nodes with total
#' # degree greater than or equal
#' # to 3 and less than or equal
#' # to 10 (the key here is to
#' # `intersect` the sets of nodes
#' # selected in the second call)
#' graph |>
#'   select_nodes_by_degree(
#'     expressions = "deg >= 3") |>
#'   select_nodes_by_degree(
#'     expressions = "deg <= 10",
#'     set_op = "intersect") |>
#'   get_selection()
#'
#' # Select all nodes with an
#' # in-degree greater than 5, then,
#' # apply a node attribute to those
#' # selected nodes (coloring the
#' # selected nodes red)
#' graph_2 <-
#'   graph |>
#'   select_nodes_by_degree(
#'     expressions = "indeg > 5") |>
#'   set_node_attrs_ws(
#'     node_attr = color,
#'     value = "red")
#'
#' # Get the selection of nodes
#' graph_2 |> get_selection()
#'
#' @export
select_nodes_by_degree <- function(
    graph,
    expressions,
    set_op = "union"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # Get a data frame with node ID and degree types
  node_degree <-
    get_node_info(graph) |>
    dplyr::select("id", "deg", "indeg", "outdeg") |>
    dplyr::filter(!!!parse_exprs(expressions))

  # Get the node ID values from the filtered table
  nodes_selected <- node_degree$id

  # If no node ID values in `nodes_selected` return
  # the graph without a changed node selection
  if (length(nodes_selected) == 0) {
    return(graph)
  }

  # Obtain vector with node ID selection of nodes
  # already present
  nodes_prev_selection <- graph$node_selection$node

  # Incorporate selected nodes into graph's
  # selection section
  if (set_op == "union") {
    nodes_combined <-
      union(nodes_prev_selection, nodes_selected)
  } else if (set_op == "intersect") {
    nodes_combined <-
      intersect(nodes_prev_selection, nodes_selected)
  } else if (set_op == "difference") {
    nodes_combined <-
      base::setdiff(nodes_prev_selection, nodes_selected)
  }

  # Add the node ID values to the active selection
  # of nodes in `graph$node_selection`
  graph$node_selection <-
    replace_graph_node_selection(
      graph = graph,
      replacement = nodes_combined)

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Obtain the output graph's node and edge
  # selection properties
  n_e_select_properties_out <-
    node_edge_selection_properties(graph = graph)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # Emit a message about the modification of a selection
  # if that option is set
  if (!is.null(graph$graph_info$display_msgs) &&
      graph$graph_info$display_msgs) {

    # Construct message body
    if (!n_e_select_properties_in[["node_selection_available"]] &&
        !n_e_select_properties_in[["edge_selection_available"]]) {

      msg_body <-
        glue::glue(
          "created a new selection of \\
        {n_e_select_properties_out[['selection_count_str']]}")

    } else if (n_e_select_properties_in[["node_selection_available"]] ||
               n_e_select_properties_in[["edge_selection_available"]]) {

      if (n_e_select_properties_in[["edge_selection_available"]]) {
        msg_body <-
          glue::glue(
            "modified an existing selection of\\
           {n_e_select_properties_in[['selection_count_str']]}:
           * {n_e_select_properties_out[['selection_count_str']]}\\
           are now in the active selection
           * used the `{set_op}` set operation")
      }

      if (n_e_select_properties_in[["node_selection_available"]]) {
        msg_body <-
          glue::glue(
            "created a new selection of\\
           {n_e_select_properties_out[['selection_count_str']]}:
           * this replaces\\
           {n_e_select_properties_in[['selection_count_str']]}\\
           in the prior selection")
      }
    }

    # Issue a message to the user
    emit_message(
      fcn_name = fcn_name,
      message_body = msg_body)
  }

  graph
}

#' Select nodes in a graph by their ID values
#'
#' @description
#'
#' Select nodes in a graph object of class `dgr_graph` by their node ID values.
#' If nodes have IDs that are monotonically increasing integer values, then
#' numeric ranges can be used for the selection.
#'
#' @inheritParams render_graph
#' @param nodes A vector of node IDs for the selection of nodes present in the
#'   graph.
#' @param set_op The set operation to perform upon consecutive selections of
#'   graph nodes. This can either be as a `union` (the default), as an
#'   intersection of selections with `intersect`, or, as a `difference` on the
#'   previous selection, if it exists.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <- create_node_df(n = 10)
#'
#' # Create a graph
#' graph <-
#'   create_graph(
#'     nodes_df = ndf)
#'
#' # Select nodes `1` to `5` and show that
#' # selection of nodes with `get_selection()`
#' graph %>%
#'   select_nodes_by_id(nodes = 1:5) %>%
#'   get_selection()
#'
#' @export
select_nodes_by_id <- function(
    graph,
    nodes,
    set_op = "union"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get a vector of node ID values from the graph
  nodes_in_graph <- get_node_ids(graph)

  # Stop function if any nodes specified are not
  # in the graph
  if (!all(nodes %in% nodes_in_graph)) {

    cli::cli_abort(
      "One of more of the nodes specified are not available in the graph.")
  }

  # Obtain vector with node ID selection of nodes
  # already present
  nodes_prev_selection <- graph$node_selection$node

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # Incorporate selected nodes into graph's
  # selection
  if (set_op == "union") {
    nodes_combined <-
      union(nodes_prev_selection, nodes)
  } else if (set_op == "intersect") {
    nodes_combined <-
      intersect(nodes_prev_selection, nodes)
  } else if (set_op == "difference") {
    nodes_combined <-
      base::setdiff(nodes_prev_selection, nodes)
  }

  # Add the node ID values to the active selection
  # of nodes in `graph$node_selection`
  graph$node_selection <-
    replace_graph_node_selection(
      graph = graph,
      replacement = nodes_combined)

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Obtain the output graph's node and edge
  # selection properties
  n_e_select_properties_out <-
    node_edge_selection_properties(graph = graph)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # Emit a message about the modification of a selection
  # if that option is set
  if (!is.null(graph$graph_info$display_msgs) &&
      graph$graph_info$display_msgs) {

    # Construct message body
    if (!n_e_select_properties_in[["node_selection_available"]] &&
        !n_e_select_properties_in[["edge_selection_available"]]) {

      msg_body <-
        glue::glue(
          "created a new selection of \\
        {n_e_select_properties_out[['selection_count_str']]}")

    } else if (n_e_select_properties_in[["node_selection_available"]] ||
               n_e_select_properties_in[["edge_selection_available"]]) {

      if (n_e_select_properties_in[["node_selection_available"]]) {
        msg_body <-
          glue::glue(
            "modified an existing selection of\\
           {n_e_select_properties_in[['selection_count_str']]}:
           * {n_e_select_properties_out[['selection_count_str']]}\\
           are now in the active selection
           * used the `{set_op}` set operation")
      }

      if (n_e_select_properties_in[["edge_selection_available"]]) {
        msg_body <-
          glue::glue(
            "created a new selection of\\
           {n_e_select_properties_out[['selection_count_str']]}:
           * this replaces\\
           {n_e_select_properties_in[['selection_count_str']]}\\
           in the prior selection")
      }
    }

    # Issue a message to the user
    emit_message(
      fcn_name = fcn_name,
      message_body = msg_body)
  }

  graph
}

#' Add an edge between nodes in a graph object
#'
#' @description
#'
#' With a graph object of class `dgr_graph`, add an edge to nodes within the
#' graph.
#'
#' @inheritParams node_edge_aes_data
#' @inheritParams render_graph
#' @param from The outgoing node from which the edge is connected. There is the
#'   option to use a node `label` value here (and this must correspondingly also
#'   be done for the `to` argument) for defining node connections. Note that
#'   this is only possible if all nodes have distinct `label` values set and
#'   none exist as an empty string.
#' @param to The incoming nodes to which each edge is connected. There is the
#'   option to use a node `label` value here (and this must correspondingly also
#'   be done for the `from` argument) for defining node connections. Note that
#'   this is only possible if all nodes have distinct `label` values set and
#'   none exist as an empty string.
#' @param rel An optional string specifying the relationship between the
#'   connected nodes.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 4 nodes
#' graph <-
#'   create_graph() %>%
#'   add_node(label = "one") %>%
#'   add_node(label = "two") %>%
#'   add_node(label = "three") %>%
#'   add_node(label = "four")
#'
#' # Add an edge between those
#' # nodes and attach a
#' # relationship to the edge
#' graph <-
#'  add_edge(
#'    graph,
#'    from = 1,
#'    to = 2,
#'    rel = "A")
#'
#' # Use the `get_edge_info()`
#' # function to verify that
#' # the edge has been created
#' graph %>%
#'   get_edge_info()
#'
#' # Add another node and
#' # edge to the graph
#' graph <-
#'   graph %>%
#'   add_edge(
#'     from = 3,
#'     to = 2,
#'     rel = "A")
#'
#' # Verify that the edge
#' # has been created by
#' # counting graph edges
#' graph %>% count_edges()
#'
#' # Add edges by specifying
#' # node `label` values; note
#' # that all nodes must have
#' # unique `label` values to
#' # use this option
#' graph <-
#'   graph %>%
#'   add_edge(
#'     from = "three",
#'     to = "four",
#'     rel = "L") %>%
#'   add_edge(
#'     from = "four",
#'     to = "one",
#'     rel = "L")
#'
#' # Use `get_edges()` to verify
#' # that the edges were added
#' graph %>% get_edges()
#'
#' # Add edge aesthetic and data
#' # attributes during edge creation
#' graph_2 <-
#'   create_graph() %>%
#'   add_n_nodes(n = 2) %>%
#'   add_edge(
#'     from = 1,
#'     to = 2,
#'     rel = "M",
#'     edge_aes = edge_aes(
#'       penwidth = 1.5,
#'       color = "blue"),
#'     edge_data = edge_data(
#'       value = 4.3))
#'
#' # Use the `get_edges()` function
#' # to verify that the attribute
#' # values were bound to the
#' # newly created edge
#' graph_2 %>% get_edge_df()
#'
#'
#' @family edge creation and removal
#'
#' @export
add_edge <- function(
    graph,
    from,
    to,
    rel = NULL,
    edge_aes = NULL,
    edge_data = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, extra_msg = "An edge cannot be added.")

  if (length(from) > 1 || length(to) > 1) {

    rlang::abort("Only one edge can be specified in `from` or `to`")
  }

  # Get the value for the latest `version_id` for
  # graph (in the `graph_log`)
  current_graph_log_version_id <-
    max(graph$graph_log$version_id)

  rel <- rel %||% NA_character_

  # Collect edge aesthetic attributes
  if (!is.null(edge_aes)) {

    edge_aes_tbl <- dplyr::as_tibble(edge_aes)

    if (nrow(edge_aes_tbl) == 1) {

      edge_aes$index__ <- 1

      edge_aes_tbl <-
        dplyr::as_tibble(edge_aes) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_aes_tbl)) {
      edge_aes_tbl$id <- NULL
    }
  }

  # Collect edge data attributes
  if (!is.null(edge_data)) {

    edge_data_tbl <- dplyr::as_tibble(edge_data)

    if (nrow(edge_data_tbl) == 1) {

      edge_data$index__ <- 1

      edge_data_tbl <-
        dplyr::as_tibble(edge_data) %>%
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_data_tbl)) {
      edge_data_tbl$id <- NULL
    }
  }

  # If `from` and `to` values provided as character
  # values, assume that these values refer to node
  # `label` attr values
  if (is.character(from) && is.character(to)) {

    # Stop function if the label for
    # `from` does not exist in the graph
    if (!(from %in% graph$nodes_df$label)) {

      rlang::abort(
        "The value provided in `from` does not exist as a node `label` value.")
    }

    # Stop function if the label for
    # `from` is not distinct in the graph
    if (graph$nodes_df %>%
        dplyr::select("label") %>%
        dplyr::filter(label == from) %>%
        nrow() > 1) {

      rlang::abort(
         "The node `label` provided in `from` is not distinct in the graph.")
    }

    # Stop function if the label for
    # `to` does not exist in the graph
    if (!(to %in% graph$nodes_df$label)) {

      rlang::abort(
        "The value provided in `to` does not exist as a node `label` value.")
    }

    # Stop function if the label for
    # `to` is not distinct in the graph
    if (graph$nodes_df %>%
        dplyr::select("label") %>%
        dplyr::filter(label == to) %>%
        nrow() > 1) {

      rlang::abort(
        "The node `label` provided in `to` is not distinct in the graph.")
    }

    # Use the `translate_to_node_id()` helper function to map
    # node `label` values to node `id` values
    from_to_node_id <-
      translate_to_node_id(
        graph = graph,
        from = from,
        to = to)

    from <- from_to_node_id$from
    to <- from_to_node_id$to
  }

  # Use `bind_rows()` to add an edge
  if (!is.null(graph$edges_df)) {

    combined_edges <-
      dplyr::bind_rows(
        graph$edges_df,
        data.frame(
          id = as.integer(graph$last_edge + 1),
          from = as.integer(from),
          to = as.integer(to),
          rel = as.character(rel),
          stringsAsFactors = FALSE))

    # Use the `combined_edges` object as a
    # replacement for the graph's internal
    # edge data frame
    graph$edges_df <- combined_edges

    if (exists("edge_aes_tbl")) {

      # If extra edge attributes available, add
      # those to the new edge
      graph <-
        suppressMessages(
          graph %>%
            select_edges_by_edge_id(
              edges = graph$edges_df$id %>% max())
        )

      # Iteratively set edge attribute values for
      # the new edge in the graph
      for (i in seq_len(ncol(edge_aes_tbl))) {
        graph <-
          graph %>%
          set_edge_attrs_ws(
            edge_attr = !!colnames(edge_aes_tbl)[i],
            value = edge_aes_tbl[1, i][[1]]
          )
      }

      # Clear the graph's active selection
      graph <-
        suppressMessages(
          clear_selection(graph))
    }

    if (exists("edge_data_tbl")) {

      # If extra edge attributes available, add
      # those to the new edge
      graph <-
        suppressMessages(
          graph %>%
            select_edges_by_edge_id(
              edges = max(graph$edges_df$id))
        )

      # Iteratively set edge attribute values for
      # the new edge in the graph
      for (i in seq_len(ncol(edge_data_tbl))) {
        graph <-
          graph %>%
          set_edge_attrs_ws(
            edge_attr = !!colnames(edge_data_tbl)[i],
            value = edge_data_tbl[1, i][[1]]
          )
      }

      # Clear the graph's active selection
      graph <-
        suppressMessages(
          clear_selection(graph))
    }

    # Modify the `last_edge` vector
    graph$last_edge <- as.integer(graph$last_edge + 1L)

    # Remove extra items from the `graph_log`
    graph$graph_log <-
      graph$graph_log %>%
      dplyr::filter(version_id <= current_graph_log_version_id)

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph$graph_log <-
      add_action_to_log(
        graph_log = graph$graph_log,
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df),
        d_e = 1)

    # Perform graph actions, if any are available
    if (nrow(graph$graph_actions) > 0) {
      graph <-
        trigger_graph_actions(graph)
    }

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }

    graph
  }
}

# Get graph properties

test_that("Getting the graph object log is possible", {

  # Create a random graph and add
  # an edge as well
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23) %>%
    add_edge(
      from = 5,
      to = 10)

  # Get the graph's log data
  graph_log <- get_graph_log(graph)

  # Expect that the graph log is provided
  # as a tibble object
  expect_s3_class(
    graph_log, "tbl_df")

  # Expect certain column names for the
  # `graph_log` object
  expect_identical(
    colnames(graph_log),
    c("version_id", "function_used", "time_modified",
      "duration", "nodes", "edges", "d_n", "d_e"))

  # Expect 3 rows in this graph log
  expect_equal(
    nrow(graph_log), 3)
})

#' Get all neighbors of one or more nodes
#'
#' @description
#'
#' With one or more nodes, get the set of all neighboring nodes.
#'
#' @inheritParams render_graph
#' @param nodes A vector of node ID values.
#'
#' @return A vector of node ID values.
#'
#' @examples
#' # Create a simple, directed graph with 5
#' # nodes and 4 edges
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 5)
#'
#' # Find all neighbor nodes for node `2`
#' graph %>% get_nbrs(nodes = 2)
#'
#' # Find all neighbor nodes for nodes `1`
#' # and `5`
#' graph %>% get_nbrs(nodes = c(1, 5))
#'
#' # Color node `3` with purple, get its
#' # neighbors and color those nodes green
#' graph <-
#'   graph %>%
#'   select_nodes_by_id(nodes = 3) %>%
#'   set_node_attrs_ws(
#'     node_attr = color,
#'     value = "purple") %>%
#'   clear_selection() %>%
#'   select_nodes_by_id(
#'     nodes = get_nbrs(
#'       graph = .,
#'       nodes = 3)) %>%
#'   set_node_attrs_ws(
#'     node_attr = color,
#'     value = "green")
#'
#' @export
get_nbrs <- function(
    graph,
    nodes
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get predecessors and successors for all nodes
  # in `nodes`
  for (i in seq_along(nodes)) {
    if (i == 1) {
      node_nbrs <- vector(mode = "numeric")
    }
    node_nbrs <-
      c(node_nbrs,
        c(get_predecessors(graph, node = nodes[i]),
          get_successors(graph, node = nodes[i])))
  }

  # Get a unique set of node ID values
  node_nbrs <- sort(unique(node_nbrs))

  # If there are no neighbors, then return NA
  if (length(node_nbrs) == 0) {
    return(NA)
  } else {
    return(node_nbrs)
  }
}

#' Razor-like template for diagram specification
#'
#' @description
#'
#' Use Razor-like syntax to define a template for use in a `grViz` diagram.
#'
#' @param spec String spec to be parsed and evaluated.
#' @inheritParams grViz
#' @examples
#' \dontrun{
#' # a simple example to use a LETTER as a node label
#' spec <- "
#'   digraph { '@@1' }
#'
#'  [1]: LETTERS[1]
#' "
#' grViz(replace_in_spec(spec))
#'
#'
#' spec <- "
#' digraph a_nice_graph {
#' node [fontname = Helvetica]
#' a [label = '@@1']
#' b [label = '@@2-1']
#' c [label = '@@2-2']
#' d [label = '@@2-3']
#' e [label = '@@2-4']
#' f [label = '@@2-5']
#' g [label = '@@2-6']
#' h [label = '@@2-7']
#' i [label = '@@2-8']
#' j [label = '@@2-9']
#' a -> { b c d e f g h i j}
#' }
#'
#' [1]: 'top'
#' [2]: 10:20
#' "
#' grViz(replace_in_spec(spec))
#' }
#'
#' @export
replace_in_spec <- function(spec, envir = parent.frame()) {

  # Directive for marking subscripted text in a label or tooltip '@_'
  if (grepl("@_", spec)) {

    spec <- gsub('(label|tooltip)[ ]*=[ ]*\'(.*?)@_\\{(.*?)\\}(.*?)\'',
                 '\\1 = <\\2<FONT POINT-SIZE=\'8\'><SUB>\\3</SUB></FONT>\\4>',
                 spec, perl = TRUE)
  }

  # Directive for marking superscripted text in a label or tooltip '@_'
  if (grepl("@\\^", spec)) {

    spec <- gsub('(label|tooltip)[ ]*=[ ]*\'(.*?)@\\^\\{(.*?)\\}(.*?)\'',
                 '\\1 = <\\2<FONT POINT-SIZE=\'8\'><SUP>\\3</SUP></FONT>\\4>',
                 spec, perl = TRUE)
  }

  # Make a second pass to add subscripts as inline HTML
  while (grepl('(label|tooltip)[ ]*=[ ]*<(.*?)@_\\{(.+?)\\}(.*?)>', spec)) {

    spec <- gsub('(label|tooltip)[ ]*=[ ]*<(.*?)@_\\{(.*?)\\}(.*?)>',
                 '\\1 = <\\2<FONT POINT-SIZE=\'8\'><SUB>\\3</SUB></FONT>\\4>',
                 spec, perl = TRUE)
  }

  # Make a second pass to add superscripts as inline HTML
  while (grepl('(label|tooltip)[ ]*=[ ]*<(.*?)@\\^\\{(.+?)\\}(.*?)>', spec)) {

    spec <- gsub('(label|tooltip)[ ]*=[ ]*<(.*?)@\\^\\{(.*?)\\}(.*?)>',
                 '\\1 = <\\2<FONT POINT-SIZE=\'8\'><SUP>\\3</SUP></FONT>\\4>',
                 spec, perl = TRUE)
  }

  # Directive for substitution of arbitrary specification text '@@'
  if (grepl("@@", spec)) {

    # Extract the spec into several pieces: first being the body,
    # subsequent pieces belonging the replacement references
    spec_body <- unlist(strsplit(x = spec, "\\n\\s*\\[1\\]:"))[1]

    spec_references <-
      paste0("[1]:", unlist(strsplit(x = spec, "\\n\\s*\\[1\\]:"))[2])

    # Split the references into a vector of R statements
    split_references <-
      gsub("\\[[0-9]+\\]:[ ]?", "", unlist(strsplit(x = spec_references, "\\n")))

    # Evaluate the expressions and save into a list object
    for (i in seq_along(split_references)) {

      if (i == 1) {
        eval_expressions <- list()
      }

      eval_expressions <-
        c(
          eval_expressions,
          list(eval(parse(text = split_references[i]), envir = envir))
        )
    }

    # Make replacements to the spec body for each replacement that has no hyphen
    for (i in seq_along(split_references)) {
      while (grepl(paste0("@@", i, "([^-0-9])"), spec_body)) {
        spec_body <-
          gsub(paste0("@@", i, "(?=[^-0-9])"), eval_expressions[[i]][1], spec_body, perl = TRUE)
      }
    }

    # If the replacement has a hyphen, then obtain the digit(s) immediately
    # following and return the value from that index
    for (i in seq_along(split_references)) {
      while (grepl(paste0("@@", i, "-", "[0-9]+"), spec_body)) {

        the_index <-
          gsub(
            "^([0-9]+)(.*)", "\\1",
            strsplit(spec_body, paste0("@@", i, "-"))[[1]][2]
          ) %>%
          as.numeric()

        if (the_index > length(eval_expressions[[i]])) {

          spec_body <-
            gsub(
              paste0("@@", i, "-", the_index, "([^0-9])"),
              paste0(eval_expressions[[i]][length(eval_expressions[[i]])], "\\1"),
              spec_body
            )

        } else {

          spec_body <-
            gsub(
              paste0("@@", i, "-", the_index, "([^0-9])"),
              paste0(eval_expressions[[i]][the_index], "\\1"),
              spec_body
            )
        }
      }
    }

    # Return the updated spec with replacements evaluated
    return(spec_body)
  }

  if (!grepl("@@", spec)) {
    return(spec)
  }
}

# Graphviz specifications

test_that("the specification tools are functional", {

  # Create a simple Graphviz graph specification and perform
  # a substitution
  spec <- "
   digraph { '@@1' }

  [1]: LETTERS[1]
  "

  # Create a 'replaced_spec' grViz object
  replaced_spec <- grViz(replace_in_spec(spec))

  # Expect that the object inherits from 'grViz' and 'htmlwidget'
  expect_s3_class(
    replaced_spec, c("grViz", "htmlwidget"))

  # Create a simple Graphviz graph specification and perform
  # substitution using a vector
  spec <- "
 digraph a_nice_graph {
 node [fontname = Helvetica]
 a [label = '@@1']
 b [label = '@@2-1']
 c [label = '@@2-2']
 d [label = '@@2-3']
 e [label = '@@2-4']
 f [label = '@@2-5']
 g [label = '@@2-6']
 h [label = '@@2-7']
 i [label = '@@2-8']
 j [label = '@@2-9']
 a -> { b c d e f g h i j}
 }

 [1]: 'top'
 [2]: 10:20
 "
  # Create a 'replaced_spec' grViz object
  replaced_spec <- grViz(replace_in_spec(spec))

  # Expect that the object inherits from 'grViz' and 'htmlwidget'
  expect_s3_class(
    replaced_spec, c("grViz", "htmlwidget"))
})

test_that("specification substitution works without quotes", {

  spec <- "digraph { @@1 }
   [1]: LETTERS[1]"
  replaced_spec <- replace_in_spec(spec)
  expect_equal(replaced_spec, "digraph { A }")

})

#' Get radiality centrality scores
#'
#' @description
#'
#' Get the radiality centrality for all nodes in a graph. These scores describe
#' the ease to which nodes can reach other nodes.
#'
#' @inheritParams render_graph
#' @param direction Using `all` (the default), the search will ignore edge
#'   direction while traversing through the graph. With `out`, measurements of
#'   paths will be from a node whereas with `in`, measurements of paths will be
#'   to a node.
#'
#' @return A data frame with radiality centrality scores for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the radiality scores for nodes in the graph
#' graph |>
#'   get_radiality()
#'
#' # Add the radiality values
#' # to the graph as a node
#' # attribute
#' graph <-
#'   graph |>
#'   join_node_attrs(
#'     df = get_radiality(graph))
#'
#' # Display the graph's node data frame
#' graph |>
#'   get_node_df()
#'
#' @export
get_radiality <- function(
    graph,
    direction = "all"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation Ensure that values provided for the
  # `direction` argument are from the
  # valid options
  rlang::arg_match0(direction, c("all", "in", "out"))

  # Get the number of nodes in the graph
  n_nodes <- count_nodes(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get a matrix of shortest paths between all
  # pairs of nodes in the graph
  sp_matrix <-
    igraph::distances(
      graph = ig_graph,
      mode = direction,
      weights = NA)

  # Get the graph diameter
  is_directed <- direction != "all"

  diam <-
    igraph::diameter(
      graph = ig_graph,
      directed = is_directed)

  # Get the radiality values for all
  # nodes in the graph
  radiality_values <-
    apply(
      X = sp_matrix,
      MARGIN = 1,
      FUN = function(x) {
        if (all(is.infinite(x))) {
          return(0)
        } else {
          return(sum(diam + 1 - x[!is.infinite(x)])/(n_nodes - 1))
        }
      })

  # Create df with radiality scores
  data.frame(
    id = radiality_values |>
      names() |>
      as.integer(),
    radiality = radiality_values |> round(4),
    stringsAsFactors = FALSE)
}

#' Set edge attribute values
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, set edge attribute values for one
#' or more edges.
#'
#' @inheritParams render_graph
#' @param edge_attr The name of the attribute to set. Some examples are located in [edge_aes()]
#' @param values The values to be set for the chosen attribute for the chosen
#'   edges.
#' @param from An optional vector of node IDs from which the edge is outgoing
#'   for filtering list of nodes with outgoing edges in the graph.
#' @param to An optional vector of node IDs from which the edge is incoming for
#'   filtering list of nodes with incoming edges in the graph.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a simple graph
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = "basic",
#'     label = TRUE,
#'     value = c(3.5, 2.6, 9.4, 2.7))
#'
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "leading_to")
#'
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Set attribute `color = "green"`
#' # for edges `1`->`4` and `3`->`1`
#' # in the graph
#' graph <-
#'   graph |>
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "green",
#'     from = c(1, 3),
#'     to = c(4, 1))
#'
#' # Set attribute `color = "blue"`
#' # for all edges in the graph
#' graph <-
#'   graph |>
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "blue")
#'
#' # Set attribute `color = "pink"`
#' # for all edges in graph outbound
#' # from node with ID value `1`
#' graph <-
#'   graph |>
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "pink",
#'     from = 1)
#'
#' # Set attribute `color = "black"`
#' # for all edges in graph inbound
#' # to node with ID `1`
#' graph <-
#'   graph |>
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "black",
#'     to = 1)
#'
#' @family edge creation and removal
#'
#' @export
set_edge_attrs <- function(
    graph,
    edge_attr,
    values,
    from = NULL,
    to = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Get the requested `edge_attr`
  edge_attr <-
    rlang::enquo(edge_attr) |> rlang::get_expr() |> as.character()

  if (edge_attr %in% c("id", "from", "to")) {

    cli::cli_abort(
      "You cannot alter edge ID values or attributes associated with node IDs.")
  }

  if (!is.null(from) && !is.null(to) && (length(from) != length(to))) {

    cli::cli_abort(
      "The number of values specified in `from` and `to` must be the same.")
  }

  # Extract the graph's edf
  edf <- graph$edges_df

  if (length(values) != 1 &&
      length(values) != nrow(edf)) {

    cli::cli_abort(
      "The length of values provided must either be 1 or that of the number of rows in the edf.")
  }

  # Get the indices for the edge data frame
  # that require modification
  if (is.null(from) && !is.null(to)) {
    indices <-
      which(edf$to %in% to)
  } else if (!is.null(from) && is.null(to)) {
    indices <-
      which(edf$from %in% from)
  } else if (is.null(from) && is.null(to)) {
    indices <- seq_len(nrow(edf))
  } else {
    indices <-
      which((edf$from %in% from) &
              (edf$to %in% to))
  }

  # Apply single value to all target edges
  if (length(values) == 1) {

    if (edge_attr %in% colnames(edf)) {

      edf[indices,
          which(colnames(edf) %in%
                  edge_attr)] <- values
    }

    if (!(edge_attr %in% colnames(edf))) {

      # Add a new column and map the edge attribute
      # value to each of the indices in `edges_df`
      edf <-
        dplyr::mutate(
          edf,
          new_attr__ = ifelse(as.numeric(row.names(edf)) %in%
                                indices, values, NA))

      colnames(edf)[ncol(edf)] <- edge_attr
    }
  }

  if (length(values) == nrow(edf)) {

    if (edge_attr %in% colnames(edf)) {
      edf[, which(colnames(edf) %in%
                    edge_attr)] <- values
    }

    if (!(edge_attr %in% colnames(edf))) {
      edf <-
        cbind(edf,
              rep(NA_character_, nrow(edf)))

      edf[, ncol(edf)] <-
        edf[, ncol(edf)]

      colnames(edf)[ncol(edf)] <- edge_attr

      edf[, ncol(edf)] <- values
    }
  }

  # Update the graph object
  graph$edges_df <- edf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Determine whether a specified node is present
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, determine whether a specified node
#' is present.
#'
#' @inheritParams render_graph
#' @param node Either a node ID value or a node label to test for presence in
#'   the graph.
#'
#' @return A logical value.
#'
#' @examples
#' # Create a simple graph with
#' # a path of four nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(
#'     n = 4,
#'     type = "path",
#'     label = c(
#'       "one", "two",
#'       "three", "four"))
#'
#' # Determine if there is a node
#' # with ID `1` in the graph
#' graph %>%
#'   is_node_present(node = 1)
#'
#' # Determine if there is a node
#' # with ID `5` in the graph
#' graph %>%
#'   is_node_present(node = 5)
#'
#' # Determine if there is a node
#' # with label `two` in the graph
#' graph %>%
#'   is_node_present(node = "two")
#' @export
is_node_present <- function(
    graph,
    node
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Stop function if `node` not a single value
  if (length(node) != 1) {

    cli::cli_abort(
      "Only a single node can be queried using `is_node_present()`")
  }

  if (inherits(node, "character")) {

    # Determine whether the label value
    # corresponds to a label in the graph
    node_is_present <- node %in% graph$nodes_df$label

  } else if (inherits(node, "numeric")) {

    # Determine whether the node ID value
    # corresponds to a node ID in the graph
    node_is_present <- node %in% get_node_ids(graph)
  }

  node_is_present
}

#' Get nodes that form the graph periphery
#'
#' @description
#'
#' Get those nodes that are part of the graph periphery (i.e., have the maximum
#' eccentricity in the graph).
#'
#' @inheritParams render_graph
#'
#' @return A vector of node IDs.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function and
#' # get the nodes in the graph periphery
#' create_graph() %>%
#'   add_gnm_graph(
#'     n = 28,
#'     m = 35,
#'     set_seed = 23) %>%
#'   get_periphery()
#'
#' @export
get_periphery <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the eccentricity for each of the graph's nodes
  eccentricity <- get_eccentricity(graph)

  # Get the node ID values for all nodes where the
  # eccentricity is equal to the graph diameter
  # (i.e., maximum eccentricity)
  eccentricity %>%
    dplyr::filter(eccentricity == get_max_eccentricity(graph)) %>%
    dplyr::pull("id") %>%
    as.integer()
}

#' Get the graph date-time or timezone
#'
#' @description
#'
#' Get the time and timezone for a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#'
#' @return A single-length `POSIXct` vector with the assigned graph time.
#'
#' @examples
#' # Create an empty graph and
#' # set the graph's time; if nothing
#' # is supplied for the `tz` argument,
#' # `GMT` is used as the time zone
#' graph <-
#'   create_graph() %>%
#'     set_graph_time(
#'       time = "2015-10-25 15:23:00")
#'
#' # Get the graph's time as a POSIXct
#' # object using `get_graph_time()`
#' graph %>% get_graph_time()
#'
#' @export
get_graph_time <- function(graph) {
  check_graph_valid(graph)
  graph$graph_info$graph_time
}

#' Create a graph object
#'
#' @description
#'
#' Generates a graph object with the option to use node data frames (ndfs)
#' and/or edge data frames (edfs) to populate the initial graph.
#'
#' @param nodes_df An optional data frame containing, at minimum, a column
#'   (called `id`) which contains node IDs for the graph. Additional columns
#'   (node attributes) can be included with values for the named node attribute.
#' @param edges_df An optional data frame containing, at minimum, two columns
#'   (called `from` and `to`) where node IDs are provided. Additional columns
#'   (edge attributes) can be included with values for the named edge attribute.
#' @param directed With `TRUE` (the default) or `FALSE`, either directed or
#'   undirected edge operations will be generated, respectively.
#' @param graph_name An optional string for labeling the graph object.
#' @param attr_theme The theme (i.e., collection of `graph`, `node`, and `edge`
#'   global graph attributes) to use for this graph. The default theme is called
#'   `default`; there are hierarchical layout themes called `lr`, `tb`, `rl`,
#'   and `bt` (these operate from left-to-right, top-to-bottom, right-to-left,
#'   and bottom-to-top); and, for larger graphs, the `fdp` theme provides a
#'   force directed layout. If this is set to `NULL` then no global graph
#'   attributes will be applied to the graph upon creation.
#' @param write_backups An option to write incremental backups of changing graph
#'   states to disk. If `TRUE`, a subdirectory within the working directory will
#'   be created and used to store `RDS` files. The default value is `FALSE` so
#'   one has to opt in to use this functionality.
#' @param display_msgs An option to display messages primarily concerned with
#'   changes in graph selections. By default, this is `FALSE`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # With `create_graph()` we can
#' # simply create an empty graph (and
#' # add in nodes and edges later
#' # with other functions)
#' graph <- create_graph()
#'
#' # A graph can be created with
#' # nodes and without having any edges;
#' # this can be done in 2 steps:
#' # 1. create a node data frame (ndf)
#' #    using `create_node_df()`
#' ndf <-
#'   create_node_df(n = 4)
#'
#' # 2. create a new graph object with
#' #    `create_graph()` and then pass
#' #    in the ndf to `nodes_df`
#' graph <-
#'   create_graph(
#'     nodes_df = ndf)
#'
#' # Get information on the graph's nodes
#' graph |>
#'   get_node_info()
#'
#' # You can create a similar graph with
#' # just nodes but also providing a
#' # range of attributes for the nodes
#' # (e.g., types, labels, or arbitrary
#' # 'values')
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     label = TRUE,
#'     type = c("type_1", "type_1",
#'              "type_5", "type_2"),
#'     shape = c("circle", "circle",
#'               "rectangle", "rectangle"),
#'     values = c(3.5, 2.6, 9.4, 2.7))
#'
#' graph <-
#'   create_graph(nodes_df = ndf)
#'
#' # Get information on the graph's
#' # internal node data frame (ndf)
#' graph |>
#'   get_node_df()
#'
#' # A graph can also be created by
#' # specifying both the nodes and
#' # edges; create an edge data frame
#' # (edf) using the `create_edge_df()`
#' # function:
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "leading_to",
#'     values = c(7.3, 2.6, 8.3))
#'
#' # Create the graph object with
#' # `create_graph()` and pass in the
#' # ndf and edf objects
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Get information on the graph's
#' # internal edge data frame (edf)
#' graph |> get_edge_df()
#'
#' # Get information on the graph's
#' # internal node data frame (ndf)
#' graph |> get_node_df()
#'
#' @export
create_graph <- function(
    nodes_df = NULL,
    edges_df = NULL,
    directed = TRUE,
    graph_name = NULL,
    attr_theme = "default",
    write_backups = FALSE,
    display_msgs = FALSE
) {

  ## DF: `graph_info`

  # Get the time of graph creation
  graph_time <- Sys.time()

  # Get the locale's time zone
  graph_tz <- Sys.timezone()

  # Generate a random 8-character, alphanumeric
  # string to use as a graph ID
  graph_id <-
    replicate(8, sample(c(LETTERS, letters, 0:9), 1)) |>
    paste(collapse = "")

  # Create the `graph_info` data frame
  graph_info <-
    data.frame(
      graph_id = as.character(graph_id),
      graph_name = as.character(paste0("graph_", graph_id)),
      graph_time = graph_time,
      graph_tz = graph_tz,
      write_backups = write_backups,
      display_msgs = display_msgs,
      stringsAsFactors = FALSE
    )

  # Insert a user-defined `graph_name` if supplied
  if (!is.null(graph_name)) {
    graph_info[1, 2] <- as.character(graph_name)
  }

  ## DF: `global_attrs`

  # Create an empty table for global graph attributes
  global_attrs <-
    data.frame(
      attr = character(0L),
      value = character(0L),
      attr_type = character(0L),
      stringsAsFactors = FALSE
    )

  # If `attr_theme` is `default` then populate the
  # `global_attrs` data frame with global graph attrs
  if (inherits(attr_theme, "character")) {

    global_attrs <-
      switch(
        attr_theme,
        default = attr_theme_default(),
        lr = attr_theme_lr(),
        tb = attr_theme_tb(),
        rl = attr_theme_rl(),
        bt = attr_theme_bt(),
        fdp = attr_theme_fdp(),
        kk = attr_theme_kk(),
        cli::cli_abort(
          "The value for `attr_theme` doesn't refer to any available theme."
        )
      )

  } else if (is.null(attr_theme)) {

    global_attrs <-
      data.frame(
        attr = character(0L),
        value = character(0L),
        attr_type = character(0L),
        stringsAsFactors = FALSE)
  }

  ## DF: `nodes_df`

  # Create an empty node data frame (`ndf`)
  ndf <-
    data.frame(
      id = integer(0L),
      type = character(0L),
      label = character(0L),
      stringsAsFactors = FALSE
    )

  ## DF: `edges_df`

  # Create an empty edge data frame (`edf`)
  edf <-
    data.frame(
      id = integer(0L),
      from = integer(0L),
      to = integer(0L),
      rel = character(0L),
      stringsAsFactors = FALSE
    )

  ## DF: `node_selection`

  # Create an empty node selection data frame (`nsdf`)
  nsdf <-
    data.frame(node = integer(0L), stringsAsFactors = FALSE)

  ## DF: `edge_selection`

  # Create an empty edge selection data frame (`esdf`)
  esdf <-
    data.frame(
      edge = integer(0L),
      from = integer(0L),
      to = integer(0L),
      stringsAsFactors = FALSE
    )

  ## DF: `graph_actions`

  # Create an empty `graph_actions` data frame
  graph_actions <-
    data.frame(
      action_index = integer(0L),
      action_name = character(0L),
      expression = character(0L),
      stringsAsFactors = FALSE
    )

  ## DF: `graph_log`

  # Create an empty `graph_log` data frame

  graph_log <-
    data.frame(
      version_id = integer(0L),
      function_used = character(0L),
      # Datetime of length 0
      time_modified = structure(numeric(0L), class = c("POSIXct", "POSIXt")),
      duration = numeric(0L),
      nodes = integer(0L),
      edges = integer(0L),
      d_n = integer(0L),
      d_e = integer(0L),
      stringsAsFactors = FALSE
    )

  ## list: `cache`

  # Create an empty `cache` list object
  cache <- list()

  ## Empty Graph

  # Initialize a graph object
  graph <-
    list(
      graph_info = graph_info,
      nodes_df = ndf,
      edges_df = edf,
      global_attrs = global_attrs,
      directed = directed, # TRUE or FALSE
      last_node = 0,
      last_edge = 0,
      node_selection = nsdf,
      edge_selection = esdf,
      cache = cache,
      graph_actions = graph_actions,
      graph_log = graph_log
    )

  attr(graph, "class") <- "dgr_graph"

  # If neither an ndf nor both ndf & edf provided,
  # create an initialized graph with no nodes or edges
  if (is.null(nodes_df) && is.null(edges_df)) {

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph_log <-
      add_action_to_log(
        graph_log = graph_log,
        version_id = 1L,
        function_used = fcn_name,
        time_modified = graph_time,
        duration = graph_function_duration(graph_time),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df),
        d_n = nrow(graph$nodes_df),
        d_e = nrow(graph$edges_df)
      )

  } else if (!is.null(nodes_df) && is.null(edges_df)) {

    # If only an ndf is provided, create a graph
    # just containing nodes

    # Transform any `tbl_df` object to a `data.frame`
    if (inherits(nodes_df, "tbl_df")) {
      nodes_df <- as.data.frame(nodes_df, stringsAsFactors = FALSE)
    }

    # Force the `type` and `label` columns
    # to be of the character class
    for (i in 2:3) {
      nodes_df[, i] <- as.character(nodes_df[, i])
    }

    # Bind the nodes to the `nodes_df` df in the graph
    graph$nodes_df <- dplyr::bind_rows(graph$nodes_df, nodes_df)

    # Modify the `last_node` vector
    graph$last_node <- nrow(nodes_df)

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph_log <-
      add_action_to_log(
        graph_log = graph_log,
        version_id = 1L,
        function_used = fcn_name,
        time_modified = graph_time,
        duration = graph_function_duration(graph_time),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df),
        d_n = nrow(graph$nodes_df),
        d_e = nrow(graph$edges_df)
      )

  } else if (!is.null(nodes_df) && !is.null(edges_df)) {

    # If an ndf and edf both provided, create a graph
    # initially populated with both nodes and edges

    # Transform any `tbl_df` object to a `data.frame`
    if (inherits(nodes_df, "tbl_df")) {
      nodes_df <- as.data.frame(nodes_df, stringsAsFactors = FALSE)
    }

    # Transform any `tbl_df` object to a `data.frame`
    if (inherits(edges_df, "tbl_df")) {
      edges_df <- as.data.frame(edges_df, stringsAsFactors = FALSE)
    }

    # Force the `type` and `label` columns
    # to be of the character class
    for (i in 2:3) {
      nodes_df[, i] <- as.character(nodes_df[, i])
    }

    # Bind the nodes to the `nodes_df` df in the graph
    graph$nodes_df <-
      dplyr::bind_rows(graph$nodes_df, nodes_df)

    # Modify the `last_node` vector
    graph$last_node <- nrow(nodes_df)

    # Ensure that the edf has the correct classes
    if (inherits(edges_df, "data.frame") && ncol(edges_df) > 2) {

      # Force the rel column to be of the character class
      edges_df$rel <- as.character(edges_df$rel)
    }

    # Bind the edges to the `edges_df` df in the graph
    graph$edges_df <-
      dplyr::bind_rows(graph$edges_df, edges_df)

    # Modify the `last_edge` vector
    graph$last_edge <- nrow(edges_df)

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph_log <-
      add_action_to_log(
        graph_log = graph_log,
        version_id = 1L,
        function_used = fcn_name,
        time_modified = graph_time,
        duration = graph_function_duration(graph_time),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df),
        d_n = nrow(graph$nodes_df),
        d_e = nrow(graph$edges_df))
  }

  # Add the `graph_log` df to the graph object
  graph$graph_log <- graph_log

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # If neither an ndf nor both ndf & edf provided,
  # return the initialized graph with no nodes or edges
  graph
}

# Get aggregate degree values

test_that("Getting aggregated indegree values is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23)

  # Expect a certain value for the
  # mean indegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "mean"), 2.2)

  # Expect a certain value for the
  # minimum indegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "min"), 1)

  # Expect a certain value for the
  # maximum indegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "max"), 4)

  # Expect a certain value for the
  # maximum indegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "median"), 2)

  # Expect a certain value for the sum
  # of indegree values from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "sum"), 22)

  # Expect a certain value for the
  # mean indegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "mean",
      conditions = value > 5.0),
    1.8, tolerance = 0.002)

  # Expect a certain value for the
  # minimum indegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "min",
      conditions = value > 5.0),
    1, tolerance = 0.002)

  # Expect a certain value for the
  # maximum indegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "max",
      conditions = value > 5.0),
    3, tolerance = 0.002)

  # Expect a certain value for the
  # maximum indegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "median",
      conditions = value > 5.0),
    2, tolerance = 0.002)

  # Expect a certain value for the sum
  # of indegree values from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_in(
      graph = graph,
      agg = "sum",
      conditions = value > 5.0),
    9, tolerance = 0.002)

  # Expect an error if supplying a
  # value for agg not in "sum",
  # "min", "max", "mean", or "median"
  expect_error(
    get_agg_degree_in(
      graph = graph,
      agg = "sam",
      conditions = value > 5.0))
})

test_that("Getting aggregated outdegree values is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23)

  # Expect a certain value for the
  # mean outdegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "mean"),
    2.2)

  # Expect a certain value for the
  # minimum outdegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "min"),
    0)

  # Expect a certain value for the
  # maximum outdegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "max"),
    5)

  # Expect a certain value for the
  # maximum outdegree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "median"),
    2)

  # Expect a certain value for the sum
  # of outdegree values from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "sum"),
    22)

  # Expect a certain value for the
  # mean outdegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "mean",
      conditions = value > 5.0),
    1.6, tolerance = 0.002)

  # Expect a certain value for the
  # minimum outdegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "min",
      conditions = value > 5.0),
    0, tolerance = 0.002)

  # Expect a certain value for the
  # maximum outdegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "max",
      conditions = value > 5.0),
    3, tolerance = 0.002)

  # Expect a certain value for the
  # maximum outdegree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "median",
      conditions = value > 5.0),
    2, tolerance = 0.002)

  # Expect a certain value for the sum
  # of outdegree values from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_out(
      graph = graph,
      agg = "sum",
      conditions = value > 5.0),
    8, tolerance = 0.002)

  # Expect an error if supplying a
  # value for agg not in "sum",
  # "min", "max", "mean", or "median"
  expect_error(
    get_agg_degree_out(
      graph = graph,
      agg = "sam",
      conditions = value > 5.0))
})

test_that("Getting aggregated total degree values is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      node_data = node_data(
        value = 1:10),
      set_seed = 23)

  # Expect a certain value for the
  # mean total degree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "mean"),
    4.4)

  # Expect a certain value for the
  # minimum total degree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "min"),
    1)

  # Expect a certain value for the
  # maximum total degree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "max"),
    9)

  # Expect a certain value for the
  # maximum total degree value from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "median"),
    4.5)

  # Expect a certain value for the sum
  # of total degree values from all
  # nodes in the graph
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "sum"),
    44)

  # Expect a certain value for the
  # mean total degree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "mean",
      conditions = value > 5.0),
    3.4, tolerance = 0.002)

  # Expect a certain value for the
  # minimum total degree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "min",
      conditions = value > 5.0),
    1, tolerance = 0.002)

  # Expect a certain value for the
  # maximum total degree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "max",
      conditions = value > 5.0),
    5, tolerance = 0.002)

  # Expect a certain value for the
  # maximum total degree value from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "median",
      conditions = value > 5.0),
    4, tolerance = 0.002)

  # Expect a certain value for the sum
  # of total degree values from all
  # nodes in the graph (considering
  # only nodes with `value > 5.0`)
  expect_equal(
    get_agg_degree_total(
      graph = graph,
      agg = "sum",
      conditions = value > 5.0),
    17, tolerance = 0.002)

  # Expect an error if supplying a
  # value for agg not in "sum",
  # "min", "max", "mean", or "median"
  expect_error(
    get_agg_degree_total(
      graph = graph,
      agg = "sam",
      conditions = value > 5.0))
})

# Adding nodes and/or edges from a table to a graph

test_that("adding nodes from a table to a graph is possible", {

  # Create a data frame for graph nodes
  node_table <-
    dplyr::tribble(
      ~iso_4217_code, ~curr_number, ~exponent,
      "AED", 784,  2,
      "AFN", 971,  2,
      "ALL",   8,  2,
      "AMD",  51,  2,
      "ANG", 532,  2,
      "AOA", 973,  2,
      "ARS",  32,  2,
      "AUD",  36,  2,
      "AWG", 533,  2,
      "AZN", 944,  2,
      "BAM", 977,  2,
      "BBD",  52,  2,
      "BDT",  50,  2,
      "BGN", 975,  2,
      "BHD",  48,  3,
      "BIF", 108,  0,
      "BMD",  60,  2,
      "BND",  96,  2,
      "BOB",  68,  2,
      "BOV", 984,  2,
      "BRL", 986,  2,
      "BSD",  44,  2,
      "BTN",  64,  2,
      "BWP",  72,  2,
      "BYR", 974,  0,
      "BZD",  84,  2,
      "CAD", 124,  2,
      "CDF", 976,  2,
      "CHE", 947,  2,
      "CHF", 756,  2,
      "CHW", 948,  2,
      "CLF", 990,  4,
      "CLP", 152,  0,
      "CNY", 156,  2,
      "COP", 170,  2,
      "COU", 970,  2,
      "CRC", 188,  2,
      "CUC", 931,  2,
      "CUP", 192,  2,
      "CVE", 132,  0,
      "CZK", 203,  2,
      "DJF", 262,  0,
      "DKK", 208,  2,
      "DOP", 214,  2,
      "DZD",  12,  2,
      "EGP", 818,  2,
      "ERN", 232,  2,
      "ETB", 230,  2,
      "EUR", 978,  2,
      "FJD", 242,  2,
      "FKP", 238,  2,
      "GBP", 826,  2,
      "GEL", 981,  2,
      "GHS", 936,  2,
      "GIP", 292,  2,
      "GMD", 270,  2,
      "GNF", 324,  0,
      "GTQ", 320,  2,
      "GYD", 328,  2,
      "HKD", 344,  2,
      "HNL", 340,  2,
      "HRK", 191,  2,
      "HTG", 332,  2,
      "HUF", 348,  2,
      "IDR", 360,  2,
      "ILS", 376,  2,
      "INR", 356,  2,
      "IQD", 368,  3,
      "IRR", 364,  2,
      "ISK", 352,  0,
      "JMD", 388,  2,
      "JOD", 400,  3,
      "JPY", 392,  0,
      "KES", 404,  2,
      "KGS", 417,  2,
      "KHR", 116,  2,
      "KMF", 174,  0,
      "KPW", 408,  2,
      "KRW", 410,  0,
      "KWD", 414,  3,
      "KYD", 136,  2,
      "KZT", 398,  2,
      "LAK", 418,  2,
      "LBP", 422,  2,
      "LKR", 144,  2,
      "LRD", 430,  2,
      "LSL", 426,  2,
      "LYD", 434,  3,
      "MAD", 504,  2,
      "MDL", 498,  2,
      "MGA", 969,  1,
      "MKD", 807,  2,
      "MMK", 104,  2,
      "MNT", 496,  2,
      "MOP", 446,  1,
      "MRO", 478,  1,
      "MUR", 480,  2,
      "MVR", 462,  2,
      "MWK", 454,  2,
      "MXN", 484,  2,
      "MXV", 979,  2,
      "MYR", 458,  2,
      "MZN", 943,  2,
      "NAD", 516,  2,
      "NGN", 566,  2,
      "NIO", 558,  2,
      "NOK", 578,  2,
      "NPR", 524,  2,
      "NZD", 554,  2,
      "OMR", 512,  3,
      "PAB", 590,  2,
      "PEN", 604,  2,
      "PGK", 598,  2,
      "PHP", 608,  2,
      "PKR", 586,  2,
      "PLN", 985,  2,
      "PYG", 600,  0,
      "QAR", 634,  2,
      "RON", 946,  2,
      "RSD", 941,  2,
      "RUB", 643,  2,
      "RWF", 646,  0,
      "SAR", 682,  2,
      "SBD",  90,  2,
      "SCR", 690,  2,
      "SDG", 938,  2,
      "SEK", 752,  2,
      "SGD", 702,  2,
      "SHP", 654,  2,
      "SLL", 694,  2,
      "SOS", 706,  2,
      "SRD", 968,  2,
      "SSP", 728,  2,
      "STD", 678,  2,
      "SYP", 760,  2,
      "SZL", 748,  2,
      "THB", 764,  2,
      "TJS", 972,  2,
      "TMT", 934,  2,
      "TND", 788,  3,
      "TOP", 776,  2,
      "TRY", 949,  2,
      "TTD", 780,  2,
      "TWD", 901,  2,
      "TZS", 834,  2,
      "UAH", 980,  2,
      "UGX", 800,  0,
      "USD", 840,  2,
      "USN", 997,  2,
      "USS", 998,  2,
      "UYI", 940,  0,
      "UYU", 858,  2,
      "UZS", 860,  2,
      "VEF", 937,  2,
      "VND", 704,  0,
      "VUV", 548,  0,
      "WST", 882,  2,
      "XAF", 950,  0,
      "XAG", 961, NA,
      "XAU", 959, NA,
      "XCD", 951,  2,
      "XDR", 960, NA,
      "XOF", 952,  0,
      "XPD", 964, NA,
      "XPF", 953,  0,
      "XPT", 962, NA,
      "XSU", 994, NA,
      "XUA", 965, NA,
      "YER", 886,  2,
      "ZAR", 710,  2,
      "ZMW", 967,  2)

  # Convert to a data frame
  node_table <-
    as.data.frame(node_table, stringsAsFactors = FALSE)

  # Add nodes directly from the df, calling the
  # `add_nodes_from_table()` function with default
  # options
  graph_1_df <-
    create_graph() %>%
    add_nodes_from_table(table = node_table)

  # Expect that the graph has the same number of nodes
  # as there are rows in the CSV
  expect_equal(
    nrow(node_table), count_nodes(graph = graph_1_df))

  # Expect certain columns to exist in the graph's
  # node data frame
  expect_equal(
    colnames(graph_1_df$nodes_df),
    c("id", "type", "label", "iso_4217_code",
      "curr_number", "exponent"))

  # Add nodes from the df, but this time apply the
  # `curr_number` column to the graph's `label` attribute
  graph_2_df <-
    create_graph() %>%
    add_nodes_from_table(
      table = node_table,
      label_col = curr_number)

  # Expect that there aren't any NA values in the
  # graph's `label` column
  expect_false(
    anyNA(graph_2_df$nodes_df[, 3]))

  # Expect that the values in the `label` are
  # of the character class
  expect_type(graph_2_df$nodes_df[, 3], "character")

  # Add nodes from the df; also apply a static value
  # for `type` as `currency`
  graph_3_df <-
    create_graph() %>%
    add_nodes_from_table(
      table = node_table,
      set_type = "currency",
      label_col = curr_number)

  # Expect that all values set for the `type`
  # attribute are `currency`
  expect_in(graph_3_df$nodes_df[, 2], "currency")

  # Add nodes from the df; drop some of the
  # incoming columns
  graph_4_df <-
    create_graph() %>%
    add_nodes_from_table(
      table = node_table,
      set_type = "currency",
      label_col = curr_number,
      drop_cols = exponent & currency_name)

  # Expect that the node attributes `exponent`
  # and `currency_name` do not appear in the graph's
  # internal node data frame
  expect_false(
    all(c("exponent", "currency_name") %in%
           colnames(graph_4_df$nodes_df)))

  # Add nodes from the df; assign a table column
  # to the `type` atttribute
  graph_5_df <-
    create_graph() %>%
    add_nodes_from_table(
      table = node_table,
      set_type = "currency",
      label_col = curr_number,
      type_col = exponent)

  # Expect that all values set for the `type`
  # attribute are have certain values (including NA)
  expect_in(
    graph_5_df$nodes_df[, 2],
    c("0", "1", "2", "3", "4", NA))
})

test_that("adding edges from a table to a graph is possible", {

  # Create a data frame for graph nodes
  node_table <-
    dplyr::tribble(
      ~iso_4217_code, ~curr_number, ~exponent,
      "AED", 784,  2,
      "AFN", 971,  2,
      "ALL",   8,  2,
      "AMD",  51,  2,
      "ANG", 532,  2,
      "AOA", 973,  2,
      "ARS",  32,  2,
      "AUD",  36,  2,
      "AWG", 533,  2,
      "AZN", 944,  2,
      "BAM", 977,  2,
      "BBD",  52,  2,
      "BDT",  50,  2,
      "BGN", 975,  2,
      "BHD",  48,  3,
      "BIF", 108,  0,
      "BMD",  60,  2,
      "BND",  96,  2,
      "BOB",  68,  2,
      "BOV", 984,  2,
      "BRL", 986,  2,
      "BSD",  44,  2,
      "BTN",  64,  2,
      "BWP",  72,  2,
      "BYR", 974,  0,
      "BZD",  84,  2,
      "CAD", 124,  2,
      "CDF", 976,  2,
      "CHE", 947,  2,
      "CHF", 756,  2,
      "CHW", 948,  2,
      "CLF", 990,  4,
      "CLP", 152,  0,
      "CNY", 156,  2,
      "COP", 170,  2,
      "COU", 970,  2,
      "CRC", 188,  2,
      "CUC", 931,  2,
      "CUP", 192,  2,
      "CVE", 132,  0,
      "CZK", 203,  2,
      "DJF", 262,  0,
      "DKK", 208,  2,
      "DOP", 214,  2,
      "DZD",  12,  2,
      "EGP", 818,  2,
      "ERN", 232,  2,
      "ETB", 230,  2,
      "EUR", 978,  2,
      "FJD", 242,  2,
      "FKP", 238,  2,
      "GBP", 826,  2,
      "GEL", 981,  2,
      "GHS", 936,  2,
      "GIP", 292,  2,
      "GMD", 270,  2,
      "GNF", 324,  0,
      "GTQ", 320,  2,
      "GYD", 328,  2,
      "HKD", 344,  2,
      "HNL", 340,  2,
      "HRK", 191,  2,
      "HTG", 332,  2,
      "HUF", 348,  2,
      "IDR", 360,  2,
      "ILS", 376,  2,
      "INR", 356,  2,
      "IQD", 368,  3,
      "IRR", 364,  2,
      "ISK", 352,  0,
      "JMD", 388,  2,
      "JOD", 400,  3,
      "JPY", 392,  0,
      "KES", 404,  2,
      "KGS", 417,  2,
      "KHR", 116,  2,
      "KMF", 174,  0,
      "KPW", 408,  2,
      "KRW", 410,  0,
      "KWD", 414,  3,
      "KYD", 136,  2,
      "KZT", 398,  2,
      "LAK", 418,  2,
      "LBP", 422,  2,
      "LKR", 144,  2,
      "LRD", 430,  2,
      "LSL", 426,  2,
      "LYD", 434,  3,
      "MAD", 504,  2,
      "MDL", 498,  2,
      "MGA", 969,  1,
      "MKD", 807,  2,
      "MMK", 104,  2,
      "MNT", 496,  2,
      "MOP", 446,  1,
      "MRO", 478,  1,
      "MUR", 480,  2,
      "MVR", 462,  2,
      "MWK", 454,  2,
      "MXN", 484,  2,
      "MXV", 979,  2,
      "MYR", 458,  2,
      "MZN", 943,  2,
      "NAD", 516,  2,
      "NGN", 566,  2,
      "NIO", 558,  2,
      "NOK", 578,  2,
      "NPR", 524,  2,
      "NZD", 554,  2,
      "OMR", 512,  3,
      "PAB", 590,  2,
      "PEN", 604,  2,
      "PGK", 598,  2,
      "PHP", 608,  2,
      "PKR", 586,  2,
      "PLN", 985,  2,
      "PYG", 600,  0,
      "QAR", 634,  2,
      "RON", 946,  2,
      "RSD", 941,  2,
      "RUB", 643,  2,
      "RWF", 646,  0,
      "SAR", 682,  2,
      "SBD",  90,  2,
      "SCR", 690,  2,
      "SDG", 938,  2,
      "SEK", 752,  2,
      "SGD", 702,  2,
      "SHP", 654,  2,
      "SLL", 694,  2,
      "SOS", 706,  2,
      "SRD", 968,  2,
      "SSP", 728,  2,
      "STD", 678,  2,
      "SYP", 760,  2,
      "SZL", 748,  2,
      "THB", 764,  2,
      "TJS", 972,  2,
      "TMT", 934,  2,
      "TND", 788,  3,
      "TOP", 776,  2,
      "TRY", 949,  2,
      "TTD", 780,  2,
      "TWD", 901,  2,
      "TZS", 834,  2,
      "UAH", 980,  2,
      "UGX", 800,  0,
      "USD", 840,  2,
      "USN", 997,  2,
      "USS", 998,  2,
      "UYI", 940,  0,
      "UYU", 858,  2,
      "UZS", 860,  2,
      "VEF", 937,  2,
      "VND", 704,  0,
      "VUV", 548,  0,
      "WST", 882,  2,
      "XAF", 950,  0,
      "XAG", 961, NA,
      "XAU", 959, NA,
      "XCD", 951,  2,
      "XDR", 960, NA,
      "XOF", 952,  0,
      "XPD", 964, NA,
      "XPF", 953,  0,
      "XPT", 962, NA,
      "XSU", 994, NA,
      "XUA", 965, NA,
      "YER", 886,  2,
      "ZAR", 710,  2,
      "ZMW", 967,  2)

  # Convert to a data frame
  node_table <-
    as.data.frame(node_table, stringsAsFactors = FALSE)

  # Create a data frame for graph edges
  edge_table <-
    dplyr::tribble(
      ~from_currency, ~to_currency, ~cost_unit,
      "USD", "ADF", 0.1672,
      "USD", "ADP", 0.00659,
      "USD", "AED", 0.2723,
      "USD", "AFN", 0.01521,
      "USD", "ALL", 0.008055,
      "USD", "AMD", 0.002107,
      "USD", "ANG", 0.565,
      "USD", "AOA", 0.006058,
      "USD", "AON", 0.006058,
      "USD", "ARS", 0.06597,
      "USD", "ATS", 0.07969,
      "USD", "AUD", 0.7604,
      "USD", "AWG", 0.5587,
      "USD", "AZM", 0.0001221,
      "USD", "AZN", 0.6105,
      "USD", "BAM", 0.5607,
      "USD", "BBD", 0.5,
      "USD", "BDT", 0.01277,
      "USD", "BEF", 0.02718,
      "USD", "BGN", 0.5635,
      "USD", "BHD", 2.6699,
      "USD", "BIF", 0.000601,
      "USD", "BMD", 1,
      "USD", "BND", 0.7184,
      "USD", "BOB", 0.1456,
      "USD", "BRL", 0.3127,
      "USD", "BSD", 1,
      "USD", "BTN", 0.01496,
      "USD", "BWP", 0.097,
      "USD", "BYR", 4.995e-05,
      "USD", "BZD", 0.5051,
      "USD", "CAD", 0.7461,
      "USD", "CDF", 0.0009891,
      "USD", "CHF", 1.0116,
      "USD", "CLP", 0.001533,
      "USD", "CNY", 0.1476,
      "USD", "COP", 0.0003334,
      "USD", "CRC", 0.001827,
      "USD", "CUC", 1,
      "USD", "CUP", 0.045,
      "USD", "CVE", 0.00995,
      "USD", "CYP", 1.8735,
      "USD", "CZK", 0.04061,
      "USD", "DEM", 0.5606,
      "USD", "DJF", 0.005654,
      "USD", "DKK", 0.1474,
      "USD", "DOP", 0.02173,
      "USD", "DZD", 0.00912,
      "USD", "ECS", 4.15e-05,
      "USD", "EEK", 0.07008,
      "USD", "EGP", 0.1129,
      "USD", "ESP", 0.00659,
      "USD", "ETB", 0.04515,
      "USD", "EUR", 1.0965,
      "USD", "FIM", 0.1844,
      "USD", "FJD", 0.4929,
      "USD", "FKP", 1.2236,
      "USD", "FRF", 0.1672,
      "USD", "GBP", 1.2191,
      "USD", "GEL", 0.4151,
      "USD", "GHC", 2.515e-05,
      "USD", "GHS", 0.2515,
      "USD", "GIP", 1.2236,
      "USD", "GMD", 0.02365,
      "USD", "GNF", 0.0001105,
      "USD", "GRD", 0.003218,
      "USD", "GTQ", 0.1331,
      "USD", "GYD", 0.005083,
      "USD", "HKD", 0.129,
      "USD", "HNL", 0.04444,
      "USD", "HRK", 0.1461,
      "USD", "HTG", 0.01582,
      "USD", "HUF", 0.003549,
      "USD", "IDR", 7.74e-05,
      "USD", "IEP", 1.3923,
      "USD", "ILS", 0.2609,
      "USD", "INR", 0.015,
      "USD", "IQD", 0.0008729,
      "USD", "IRR", 3.328e-05,
      "USD", "ISK", 0.008882,
      "USD", "ITL", 0.0005663,
      "USD", "JMD", 0.007735,
      "USD", "JOD", 1.413,
      "USD", "JPY", 0.009536,
      "USD", "KES", 0.01003,
      "USD", "KGS", 0.01456,
      "USD", "KHR", 0.0002483,
      "USD", "KMF", 0.002281,
      "USD", "KPW", 0.007407,
      "USD", "KRW", 0.0008803,
      "USD", "KWD", 3.3113,
      "USD", "KYD", 1.2293,
      "USD", "KZT", 0.002981,
      "USD", "LAK", 0.0001229,
      "USD", "LBP", 0.0006638,
      "USD", "LKR", 0.006768,
      "USD", "LRD", 0.01111,
      "USD", "LSL", 0.073291,
      "USD", "LTL", 0.3176,
      "USD", "LUF", 0.02718,
      "USD", "LVL", 1.5602,
      "USD", "LYD", 0.7135,
      "USD", "MAD", 0.1015,
      "USD", "MDL", 0.05,
      "USD", "MGA", 0.0003137,
      "USD", "MGF", 0.0001093,
      "USD", "MKD", 0.01786,
      "USD", "MMK", 0.0007776,
      "USD", "MNT", 0.0004239,
      "USD", "MOP", 0.1285,
      "USD", "MRO", 0.002817,
      "USD", "MTL", 2.5541,
      "USD", "MUR", 0.028,
      "USD", "MVR", 0.06623,
      "USD", "MWK", 0.001404,
      "USD", "MXN", 0.05281,
      "USD", "MYR", 0.2385,
      "USD", "MZM", 1.299e-05,
      "USD", "MZN", 0.01299,
      "USD", "NAD", 0.073291,
      "USD", "NGN", 0.003284,
      "USD", "NIO", 0.03438,
      "USD", "NLG", 0.4976,
      "USD", "NOK", 0.1211,
      "USD", "NPR", 0.009368,
      "USD", "NZD", 0.7152,
      "USD", "OMR", 2.5974,
      "USD", "PAB", 1,
      "USD", "PEN", 0.2974,
      "USD", "PGK", 0.3222,
      "USD", "PHP", 0.02064,
      "USD", "PKR", 0.009545,
      "USD", "PLN", 0.2536,
      "USD", "PTE", 0.005469,
      "USD", "PYG", 0.0001751,
      "USD", "QAR", 0.2746,
      "USD", "ROL", 2.436e-05,
      "USD", "RON", 0.2436,
      "USD", "RSD", 0.008938,
      "USD", "RUB", 0.01587,
      "USD", "RWF", 0.001231,
      "USD", "SAR", 0.2668,
      "USD", "SBD", 0.1275,
      "USD", "SCR", 0.07586,
      "USD", "SDD", 0.001576,
      "USD", "SDG", 0.1576,
      "USD", "SDP", 0.0004423,
      "USD", "SEK", 0.1109,
      "USD", "SGD", 0.7184,
      "USD", "SHP", 1.8744,
      "USD", "SIT", 0.004576,
      "USD", "SKK", 0.0364,
      "USD", "SLL", 0.0001786,
      "USD", "SOS", 0.001799,
      "USD", "SRD", 0.146,
      "USD", "SRG", 0.000146,
      "USD", "STD", 4.454e-05,
      "USD", "SVC", 0.1147,
      "USD", "SYP", 0.004661,
      "USD", "SZL", 0.073291,
      "USD", "THB", 0.0286,
      "USD", "TJS", 0.127,
      "USD", "TMM", 5.7e-05,
      "USD", "TMT", 0.2857,
      "USD", "TND", 0.4441,
      "USD", "TOP", 0.4458,
      "USD", "TRL", 3.222e-07,
      "USD", "TRY", 0.3222,
      "USD", "TTD", 0.1497,
      "USD", "TWD", 0.03167,
      "USD", "TZS", 0.0004587,
      "USD", "UAH", 0.03922,
      "USD", "UGX", 0.0002929,
      "USD", "USD", 1,
      "USD", "UYU", 0.03524,
      "USD", "UZS", 0.0003225,
      "USD", "VEB", 0.0001003,
      "USD", "VEF", 0.1003,
      "USD", "VND", 4.48e-05,
      "USD", "VUV", 0.009526,
      "USD", "WST", 0.3951,
      "USD", "XAF", 0.001674,
      "USD", "XAG", 17.8532,
      "USD", "XAU", 1275.81,
      "USD", "XCD", 0.372,
      "USD", "XEU", 1.0965,
      "USD", "XOF", 0.001669,
      "USD", "XPD", 624,
      "USD", "XPF", 0.009255,
      "USD", "XPT", 983.5,
      "USD", "YER", 0.004001,
      "USD", "YUN", 0.008938,
      "USD", "ZAR", 0.07329,
      "USD", "ZMK", 0.0001932,
      "USD", "ZMW", 0.1037,
      "USD", "ZWD", 0.002679)

  # Convert to a data frame
  edge_table <-
    as.data.frame(edge_table, stringsAsFactors = FALSE)

  # Add nodes directly from the CSV file, calling the
  # `add_nodes_from_table()` function with default
  # options
  graph <-
    create_graph() %>%
    add_nodes_from_table(table = node_table)

  # Augment the graph by adding edges from a table
  # with the `add_edges_from_table()` function
  graph_nodes_edges <-
    graph %>%
    add_edges_from_table(
      table = edge_table,
      from_col = from_currency,
      to_col = to_currency,
      from_to_map = iso_4217_code)

  # Expect that the graph has a certain number of edges
  expect_equal(
    count_edges(graph = graph_nodes_edges), 157)

  # Expect certain columns to exist in the graph's
  # edge data frame
  expect_named(
    graph_nodes_edges$edges_df,
    c("id", "from", "to", "rel", "cost_unit"))

  # Expect an error if value for `from_col` or `to_col` or `from_to_map` is
  # not in the table
  expect_snapshot(error = TRUE, {
    graph %>%
      add_edges_from_table(
        edge_table,
        from_col = from,
        to_col = to_currency,
        from_to_map = iso_4217_code)

    graph %>%
      add_edges_from_table(
        edge_table,
        from_col = from_currency,
        to_col = to,
        from_to_map = iso_4217_code)

    graph %>%
      add_edges_from_table(
        edge_table,
        from_col = from_currency,
        to_col = to_currency,
        from_to_map = iso_4217)
  })

  # Augment the graph by first
  # adding edges from a table
  # with `add_edges_from_table()` and
  # then dropping a column
  graph_nodes_edges_drop <-
    graph %>%
    add_edges_from_table(
      table = edge_table,
      from_col = from_currency,
      to_col = to_currency,
      drop_cols = cost_unit,
      from_to_map = iso_4217_code)

  # Expect that the graph has a certain number of edges
  expect_equal(
    count_edges(graph = graph_nodes_edges_drop), 157)

  # Expect certain columns to exist in the graph's
  # edge data frame
  expect_named(
    graph_nodes_edges_drop$edges_df,
    c("id", "from", "to", "rel"))

  # Augment the graph by first
  # adding edges from a table
  # with `add_edges_from_table()` and
  # then setting a static `rel` value
  graph_nodes_edges_set_rel <-
    graph %>%
    add_edges_from_table(
      table = edge_table,
      from_col = from_currency,
      to_col = to_currency,
      from_to_map = iso_4217_code,
      set_rel = "change_to")

  # Expect that the graph has a certain number of edges
  expect_equal(
    count_edges(graph = graph_nodes_edges_set_rel), 157)

  # Expect certain columns to exist in the graph's
  # edge data frame
  expect_named(
    graph_nodes_edges_set_rel$edges_df,
    c("id", "from", "to", "rel", "cost_unit"))

  # Expect the same value (repeated down)
  # for the `rel` edge attribute
  expect_in(
    graph_nodes_edges_set_rel$edges_df$rel,
    "change_to")
})

test_that("adding nodes from several table columns to a graph is possible", {

  # Create a simple graph
  graph <-
    create_graph() %>%
    add_path(n = 2)

  # Create a data frame from which several
  # columns have values designated as graph nodes
  df <-
    data.frame(
      col_1 = c("f", "p", "q"),
      col_2 = c("q", "x", "f"),
      col_3 = c(1, 5, 3),
      col_4 = c("a", "v", "h"),
      stringsAsFactors = FALSE)

  # Add nodes from columns `col_1` and `col_2`
  # from the data frame to the graph object
  graph <-
    graph %>%
    add_nodes_from_df_cols(
      df = df,
      columns = c("col_1", "col_2"))

  # Expect a certain sequence of node `label` values
  expect_equal(
    get_node_df(graph)$label,
    c("1", "2", "f", "p", "q", "x"))

  # Add new nodes from columns 3 and 4; we are here
  # specifying the columns by their numbers
  graph <-
    graph %>%
    add_nodes_from_df_cols(
      df = df,
      columns = 3:4)

  # Expect a certain sequence of node `label` values
  expect_equal(
    get_node_df(graph)$label,
    c("1", "2", "f", "p", "q", "x",
      "a", "v", "h"))

  # Add column 4's values as labels/nodes again
  graph <-
    graph %>%
    add_nodes_from_df_cols(
      df = df,
      columns = 4)

  # Expect no change in the graph
  expect_equal(
    get_node_df(graph)$label,
    c("1", "2", "f", "p", "q", "x",
      "a", "v", "h"))

  # Add column 4's values as labels/nodes except with
  # the `keep_duplicates` argument set to TRUE
  graph <-
    graph %>%
    add_nodes_from_df_cols(
      df = df,
      columns = 4,
      keep_duplicates = TRUE)

  # Expect duplicated labels in the graph
  expect_equal(
    get_node_df(graph)$label,
    c("1", "2", "f", "p", "q", "x",
      "a", "v", "h", "a", "v", "h"))

  # Add column 4's values as labels/nodes with
  # the `keep_duplicates = TRUE` and a `type` value
  # of `new`
  graph <-
    graph %>%
    add_nodes_from_df_cols(
      df = df,
      columns = 4,
      type = "new",
      keep_duplicates = TRUE)

  # Expect more duplicated labels in the graph
  expect_equal(
    get_node_df(graph)$label,
    c("1", "2", "f", "p", "q", "x",
      "a", "v", "h", "a", "v", "h",
      "a", "v", "h"))

  # Expect the `type` value of `new` to appear
  # for the last three nodes (others are not set)
  expect_equal(
    get_node_df(graph)$type,
    c(rep(NA_character_, 12),
      rep("new", 3)))
})

# Rescaling node or edge attribute values in a graph object

test_that("rescaling node attributes in a graph is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10, m = 22,
      node_data = node_data(value = 1:10),
      set_seed = 23
    )

  # Rescale the `value` node attribute, so that
  # its values are rescaled between 0 and 1
  graph_r_value_0_1 <-
    graph %>%
    rescale_node_attrs(
      node_attr_from = value,
      to_lower_bound = 0,
      to_upper_bound = 1
    )

  # Expect that certain (rescaled) values are now
  # available in the graph's ndf
  expect_equal(
    graph_r_value_0_1$nodes_df$value,
    c(0.000, 0.111, 0.222, 0.333, 0.444,
      0.556, 0.667, 0.778, 0.889, 1.000))

  # Scale the values in the `value` node attribute
  # to different shades of gray for the `fillcolor`
  # and `fontcolor` node attributes
  graph_r_value_fill_font_color <-
    graph %>%
    rescale_node_attrs(
      node_attr_from = value,
      to_lower_bound = "gray80",
      to_upper_bound = "gray20",
      node_attr_to = fillcolor) %>%
    rescale_node_attrs(
      node_attr_from = value,
      to_lower_bound = "gray5",
      to_upper_bound = "gray95",
      node_attr_to = fontcolor)

  # Expect that the `fillcolor` and `fontcolor` node
  # attribute columns are now available
  expect_contains(
    colnames(graph_r_value_fill_font_color$nodes_df),
    c("fillcolor", "fontcolor"))

  # Expect that certain (rescaled) values are now
  # available in the graph's ndf
  expect_equal(
    graph_r_value_fill_font_color$nodes_df$fillcolor,
    c("#CCCCCC", "#B9B9B9", "#A7A7A7", "#959595", "#848484",
      "#737373", "#626262", "#525252", "#424242", "#333333"))

  expect_equal(
    graph_r_value_fill_font_color$nodes_df$fontcolor,
    c("#0D0D0D", "#232323", "#393939", "#515151", "#696969",
      "#838383", "#9D9D9D", "#B9B9B9", "#D5D5D5", "#F2F2F2"))

  # Expect an error if using supplying a node attribute
  # that doesn't exist (`values` instead of `value`)
  expect_error(
    graph %>%
      rescale_node_attrs(
        node_attr_from = values))
})

test_that("rescaling edge attributes in a graph is possible", {

  # Create a randomized graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 5,
      m = 10,
      edge_data = edge_data(
        weight = 1:10),
      set_seed = 23)

  # Rescale the `weight` edge attribute, so that
  # its values are rescaled between 0 and 1
  graph_r_value_0_1 <-
    graph %>%
    rescale_edge_attrs(
      edge_attr_from = weight,
      to_lower_bound = 0,
      to_upper_bound = 1)

  # Expect that certain (rescaled) values are now
  # available in the graph's edf
  expect_equal(
    graph_r_value_0_1$edges_df$weight,
    c(0.000, 0.111, 0.222, 0.333, 0.444,
      0.556, 0.667, 0.778, 0.889, 1.000))

  # Scale the values in the `weight` edge attribute
  # to different shades of gray for the `fillcolor`
  # and `fontcolor` edge attributes
  graph_r_value_fill_font_color <-
    graph %>%
    rescale_edge_attrs(
      edge_attr_from = weight,
      to_lower_bound = "gray80",
      to_upper_bound = "gray20",
      edge_attr_to = fillcolor) %>%
    rescale_edge_attrs(
      edge_attr_from = weight,
      to_lower_bound = "gray5",
      to_upper_bound = "gray95",
      edge_attr_to = fontcolor)

  # Expect that the `fillcolor` and `fontcolor` node
  # attribute columns are now available
  expect_true(
    all(
      c("fillcolor", "fontcolor") %in%
        colnames(graph_r_value_fill_font_color$edges_df)))

  # Expect that certain (rescaled) values are now
  # available in the graph's ndf
  expect_equal(
    graph_r_value_fill_font_color$edges_df$fillcolor,
    c("#CCCCCC", "#B9B9B9", "#A7A7A7", "#959595", "#848484",
      "#737373", "#626262", "#525252", "#424242", "#333333"))

  expect_equal(
    graph_r_value_fill_font_color$edges_df$fontcolor,
    c("#0D0D0D", "#232323", "#393939", "#515151", "#696969",
      "#838383", "#9D9D9D", "#B9B9B9", "#D5D5D5", "#F2F2F2"))

  # Expect an error if using supplying a node attribute
  # that doesn't exist (`weights` instead of `weight`)
  expect_error(
    graph %>%
      rescale_edge_attrs(
        edge_attr_from = weights))
})

test_that("the dice similarity algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Get the Dice similarity values for
  # nodes `5`, `6`, and `7`; all directions
  dice_all <-
    get_dice_similarity(
      graph = graph,
      nodes = 5:7,
      direction = "all")

  # Get the Dice similarity values for
  # nodes `5`, `6`, and `7`; `out` direction
  dice_out <-
    get_dice_similarity(
      graph = graph,
      nodes = 5:7,
      direction = "out")

  # Get the Dice similarity values for
  # nodes `5`, `6`, and `7`; `in` direction
  dice_in <-
    get_dice_similarity(
      graph = graph,
      nodes = 5:7,
      direction = "in")

  # Expect that a `matrix` object is returned
  expect_true(is.matrix(dice_all))
  expect_true(is.matrix(dice_out))
  expect_true(is.matrix(dice_in))

  # Expect a square matrix of 3 columns and 3 rows
  expect_equal(dim(dice_all), c(3, 3))
  expect_equal(dim(dice_out), c(3, 3))
  expect_equal(dim(dice_in), c(3, 3))

  # Expect all columns to be numeric
  expect_type(dice_all[, 1], "double")
  expect_type(dice_all[, 2], "double")
  expect_type(dice_all[, 3], "double")

  expect_type(dice_out[, 1], "double")
  expect_type(dice_out[, 2], "double")
  expect_type(dice_out[, 3], "double")

  expect_type(dice_in[, 1], "double")
  expect_type(dice_in[, 2], "double")
  expect_type(dice_in[, 3], "double")

  # Expect specific column names in this data frame
  expect_equal(
    colnames(dice_all), c("5", "6", "7"))

  expect_equal(
    colnames(dice_out), c("5", "6", "7"))

  expect_equal(
    colnames(dice_in), c("5", "6", "7"))

  # Expect specific row names in this data frame
  expect_equal(
    rownames(dice_all), c("5", "6", "7"))

  expect_equal(
    rownames(dice_out), c("5", "6", "7"))

  expect_equal(
    rownames(dice_in), c("5", "6", "7"))

  # Expect all values in the matrix to be less than
  # or equal to 1.0
  expect_true(
    all(as.numeric(dice_all) <= 1.0))

  expect_true(
    all(as.numeric(dice_out) <= 1.0))

  expect_true(
    all(as.numeric(dice_in) <= 1.0))

  # Get the Dice similarity values for all nodes,
  # all directions
  dice_all_all <-
    get_dice_similarity(
      graph = graph,
      direction = "all")

  # Expect the same number of matrix columns
  # and rows as there are nodes
  expect_equal(
    ncol(dice_all_all), count_nodes(graph = graph))

  expect_equal(
    nrow(dice_all_all), count_nodes(graph = graph))

  # Expect an error if one or more node IDs
  # provided are not in the graph
  expect_snapshot(error = TRUE,
    get_dice_similarity(
      graph = graph,
      nodes = 8:12,
      direction = "all"))
})

test_that("the Jaccard similarity algorithm is functional", {

  # Create a random graph
  graph <-
    create_graph() %>%
    add_gnm_graph(
      n = 10,
      m = 22,
      set_seed = 23)

  # Expect an error when using a value
  # for `direction` that is not one of
  # three accepted values
  expect_snapshot(error = TRUE,
    get_jaccard_similarity(
      graph = graph,
      nodes = 5:7,
      direction = "away"))

  # Get the Jaccard similarity values for
  # nodes `5`, `6`, and `7`; all directions
  jaccard_all <-
    get_jaccard_similarity(
      graph = graph,
      nodes = 5:7,
      direction = "all")

  # Get the Jaccard similarity values for
  # nodes `5`, `6`, and `7`; `out` direction
  jaccard_out <-
    get_jaccard_similarity(
      graph = graph,
      nodes = 5:7,
      direction = "out")

  # Get the Jaccard similarity values for
  # nodes `5`, `6`, and `7`; `in` direction
  jaccard_in <-
    get_jaccard_similarity(
      graph = graph,
      nodes = 5:7,
      direction = "in")

  # Expect that a `matrix` object is returned
  expect_true(is.matrix(jaccard_all))
  expect_true(is.matrix(jaccard_out))
  expect_true(is.matrix(jaccard_in))

  # Expect a square matrix of 3 columns and 3 rows
  expect_equal(dim(jaccard_all), c(3, 3))
  expect_equal(dim(jaccard_out), c(3, 3))
  expect_equal(dim(jaccard_in), c(3, 3))

  # Expect all columns to be numeric
  expect_type(jaccard_all[, 1], "double")
  expect_type(jaccard_all[, 2], "double")
  expect_type(jaccard_all[, 3], "double")

  expect_type(jaccard_out[, 1], "double")
  expect_type(jaccard_out[, 2], "double")
  expect_type(jaccard_out[, 3], "double")

  expect_type(jaccard_in[, 1], "double")
  expect_type(jaccard_in[, 2], "double")
  expect_type(jaccard_in[, 3], "double")

  # Expect specific column names in this matrix
  # Expect specific row names in this matrix
  m_names <- list(
    c("5", "6", "7"),
    c("5", "6", "7")
  )

  expect_equal(dimnames(jaccard_all), m_names)
  expect_equal(dimnames(jaccard_out), m_names)
  expect_equal(dimnames(jaccard_in), m_names)

  # Expect all values in the matrix to be less than
  # or equal to 1.0
  expect_true(
    all(as.numeric(jaccard_all) <= 1.0))

  expect_true(
    all(as.numeric(jaccard_out) <= 1.0))

  expect_true(
    all(as.numeric(jaccard_in) <= 1.0))

  # Get the Jaccard similarity values for all nodes,
  # all directions
  jaccard_all_all <-
    get_jaccard_similarity(
      graph = graph,
      direction = "all")

  # Expect the same number of matrix columns
  # and rows as there are nodes
  expect_equal(
    ncol(jaccard_all_all), count_nodes(graph = graph))

  expect_equal(
    nrow(jaccard_all_all), count_nodes(graph = graph))

  # Expect an error if one or more node IDs
  # provided are not in the graph
  expect_error(
    get_jaccard_similarity(
      graph = graph,
      nodes = 8:12,
      direction = "all"))
})

#' Create a graph series object
#'
#' @description
#'
#' Create a graph series object for the storage of multiple graphs across a
#' sequential or temporal one-dimensional array.
#'
#' @param graph A graph object to add to the new graph series object.
#' @param series_name An optional name to ascribe to the series.
#' @param series_type Either a `sequential` type (the default) or a `temporal`
#'   type (which requires date-time strings and time zone codes to be supplied).
#'
#' @return A graph series object of type `dgr_graph_1D`.
#'
#' @examples
#' # Create three graphs
#' graph_1 <-
#'   create_graph() %>%
#'   add_path(n = 4)
#'
#' graph_2 <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' graph_3 <-
#'   create_graph() %>%
#'   add_star(n = 6)
#'
#' # Create an empty graph series
#' # and add the graphs
#' series <-
#'   create_graph_series() %>%
#'   add_graph_to_graph_series(
#'     graph = graph_1) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_2) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_3)
#'
#' # Count the number of graphs
#' # in the graph series
#' series %>%
#'   count_graphs_in_graph_series()
#'
#' @export
create_graph_series <- function(
    graph = NULL,
    series_name = NULL,
    series_type = "sequential"
) {

  # Validation: Graph object is valid
  if (!is.null(graph))  {
    check_graph_valid(graph)
  }

  # Initialize an empty graph series object
  graph_series <-
    list(
      graphs = NULL,
      series_name = series_name,
      series_type = series_type)

  attr(graph_series, "class") <- "dgr_graph_1D"

  if (is.null(graph)) {
    return(graph_series)
  }

  # Add a graph to the initialized graph series
  graph_series$graphs[[length(graph_series$graphs) + 1]] <- graph

  graph_series
}

#' Add edges and attributes to graph from a table
#'
#' @description
#'
#' Add edges and their attributes to an existing graph object from data in a CSV
#' file or a data frame.
#'
#' @inheritParams render_graph
#' @param table Either a path to a CSV file, or, a data frame object.
#' @param from_col The name of the table column from which edges originate.
#' @param to_col The name of the table column to which edges terminate.
#' @param from_to_map A single character value for the mapping of the `from` and
#'   `to` columns in the external table (supplied as `from_col` and `to_col`,
#'   respectively) to a column in the graph's internal node data frame (ndf).
#' @param rel_col An option to apply a column of data in the table as `rel`
#'   attribute values.
#' @param set_rel an optional string to apply a `rel` attribute to all edges
#'   created from the table records.
#' @param drop_cols An optional column selection statement for dropping columns
#'   from the external table before inclusion as attributes in the graph's
#'   internal edge data frame. Several columns can be dropped by name using the
#'   syntax `col_1 & col_2 & ...`. Columns can also be dropped using a numeric
#'   column range with `:` (e.g., `5:8`), or, by using the `:` between column
#'   names to specify the range (e.g., `col_5_name:col_8_name`).
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph and then
#' # add nodes to it from the
#' # `currencies` dataset available
#' # in the package
#' graph <-
#'   create_graph() |>
#'   add_nodes_from_table(
#'     table = currencies)
#'
#' # Now we want to add edges to the
#' # graph using an included dataset,
#' # `usd_exchange_rates`, which has
#' # exchange rates between USD and
#' # many other currencies; the key
#' # here is that the data in the
#' # `from` and `to` columns in the
#' # external table maps to graph
#' # node data available in the
#' # `iso_4217_code` column of the
#' # graph's internal node data frame
#' graph_1 <-
#'   graph |>
#'     add_edges_from_table(
#'       table = usd_exchange_rates,
#'       from_col = from_currency,
#'       to_col = to_currency,
#'       from_to_map = iso_4217_code)
#'
#' # View part of the graph's
#' # internal edge data frame
#' graph_1 |>
#'   get_edge_df() |>
#'   head()
#'
#' # If you would like to assign
#' # any of the table's columns as the
#' # `rel` attribute, this can done
#' # with the `rel_col` argument; to
#' # set a static `rel` attribute for
#' # all edges created, use `set_rel`
#' graph_2 <-
#'   graph |>
#'     add_edges_from_table(
#'       table = usd_exchange_rates,
#'       from_col = from_currency,
#'       to_col = to_currency,
#'       from_to_map = iso_4217_code,
#'       set_rel = "from_usd")
#'
#' # View part of the graph's internal
#' # edge data frame (edf)
#' graph_2 |>
#'   get_edge_df() |>
#'   head()
#'
#' @family edge creation and removal
#'
#' @export
add_edges_from_table <- function(
    graph,
    table,
    from_col,
    to_col,
    from_to_map,
    rel_col = NULL,
    set_rel = NULL,
    drop_cols = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, "so, edges cannot be added.")

  # Get the requested `from_col`
  from_col <-
    rlang::ensym(from_col) |> rlang::as_string()

  # Get the requested `to_col`
  to_col <-
    rlang::ensym(to_col) |> rlang::as_string()

  # Get the requested `from_to_map`
  from_to_map <-
    rlang::ensym(from_to_map) |> rlang::as_string()

  # Get the requested `rel_col`
  if (!rlang::quo_is_null(rlang::enquo(rel_col))) {
    rel_col <-
      rlang::ensym(rel_col) |> rlang::as_string()
  }

  # Get the requested `drop_cols`
  if (!rlang::quo_is_null(rlang::enquo(drop_cols))) {
    drop_cols <-
      rlang::ensym(drop_cols) |> rlang::as_string()
  }

  # Determine whether the table is a file connection
  # to a CSV file or a data frame
  if (inherits(table, "character")) {

    # Load in CSV file
    csv <- utils::read.csv(table, stringsAsFactors = FALSE)

  } else if (inherits(table, "data.frame")) {

    # Rename `table` object as `csv`
    csv <- table
  }

  # Verify that value for `from_col` is in the table
  if (!(from_col %in% colnames(csv))) {

    abort(
      "The value specified in `from_col` is not in the table.")
  }

  # Verify that value for `to_col` is in the table
  if (!(to_col %in% colnames(csv))) {

    abort(
      "The value specified in `to_col` is not in the table.")
  }

  # Verify that value for `from_to_map` is in the
  # graph's ndf
  if (!(from_to_map %in% colnames(get_node_df(graph)))) {

    abort(
      "The value specified in `from_to_map` is not in the graph.")
  }

  # Optionally set the `rel` attribute from a
  # specified column in the CSV
  if (!is.null(rel_col) && any(colnames(csv) == rel_col)) {

    colnames(csv)[which(colnames(csv) == rel_col)] <- "rel"
    csv$rel <- as.character(csv$rel)
  }

  # Extract the ndf from the graph
  ndf <- graph$nodes_df

  # Exclude the `from` and `to` columns
  # from the `csv` table
  csv_data_excluding_from_to <-
    csv |>
    dplyr::select(setdiff(colnames(csv), c(from_col, to_col)))

  # Get the `from` col
  col_from <-
    dplyr::as_tibble(csv) |>
    dplyr::select(!!from_col) |>
    dplyr::left_join(
      ndf |> dplyr::select("id", !!from_to_map),
      by = stats::setNames(from_to_map, from_col)) |>
    dplyr::select(from = "id") |>
    dplyr::mutate(from = as.integer(from))

  # Get the `to` col
  col_to <-
    dplyr::as_tibble(csv) |>
    dplyr::select(!!to_col) |>
    dplyr::left_join(
      ndf |> dplyr::select("id", !!from_to_map),
      by = stats::setNames(from_to_map, to_col)) |>
    dplyr::select(to = "id") |>
    dplyr::mutate(to = as.integer(to))

  # Combine the `from` and `to` columns together along
  # with a new `rel` column (filled with NAs) and additional
  # columns from the CSV
  edf <-
    col_from |>
    dplyr::bind_cols(col_to) |>
    dplyr::bind_cols(csv_data_excluding_from_to)

  # Add in a `rel` column (filled with NAs) if it's not
  # already in the table
  if (!("rel" %in% colnames(edf))) {
    edf$rel <- NA_character_
  }

  # Use the `select()` function to arrange the
  # column rows and then convert to a data frame
  edf <-
    edf |>
    dplyr::relocate("from", "to", "rel") |>
    as.data.frame(stringsAsFactors = FALSE)

  # Remove any rows where there is an NA in either
  # `from` or `to`
  edf <- edf[which(!is.na(edf$from) & !is.na(edf$to)), ]
  rownames(edf) <- NULL

  # Add in an `id` column
  edf <-
    dplyr::bind_cols(
      data.frame(id = seq_len(nrow(edf)) + as.integer(graph$last_edge)),
      edf)

  # Optionally set the `rel` attribute with a single
  # value repeated down
  if (is.null(rel_col) && !is.null(set_rel)) {
    edf$rel <- as.character(set_rel)
  }

  # If values for `drop_cols` provided, filter the CSV
  # columns by those named columns
  if (!is.null(drop_cols)) {

    col_selection <- get_col_selection(col_selection_stmt = drop_cols)

    if (col_selection[["selection_type"]] == "column_range") {
      col_index_1 <- which(colnames(edf) == col_selection[["column_selection"]][1])
      col_index_2 <- which(colnames(edf) == col_selection[["column_selection"]][2])

      col_indices <- col_index_1:col_index_2 |> sort()

      columns_retained <- base::setdiff(colnames(edf), colnames(edf)[col_indices])

    } else if (col_selection[["selection_type"]] == "column_index_range") {

      col_indices <- col_selection[["column_selection"]] |> sort()

      columns_retained <- base::setdiff(colnames(edf), colnames(edf)[col_indices])

    } else if (col_selection[["selection_type"]] %in% c("single_column_name", "column_names")) {

      columns_retained <- base::setdiff(colnames(edf), col_selection[["column_selection"]])

    } else if (length(col_selection) == 0) {

      columns_retained <- colnames(edf)
    }

    edf <- edf[, c(columns_retained)]
  }

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Add the edf to the graph object
  if (is.null(graph$edges_df)) {
    graph$edges_df <- edf
  } else {
    graph$edges_df <- dplyr::bind_rows(graph$edges_df, edf)
  }

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph |> count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Update the `last_edge` value in the graph
  graph$last_edge <- nrow(graph$edges_df)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Invert selection of nodes or edges in a graph
#'
#' @description
#'
#' Modify the selection of nodes or edges within a graph object such that all
#' nodes or edges previously not selected will now be selected and vice versa.
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = "standard")
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "leading_to")
#'
#' # Create a graph
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Select nodes with ID
#' # values `1` and `3`
#' graph <-
#'   graph |>
#'   select_nodes(
#'     nodes = c(1, 3))
#'
#' # Verify that a node
#' # selection has been made
#' graph |> get_selection()
#'
#' # Invert the selection
#' graph <-
#'   graph |>
#'   invert_selection()
#'
#' # Verify that the node
#' # selection has been changed
#' graph |> get_selection()
#'
#' @export
invert_selection <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph object has valid selection of
  # nodes or edges
  if (!graph_contains_node_selection(graph) &&
      !graph_contains_edge_selection(graph)) {

    cli::cli_abort(
      "There is no selection of nodes or edges available.")
  }

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # Invert the nodes in the selection
  if (nrow(graph$node_selection) > 0) {

    selection_nodes <- graph$node_selection$node

    ndf <- graph$nodes_df

    inverted_nodes <-
      ndf |>
      dplyr::filter(!(id %in% selection_nodes)) |>
      dplyr::select("id")

    # Add the node ID values to the active selection
    # of nodes in `graph$node_selection`
    graph$node_selection <-
      replace_graph_node_selection(
        graph = graph,
        replacement = inverted_nodes$id)

    # Replace `graph$edge_selection` with an empty df
    graph$edge_selection <- create_empty_esdf()
  }

  # Invert the edges in the selection
  if (nrow(graph$edge_selection) > 0) {

    selection_edges <- graph$edge_selection$edge

    edf <- graph$edges_df

    inverted_edges <-
      edf |>
      dplyr::filter(!(id %in% selection_edges)) |>
      dplyr::select("id", "from", "to")

    # Add the node ID values to the active selection
    # of nodes in `graph$node_selection`
    graph$edge_selection <-
      replace_graph_edge_selection(
        graph = graph,
        edge_id = inverted_edges$id,
        from_node = inverted_edges$from,
        to_node = inverted_edges$to)

    # Replace `graph$node_selection` with an empty df
    graph$node_selection <- create_empty_nsdf()
  }

  # Obtain the output graph's node and edge
  # selection properties
  n_e_select_properties_out <-
    node_edge_selection_properties(graph = graph)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # Emit a message about the modification of a selection
  # if that option is set
  if (!is.null(graph$graph_info$display_msgs) &&
      graph$graph_info$display_msgs) {

    # Construct message body
    msg_body <-
      glue::glue(
        "inverted an existing selection of \\
       {n_e_select_properties_in[['selection_count_str']]}:
       * {n_e_select_properties_out[['selection_count_str']]} \\
       are now in the active selection")

    # Issue a message to the user
    emit_message(
      fcn_name = fcn_name,
      message_body = msg_body)
  }

  graph
}

#' Set node attributes with a node selection
#'
#' @description
#'
#' From a graph object of class `dgr_graph` or a node data frame, set node
#' attribute properties for nodes present in a node selection.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param node_attr The name of the attribute to set.
#' @param value The value to be set for the chosen attribute for the nodes in
#'   the current selection.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a simple graph
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 6)
#'
#' # Select specific nodes from the graph and
#' # apply the node attribute `color = blue` to
#' # those selected nodes
#' graph <-
#'   graph %>%
#'   select_nodes_by_id(
#'     nodes = 1:4) %>%
#'   trav_out() %>%
#'   set_node_attrs_ws(
#'     node_attr = color,
#'     value = "blue")
#'
#' # Show the internal node data frame to verify
#' # that the node attribute has been set for
#' # specific node
#' graph %>% get_node_df()
#'
#' @family node creation and removal
#'
#' @export
set_node_attrs_ws <- function(
    graph,
    node_attr,
    value
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # Get the requested `node_attr`
  node_attr <-
    rlang::enquo(node_attr) %>% rlang::get_expr() %>% as.character()

  # Get vector of node ID values
  nodes <- graph$node_selection$node

  node_attr_2 <- rlang::enquo(node_attr)

  # Call the `set_node_attrs()` function
  # and update the graph
  graph <-
    set_node_attrs(
      graph = graph,
      node_attr = !!node_attr_2,
      values = value,
      nodes = nodes
    )

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    graph$graph_log[-nrow(graph$graph_log), ] %>%
    add_action_to_log(
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Print the graph to the terminal
#'
#' This function will provide a summary of the graph.
#'
#' @param x A graph object of class `dgr_graph`.
#' @examples
#' \dontrun{
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get a summary of the graph
#' graph
#' }
#' @keywords internal
#' @export
print.dgr_graph <- function(x, ...) {

  args <- list(...)
  args <- NULL

  # Get the console width
  console_width <- getOption("width")

  # Create vector of information labels
  info_labels <-
    c("get_node_df" = " info: `get_node_df()`",
      "get_edge_df" = " info: `get_edge_df()`",
      "get_selection" = " info: `get_selection()`",
      "get_cache" = " info: `get_cache()",
      "get_global_graph_attr_info" = " info: `get_global_graph_attr_info()`")

  # Get a count of all nodes in the graph
  node_count <- x |> count_nodes()

  # Get a count of all edges in the graph
  edge_count <- x |> count_edges()

  # Get the node `type` status
  if (all(is.na(x$nodes_df$type))) {

    node_type_status <- "<unused>"

  } else if (!all(is.na(x$nodes_df$type)) && anyNA(x$nodes_df$type)) {

    node_type_status <-
      paste0(
        x$nodes_df$type |>
          unique() |>
          base::setdiff(NA_character_) |>
          length(), " val",
        ifelse(
          x$nodes_df$type |>
            unique() |>
            base::setdiff(NA_character_) |>
            length() > 1, "s", ""))

  } else if (!anyNA(x$nodes_df$type)) {

    node_type_status <-
      paste0(
        x$nodes_df$type |>
          unique() |>
          base::setdiff(NA_character_) |>
          length(), " vals - complete")
  }

  # Get the node `label` status
  if (all(is.na(x$nodes_df$label))) {

    node_label_status <- "<unused>"

  } else if (!all(is.na(x$nodes_df$label)) && anyNA(x$nodes_df$label)) {
    node_label_status <-
      paste0(
        x$nodes_df$label |>
          unique() |>
          base::setdiff(NA_character_) |>
          length(), " val",
        ifelse(
          x$nodes_df$label |>
            unique() |>
            base::setdiff(NA_character_) |>
            length() > 1, "s", ""))

  } else if (!anyNA(x$nodes_df$label)) {

    node_label_status <-
      paste0(
        x$nodes_df$label |>
          unique() |>
          base::setdiff(NA_character_) |>
          length(), " vals - complete")

    if (!anyDuplicated(x$nodes_df$label)) {

      node_label_status <-
        paste0(
          node_label_status, " & unique")
    }
  }

  # Get a list of extra node attributes
  node_extra_attrs <-
    x$nodes_df |>
    colnames() |>
    base::setdiff(c("id", "type", "label"))

  if (length(node_extra_attrs) > 0) {

    count_extra_attrs <- length(node_extra_attrs)

    if (count_extra_attrs == 1) {

      node_extra_attrs_str <-
        paste0(
          "1 additional node attribute (",
          node_extra_attrs[1], ")")

    } else if (count_extra_attrs > 1 && count_extra_attrs <= 3) {

      node_extra_attrs_str <-
        paste0(
          count_extra_attrs,
          " additional node attributes (",
          paste(node_extra_attrs, collapse = ", "),
          ")")

    } else if (count_extra_attrs > 3) {

      node_extra_attrs_str <-
        paste0(
          count_extra_attrs,
          " additional node attributes (",
          paste(node_extra_attrs[1:3], collapse = ", "),
          " + ", count_extra_attrs - 3, " more)")
    }
  } else {

    node_extra_attrs_str <- "no additional node attributes"
  }

  # Get the edge `rel` status
  if (all(is.na(x$edges_df$rel))) {

    edge_rel_status <- "<unused>"

  } else if (!all(is.na(x$edges_df$rel)) && anyNA(x$edges_df$rel)) {

    edge_rel_status <-
      paste0(
        x$edges_df$rel |>
          unique() |>
          base::setdiff(NA_character_) |>
          length(), " val",
        ifelse(
          x$edges_df$rel |>
            unique() |>
            base::setdiff(NA_character_) |>
            length() > 1, "s", ""))

  } else if (!anyNA(x$edges_df$rel)) {
    edge_rel_status <-
      paste0(
        x$edges_df$rel |>
          unique() |>
          base::setdiff(NA_character_) |>
          length(), " vals - complete")
  }

  # Get a list of extra edge attributes
  edge_extra_attrs <-
    x$edges_df |>
    colnames() |>
    base::setdiff(c("id", "from", "to", "rel"))

  if (length(edge_extra_attrs) > 0) {

    count_extra_attrs <- length(edge_extra_attrs)

    if (count_extra_attrs == 1) {

      edge_extra_attrs_str <-
        paste0(
          "1 additional edge attribute (",
          edge_extra_attrs[1], ")")

    } else if (count_extra_attrs > 1 && count_extra_attrs <= 3) {

      edge_extra_attrs_str <-
        paste0(
          count_extra_attrs,
          " additional edge attributes (",
          paste(edge_extra_attrs, collapse = ", "),
          ")")

    } else if (count_extra_attrs > 3) {

      edge_extra_attrs_str <-
        paste0(
          count_extra_attrs,
          " additional edge attributes (",
          paste(edge_extra_attrs[1:3], collapse = ", "),
          " + ", count_extra_attrs - 3, " more)")
    }
  } else {

    edge_extra_attrs_str <- "no additional edge attributes"
  }

  # Determine if the graph is directed
  directed_undirected <- ifelse(
    is_graph_directed(x),
    "directed",
    "undirected"
  )

  # Determine if the graph is weighted
  weighted_graph_status <- is_graph_weighted(x)

  # Determine if the graph is a DAG
  dag_graph_status <- is_graph_dag(x)

  # Determine if the graph is a property graph
  property_graph_status <- is_property_graph(x)

  # Determine if the graph is a simple graph
  simple_graph_status <- is_graph_simple(x)

  # Determine if the graph is connected
  # or disconnected
  connected_graph_status <- is_graph_connected(x)

  # Generate a string describing the number of
  # nodes in the graph
  if (node_count > 1) {
    node_count_str <- paste0(node_count, " nodes")
  } else if (node_count == 1) {
    node_count_str <- paste0(node_count, " node")
  } else if (node_count == 0) {
    node_count_str <- "no nodes"
  }

  # Generate a string describing the number of
  # edges in the graph
  if (node_count > 0) {
    if (edge_count > 1) {
      edge_count_str <- paste0(edge_count, " edges")
    } else if (edge_count == 1) {
      edge_count_str <- paste0(edge_count, " edge")
    } else if (edge_count == 0) {
      edge_count_str <- "no edges"
    }
  } else {
    edge_count_str <- ""
  }

  # Generate the header string for printing
  if (edge_count > 0) {

    # Generate a string describing the density of
    # the graph
    # density_str <-
    #   paste0(
    #     "density: ", get_graph_info(x)$dens)

    header_str <-
      paste0(
        "DiagrammeR Graph // ", node_count_str,
        " / ", edge_count_str)#,
    # " / ", density_str)

  } else if (edge_count == 0) {

    header_str <-
      paste0(
        "DiagrammeR Graph // ", node_count_str)
  }

  # Generate the subheader string for printing
  if (is_graph_empty(x)) {

    sub_header_str <-
      paste0(
        "  -- empty graph",
        " (mode: ", directed_undirected, ")")

  } else {

    sub_header_str <-
      paste0(
        "  -- ",
        directed_undirected,
        ifelse(connected_graph_status, " / connected", " / disconnected"),
        ifelse(weighted_graph_status, " / weighted", ""),
        ifelse(dag_graph_status, " / DAG", ""),
        ifelse(property_graph_status, " / property graph", ""),
        ifelse(simple_graph_status, " / simple", ""))
  }

  #
  # Create strings for node-specific information
  #

  node_detail_str_1 <-
    paste0(
      "  NODES / ",
      "type: ", node_type_status,
      " / ",
      "label: ", node_label_status)

  node_detail_str_2 <-
    paste0(
      "    -- ",
      node_extra_attrs_str)

  #
  # Create strings for edge-specific information
  #

  edge_detail_str_1 <-
    paste0(
      "  EDGES / ",
      "rel: ", edge_rel_status)

  edge_detail_str_2 <-
    paste0(
      "    -- ",
      edge_extra_attrs_str)

  #
  # Create string for active selections
  #

  if (all(is.na(suppressMessages(get_selection(x)))) &&
      length(suppressMessages(get_selection(x))) == 1) {

    selection_str <- "<none>"

  } else {

    if (nrow(x$node_selection) > 0) {

      selection_type <- "node"

      selection_count <-
        nrow(x$node_selection)

      # selection_all is TRUE if selection_count = number of nodes
      selection_all <-
        selection_count == nrow(x$nodes_df)
    }

    if (nrow(x$edge_selection) > 0) {

      selection_type <- "edge"

      selection_count <-
        nrow(x$edge_selection)

      # selection_all is TRUE if selection_count = number of edges
      selection_all <-
        selection_count == nrow(x$edges_df)
    }

    selection_str <-
      paste0(
        selection_count, " ",
        selection_type,
        ifelse(selection_count > 1, "s", ""),
        " selected",
        ifelse(selection_all, " (ALL)", ""))
  }

  selection_detail_str <-
    paste0(
      "  SELECTION / ",
      selection_str)

  #
  # Create string for active caches
  #

  if (all(is.na(get_cache(x)))) {

    cache_str <- "<none>"

  } else {

    cache_count <-
      length(x$cache)

    cache_str <-
      paste0(
        cache_count,
        " cache",
        ifelse(cache_count > 1, "s", ""))
  }

  cache_detail_str <-
    paste0(
      "  CACHE / ",
      cache_str)

  #
  # Create string for global attributes
  #

  if (nrow(x$global_attrs) == 0) {

    global_attrs_str <- "<none>"

  } else {

    global_attrs_count <-
      nrow(x$global_attrs)

    global_attrs_str <-
      paste0(
        global_attrs_count,
        ifelse(global_attrs_count > 1, " are set", " is set"))
  }

  global_attrs_detail_str <-
    paste0(
      "  GLOBAL ATTRS / ",
      global_attrs_str)

  #
  # Create string for graph actions
  #

  if (nrow(x$graph_actions) == 0) {

    graph_actions_str <- "<none>"

  } else {

    graph_actions_count <-
      nrow(x$graph_actions)

    graph_actions_str <-
      paste0(
        graph_actions_count,
        ifelse(graph_actions_count > 1, " are set", " is set"))
  }

  graph_actions_detail_str <-
    paste0(
      "  GRAPH ACTIONS / ",
      graph_actions_str)

  #
  # Create string of the recent graph history
  #

  number_of_actions_logged <- nrow(x$graph_log)

  tail_actions_logged <-
    x$graph_log |>
    utils::tail(3)

  number_of_actions_in_tail <-
    nrow(tail_actions_logged)

  if (number_of_actions_in_tail == number_of_actions_logged) {
    last_actions_performed_str <-
      paste0(
        paste(tail_actions_logged$function_used, collapse = "() -> "), "()")
  } else if (number_of_actions_in_tail < number_of_actions_logged) {
    last_actions_performed_str <-
      paste0(
        "<", number_of_actions_logged - number_of_actions_in_tail, " action",
        ifelse((number_of_actions_logged - number_of_actions_in_tail) > 1, "s", ""),
        "> -> ",
        paste0(
          paste(
            tail_actions_logged$function_used, collapse = "() -> "), "()"))
  }

  graph_history_detail_str <-
    paste0(
      "  GRAPH LOG / ",
      last_actions_performed_str)

  node_detail_str_1_length <- nchar(node_detail_str_1)
  edge_detail_str_1_length <- nchar(edge_detail_str_1)
  selection_detail_str_length <- nchar(selection_detail_str)
  cache_detail_str_length <- nchar(cache_detail_str)
  global_attrs_detail_str_length <- nchar(global_attrs_detail_str)

  info_labels_node_df_length <- nchar(info_labels["get_node_df"])[[1]]
  info_labels_edge_df_length <- nchar(info_labels["get_edge_df"])[[1]]
  info_labels_selection_length <- nchar(info_labels["get_selection"])[[1]]
  info_labels_cache_length <- nchar(info_labels["get_cache"])[[1]]
  info_labels_global_graph_attrs <- nchar(info_labels["get_global_graph_attr_info"])[[1]]

  if (console_width - node_detail_str_1_length - info_labels_node_df_length >= 5) {

    node_detail_str_1 <-
      paste0(
        node_detail_str_1,
        paste(
          rep(
            x = " ",
            times = (console_width -
                       node_detail_str_1_length -
                       info_labels_node_df_length)),
          collapse = ""),
        info_labels["get_node_df"])
  }

  if (console_width - edge_detail_str_1_length - info_labels_edge_df_length >= 5) {

    edge_detail_str_1 <-
      paste0(
        edge_detail_str_1,
        paste(
          rep(
            x = " ",
            times = (console_width -
                       edge_detail_str_1_length -
                       nchar(info_labels["get_edge_df"])[[1]])),
          collapse = ""),
        info_labels["get_edge_df"])
  }

  if (console_width -
      selection_detail_str_length -
      info_labels_selection_length >= 5 &
      !is.na(suppressMessages(get_selection(x)))[1]) {

    selection_detail_str <-
      paste0(
        selection_detail_str,
        paste(
          rep(
            x = " ",
            times = (console_width -
                       selection_detail_str_length -
                       info_labels_selection_length)),
          collapse = ""),
        info_labels["get_selection"])
  }

  if (console_width -
      cache_detail_str_length -
      info_labels_cache_length >= 5 &&
      !is.na(get_cache(x))[1]) {

    cache_detail_str <-
      paste0(
        cache_detail_str,
        paste(
          rep(
            x = " ",
            times = (console_width -
                       cache_detail_str_length -
                       info_labels_cache_length)),
          collapse = ""),
        info_labels["get_cache"])
  }

  if (console_width -
      global_attrs_detail_str_length -
      info_labels_global_graph_attrs >= 5) {

    if (inherits(get_global_graph_attr_info(x), "data.frame")) {

      global_attrs_detail_str <-
        paste0(
          global_attrs_detail_str,
          paste(
            rep(
              x = " ",
              times = (console_width -
                         global_attrs_detail_str_length -
                         info_labels_global_graph_attrs)),
            collapse = ""),
          info_labels["get_global_graph_attr_info"])
    }
  }

  # Generate the complete statement for printing
  complete_stmt <-
    paste0(
      header_str, "\n",
      sub_header_str, "\n",
      "\n",
      node_detail_str_1, "\n",
      node_detail_str_2, "\n",
      edge_detail_str_1, "\n",
      edge_detail_str_2, "\n",
      selection_detail_str, "\n",
      cache_detail_str, "\n",
      global_attrs_detail_str, "\n",
      graph_actions_detail_str, "\n",
      graph_history_detail_str, "\n"
    )

  cat(complete_stmt)
}

#' Layout nodes using a text-based schematic
#'
#' @description
#'
#' Layout one or several groups of nodes using a text-based schematic. The
#' option is available to apply sorting to each of the groups.
#'
#' @inheritParams render_graph
#' @param layout A layout character string that provides a schematic for the
#'   layout. This consists of a rectangular collection of `-` characters (for no
#'   node placement), and numbers from `1` to `9` (representing different
#'   groupings of nodes, further described in the `nodes` argument).
#' @param nodes A named vector of the form: `c("1" = "[node_attr]:[value]",
#'   ...)`. The LHS corresponds to the numbers used in the `layout` schematic.
#'   The RHS provides a shorthand for the node attribute and a value for
#'   grouping together nodes (separated by a colon). For instance, with
#'   `"type:a"` in the RHS (and `"1"` in the LHS) we would target all nodes with
#'   a `type` attribute equal to `a` for positioning in the graph as described
#'   by the `1`s in the `layout`.
#' @param sort An optional sorting method to apply to the collection of nodes
#'   before assigning positional information. Like `nodes`, this is a named
#'   vector of the form: `c("1" = "[node_attr]:asc|desc", ...)`. The `node_attr`
#'   in this case should be different than that used in `nodes`. Ideally, this
#'   node attribute should have unique values. Choose either `asc` or `desc`
#'   right of the colon for ascending or descending sorts.
#' @param width The width of the `layout` diagram.
#' @param height The height of the `layout` diagram.
#' @param ll A vector describing the the lower-left coordinates of the `layout`
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with unique labels and
#' # several node `type` groupings
#' graph <-
#'   create_graph() |>
#'   add_node(type = "a", label = "a") |>
#'   add_node(type = "a", label = "b") |>
#'   add_node(type = "b", label = "c") |>
#'   add_node(type = "b", label = "d") |>
#'   add_node(type = "b", label = "e") |>
#'   add_node(type = "c", label = "f") |>
#'   add_node(type = "c", label = "g")
#'
#' # Define a 'layout' for groups of nodes
#' # using a text string (dashes are empty
#' # grid cells, numbers--representing
#' # ad-hoc groupings--correspond to
#' # individual nodes); here, define a layout
#' # with 3 groups of nodes
#' layout <-
#' "
#' 1--------
#' 1--------
#' ---222---
#' --------3
#' --------3
#' "
#'
#' # Use the `layout` along with what nodes
#' # the numbers correspond to in the graph
#' # with the `nodes` named vectors; the
#' # optional `sort` vector describes how
#' # we should sort the collection of node
#' # before adding position information
#' graph <-
#'   graph |>
#'   layout_nodes_w_string(
#'     layout = layout,
#'     nodes = c("1" = "type:a",
#'               "2" = "type:b",
#'               "3" = "type:c"),
#'     sort = c("1" = "label:asc",
#'              "2" = "label:desc",
#'              "3" = "label:desc"))
#'
#' # Show the graph's node data frame
#' # to confirm that `x` and `y` values
#' # were added to each of the nodes
#' graph |> get_node_df()
#' @family node creation and removal
#' @export
layout_nodes_w_string <- function(
    graph,
    layout,
    nodes,
    sort = NULL,
    width = 8,
    height = 8,
    ll = c(0, 0)
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the graph's internal node data frame
  # (ndf) into a separate object
  ndf <- graph$nodes_df

  # Get the total number of groups described
  # in the layout
  node_group_count <- length(nodes)

  # Parse the `layout` object to get a vector
  # of rows
  layout <-
    gsub(" ", "", layout) |>
    stringr::str_split("\n") |>
    unlist()

  layout <-
    layout[which(nzchar(layout))]

  # Determine the row length from the layout text
  layout_row_length <- vector(mode = "numeric")

  for (i in seq_along(layout)) {
    layout_row_length <-
      c(layout_row_length, nchar(layout[i]))
  }

  # Stop function if not all rows are of equal length
  if (mean(layout_row_length) != layout_row_length[1]) {

    cli::cli_abort(
      "Each row must have the same length ({layout_row_length[1]}).")
  }

  layout_column_number <- layout_row_length <- layout_row_length[1]
  layout_row_number <- layout_column_length <- length(layout)

  # Define the exact `x` and `y` positions in which
  # the nodes could be placed
  x_pts <- seq(0, width, width / (layout_column_number - 1)) + ll[1]
  y_pts <- rev(seq(0, height, height / (layout_row_number - 1))) + ll[2]

  # Create a tibble called `ndf_parts`
  ndf_parts <- dplyr::tibble()

  for (k in seq_len(node_group_count)) {

    # Create empty table with position and node ID
    position_table <- dplyr::tibble(x = numeric(), y = numeric())

    node_group <- names(nodes)[k]
    node_select <- nodes[[k]]

    node_attr <- unlist(stringr::str_split(node_select, ":"))[1]
    node_attr_val <- unlist(stringr::str_split(node_select, ":"))[2]

    sort_attr <- unlist(stringr::str_split(sort[k], ":"))[1]
    sort_dir <- unlist(stringr::str_split(sort[k], ":"))[2]

    for (i in seq_len(layout_row_number)) {
      for (j in seq_len(layout_column_number)) {

        if (unlist(stringr::str_split(layout[i], ""))[j] == k) {

          item_table <-
            dplyr::tibble(
              x = x_pts[j],
              y = y_pts[i])

          position_table <-
            position_table |>
            dplyr::bind_rows(item_table)
        }
      }
    }

    # Filter the graph `ndf`
    ndf_part <-
      ndf |>
      dplyr::filter(!!rlang::parse_expr(paste0(node_attr, " == '", node_attr_val, "'")))

    # Optionally apply sorting
    if (!is.null(sort)) {
      if (sort_dir == "desc") {
        ndf_part <-
          ndf_part |>
          dplyr::arrange(dplyr::desc(!!sym(sort_attr)))

      } else {
        ndf_part <-
          ndf_part |>
          dplyr::arrange(!!sym(sort_attr))
      }
    }

    # Trim rows from the position table if its
    # row count is greater than that of `ndf_part`
    if (nrow(position_table) > nrow(ndf_part)) {
      position_table <- position_table[seq_len(nrow(ndf_part)), ]
    }

    # Bind `position_table` to the split `ndf_part`
    ndf_part <-
      dplyr::bind_cols(position_table, ndf_part)

    # Bind `ndf_part` to `ndf_parts`
    ndf_parts <-
      dplyr::bind_rows(ndf_parts, ndf_part)
  }

  # Join the `ndf_parts` to the main `ndf`
  ndf <-
    ndf |>
    dplyr::left_join(
      ndf_parts,
      by = c("id", "type", "label"))

  # Replace the graph's ndf with the revised version
  graph$nodes_df <- ndf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Apply colors based on node attribute values
#'
#' @description
#'
#' Within a graph's internal node data frame (ndf), use a categorical node
#' attribute to generate a new node attribute with color values.
#'
#' @inheritParams render_graph
#' @param node_attr_from The name of the node attribute column from which color
#'   values will be based.
#' @param node_attr_to The name of the new node attribute to which the color
#'   values will be applied.
#' @param cut_points An optional vector of numerical breaks for bucketizing
#'   continuous numerical values available in a edge attribute column.
#' @param palette Can either be: (1) a palette name from the RColorBrewer
#'   package (e.g., `Greens`, `OrRd`, `RdYlGn`), (2) `viridis`, which indicates
#'   use of the `viridis` color scale from the package of the same name, or (3)
#'   a vector of hexadecimal color names.
#' @param alpha An optional alpha transparency value to apply to the generated
#'   colors. Should be in the range of `0` (completely transparent) to `100`
#'   (completely opaque).
#' @param reverse_palette An option to reverse the order of colors in the chosen
#'   palette. The default is `FALSE`.
#' @param default_color A hexadecimal color value to use for instances when the
#'   values do not fall into the bucket ranges specified in the `cut_points`
#'   vector.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 8
#' # nodes and 7 edges
#' graph <-
#'   create_graph() |>
#'   add_path(n = 8) |>
#'   set_node_attrs(
#'     node_attr = weight,
#'     values = c(
#'       8.2, 3.7, 6.3, 9.2,
#'       1.6, 2.5, 7.2, 5.4))
#'
#' # Find group membership values for all nodes
#' # in the graph through the Walktrap community
#' # finding algorithm and join those group values
#' # to the graph's internal node data frame (ndf)
#' # with the `join_node_attrs()` function
#' graph <-
#'   graph |>
#'   join_node_attrs(
#'     df = get_cmty_walktrap(graph))
#'
#' # Inspect the number of distinct communities
#' graph |>
#'   get_node_attrs(
#'     node_attr = walktrap_group) |>
#'   unique() |>
#'   sort()
#'
#' # Visually distinguish the nodes in the different
#' # communities by applying colors using the
#' # `colorize_node_attrs()` function; specifically,
#' # set different `fillcolor` values with an alpha
#' # value of 90 and apply opaque colors to the node
#' # border (with the `color` node attribute)
#' graph <-
#'   graph |>
#'   colorize_node_attrs(
#'     node_attr_from = walktrap_group,
#'     node_attr_to = fillcolor,
#'     palette = "Greens",
#'     alpha = 90) |>
#'   colorize_node_attrs(
#'     node_attr_from = walktrap_group,
#'     node_attr_to = color,
#'     palette = "viridis",
#'     alpha = 80)
#'
#' # Show the graph's internal node data frame
#' graph |> get_node_df()
#'
#' # Create a graph with 8 nodes and 7 edges
#' graph <-
#'   create_graph() |>
#'   add_path(n = 8) |>
#'   set_node_attrs(
#'     node_attr = weight,
#'     values = c(
#'       8.2, 3.7, 6.3, 9.2,
#'       1.6, 2.5, 7.2, 5.4))
#'
#' # We can bucketize values in `weight` using
#' # `cut_points` and assign colors to each of the
#' # bucketed ranges (for values not part of any
#' # bucket, a gray color is assigned by default)
#' graph <-
#'   graph |>
#'   colorize_node_attrs(
#'     node_attr_from = weight,
#'     node_attr_to = fillcolor,
#'     cut_points = c(1, 3, 5, 7, 9))
#'
#' # Now there will be a `fillcolor` node attribute
#' # with distinct colors (the `#D9D9D9` color is
#' # the default `gray85` color)
#' graph |> get_node_df()
#'
#' @family node creation and removal
#' @export
colorize_node_attrs <- function(
    graph,
    node_attr_from,
    node_attr_to,
    cut_points = NULL,
    palette = "Spectral",
    alpha = NULL,
    reverse_palette = FALSE,
    default_color = "#D9D9D9"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the requested `node_attr_from`
  node_attr_from <-
    rlang::enquo(node_attr_from) |> rlang::get_expr() |> as.character()

  # Get the requested `node_attr_to`
  node_attr_to <-
    rlang::enquo(node_attr_to) |> rlang::get_expr() |> as.character()

  # Extract ndf from graph
  nodes_df <- graph$nodes_df

  # Get the column number in the ndf from which to
  # recode values
  col_to_recode_no <-
    which(colnames(nodes_df) %in% node_attr_from)

  # Get the number of recoded values
  if (is.null(cut_points)) {
    num_recodings <-
      nrow(unique(nodes_df[col_to_recode_no]))
  } else {
    num_recodings <- length(cut_points) - 1
  }

  # Handle vector of hexadecimal or named colors
  if (length(palette) > 1) {
    # Verify colors are valid
    is_valid_hex <- grepl(toupper(palette), pattern = "#[0-9A-F]{6}")

    if (!all(is_valid_hex)) {
      cli::cli_abort(
        "The color palette contains invalid hexadecimal values.")
    }

    if (length(palette) < num_recodings) {
      # Revert to viridis if provided color vector is too short
      palette <- "viridis"
    } else {
      color_palette <- toupper(palette)[1:num_recodings]
    }
  }

  # Handle viridis and ColorBrewer palette name input
  if (length(palette) == 1) {
    # If the number of recodings lower than any Color
    # Brewer palette, shift palette to `viridis`
    if ((num_recodings < 3 || num_recodings > 10) &&
        palette %in% c(row.names(RColorBrewer::brewer.pal.info))) {
      palette <- "viridis"
    }

    # or any of the RColorBrewer palettes
    if (!(palette %in% c(row.names(RColorBrewer::brewer.pal.info),
                         "viridis"))) {
      cli::cli_abort(
        "The color palette is not an `RColorBrewer` or `viridis` palette.")
    }

    # Obtain a color palette
    if (palette %in% row.names(RColorBrewer::brewer.pal.info)) {
      color_palette <- RColorBrewer::brewer.pal(num_recodings, palette)
    } else if (palette == "viridis") {
      color_palette <- viridisLite::viridis(num_recodings)
      color_palette <- gsub("..$", "", color_palette)
    }
  }

  # Reverse color palette if `reverse_palette = TRUE`
  if (reverse_palette) {
    color_palette <- rev(color_palette)
  }

  # Create a data frame with initial values
  new_node_attr_col <-
    data.frame(
      node_attr_to = rep(default_color, nrow(nodes_df)),
      stringsAsFactors = FALSE)

  # Get the column number for the new node attribute
  to_node_attr_colnum <- ncol(nodes_df) + 1

  # Bind the current ndf with the new column
  nodes_df <- cbind(nodes_df, new_node_attr_col)

  # Rename the new column with the target node attr name
  colnames(nodes_df)[to_node_attr_colnum] <- node_attr_to

  # Get a data frame of recodings
  if (is.null(cut_points)) {

    recode_df <-
      data.frame(
        to_recode = names(table(nodes_df[, col_to_recode_no])),
        colors = color_palette,
        stringsAsFactors = FALSE)

    # Recode rows in the new node attribute
    for (i in seq_along(names(table(nodes_df[, col_to_recode_no])))) {

      recode_rows <-
        which(nodes_df[, col_to_recode_no] %in%
                recode_df[i, 1])

      if (is.null(alpha)) {
        nodes_df[recode_rows, to_node_attr_colnum] <-
          color_palette[i]
      } else if (!is.null(alpha)) {
        if (alpha < 100) {
          nodes_df[recode_rows, to_node_attr_colnum] <-
            gsub("$", alpha, color_palette[i])
        } else if (alpha == 100) {
          nodes_df[recode_rows, to_node_attr_colnum] <-
            gsub("$", "", color_palette[i])
        }
      }
    }
  }

  # Recode according to provided cut points
  if (!is.null(cut_points)) {
    for (i in seq_len(length(cut_points) - 1)) {
      recode_rows <-
        which(
          as.numeric(nodes_df[, col_to_recode_no]) >=
            cut_points[i] &
            as.numeric(nodes_df[, col_to_recode_no]) <
            cut_points[i + 1])

      nodes_df[recode_rows, to_node_attr_colnum] <-
        color_palette[i]
    }

    if (!is.null(alpha)) {
      if (alpha < 100) {
        nodes_df[, to_node_attr_colnum] <-
          gsub("$", alpha, nodes_df[, to_node_attr_colnum])
      } else if (alpha == 100) {
        nodes_df[, to_node_attr_colnum] <-
          gsub("$", "", nodes_df[, to_node_attr_colnum])
      }
    }
  }

  # Get the finalized column of values
  nodes_attr_vector_colorized <- nodes_df[, ncol(nodes_df)]

  node_attr_to_2 <- rlang::enquo(node_attr_to)

  # Set the node attribute values for nodes specified
  # in selection
  graph <-
    set_node_attrs(
      graph = graph,
      node_attr = !!node_attr_to_2,
      values = nodes_attr_vector_colorized
    )

  # Remove last action from the `graph_log`
  graph$graph_log <- graph$graph_log[1:(nrow(graph$graph_log) - 1), ]

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get count of all loop edges
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get a count of all loop edges in
#' the graph.
#'
#' @inheritParams render_graph
#'
#' @return A numeric vector of single length.
#'
#' @examples
#' # Create an undirected, full graph
#' # with 3 nodes and all possible
#' # edges, including loop edges
#' graph <-
#'   create_graph(
#'     directed = FALSE) %>%
#'   add_full_graph(
#'     n = 3,
#'     keep_loops = TRUE)
#'
#' # Get a count of all loop edges
#' # in the graph
#' graph %>% count_loop_edges()
#'
#' @export
count_loop_edges <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If graph is empty, return 0
  if (is_graph_empty(graph)) {
    return(0L)
  }

  n_match <- length(which(graph$edges_df$from == graph$edges$to))
  n_match
}

#' Is the edge mutual with another edge?
#'
#' @description
#'
#' Determines whether an edge definition has a mutual analogue with the same
#' node pair.
#'
#' @inheritParams render_graph
#' @param edge A numeric edge ID value.
#'
#' @return A logical value.
#'
#' @examples
#' # Create a graph that has mutual
#' # edges across some node pairs
#' graph <-
#'   create_graph() |>
#'   add_path(n = 4) |>
#'   add_edge(
#'     from = 4,
#'     to = 3) |>
#'   add_edge(
#'     from = 2,
#'     to = 1)
#'
#' # Get the graph's internal
#' # edge data frame
#' graph |> get_edge_df()
#'
#' # Determine if edge `1` has
#' # a mutual edge
#' graph |>
#'   is_edge_mutual(edge = 1)
#'
#' # Determine if edge `2` has
#' # a mutual edge
#' graph |>
#'   is_edge_mutual(edge = 2)
#'
#' @export
is_edge_mutual <- function(
    graph,
    edge
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Stop function if `edge` is not a single numeric value.
  check_number_decimal(edge)

  # Obtain the graph's edf
  edf <- graph$edges_df

  # Stop function if the edge ID provided
  # is not a valid edge ID
  if (!(edge %in% edf$id)) {

    cli::cli_abort(
      "The provided edge ID is not present in the graph.")
  }

  # Obtain the edge definition
  from <-
    edf |>
    dplyr::filter(id == !!edge) |>
    dplyr::pull("from")

  to <-
    edf |>
    dplyr::filter(id == !!edge) |>
    dplyr::pull("to")

  # Determine if there is any row where
  # the definition of `from` and `to` is
  # reversed
  mutual_edges <-
    edf |>
    dplyr::filter(from == !!to, to == !!from)

  res <- nrow(mutual_edges) > 0
  res
}

#' Get a graph available in a series
#'
#' @description
#'
#' Using a graph series object of type `dgr_graph_1D`, get a graph object.
#'
#' @param graph_series A graph series object of type `dgr_graph_1D`.
#' @param graph_no The index of the graph in the graph series.
#'
#' @examples
#' # Create three graphs
#' graph_1 <-
#'   create_graph() |>
#'   add_path(n = 4)
#'
#' graph_2 <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' graph_3 <-
#'   create_graph() |>
#'   add_star(n = 6)
#'
#' # Create an empty graph series
#' # and add the graphs
#' series <-
#'   create_graph_series() |>
#'   add_graph_to_graph_series(
#'     graph = graph_1) |>
#'   add_graph_to_graph_series(
#'     graph = graph_2) |>
#'   add_graph_to_graph_series(
#'     graph = graph_3)
#'
#' # Get the second graph in the series
#' extracted_graph <-
#'   series |>
#'   get_graph_from_graph_series(
#'     graph_no = 2)
#'
#' @export
get_graph_from_graph_series <- function(
    graph_series,
    graph_no
) {

  # Stop function if no graphs are available
  if (is.null(graph_series$graphs)) {

    rlang::abort("There are no graphs in this graph series.")
  }

  # Stop function if `graph_no` out of range
  if (!(graph_no %in% seq_len(count_graphs_in_graph_series(graph_series)))) {

    rlang::abort(
      "The index chosen doesn't correspond to that of a graph in the series.")
  }

  # Extract the specified graph from the series
  graph_series$graphs[[graph_no]]
}

#' R + mermaid.js
#'
#' Make diagrams in R using
#' \href{https://github.com/mermaid-js/mermaid/wiki}{mermaid.js} with
#' infrastructure provided by \href{http://www.htmlwidgets.org/}{htmlwidgets}.
#'
#' @param diagram Diagram in mermaid markdown-like language or file (as a
#'   connection or file name) containing a diagram specification. If no diagram
#'   is provided `diagram = ""` then the function will assume that a diagram
#'   will be provided by [htmltools::tags()] and `DiagrammeR` is just being used
#'   for dependency injection.
#' @param ... Other arguments and parameters you would like to send to
#'   JavaScript.
#' @param width The width of the resulting graphic in pixels.
#' @param height The height of the resulting graphic in pixels.
#'
#' @return An object of class `htmlwidget` that will intelligently print itself
#'   into HTML in a variety of contexts including the R console, within R
#'   Markdown documents, and within Shiny output bindings.
#'
#' @examples
#' # Create a simple graph running left to right (note
#' # that the whitespace is not important)
#' # DiagrammeR("
#' #   graph LR
#' #     A-->B
#' #     A-->C
#' #     C-->E
#' #     B-->D
#' #     C-->D
#' #     D-->F
#' #     E-->F
#' # ")
#'
#' # Create the equivalent graph but have it running
#' # from top to bottom
#' # DiagrammeR("
#' #    graph TB
#' #    A-->B
#' #    A-->C
#' #    C-->E
#' #    B-->D
#' #    C-->D
#' #    D-->F
#' #    E-->F
#' # ")
#'
#' # Create a graph with different node shapes and
#' # provide fill styles for each node
#' # DiagrammeR("graph LR;A(Rounded)-->B[Squared];B-->C{A Decision};
#' #  C-->D[Square One];C-->E[Square Two];
#' #  style A fill:#E5E25F;  style B fill:#87AB51; style C fill:#3C8937;
#' #  style D fill:#23772C;  style E fill:#B6E6E6;"
#' # )
#'
#' # Load in the 'mtcars' dataset
#' # data(mtcars)
#' # connections <- sapply(
#' #  1:ncol(mtcars)
#' #   ,function(i) {
#' #      paste0(
#' #         i
#' #       ,"(",colnames(mtcars)[i],")---"
#' #       ,i,"-stats("
#' #       ,paste0(
#' #         names(summary(mtcars[,i]))
#' #         ,": "
#' #         ,unname(summary(mtcars[,i]))
#' #         ,collapse="<br/>"
#' #       )
#' #       ,")"
#' #    )
#' #  }
#' # )
#'
#' # Create a diagram using the 'connections' object
#' # DiagrammeR(
#' #    paste0(
#' #      "graph TD;", "\n",
#' #      paste(connections, collapse = "\n"),"\n",
#' #      "classDef column fill:#0001CC, stroke:#0D3FF3, stroke-width:1px;" ,"\n",
#' #      "class ", paste0(1:length(connections), collapse = ","), " column;"
#' #    )
#' #  )
#'
#' # Also with `DiagrammeR()`, you can use tags
#' # from `htmltools` (just make sure to use
#' # `class = "mermaid"`)
#' library(htmltools)
#' # diagramSpec = "
#' # graph LR;
#' #   id1(Start)-->id2(Stop);
#' #   style id1 fill:#f9f,stroke:#333,stroke-width:4px;
#' #   style id2 fill:#ccf,stroke:#f66,stroke-width:2px,stroke-dasharray: 5, 5;
#' # "
#' # html_print(tagList(
#' #   tags$h1("R + mermaid.js = Something Special")
#' #   ,tags$pre(diagramSpec)
#' #   ,tags$div(class="mermaid", diagramSpec)
#' #   ,DiagrammeR()
#' # ))
#'
#' # Create a sequence diagram
#' # DiagrammeR("
#' # sequenceDiagram;
#' #    customer->>ticket seller: ask for a ticket;
#' #    ticket seller->>database: seats;
#' #    alt tickets available
#' #      database->>ticket seller: ok;
#' #      ticket seller->>customer: confirm;
#' #      customer->>ticket seller: ok;
#' #      ticket seller->>database: book a seat;
#' #      ticket seller->>printer: print a ticket;
#' #    else sold out
#' #      database->>ticket seller: none left;
#' #      ticket seller->>customer: sorry;
#' #    end
#' # ")
#'
#' @export
mermaid <- function(
    diagram = "",
    ...,
    width = NULL,
    height = NULL
) {

  # Check for a connection or file
  is_connection_or_file <-
    inherits(diagram[1], "connection") || file.exists(diagram[1])

  # Obtain the diagram text via `readLines()`
  if (is_connection_or_file) {

    diagram <- readLines(diagram, encoding = "UTF-8", warn = FALSE)
    diagram <- paste0(diagram, collapse = "\n")

  } else {

    # Check for vector with length > 1 and concatenate
    if (length(diagram) > 1) {

      nosep <- grep("[;\n]", diagram)

      if (length(nosep) < length(diagram)) {

        diagram[-nosep] <-
          sapply(diagram[-nosep], function(c) paste0(c, ";"))
      }

      diagram <- paste0(diagram, collapse = "")
    }
  }

  # Forward options using x
  x <- list(diagram = diagram)

  # Create widget
  htmlwidgets::createWidget(
    name = "DiagrammeR",
    x = x,
    width = width,
    height = height,
    package = "DiagrammeR"
  )
}

# Creating objects with DiagrammeR, mermaid, and grViz")

test_that("htmlwidgets object can be created", {

  diagrammer_htmlwidget <-
    DiagrammeR("
  graph LR
    A-->B
    A-->C
    C-->E
    B-->D
    C-->D
    D-->F
    E-->F")

  # Expect that the object inherits
  # from `DiagrammeR` and `htmlwidget`
  expect_s3_class(
    diagrammer_htmlwidget, c("DiagrammeR", "htmlwidget"))

  mermaid_htmlwidget <-
    mermaid("
  graph LR
    A-->B
    A-->C
    C-->E
    B-->D
    C-->D
    D-->F
    E-->F")

  # Expect that the object inherits
  # from `DiagrammeR` and `htmlwidget`
  expect_s3_class(
    mermaid_htmlwidget, c("DiagrammeR", "htmlwidget"))
})

#' Get Dice similarity coefficient scores
#'
#' @description
#'
#' Get the Dice similarity coefficient scores for one or more nodes in a graph.
#'
#' @inheritParams render_graph
#' @param nodes an optional vector of node IDs to consider for Dice similarity
#'   scores. If not supplied, then similarity scores will be provided for every
#'   pair of nodes in the graph.
#' @param direction using `all` (the default), the function will ignore
#'   edge direction when determining scores for neighboring nodes. With
#'   `out` and `in`, edge direction for neighboring nodes will be
#'   considered.
#' @param round_to the maximum number of decimal places to retain for the Dice
#'   similarity coefficient scores. The default value is `3`.
#'
#' @return A matrix with Dice similarity values for each pair of nodes
#'   considered.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the Dice similarity
#' # values for nodes `5`, `6`,
#' # and `7`
#' graph |>
#'   get_dice_similarity(
#'     nodes = 5:7)
#'
#' @export
get_dice_similarity <- function(
    graph,
    nodes = NULL,
    direction = "all",
    round_to = 3
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  if (is.null(nodes)) {
    ig_nodes <- igraph::V(ig_graph)
  }

  if (!is.null(nodes)) {

    # Stop function if nodes provided not in
    # the graph
    if (!all(nodes %in% get_node_ids(graph))) {

      cli::cli_abort(
        "One or more nodes provided not in graph.")
    }

    # Get an igraph representation of node ID values
    ig_nodes <- igraph::V(ig_graph)[nodes]
  }

  # Get the Dice similarity scores
  if (direction == "all") {
    d_sim_values <-
      as.matrix(
        igraph::similarity(
          ig_graph,
          vids = ig_nodes,
          mode = "all",
          method = "dice"))
  }

  if (direction == "out") {
    d_sim_values <-
      as.matrix(
        igraph::similarity(
          ig_graph,
          vids = ig_nodes,
          mode = "out",
          method = "dice"))
  }

  if (direction == "in") {
    d_sim_values <-
      as.matrix(
        igraph::similarity(
          ig_graph,
          vids = ig_nodes,
          mode = "in",
          method = "dice"))
  }

  if (is.null(nodes)) {
    row.names(d_sim_values) <- graph$nodes_df$id
    colnames(d_sim_values) <- graph$nodes_df$id
  }

  if (!is.null(nodes)) {
    row.names(d_sim_values) <- graph$nodes_df$id[nodes]
    colnames(d_sim_values) <- graph$nodes_df$id[nodes]
  }

  # Round all values in matrix to set SD
  round(d_sim_values, round_to)
}

#' Create a complement of a graph
#'
#' @description
#'
#' Create a complement graph which contains only edges not present in the input
#' graph. It's important to nodes that any edge attributes in the input graph's
#' edges will be lost. Node attributes will be retained, since they are not
#' affected by this transformation.
#'
#' @inheritParams render_graph
#' @param loops An option for whether loops should be generated in the
#'   complement graph.
#'
#' @return a graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a simple graph
#' # with a single cycle
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 4)
#'
#' # Get the graph's edge
#' # data frame
#' graph |> get_edge_df()
#'
#' # Create the complement
#' # of the graph
#' graph_c <-
#'   graph |>
#'     transform_to_complement_graph()
#'
#' # Get the edge data frame
#' # for the complement graph
#' graph_c |> get_edge_df()
#'
#' @export
transform_to_complement_graph <- function(
    graph,
    loops = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph |> count_nodes()

  # Get the number of edges ever created for
  # this graph
  edges_created <- graph$last_edge

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the complement graph
  ig_graph <- igraph::complementer(ig_graph, loops = loops)

  # Get the edge data frame for the complement graph
  edf_new <- from_igraph(ig_graph) |> get_edge_df()

  # Add edge ID values to the complement graph edf
  edf_new$id <- seq_len(nrow(edf_new))

  # Replace the input graph's edf with its complement
  graph$edges_df <- edf_new

  # Manually update the graph's edge counter
  graph$last_edge <- nrow(edf_new)

  # Scavenge any invalid, linked data frames
  graph <-
    remove_linked_dfs(graph)

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph |> count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph |> count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_added,
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Join new edge attribute values using a data frame
#'
#' @description
#'
#' Join new edge attribute values in a left join using a data frame. The use of
#' a left join in this function allows for no possibility that edges in the
#' graph might be removed after the join.
#'
#' @inheritParams render_graph
#' @param df The data frame to use for joining.
#' @param by_graph Optional specification of the column in the graph's internal
#'   edge data frame for the left join. If both `by_graph` and `by_df` are not
#'   provided, then a natural join will occur if there are columns in the
#'   graph's edf and in `df` with identical names.
#' @param by_df Optional specification of the column in `df` for the left join.
#'   If both `by_graph` and `by_df` are not provided, then a natural join will
#'   occur if there are columns in the graph's edf and in `df` with identical
#'   names.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a simple graph
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(n = 5) %>%
#'   add_edges_w_string(
#'     edges = "1->2 1->3 2->4 2->5 3->5")
#'
#' # Create a data frame with node ID values
#' # representing the graph edges (with `from` and `to`
#' # columns), and, a set of numeric values
#' df <-
#'   data.frame(
#'     from = c(1, 1, 2, 2, 3),
#'     to = c(2, 3, 4, 5, 5),
#'     values = rnorm(5, 5))
#'
#' # Join the values in the data frame to the
#' # graph's edges; this works as a left join using
#' # identically-named columns in the graph and the df
#' # (in this case `from` and `to` are common to both)
#' graph <-
#'   graph %>%
#'   join_edge_attrs(
#'     df = df)
#'
#' # Get the graph's internal edf to show that the
#' # join has been made
#' graph %>% get_edge_df()
#' @family edge creation and removal
#' @export
join_edge_attrs <- function(
    graph,
    df,
    by_graph = NULL,
    by_df = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  if (is.null(by_graph) && !is.null(by_df)) {

    cli::cli_abort(
      "Both column specifications must be provided.")
  }

  if (!is.null(by_graph) && is.null(by_df)) {

    cli::cli_abort(
      "Both column specifications must be provided.")
  }

  # Extract the graph's edf
  edges <- get_edge_df(graph)

  # Get column names from the graph's edf
  column_names <- colnames(edges)

  if (is.null(by_graph) && is.null(by_df)) {

    # Perform a left join on the `edges` data frame
    edges <- merge(edges, df, all.x = TRUE)
  }

  if (!is.null(by_graph) && !is.null(by_df)) {

    # Perform a left join on the `edges` data frame
    edges <-
      merge(
        edges, df,
        all.x = TRUE,
        by.x = by_graph,
        by.y = by_df)
  }

  # Get new column names in the revised edf
  new_col_names <-
    base::setdiff(colnames(edges), column_names)

  # Sort the columns in `edges`
  edges <-
    edges %>% dplyr::relocate("id", "from", "to", "rel")

  # Modify the graph object
  graph$edges_df <- edges

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' R + viz.js
#'
#' Make diagrams in R using `viz.js` with infrastructure provided by
#' \pkg{htmlwidgets}.
#'
#' @param diagram Spec for a diagram as either text, filename string, or file
#'   connection.
#' @param engine String for the Graphviz layout engine; can be `dot` (default),
#'   `neato`, `circo`, or `twopi`.
#' @param allow_subst A boolean that enables/disables substitution
#'   functionality.
#' @param options Parameters supplied to the htmlwidgets framework.
#' @param width An optional parameter for specifying the width of the resulting
#'   graphic in pixels.
#' @param height An optional parameter for specifying the height of the
#'   resulting graphic in pixels.
#' @param envir The environment in which substitution functionality takes place.
#'
#' @return An object of class `htmlwidget` that will intelligently print itself
#'   into HTML in a variety of contexts including the R console, within R
#'   Markdown documents, and within Shiny output bindings.
#'
#' @export
grViz <- function(diagram = "",
                  engine = "dot",
                  allow_subst = TRUE,
                  options = NULL,
                  width = NULL,
                  height = NULL,
                  envir = parent.frame()) {

  # Check for a connection or file
  is_connection_or_file <-
    inherits(diagram[1], "connection") || file.exists(diagram[1])

  # Obtain the diagram text via `readLines()`
  if (is_connection_or_file) {
    diagram <- readLines(diagram, encoding = "UTF-8", warn = FALSE)
  }

  diagram <- paste0(diagram, collapse = "\n")

  # If allow_subst = TRUE
  if (allow_subst) {
    diagram <- replace_in_spec(diagram, envir = envir)
  }

  # Single quotes within a diagram spec are problematic
  # so try to replace with `\"`
  diagram <- gsub(x = diagram, "'", "\"")

  # Forward options using `x`
  x <-
    list(
      diagram = diagram,
      config = list(
        engine = engine,
        options = options
      )
    )

  # Only use the Viewer for newer versions of RStudio,
  # but allow other, non-RStudio viewers
  viewer.suppress <-
    rstudioapi::isAvailable() &&
    !rstudioapi::isAvailable("0.99.120")

  # Create the widget
  htmlwidgets::createWidget(
    name = "grViz",
    x = x,
    width = width,
    height = height,
    package = "DiagrammeR",
    htmlwidgets::sizingPolicy(viewer.suppress = viewer.suppress)
  )
}

#' Widget output function for use in Shiny
#'
#' @param outputId Output variable to read from.
#' @param width A valid CSS unit for the width or a number, which will be
#'   coerced to a string and have `px` appended.
#' @param height A valid CSS unit for the height or a number, which will be
#'   coerced to a string and have `px` appended.
#'
#' @export
grVizOutput <- function(outputId,
                        width = "100%",
                        height = "400px") {

  htmlwidgets::shinyWidgetOutput(
    outputId = outputId,
    name = "grViz",
    width = width,
    height = height,
    package = "DiagrammeR"
  )
}

#' Widget render function for use in Shiny
#'
#' @param expr an expression that generates a DiagrammeR graph
#' @param env the environment in which to evaluate expr.
#' @param quoted is expr a quoted expression (with quote())? This is useful if
#'   you want to save an expression in a variable.
#' @seealso [grVizOutput()] for an example in Shiny.
#' @export
renderGrViz <- function(expr,
                        env = parent.frame(),
                        quoted = FALSE) {

  if (!quoted) expr <- substitute(expr)

  htmlwidgets::shinyRenderWidget(
    expr = expr,
    outputFunction = grVizOutput,
    env = env,
    quoted = TRUE
  )
}

#' Add MathJax-formatted equation text
#'
#' @param gv A `grViz` htmlwidget.
#' @param include_mathjax A `logical` to add mathjax JS. Change to `FALSE` if
#'   using with \pkg{rmarkdown} since MathJax will likely already be added.
#'
#' @return A `grViz` htmlwidget
#'
#' @keywords internal
#' @export
add_mathjax <- function(gv = NULL,
                        include_mathjax = TRUE) {

  stopifnot(!is.null(gv), inherits(gv, "grViz"))

  gv$dependencies <-
    c(
      gv$dependencies,
      list(htmltools::htmlDependency(
        name = "svg_mathjax2",
        version = "0.1.0",
        src = c(href = "https://cdn.rawgit.com/timelyportfolio/svg_mathjax2/master/"),
        script = "svg_mathjax2.js")))

  if (include_mathjax) {
    htmltools::browsable(
      htmltools::tagList(
        gv,
        htmltools::tags$script(src = "http://cdn.mathjax.org/mathjax/latest/MathJax.js?config=TeX-AMS-MML_SVG"),
        htmlwidgets::onStaticRenderComplete(
          "setTimeout(function(){new Svg_MathJax().install()}, 4000);"
        )
      ))
  } else {
    htmltools::browsable(htmltools::tagList(
      gv,
      htmlwidgets::onStaticRenderComplete(
        "setTimeout(function(){new Svg_MathJax().install()}, 4000);"
      )
    ))
  }
}

#' Get the minimum graph eccentricity
#'
#' @description
#'
#' Get the radius of a graph, which is the smallest eccentricity in the graph.
#' The graph eccentricity of a node is its shortest path from the farthest other
#' node in the graph.
#'
#' @inheritParams render_graph
#' @param direction Using `all` (the default), the search will ignore edge
#'   direction while traversing through the graph. With `out`, measurements of
#'   paths will be from a node whereas with `in`, measurements of paths will be
#'   to a node.
#'
#' @return A single numeric value representing the minimum eccentricity of the
#'   graph.
#'
#' @examples
#' # Create a cycle graph
#' graph <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' # Determine the graph's minimum
#' # eccentricity
#' graph %>%
#'   get_min_eccentricity()
#'
#' # Create a full graph and then
#' # get the minimum eccentricity
#' # value for that
#' create_graph() %>%
#'   add_full_graph(n = 10) %>%
#'   get_min_eccentricity()
#'
#' @export
get_min_eccentricity <- function(
    graph,
    direction = "all"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Ensure that values provided for the
  # `direction` argument are from the
  # valid options
  rlang::arg_match0(direction, c("all", "in", "out"))

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the minimum eccentricity of the graph
  igraph::radius(ig_graph, mode = direction)
}

#' Get the graph's edf filtered by a selection of edges
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get the graph's internal edge data
#' frame that is filtered by the edge ID values currently active as a selection.
#'
#' This function makes use of an active selection of edges (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#'
#' @return an edge data frame.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 4,
#'     m = 4,
#'     set_seed = 23) |>
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = c(2.5, 8.2, 4.2, 2.4))
#'
#' # Select edges with ID values
#' # `1` and `3`
#' graph <-
#'   graph |>
#'   select_edges_by_edge_id(
#'     edges = c(1, 3))
#'
#' # Get the edge data frame that's
#' # limited to the rows that correspond
#' # to the edge selection
#' graph |> get_edge_df_ws()
#'
#' @export
get_edge_df_ws <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph object has a valid edge selection
  check_graph_contains_edge_selection(graph)

  # Extract the edge data frame (edf)
  # from the graph and get only those edges
  # from the edges selection
  graph$edges_df |>
    dplyr::filter(id %in% graph$edge_selection$edge)
}

#' Insert edge data attributes during edge creation
#'
#' @description
#'
#' This helper function should be invoked to provide values for the namesake
#' `edge_data` argument, which is present in any function where edges are
#' created.
#'
#' @param ... Edge data attributes provided as one or more named vectors.
#'
#' @examples
#' \dontrun{
#' # Create a new graph and add
#' # a path with several edge
#' # data attributes
#' graph <-
#'   create_graph() %>%
#'   add_path(
#'     n = 3,
#'     type = "path",
#'     edge_data = edge_data(
#'       hour = 5,
#'       index = c(1, 2)))
#'
#' # View the graph's internal
#' # edge data frame; the edge
#' # data attributes have
#' # been inserted
#' graph %>% get_edge_df()
#' }
#'
#' @family edge creation and removal
#'
#' @export
edge_data <- function(...) {

  # Collect vectors of edge data
  # attribute values into a list object
  edge_data_values <- list(...)

  if (any(names(edge_data_values) %in% gv_edge_attributes())) {

    cli::cli_abort(
      "Names for edge data attributes shouldn't be any of those reserved for edge aesthetic attributes.")
  }

  edge_data_values
}

#' Insert node data attributes during node creation
#'
#' @description
#'
#' This helper function should be invoked to provide values for the namesake
#' `node_data` argument, which is present in any function where nodes are
#' created.
#'
#' @param ... Node data attributes provided as one or more named vectors.
#' @examples
#' # Create a new graph and add
#' # a path with several node
#' # data attributes
#' graph <-
#'   create_graph() |>
#'   add_path(
#'     n = 3,
#'     type = "path",
#'     node_data = node_data(
#'       hour = 5,
#'       index = c(1, 3, 2)))
#'
#' # View the graph's internal
#' # node data frame; the node
#' # data attributes have been
#' # inserted
#' graph |> get_node_df()
#' @family node creation and removal
#' @export
node_data <- function(...) {

  # Collect vectors of node data
  # attribute values into a list object
  node_data_values <- list(...)

  if (any(names(node_data_values) %in% c(gv_node_attributes(), "x", "y"))) {

    cli::cli_abort(
      "Names for node data attributes shouldn't be any of those reserved for node aesthetic attributes.")
  }

  node_data_values
}

#' Add clones of a selection of nodes
#'
#' @description
#'
#' Add new nodes to a graph object of class `dgr_graph` which are clones of
#' nodes in an active selection of nodes. All node attributes are preserved
#' except for the node `label` attribute (to maintain the uniqueness of non-`NA`
#' node label values). A vector of node `label` can be provided to bind new
#' labels to the cloned nodes.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param add_edges An option for whether to add edges from the selected nodes
#'   to each of their clones, or, in the opposite direction.
#' @param direction Using `from` will create new edges from existing nodes to
#'   the new, cloned nodes. The `to` option will create new edges directed
#'   toward the existing nodes.
#' @param label An optional vector of node label values. The vector length
#'   should correspond to the number of nodes in the active selection of nodes.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with a path of
#' # nodes; supply `label`, `type`,
#' # and `value` node attributes,
#' # and select the created nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(
#'     n = 3,
#'     label = c("d", "g", "r"),
#'     type = c("a", "b", "c")) %>%
#'   select_last_nodes_created()
#'
#' # Display the graph's internal
#' # node data frame
#' graph %>% get_node_df()
#'
#' # Create clones of all nodes
#' # in the selection but assign
#' # new node label values
#' # (leaving `label` as NULL
#' # yields NA values)
#' graph <-
#'   graph %>%
#'   add_node_clones_ws(
#'     label = c("a", "b", "v"))
#'
#' # Display the graph's internal
#' # node data frame: nodes `4`,
#' # `5`, and `6` are clones of
#' # `1`, `2`, and `3`
#' graph %>% get_node_df()
#'
#' # Select the last nodes
#' # created (`4`, `5`, and `6`)
#' # and clone those nodes and
#' # their attributes while
#' # creating new edges between
#' # the new and existing nodes
#' graph <-
#'   graph %>%
#'   select_last_nodes_created() %>%
#'   add_node_clones_ws(
#'     add_edges = TRUE,
#'     direction = "to",
#'     label = c("t", "z", "s"))
#'
#' # Display the graph's internal
#' # edge data frame; there are
#' # edges between the selected
#' # nodes and their clones
#' graph %>% get_edge_df()
#'
#' @family node creation and removal
#'
#' @export
add_node_clones_ws <- function(
    graph,
    add_edges = FALSE,
    direction = NULL,
    label = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, extra_msg = "So, clones of nodes cannot be added.")

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # # Stop function if vector provided for label but it
  # # is not of length `n`
  # if (!is.null(label)) {
  #   if (length(label) != n) {
  #     stop(
  #       "The vector provided for `label` is not the same length as the value of `n`."),
  #       call. = FALSE
  #   }
  # }

  # Get the value for the latest `version_id` for
  # graph (in the `graph_log`)
  current_graph_log_version_id <-
    max(graph$graph_log$version_id)

  # Get the number of columns in the graph's
  # internal node data frame
  n_col_ndf <-
    graph %>%
    get_node_df() %>%
    ncol()

  # Get the node ID values for
  # the nodes in the active selection
  selected_nodes <- suppressMessages(get_selection(graph))

  # Clear the graph's selection
  graph <-
    suppressMessages(
      clear_selection(graph))

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph %>% count_nodes()

  # Get the number of edges in the graph
  edges_graph_1 <- graph %>% count_edges()

  node_id_value <- graph$last_node

  for (i in seq_along(selected_nodes)) {

    # Extract all of the node attributes
    # (`type` and additional node attrs)
    node_attr_vals <-
      graph %>%
      get_node_df() %>%
      dplyr::filter(id %in% selected_nodes[i]) %>%
      dplyr::select(-"id", -"label")

    # Create a clone of the selected
    # node in the graph
    graph <-
      graph %>%
      add_node(
        label = label[i])

    # Obtain the node ID value for
    # the new node
    new_node_id <-
      graph$nodes_df[nrow(graph$nodes_df), 1]

    # Create a node selection for the
    # new nodes in the graph
    graph <-
      graph %>%
      select_nodes_by_id(
        nodes = new_node_id)

    # Iteratively set node attribute values for
    # the new nodes in the graph
    for (j in seq_len(ncol(node_attr_vals))) {
      for (k in seq_along(new_node_id)) {

        graph$nodes_df[
          which(graph$nodes_df[, 1] == new_node_id[k]),
          which(colnames(graph$nodes_df) == colnames(node_attr_vals)[j])] <-
          node_attr_vals[[j]]
      }
    }

    # Create an edge if `add_edges = TRUE`
    if (add_edges) {

      if (direction == "from") {
        graph <-
          graph %>%
          add_edge(
            from = new_node_id,
            to = selected_nodes[i])
      } else {
        graph <-
          graph %>%
          add_edge(
            from = selected_nodes[i],
            to = new_node_id)
      }
    }

    # Increment the node ID value
    node_id_value <- node_id_value + 1

    # Clear the graph's active selection
    graph <-
      suppressMessages(
        clear_selection(graph))
  }

  # Remove extra items from the `graph_log`
  graph$graph_log <-
    graph$graph_log %>%
    dplyr::filter(version_id <= current_graph_log_version_id)

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph %>% count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Update the `last_node` value
  graph$last_node <- max(graph$nodes_df$id)

  # Update the `last_edge` value
  graph$last_edge <- max(graph$edges_df$id)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_added,
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get a node data frame from a graph
#'
#' @description
#'
#' From a graph, obtain a node data frame with all current node attributes.
#'
#' @inheritParams render_graph
#'
#' @return A node data frame.
#'
#' @examples
#' # Create a graph
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(
#'     n = 1,
#'     type = "a") %>%
#'   select_last_nodes_created() %>%
#'   add_n_nodes_ws(
#'     n = 5,
#'     direction = "from",
#'     type = "b") %>%
#'   select_nodes_by_id(
#'     nodes = 1) %>%
#'   set_node_attrs_ws(
#'     node_attr = value,
#'     value = 25.3) %>%
#'   clear_selection() %>%
#'   select_nodes_by_id(
#'     nodes = 2:4) %>%
#'   set_node_attrs_ws(
#'     node_attr = color,
#'     value = "grey70") %>%
#'   invert_selection() %>%
#'   set_node_attrs_ws(
#'     node_attr = color,
#'     value = "grey80") %>%
#'   clear_selection()
#'
#' # Get the graph's internal node
#' # data frame (ndf)
#' graph %>%
#'   get_node_df()
#'
#' @export
get_node_df <- function(graph) {
  check_graph_valid(graph)
  graph$nodes_df
}

#' Count graphs in a graph series object
#'
#' @description
#'
#' Counts the total number of graphs in a graph series object.
#'
#' @param graph_series A graph series object of type `dgr_graph_1D`
#'
#' @return A numeric vector representing a count of graphs in a graph series
#'   object.
#'
#' @examples
#' # Create three graphs
#' graph_1 <-
#'   create_graph() %>%
#'   add_path(n = 4)
#'
#' graph_2 <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' graph_3 <-
#'   create_graph() %>%
#'   add_star(n = 6)
#'
#' # Create an empty graph series
#' # and add the graphs
#' series <-
#'   create_graph_series() %>%
#'   add_graph_to_graph_series(
#'     graph = graph_1) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_2) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_3)
#'
#' # Count the number of graphs
#' # in the graph series
#' series %>%
#'   count_graphs_in_graph_series()
#'
#' @export
count_graphs_in_graph_series <- function(graph_series) {

  if (inherits(graph_series, "dgr_graph_1D")) {
    if (is.null(graph_series$graphs)) {
      return(0L)
    }
    return(length(graph_series$graphs))
  }
}

#' Import a graph from various graph formats
#'
#' @description
#'
#' Import a variety of graphs from different graph formats and create a graph
#' object.
#'
#' @param graph_file A connection to a graph file. When provided as a path to a
#'   file, it will read the file from disk. Files starting with `http://`,
#'   `https://`, `ftp://`, or `ftps://` will be automatically downloaded.
#' @param file_type The type of file to be imported. Options are: `gml` (GML),
#'   `sif` (SIF), `edges` (a .edges file), and `mtx` (MatrixMarket format). If
#'   not supplied, the type of graph file will be inferred by its file
#'   extension.
#' @param edges_extra_attr_names For `edges` files, a vector of attribute names
#'   beyond the `from` and `to` data columns can be provided in the order they
#'   appear in the input data file.
#' @param edges_extra_attr_coltypes For `edges` files, this is a string of
#'   column types for any attribute columns provided for
#'   `edges_extra_attr_names`. This string representation is where each
#'   character represents each of the extra columns of data and the mappings
#'   are: `c` -> character, `i` -> integer, `n` -> number, `d` -> double, `l` ->
#'   logical, `D` -> date, `T` -> date time, `t` -> time, `?` -> guess, or
#'   `_/-`, which skips the column.
#' @inheritParams create_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' \dontrun{
#' # Import a GML graph file
#' gml_graph <-
#'   import_graph(
#'     system.file(
#'       "extdata/karate.gml",
#'       package = "DiagrammeR"))
#'
#' # Get a count of the graph's nodes
#' gml_graph |>
#'   count_nodes()
#'
#' # Get a count of the graph's edges
#' gml_graph |>
#'   count_edges()
#' }
#'
#' @export
import_graph <- function(
    graph_file,
    file_type = NULL,
    edges_extra_attr_names = NULL,
    edges_extra_attr_coltypes = NULL,
    graph_name = NULL,
    attr_theme = "default",
    write_backups = FALSE,
    display_msgs = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Stop function if `file_type` specified is not part
  # of the group that can be imported
  if (!is.null(file_type)) {
    if (!(tolower(file_type) %in%
          c("gml", "sif", "edges", "mtx"))) {

      cli::cli_abort(
        "The file type as specified cannot be imported.")
    }
  }

  # Stop function if file doesn't exist
  if (!grepl("(^http:|^https:|^ftp:|^ftp:)", graph_file)) {
    if (!file.exists(graph_file)) {

      cli::cli_abort(
        "The file as specified doesn't exist.")
    }
  }

  if (grepl("(^http:|^https:|^ftp:|^ftp:)", graph_file)) {

    dest_file <-
      unlist(strsplit(graph_file, "/"))[
        length(unlist(strsplit(graph_file, "/")))]

    # Download the file
    rlang::check_installed("curl", "to download a graph file.")
    curl::curl_download(graph_file, destfile = dest_file)

    # Extract the file and get the filename of the extracted file
    if (gsub(".*\\.([a-zA-Z]*?)", "\\1", graph_file) == "zip") {

      # Extract from the .zip archive
      utils::unzip(zipfile = dest_file)

      # Get the file name
      base_name_parts <- strsplit(dest_file, split = "\\.") |> unlist()
      base_name <- base_name_parts[[1]]

      graph_file <-
        list.files(pattern = paste0(base_name, ".*"))[
        !grepl(".*\\.zip$", list.files(pattern = paste0(base_name, ".*")))][1]
    }
  }

  # Obtain file extension if no value supplied
  # for `file_type`
  if (is.null(file_type)) {
    file_extension <- gsub(".*\\.([a-zA-Z]*?)", "\\1", graph_file)

    # Determine file type from file extension
    rlang::arg_match0(file_extension, c("gml", "sif", "edges", "mtx"))

    # one of gml, sif, edges, or mtx
    file_type <- file_extension
  }

  if (file_type == "edges") {

    # Read in the .edges document as a vector object
    edges_document <- readLines(graph_file, 10)

    # Determine which line the data fields begin
    first_line <- grep("^[^%].*", edges_document)[1]

    # Determine the number of lines to skip
    lines_to_skip <- first_line - 1L

    # Set default attribute names and column types
    # for an `edges` file
    attr_names <- c("from", "to")
    attr_coltypes <- c("i", "i")

    if (!is.null(edges_extra_attr_names)) {
      attr_names <- c(attr_names, edges_extra_attr_names)
    }

    if (!is.null(edges_extra_attr_coltypes)) {
      attr_coltypes <- paste(c(attr_coltypes, edges_extra_attr_coltypes), collapse = "")
    }

    # Create an edge data frame
    edges <-
      readr::read_delim(
        file = graph_file,
        delim = " ",
        skip = lines_to_skip,
        col_names = attr_names,
        col_types = attr_coltypes,
        progress = FALSE)

    n_rows <- nrow(edges)

    edges <-
      edges |>
      dplyr::mutate(
        id = seq_len(n_rows),
        rel = NA_character_
        ) |>
      dplyr::relocate("id", "from", "to", "rel") |>
      as.data.frame(stringsAsFactors = FALSE)

    # Create a node data frame
    # taking edges$from and edges$to
    edge_id_from <- as.integer(edges$from)
    edge_id_to <- as.integer(edges$to)

    nodes_raw <- unique(c(edge_id_from, edge_id_to))

    nodes <-
      data.frame(
        id = nodes_raw,
        type = NA_character_,
        label = as.character(nodes_raw),
        stringsAsFactors = FALSE
    )
    nodes <- dplyr::arrange(nodes, .data$id)

    # Create the graph
    the_graph <-
      create_graph(
        nodes_df = nodes,
        edges_df = edges,
        graph_name = graph_name,
        write_backups = write_backups,
        display_msgs = display_msgs
      )

    # Return the graph
    return(the_graph)
  }

  if (file_type == "mtx") {

    # Read in the .mtx document as a vector object
    mtx_document <- readLines(graph_file)

    # Determine which line the data fields begin
    first_line <- grep("^(\\w*) (\\w*)$", mtx_document)[1]

    # Create an edge data frame
    edges <-
      create_edge_df(
        from = sapply(
          strsplit(
            mtx_document[first_line:length(mtx_document)],
            " "), "[[", 1),
        to = sapply(
          strsplit(
            mtx_document[first_line:length(mtx_document)],
            " "), "[[", 2))

    # Create a node data frame
    nodes <-
      dplyr::tibble(
        id = as.integer(unique(
          unlist(
            strsplit(
              mtx_document[first_line:length(mtx_document)],
              " ")))),
        type = NA_character_,
        label = as.integer(unique(
          unlist(
            strsplit(
              mtx_document[first_line:length(mtx_document)],
              " "))))) |>
      as.data.frame(stringsAsFactors = FALSE)

    # Create the graph
    the_graph <-
      create_graph(
        nodes_df = nodes,
        edges_df = edges,
        graph_name = graph_name,
        write_backups = write_backups,
        display_msgs = display_msgs
      )

    # Return the graph
    return(the_graph)
  }

  if (file_type == "gml") {

    # Read in the .gml document as a vector object
    gml_document <-
      paste(readLines(graph_file), collapse = "")

    # Extract information on whether graph is directed
    graph_directed <-
      unlist(
        stringr::str_remove_all(
          stringr::str_extract_all(gml_document,
                          "directed [0-1]"),
          "directed "))

    # Extract all node definitions
    node_defs <-
      unlist(
        stringr::str_extract_all(gml_document,
                        "node[ ]*?\\[.*?\\]"))

    # Get all node ID values
    node_id <-
      as.integer(
        stringr::str_remove_all(
          stringr::str_extract_all(
            node_defs,
            "id [a-z0-9_]*"),
          "id "))

    # Get all node label values, if they exist
    if (any(stringr::str_detect(node_defs, "label"))) {
      node_label <-
        stringr::str_replace_all(
          stringr::str_replace_all(
            stringr::str_extract_all(
              node_defs,
              "label \\\".*?\\\""),
            "label \"", ""),
          "\"", "")
    }

    # Extract all edge definitions
    edge_defs <-
      unlist(stringr::str_extract_all(
        gml_document,
        "edge[ ]*?\\[.*?\\]"))

    edges_from <-
      as.integer(
        stringr::str_remove_all(
          stringr::str_extract_all(
            edge_defs,
            "source [a-z0-9_]*"),
          "source "))

    edges_to <-
      as.integer(
        stringr::str_remove_all(
          stringr::str_extract_all(
            edge_defs,
            "target [a-z0-9_]*"),
          "target "))


    if (any(stringr::str_detect(edge_defs, "label"))) {
      edge_label <-
        stringr::str_replace_all(
          stringr::str_replace_all(
            stringr::str_extract_all(
              edge_defs,
              "label \\\".*?\\\""),
            "label \"", ""),
          "\"", "")
    }

    if (any(stringr::str_detect(edge_defs, "value"))) {
      edge_value <-
        stringr::str_remove_all(
          stringr::str_extract_all(
            edge_defs,
            "value [a-z0-9\\.]*"),
          "value ")
    }

    # Create all nodes for graph
    all_nodes <-
      data.frame(
        id = node_id,
        type = NA_character_,
        label = NA_character_,
        stringsAsFactors = FALSE)

    if (exists("node_label")) {
      all_nodes$label <- node_label
    }

    # Create all edges for graph
    all_edges <-
      create_edge_df(
        from = edges_from,
        to = edges_to)

    if (exists("edge_label")) {
      all_edges$label <- edge_label
    }

    if (exists("edge_value")) {
      all_edges$data_value <- edge_value
    }

    # Create the graph
    the_graph <-
      create_graph(
        nodes_df = all_nodes,
        edges_df = all_edges,
        directed = graph_directed == "1",
        graph_name = graph_name,
        write_backups = write_backups,
        display_msgs = display_msgs
      )

    # Return the graph
    return(the_graph)
  }

  if (file_type == "sif") {

    # Read in the SIF document as a vector object
    sif_document <- readLines(graph_file)

    # Initialize the vector for a node data frame
    nodes <- vector(mode = "character")

    # Determine which nodes are present in the graph
    for (i in seq_along(sif_document)) {
      nodes <-
        c(nodes,
          ifelse(
            length(
              unlist(stringr::str_split(sif_document[i], "\t"))) == 1,
            unlist(stringr::str_split(sif_document[i], "\t"))[1],
            unlist(stringr::str_split(sif_document[i], "\t"))[-2]))
    }

    # Obtain a unique vector of nodes in the graph
    nodes <- unique(nodes)

    # Create a node data frame
    ndf <-
      create_node_df(
        n = length(nodes),
        label = nodes)

    # Determine which lines have single nodes
    if (!all(stringr::str_detect(sif_document, "\\t"))) {
      single_nodes <- which(!stringr::str_detect(sif_document, "\\t"))
    }

    # Initialize vectors for an edge data frame
    from <- to <- vector(mode = "integer")
    rel <- vector(mode = "character")

    # Obtain complete vectors for the edge data frame
    for (i in which(stringr::str_count(sif_document, "\\t") > 1)) {
      length_stmt <- length(stringr::str_split(sif_document[i], "\t")[[1]])
      from <- c(from, stringr::str_split(sif_document[i], "\t")[[1]][1])
      rel <- c(rel, stringr::str_split(sif_document[i], "\t")[[1]][2])
      to <- c(to, stringr::str_split(sif_document[i], "\t")[[1]][3:length_stmt])
    }

    # Create an edge data frame
    edf <-
      dplyr::tibble(
        from_label = from,
        to_label = to,
        rel = rel) |>
      dplyr::right_join(ndf, c("from_label" = "label")) |>
      dplyr::select(from = "id", "to_label", "rel") |>
      dplyr::right_join(ndf, c("to_label" = "label")) |>
      dplyr::select("from", to = "id", "rel") |>
      as.data.frame(stringsAsFactors = FALSE)

    # Create a graph object
    the_graph <-
      create_graph(
        nodes_df = ndf,
        edges_df = edf,
        graph_name = graph_name,
        write_backups = write_backups,
        display_msgs = display_msgs
      )

    # Return the graph
    return(the_graph)
  }
}

# Importing graphs from various formats

test_that("importing a GML graph applies edge labels", {

  # Create a temporary GML file with edge labels
  gml_text <- c(
    "graph",
    "[",
    "  directed 1",
    "  node",
    "  [",
    "    id 1",
    "    label \"A\"",
    "  ]",
    "  node",
    "  [",
    "    id 2",
    "    label \"B\"",
    "  ]",
    "  node",
    "  [",
    "    id 3",
    "    label \"C\"",
    "  ]",
    "  edge",
    "  [",
    "    source 1",
    "    target 2",
    "    label \"e1\"",
    "  ]",
    "  edge",
    "  [",
    "    source 2",
    "    target 3",
    "    label \"e2\"",
    "  ]",
    "]"
  )

  gml_file <- tempfile(fileext = ".gml")
  writeLines(gml_text, gml_file)
  on.exit(unlink(gml_file), add = TRUE)

  # Import the GML graph
  graph <- import_graph(gml_file, file_type = "gml")

  # Expect 3 nodes and 2 edges

  expect_equal(count_nodes(graph), 3)
  expect_equal(count_edges(graph), 2)

  # Expect node labels are applied
  expect_equal(graph$nodes_df$label, c("A", "B", "C"))

  # Expect edge labels are applied
  expect_equal(graph$edges_df$label, c("e1", "e2"))
})

test_that("importing a GML graph applies edge values", {

  gml_text <- c(
    "graph",
    "[",
    "  directed 0",
    "  node",
    "  [",
    "    id 1",
    "  ]",
    "  node",
    "  [",
    "    id 2",
    "  ]",
    "  edge",
    "  [",
    "    source 1",
    "    target 2",
    "    label \"x\"",
    "    value 3.5",
    "  ]",
    "]"
  )

  gml_file <- tempfile(fileext = ".gml")
  writeLines(gml_text, gml_file)
  on.exit(unlink(gml_file), add = TRUE)

  graph <- import_graph(gml_file, file_type = "gml")

  # Expect edge label and value are both applied
  expect_equal(graph$edges_df$label, "x")
  expect_equal(graph$edges_df$data_value, "3.5")

  # Expect the graph is undirected
  expect_false(graph$directed)
})

#' Select nodes in a graph
#'
#' @description
#'
#' Select nodes from a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#' @param conditions An option to use filtering conditions for the retrieval of
#'   nodes.
#' @param set_op The set operation to perform upon consecutive selections of
#'   graph nodes. This can either be as a `union` (the default), as an
#'   intersection of selections with `intersect`, or, as a `difference` on the
#'   previous selection, if it exists.
#' @param nodes An optional vector of node IDs for filtering the list of nodes
#'   present in the graph.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = c("a", "a", "z", "z"),
#'     label = TRUE,
#'     value = c(3.5, 2.6, 9.4, 2.7))
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = c("a", "z", "a"))
#'
#' # Create a graph with the ndf and edf
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Explicitly select nodes `1` and `3`
#' graph <-
#'   graph %>%
#'   select_nodes(nodes = c(1, 3))
#'
#' # Verify that the node selection has been made
#' # using the `get_selection()` function
#' graph %>% get_selection()
#'
#' # Select nodes based on the node `type`
#' # being `z`
#' graph <-
#'   graph %>%
#'   clear_selection() %>%
#'   select_nodes(
#'     conditions = type == "z")
#'
#' # Verify that an node selection has been made, and
#' # recall that the `3` and `4` nodes are of the
#' # `z` type
#' graph %>% get_selection()
#'
#' # Select edges based on the node value attribute
#' # being greater than 3.0 (first clearing the current
#' # selection of nodes)
#' graph <-
#'   graph %>%
#'   clear_selection() %>%
#'   select_nodes(
#'     conditions = value > 3.0)
#'
#' # Verify that the correct node selection has been
#' # made; in this case, nodes `1` and `3` have values
#' # for `value` greater than 3.0
#' graph %>% get_selection()
#'
#' @export
select_nodes <- function(
    graph,
    conditions = NULL,
    set_op = "union",
    nodes = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Stop function if all `nodes` refer to node ID
  # values that are not in the graph
  # If there is one node in graph and one out of bound, no error.
  if (!is.null(nodes) && !any(nodes %in% graph$nodes_df$id)) {

    cli::cli_abort(c(
      "`nodes` must correspond to values in the graph.",
      i = "`Graph values IDs include {unique(graph$nodes_df$id)}, not {nodes}."))
  }

  # Extract the graph's internal ndf
  nodes_df <- graph$nodes_df

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # If conditions are provided then
  # pass in those conditions and filter the
  # data frame of `nodes_df`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    nodes_df <- dplyr::filter(.data = nodes_df, {{ conditions }})
  }

  # Get the nodes as a vector
  nodes_selected <-
    nodes_df %>%
    dplyr::pull("id")

  # If a `nodes` vector provided, get the intersection
  # of that vector with the filtered node IDs
  if (!is.null(nodes)) {
    nodes_selected <- intersect(nodes, nodes_selected)
  }

  # Obtain vector with node ID selection of nodes
  # already present
  nodes_prev_selection <- graph$node_selection$node

  # Incorporate the selected nodes into the
  # graph's selection
  if (set_op == "union") {
    nodes_combined <- union(nodes_prev_selection, nodes_selected)
  } else if (set_op == "intersect") {
    nodes_combined <- intersect(nodes_prev_selection, nodes_selected)
  } else if (set_op == "difference") {
    nodes_combined <- base::setdiff(nodes_prev_selection, nodes_selected)
  }

  # Add the node ID values to the active selection
  # of nodes in `graph$node_selection`
  graph$node_selection <-
    replace_graph_node_selection(
      graph = graph,
      replacement = nodes_combined)

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Obtain the output graph's node and edge
  # selection properties
  n_e_select_properties_out <-
    node_edge_selection_properties(graph = graph)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # Emit a message about the modification of a selection
  # if that option is set
  if (!is.null(graph$graph_info$display_msgs) &&
      graph$graph_info$display_msgs) {

    # Construct message body
    if (!n_e_select_properties_in[["node_selection_available"]] &&
        !n_e_select_properties_in[["edge_selection_available"]]) {

      msg_body <-
        glue::glue(
          "created a new selection of \\
         {n_e_select_properties_out[['selection_count_str']]}")

    } else if (n_e_select_properties_in[["node_selection_available"]] ||
               n_e_select_properties_in[["edge_selection_available"]]) {

      if (n_e_select_properties_in[["node_selection_available"]]) {
        msg_body <-
          glue::glue(
            "modified an existing selection of \\
           {n_e_select_properties_in[['selection_count_str']]}:
           * {n_e_select_properties_out[['selection_count_str']]} \\
           are now in the active selection
           * used the `{set_op}` set operation")
      }

      if (n_e_select_properties_in[["edge_selection_available"]]) {
        msg_body <-
          glue::glue(
            "created a new selection of \\
           {n_e_select_properties_out[['selection_count_str']]}:
           * this replaces \\
           {n_e_select_properties_in[['selection_count_str']]} \\
           in the prior selection")
      }
    }

    # Issue a message to the user
    emit_message(
      fcn_name = fcn_name,
      message_body = msg_body)
  }

  graph
}

#' Add a node to an existing graph object
#'
#' @description
#'
#' With a graph object of class `dgr_graph`, add a new node to the graph. One
#' can optionally provide node attributes for the created node. There is also
#' the option to create edges to and from existing nodes in the graph. Because
#' new edges can also be created through this function, there is the possibility
#' to set edge attributes for any new graph edges.
#'
#' @inheritParams node_edge_aes_data
#' @inheritParams render_graph
#' @param type An optional character object that acts as a group identifier for
#'   the node to be added.
#' @param label An optional character object that describes the node.
#' @param from An optional vector containing node IDs from which edges will be
#'   directed to the new node.
#' @param to An optional vector containing node IDs to which edges will be
#'   directed from the new node.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph and add 2 nodes by using
#' # the `add_node()` function twice
#' graph <-
#'   create_graph() |>
#'   add_node() |>
#'   add_node()
#'
#' # Get a count of all nodes
#' # in the graph
#' graph |> count_nodes()
#'
#' # The nodes added were given
#' # ID values `1` and `2`; obtain
#' # the graph's node IDs
#' graph |> get_node_ids()
#'
#' # Add a node with a `type`
#' # value defined
#' graph <-
#'   graph |>
#'   add_node(type = "person")
#'
#' # View the graph's internal
#' # node data frame (ndf)
#' graph |> get_node_df()
#'
#' @family node creation and removal
#'
#' @export
add_node <- function(
    graph,
    type = NULL,
    label = NULL,
    from = NULL,
    to = NULL,
    node_aes = NULL,
    edge_aes = NULL,
    node_data = NULL,
    edge_data = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the node ID for the node to be added
  node <- graph$last_node + 1L

  # Get the number of edges in the graph
  edges_in_graph <- nrow(graph$edges_df)

  # Get the number of edges ever created for
  # this graph
  edges_created <- graph$last_edge

  # If `type` or `label` is NULL, convert to NA
  type <- type %||% NA_character_
  label <- label %||% NA_character_

  # Collect node aesthetic attributes
  if (!is.null(node_aes)) {

    node_aes_tbl <- dplyr::as_tibble(node_aes)

    if (nrow(node_aes_tbl) == 1) {

      node_aes$index__ <- 1

      node_aes_tbl <-
        dplyr::as_tibble(node_aes) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_aes_tbl)) {
      node_aes_tbl$id <- NULL
    }
  }

  # Collect node data attributes
  if (!is.null(node_data)) {

    node_data_tbl <- dplyr::as_tibble(node_data)

    if (nrow(node_data_tbl) == 1) {

      node_data$index__ <- 1

      node_data_tbl <-
        dplyr::as_tibble(node_data) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_data_tbl)) {
      node_data_tbl$id <- NULL
    }
  }

  # Modify graph if neither `to` nor `from`
  # values provided
  if (is.null(from) && is.null(to)) {

    new_node <-
      create_node_df(
        n = 1,
        label = as.character(label),
        type = as.character(type))

    new_node[1, 1] <- node

    # Add node aesthetics if available
    if (exists("node_aes_tbl")) {

      new_node <-
        new_node |>
        dplyr::bind_cols(node_aes_tbl)
    }

    # Add node data if available
    if (exists("node_data_tbl")) {

      new_node <-
        new_node |>
        dplyr::bind_cols(node_data_tbl)
    }

    combined_nodes <-
      combine_ndfs(graph$nodes_df, new_node)

    edges_in_graph_2 <- nrow(graph$edges_df)

    graph$nodes_df <- combined_nodes
    graph$last_node <- graph$last_node + 1

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph$graph_log <-
      add_action_to_log(
        graph_log = graph$graph_log,
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df),
        d_n = 1)

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }

    return(graph)
  }

  # Modify graph if only `from` values provided
  if (!is.null(from) && is.null(to)) {

    from_nodes_available <-
      all(from %in% get_node_ids(graph))

    if (!from_nodes_available) {

      cli::cli_abort(c(
        "The nodes from which edges should be applied to the new node are not available."
      ))
    }

    # from_nodes_available

    new_node <-
      create_node_df(
        n = 1,
        label = as.character(label),
        type = as.character(type)
      )

    new_node[1, 1] <- node

    # Add node aesthetics if available
    if (exists("node_aes_tbl")) {
      new_node <-
        new_node |>
        dplyr::bind_cols(node_aes_tbl)
    }

    # Add node data if available
    if (exists("node_data_tbl")) {
      new_node <-
        new_node |>
        dplyr::bind_cols(node_data_tbl)
    }

    # Combine the new nodes with those in the graph
    combined_nodes <-
      combine_ndfs(graph$nodes_df, new_node)

    # Collect edge aesthetic attributes
    if (!is.null(edge_aes)) {

      edge_aes_tbl <- dplyr::as_tibble(edge_aes)

      if (nrow(edge_aes_tbl) < length(from)) {
        edge_aes$index__ <- seq_along(from)

        edge_aes_tbl <-
          dplyr::as_tibble(edge_aes) |>
          dplyr::select(-"index__")
      }

      if ("id" %in% colnames(edge_aes_tbl)) {
        edge_aes_tbl$id <- NULL
      }
    }

    # Collect edge data attributes
    if (!is.null(edge_data)) {
      edge_data_tbl <- dplyr::as_tibble(edge_data)

      if (nrow(edge_data_tbl) < length(from)) {
        edge_data$index__ <- seq_along(from)

        edge_data_tbl <-
          dplyr::as_tibble(edge_data) |>
          dplyr::select(-"index__")
      }

      if ("id" %in% colnames(edge_data_tbl)) {
        edge_data_tbl$id <- NULL
      }
    }

    new_edges <-
      create_edge_df(
        from = from,
        to = rep(node, length(from))
      )

    new_edges$id <-
      new_edges$id + edges_created

    # Add edge aesthetics if available
    if (exists("edge_aes_tbl")) {
      new_edges <-
        new_edges |>
        dplyr::bind_cols(edge_aes_tbl)
    }

    # Add edge data if available
    if (exists("edge_data_tbl")) {
      new_edges <-
        new_edges |>
        dplyr::bind_cols(edge_data_tbl)
    }

    # Use `bind_rows()` to add the
    # new edges
    combined_edges <-
      dplyr::bind_rows(
        graph$edges_df,
        new_edges
      )

    edges_in_graph_2 <- nrow(combined_edges)
    edges_added <- edges_in_graph_2 - edges_in_graph

    # Create a revised graph
    graph$nodes_df <- combined_nodes
    graph$edges_df <- combined_edges
    graph$last_node <- graph$last_node + 1
    graph$last_edge <- graph$last_edge + edges_added

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph$graph_log <-
      add_action_to_log(
        graph_log = graph$graph_log,
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df),
        d_n = 1,
        d_e = length(from)
      )

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }

    return(graph)
  }

  # Modify graph if only `to` values provided
  if (is.null(from) && !is.null(to)) {
    to_nodes_available <-
      all(to %in% get_node_ids(graph))

    if (!to_nodes_available) {
      cli::cli_abort(
        "The nodes to which edges should be applied from the new node are not available.")
    }

    # Assume to_nodes is available.
    new_node <-
      create_node_df(
        n = 1,
        label = as.character(label),
        type = as.character(type))

    new_node[1, 1] <- node

    # Add node aesthetics if available
    if (exists("node_aes_tbl")) {

      new_node <-
        new_node |>
        dplyr::bind_cols(node_aes_tbl)
    }

    # Add node data if available
    if (exists("node_data_tbl")) {

      new_node <-
        new_node |>
        dplyr::bind_cols(node_data_tbl)
    }

    combined_nodes <-
      combine_ndfs(graph$nodes_df, new_node)

    # Collect edge aesthetic attributes
    if (!is.null(edge_aes)) {

      edge_aes_tbl <- dplyr::as_tibble(edge_aes)

      if (nrow(edge_aes_tbl) < length(from)) {
        edge_aes$index__ <- seq_along(from)

        edge_aes_tbl <-
          dplyr::as_tibble(edge_aes) |>
          dplyr::select(-"index__")
      }

      if ("id" %in% colnames(edge_aes_tbl)) {
        edge_aes_tbl$id <- NULL
      }
    }

    # Collect edge data attributes
    if (!is.null(edge_data)) {

      edge_data_tbl <- dplyr::as_tibble(edge_data)

      if (nrow(edge_data_tbl) < length(from)) {
        edge_data$index__ <- seq_along(from)

        edge_data_tbl <-
          dplyr::as_tibble(edge_data) |>
          dplyr::select(-"index__")
      }

      if ("id" %in% colnames(edge_data_tbl)) {
        edge_data_tbl$id <- NULL
      }
    }

    new_edges <-
      create_edge_df(
        from = rep(node, length(to)),
        to = to)

    new_edges$id <-
      new_edges$id + edges_created

    # Add edge aesthetics if available
    if (exists("edge_aes_tbl")) {

      new_edges <-
        new_edges |>
        dplyr::bind_cols(edge_aes_tbl)
    }

    # Add edge data if available
    if (exists("edge_data_tbl")) {
      new_edges <-
        new_edges |>
        dplyr::bind_cols(edge_data_tbl)
    }

    # Use `bind_rows()` to add the
    # new edges
    combined_edges <-
      dplyr::bind_rows(
        graph$edges_df,
        new_edges)

    edges_in_graph_2 <- nrow(combined_edges)
    edges_added <- edges_in_graph_2 - edges_in_graph

    # Create a revised graph
    graph$nodes_df <- combined_nodes
    graph$edges_df <- combined_edges
    graph$last_node <- graph$last_node + 1
    graph$last_edge <- graph$last_edge + edges_added

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph$graph_log <-
      add_action_to_log(
        graph_log = graph$graph_log,
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df),
        d_n = 1,
        d_e = length(to))

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }

    return(graph)
  }

  # Modify graph if both `to` and `from` values provided
  if (!is.null(from) && !is.null(to)) {

    from_nodes_available <-
      all(from %in% get_node_ids(graph))

    to_nodes_available <-
      all(to %in% get_node_ids(graph))

    if (!from_nodes_available) {

      cli::cli_abort(
        "The nodes from which edges should be applied to the new node are not available.")
    }

    if (!to_nodes_available) {

      cli::cli_abort(
        "The nodes to which edges should be applied from the new node are not available."
      )
    }

    # If both to_nodes and from_nodes are available, continue.
      new_node <-
        create_node_df(
          n = 1,
          label = as.character(label),
          type = as.character(type))

      new_node[1, 1] <- node

      # Add node aesthetics if available
      if (exists("node_aes_tbl")) {
        new_node <-
          new_node |>
          dplyr::bind_cols(node_aes_tbl)
      }

      # Add node data if available
      if (exists("node_data_tbl")) {

        new_node <-
          new_node |>
          dplyr::bind_cols(node_data_tbl)
      }

      # Combine the new nodes with those in the graph
      combined_nodes <-
        combine_ndfs(graph$nodes_df, new_node)

      new_edges <-
        combine_edfs(
          create_edge_df(
            from = rep(node, length(to)),
            to = to),
          create_edge_df(
            from = from,
            to = rep(node, length(from))))

      new_edges$id <-
        new_edges$id + edges_created

      # Collect edge aesthetic attributes
      if (!is.null(edge_aes)) {

        edge_aes_tbl <- dplyr::as_tibble(edge_aes)

        if (nrow(edge_aes_tbl) < (length(from) + length(to))) {
          edge_aes$index__ <- seq_len(length(from) + length(to))

          edge_aes_tbl <-
            dplyr::as_tibble(edge_aes) |>
            dplyr::select(-"index__")
        }

        if ("id" %in% colnames(edge_aes_tbl)) {
          edge_aes_tbl$id <- NULL
        }
      }

      # Collect edge data attributes
      if (!is.null(edge_data)) {

        edge_data_tbl <- dplyr::as_tibble(edge_data)

        if (nrow(edge_data_tbl) < (length(from) + length(to))) {

          edge_data$index__ <- seq_len(length(from) + length(to))

          edge_data_tbl <-
            dplyr::as_tibble(edge_data) |>
            dplyr::select(-"index__")
        }

        if ("id" %in% colnames(edge_data_tbl)) {
          edge_data_tbl$id <- NULL
        }
      }

      # Add edge aesthetics if available
      if (exists("edge_aes_tbl")) {

        new_edges <-
          new_edges |>
          dplyr::bind_cols(edge_aes_tbl)
      }

      # Add edge data if available
      if (exists("edge_data_tbl")) {

        new_edges <-
          new_edges |>
          dplyr::bind_cols(edge_data_tbl)
      }

      # Use `bind_rows()` to add the
      # new edges
      combined_edges <-
        dplyr::bind_rows(
          graph$edges_df,
          new_edges)

      edges_in_graph_2 <- nrow(combined_edges)
      edges_added <- edges_in_graph_2 - edges_in_graph

      # Create a revised graph and return that graph
      graph$nodes_df <- combined_nodes
      graph$edges_df <- combined_edges
      graph$last_node <- graph$last_node + 1
      graph$last_edge <- graph$last_edge + edges_added

      # Get the name of the function
      fcn_name <- get_calling_fcn()

      # Update the `graph_log` df with an action
      graph$graph_log <-
        add_action_to_log(
          graph_log = graph$graph_log,
          version_id = nrow(graph$graph_log) + 1L,
          function_used = fcn_name,
          time_modified = time_function_start,
          duration = graph_function_duration(time_function_start),
          nodes = nrow(graph$nodes_df),
          edges = nrow(graph$edges_df),
          d_n = 1,
          d_e = length(from) + length(to))

      # Perform graph actions, if any are available
      if (nrow(graph$graph_actions) > 0) {
        graph <-
          trigger_graph_actions(graph)
      }

      # Write graph backup if the option is set
      if (graph$graph_info$write_backups) {
        save_graph_as_rds(graph = graph)
      }

      return(graph)
  }
}

#' Get edge attribute values
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get edge attribute values for one
#' or more edges.
#'
#' @inheritParams render_graph
#' @param edge_attr the name of the attribute for which to get values.
#' @param from an optional vector of node IDs from which the edge is outgoing
#'   for filtering the list of edges.
#' @param to an optional vector of node IDs from which the edge is incoming for
#'   filtering the list of edges.
#'
#' @return A named vector of edge attribute values for the attribute given by
#'   `edge_attr` by edge.
#'
#' @examples
#' # Create a simple graph where
#' # edges have an edge attribute
#' # named `value`
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(n = 4) %>%
#'   {
#'     edges <-
#'       create_edge_df(
#'         from = c(1, 2, 1, 4),
#'           to = c(2, 3, 4, 3),
#'          rel = "rel")
#'     add_edge_df(
#'       graph = .,
#'       edge_df = edges)
#'   } %>%
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 1.6,
#'     from = 1,
#'       to = 2) %>%
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 4.3,
#'     from = 1,
#'       to = 4) %>%
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 2.9,
#'     from = 2,
#'       to = 3) %>%
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 8.4,
#'     from = 4,
#'       to = 3)
#'
#' # Get the values for the
#' # `value` edge attribute
#' graph %>%
#'   get_edge_attrs(
#'     edge_attr = value)
#'
#' # To only return edge attribute
#' # values for specified edges, use
#' # the `from` and `to` arguments
#' graph %>%
#'   get_edge_attrs(
#'     edge_attr = value,
#'     from = c(1, 2),
#'       to = c(2, 3))
#'
#' @export
get_edge_attrs <- function(
    graph,
    edge_attr,
    from = NULL,
    to = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  edge_attr <- rlang::enquo(edge_attr)

  if (rlang::get_expr(edge_attr) %>%
      as.character() %in% c("id", "from", "to")) {

    cli::cli_abort(
      "This is not an edge attribute.")
  }

  if (length(from) != length(to)) {
      cli::cli_abort(
        "The number of nodes in `from` and `to` must be the same.")
  }

  # Extract the edge data frame (ndf)
  # from the graph
  edf <- graph$edges_df

  if (is.null(from) || is.null(to)) {

    # Extract the edge attribute values
    edge_attr_vals <-
      edf %>%
      dplyr::pull(!!edge_attr)

    # Extract the edge names
    edge_names <-
      paste(edf$from, edf$to, sep = "->")

    # Assign edge names
    names(edge_attr_vals) <- edge_names
  }

  if (!is.null(from) && !is.null(to)) {

    # Get edges as strings for filtering
    # the `edf` object
    edges <- paste(from, to, sep = "->")

    # Filter the edf by the supplied
    # edge definitions
    edf <-
      edf %>%
      dplyr::mutate(from_to = paste(from, to, sep = "->")) %>%
      dplyr::filter(from_to %in% edges)

    # Extract the edge attribute values
    edge_attr_vals <-
      edf %>%
      dplyr::pull(!!edge_attr)

    # Extract the edge names
    edge_names <-
      paste(edf$from, edf$to, sep = "->")

    # Assign edge names
    names(edge_attr_vals) <- edge_names
  }

  edge_attr_vals
}

#' Save a graph or graph series to disk
#'
#' @description
#'
#' Save a graph or a graph series object to disk.
#'
#' @param x A graph object of class `dgr_graph` or a graph series object of type
#'   `dgr_graph_1D`.
#' @param file A file name for the graph or graph series. Provide a character
#'   string and the `.dgr` extension will be applied to it.
#'
#' @examples
#' # Create an undirected GNP
#' # graph with 100 nodes using
#' # a probability value of 0.05
#' gnp_graph <-
#'   create_graph(
#'     directed = FALSE) %>%
#'   add_gnp_graph(
#'     n = 100,
#'     p = 0.05)
#'
#' # Save the graph to disk; use
#' # the file name `gnp_graph.dgr`
#' # save_graph(
#' #   x = gnp_graph,
#' #   file = "gnp_graph"
#' # )
#'
#' # To read the graph file from
#' # disk, use `open_graph()`
#' # gnp_graph_2 <-
#' #   open_graph(
#' #     file = "gnp_graph.dgr"
#' # )
#'
#' @family Display and Save
#' @export
save_graph <- function(
    x,
    file
) {

  check_string(file)

  # Check if the object is a graph or graph series.
  if (!rlang::inherits_any(x, c("dgr_graph", "dgr_graph_1D"))) {
    cli::cli_abort(
      "The object provided is not a graph or graph series."
    )
  }

  # Append the file extension to the file path
  file_name <- file %>% paste0(".dgr")

  # Save the graph or graph series
  saveRDS(x, file = file_name, compress = "xz")
}

#' Get the current selection available in a graph object
#'
#' @description
#'
#' Get the current selection of node IDs or edge IDs from a graph object of
#' class `dgr_graph`.
#'
#' @inheritParams render_graph
#'
#' @return A vector with the current selection of node or edge ID values.
#'
#' @examples
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_path(n = 6)
#'
#' # Select node `4`, then select
#' # all nodes a distance of 1 away
#' # from node `4`, and finally
#' # return the selection of nodes as
#' # a vector object
#' graph |>
#'   select_nodes(nodes = 4) |>
#'   select_nodes_in_neighborhood(
#'     node = 4,
#'     distance = 1) |>
#'   get_selection()
#'
#' # Select edges associated with
#' # node `4` and return the
#' # selection of edges
#' graph |>
#'   select_edges_by_node_id(
#'     nodes = 4) |>
#'   get_selection()
#'
#' @export
get_selection <- function(graph) {

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # If there is no selection available, return NA
  if (is.na(n_e_select_properties_in[["selection_type"]])) {

    # Emit a message about the modification of a selection
    # if that option is set
    if (isTRUE(graph$graph_info$display_msgs)) {

      # Issue a message to the user
      emit_message(
        fcn_name = fcn_name,
        message_body = "There is no active selection of nodes or edges.")
    }

    return(NA)
  }

  # For a selection of nodes, return a vector of node
  # ID values
  if (n_e_select_properties_in[["selection_type"]] == "node") {

    # Emit a message about the modification of a selection
    # if that option is set
    if (isTRUE(graph$graph_info$display_msgs)) {

    # Issue a message to the user
    emit_message(
      fcn_name = fcn_name,
      message_body = glue::glue(
        "there is an active selection of \\
        {n_e_select_properties_in[['selection_count_str']]}"))
    }

    return(as.integer(sort(graph$node_selection$node)))
  }

  # For a selection of edges, return a vector of edge
  # ID values
  if (n_e_select_properties_in[["selection_type"]] == "edge") {

    # Emit a message about the modification of a selection
    # if that option is set
    if (isTRUE(graph$graph_info$display_msgs)) {

      # Issue a message to the user
      emit_message(
        fcn_name = fcn_name,
        message_body = glue::glue(
          "there is an active selection of\\
        {n_e_select_properties_in[['selection_count_str']]}"))
    }

    return(as.integer(sort(graph$edge_selection$edge)))
  }
}

#' Get global graph attributes
#'
#' @description
#'
#' Get the available global attributes for a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#'
#' @return A data frame containing global attributes for the graph.
#'
#' @examples
#' # Create a new, empty graph
#' graph <- create_graph()
#'
#' # View the graph's set of
#' # global attributes
#' graph |>
#'   get_global_graph_attr_info()
#'
#' @export
get_global_graph_attr_info <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  if (nrow(graph$global_attrs) == 0) {

    global_graph_attrs_tbl <-
      dplyr::tibble(
        attr = character(),
        value = character(),
        attr_type = character()
      )

  } else if (nrow(graph$global_attrs) > 0) {
    global_graph_attrs_tbl <- dplyr::as_tibble(graph$global_attrs)
  }

  global_graph_attrs_tbl
}

#' Get edge attribute values from a selection of edges
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get edge attribute values for one
#' or more edges.
#'
#' This function makes use of an active selection of edges (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#' @param edge_attr the name of the attribute for which to get values.
#'
#' @return A named vector of edge attribute values for the attribute given by
#'   `edge_attr` by edge.
#'
#' @examples
#' # Create a simple graph where
#' # edges have an edge attribute
#' # named `value`
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(n = 4)
#'
#' edges <-
#'   create_edge_df(
#'     from = c(1, 2, 1, 4),
#'       to = c(2, 3, 4, 3),
#'      rel = "rel")
#'
#' graph <-
#'   add_edge_df(
#'     graph = graph,
#'     edge_df = edges) |>
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 1.6,
#'     from = 1,
#'       to = 2) |>
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 4.3,
#'     from = 1,
#'       to = 4) |>
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 2.9,
#'     from = 2,
#'       to = 3) |>
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 8.4,
#'     from = 4,
#'       to = 3)
#'
#' # Select the edges defined as
#' # `1`->`3` and `2`->`3`
#' graph <-
#'   graph |>
#'   select_edges(
#'     from = c(1, 2),
#'     to = c(2, 3))
#'
#' # Get the edge attribute values
#' # for the `value` attribute, limited
#' # to the current edge selection
#' graph |>
#'   get_edge_attrs_ws(
#'     edge_attr = value)
#'
#' @export
get_edge_attrs_ws <- function(
    graph,
    edge_attr
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph object has a valid edge selection
  check_graph_contains_edge_selection(graph)

  edge_attr <- rlang::enquo(edge_attr)

  if (rlang::enquo(edge_attr) |>
      rlang::get_expr() |>
      as.character() %in% c("id", "from", "to")) {

    cli::cli_abort(
      "This is not an edge attribute.")
  }

  # Extract the edge data frame (edf)
  # from the graph
  edf <- graph$edges_df

  # Get the edge IDs from the edge selection
  edges <- sort(graph$edge_selection$edge)

  # Filter the edf by the supplied
  # edge ID values
  edf <-
    edf |>
    dplyr::filter(id %in% edges)

  # Extract the edge attribute values
  edge_attr_vals <- edf |> dplyr::pull(!!edge_attr)

  # Extract the edge names
  edge_names <- paste(edf$from, edf$to, sep = "->")

  # Assign edge names
  names(edge_attr_vals) <- edge_names

  edge_attr_vals
}

#' Select edges in a graph using node ID values
#'
#' @description
#'
#' Select edges in a graph object of class `dgr_graph` using node ID values. All
#' edges associated with the provided nodes will be included in the selection.
#'
#' @inheritParams render_graph
#' @param nodes A vector of node IDs for the selection of edges present in the
#'   graph.
#' @param set_op The set operation to perform upon consecutive selections of
#'   graph edges This can either be as a `union` (the default), as an
#'   intersection of selections with `intersect`, or, as a `difference` on the
#'   previous selection, if it exists.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 5 nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 5)
#'
#' # Create a graph selection by selecting edges
#' # associated with nodes `1` and `2`
#' graph <-
#'   graph %>%
#'   select_edges_by_node_id(
#'     nodes = 1:2)
#'
#' # Get the selection of edges
#' graph %>% get_selection()
#'
#' # Perform another selection of edges, with nodes
#' # `1`, `2`, and `4`
#' graph <-
#'   graph %>%
#'   clear_selection() %>%
#'   select_edges_by_node_id(
#'     nodes = c(1, 2, 4))
#'
#' # Get the selection of edges
#' graph %>% get_selection()
#'
#' # Get a fraction of the edges selected over all
#' # the edges in the graph
#' graph %>%
#'   {
#'     l <- get_selection(.) %>%
#'       length(.)
#'     e <- count_edges(.)
#'     l/e
#'   }
#'
#' @export
select_edges_by_node_id <- function(
    graph,
    nodes,
    set_op = "union"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Extract the graph's internal edf
  edges_df <- graph$edges_df

  # Filter the edf by the requested node IDs
  edges_selected <-
    edges_df %>%
    dplyr::filter(from %in% nodes | to %in% nodes)

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # Create an integer vector representing edges
  edges_selected <- edges_selected$id

  # Obtain vector with node ID selection of edges
  # already present
  edges_prev_selection <- graph$edge_selection$edge

  # Incorporate the selected edges into the
  # graph's selection
  if (set_op == "union") {
    edges_combined <-
      union(edges_prev_selection, edges_selected)
  } else if (set_op == "intersect") {
    edges_combined <-
      intersect(edges_prev_selection, edges_selected)
  } else if (set_op == "difference") {
    edges_combined <-
      base::setdiff(edges_prev_selection, edges_selected)
  }

  # Filter `edges_df` to provide the correct esdf
  edges_combined <-
    graph$edges_df %>%
    dplyr::filter(id %in% edges_combined) %>%
    dplyr::select(edge = "id", "from", "to")

  # Add the edge ID values to the active selection
  # of edges in `graph$edge_selection`
  graph$edge_selection <- edges_combined

  # Replace `graph$node_selection` with an empty df
  graph$node_selection <- create_empty_nsdf()

  # Obtain the output graph's node and edge
  # selection properties
  n_e_select_properties_out <-
    node_edge_selection_properties(graph = graph)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # Emit a message about the modification of a selection
  # if that option is set
  if (!is.null(graph$graph_info$display_msgs) &&
      graph$graph_info$display_msgs) {

    # Construct message body
    if (!n_e_select_properties_in[["node_selection_available"]] &&
        !n_e_select_properties_in[["edge_selection_available"]]) {

      msg_body <-
        glue::glue(
          "created a new selection of \\
        {n_e_select_properties_out[['selection_count_str']]}")

    } else if (n_e_select_properties_in[["node_selection_available"]] ||
               n_e_select_properties_in[["edge_selection_available"]]) {

      if (n_e_select_properties_in[["edge_selection_available"]]) {
        msg_body <-
          glue::glue(
            "modified an existing selection of\\
           {n_e_select_properties_in[['selection_count_str']]}:
           * {n_e_select_properties_out[['selection_count_str']]}\\
           are now in the active selection
           * used the `{set_op}` set operation")
      }

      if (n_e_select_properties_in[["node_selection_available"]]) {
        msg_body <-
          glue::glue(
            "created a new selection of\\
           {n_e_select_properties_out[['selection_count_str']]}:
           * this replaces\\
           {n_e_select_properties_in[['selection_count_str']]}\\
           in the prior selection")
      }
    }

    # Issue a message to the user
    emit_message(
      fcn_name = fcn_name,
      message_body = msg_body)
  }

  graph
}

#' Create a graph using an adjacency matrix
#'
#' @description
#'
#' Using an adjacency matrix object, generate a graph of class `dgr_graph`.
#'
#' @param x A square `matrix` object serving as the adjacency matrix.
#' @param mode The method in which to interpret the input adjacency matrix.
#'   Options include: `undirected`, `directed`, `upper`, `lower`, `max`, `min`,
#'   and `plus`.
#' @param weighted Whether to create a weighted graph from the adjacency matrix.
#' @param use_diag Whether to use the diagonal of the adjacency matrix in
#'   calculations. If `TRUE` then the diagonal values will be included as is. If
#'   `FALSE` then the diagonal values will be replaced with zero values before
#'   inclusion in any calculations.
#' @inheritParams create_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an adjacency matrix
#' adj_matrix <-
#'   sample(
#'     0:1, 100,
#'     replace = TRUE,
#'     prob = c(0.9,0.1)
#'   ) %>%
#'   matrix(ncol = 10)
#'
#' # Create a graph from the adjacency matrix
#' graph <- from_adj_matrix(adj_matrix)
#'
#' @export
from_adj_matrix <- function(
    x,
    mode = "undirected",
    weighted = FALSE,
    use_diag = TRUE,
    graph_name = NULL,
    write_backups = FALSE,
    display_msgs = FALSE
) {

  # Stop function if x is not a matrix object
  if (!inherits(x, "matrix")) {

    cli::cli_abort(
      "The input for this function must be a square matrix object.")
  }

  # Stop function if the matrix is not a square matrix
  if (ncol(x) != nrow(x)) {

    cli::cli_abort(
      "The input matrix must be a square matrix.")
  }

  # If FALSE provided for `weighted`, change value to
  # NULL for `graph_from_adjacency_matrix()`
  if (!weighted) {
    weighted <- NULL
  }

  # Generate an igraph graph from the adjacency matrix
  if (is.null(colnames(x))) {

    igraph <-
      igraph::graph_from_adjacency_matrix(
        adjmatrix = x,
        mode = mode,
        weighted = weighted,
        diag = use_diag)

  } else if (!is.null(colnames(x))) {

    igraph <-
      igraph::graph_from_adjacency_matrix(
        adjmatrix = x,
        mode = mode,
        weighted = weighted,
        diag = use_diag,
        add.colnames = "label")
  }

  # Generate the graph object from an igraph graph
  graph <-
    from_igraph(
      igraph,
      graph_name = graph_name,
      write_backups = write_backups,
      display_msgs = display_msgs
    )

  # Add edge ID values to `graph`
  if (nrow(graph$edges_df) > 0) {
    graph$edges_df$id <- seq_len(nrow(graph$edges_df))
  }

  graph
}

#' Get node attribute values from a selection of nodes
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get node attribute values from
#' nodes currently active as a selection.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param node_attr The name of the attribute for which to get values.
#'
#' @return A named vector of node attribute values for the attribute given by
#'   `node_attr` by node ID.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 4,
#'     m = 4,
#'     set_seed = 23) |>
#'   set_node_attrs(
#'     node_attr = value,
#'     values = c(2.5, 8.2, 4.2, 2.4))
#'
#' # Select nodes with ID values
#' # `1` and `3`
#' graph <-
#'   graph |>
#'   select_nodes_by_id(
#'     nodes = c(1, 3))
#'
#' # Get the node attribute values
#' # for the `value` attribute, limited
#' # to the current node selection
#' graph |>
#'   get_node_attrs_ws(
#'     node_attr = value)
#'
#' @export
get_node_attrs_ws <- function(
    graph,
    node_attr
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph object has a valid node selection
  check_graph_contains_node_selection(graph)

  node_attr <- rlang::enquo(node_attr)

  if (rlang::enquo(node_attr) |>
      rlang::get_expr() |>
      as.character() %in% c("id", "nodes")) {

    cli::cli_abort(
      "This is not a node attribute.")
  }

  # Extract the node data frame (ndf)
  # from the graph
  ndf <- graph$nodes_df

  # Get the node IDs from the node selection
  nodes <- sort(graph$node_selection$node)

  # Filter the ndf by the supplied
  # node ID values
  ndf <-
    ndf |>
    dplyr::filter(id %in% nodes)

  # Extract the node attribute values
  node_attr_vals <- ndf |> dplyr::pull(!!node_attr)

  # Add names to each of the values
  names(node_attr_vals) <- nodes

  node_attr_vals
}

#' Get an aggregate value from the indegree of nodes
#'
#' @description
#'
#' Get a single, aggregate value from the indegree values for all nodes in a
#' graph, or, a subset of graph nodes.
#'
#' @inheritParams render_graph
#' @param agg The aggregation function to use for summarizing indegree values
#'   from graph nodes. The following aggregation functions can be used: `sum`,
#'   `min`, `max`, `mean`, or `median`.
#' @param conditions An option to use filtering conditions for the nodes to
#'   consider.
#'
#' @return A vector with an aggregate indegree value.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 20,
#'     m = 35,
#'     set_seed = 23) %>%
#'   set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = count_nodes(.),
#'       mean = 5,
#'       sd = 1) %>% round(1))
#'
#' # Get the mean indegree value
#' # from all nodes in the graph
#' graph %>%
#'   get_agg_degree_in(
#'     agg = "mean")
#'
#' # Other aggregation functions
#' # can be used (`min`, `max`,
#' # `median`, `sum`); let's get
#' # the median in this example
#' graph %>%
#'   get_agg_degree_in(
#'     agg = "median")
#'
#' # The aggregation of indegree
#' # can occur for a subset of the
#' # graph nodes and this is made
#' # possible by specifying
#' # `conditions` for the nodes
#' graph %>%
#'   get_agg_degree_in(
#'     agg = "mean",
#'     conditions = value > 5.0)
#'
#' @export
get_agg_degree_in <- function(
    graph,
    agg,
    conditions = NULL
) {


  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If filtering conditions are provided then
  # pass in those conditions and filter the ndf
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    # Extract the node data frame from the graph
    ndf <- get_node_df(graph)
    ndf <- dplyr::filter(.data = ndf, {{ conditions }})

    # Get a vector of node ID values
    node_ids <-
      ndf %>%
      dplyr::pull("id")
  }

  # Get a data frame with indegree values for
  # all nodes in the graph
  indegree_df <- get_degree_in(graph)

  if (exists("node_ids")) {
    indegree_df <-
      indegree_df %>%
      dplyr::filter(id %in% node_ids)
  }

  # Verify that the value provided for `agg`
  # is one of the accepted aggregation types
  rlang::arg_match0(agg, c("sum", "min", "max", "mean", "median"))

  # Get the aggregate value of total degree based
  # on the aggregate function provided
  fun <- match.fun(agg)

  indegree_agg <-
    indegree_df %>%
    dplyr::group_by() %>%
    dplyr::summarize(fun(indegree, na.rm = TRUE), .groups = "drop") %>%
    purrr::flatten_dbl()

  indegree_agg
}

#' Export a graph to CSV files
#'
#' @description
#'
#' Export a graph to separate CSV files for nodes and edges.
#'
#' @inheritParams render_graph
#' @param ndf_name The name to provide to the CSV file containing node
#'   information. By default this CSV will be called `nodes.csv`.
#' @param edf_name The name to provide to the CSV file containing edge
#'   information. By default this CSV will be called `edges.csv`.
#' @param output_path The path to which the CSV files will be placed. By
#'   default, this is the current working directory.
#' @param colnames_type Provides options to modify CSV column names to allow for
#'   easier import into other graph systems. The `neo4j` option modifies column
#'   names to allow for direct import of CSVs into Neo4J with the `LOAD CSV`
#'   clause. The `graphframes` option modifies column names to match those
#'   required by the Spark GraphFrames package.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = c("a", "a", "z", "z"),
#'     label = TRUE,
#'     value = c(3.5, 2.6, 9.4, 2.7)
#'   )
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = c("rel_a", "rel_z", "rel_a")
#'   )
#'
#' # Create a graph with the ndf and edf
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf
#'   )
#'
#' # Create separate `nodes.csv` and
#' # `edges.csv` files
#' # graph %>% export_csv()
#'
#' @export
export_csv <- function(
    graph,
    ndf_name = "nodes.csv",
    edf_name = "edges.csv",
    output_path = getwd(),
    colnames_type = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  nodes_df <- get_node_df(graph)
  edges_df <- get_edge_df(graph)

  if (!is.null(colnames_type)) {

    # Modify column names for easier import into Neo4J
    # via `LOAD CSV` or the `neo4j-import` tool
    if (colnames_type == "neo4j") {

      # Modify column names in the ndf
      colnames(nodes_df)[1:3] <-
        c("nodes:ID", ":LABEL", "label")

      # Modify column names in the edf
      colnames(edges_df)[1:3] <-
        c(":START_ID", ":END_ID", ":TYPE")
    }

    # Modify column names for easier import into a
    # Spark GraphFrame using the `spark-csv` pkg with
    # `sqlContext.read.format('com.databricks.spark.csv')`
    if (colnames_type == "graphframes") {

      # Modify column names in the ndf
      colnames(nodes_df)[1] <- "id"

      # Modify column names in the edf
      colnames(edges_df)[1:2] <-
        c("src", "dst")
    }
  }

  # Write the CSV files to the output directory
  utils::write.csv(
    nodes_df,
    file = file.path(output_path, ndf_name, fsep = "/"),
    row.names = FALSE, quote = FALSE)

  utils::write.csv(
    edges_df,
    file = file.path(output_path, edf_name, fsep = "/"),
    row.names = FALSE, quote = FALSE)
}

###
# Graph validation functions
###

#' Check whether a graph object is valid
#'
#' @noRd
graph_object_valid <- function(graph) {

  # Check for all component names to be present
  if (!all(c("graph_info", "nodes_df", "edges_df",
             "global_attrs", "directed",
             "last_node", "last_edge",
             "node_selection", "edge_selection",
             "cache", "graph_log") %in%
           names(graph))) {

    return(FALSE)
  }

  # Check for specific graph classes
  graph_valid <- all(
    inherits(graph$graph_info, "data.frame"),
    inherits(graph$nodes_df, "data.frame"),
    inherits(graph$edges_df, "data.frame"),
    inherits(graph$global_attrs, "data.frame"),
    inherits(graph$global_attrs$attr, "character"),
    inherits(graph$global_attrs$value, "character"),
    inherits(graph$global_attrs$attr_type, "character"),
    inherits(graph$directed, "logical"),
    inherits(graph$node_selection, "data.frame"),
    inherits(graph$edge_selection, "data.frame"),
    inherits(graph$cache, "list"),
    inherits(graph$graph_log, "data.frame")
  )

  graph_valid
}

#' Check whether a graph contains any nodes
#'
#' @noRd
graph_contains_nodes <- function(graph) {

  nrow(graph$nodes_df > 0)
}

#' Check whether a graph contains any edges
#'
#' @noRd
graph_contains_edges <- function(graph) {

  nrow(graph$edges_df) > 0
}

#' Check whether a graph contains a valid node selection
#'
#' @noRd
graph_contains_node_selection <- function(graph) {

  nrow(graph$node_selection) > 0
}

#' Check whether a graph contains a valid edge selection
#'
#' @noRd
graph_contains_edge_selection <- function(graph) {

  nrow(graph$edge_selection) > 0
}

#' Get a list of node/edge selection properties
#'
#' @noRd
node_edge_selection_properties <- function(graph) {

  # Determine if there is an existing
  # selection of nodes
  node_selection_available <-
    graph_contains_node_selection(graph = graph)

  # Determine if there is an existing
  # selection of edges
  edge_selection_available <-
    graph_contains_edge_selection(graph)

  # Get the existing node/edge count
  if (node_selection_available || edge_selection_available) {

    selection_type <-
      ifelse(node_selection_available, "node", "edge")

    if (selection_type == "node") {
      selection_count <- nrow(graph$node_selection)
      selection_count_str <-
        paste0(selection_count, " node", ifelse(selection_count > 1, "s", ""))

    } else {

      selection_count <- nrow(graph$edge_selection)
      selection_count_str <-
        paste0(selection_count, " edge", ifelse(selection_count > 1, "s", ""))
    }

  } else {

    selection_type <- NA_character_
    selection_count <- 0
    selection_count_str <- "no nodes or edges"
  }

  list(
    node_selection_available = node_selection_available,
    edge_selection_available = edge_selection_available,
    selection_type = selection_type,
    selection_count = selection_count,
    selection_count_str = selection_count_str)
}

#' Replace the `node_selection` df with different node ID values
#'
#' @noRd
replace_graph_node_selection <- function(graph,
                                         replacement) {

  # Create a new table for the node selection
  dplyr::bind_rows(
    create_empty_nsdf(),
    dplyr::tibble(node = as.integer(replacement))
  )
}

#' Replace the `edge_selection` df with different edge ID values
#'
#' @noRd
replace_graph_edge_selection <- function(graph,
                                         edge_id,
                                         from_node,
                                         to_node) {

  # Create a new table for the edge selection
  dplyr::bind_rows(
    create_empty_esdf(),
    dplyr::tibble(
      edge = as.integer(edge_id),
      from = as.integer(from_node),
      to = as.integer(to_node)
    )
  )
}

create_empty_nsdf <- function() {

  # Create empty `nsdf`
  dplyr::tibble(node = integer()) %>%
    as.data.frame(stringsAsFactors = FALSE)
}

create_empty_esdf <- function() {

  # Create empty `esdf`
  data.frame(
    edge = integer(),
    from = integer(),
    to = integer()
  )
}

#' Is an attribute unique and fully free of `NA`s?
#'
#' This function determines whether a node or edge attribute has values that
#' are all non-`NA` and are unique
#'
#' @noRd
is_attr_unique_and_non_na <- function(graph,
                                      which_graph_df,
                                      attr) {

  if (which_graph_df == "ndf") {
    df <- graph$nodes_df
  } else if (which_graph_df == "edf") {
    df <- graph$edges_df
  } else {
    stop(
      "The `which_graph_df` argument must be either `ndf` or `edf`.",
      call. = FALSE)
  }

  if (!(attr %in% colnames(df))) {
    stop(
      "The `attr` provided is not available.",
      call. = FALSE)
  }

  # Are all values not NA?
  all_is_not_na <-
    !all(is.na(dplyr::pull(df, !!enquo(attr))))

  # Are all values distinct?
  all_values_distinct <-
    dplyr::pull(df, !!enquo(attr)) %>% dplyr::n_distinct() ==
    nrow(df)

  all_is_not_na && all_values_distinct
}

###
# Graph transformation functions
###

#' Transform `from` and `to` edge attributes
#'
#' This function takes a graph object and labels for `from` and `to` values and
#' then translates the `from`/`to` label values to node ID values
#'
#' @noRd
translate_to_node_id <- function(graph, from, to) {

  # Check that node labels are unique
  if (length(unique(graph$nodes_df$label)) !=
      count_nodes(graph)) {

    stop(
      "You cannot use labels to define edges because they are not distinct.",
      call. = FALSE)
  }

  # No node labels can be empty
  if (any(graph$nodes_df$label == "")) {

    stop(
      "You cannot use labels to define edges since there are empty strings for labels.",
      call. = FALSE)
  }

  # Create the `from_id` and `to_id` vectors
  from_id <- vector("integer")
  to_id <- vector("integer")

  # Get an ordered vector of node ID values
  # as `from` nodes
  for (i in seq_along(from)) {
    from_id <-
      c(from_id,
        graph$nodes_df[
          which(graph$nodes_df$label %in% from[i]), 1])
  }

  # Get an ordered vector of node ID values
  # as `to` nodes
  for (i in seq_along(to)) {
    to_id <-
      c(to_id,
        graph$nodes_df[
          which(graph$nodes_df$label %in% to[i]), 1])
  }

  # Reassign these nodes back to `from` and `to`
  from <- from_id
  to <- to_id

  list(from = from_id, to = to_id)
}


###
# Graph logging functions
###

#' Get the time of the graph function invocation
#'
#' @noRd
graph_function_sys_time <- function() {
  Sys.time()
}

#' Get the time taken for a graph function to execute
#'
#' This calculates the difference from the start of the function call (relies on
#' a call of the `graph_function_sys_time()` function) to the time of invoking
#' this function
#'
#' @noRd
graph_function_duration <- function(start_time) {
  end_time <- Sys.time()
  time_diff_s <- (end_time - start_time)[[1]]

  time_diff_s
}

#' Add a log line for a graph `action`
#'
#' @noRd
add_action_to_log <- function(graph_log,
                              version_id,
                              function_used,
                              time_modified,
                              duration,
                              nodes,
                              edges,
                              d_n = 0L,
                              d_e = 0L) {

  # Ensure that `time_modified` inherits from POSIXct
  if (!inherits(time_modified, "POSIXct")) {

    cli::cli_abort(
      "The `time_modified` value must inherit from POSIXct.",
      call. = FALSE)
  }

  # Create a log line
  graph_log_line <-
    data.frame(
      version_id = version_id,
      function_used = as.character(function_used),
      time_modified = time_modified,
      duration = as.numeric(duration),
      nodes = nodes,
      edges = edges,
      d_n = as.integer(d_n),
      d_e = as.integer(d_e),
      stringsAsFactors = FALSE)

  # Append the log line to `graph_log`
  dplyr::bind_rows(graph_log, graph_log_line)
}

#' Save the graph as an RDS file
#'
#' @noRd
save_graph_as_rds <- function(graph) {

  # Construct a file name for the RDS
  rds_filename <-
    paste0(
      graph$graph_info$graph_id, "_",
      formatC(
        graph$graph_log$version_id[nrow(graph$graph_log)],
        width = 6, format = "d", flag = "0"), "_",
      round(
        as.integer(
          graph$graph_log$time_modified[nrow(graph$graph_log)]), 4),
      ".rds")

  # Construct the subdirectory name for the RDS-based
  # graph backups
  rds_dir_name <-
    paste0("backup_", graph$graph_info$graph_id)

  # If directory doesn't exist, create the directory
  # inside of the working directory
  if (!dir.exists(rds_dir_name)) {
    dir.create(rds_dir_name)
  }

  # Save the graph as an RDS file in the subdirectory
  saveRDS(graph, file = file.path(rds_dir_name, rds_filename, fsep = "/"))
}

###
# Selection helper functions
###

#' Get column names from a column selection statement
#'
#' @noRd
get_col_selection <- function(col_selection_stmt) {

  if (length(col_selection_stmt) == 1 &&
      all(stringr::str_detect(
        string = col_selection_stmt,
        pattern = "^([a-zA-Z_\\.][a-zA-Z0-9_\\.]*?|`.*?`)$")
        )) {

    selection_type <- "single_column_name"

    # Get the column names
    column_selection <- col_selection_stmt

  } else if (length(col_selection_stmt) == 1 &&
             all(stringr::str_detect(
               string = col_selection_stmt,
               pattern = "(.* & ).*")
               )) {

    selection_type <- "column_names"

    # Get the column names
    column_selection <-
      stringr::str_split(
        string = col_selection_stmt,
        pattern = " & ") %>%
      unlist()

  } else if (any(stringr::str_detect(
    string = col_selection_stmt,
    pattern = "^[0-9]*?:[0-9]*?$"))) {

    selection_type <- "column_index_range"

    # Get the column indices
    column_selection <-
      seq(
        from = (stringr::str_split(
          string = col_selection_stmt,
          pattern = ":") %>%
            unlist())[1] %>%
          as.numeric(),
        to = (stringr::str_split(
          string = col_selection_stmt,
          pattern = ":") %>%
            unlist())[2] %>%
          as.numeric())

  } else if (any(
    stringr::str_detect(
      string = col_selection_stmt,
      pattern = "^([a-zA-Z_\\.][a-zA-Z0-9_\\.]*?|`.*?`):([a-zA-Z_\\.][a-zA-Z0-9_\\.]*?|`.*?`)$"))
  ) {

    selection_type <- "column_range"

    # Get the first and last column names
    column_selection <-
      c(
        (stringr::str_split(
          string = col_selection_stmt,
          pattern = ":") %>%
           unlist())[1],
        (stringr::str_split(
          string = col_selection_stmt,
          pattern = ":") %>%
           unlist())[2])
  } else {
    return(list())
  }

  list(
    selection_type = selection_type,
    column_selection = column_selection)
}

###
# Aesthetic attribute functions
###

#' Get the ideal contrasting text color given the fill color of a node
#'
#' @noRd
contrasting_text_color <- function(background_color) {

  rgb_colors <-
    ((grDevices::col2rgb(background_color) %>%
        as.numeric()) / 255)^2.2

  luminance <-
    (0.2126 * rgb_colors[1]) +
    (0.7152 * rgb_colors[2]) +
    (0.0722 * rgb_colors[3])

  saturation <-
    (max(rgb_colors) - min(rgb_colors) + 0.00001) /
    (max(rgb_colors) + 0.00001)

  if (saturation < 0.35) {
    if (luminance > 0.5) {
      contrasting_color <- "#000000"
    }
  } else {
    contrasting_color <- "#FFFFFF"
  }

  contrasting_color
}

###
# Functions that help build info messages
###

#' Construct a consistent message string, passing it to `message()`
#'
#' @noRd
emit_message <- function(fcn_name,
                         message_body) {

  glue::glue("`{fcn_name}()` INFO: {message_body}") %>%
    as.character() %>%
    message()
}

#' Get the calling function as a formatted character string
#'
#' @noRd
get_calling_fcn <- function() {

  calling_fcn <- deparse(sys.call(-1))

  stringr::str_replace_all(
    calling_fcn,
    pattern = "([a-z0-9_]*)(.*)",
    replacement = "\\1")
}

###
# Functions that produce useful vectors
###


#' Produce a vector of Graphviz graph attribute names
#'
#' @noRd
gv_graph_attributes <- function() {

  c("layout", "bgcolor", "rankdir",
    "overlap", "outputorder", "fixedsize",
    "mindist", "nodesep", "ranksep",
    "stylesheet")
}

#' Produce a vector of Graphviz node attribute names
#'
#' @noRd
gv_node_attributes <- function() {

  c("shape", "style", "penwidth", "color", "fillcolor",
    "fontname", "fontsize", "fontcolor", "image", "fa_icon",
    "height", "width", "group",
    "tooltip", "xlabel", "URL",
    "distortion", "sides", "skew", "peripheries",
    "gradientangle", "label", "fixedsize",
    "labelloc", "margin", "orientation", "pos")
}

#' Produce a vector of Graphviz edge attribute names
#'
#' @noRd
gv_edge_attributes <- function() {

  c("style", "penwidth", "color", "arrowsize",
    "arrowhead", "arrowtail",
    "fontname", "fontsize", "fontcolor",
    "len", "tooltip", "URL",
    "label", "labelfontname", "labelfontsize",
    "labelfontcolor", "labeltooltip", "labelURL",
    "edgetooltip", "edgeURL",
    "headtooltip", "headURL",
    "headclip", "headlabel", "headport",
    "tailtooltip", "tailURL",
    "tailclip",  "taillabel", "tailport",
    "dir", "decorate")
}

#' Produce a vector of node creation function names
#'
#' @noRd
node_creation_functions <- function() {

  c("add_node", "add_n_nodes", "add_n_node_clones",
    "add_node_clones_ws", "add_n_nodes_ws",
    "add_node_df", "add_nodes_from_df_cols",
    "add_nodes_from_table", "add_full_graph",
    "add_balanced_tree", "add_cycle",
    "add_path", "add_prism", "add_star",
    "add_grid_2d", "add_grid_3d", "add_gnm_graph",
    "add_gnp_graph", "add_pa_graph",
    "add_smallworld_graph", "add_growing_graph",
    "add_islands_graph")
}

#' Produce a vector of node deletion function names
#'
#' @noRd
node_deletion_functions <- function() {
  # previously, create_subgraph_ws, and create_complement_graph?
  c("transform_to_subgraph_ws", "transform_to_complement_graph",
    "delete_node", "delete_nodes_ws")
}

#' Produce a vector of edge creation function names
#'
#' @noRd
edge_creation_functions <- function() {

  c("add_edge", "add_edge_clone", "add_edges_w_string",
    "add_edge_df", "add_forward_edges_ws",
    "add_reverse_edges_ws",
    "add_edges_from_table", "add_full_graph",
    "add_balanced_tree", "add_cycle",
    "add_path", "add_prism", "add_star",
    "add_grid_2d", "add_grid_3d", "add_gnm_graph",
    "add_gnp_graph", "add_pa_graph",
    "add_smallworld_graph", "add_growing_graph",
    "fully_connect_nodes_ws")
}

#' Produce a vector of edge deletion function names
#'
#' @noRd
edge_deletion_functions <- function() {

  c("delete_edge", "delete_edges_ws",
    "create_subgraph_ws", "create_complement_graph",
    "delete_node", "delete_nodes_ws", "delete_loop_edges_ws",
    "fully_disconnect_nodes_ws")
}

#' Produce vector of graph initializing function names
#'
#' @noRd
graph_init_functions <- function() {
  c("create_graph",
    "from_igraph", "from_adj_matrix",
    "import_graph")
}

#' Produce a list of data frames for functions that get node properties
#'
#' @noRd
value_per_node_functions <- function() {

  list(
    "get_alpha_centrality" =
      data.frame(
        arg = c("alpha", "exo", "weights_attr", "tol"),
        value_type = c("numeric", "numeric", "character", "numeric"),
        stringsAsFactors = FALSE
      ),
    "get_authority_centrality" =
      data.frame(
        arg = "weights_attr",
        value_type = "character",
        stringsAsFactors = FALSE
      ),
    "get_betweenness" =
      data.frame(
        arg = NULL,
        value_type = NULL
      ),
    "get_closeness" =
      data.frame(
        arg = "direction",
        value_type = "character",
        stringsAsFactors = FALSE
      ),
    "get_cmty_edge_btwns" =
      data.frame(
        arg = NULL,
        value_type = NULL
      ),
    "get_cmty_fast_greedy" =
      data.frame(
        arg = NULL,
        value_type = NULL
      ),
    "get_cmty_l_eigenvec" =
      data.frame(
        arg = NULL,
        value_type = NULL
      ),
    "get_cmty_louvain" =
      data.frame(
        arg = NULL,
        value_type = NULL
      ),
    "get_cmty_walktrap" =
      data.frame(
        arg = "steps",
        value_type = "numeric",
        stringsAsFactors = FALSE
      ),
    "get_degree_distribution" =
      data.frame(
        arg = "mode",
        value_type = "character",
        stringsAsFactors = FALSE
      ),
    "get_degree_histogram" =
      data.frame(
        arg = "mode",
        value_type = "character",
        stringsAsFactors = FALSE
      ),
    "get_degree_in" =
      data.frame(
        arg = "normalized",
        value_type = "logical",
        stringsAsFactors = FALSE
      ),
    "get_degree_out" =
      data.frame(
        arg = "normalized",
        value_type = "logical",
        stringsAsFactors = FALSE
      ),
    "get_degree_total" =
      data.frame(
        arg = "normalized",
        value_type = "logical",
        stringsAsFactors = FALSE
      ),
    "get_eccentricity" =
      data.frame(
        arg = NULL,
        value_type = NULL
      ),
    "get_eigen_centrality" =
      data.frame(
        arg = "weights_attr",
        value_type = "character",
        stringsAsFactors = FALSE
      ),
    "get_pagerank" =
      data.frame(
        arg = c("directed", "damping"),
        value_type = c("logical", "numeric"),
        stringsAsFactors = FALSE
      ),
    "get_s_connected_cmpts" =
      data.frame(
        arg = NULL,
        value_type = NULL
      ),
    "get_w_connected_cmpts" =
      data.frame(
        arg = NULL,
        value_type = NULL
      )
  )
}


#' Get the `df_id` values for a graph's internal NDF or EDF
#'
#' @noRd
get_df_ids <- function(graph_df) {

  if (nrow(graph_df) > 0) {

    if ("df_id" %in% colnames(graph_df)) {

      graph_df %>%
        dplyr::select("df_id") %>%
        dplyr::filter(!is.na(df_id)) %>%
        dplyr::pull("df_id")
    } else {
      return(NA_character_)
    }

  } else if (nrow(graph_df) == 0) {
    return(NA_character_)
  }
}


#' Scavenge the `graph$df_storage` list
#'
#' This removes any linked data frames if the associated nodes or edges no
#' longer exist.
#'
#' @noRd
remove_linked_dfs <- function(graph) {

  if (is.null(graph$df_storage)) {
    return(graph)
  }

  ndf_df_ids <-
    graph %>%
    get_node_df() %>%
    get_df_ids()

  edf_df_ids <-
    graph %>%
    get_edge_df() %>%
    get_df_ids()

  # Determine if any of the stored
  # data frames are not available in
  # the graph's internal node data frame
  if (length(graph$df_storage) > 0) {

    ndf_df_id_to_remove <-
      graph$df_storage %>%
      dplyr::bind_rows() %>%
      dplyr::filter(node_edge__ == "node") %>%
      dplyr::select("df_id__") %>%
      dplyr::distinct() %>%
      dplyr::pull("df_id__") %>%
      base::setdiff(ndf_df_ids)

    # If any stored data frames are associated
    # with edges that no longer exist, remove them
    if (length(ndf_df_id_to_remove) > 0) {

      for (i in seq_along(ndf_df_id_to_remove)) {

        graph$df_storage[ndf_df_id_to_remove[i]] <- NULL
      }
    }
  }

  # Determine if any of the stored
  # data frames are not available in
  # the graph's internal edge data frame
  if (length(graph$df_storage) > 0) {
    edf_df_id_to_remove <-
      graph$df_storage %>%
      dplyr::bind_rows() %>%
      dplyr::filter(node_edge__ == "edge") %>%
      dplyr::select("df_id__") %>%
      dplyr::distinct() %>%
      dplyr::pull("df_id__") %>%
      base::setdiff(edf_df_ids)


    # If any stored data frames are associated
    # with edges that no longer exist, remove them
    if (length(edf_df_id_to_remove) > 0) {

      for (i in seq_along(edf_df_id_to_remove)) {

        graph$df_storage[edf_df_id_to_remove[i]] <- NULL
      }
    }
  }

  # Check the type of list that remains
  if (length(graph$df_storage) == 0) {
    graph$df_storage <-
      unname(graph$df_storage)
  }

  # Remove the `df_id` column from the
  # graph's ndf if there are no referenced
  # data frames within (i.e., all NA)
  if (rlang::has_name(graph$nodes_df, "df_id") &&
      all(is.na(graph$nodes_df$df_id))) {

    graph$nodes_df$df_id <- NULL
  }

  # Remove the `df_id` column from the
  # graph's edf if there are no referenced
  # data frames within (i.e., all NA)
  if (rlang::has_name(graph$edges_df, "df_id") &&
      all(is.na(graph$edges_df$df_id))) {

    graph$edges_df$df_id <- NULL
  }

  graph
}

###
# Functions for working with graph SVG output
###

#' Create a table of SVG attributes
#'
#' @noRd
get_svg_tbl <- function(svg_vec) {

  svg_tbl <-
    dplyr::tibble(
      index = integer(),
      type = character(),
      node_id = integer(),
      from = integer(),
      to = integer())

  for (i in seq(svg_vec)) {

    line <- svg_vec[i]

    if (grepl("^<\\?xml.*", line)) {
      rec <- dplyr::tibble(index = i, type = "xml_declaration")
    } else if (grepl("(^<!DOCTYPE.*|.*http://www.w3.org.*)", line)) {
      rec <- dplyr::tibble(index = i, type = "doctype_declaration")
    } else if (grepl("(<!-- Generated by|^ -->$)", line)) {
      rec <- dplyr::tibble(index = i, type = "generator_comment")
    } else if (grepl("<!-- Title:", line)) {
      rec <- dplyr::tibble(index = i, type = "title_block")
    } else if (grepl("^<svg", line)) {
      rec <- dplyr::tibble(index = i, type = "svg") %>% dplyr::bind_cols(get_attr_tbl(line))
    } else if (grepl(" viewBox", line)) {
      rec <- dplyr::tibble(index = i, type = "viewbox_info")
    } else if (grepl("^<g ", line)) {
      rec <- dplyr::tibble(index = i, type = "g") %>% dplyr::bind_cols(get_attr_tbl(line))
    } else if (grepl("^<title>", line)) {
      rec <- dplyr::tibble(index = i, type = "title") %>% dplyr::bind_cols(get_inner_html(line))
    } else if (grepl("^<polygon", line)) {
      rec <- dplyr::tibble(index = i, type = "polygon") %>% dplyr::bind_cols(get_attr_tbl(line))
    } else if (grepl("^<path", line)) {
      rec <- dplyr::tibble(index = i, type = "path") %>% dplyr::bind_cols(get_attr_tbl(line))
    } else if (grepl("^<ellipse", line)) {
      rec <- dplyr::tibble(index = i, type = "ellipse") %>% dplyr::bind_cols(get_attr_tbl(line))
    } else if (grepl("<!-- [0-9]*? -->", line)) {
      node_id <- gsub("(<!-- | -->)", "", line) %>% as.integer()
      rec <- dplyr::tibble(index = i, type = "node_block", node_id = node_id)
    } else if (grepl("<!-- [0-9]*?&#45;&gt;[0-9]*? -->", line)) {
      from_node_id <- gsub("(<!-- |?&#45;&gt;[0-9]*? -->)", "", line) %>% as.integer()
      to_node_id <- gsub("(<!-- [0-9]*?&#45;&gt;| -->)", "", line) %>% as.integer()
      rec <- dplyr::tibble(index = i, type = "edge_block", from = from_node_id, to = to_node_id)
    } else if (grepl("<!-- [0-9]*?&#45;&#45;[0-9]*? -->", line)) {
      from_node_id <- gsub("(<!-- |?&#45;&#45;[0-9]*? -->)", "", line) %>% as.integer()
      to_node_id <- gsub("(<!-- [0-9]*?&#45;&#45;| -->)", "", line) %>% as.integer()
      rec <- dplyr::tibble(index = i, type = "edge_block", from = from_node_id, to = to_node_id)
    } else if (grepl("^<text ", line)) {
      rec <- dplyr::tibble(index = i, type = "text") %>%
        dplyr::bind_cols(get_attr_tbl(line)) %>%
        dplyr::bind_cols(get_inner_html(line))
    } else if (grepl("</g>", line)) {
      rec <- dplyr::tibble(index = i, type = "g_close")
    } else if (grepl("</svg>", line)) {
      rec <- dplyr::tibble(index = i, type = "svg_close")
    } else {
      rec <- dplyr::tibble(index = i, type = "unknown")
    }

    svg_tbl <- dplyr::bind_rows(svg_tbl, rec)
  }

  svg_tbl %>% tidyr::fill(node_id)
}

#' Function to create a one-row table of attr-value pairs
#'
#' @noRd
get_attr_tbl <- function(line) {

  line <- gsub("<[a-z]*? ", "", line) %>% gsub("\"", "'", .) %>% gsub("(/>|>|>.*)", "", .)

  el_attrs <-
    strsplit(line, "' ") %>%
    unlist() %>%
    gsub("'", "", .) %>%
    strsplit("=")

  stats::setNames(
    sapply(el_attrs, `[[`, 2),
    sapply(el_attrs, `[[`, 1)) %>%
    as.list() %>%
    dplyr::as_tibble()
}

#' Function to create a one-row, one-column table of inner HTML
#'
#' @noRd
get_inner_html <- function(line) {

  inner_html <- gsub("(^<.*?>|</.*?>)", "", line)

  dplyr::tibble(inner_html = inner_html)
}

#' Clear an active selection of nodes or edges
#'
#' @description
#'
#' Clear the selection of nodes or edges within a graph object.
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with
#' # a single path
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 5)
#'
#' # Select nodes with IDs `1`
#' # and `3`
#' graph <-
#'   graph %>%
#'   select_nodes(
#'     nodes = c(1, 3))
#'
#' # Verify that a node selection
#' # has been made
#' graph %>% get_selection()
#'
#' # Clear the selection with
#' # `clear_selection()`
#' graph <-
#'   graph %>%
#'   clear_selection()
#'
#' # Verify that the node
#' # selection has been cleared
#' graph %>% get_selection()
#'
#' @export
clear_selection <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # Clear the selection of nodes and edges in the graph
  graph$node_selection <- create_empty_nsdf()
  graph$edge_selection <- create_empty_esdf()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # Emit a message about the modification of a selection
  # if that option is set
  if (isTRUE(graph$graph_info$display_msgs)) {

    # Issue a message to the user
    if (n_e_select_properties_in[["selection_count"]] > 0) {

      emit_message(
        fcn_name = fcn_name,
        message_body = glue::glue(
          "cleared an existing selection of \\
       {n_e_select_properties_in[['selection_count_str']]}"))

    } else {

      emit_message(
        fcn_name = fcn_name,
        message_body = "no existing selection to clear; graph unchanged")
    }
  }

  graph
}

# Function for checking graph
check_graph_valid <- function(graph, error_call = rlang::caller_env()) {
  if (!graph_object_valid(graph)) {
    rlang::abort(
      "The graph is not valid.",
      call = error_call)
  }
}

check_graph_contains_nodes <- function(graph,
                                       extra_msg = NULL,
                                       error_call = rlang::caller_env()
) {
  if (!graph_contains_nodes(graph)) {
    rlang::abort(c(
      "The graph does not contain nodes.",
      extra_msg),
      call = error_call)
  }
}

check_graph_contains_edges <- function(graph, error_call = rlang::caller_env()) {
  if (!graph_contains_edges(graph)) {
    rlang::abort(
      "The graph contains no edges.",
      call = error_call)
  }
}

check_graph_contains_node_selection <- function(graph,
                                                extra_msg = NULL,
                                                error_call = rlang::caller_env()
) {
  if (!graph_contains_node_selection(graph)) {
    rlang::abort(c(
      "There is no selection of nodes available.",
      extra_msg),
      call = error_call)
  }
}

check_graph_contains_edge_selection <- function(graph,
                                                extra_msg = NULL,
                                                error_call = caller_env()) {
  if (!graph_contains_edge_selection(graph)) {
    rlang::abort(c(
      "The graph contains no selection of edges.",
      extra_msg),
      call = error_call
    )
  }
}

check_property_graph <- function(graph,
                                 error_call = caller_env()) {
  if (!is_property_graph(graph)) {
    abort(c(
      "The graph object is not a property graph"),
      call = error_call
    )
  }
}

#' Delete a node from an existing graph object
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, delete an existing node by
#' specifying its node ID.
#'
#' @inheritParams render_graph
#' @param node A node ID for the node to be deleted from the graph.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 5 nodes and
#' # edges between each in a path
#' graph <-
#'   create_graph() |>
#'   add_path(n = 5)
#'
#' # Delete node with ID `3`
#' graph <- delete_node(graph, node = 3)
#'
#' # Verify that the node with ID `3`
#' # is no longer in the graph
#' graph |> get_node_ids()
#'
#' # Also note that edges are removed
#' # since there were edges between the
#' # removed node to and from other nodes
#' graph |> get_edges()
#' @family node creation and removal
#' @export
delete_node <- function(
    graph,
    node
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, "So, a node cannot be deleted.")

  # Verify that `node` is given as a single value
  node_is_single_value <- length(node) == 1

  # Stop function if node not a single value
  if (!node_is_single_value) {

    cli::cli_abort(
      "Only a single node can be deleted.")
  }

  # Stop function if node is not in the graph
  if (!(node %in% get_node_ids(graph))) {

    cli::cli_abort(
      "The specified node is not available in the graph.")
  }

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph |> count_nodes()

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Get the graph's node data frame
  ndf <- graph$nodes_df

  # Get the graph's edge data frame
  edf <- graph$edges_df

  # Remove node from `ndf`
  ndf <-
    ndf |>
    dplyr::filter(id != node)

  # Remove any edges connected to `node`
  # in the `edf`
  edf <-
    edf |>
    dplyr::filter(!(from == node | to == node))

  # Reset the row names in the ndf and the edf
  row.names(ndf) <- NULL
  row.names(edf) <- NULL

  # Update the graph's node and edge data frames
  graph$nodes_df <- ndf
  graph$edges_df <- edf

  # Scavenge any invalid, linked data frames
  graph <-
    graph |>
    remove_linked_dfs()

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph |> count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_deleted <- nodes_graph_2 - nodes_graph_1

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph |> count_edges()

  # Get the number of edges added to
  # the graph
  edges_deleted <- edges_graph_2 - edges_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_deleted,
      d_e = edges_deleted)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get non-neighbors of a node in a graph
#'
#' @description
#'
#' Get the set of all nodes not neighboring a single graph node.
#'
#' @inheritParams render_graph
#' @param node A single-length vector containing a node ID value.
#'
#' @return A vector of node ID values.
#'
#' @examples
#' # Create a simple, directed graph with 5
#' # nodes and 4 edges
#' graph <-
#'   create_graph() |>
#'   add_path(n = 5)
#'
#' # Find all non-neighbors of node `2`
#' graph |> get_non_nbrs(node = 2)
#'
#' @export
get_non_nbrs <- function(
    graph,
    node
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get predecessors and successors for the `node`
  node_nbrs <-
    c(get_predecessors(graph, node = node),
      get_successors(graph, node = node))

  # Get all non-neighbors to the `node`
  node_non_nbrs <-
    base::setdiff(
      base::setdiff(get_node_ids(graph), node),
      node_nbrs)

  # Get a unique set of node ID values
  node_non_nbrs <- sort(unique(node_non_nbrs))

  # If there are no non-neighbors, then return `NA`
  if (length(node_non_nbrs) == 0) {
    return(NA)
  }

  node_non_nbrs
}

#' Add a cycle of nodes to the graph
#'
#' @description
#'
#' With a graph object of class `dgr_graph`, add a node cycle to the graph.
#'
#' @inheritParams node_edge_aes_data
#' @inheritParams render_graph
#' @param n The number of nodes comprising the cycle.
#' @param type An optional string that describes the entity type for the nodes
#'   to be added.
#' @param label Either a vector object of length `n` that provides optional
#'   labels for the new nodes, or, a boolean value where setting to `TRUE`
#'   ascribes node IDs to the label and `FALSE` yields a blank label.
#' @param rel An optional string for providing a relationship label to all new
#'   edges created in the node cycle.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a new graph and
#' # add a cycle of nodes to it
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 6)
#'
#' # Get node information
#' # from this graph
#' graph |>
#'   get_node_info()
#'
#' # Node and edge aesthetic and data
#' # attributes can be specified in
#' # the `node_aes`, `edge_aes`,
#' # `node_data`, and `edge_data`
#' # arguments
#'
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' graph_w_attrs <-
#'   create_graph() |>
#'   add_cycle(
#'     n = 3,
#'     label = c(
#'       "one", "two", "three"),
#'     type = c(
#'       "a", "a", "b"),
#'     rel = "A",
#'     node_aes = node_aes(
#'       fillcolor = "steelblue"),
#'     edge_aes = edge_aes(
#'       color = "red",
#'       penwidth = 1.2),
#'     node_data = node_data(
#'       value = c(
#'         1.6, 2.8, 3.4)),
#'     edge_data = edge_data(
#'       value =
#'         rnorm(
#'           n = 3,
#'           mean = 5.0,
#'           sd = 1.0)))
#'
#' # Get the graph's node data frame
#' graph_w_attrs |> get_node_df()
#'
#' # Get the graph's edge data frame
#' graph_w_attrs |> get_edge_df()
#'
#' @export
add_cycle <- function(
    graph,
    n,
    type = NULL,
    label = TRUE,
    rel = NULL,
    node_aes = NULL,
    edge_aes = NULL,
    node_data = NULL,
    edge_data = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Stop if n is too small
  check_number_whole(n, min = 3)

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  # Get the number of edges ever created for
  # this graph
  edges_created <- graph$last_edge

  # Get the graph's global attributes
  global_attrs <- graph$global_attrs

  # Get the graph's log
  graph_log <- graph$graph_log

  # Get the graph's info
  graph_info <- graph$graph_info

  # Get the graph's state of being directed
  # or undirected
  graph_directed <- graph$directed

  # Get the sequence of nodes required
  nodes <- seq_len(n)

  # Collect node aesthetic attributes
  if (!is.null(node_aes)) {

    node_aes_tbl <- dplyr::as_tibble(node_aes)

    if (nrow(node_aes_tbl) < n) {

      node_aes$index__ <- seq_len(n)

      node_aes_tbl <-
        dplyr::as_tibble(node_aes) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_aes_tbl)) {
      node_aes_tbl$id <- NULL
    }
  }

  # Collect edge aesthetic attributes
  if (!is.null(edge_aes)) {

    edge_aes_tbl <- dplyr::as_tibble(edge_aes)

    if (nrow(edge_aes_tbl) < n) {

      edge_aes$index__ <- seq_len(n)

      edge_aes_tbl <-
        dplyr::as_tibble(edge_aes) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_aes_tbl)) {
      edge_aes_tbl$id <- NULL
    }
  }

  # Collect node data attributes
  if (!is.null(node_data)) {

    node_data_tbl <- dplyr::as_tibble(node_data)

    if (nrow(node_data_tbl) < n) {

      node_data$index__ <- seq_len(n)

      node_data_tbl <-
        dplyr::as_tibble(node_data) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_data_tbl)) {
      node_data_tbl$id <- NULL
    }
  }

  # Collect edge data attributes
  if (!is.null(edge_data)) {

    edge_data_tbl <- dplyr::as_tibble(edge_data)

    if (nrow(edge_data_tbl) < n) {

      edge_data$index__ <- seq_len(n)

      edge_data_tbl <-
        dplyr::as_tibble(edge_data) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_data_tbl)) {
      edge_data_tbl$id <- NULL
    }
  }

  # Create a node data frame for the cycle graph
  cycle_nodes <-
    create_node_df(
      n = n,
      type = type,
      label = label)

  # Add node aesthetics if available
  if (exists("node_aes_tbl")) {

    cycle_nodes <-
      cycle_nodes |>
      dplyr::bind_cols(node_aes_tbl)
  }

  # Add node data if available
  if (exists("node_data_tbl")) {

    cycle_nodes <-
      cycle_nodes |>
      dplyr::bind_cols(node_data_tbl)
  }

  # Create an edge data frame for the cycle graph
  cycle_edges <-
    create_edge_df(
      from = nodes,
      to = c(nodes[2:length(nodes)], nodes[1]),
      rel = rel)

  # Add edge aesthetics if available
  if (exists("edge_aes_tbl")) {

    cycle_edges <-
      cycle_edges |>
      dplyr::bind_cols(edge_aes_tbl)
  }

  # Add edge data if available
  if (exists("edge_data_tbl")) {

    cycle_edges <-
      cycle_edges |>
      dplyr::bind_cols(edge_data_tbl)
  }

  # Create the cycle graph
  cycle_graph <-
    create_graph(
      directed = graph_directed,
      nodes_df = cycle_nodes,
      edges_df = cycle_edges)

  # If the input graph is not empty, combine graphs
  # using the `combine_graphs()` function
  if (is_graph_empty(graph)) {
    graph <- cycle_graph
  } else {
    graph <- combine_graphs(graph, cycle_graph)
  }

  # Update the `last_node` counter
  graph$last_node <- nodes_created + nrow(cycle_nodes)

  # Update the `last_edge` counter
  graph$last_edge <- edges_created + nrow(cycle_edges)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph_log <-
    add_action_to_log(
      graph_log = graph_log,
      version_id = nrow(graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = n,
      d_e = n)

  graph$global_attrs <- global_attrs
  graph$graph_log <- graph_log
  graph$graph_info <- graph_info

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Add nodes and attributes to graph from a table
#'
#' @description
#'
#' Add nodes and their attributes to an existing graph object from data in a CSV
#' file or a data frame.
#'
#' @inheritParams render_graph
#' @param table Either a path to a CSV file, or, a data frame object.
#' @param label_col An option to apply a column of data in the table as `label`
#'   attribute values.
#' @param type_col An option to apply a column of data in the table as `type`
#'   attribute values.
#' @param set_type An optional string to apply a `type` attribute to all nodes
#'   created from data in the external table.
#' @param drop_cols An optional column selection statement for dropping columns
#'   from the external table before inclusion as attributes in the graph's
#'   internal node data frame. Several columns can be dropped by name using the
#'   syntax `col_1 & col_2 & ...`. Columns can also be dropped using a numeric
#'   column range with `:` (e.g., `5:8`), or, by using the `:` between column
#'   names to specify the range (e.g., `col_5_name:col_8_name`).
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # To add nodes from the dataset called
#' # `currencies` (available as a dataset
#' # in the package), call the
#' # `add_nodes_from_table()` function
#' # after creating an empty graph; new
#' # node ID values will be created as
#' # monotonically-increasing values
#' graph_1 <-
#'   create_graph() %>%
#'   add_nodes_from_table(
#'     table = currencies)
#'
#' # View part of the graph's internal
#' # node data frame (ndf)
#' graph_1 %>%
#'   get_node_df() %>%
#'   .[, 1:5] %>%
#'   head()
#'
#' # If you would like to assign
#' # any of the table's columns as
#' # `type` or `label` attributes,
#' # this can be done with the `type_col`
#' # and `label_col` arguments; to set
#' # a static `type` attribute for all
#' # of the table records, use `set_type`
#' graph_2 <-
#'   create_graph() %>%
#'   add_nodes_from_table(
#'     table = currencies,
#'     label_col = iso_4217_code,
#'     set_type = currency)
#'
#' # View part of the graph's internal ndf
#' graph_2 %>%
#'   get_node_df() %>%
#'   .[, 1:5] %>%
#'   head()
#'
#' # Suppose we would like to not
#' # include certain columns from the
#' # external table in the resulting
#' # graph; we can use the `drop_cols`
#' # argument to choose which columns
#' # to not include as attributes
#' graph_3 <-
#'   create_graph() %>%
#'   add_nodes_from_table(
#'     table = currencies,
#'     label_col = iso_4217_code,
#'     set_type = currency,
#'     drop_cols = exponent & currency_name)
#'
#' # Show the node attribute names
#' # for the graph; note that the
#' # `exponent` and `currency_name`
#' # columns are not attributes in the
#' # graph's internal node data frame
#' graph_3 %>%
#'   get_node_df() %>%
#'   colnames()
#'
#' @family node creation and removal
#'
#' @export
add_nodes_from_table <- function(
    graph,
    table,
    label_col = NULL,
    type_col = NULL,
    set_type = NULL,
    drop_cols = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)
  # TODO use new technique to convert to string.
  # Get the requested `label_col`
  label_col <-
    rlang::enquo(label_col) %>% rlang::get_expr() %>% as.character()

  # Get the requested `type_col`
  type_col <-
    rlang::enquo(type_col) %>% rlang::get_expr() %>% as.character()

  # Get the requested `set_type`
  set_type <-
    rlang::enquo(set_type) %>% rlang::get_expr() %>% as.character()

  # Get the requested `drop_cols`
  drop_cols <-
    rlang::enquo(drop_cols) %>% rlang::get_expr() %>% as.character()

  if (length(label_col) == 0) {
    label_col <- NULL
  }

  if (length(type_col) == 0) {
    type_col <- NULL
  }

  if (length(set_type) == 0) {
    set_type <- NULL
  }

  if (length(drop_cols) == 0) {
    drop_cols <- NULL
  } else {
    drop_cols <- drop_cols[drop_cols != "&"]
    drop_cols <- paste(drop_cols, collapse = " & ")
  }

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  if (inherits(table, "character")) {

    # Load in CSV file
    csv <- utils::read.csv(table, stringsAsFactors = FALSE)

  } else if (inherits(table, "data.frame")) {

    # Rename `table` object as `csv`
    csv <- as.data.frame(table, stringsAsFactors = FALSE)
  }

  # Get number of rows in the CSV
  rows_in_csv <- nrow(csv)

  # Create an empty ndf and bind those columns
  # with the table data
  ndf <- create_node_df(n = rows_in_csv)

  # Optionally set the `label` attribute from a
  # specified column in the CSV (this copies data into
  # the `label` column)
  if (!is.null(label_col)) {

    colnames(csv)[which(colnames(csv) == label_col)] <- "label"

    ndf$label <- as.character(csv[, which(colnames(csv) == "label")])
  }

  # Optionally set the `type` attribute from a
  # specified column in the CSV (this copies data into
  # the `type` column)
  if (!is.null(type_col)) {

    colnames(csv)[which(colnames(csv) == type_col)] <- "type"

    ndf$type <- as.character(csv[, which(colnames(csv) == "type")])
  }

  # If an `id` column exists in the external
  # table, copy those values in as `id_external`
  if ("id" %in% colnames(csv)) {

    colnames(csv)[which(colnames(csv) == "id")] <- "id_external"

    ndf$id_external <- csv$id_external
  }

  # Optionally set the `type` attribute with a single
  # value repeated down
  if (is.null(type_col) && !is.null(set_type)) {
    ndf$type <- as.character(set_type)
  }

  # Get the remaining columns from `csv`
  # to add to the ndf
  columns_to_add <-
    base::setdiff(colnames(csv), colnames(ndf))

  # If values for `drop_cols` provided,
  # further filter the list of column data
  # to migrate to `ndf`
  if (!is.null(drop_cols)) {

    col_selection <- get_col_selection(col_selection_stmt = drop_cols)

    if (col_selection[["selection_type"]] == "column_range") {
      col_index_1 <- which(colnames(csv) == col_selection[["column_selection"]][1])
      col_index_2 <- which(colnames(csv) == col_selection[["column_selection"]][2])

      col_indices <- col_index_1:col_index_2 %>% sort()

      columns_to_add <- base::setdiff(columns_to_add, colnames(csv)[col_indices])

    } else if (col_selection[["selection_type"]] == "column_index_range") {

      col_indices <- col_selection[["column_selection"]] %>% sort()

      columns_to_add <- base::setdiff(columns_to_add, colnames(csv)[col_indices])

    } else if (col_selection[["selection_type"]] %in% c("single_column_name", "column_names")) {

      columns_to_add <- base::setdiff(columns_to_add, col_selection[["column_selection"]])

    } else if (length(col_selection) == 0) {

      columns_to_add <- columns_to_add
    }
  }

  # Move any additional columns from the
  # external table to `ndf`
  ndf <-
    dplyr::bind_cols(
      ndf,
      dplyr::select(csv, dplyr::all_of(columns_to_add))
      )

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph %>% count_nodes()

  # Add node data frame `ndf` to the graph
  graph <- add_node_df(graph, ndf)

  # Redact the signing of the action to the log
  graph$graph_log <-
    graph$graph_log[-nrow(graph$graph_log), ]

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph %>% count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Update the `last_node` counter
  graph$last_node <- nodes_created + nodes_added

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      graph %>%
      trigger_graph_actions()
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Rescale numeric node attribute values
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, take a set of numeric values for a
#' node attribute, rescale to a new numeric or color range, then write to the
#' same node attribute or to a new node attribute column.
#'
#' @inheritParams render_graph
#' @param node_attr_from The node attribute containing numeric data that is to
#'   be rescaled to new numeric or color values.
#' @param to_lower_bound The lower bound value for the set of rescaled values.
#'   This can be a numeric value or an X11 color name.
#' @param to_upper_bound The upper bound value for the set of rescaled values.
#'   This can be a numeric value or an X11 color name.
#' @param node_attr_to An optional name of a new node attribute to which the
#'   recoded values will be applied. This will retain the original node
#'   attribute and its values.
#' @param from_lower_bound An optional, manually set lower bound value for the
#'   rescaled values. If not set, the minimum value from the set will be used.
#' @param from_upper_bound An optional, manually set upper bound value for the
#'   rescaled values. If not set, the minimum value from the set will be used.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 5,
#'     m = 10,
#'     set_seed = 23) |>
#'   set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = 5,
#'       mean = 5,
#'       sd = 1) |> round(1))
#'
#' # Get the graph's internal ndf
#' # to show which node attributes
#' # are available
#' graph |> get_node_df()
#'
#' # Rescale the `value` node
#' # attribute, so that its values
#' # are rescaled between 0 and 1
#' graph <-
#'   graph |>
#'   rescale_node_attrs(
#'     node_attr_from = value,
#'     to_lower_bound = 0,
#'     to_upper_bound = 1)
#'
#' # Get the graph's internal ndf
#' # to show that the node attribute
#' # values had been rescaled
#' graph |> get_node_df()
#'
#' # Scale the values in the `value`
#' # node attribute to different
#' # shades of gray for the `fillcolor`
#' # and `fontcolor` node attributes
#' graph <-
#'   graph |>
#'   rescale_node_attrs(
#'     node_attr_from = value,
#'     to_lower_bound = "gray80",
#'     to_upper_bound = "gray20",
#'     node_attr_to = fillcolor) |>
#'   rescale_node_attrs(
#'     node_attr_from = value,
#'     to_lower_bound = "gray5",
#'     to_upper_bound = "gray95",
#'     node_attr_to = fontcolor)
#'
#' # Get the graph's internal ndf
#' # once more to show that scaled
#' # grayscale colors are now available
#' # in the `fillcolor` and `fontcolor`
#' # node attributes
#' graph |> get_node_df()
#'
#' @family node creation and removal
#'
#' @export
rescale_node_attrs <- function(
    graph,
    node_attr_from,
    to_lower_bound = 0,
    to_upper_bound = 1,
    node_attr_to = NULL,
    from_lower_bound = NULL,
    from_upper_bound = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get the requested `node_attr_from`
  node_attr_from <-
    rlang::enquo(node_attr_from) |> rlang::get_expr() |> as.character()

  # Get the requested `node_attr_to`
  node_attr_to <-
    rlang::enquo(node_attr_to) |> rlang::get_expr() |> as.character()

  if (length(node_attr_to) == 0) {
    node_attr_to <- NULL
  }

  # Extract the graph's ndf
  nodes <- get_node_df(graph)

  # Get column names from the graph's ndf
  column_names_graph <- colnames(nodes)

  # Stop function if `node_attr_from` is not one
  # of the graph's node attributes
  if (!any(column_names_graph %in% node_attr_from)) {

    cli::cli_abort(
      "The node attribute to rescale is not in the ndf.")
  }

  # Extract the vector to rescale from the `nodes` df
  vector_to_rescale <-
    nodes |>
    dplyr::mutate(
      dplyr::across(
        dplyr::all_of(node_attr_from),
        .fns = as.numeric)
      ) |>
    dplyr::pull(var = !!node_attr_from)

  # TODO condition could be simplified to
  # is.null(from_lower) || (!is.null(from_lower) && is.null(from_upper))?
  if ((!is.null(from_lower_bound) &&
       is.null(from_upper_bound)) ||
      (is.null(from_lower_bound) &&
       !is.null(from_upper_bound)) ||
      (is.null(from_lower_bound) &&
       is.null(from_upper_bound))) {

    from <- range(vector_to_rescale, na.rm = TRUE, finite = TRUE)

  } else {
    from <- c(from_lower_bound, from_upper_bound)
  }

  # Get vector of rescaled, numeric node
  # attribute values
  if (is.numeric(to_lower_bound) &&
      is.numeric(to_upper_bound)) {

    nodes_attr_vector_rescaled <-
      round(
        scales::rescale(
          x = vector_to_rescale,
          to = c(to_lower_bound,
                 to_upper_bound),
          from = from),
        3)
  }

  # Get vector of rescaled, node attribute color values
  if ((to_lower_bound %in% grDevices::colors()) &&
      (to_upper_bound %in% grDevices::colors())) {

    nodes_attr_vector_rescaled <-
      scales::cscale(
        x = vector_to_rescale,
        palette = scales::seq_gradient_pal(
          to_lower_bound,
          to_upper_bound))
  }

  # If a new node attribute name was not provided,
  # overwrite the source node attribute with the
  # rescaled values
  if (is.null(node_attr_to)) {
    node_attr_to <- node_attr_from
  }

  node_attr_to_2 <- rlang::enquo(node_attr_to)

  # Set the node attribute values for nodes specified
  # in selection
  graph <-
    set_node_attrs(
      graph = graph,
      node_attr = !!node_attr_to_2,
      values = nodes_attr_vector_rescaled
    )

  # Remove last action from the `graph_log`
  graph$graph_log <- graph$graph_log[1:(nrow(graph$graph_log) - 1), ]

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Delete vectors cached in a graph object
#'
#' @description
#'
#' Delete vectors cached in a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#' @param name One or more name of vector objects to delete from the cache. If
#'   none supplied, all cached vectors available in the graph will be deleted.
#'
#' @return A vector.
#'
#' @examples
#' # Create an empty graph
#' graph <-
#'   create_graph()
#'
#' # Cache 3 different vectors inside
#' # the graph object
#' graph <-
#'   graph %>%
#'   set_cache(
#'     name = "a",
#'     to_cache = 1:4) %>%
#'   set_cache(
#'     name = "b",
#'     to_cache = 5:9) %>%
#'   set_cache(
#'     name = "c",
#'     to_cache = 10:14)
#'
#' # Delete cache `b`
#' graph <-
#'   graph %>%
#'   delete_cache(name = "b")
#'
#' # Delete remaining cached vectors
#' graph <-
#'   graph %>%
#'   delete_cache()
#'
#' @export
delete_cache <- function(
    graph,
    name = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If there are no cached vectors available,
  # return the graph unchanged with a warning
  if (length(graph$cache) == 0) {

    cli::cli_warn(
        "There are no cached vectors, so, the graph is unchanged.",
        call = current_env()
      )

    return(graph)
  }

  if (length(graph$cache) > 0) {

    if (is.null(name)) {

      graph$cache <- list()

    } else {

      if (name %in% names(graph$cache)) {

        graph$cache[name] <- NULL

      } else {

        cli::cli_warn(c(
          "The supplied `name` (`{name}`) does not match the name of a cached vector.",
          "So, the graph is unchanged."
        ), call = current_env())
        return(graph)
      }
    }
  }

  # Get current function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Deselect any selected nodes in a graph
#'
#' @description
#'
#' Deselect nodes in a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#' @param nodes A vector of node IDs that should be deselected.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = c("a", "a", "z", "z"),
#'     label = TRUE,
#'     value = c(3.5, 2.6, 9.4, 2.7))
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = c("a", "z", "a"))
#'
#' # Create a graph with the ndf and edf
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Explicitly select nodes `1` and `3`
#' graph <-
#'   graph |>
#'   select_nodes(nodes = c(1, 3)) |>
#'   deselect_nodes(nodes = 1)
#'
#' # Verify that the node selection
#' # has been made for nodes `1` and
#' # `3` and that node `1` has been
#' # deselected (leaving only `3`)
#' graph |> get_selection()
#'
#' @export
deselect_nodes <- function(
    graph,
    nodes
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If no node selection available, return graph
  if (!graph_contains_node_selection(graph)) {
    return(graph)
  }

  # Extract the graph's node selection
  graph$node_selection <-
    graph$node_selection |>
    dplyr::filter(!(node %in% nodes))

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get the mean distance
#'
#' @description
#'
#' Get the mean distance of a graph, which is the average path length in the
#' graph. This operates through calculation of the shortest paths between all
#' pairs of nodes.
#'
#' @inheritParams render_graph
#'
#' @return A single numeric value representing the mean distance of the graph.
#'
#' @examples
#' # Create a cycle graph
#' graph <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' # Determine the mean distance
#' graph %>%
#'   get_mean_distance()
#'
#' # Create a full graph and then
#' # get the mean distance value
#' create_graph() %>%
#'   add_full_graph(n = 10) %>%
#'   get_mean_distance()
#'
#' @export
get_mean_distance <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the maximum eccentricity of the graph
  igraph::mean_distance(ig_graph)
}

#' Get detailed information on nodes
#'
#' @description
#'
#' Obtain a data frame with detailed information on nodes and their
#' interrelationships within the graph.
#'
#' @param graph A graph object of class `dgr_graph`.
#'
#' @return A data frame containing information specific to each node within the
#'   graph.
#'
#' @examples
#' # Create a simple graph
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 5, m = 10,
#'     set_seed = 23)
#'
#' # Get information on the graph's nodes
#' graph %>% get_node_info()
#'
#' @export
get_node_info <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If graph is empty, return NULL
  if (is_graph_empty(graph)) {
    return(NULL)
  }

  # Get vectors of nodes in edges and
  # node `type` values
  edge_from <- graph$edges_df$from
  edge_to <- graph$edges_df$to
  type <- graph$nodes_df$type

  # Get vector of all node IDs and all labels
  all_nodes <- graph$nodes_df$id
  labels <- graph$nodes_df$label

  # For graphs with no edges, create a
  # `node_properties` data frame that doesn't
  # need to consider any edge information
  if (nrow(graph$edges_df) == 0) {

    node_properties <-
      as.data.frame(
        mat.or.vec(
          nr = length(all_nodes),
          nc = 7),
        stringsAsFactors = FALSE)

    colnames(node_properties) <-
      c("id", "type", "label", "deg",
        "indeg", "outdeg", "loops")

    node_properties[, 1] <- graph$nodes_df[, 1]
    node_properties[, 2] <- graph$nodes_df[, 2]
    node_properties[, 3] <- graph$nodes_df[, 3]

    # Ensure that the `id` column is an integer
    node_properties <-
      dplyr::mutate(node_properties, id = as.integer(id))

    # Arrange the table by `id` ascending
    node_properties <-
      dplyr::arrange(node_properties, id)

  } else if (!is.null(graph$edges_df)) {

    # Get vector of the top-level nodes
    top_nodes <-
      unique(
        edge_from[which(!(edge_from %in%
                            edge_to))])

    # Get vector of the bottom-level nodes
    bottom_nodes <-
      unique(
        edge_to[which(!(edge_to %in%
                          edge_from))])

    # Get vector of all nodes neither at the top nor
    # the bottom level
    between_nodes <-
      all_nodes[which(!(all_nodes %in%
                          c(top_nodes,
                            bottom_nodes)))]

    # Place the nodes in order
    ordered_nodes <-
      c(top_nodes, between_nodes, bottom_nodes)

    # Create data frame of node properties
    for (i in seq(ordered_nodes)) {
      if (i == 1) {
        node_properties <-
          as.data.frame(
            mat.or.vec(nr = 0,
                       nc = 7),
            stringsAsFactors = FALSE)

        colnames(node_properties) <-
          c("id", "type", "label", "deg",
            "indeg", "outdeg", "loops")
      }

      # Get degree for each node
      degree <-
        sum(c(graph$edges_df$from,
              graph$edges_df$to) %in%
              ordered_nodes[i])

      # Get indegree for each node
      if (ordered_nodes[i] %in% top_nodes ||
          degree == 0) {
        indegree <- 0
      }

      if (!(ordered_nodes[i] %in% top_nodes) &&
          degree != 0) {
        for (j in 1:sum(edge_to %in%
                        ordered_nodes[i])) {
          if (j == 1) {
            indegree <- vector(mode = "character")
          }
          indegree <-
            c(indegree,
              edge_from[which(edge_to %in%
                                ordered_nodes[i])[j]])
        }
        indegree <- length(indegree)
      }

      # Get outdegree for each node
      if (ordered_nodes[i] %in% bottom_nodes ||
          degree == 0) {
        outdegree <- 0
      }

      if (!(ordered_nodes[i] %in% bottom_nodes) &&
          degree != 0) {
        for (j in 1:sum(edge_from %in%
                        ordered_nodes[i])) {
          if (j == 1) {
            outdegree <- vector(mode = "character")
          }
          outdegree <-
            c(outdegree,
              edge_from[which(edge_from %in%
                                ordered_nodes[i])[j]])
        }
        outdegree <- length(outdegree)
      }

      # Get number of loops for each node
      loops <-
        sum(graph$edges_df$from == graph$edges_df$to &
              graph$edges_df$to == ordered_nodes[i])

      # Collect information into `node_properties`
      node_properties[i, 1] <-
        ordered_nodes[i]

      node_properties[i, 2] <-
        ifelse(exists("type"),
               type[which(all_nodes %in%
                            ordered_nodes[i])],
               rep(NA, length(ordered_nodes)))

      node_properties[i, 3] <-
        ifelse(!is.null(
          labels[which(all_nodes %in%
                         ordered_nodes[i])]),
          labels[which(all_nodes %in%
                         ordered_nodes[i])],
          NA)

      node_properties[i, 4] <- degree
      node_properties[i, 5] <- indegree
      node_properties[i, 6] <- outdegree
      node_properties[i, 7] <- loops
    }

    # Ensure that the `id` column is an integer
    node_properties <-
      dplyr::mutate(
        node_properties, id = as.integer(id))

    # Arrange the table by `id` ascending
    node_properties <-
      dplyr::arrange(
        node_properties, id)
  }

  node_properties
}

#' Get information on a graph series
#'
#' @description
#'
#' Obtain a data frame with information on the graphs within a graph series.
#'
#' @param graph_series A graph series object of type `dgr_graph_1D`.
#'
#' @return A data frame containing information on the graphs within the supplied
#'   graph series.
#'
#' @examples
#' # Create three graphs
#' graph_1 <-
#'   create_graph() |>
#'   add_path(n = 4)
#'
#' graph_2 <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' graph_3 <-
#'   create_graph() |>
#'   add_star(n = 6)
#'
#' # Create an empty graph series
#' # and add the graphs
#' series <-
#'   create_graph_series() |>
#'   add_graph_to_graph_series(
#'     graph = graph_1) |>
#'   add_graph_to_graph_series(
#'     graph = graph_2) |>
#'   add_graph_to_graph_series(
#'     graph = graph_3)
#'
#' # Get information on the graphs in the series
#' series |> get_graph_series_info()
#'
#' @export
get_graph_series_info <- function(graph_series) {

  graphs_in_series <-
    count_graphs_in_graph_series(graph_series)

  series_info_df <-
    dplyr::tibble(
      graph = integer(),
      name = character(),
      date_time = Sys.time(),
      tz = character(),
      nodes = integer(),
      edges = integer(),
      directed = logical()
    ) |>
    as.data.frame(stringsAsFactors = FALSE)

  if (graphs_in_series == 0) {
    return(series_info_df)
  }

  for (i in 1:graphs_in_series) {

    # Add a graph index number
    series_info_df[i, 1] <- i

    # Add the graph name
    series_info_df[i, 2] <-
      graph_series$graphs[[i]]$graph_info$graph_name

    # Add the graph time
    series_info_df[i, 3] <-
      as.POSIXct(
        graph_series$graphs[[i]]$graph_info$graph_time,
        tz = graph_series$graphs[[i]]$graph_info$graph_tz,
        origin = "1970-01-01")

    # Add the graph time zone
    series_info_df[i, 4] <-
      graph_series$graphs[[i]]$graph_info$graph_tz

    # Add the count of nodes in the graph
    series_info_df[i, 5] <-
      nrow(graph_series$graphs[[i]]$nodes_df)

    # Add the count of nodes in the graph
    series_info_df[i, 6] <-
      nrow(graph_series$graphs[[i]]$edges_df)

    # Add TRUE/FALSE value on whether graph is directed
    series_info_df[i, 7] <-
      is_graph_directed(graph_series$graphs[[i]])
  }

  series_info_df
}

#' Is the graph a simple graph?
#'
#' @description
#'
#' Determine whether the graph is a simple graph. A simple graph is one that
#' does not contain any loops nor any multiple edges.
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create a graph with 2 cycles
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 4) |>
#'   add_cycle(n = 3)
#'
#' # Check if the graph is simple
#' graph |> is_graph_simple()
#'
#' @export
is_graph_simple <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Determine whether the graph is
  # a simple graph
  igraph::is_simple(ig_graph)
}

#' R + mermaid.js
#'
#' @description
#'
#' Make diagrams in R using \href{https://github.com/mdaines/viz-js}{viz.js} or
#' \href{https://github.com/mermaid-js/mermaid}{mermaid.js} with infrastructure
#' provided by \href{http://www.htmlwidgets.org/}{htmlwidgets}.
#'
#' @param diagram The diagram in `graphviz` or `mermaid` format, or, a file (as
#'   a connection or file name) containing a diagram specification. The
#'   recommended filename extensions are `.gv` and `.mmd` for the Graphviz and
#'   the mermaid diagram specifications, respectively. If no diagram is provided
#'   (`diagram = ""`) then the function will assume that a diagram will be
#'   provided by [htmltools::tags()] and `DiagrammeR` is just being used for
#'   dependency injection.
#' @param type A string, either `mermaid` (default) or `grViz` indicating the
#'   type of diagram spec and the desired parser/renderer.
#' @param ... Any other parameters to pass to `grViz` or `mermaid`
#'
#' @return An object of class `htmlwidget` that will intelligently print itself
#'   into HTML in a variety of contexts including the R console, within R
#'   Markdown documents, and within Shiny output bindings.
#'
#' @examples
#' \dontrun{
#' # note the whitespace is not important
#' DiagrammeR("
#'   graph LR
#'     A-->B
#'     A-->C
#'     C-->E
#'     B-->D
#'     C-->D
#'     D-->F
#'     E-->F
#' ")
#'
#' DiagrammeR("
#'    graph TB
#'    A-->B
#'    A-->C
#'    C-->E
#'    B-->D
#'    C-->D
#'    D-->F
#'    E-->F
#' ")
#'
#' DiagrammeR("graph LR;A(Rounded)-->B[Squared];B-->C{A Decision};
#'  C-->D[Square One];C-->E[Square Two];
#'  style A fill:#E5E25F;  style B fill:#87AB51; style C fill:#3C8937;
#'  style D fill:#23772C;  style E fill:#B6E6E6;"
#' )
#'
#' # Load in the 'mtcars' dataset
#' mtcars
#' connections <- sapply(
#'  1:ncol(mtcars)
#'   ,function(i) {
#'      paste0(
#'         i
#'       ,"(",colnames(mtcars)[i],")---"
#'       ,i,"-stats("
#'       ,paste0(
#'         names(summary(mtcars[,i]))
#'         ,": "
#'         ,unname(summary(mtcars[,i]))
#'         ,collapse="<br/>"
#'       )
#'       ,")"
#'    )
#'  }
#' )
#'
#' DiagrammeR(
#'    paste0(
#'      "graph TD;", "\n",
#'      paste(connections, collapse = "\n"),"\n",
#'      "classDef column fill:#0001CC, stroke:#0D3FF3, stroke-width:1px;" ,"\n",
#'      "class ", paste0(1:length(connections), collapse = ","), " column;"
#'    )
#'  )
#'
#' # also with DiagrammeR() you can use tags from htmltools
#' # just make sure to use class = "mermaid"
#' library(htmltools)
#' diagramSpec = "
#' graph LR;
#'   id1(Start)-->id2(Stop);
#'   style id1 fill:#f9f,stroke:#333,stroke-width:4px;
#'   style id2 fill:#ccf,stroke:#f66,stroke-width:2px,stroke-dasharray: 5, 5;
#' "
#' html_print(tagList(
#'   tags$h1("R + mermaid.js = Something Special")
#'   ,tags$pre(diagramSpec)
#'   ,tags$div(class="mermaid",diagramSpec)
#'   ,DiagrammeR()
#' ))
#'
#' # sequence diagrams
#' # Using this "How to Draw a Sequence Diagram"
#' #   http://www.cs.uku.fi/research/publications/reports/A-2003-1/page91.pdf
#' # draw some sequence diagrams with DiagrammeR
#'
#' DiagrammeR("
#' sequenceDiagram;
#'    customer->>ticket seller: ask for ticket;
#'    ticket seller->>database: seats;
#'    alt tickets available
#'      database->>ticket seller: ok;
#'      ticket seller->>customer: confirm;
#'      customer->>ticket seller: ok;
#'      ticket seller->>database: book a seat;
#'      ticket seller->>printer: print ticket;
#'    else sold out
#'      database->>ticket seller: none left;
#'      ticket seller->>customer:  sorry;
#'    end
#' ")
#' }
#'
#' @export
DiagrammeR <- function(
    diagram = "",
    type = "mermaid",
    ...
) {

  # DiagrammeR will serve as a wrapper function for mermaid and grVis
  if (grepl(x = type, pattern = "[m,M](erm).*")) {

    mermaid(diagram, ...)

  } else if (grepl(x = type, pattern = "[g,G]?[r,R]?.*[v,V][i].*")) {
    grViz(diagram, ...)

  } else {
    abort("The type should be `mermaid` or `grViz`.")
  }
}

#' Widget output function for use in Shiny
#'
#' @param outputId Output variable to read from
#' @param width A valid CSS unit for the width or a number, which will be
#'   coerced to a string and have `px` appended.
#' @param height A valid CSS unit for the height or a number, which will be
#'   coerced to a string and have `px` appended.
#'
#' @export
DiagrammeROutput <- function(
    outputId,
    width = "100%",
    height = "auto"
) {

  htmlwidgets::shinyWidgetOutput(
    outputId,
    "DiagrammeR",
    width,
    height,
    package = "DiagrammeR"
  )
}

#' Widget render function for use in Shiny
#'
#' @param expr An expression that generates a DiagrammeR graph
#' @param env The environment in which to evaluate expr.
#' @param quoted Is expr a quoted expression (with quote())? This is useful if
#'   you want to save an expression in a variable.
#'
#' @export
renderDiagrammeR <- function(
    expr,
    env = parent.frame(),
    quoted = FALSE
) {

  if (!quoted) expr <- substitute(expr)

  htmlwidgets::shinyRenderWidget(
    expr,
    DiagrammeROutput,
    env,
    quoted = TRUE
  )
}

#' Add a fully connected graph
#'
#' @description
#'
#' With a graph object of class `dgr_graph`, add a fully connected graph either
#' with or without loops. If the graph object set as directed, the added graph
#' will have edges to and from each pair of nodes. In the undirected case, a
#' single edge will link each pair of nodes.
#'
#' @inheritParams node_edge_aes_data
#' @inheritParams render_graph
#' @param n The number of nodes comprising the fully connected graph.
#' @param type An optional string that describes the entity type for the nodes
#'   to be added.
#' @param label Either a vector object of length `n` that provides optional
#'   labels for the new nodes, or, a boolean value where setting to `TRUE`
#'   ascribes node IDs to the label and `FALSE` or `NULL` yields a blank label.
#' @param rel An optional string for providing a relationship label to all new
#'   edges created in the connected graph.
#' @param edge_wt_matrix An optional matrix of `n` by `n` dimensions containing
#'   values to apply as edge weights. If the matrix has row names or column
#'   names and `label = TRUE`, those row or column names will be used as node
#'   label values.
#' @param keep_loops An option to simplify the fully connected graph by removing
#'   loops (edges from and to the same node). The default value is `FALSE`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a new graph object
#' # and add a directed and fully
#' # connected graph with 3 nodes
#' # and edges to and from all
#' # pairs of nodes; with the option
#' # `keep_loops = TRUE` nodes
#' # will also have edges from
#' # and to themselves
#' graph <-
#'   create_graph() |>
#'   add_full_graph(
#'     n = 3, keep_loops = TRUE
#'   )
#'
#' # Get node information
#' # from this graph
#' graph |> get_node_info()
#'
#' # Using `keep_loops = FALSE`
#' # (the default) will remove
#' # the loops
#' create_graph() |>
#'   add_full_graph(n = 3) |>
#'   get_node_info()
#'
#' # Values can be set for
#' # the node `label`, node
#' # `type`, and edge `rel`
#' graph <-
#'   create_graph() |>
#'   add_full_graph(
#'     n = 3,
#'     type = "connected",
#'     label = c("1st", "2nd", "3rd"),
#'     rel = "connected_to"
#'   )
#'
#' # Show the graph's node
#' # data frame (ndf)
#' graph |> get_node_df()
#'
#' # Show the graph's edge
#' # data frame (edf)
#' graph |> get_edge_df()
#'
#' # Create a fully-connected and
#' # directed graph with 3 nodes,
#' # and, where a matrix provides
#' # edge weights; first, create the
#' # matrix (with row names to be
#' # used as node labels)
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' edge_wt_matrix <-
#'   rnorm(100, 5, 2) |>
#'   sample(9, FALSE) |>
#'   round(2) |>
#'   matrix(
#'     ncol = 3,
#'     nrow = 3,
#'     dimnames = list(c("a", "b", "c"))
#'   )
#'
#' # Create the fully-connected
#' # graph (without loops however)
#' graph <-
#'   create_graph() |>
#'   add_full_graph(
#'     n = 3,
#'     type = "weighted",
#'     label = TRUE,
#'     rel = "related_to",
#'     edge_wt_matrix = edge_wt_matrix,
#'     keep_loops = FALSE
#'   )
#'
#' # Show the graph's node
#' # data frame (ndf)
#' graph |> get_node_df()
#'
#' # Show the graph's edge
#' # data frame (edf)
#' graph |> get_edge_df()
#'
#' # An undirected graph can
#' # also use a matrix with
#' # edge weights, but only
#' # the lower triangle of
#' # that matrix will be used
#' create_graph(directed = FALSE) |>
#'   add_full_graph(
#'     n = 3,
#'     type = "weighted",
#'     label = TRUE,
#'     rel = "related_to",
#'     edge_wt_matrix = edge_wt_matrix,
#'     keep_loops = FALSE
#'   ) |>
#'   get_edge_df()
#'
#' @export
add_full_graph <- function(
    graph,
    n,
    type = NULL,
    label = TRUE,
    rel = NULL,
    edge_wt_matrix = NULL,
    keep_loops = FALSE,
    node_aes = NULL,
    edge_aes = NULL,
    node_data = NULL,
    edge_data = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  # Get the number of edges ever created for
  # this graph
  edges_created <- graph$last_edge

  # Get the graph's global attributes
  global_attrs <- graph$global_attrs

  # Get the graph's log
  graph_log <- graph$graph_log

  # Get the graph's info
  graph_info <- graph$graph_info

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph |> count_nodes()

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Create initial adjacency matrix
  adj_matrix <- matrix(1, nrow = n, ncol = n)

  # Remove loops by making the diagonal of the
  # adjacency matrix all 0
  if (!keep_loops) {
    adj_matrix <-
      adj_matrix -
      diag(1, nrow = nrow(adj_matrix), ncol = ncol(adj_matrix))
  }

  if (is_graph_directed(graph)) {

    # Create a new directed graph based on the
    # adjacency matrix `adj_matrix`
    new_graph <-
      from_adj_matrix(adj_matrix, mode = "directed")

    # If a matrix of edge weights provided, apply those
    # to each of the edges in a row-major fashion
    if (!is.null(edge_wt_matrix)) {

      new_graph <-
        set_edge_attrs(
          new_graph,
          edge_attr = "weight",
          values = as.numeric(edge_wt_matrix)[
            which(as.numeric(adj_matrix) == 1)
          ]
        )
    }
  }

  if (!is_graph_directed(graph)) {

    new_graph <-
      from_adj_matrix(
        adj_matrix,
        mode = "undirected")

    # If a matrix of edge weights provided, apply those
    # from the bottom triangle to each of the edges in a
    # row-major fashion
    if (!is.null(edge_wt_matrix)) {

      new_graph <-
        set_edge_attrs(
          new_graph,
          edge_attr = "weight",
          values = edge_wt_matrix[
            lower.tri(
              edge_wt_matrix,
              diag = keep_loops # TRUE or FALSE
            )])
    }
  }

  # Add label values to nodes
  if (length(label) == 1) {
    if (label) {
      new_graph$nodes_df[, 3] <- new_graph$nodes_df[, 1]
    } else {
      new_graph$nodes_df[, 3] <-
        as.character(new_graph$nodes_df[, 1])
    }
  }

  if (n > 1 && length(label) == n) {
    new_graph$nodes_df[, 3] <- label
  }

  if (isTRUE(label) && !is.null(edge_wt_matrix)) {

    if (!is.null(colnames(edge_wt_matrix))) {
      ewm_names <- colnames(edge_wt_matrix)
    }

    if (!is.null(rownames(edge_wt_matrix))) {
      ewm_names <- rownames(edge_wt_matrix)
    }

    if (length(ewm_names) == n) {
      new_graph$nodes_df[, 3] <- ewm_names
    }
  }

  # Add `type` values to all new nodes
  if (length(type) == 1) {
    new_graph$nodes_df[, 2] <- type
  }

  # Add `rel` values to all new edges
  # NULL is length 0
  if (length(rel) == 1) {
    new_graph$edges_df[, 4] <- rel
  }

  # Collect node aesthetic attributes
  if (!is.null(node_aes)) {

    node_aes_tbl <- dplyr::as_tibble(node_aes)

    if (nrow(node_aes_tbl) < nrow(new_graph$nodes_df)) {

      node_aes$index__ <- seq_len(nrow(new_graph$nodes_df))

      node_aes_tbl <-
        dplyr::as_tibble(node_aes) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_aes_tbl)) {
      node_aes_tbl$id <- NULL
    }
  }

  # Collect node data attributes
  if (!is.null(node_data)) {

    node_data_tbl <- dplyr::as_tibble(node_data)

    if (nrow(node_data_tbl) < nrow(new_graph$nodes_df)) {

      node_data$index__ <- seq_len(nrow(new_graph$nodes_df))

      node_data_tbl <-
        dplyr::as_tibble(node_data) |>
        dplyr::select(-index__)
    }

    if ("id" %in% colnames(node_data_tbl)) {
      node_data_tbl$id <- NULL
    }
  }

  # Collect edge aesthetic attributes
  if (!is.null(edge_aes)) {

    edge_aes_tbl <- dplyr::as_tibble(edge_aes)

    if (nrow(edge_aes_tbl) < nrow(new_graph$edges_df)) {

      edge_aes$index__ <- seq_len(nrow(new_graph$edges_df))

      edge_aes_tbl <-
        dplyr::as_tibble(edge_aes) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_aes_tbl)) {
      edge_aes_tbl$id <- NULL
    }
  }

  # Collect edge data attributes
  if (!is.null(edge_data)) {

    edge_data_tbl <- dplyr::as_tibble(edge_data)

    if (nrow(edge_data_tbl) < nrow(new_graph$edges_df)) {

      edge_data$index__ <- seq_len(nrow(new_graph$edges_df))

      edge_data_tbl <-
        dplyr::as_tibble(edge_data) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_data_tbl)) {
      edge_data_tbl$id <- NULL
    }
  }

  # Add node aesthetics if available
  if (exists("node_aes_tbl")) {

    new_graph$nodes_df <-
      new_graph$nodes_df |>
      dplyr::bind_cols(node_aes_tbl)
  }

  # Add node data if available
  if (exists("node_data_tbl")) {

    new_graph$nodes_df <-
      new_graph$nodes_df |>
      dplyr::bind_cols(node_data_tbl)
  }

  # Add edge aesthetics if available
  if (exists("edge_aes_tbl")) {

    new_graph$edges_df <-
      new_graph$edges_df |>
      dplyr::bind_cols(edge_aes_tbl)
  }

  # Add edge data if available
  if (exists("edge_data_tbl")) {

    new_graph$edges_df <-
      new_graph$edges_df |>
      dplyr::bind_cols(edge_data_tbl)
  }

  # If the input graph is not empty, combine graphs
  # using the `combine_graphs()` function
  if (!is_graph_empty(graph)) {

    combined_graph <- combine_graphs(graph, new_graph)

    # Update the `last_node` counter
    combined_graph$last_node <- nodes_created + n

    # Get the updated number of nodes in the graph
    nodes_graph_2 <- combined_graph |> count_nodes()

    # Get the number of nodes added to
    # the graph
    nodes_added <- nodes_graph_2 - nodes_graph_1

    # Get the updated number of edges in the graph
    edges_graph_2 <- combined_graph |> count_edges()

    # Get the number of edges added to
    # the graph
    edges_added <- edges_graph_2 - edges_graph_1

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph_log <-
      add_action_to_log(
        graph_log = graph_log,
        version_id = nrow(graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(combined_graph$nodes_df),
        edges = nrow(combined_graph$edges_df),
        d_n = nodes_added,
        d_e = edges_added)

    combined_graph$global_attrs <- global_attrs
    combined_graph$graph_log <- graph_log
    combined_graph$graph_info <- graph_info

    # Write graph backup if the option is set
    if (combined_graph$graph_info$write_backups) {
      save_graph_as_rds(graph = combined_graph)
    }

    return(combined_graph)
  } else {

    # Get the updated number of nodes in the graph
    nodes_graph_2 <- new_graph |> count_nodes()

    # Get the number of nodes added to
    # the graph
    nodes_added <- nodes_graph_2 - nodes_graph_1

    # Get the updated number of edges in the graph
    edges_graph_2 <- new_graph |> count_edges()

    # Get the number of edges added to
    # the graph
    edges_added <- edges_graph_2 - edges_graph_1

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph_log <-
      add_action_to_log(
        graph_log = graph_log,
        version_id = nrow(graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(new_graph$nodes_df),
        edges = nrow(new_graph$edges_df),
        d_n = nodes_added,
        d_e = edges_added)

    new_graph$global_attrs <- global_attrs
    new_graph$graph_log <- graph_log
    new_graph$graph_info <- graph_info

    # Perform graph actions, if any are available
    if (nrow(graph$graph_actions) > 0) {
      graph <-
        trigger_graph_actions(graph)
    }

    # Write graph backup if the option is set
    if (new_graph$graph_info$write_backups) {
      save_graph_as_rds(graph = new_graph)
    }

    return(new_graph)
  }
}

#' Get the authority scores for all nodes
#'
#' @description
#'
#' Get the Kleinberg authority centrality scores for all nodes in the graph.
#'
#' @inheritParams render_graph
#' @param weights_attr an optional name of the edge attribute to use in the
#' adjacency matrix. If `NULL` then, if it exists, the `weight` edge
#' attribute of the graph will be used.
#'
#' @return a data frame with authority scores for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the authority centrality scores
#' # for all nodes in the graph
#' graph |>
#'   get_authority_centrality()
#'
#' # Add the authority centrality
#' # scores to the graph as a node
#' # attribute
#' graph <-
#'   graph |>
#'   join_node_attrs(
#'     df = get_authority_centrality(graph))
#'
#' # Display the graph's node data frame
#' graph |> get_node_df()
#'
#' @export
get_authority_centrality <- function(
    graph,
    weights_attr = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # if weights_attr is not NULL and character
  if (inherits(weights_attr, "character")) {

    # Stop function if the edge attribute does not exist
    if (!(weights_attr %in% colnames(graph$edges_df))) {

      cli::cli_abort(
        "The edge attribute to be used as weights must exist in the graph.")
    }

    # Stop function if the edge attribute is not numeric
    if (!is.numeric(graph$edges_df[, which(colnames(graph$edges_df) == weights_attr)])) {

      cli::cli_abort(
        "The edge attribute to be used as weights must be numeric.")
    }

    weights_attr <- graph$edges_df[, which(colnames(graph$edges_df) == weights_attr)]
  }

  # Get the authority centrality values for
  # each of the graph's nodes
  if (igraph::igraph_version() >= "2.1.0" &&
      "hits_scores" %in% getNamespaceExports("igraph")) {
    hits_scores_fn <- utils::getFromNamespace("hits_scores", "igraph")
    authority_centrality_values <-
      hits_scores_fn(
        graph = ig_graph,
        weights = weights_attr)

    authority_vector <- authority_centrality_values$authority
  } else {
    authority_centrality_values <-
      igraph::authority_score(
        graph = ig_graph,
        weights = weights_attr)

    authority_vector <- authority_centrality_values$vector
  }

  # Create df with authority centrality values
  data.frame(
    id = authority_vector |>
      names() |>
      as.integer(),
    authority_centrality = unname(authority_vector),
    stringsAsFactors = FALSE)
}

#' Render the graph in various formats
#'
#' @description
#'
#' Using a `dgr_graph` object, render the graph in the RStudio Viewer.
#'
#' @param graph A graph object of class `dgr_graph`.
#' @param layout A string specifying a layout type to use for node placement in
#'   this rendering. Possible layouts include: `nicely`, `circle`, `tree`, `kk`,
#'   and `fr`.
#' @param output A string specifying the output type; `graph` (the default)
#'   renders the graph using the [grViz()] function and `visNetwork` renders the
#'   graph using the [visnetwork()] function.
#' @param as_svg An option to render the graph as an SVG document.
#' @param title An optional title for a graph when using `output = "graph"`.
#' @param width An optional parameter for specifying the width of the resulting
#'   graphic in pixels.
#' @param height An optional parameter for specifying the height of the
#'   resulting graphic in pixels.
#' @examples
#' if (interactive()) {
#'
#'   # Render a graph that's a
#'   # balanced tree
#'   create_graph() |>
#'     add_balanced_tree(
#'       k = 2, h = 3
#'     ) |>
#'     render_graph()
#'
#'   # Use the `tree` layout for
#'   # better node placement in this
#'   # hierarchical graph
#'   create_graph() |>
#'     add_balanced_tree(
#'       k = 2, h = 3
#'     ) |>
#'     render_graph(layout = "tree")
#'
#'   # Plot the same tree graph but
#'   # don't show the node ID values
#'   create_graph() |>
#'     add_balanced_tree(
#'       k = 2, h = 3
#'     ) |>
#'     set_node_attr_to_display() |>
#'     render_graph(layout = "tree")
#'
#'   # Create a circle graph
#'   create_graph() |>
#'     add_gnm_graph(
#'       n = 55,
#'       m = 75,
#'       set_seed = 23
#'     ) |>
#'     render_graph(
#'       layout = "circle"
#'     )
#'
#'   # Render the graph using the
#'   # `visNetwork` output option
#'   create_graph() |>
#'     add_balanced_tree(
#'       k = 2, h = 3
#'     ) |>
#'     render_graph(
#'       output = "visNetwork"
#'     )
#' }
#'
#' @family Display and Save
#'
#' @export
render_graph <- function(
    graph,
    layout = NULL,
    output = NULL,
    as_svg = FALSE,
    title = NULL,
    width = NULL,
    height = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  output <- output %||% "graph"

  # Not allowing partial matching.
  output <- rlang::arg_match0(output, c("graph", "visNetwork"))
  # Check layout is a single string or NULL
  check_string(layout, allow_null = TRUE)

  if (!is.null(layout)) {
    rlang::arg_match0(layout, c("circle", "tree", "kk", "fr", "nicely", "neato"))
  }

  # Return early if output is visNetwork.
  if (output == "visNetwork") {
    return(visnetwork(graph))
  }

  # output = "graph" code with out = grViz

  # Add title as attribute
  if (!is.null(title)) {

    graph <- add_global_graph_attrs(graph, "label", title, "graph")
    graph <- add_global_graph_attrs(graph, "labelloc", "t", "graph")
    graph <- add_global_graph_attrs(graph, "labeljust", "c", "graph")
    graph <- add_global_graph_attrs(graph, "fontname", "Helvetica", "graph")
    graph <- add_global_graph_attrs(graph, "fontcolor", "gray30", "graph")
  }

  # If no fillcolor provided, use default; if no default available, use white
  if (nrow(graph$nodes_df) > 0 &&
      !("fillcolor" %in% colnames(graph$nodes_df))) {

    if ("fillcolor" %in% graph$global_attrs$attr) {

      graph$nodes_df$fillcolor <-
        graph$global_attrs |>
        dplyr::filter(attr == "fillcolor", attr_type == "node") |>
        dplyr::pull("value") |>
        as.character()
    } else {
      graph$nodes_df$fillcolor <- "white"
    }
  }

  # If fillcolor is available and there are NA values,
  # replace NAs with default color if available
  if (nrow(graph$nodes_df) > 0 &&
      rlang::has_name(graph$nodes_df, "fillcolor") &&
      "fillcolor" %in% graph$global_attrs$attr) {

    graph$nodes_df$fillcolor[which(is.na(graph$nodes_df$fillcolor))] <-
      graph$global_attrs[which(graph$global_attrs$attr == "fillcolor"), 2]
  }

  # Translate X11 colors to hexadecimal colors
  if ("fillcolor" %in% colnames(graph$nodes_df)) {

    graph$nodes_df <-
      graph$nodes_df |>
      dplyr::left_join(
        x11_hex() |>
          dplyr::as_tibble() |>
          dplyr::mutate(hex = toupper(hex)),
        by = c("fillcolor" = "x11_name")
      ) |>
      dplyr::mutate(
        fillcolor = dplyr::coalesce(hex, fillcolor),
        .keep = "unused"
      )
  }

  # Use adaptive font coloring for nodes that have a fill color
  if ("fillcolor" %in% colnames(graph$nodes_df) &&
      !"fontcolor" %in% colnames(graph$nodes_df)
  ) {

    graph$nodes_df$fontcolor <-
      tibble::tibble(value = graph$nodes_df$fillcolor) |>
      dplyr::mutate(value_x = contrasting_text_color(background_color = value)) |>
      dplyr::pull("value_x")
  }

  # Modify nodes df if a specific layout is requested.
  # and is one of the accepted values ("circle", "tree", "kk", "fr", "nicely")
  if (!is.null(layout)) {

    graph <-
      add_global_graph_attrs(
        graph,
        attr = "layout",
        value = "neato",
        attr_type = "graph"
      )

    # Remove existing x and y columns in nodes_df
    # to replace them with the layout coords
    if ("x" %in% colnames(graph$nodes_df)) {
      graph$nodes_df$x <- NULL
    }

    if ("y" %in% colnames(graph$nodes_df)) {
      graph$nodes_df$y <- NULL
    }

    # layout = "tree" is special, because layout_with_sugiyama creates two
    # different layouts
    if (layout == "tree") {
      m_coords <-
        to_igraph(graph) |>
        igraph::layout_with_sugiyama()
      m_coords <- m_coords[["layout"]]

      # Safety
      if (!is.matrix(m_coords) && nrow(m_coords) == 0) {

        cli::cli_abort("The tree coords should be a matrix", .internal = TRUE)
      }

      coords <- data.frame(
        x = m_coords[, 1],
        y = m_coords[, 2],
        stringsAsFactors = FALSE
      )
    } else {
      # Simple cases using defaults for kk, fr, nicely, and circle.
      fn_igraph <- switch(layout,
                          "kk" = igraph::layout_with_kk,
                          "fr" = igraph::layout_with_fr,
                          "nicely" = igraph::layout_nicely,
                          "circle" = igraph::layout_in_circle,
                          "neato" = igraph::layout_with_fr
      )

      m_coords <- graph |>
        to_igraph() |>
        fn_igraph()

      if (!is.matrix(m_coords)) {
        cli::cli_abort("The {.val {layout}} coords should be a matrix", .internal = TRUE)
      }

      coords <- data.frame(
        x = m_coords[, 1],
        y = m_coords[, 2],
        stringsAsFactors = FALSE
      )
    }

    # Corrections for layout = "circle"
    if (layout == "circle") {
      n_nodes <- count_nodes(graph)

      if (n_nodes == 0) {
        cli::cli_warn("No nodes exist? in the circle graph?")
        coords$x <- NaN
        coords$y <- NaN
      } else {
        # Previously as x * (count_nodes(graph) + (0.25 * count_nodes(graph))) / count_nodes(graph)
        # which can be simplified to x * 1.25? if n_nodes > 0
        # coords$x <- coords$x * (n_nodes + 0.25 * n_nodes) / n_nodes
        coords$x <- coords$x * 1.25
        coords$y <- coords$y * 1.25
      }

    }
    # Bind (x, y) coordinates to the graph's
    # internal NDF
    graph$nodes_df <-
      dplyr::bind_cols(graph$nodes_df, coords)
  }

  if (as_svg || any(c("image", "fa_icon") %in% colnames(get_node_df(graph)))) {

    if (as_svg && !rlang::is_installed("DiagrammeRsvg")) {
      rlang::inform("Use `as_svg = FALSE` if you don't want to install DiagrammeRsvg.")
    }

    # Stop function if `DiagrammeRsvg` package is not available
    rlang::check_installed("DiagrammeRsvg", "to render the graph to SVG.")

    # Generate DOT code
    dot_code <- generate_dot(graph)

    # create display to make return work when `as_svg = TRUE` #482
    display <- grViz(diagram = dot_code, width = width, height = height)

    # Get a vector of SVG lines
    svg_vec <-
      strsplit(DiagrammeRsvg::export_svg(
        grViz(diagram = dot_code)
      ), "\n") |>
      unlist()

    # Get a tibble of SVG data
    svg_tbl <- get_svg_tbl(svg_vec)

    svg_lines <-
      "<svg display=\"block\" margin=\"0 auto\" position=\"absolute\" width=\"100%\" height=\"100%\""

    svg_line_no <- svg_tbl |>
      dplyr::filter(type == "svg") |>
      dplyr::pull("index")

    # Modify <svg> attrs
    svg_vec[svg_line_no] <- svg_lines

    if ("image" %in% colnames(graph |> get_node_df())) {
      node_id_images <-
        graph |>
        get_node_df() |>
        dplyr::select("id", "image") |>
        dplyr::filter(nzchar(image)) |>
        dplyr::pull("id")

      filter_lines <-
        graph |>
        get_node_df() |>
        dplyr::select("id", "image") |>
        dplyr::filter(nzchar(image)) |>
        dplyr::mutate(filter_lines = as.character(glue::glue("<filter id=\"{id}\" x=\"0%\" y=\"0%\" width=\"100%\" height=\"100%\"><feImage xlink:href=\"{image}\"/></filter>"))) |>
        dplyr::pull("filter_lines") |>
        paste(collapse = "\n")

      filter_shape_refs <- as.character(glue::glue(" filter=\"url(#{node_id_images})\" "))

      svg_shape_nos <-
        svg_tbl |>
        dplyr::filter(node_id %in% node_id_images) |>
        dplyr::filter(type == "node_block") |>
        dplyr::pull("index")

      svg_shape_nos <- svg_shape_nos + 3
      svg_text_nos <- svg_shape_nos + 1

      # Modify shape lines
      for (i in seq(node_id_images)) {
        svg_vec[svg_shape_nos[i]] <-
          sub(" ", paste0(filter_shape_refs[i]), svg_vec[svg_shape_nos[i]])

        svg_vec[svg_text_nos[i]] <- ""
      }

      # Add in <filter> lines
      svg_vec[svg_line_no + 1] <-
        paste0(svg_vec[svg_line_no + 1], "\n\n", filter_lines, "\n")
    }

    # # Get the name of the function
    # if ("fa_icon" %in% colnames(graph |> get_node_df())) {
    #
    #   # Using a fontawesome icon requires the fontawesome package;
    #   # if it's not present, stop with a message
    #   if (requireNamespace("fontawesome", quietly = TRUE)) {
    #
    #     node_id_fa <-
    #       graph |>
    #       get_node_df() |>
    #       dplyr::select(id, fa_icon) |>
    #       dplyr::filter(fa_icon != "") |>
    #       dplyr::filter(!is.na(fa_icon)) |>
    #       dplyr::mutate(fa_uri = NA_character_)
    #
    #     node_id_svg <-
    #       node_id_fa |>
    #       dplyr::pull(id)
    #
    #     for (i in seq(nrow(node_id_fa))) {
    #
    #       random_name <- paste(sample(letters[1:10], 10), collapse = "")
    #       tmp_svg_file <- paste0(random_name, ".svg")
    #
    #       fa_icon <- node_id_fa[i, ]$fa_icon
    #       id <- node_id_fa[i, ]$id
    #
    #       writeLines(fontawesome::fa(fa_icon), tmp_svg_file)
    #
    #       svg_uri <- get_image_uri(tmp_svg_file)
    #
    #       file.remove(tmp_svg_file)
    #
    #       node_id_fa[i, "fa_uri"] <-
    #         as.character(glue::glue("<filter id=\"{id}\" x=\"0%\" y=\"0%\" width=\"100%\" height=\"100%\"><feImage xlink:href=\"{svg_uri}\"/></filter>"))
    #     }
    #
    #     filter_lines <-
    #       node_id_fa |>
    #       dplyr::pull(fa_uri) |>
    #       paste(collapse = "\n")
    #
    #     filter_shape_refs <- as.character(glue::glue(" filter=\"url(#{node_id_svg})\" "))
    #
    #     svg_shape_nos <-
    #       svg_tbl |>
    #       dplyr::filter(node_id %in% node_id_svg) |>
    #       dplyr::filter(type == "node_block") |>
    #       dplyr::pull(index)
    #
    #     svg_shape_nos <- svg_shape_nos + 3
    #     svg_text_nos <- svg_shape_nos + 1
    #
    #     # Modify shape lines
    #     for (i in seq_len(node_id_svg)) {
    #
    #       svg_vec[svg_shape_nos[i]] <-
    #         sub(" ", paste0(filter_shape_refs[i]), svg_vec[svg_shape_nos[i]])
    #
    #       svg_vec[svg_text_nos[i]] <- ""
    #     }
    #
    #     # Add in <filter> lines
    #     svg_vec[svg_line_no + 1] <-
    #       paste0(svg_vec[svg_line_no + 1], "\n\n", filter_lines, "\n")
    #   }
    #
    #   svg_vec_1 <- paste(svg_vec, collapse = "\n")
    #
    #   display <- htmltools::browsable(htmltools::HTML(svg_vec_1))
    #
    # } else {
    #
    #   cli::cli_abort(
    #     c(
    #       "Cannot currently render FontAwesome icons.",
    #       "please install the `fontawesome` package and retry",
    #       "pkg installed using `devtools::install_github('rstudio/fontawesome')`"))
    # }
  } else {
    # Generate DOT code
    dot_code <- generate_dot(graph)

    # Generate a `grViz` object
    grVizObject <-
      grViz(
        diagram = dot_code,
        width = width,
        height = height
      )

    display <- grVizObject
  }

  display
}

#' Apply a layout position to a single node
#'
#' @description
#'
#' Apply position information for a single node. This is done by setting the `x`
#' and `y` attrs for a node `id` or node `label` supplied in `node`. When
#' rendering the graph, nodes with attribute values set for `x` and `y` will be
#' fixed to those positions on the graph canvas.
#'
#' @inheritParams render_graph
#' @param node A single-length vector containing either a node ID value
#'   (integer) or a node label (character) for which position information should
#'   be applied.
#' @param x The x coordinate to set for the node.
#' @param y The y coordinate to set for the node.
#' @param use_labels An option to use a node `label` value in `node`. Note that
#'   this is only possible if all nodes have distinct `label` values set and
#'   none exist as an NA value.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a simple graph with 4 nodes
#' graph <-
#'   create_graph() %>%
#'   add_node(label = "one") %>%
#'   add_node(label = "two") %>%
#'   add_node(label = "three") %>%
#'   add_node(label = "four")
#'
#' # Add position information to each of
#' # the graph's nodes
#' graph <-
#'   graph %>%
#'   set_node_position(
#'     node = 1,
#'     x = 1, y = 1) %>%
#'   set_node_position(
#'     node = 2,
#'     x = 2, y = 2) %>%
#'   set_node_position(
#'     node = 3,
#'     x = 3, y = 3) %>%
#'   set_node_position(
#'     node = 4,
#'     x = 4, y = 4)
#'
#' # View the graph's node data frame to
#' # verify that the `x` and `y` node
#' # attributes are available and set to
#' # the values provided
#' graph %>% get_node_df()
#'
#' # The same function can modify the data
#' # in the `x` and `y` attributes
#' graph <-
#'   graph %>%
#'   set_node_position(
#'     node = 1,
#'     x = 1, y = 4) %>%
#'   set_node_position(
#'     node = 2,
#'     x = 3, y = 3) %>%
#'   set_node_position(
#'     node = 3,
#'     x = 3, y = 2) %>%
#'   set_node_position(
#'     node = 4,
#'     x = 4, y = 1)
#'
#' # View the graph's node data frame
#' graph %>% get_node_df()
#'
#' # Position changes can also be made by
#' # supplying a node `label` value (and setting
#' # `use_labels` to TRUE). For this to work,
#' # all `label` values in the graph's ndf must
#' # be unique and non-NA
#' graph <-
#'   graph %>%
#'   set_node_position(
#'     node = "one",
#'     x = 1, y = 1,
#'     use_labels = TRUE) %>%
#'   set_node_position(
#'     node = "two",
#'     x = 2, y = 2,
#'     use_labels = TRUE)
#'
#' # View the graph's node data frame
#' graph %>% get_node_df()
#'
#' @family node creation and removal
#'
#' @export
set_node_position <- function(
    graph,
    node,
    x,
    y,
    use_labels = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get the graph's node data frame
  ndf <- graph$nodes_df

  # Stop function if the node ID provided doesn't
  # exist in the graph
  if (!use_labels && !(node %in% graph$nodes_df[, 1])) {

    cli::cli_abort(
      "The node ID provided doesn't exist in the graph.")
  }

  # If the `x` node attribute doesn't exist, create
  # that column in the `ndf`
  if (!("x" %in% colnames(ndf))) {
    ndf$x <- NA_real_
  }

  # If the `y` node attribute doesn't exist, create
  # that column in the `ndf`
  if (!("y" %in% colnames(ndf))) {
    ndf$y <- NA_real_
  }

  if (use_labels) {

    # Ensure that the label column contains unique,
    # non-NA values
    unique_labels_available <-
      is_attr_unique_and_non_na(
        graph = graph,
        which_graph_df = "ndf",
        attr = "label")

    # Stop function if `label` doesn't contain
    # unique, non-NA values
    if (!unique_labels_available) {

      cli::cli_abort(
        "The `label` attribute in the graph's ndf must contain unique, non-NA values.")
    }

    # Use `case_when` statements to selectively perform
    # a vectorized `if` statement across all nodes for
    # the `x` and `y` node attribute
    x_attr_new <-
      dplyr::case_when(
        ndf$label == node ~ x,
        .default = as.numeric(ndf$x))

    y_attr_new <-
      dplyr::case_when(
        ndf$label == node ~ y,
        .default = as.numeric(ndf$y))
  }

  if (!use_labels) {

    # Use `case_when` statements to selectively perform
    # a vectorized `if` statement across all nodes for
    # the `x` and `y` node attribute
    x_attr_new <-
      dplyr::case_when(
        ndf$id == node ~ x,
        .default = as.numeric(ndf$x))

    y_attr_new <-
      dplyr::case_when(
        ndf$id == node ~ y,
        .default = as.numeric(ndf$y))
  }

  # Replace the `x` column to the ndf with a
  # coalesced version of the column contents
  ndf$x <- dplyr::coalesce(x_attr_new, ndf$x)

  # Replace the `y` column to the ndf with a
  # coalesced version of the column contents
  ndf$y <- dplyr::coalesce(y_attr_new, ndf$y)

  # Replace the graph's node data frame with `ndf`
  graph$nodes_df <- ndf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Insert node aesthetic attributes during node creation
#'
#' @description
#'
#' This helper function should be invoked to provide values for the namesake
#' `node_aes` argument, which is present in any function where nodes are
#' created.
#'
#' @param shape The shape to use for the node. Some possible `shape` types
#'   include: `circle`, `rectangle`, `triangle`, `plaintext`, `square`, and
#'   `polygon`.
#' @param style The node line style. The `style` types that can be used are:
#'   `filled`, `invisible`, `diagonals`, `rounded`, `dashed`, `dotted`, `solid`,
#'   and `bold`.
#' @param penwidth The thickness of the stroke line (in pt units) for the node
#'   shape. The default value is `1.0`.
#' @param color The color of the node's outline. Can be any of the named colors
#'   that R knows about (obtained using the `colors()` function), or, a
#'   hexadecimal color code.
#' @param fillcolor The color with which to fill the shape of the node. Can be
#'   any of the named colors that R knows about (obtained using the `colors()`
#'   function), or, a hexadecimal color code.
#' @param image A reference to an image location.
#' @param fontname The name of the system font that will be used for any node
#'   text.
#' @param fontsize The point size of the font used for any node text.
#' @param fontcolor The color used for any node text. Can be any of the named
#'   colors that R knows about (obtained using the `colors()` function), or, a
#'   hexadecimal color code.
#' @param peripheries The repeated number of node shapes (of increasing size) to
#'   draw at the node periphery.
#' @param height The height of the node shape, in inches. The default value is
#'   `0.5` whereas the minimum value is `0.02`. This is understood as the
#'   initial, minimum height of the node. If `fixedsize` is set to `TRUE`, this
#'   will be the final height of the node. Otherwise, if the node label requires
#'   more height to fit, the node's height will be increased to contain the
#'   label.
#' @param width The width of the node shape, in inches. The default value is
#'   `0.5` whereas the minimum value is `0.02`. This is understood as the
#'   initial, minimum width of the node. If `fixedsize` is set to `TRUE`, this
#'   will be the final width of the node. Otherwise, if the node label requires
#'   more width to fit, the node's width will be increased to contain the label.
#' @param x The fixed position of the node in the x direction. Any integer-based
#'   or floating point value will be accepted.
#' @param y The fixed position of the node in the y direction. Any integer-based
#'   or floating point value will be accepted.
#' @param group The node group.
#' @param tooltip Text for a node tooltip.
#' @param xlabel External label for a node. The label will be placed outside of
#'   the node but near it. These labels are added after all nodes and edges have
#'   been placed. The labels will be placed so that they do not overlap any node
#'   or label. This means it may not be possible to place all of them.
#' @param URL A URL to associate with a node. Upon rendering the plot, clicking
#'   nodes with any associated URLs will open the URL in the default browser.
#' @param sides When using the shape `polygon`, this value will provide the
#'   number of sides for that polygon.
#' @param orientation This is the angle, in degrees, that is used to rotate
#'   nodes that have a `shape` of `polygon`. Not that for any of the polygon
#'   shapes (set by the `sides` node attribute), a value for `orientation` that
#'   is `0` results in a flat base.
#' @param skew A `0-1` value that will result in the node shape being skewed to
#'   the right (from bottom to top). A value in the range `0` to `-1` will skew
#'   the shape to the left.
#' @param distortion A distortion factor that is used only when a `shape` of
#'   `polygon` is used. A `0-1` value will increasingly result in the top part
#'   of the node polygon shape to be larger than the bottom. Moving from `0`
#'   toward `-1` will result in the opposite distortion effect.
#' @param gradientangle The path angle for the node color fill gradient.
#' @param fixedsize If set to `FALSE`, the size of a node is determined by
#'   smallest width and height needed to contain its label, if any, with a
#'   margin specified by the `margin` node attribute. The width and height must
#'   also be at least as large as the sizes specified by the `width` and
#'   `height` node attributes, which specify the minimum values. If set to
#'   `TRUE`, the node size is entirely specified by the values of the `width`
#'   and `height` node attributes (i.e., the node is not expanded in size to
#'   contain the text label).
#' @param labelloc Sets the vertical placement of labels for nodes and clusters.
#'   This attribute is used only when the height of the node is larger than the
#'   height of its label. The `labelloc` node attribute can be set to either `t`
#'   (top), `c` (center), or `b` (bottom). By default, the label is vertically
#'   centered.
#' @param margin Sets the amount of space around the node's label. By default,
#'   the value is `0.11,0.055`.
#' @examples
#' # Create a new graph and add
#' # a path with several node
#' # aesthetic attributes
#' graph <-
#'   create_graph() |>
#'   add_path(
#'     n = 3,
#'     type = "path",
#'     node_aes = node_aes(
#'       shape = "circle",
#'       x = c(1, 3, 2),
#'       y = c(4, -1, 3)
#'     )
#'   )
#'
#' # View the graph's internal
#' # node data frame; the node
#' # aesthetic attributes have
#' # been inserted
#' graph |> get_node_df()
#'
#' # Create a new graph which is
#' # fully connected
#' graph <-
#'   create_graph() |>
#'   add_full_graph(
#'     n = 4,
#'     node_data = node_data(value = 1:4),
#'     node_aes = node_aes(
#'       x = c(2, 1, 3, 2),
#'       y = c(3, 2, 2, 1)
#'     ),
#'     edge_aes = edge_aes(color = "blue")
#'   )
#' @family aesthetics
#' @export
node_aes <- function(shape = NULL,
                     style = NULL,
                     penwidth = NULL,
                     color = NULL,
                     fillcolor = NULL,
                     image = NULL,
                     #fa_icon = NULL,
                     fontname = NULL,
                     fontsize = NULL,
                     fontcolor = NULL,
                     peripheries = NULL,
                     height = NULL,
                     width = NULL,
                     x = NULL,
                     y = NULL,
                     group = NULL,
                     tooltip = NULL,
                     xlabel = NULL,
                     URL = NULL,
                     sides = NULL,
                     orientation = NULL,
                     skew = NULL,
                     distortion = NULL,
                     gradientangle = NULL,
                     fixedsize = NULL,
                     labelloc = NULL,
                     margin = NULL) {

  # Collect vectors of node aesthetic
  # attribute values into a list object
  attr_values <-
    list(
      shape = shape,
      style = style,
      penwidth = penwidth,
      color = color,
      fillcolor = fillcolor,
      image = image,
      #fa_icon = fa_icon,
      fontname = fontname,
      fontsize = fontsize,
      fontcolor = fontcolor,
      peripheries = peripheries,
      height = height,
      width = width,
      x = x,
      y = y,
      group = group,
      tooltip = tooltip,
      xlabel = xlabel,
      URL = URL,
      sides = sides,
      orientation = orientation,
      skew = skew,
      distortion = distortion,
      gradientangle = gradientangle,
      fixedsize = fixedsize,
      labelloc = labelloc,
      margin = margin)

  non_null_attrs <-
    seq_along(attr_values) |>
    purrr::map_chr(.f = function(x) {
      if (!is.null(attr_values[[x]])) {
        names(attr_values[x])
      } else {
        NA
      }
    })

  non_null_attrs <- non_null_attrs[which(!is.na(non_null_attrs))]

  attr_values[non_null_attrs]
}

#' Delete all selected edges in an edge selection
#'
#' @description
#'
#' In a graph object of class `dgr_graph`, delete all edges present in a
#' selection.
#'
#' This function makes use of an active selection of edges (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(n = 3) %>%
#'   add_edges_w_string(
#'     edges = "1->3 1->2 2->3")
#'
#' # Select edges attached to
#' # node with ID `3` (these are
#' # `1`->`3` and `2`->`3`)
#' graph <-
#'   graph %>%
#'   select_edges_by_node_id(nodes = 3)
#'
#' # Delete edges in selection
#' graph <-
#'   graph %>%
#'   delete_edges_ws()
#'
#' # Get a count of edges in the graph
#' graph %>% count_edges()
#'
#' @family edge creation and removal
#'
#' @export
delete_edges_ws <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, "So, there cannot be edges to delete.")

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid edge selection
  check_graph_contains_edge_selection(graph)

  # Get vectors of the nodes in edges to be deleted
  from_delete <- graph$edge_selection$from
  to_delete <- graph$edge_selection$to

  # Get the number of edges in the graph
  edges_graph_1 <- graph %>% count_edges()

  # Delete all edges in selection
  for (i in seq_along(from_delete)) {
    graph <-
      delete_edge(
        graph = graph,
        from = from_delete[i],
        to = to_delete[i])
  }

  # Replace `graph$node_selection` with an empty df
  graph$node_selection <- create_empty_nsdf()

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Remove any `delete_edge` records from the graph log
  graph$graph_log <-
    graph$graph_log[-((nrow(graph$graph_log) - (i - 1)):nrow(graph$graph_log)), ]

  # Scavenge any invalid, linked data frames
  graph <-
    remove_linked_dfs(graph)

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_removed <- edges_graph_2 - edges_graph_1

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_removed)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph)
  }

  graph
}

#' Get the last set of nodes created in a graph
#'
#' @description
#'
#' Get the last nodes that were created in a graph object of class `dgr_graph`.
#' Provides a vector of node ID values. This function should ideally be used
#' just after creating the nodes.
#'
#' @inheritParams render_graph
#'
#' @return A vector of node ID values.
#'
#' @examples
#' # Create a graph and add 4 nodes
#' # in 2 separate function calls
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(
#'     n = 2,
#'     type = "a",
#'     label = c("a_1", "a_2")) %>%
#'   add_n_nodes(
#'     n = 2,
#'     type = "b",
#'     label = c("b_1", "b_2"))
#'
#' # Get the last nodes created (2 nodes
#' # from the last function call)
#' graph %>% get_last_nodes_created()
#'
#' @export
get_last_nodes_created <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  graph_transform_steps <-
    graph$graph_log %>%
    dplyr::mutate(
      step_created_nodes = as.integer(function_used %in% node_creation_functions()),
      step_deleted_nodes = as.integer(function_used %in% node_deletion_functions()),
      step_init_with_nodes = as.integer(function_used %in% graph_init_functions() &
                                          nodes > 0)
    ) %>%
    dplyr::filter(
      # if any step is TRUE (1)
      dplyr::if_any(
        .cols = c(step_created_nodes, step_deleted_nodes, step_init_with_nodes),
        .fns = function(x) x == 1
        )
    ) %>%
    dplyr::select(-"version_id", -"time_modified", -"duration")

  if (nrow(graph_transform_steps) > 0) {

    if (graph_transform_steps %>%
        utils::tail(1) %>%
        dplyr::pull(step_deleted_nodes) == 1) {

      cli::cli_abort(
        "The previous graph transformation function resulted in a removal of nodes.")

    } else {
      if (nrow(graph_transform_steps) > 1) {
        number_of_nodes_created <-
          (graph_transform_steps %>%
             dplyr::select(nodes) %>%
             utils::tail(2) %>%
             dplyr::pull(nodes))[2] -
          (graph_transform_steps %>%
             dplyr::select(nodes) %>%
             utils::tail(2) %>%
             dplyr::pull(nodes))[1]
      } else {
        number_of_nodes_created <-
          graph_transform_steps %>%
          dplyr::pull("nodes")
      }
    }

    node_id_values <-
      graph$nodes_df %>%
      dplyr::select("id") %>%
      utils::tail(number_of_nodes_created) %>%
      dplyr::pull("id")
  } else {
    node_id_values <- NA
  }

  node_id_values
}

#' Get the number of automorphisms
#'
#' @description
#'
#' Get the number of automorphisms the graph contains. An automorphism of a
#' graph is a form of symmetry in which the graph is mapped onto itself while
#' preserving edge-node connectivity.
#'
#' @inheritParams render_graph
#'
#' @return A single numeric value representing the number of automorphisms the
#'   graph contains.
#'
#' @examples
#' # Create a cycle graph
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' # Get a count of automorphisms
#' graph |>
#'   count_automorphisms()
#'
#' # Create a full graph and then
#' # count the automorphisms
#' create_graph() |>
#'   add_full_graph(n = 10) |>
#'   count_automorphisms()
#'
#' @export
count_automorphisms <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an undirected graph
  graph_undirected <- set_graph_undirected(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph_undirected)

  # Get the number of automorphisms in
  # the graph
  res <- unlist(igraph::count_automorphisms(
    graph = ig_graph,
    sh = "fm")["group_size"])

  as.numeric(res)
}

#' Convert an igraph graph to a DiagrammeR one
#'
#' Convert an igraph graph to a DiagrammeR graph object.
#'
#' @param igraph An \pkg{igraph} graph object.
#' @inheritParams create_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a DiagrammeR graph object
#' dgr_graph_orig <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 36,
#'     m = 50,
#'     set_seed = 23)
#'
#' # Convert the DiagrammeR
#' # graph to an igraph object
#' ig_graph <-
#'   dgr_graph_orig |>
#'   to_igraph()
#'
#' # Convert the igraph graph
#' # back to a DiagrammeR graph
#' dgr_graph_new <-
#'   ig_graph |>
#'   from_igraph()
#'
#' # Get some graph information
#' graph_info <- dgr_graph_new |>
#'   get_graph_info()
#'
#' graph_info[, 1:6]
#'
#' @export
from_igraph <- function(igraph,
                        graph_name = NULL,
                        write_backups = FALSE,
                        display_msgs = FALSE) {

  # Get vectors of all node and edge attributes
  node_attrs <- igraph::vertex_attr_names(igraph)
  edge_attrs <- igraph::edge_attr_names(igraph)

  # Generate a single-column ndf with node ID values
  nodes_df <-
    data.frame(
      id = as.integer(igraph::V(igraph)),
      stringsAsFactors = FALSE)

  # If the `type` attr exists, add that to the ndf
  if ("type" %in% node_attrs) {
    nodes_df <-
      cbind(
        nodes_df,
        data.frame(
          type = igraph::vertex_attr(igraph, "type"),
          stringsAsFactors = FALSE))
  } else {
    nodes_df <-
      cbind(
        nodes_df,
        data.frame(
          type = rep(NA_character_, nrow(nodes_df)),
          stringsAsFactors = FALSE))
  }

  # If the `label` attr exists, add that to the ndf
  if ("label" %in% node_attrs) {
    nodes_df <-
      cbind(
        nodes_df,
        data.frame(
          label = igraph::vertex_attr(igraph, "label"),
          stringsAsFactors = FALSE))
  } else {
    nodes_df <-
      cbind(
        nodes_df,
        data.frame(
          label = rep(NA_character_, nrow(nodes_df)),
          stringsAsFactors = FALSE))
  }

  # Determine if there are any extra node attrs
  extra_node_attrs <-
    base::setdiff(node_attrs, c("name", "type", "label"))

  # If there are extra node attrs, add to the ndf
  if (length(extra_node_attrs) > 0) {

    for (i in seq_along(extra_node_attrs)) {

      df_col <-
        data.frame(
          igraph::vertex_attr(igraph, extra_node_attrs[i]),
          stringsAsFactors = FALSE)

      colnames(df_col) <- extra_node_attrs[i]

      nodes_df <-
        cbind(nodes_df, df_col)
    }
  }

  # Generate a 3 column edf with `id`, `to`, and `from` values
  edges_df <-
    data.frame(
      id = as.integer(igraph::E(igraph)),
      from = as.integer(igraph::ends(igraph, igraph::E(igraph))[, 1]),
      to = as.integer(igraph::ends(igraph, igraph::E(igraph))[, 2]),
      stringsAsFactors = FALSE)

  # If the `rel` attr exists, add that to the edf
  if ("rel" %in% edge_attrs) {
    edges_df <-
      cbind(
        edges_df,
        data.frame(
          rel = igraph::edge_attr(igraph, "rel"),
          stringsAsFactors = FALSE))
  } else {
    edges_df <-
      cbind(
        edges_df,
        data.frame(
          rel = rep(NA_character_, nrow(edges_df)),
          stringsAsFactors = FALSE))
  }

  # Determine if there are any extra edge attrs
  extra_edge_attrs <- base::setdiff(edge_attrs, "rel")

  # If there are extra edge attrs, add to the edf
  if (length(extra_edge_attrs) > 0) {

    for (i in seq_along(extra_edge_attrs)) {

      df_col <-
        data.frame(
          igraph::edge_attr(igraph, extra_edge_attrs[i]),
          stringsAsFactors = FALSE)

      colnames(df_col) <- extra_edge_attrs[i]

      edges_df <- cbind(edges_df, df_col)
    }
  }

  # Ensure that the ndf is sorted ascending by node ID
  nodes_df <- dplyr::arrange(nodes_df, id)

  # Create a DiagrammeR graph object
  graph <-
    create_graph(
      nodes_df = nodes_df,
      edges_df = edges_df,
      directed = igraph::is_directed(igraph),
      graph_name = graph_name,
      write_backups = write_backups,
      display_msgs = display_msgs)

  graph
}

#' Get node IDs associated with edges
#'
#' @description
#'
#' Obtain a vector, data frame, or list of node IDs associated with edges in a
#' graph object. An optional filter by edge attribute can limit the set of edges
#' returned.
#'
#' @inheritParams render_graph
#' @param conditions an option to use filtering conditions for the retrieval of
#'   edges.
#' @param return_type using `vector` (the default), a vector of character
#'   objects representing the edges is provided. With `list` a list object will
#'   be provided that contains vectors of outgoing and incoming node IDs
#'   associated with edges. With `df`, a data frame containing outgoing and
#'   incoming node IDs associated with edges.
#' @param return_values using `id` (the default) results in node ID values
#'   returned in the edge definitions. With `label`, the node labels will
#'   instead be used to define edges.
#'
#' @return A list, data frame, or a vector object, depending on the value given
#'   to `return_type`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     label = c("one", "two", "three", "four"),
#'     type = "letter",
#'     color = c("red", "green", "grey", "blue"),
#'     value = c(3.5, 2.6, 9.4, 2.7))
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "leading_to",
#'     color = c("pink", "blue", "blue"),
#'     value = c(3.9, 2.5, 7.3))
#'
#' # Create a graph
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Get all edges within a graph, returned as a list
#' graph %>%
#'   get_edges(
#'     return_type = "vector")
#'
#' # Get all edges within a graph, returned as a
#' # data frame
#' graph %>%
#'   get_edges(
#'     return_type = "df")
#'
#' # Get all edges returned as a list
#' graph %>%
#'   get_edges(
#'     return_type = "list")
#'
#' # Get a vector of edges using
#' # a numeric comparison (i.e.,
#' # all edges with a `value`
#' # attribute greater than 3)
#' graph %>%
#'   get_edges(
#'     conditions = value > 3,
#'     return_type = "vector")
#'
#' # Get a vector of edges using
#' # a matching condition
#' graph %>%
#'   get_edges(
#'     conditions = color == "pink",
#'     return_type = "vector")
#'
#' # Use multiple conditions to
#' # return edges with the
#' # desired attribute values
#' graph %>%
#'   get_edges(
#'     conditions =
#'       color == "blue" &
#'       value > 3,
#'     return_type = "vector")
#'
#' # Use `return_values = "label"`
#' # to return the labels of the
#' # connected nodes
#' graph %>%
#'   get_edges(
#'     conditions =
#'       color == "blue" &
#'       value > 3,
#'     return_type = "vector",
#'     return_values = "label")
#'
#' @export
get_edges <- function(
    graph,
    conditions = NULL,
    return_type = "vector",
    return_values = "id"
) {

  # Extract edge data frame from the graph
  edges_df <- graph$edges_df

  if (return_values == "label") {
    edges_df <-
      edges_df %>%
      dplyr::left_join(graph$nodes_df %>% dplyr::select("id", from_label_ = "label"), by = c("from" = "id")) %>%
      dplyr::left_join(graph$nodes_df %>% dplyr::select("id", to_label_ = "label"), by = c("to" = "id"))
  }


  # If conditions are provided then
  # pass in those conditions and filter the
  # data frame of `edges_df`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    edges_df <- dplyr::filter(.data = edges_df, {{ conditions }})
  }

  # If no edges remain then return NA
  if (nrow(edges_df) == 0) {
    return(NA)
  }

  if (return_type == "list") {

    edges_list <- vector(mode = "list")
    edges_list[[1]] <- edges_list[[2]] <- vector(mode = "integer")

    if (return_values == "id") {
      edges_list[[1]] <- c(edges_list[[1]], edges_df$from)
      edges_list[[2]] <- c(edges_list[[2]], edges_df$to)
    } else if (return_values == "label") {
      edges_list[[1]] <- c(edges_list[[1]], edges_df$from_label_)
      edges_list[[2]] <- c(edges_list[[2]], edges_df$to_label_)
    }

    return(edges_list)
  }

  if (return_type == "df") {

    if (return_values == "id") {
      edges_df <-
        edges_df %>%
        dplyr::select("from", "to")
    } else if (return_values == "label") {
      edges_df <-
        edges_df %>%
        dplyr::select("from_label_", "to_label_")
    }

    return(edges_df)
  }

  if (return_type == "vector") {

    if (return_values == "id") {
      edges_vector <-
        paste0(edges_df$from, "->", edges_df$to)
    } else if (return_values == "label") {
      edges_vector <-
        paste0(edges_df$from_label_, "->", edges_df$to_label_)
    }

    return(edges_vector)
  }
}

#' Traverse from one or more selected nodes onto adjacent edges
#'
#' @description
#'
#' From a graph object of class `dgr_graph` move to adjacent edges from a
#' selection of one or more selected nodes, thereby creating a selection of
#' edges. An optional filter by edge attribute can limit the set of edges
#' traversed to.
#'
#' This traversal function makes use of an active selection of nodes. After the
#' traversal, depending on the traversal conditions, there will either be a
#' selection of edges or no selection at all.
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param conditions An option to use filtering conditions for the traversal.
#' @param copy_attrs_from Providing a node attribute name will copy those node
#'   attribute values to the traversed edges. If the edge attribute already
#'   exists, the values will be merged to the traversed edges; otherwise, a new
#'   edge attribute will be created.
#' @param copy_attrs_as If a node attribute name is provided in
#'   `copy_attrs_from`, this option will allow the copied attribute values
#'   to be written under a different edge attribute name. If the attribute name
#'   provided in `copy_attrs_as` does not exist in the graph's edf, the new
#'   edge attribute will be created with the chosen name.
#' @param agg If a node attribute is provided to `copy_attrs_from`, then an
#'   aggregation function is required since there may be cases where multiple
#'   node attribute values will be passed onto the traversed edge(s). To pass
#'   only a single value, the following aggregation functions can be used:
#'   `sum`, `min`, `max`, `mean`, or `median`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(
#'     n = 2,
#'     type = "a",
#'     label = c("asd", "iekd")) |>
#'   add_n_nodes(
#'     n = 3,
#'     type = "b",
#'     label = c("idj", "edl", "ohd")) |>
#'   add_edges_w_string(
#'     edges = "1->2 1->3 2->4 2->5 3->5",
#'     rel = c(NA, "A", "B", "C", "D"))
#'
#' # Create a data frame with node ID values
#' # representing the graph edges (with `from`
#' # and `to` columns), and, a set of numeric values
#' df <-
#'   data.frame(
#'     from = c(1, 1, 2, 2, 3),
#'     to = c(2, 3, 4, 5, 5),
#'     values = round(rnorm(5, 5), 2))
#'
#' # Join the data frame to the graph's internal
#' # edge data frame (edf)
#' graph <-
#'   graph |>
#'   join_edge_attrs(df = df)
#'
#' # Show the graph's internal edge data frame
#' graph |> get_edge_df()
#'
#' # Perform a simple traversal from nodes to
#' # adjacent edges with no conditions on the
#' # nodes traversed to
#' graph |>
#'   select_nodes_by_id(nodes = 3) |>
#'   trav_both_edge() |>
#'   get_selection()
#'
#' # Traverse from node `2` to any adjacent
#' # edges, filtering to those edges that have
#' # NA values for the `rel` edge attribute
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_both_edge(
#'     conditions = is.na(rel)) |>
#'   get_selection()
#'
#' # Traverse from node `2` to any adjacent
#' # edges, filtering to those edges that have
#' # numeric values greater than `6.5` for
#' # the `rel` edge attribute
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_both_edge(
#'     conditions = values > 6.5) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any adjacent
#' # edges, filtering to those edges that
#' # have values equal to `C` for the `rel`
#' # edge attribute
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_both_edge(
#'     conditions = rel == "C") |>
#'   get_selection()
#'
#' # Traverse from node `2` to any adjacent
#' # edges, filtering to those edges that
#' # have values in the set `B` and `C` for
#' # the `rel` edge attribute
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_both_edge(
#'     conditions = rel %in% c("B", "C")) |>
#'   get_selection()
#'
#' # Traverse from node `2` to any adjacent
#' # edges, and use multiple conditions for the
#' # traversal
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_both_edge(
#'     conditions =
#'       rel %in% c("B", "C") &
#'       values > 4.0) |>
#'   get_selection()
#'
#' # Traverse from node `2` to any adjacent
#' # edges, and use multiple conditions with
#' # a single-length vector
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_both_edge(
#'     conditions =
#'       rel %in% c("B", "C") |
#'       values > 4.0) |>
#'   get_selection()
#'
#' # Traverse from node `2` to any adjacent
#' # edges, and use a regular expression as
#' # a filtering condition
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_both_edge(
#'     conditions = grepl("B|C", rel)) |>
#'   get_selection()
#'
#' # Create another simple graph to demonstrate
#' # copying of node attribute values to traversed
#' # edges
#' graph <-
#'   create_graph() |>
#'   add_path(n = 4) |>
#'   select_nodes_by_id(nodes = 2:3) |>
#'   set_node_attrs_ws(
#'     node_attr = value,
#'     value = 5)
#'
#' # Show the graph's internal edge data frame
#' graph |>get_edge_df()
#'
#' # Show the graph's internal node data frame
#' graph |> get_node_df()
#'
#' # Perform a traversal from the nodes to
#' # the adjacent edges while also applying
#' # the node attribute `value` to the edges (in
#' # this case summing the `value` of 5 from
#' # all contributing nodes adding as an edge
#' # attribute)
#' graph <-
#'   graph |>
#'   trav_both_edge(
#'     copy_attrs_from = value,
#'     agg = "sum")
#'
#' # Show the graph's internal edge data frame
#' # after this change
#' graph |> get_edge_df()
#'
#' @export
trav_both_edge <- function(
    graph,
    conditions = NULL,
    copy_attrs_from = NULL,
    copy_attrs_as = NULL,
    agg = "sum"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(
    graph,
    c("Any traversal requires an active selection.",
      "This type of traversal requires a selection of nodes."))

  # Get the requested `copy_attrs_from`
  copy_attrs_from <-
    rlang::enquo(copy_attrs_from) |> rlang::get_expr() |> as.character()

  # Get the requested `copy_attrs_as`
  copy_attrs_as <-
    rlang::enquo(copy_attrs_as) |> rlang::get_expr() |> as.character()

  if (length(copy_attrs_from) == 0) {
    copy_attrs_from <- NULL
  }

  if (length(copy_attrs_as) == 0) {
    copy_attrs_as <- NULL
  }

  if (!is.null(copy_attrs_as) && !is.null(copy_attrs_from) &&
      copy_attrs_as == copy_attrs_from
      ) {
    copy_attrs_as <- NULL
  }

  # Get the selection of nodes as the starting
  # nodes for the traversal
  starting_nodes <- graph$node_selection$node

  # Get the graph's edge data frame
  edf <- graph$edges_df

  # Get the graph's node data frame
  ndf <- graph$nodes_df

  # Find all edges that lead away from the
  # starting nodes and remove edges that
  # are loops
  valid_edges <-
    edf |>
    dplyr::filter(from %in% starting_nodes |
                    to %in% starting_nodes) |>
    dplyr::filter(to != from)

  # If traversal conditions are provided then
  # pass in those conditions and filter the
  # data frame of `valid_edges`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    valid_edges <- dplyr::filter(.data = valid_edges, {{ conditions }})
  }

  # If no rows returned, then there are no
  # valid traversals, so return the same graph
  # object without modifying the selection
  if (nrow(valid_edges) == 0) {
    return(graph)
  }

  # If the option is taken to copy node attribute
  # values to the traversed edges, perform the join
  # operation
  if (!is.null(copy_attrs_from)) {

    if (!(copy_attrs_from %in% colnames(edf))) {

      from_join <-
        ndf |>
        dplyr::select("id", !!enquo(copy_attrs_from)) |>
        dplyr::filter(id %in% starting_nodes) |>
        dplyr::right_join(
          valid_edges |>
            dplyr::select(-"rel") |>
            dplyr::rename(e_id = "id"),
          by = c("id" = "from")) |>
        dplyr::select("e_id", !!enquo(copy_attrs_from))

      to_join <-
        ndf |>
        dplyr::select("id", !!enquo(copy_attrs_from)) |>
        dplyr::filter(id %in% starting_nodes) |>
        dplyr::right_join(
          valid_edges |>
            dplyr::select(-"rel") |>
            dplyr::rename(e_id = "id"),
          by = c("id" = "to")) |>
        dplyr::select("e_id", !!enquo(copy_attrs_from))

      if (!is.null(copy_attrs_as)) {

        if (copy_attrs_as %in% c("id", "from", "to")) {

          cli::cli_abort(
            "Copied attributes should not overwrite either of the `id`, `from`, or `to` edge attributes.")
        }

        colnames(from_join)[2] <-
          colnames(to_join)[2] <-
          copy_attrs_from <- copy_attrs_as
      }

      edges <-
        dplyr::bind_rows(
          from_join, to_join) |>
        dplyr::left_join(edf, by = c("e_id" = "id")) |>
        dplyr::rename(id = "e_id") |>
        dplyr::group_by(id) |>
        dplyr::summarize(!!copy_attrs_from :=
                           match.fun(!!agg)(!!as.name(copy_attrs_from),
                                             na.rm = TRUE)) |>
        dplyr::right_join(edf, by = "id") |>
        dplyr::relocate("id", "from", "to", "rel") |>
        as.data.frame(stringsAsFractions = FALSE)
    }

    if (copy_attrs_from %in% colnames(edf)) {

      from_join <-
        ndf |>
        dplyr::select("id", !!enquo(copy_attrs_from)) |>
        dplyr::filter(id %in% starting_nodes)

      if (!is.null(copy_attrs_as)) {

        if (copy_attrs_as %in% c("id", "from", "to")) {

          cli::cli_abort(
            "Copied attributes should not overwrite either of the `id`, `from`, or `to` edge attributes.")
        }

        colnames(from_join)[2] <- copy_attrs_as
      }

      from_join <-
        from_join |>
        dplyr::right_join(
          valid_edges |>
            dplyr::select(-"rel") |>
            dplyr::rename(e_id = "id"),
          by = c("id" = "from"))

      # Get column numbers that end with ".x" or ".y"
      split_var_x_col <-
        grep("\\.x$", colnames(from_join))

      split_var_y_col <-
        grep("\\.y$", colnames(from_join))

      # Selectively merge in values to the existing
      # edge attribute column
      for (i in seq_len(nrow(from_join))) {
        if (!is.na(from_join[i, split_var_x_col])) {
          from_join[i, split_var_y_col] <- from_join[i, split_var_x_col]
        }
      }

      # Rename the ".y" column
      colnames(from_join)[split_var_y_col] <- copy_attrs_from

      # Drop the ".x" column
      from_join <-
        from_join |>
        dplyr::select("e_id", !!enquo(copy_attrs_from))

      to_join <-
        ndf |>
        dplyr::select("id", !!enquo(copy_attrs_from)) |>
        dplyr::filter(id %in% starting_nodes)

      if (!is.null(copy_attrs_as)) {

        colnames(to_join)[2] <- copy_attrs_as
      }

      to_join <-
        to_join |>
        dplyr::right_join(
          valid_edges |>
            dplyr::select(-"rel") |>
            dplyr::rename(e_id = "id"),
          by = c("id" = "to"))

      if (!is.null(copy_attrs_as)) {

        copy_attrs_from <- copy_attrs_as
      }

      # Get column numbers that end with ".x" or ".y"
      split_var_x_col <-
        grep("\\.x$", colnames(to_join))

      split_var_y_col <-
        grep("\\.y$", colnames(to_join))

      # Selectively merge in values to the existing
      # edge attribute column
      for (i in seq_len(nrow(to_join))) {
        if (!is.na(to_join[i, split_var_x_col])) {
          to_join[i, split_var_y_col] <- to_join[i, split_var_x_col]
        }
      }

      # Rename the ".y" column
      colnames(to_join)[split_var_y_col] <- copy_attrs_from

      # Drop the ".x" column
      to_join <-
        to_join |>
        dplyr::select("e_id", !!enquo(copy_attrs_from))

      edges <-
        dplyr::bind_rows(
          from_join, to_join) |>
        dplyr::left_join(edf, by = c("e_id" = "id"))

      # Get column numbers that end with ".x" or ".y"
      split_var_x_col <-
        grep("\\.x$", colnames(edges))

      split_var_y_col <-
        grep("\\.y$", colnames(edges))

      # Selectively merge in values to the existing
      # edge attribute column
      for (i in seq_len(nrow(edges))) {
        if (!is.na(edges[i, split_var_x_col])) {
          edges[i, split_var_y_col] <- edges[i, split_var_x_col]
        }
      }

      # Rename the ".y" column
      colnames(edges)[split_var_y_col] <- copy_attrs_from

      # Drop the ".x" column
      edges <- edges[-split_var_x_col]

      joined_edges <-
        edges |>
        dplyr::arrange(e_id) |>
        dplyr::rename(id = "e_id") |>
        dplyr::group_by(id) |>
        dplyr::summarize(!!copy_attrs_from :=
                            match.fun(!!agg)(!!as.name(copy_attrs_from),
                                              na.rm = TRUE), .groups = "drop") |>
        dplyr::right_join(edf, by = "id")

      # Get column numbers that end with ".x" or ".y"
      split_var_x_col <-
        grep("\\.x$", colnames(joined_edges))

      split_var_y_col <-
        grep("\\.y$", colnames(joined_edges))

      # Selectively merge in values to the existing
      # edge attribute column
      for (i in seq_len(nrow(joined_edges))) {
        if (!is.na(joined_edges[i, split_var_x_col])) {
          joined_edges[i, split_var_y_col] <- joined_edges[i, split_var_x_col]
        }
      }

      # Rename the ".y" column
      colnames(joined_edges)[split_var_y_col] <- copy_attrs_from

      # Drop the ".x" column
      joined_edges <- joined_edges[-split_var_x_col]

      edges <-
        joined_edges |>
        dplyr::relocate("id", "from", "to", "rel") |>
        dplyr::arrange(id) |>
        as.data.frame(stringsAsFractions = FALSE)
    }

    # Update the graph's internal node data frame
    graph$edges_df <- edges
  }

  # Add the edge information to the active selection
  # of edges in `graph$edge_selection`
  graph$edge_selection <-
    replace_graph_edge_selection(
      graph = graph,
      edge_id = valid_edges$id,
      from_node = valid_edges$from,
      to_node = valid_edges$to)

  # Replace `graph$node_selection` with an empty df
  graph$node_selection <- create_empty_nsdf()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get detailed information on edges
#'
#' @description
#'
#' Obtain a data frame with detailed information on edges and their
#' interrelationships within the graph.
#'
#' @inheritParams render_graph
#'
#' @return A data frame containing information specific to each edge within the
#'   graph.
#'
#' @examples
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 5, m = 10,
#'     set_seed = 23)
#'
#' # Get information on the
#' # graph's edges
#' graph |> get_edge_info()
#'
#' @export
get_edge_info <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # For graphs with no edges, return NA
  if (nrow(graph$edges_df) == 0) {
    return(NA)
  }

  # Extract only the first 4 columns of the
  # edge data frame
  graph$edges_df |>
    dplyr::select(id, from, to, rel)
}

#' Set edge attributes with an edge selection
#'
#' @description
#'
#' From a graph object of class `dgr_graph` or an edge data frame, set edge
#' attribute properties for one or more edges.
#'
#' This function makes use of an active selection of edges (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#' @param edge_attr The name of the attribute to set.
#' @param value The value to be set for the chosen attribute for the edges in
#'   the current selection.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_path(n = 6)
#'
#' # Select specific edges from
#' # the graph and apply the edge
#' # attribute `color = blue` to
#' # those selected edges
#' graph <-
#'   graph |>
#'   select_nodes_by_id(nodes = 2:4) |>
#'   trav_out_edge() |>
#'   set_edge_attrs_ws(
#'     edge_attr = color,
#'     value = "blue")
#'
#' # Show the internal edge data
#' # frame to verify that the
#' # edge attribute has been set
#' # for specific edges
#' graph |> get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
set_edge_attrs_ws <- function(
    graph,
    edge_attr,
    value
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid edge selection
  check_graph_contains_edge_selection(graph)

  # Get the requested `edge_attr`
  edge_attr <-
    rlang::enquo(edge_attr) |> rlang::get_expr() |> as.character()

  # Get vectors of edge ID values for the
  # edge selection
  edge_ids <- graph$edge_selection$edge

  # Update the graph's internal edf
  if (edge_attr %in% colnames(graph$edges_df)) {

    graph$edges_df[
      which(graph$edges_df[, 1] %in% edge_ids),
      which(colnames(graph$edges_df) %in% edge_attr)] <- value

  } else {
    graph$edges_df <-
      graph$edges_df |>
      dplyr::mutate(edge_attr__ = dplyr::case_when(
        id %in% edge_ids ~ value))

    colnames(graph$edges_df)[length(colnames(graph$edges_df))] <-
      edge_attr
  }

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Is the edge a loop edge?
#'
#' @description
#'
#' Determines whether an edge definition is a loop edge.
#'
#' @inheritParams render_graph
#' @param edge A numeric edge ID value.
#'
#' @return A logical value.
#'
#' @examples
#' # Create a graph that has multiple
#' # loop edges
#' graph <-
#'   create_graph() |>
#'   add_path(n = 4) |>
#'   add_edge(
#'     from = 1,
#'     to = 1) |>
#'   add_edge(
#'     from = 3,
#'     to = 3)
#'
#' # Get the graph's internal
#' # edge data frame
#' graph |> get_edge_df()
#'
#' # Determine if edge `4` is
#' # a loop edge
#' graph |> is_edge_loop(edge = 4)
#'
#' # Determine if edge `2` is
#' # a loop edge
#' graph |> is_edge_loop(edge = 2)
#'
#' @export
is_edge_loop <- function(
    graph,
    edge
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Stop function if `edge` is not a single numeric value.
  check_number_decimal(edge)

  # Obtain the graph's edf
  edf <- graph$edges_df

  # Stop function if the edge ID provided
  # is not a valid edge ID
  if (!(edge %in% edf$id)) {

    cli::cli_abort(
      "The provided edge ID is not present in the graph.")
  }

  # Obtain the edge definition
  from <-
    edf |>
    dplyr::filter(id == !!edge) |>
    dplyr::pull("from")

  to <-
    edf |>
    dplyr::filter(id == !!edge) |>
    dplyr::pull("to")

  # If the `from` and `to` node IDs
  # are the same then this is a loop edge
  if (from == to) {
    return(TRUE)
  } else {
    return(FALSE)
  }
}

#' Get all common neighbors between two or more nodes
#'
#' @description
#'
#' With two or more nodes, get the set of common neighboring nodes.
#'
#' @inheritParams render_graph
#' @param nodes a vector of node ID values of length at least 2.
#'
#' @return a vector of node ID values.
#'
#' @examples
#' # Create a directed graph with 5 nodes
#' graph <-
#'   create_graph() |>
#'   add_path(n = 5)
#'
#' # Find all common neighbor nodes
#' # for nodes `1` and `2` (there are no
#' # common neighbors amongst them)
#' graph |>
#'   get_common_nbrs(
#'     nodes = c(1, 2))
#'
#' # Find all common neighbor nodes for
#' # nodes `1` and `3`
#' graph |>
#'   get_common_nbrs(
#'     nodes = c(1, 3))
#'
#' @export
get_common_nbrs <- function(
    graph,
    nodes
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get predecessors and successors for all nodes
  # in `nodes`
  for (i in seq_along(nodes)) {
    if (i == 1) {
      nbrs <- list()
    }

    nbrs[[i]] <-
      c(sort(get_predecessors(graph, node = nodes[i])),
        sort(get_successors(graph, node = nodes[i])))
  }

  common_nbrs <- nbrs[[1]]

  for (i in nbrs[-1]) {
    common_nbrs <- intersect(common_nbrs, i)
  }

  if (length(common_nbrs) == 0) {
    return(NA)
  }

  sort(as.integer(common_nbrs))
}

#' Traverse from one or more selected edges onto adjacent, inward nodes
#'
#' @description
#'
#' From a graph object of class `dgr_graph` with an active selection of edges
#' move with the edge direction to connected nodes, replacing the current edges
#' in the selection with those nodes traversed to. An optional filter by node
#' attribute can limit the set of nodes traversed to.
#'
#' This traversal function makes use of an active selection of edges. After the
#' traversal, depending on the traversal conditions, there will either be a
#' selection of nodes or no selection at all.
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#' @param conditions An option to use filtering conditions for the traversal.
#' @param copy_attrs_from Providing an edge attribute name will copy those edge
#'   attribute values to the traversed nodes. If the edge attribute already
#'   exists, the values will be merged to the traversed nodes; otherwise, a new
#'   node attribute will be created.
#' @param copy_attrs_as If an edge attribute name is provided in
#'   `copy_attrs_from`, this option will allow the copied attribute values to be
#'   written under a different node attribute name. If the attribute name
#'   provided in `copy_attrs_as` does not exist in the graph's ndf, the new node
#'   attribute will be created with the chosen name.
#' @param agg If an edge attribute is provided to `copy_attrs_from`, then an
#'   aggregation function is required since there may be cases where multiple
#'   edge attribute values will be passed onto the traversed node(s). To pass
#'   only a single value, the following aggregation functions can be used:
#'   `sum`, `min`, `max`, `mean`, or `median`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(
#'     n = 2,
#'     type = "a",
#'     label = c("asd", "iekd")) |>
#'   add_n_nodes(
#'     n = 3,
#'     type = "b",
#'     label = c("idj", "edl", "ohd")) |>
#'   add_edges_w_string(
#'     edges = "1->2 1->3 2->4 2->5 3->5",
#'     rel = c(NA, "A", "B", "C", "D"))
#'
#' # Create a data frame with node ID values
#' # representing the graph edges (with `from`
#' # and `to` columns), and, a set of numeric values
#' df_edges <-
#'   data.frame(
#'     from = c(1, 1, 2, 2, 3),
#'     to = c(2, 3, 4, 5, 5),
#'     values = round(rnorm(5, 5), 2))
#'
#' # Create a data frame with node ID values
#' # representing the graph nodes (with the `id`
#' # columns), and, a set of numeric values
#' df_nodes <-
#'   data.frame(
#'     id = 1:5,
#'     values = round(rnorm(5, 7), 2))
#'
#' # Join the data frame to the graph's internal
#' # edge data frame (edf)
#' graph <-
#'   graph |>
#'   join_edge_attrs(df = df_edges) |>
#'   join_node_attrs(df = df_nodes)
#'
#' # Show the graph's internal node data frame
#' graph |> get_node_df()
#'
#' # Show the graph's internal edge data frame
#' graph |> get_edge_df()
#'
#' # Perform a simple traversal from the
#' # edge `1`->`3` to the attached node
#' # in the direction of the edge; here, no
#' # conditions are placed on the nodes
#' # traversed to
#' graph |>
#'   select_edges(
#'     from = 1,
#'     to = 3) |>
#'   trav_in_node() |>
#'   get_selection()
#'
#' # Traverse from edges `2`->`5` and
#' # `3`->`5` to the attached node along
#' # the direction of the edge; both
#' # traversals lead to the same node
#' graph |>
#'   select_edges(
#'     from = 2,
#'     to = 5) |>
#'   select_edges(
#'     from = 3,
#'     to = 5) |>
#'   trav_in_node() |>
#'   get_selection()
#'
#' # Traverse from the edge `1`->`3`
#' # to the attached node where the edge
#' # is incoming, this time filtering
#' # numeric values greater than `5.0` for
#' # the `values` node attribute
#' graph |>
#'   select_edges(
#'     from = 1,
#'     to = 3) |>
#'   trav_in_node(
#'     conditions = values > 5.0) |>
#'   get_selection()
#'
#' # Traverse from the edge `1`->`3`
#' # to the attached node where the edge
#' # is incoming, this time filtering
#' # numeric values less than `5.0` for
#' # the `values` node attribute (the
#' # condition is not met so the original
#' # selection of edge `1` -> `3` remains)
#' graph |>
#'   select_edges(
#'     from = 1,
#'     to = 3) |>
#'   trav_in_node(
#'     conditions = values < 5.0) |>
#'   get_selection()
#'
#' # Traverse from the edge `1`->`2` to
#' # the node `2` using multiple conditions
#' # with a single-length vector
#' graph |>
#'   select_edges(
#'     from = 1,
#'     to = 2) |>
#'   trav_in_node(
#'     conditions =
#'       grepl(".*d$", label) |
#'       values < 6.0) |>
#'   get_selection()
#'
#' # Create another simple graph to demonstrate
#' # copying of edge attribute values to traversed
#' # nodes
#' graph <-
#'   create_graph() |>
#'   add_node() |>
#'   select_nodes() |>
#'   add_n_nodes_ws(
#'     n = 2,
#'     direction = "to") |>
#'   clear_selection() |>
#'   select_nodes_by_id(nodes = 2) |>
#'   set_node_attrs_ws(
#'     node_attr = value,
#'     value = 8) |>
#'   clear_selection() |>
#'   select_edges_by_edge_id(edges = 1) |>
#'   set_edge_attrs_ws(
#'     edge_attr = value,
#'     value = 5) |>
#'   clear_selection() |>
#'   select_edges_by_edge_id(edges = 2) |>
#'   set_edge_attrs_ws(
#'     edge_attr = value,
#'     value = 5) |>
#'   clear_selection() |>
#'   select_edges()
#'
#' # Show the graph's internal edge data frame
#' graph |> get_edge_df()
#'
#' # Show the graph's internal node data frame
#' graph |> get_node_df()
#'
#' # Perform a traversal from the edges to
#' # the central node (`1`) while also applying
#' # the edge attribute `value` to the node (in
#' # this case summing the `value` of 5 from
#' # both edges before adding as a node attribute)
#' graph <-
#'   graph |>
#'   trav_in_node(
#'     copy_attrs_from = value,
#'     agg = "sum")
#'
#' # Show the graph's internal node data frame
#' # after this change
#' graph |> get_node_df()
#'
#' @export
trav_in_node <- function(
    graph,
    conditions = NULL,
    copy_attrs_from = NULL,
    copy_attrs_as = NULL,
    agg = "sum"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid edge selection
  check_graph_contains_edge_selection(
    graph,
    c("Any traversal requires an active selection.",
      "This type of traversal requires a selection of edges."))

  # Get the requested `copy_attrs_from`
  copy_attrs_from <-
    rlang::enquo(copy_attrs_from) |> rlang::get_expr() |> as.character()

  # Get the requested `copy_attrs_as`
  copy_attrs_as <-
    rlang::enquo(copy_attrs_as) |> rlang::get_expr() |> as.character()

  if (length(copy_attrs_from) == 0) {
    copy_attrs_from <- NULL
  }

  if (length(copy_attrs_as) == 0) {
    copy_attrs_as <- NULL
  }

  if (!is.null(copy_attrs_as) && !is.null(copy_attrs_from)) {
    if (copy_attrs_as == copy_attrs_from) {
      copy_attrs_as <- NULL
    }
  }

  # Get the selection of edges
  starting_edges <- graph$edge_selection

  # Get the graph's node data frame
  ndf <- graph$nodes_df

  # Get the graph's edge data frame
  edf <- graph$edges_df

  # Find all nodes that are connected to the
  # starting edges
  valid_nodes <-
    starting_edges |>
    dplyr::distinct(to) |>
    dplyr::left_join(ndf, by = c("to" = "id"))

  # If traversal conditions are provided then
  # pass in those conditions and filter the
  # data frame of `valid_nodes`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    valid_nodes <- dplyr::filter(.data = valid_nodes, {{ conditions }})
  }

  # If no rows returned, then there are no
  # valid traversals, so return the same graph
  # object without modifying the selection
  if (nrow(valid_nodes) == 0) {
    return(graph)
  }

  # If the option is taken to copy edge attribute
  # values to the traversed nodes, perform the join
  # operation
  if (!is.null(copy_attrs_from)) {

    nodes <-
      starting_edges |>
      dplyr::semi_join(valid_nodes, by = "to") |>
      dplyr::left_join(edf, by = c("edge" = "id")) |>
      dplyr::select("to.y", !!enquo(copy_attrs_from))


    if (!is.null(copy_attrs_as)) {

      if (copy_attrs_as %in% c("id", "from", "to")) {

        cli::cli_abort(
          "Copied attributes should not overwrite either of the `id`, `from`, or `to` edge attributes.")
      }

      colnames(nodes)[2] <- copy_attrs_from <- copy_attrs_as
    }

    nodes <-
      nodes |>
      dplyr::rename(id = to.y) |>
      dplyr::group_by(id) |>
      dplyr::summarize(!!copy_attrs_from :=
                         match.fun(!!agg)(!!as.name(copy_attrs_from),
                                           na.rm = TRUE)) |>
      dplyr::right_join(ndf, by = "id") |>
      dplyr::relocate("id", "type", "label") |>
      as.data.frame(stringsAsFactors = FALSE)

    # If edge attribute exists as a column in the ndf
    if (copy_attrs_from %in% colnames(ndf)) {

      # Get column numbers that end with ".x" or ".y"
      split_var_x_col <-
        grep("\\.x$", colnames(nodes))

      split_var_y_col <-
        grep("\\.y$", colnames(nodes))

      # Selectively merge in values to the existing
      # edge attribute column
      for (i in seq_len(nrow(nodes))) {
        if (!is.na(nodes[i, split_var_x_col])) {
          nodes[i, split_var_y_col] <- nodes[i, split_var_x_col]
        }
      }

      # Rename the ".y" column
      colnames(nodes)[split_var_y_col] <- copy_attrs_from

      # Drop the ".x" column
      nodes <- nodes[-split_var_x_col]

      # Reorder columns
      nodes <-
        nodes |>
        dplyr::relocate("id", "type", "label")
    }

    # Update the graph's internal node data frame
    graph$nodes_df <- nodes
  }

  # Add the node ID values to the active selection
  # of nodes in `graph$node_selection`
  graph$node_selection <-
    replace_graph_node_selection(
      graph = graph,
      replacement = valid_nodes$to)

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Combine multiple edge data frames into a single edge data frame
#'
#' @description
#'
#' Combine several edge data frames in the style of `rbind()`, except, it works
#' regardless of the number and ordering of the columns.
#'
#' @param ... Two or more edge data frames, which contain edge IDs and
#'   associated attributes.
#'
#' @return A combined edge data frame.
#'
#' @examples
#' # Create an edge data frame (edf)
#' edf_1 <-
#'   create_edge_df(
#'     from = c(1, 1, 2, 3),
#'     to = c(2, 4, 4, 1),
#'     rel = "requires",
#'     color = "green",
#'     data = c(2.7, 8.9, 2.6, 0.6))
#'
#' # Create a second edge data frame
#' edf_2 <-
#'   create_edge_df(
#'     from = c(5, 7, 8, 8),
#'     to = c(7, 8, 6, 5),
#'     rel = "receives",
#'     arrowhead = "dot",
#'     color = "red")
#'
#' # Combine the two edge data frames
#' all_edges <- combine_edfs(edf_1, edf_2)
#'
#' # View the combined edge data frame
#' all_edges
#'
#' @export
combine_edfs <- function(...) {

  data_frames <- list(...)

  for (l in seq_along(data_frames)) {
    if (l == 1) {
      df1 <- data_frames[l][[1]]
      df2 <- data_frames[l + 1][[1]]
    }
    if (l > 1 && l < length(data_frames)) {
      df1 <- edf_new
      df2 <- data_frames[l + 1][[1]]
    }

    # Bind rows from df1 and df2
    edf_new <- dplyr::bind_rows(df1, df2)

    if (l == length(data_frames)) {
      break
    }
  }

  edf_new$id <- seq_len(nrow(edf_new))

  edf_new
}

#' Get community membership by Louvain optimization
#'
#' @description
#'
#' Through the use of multi-level optimization of a modularity score, obtain the
#' group membership values for each of the nodes in the graph.
#'
#' @inheritParams render_graph
#'
#' @return A data frame with group membership assignments for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the group membership values
#' # for all nodes in the graph
#' # through the multi-level
#' # optimization of modularity
#' # algorithm
#' graph |>
#'   get_cmty_louvain()
#'
#' # Add the group membership
#' # values to the graph as a
#' # node attribute
#' graph <-
#'   graph |>
#'   join_node_attrs(
#'     df = get_cmty_louvain(graph))
#'
#' # Display the graph's
#' # node data frame
#' graph |> get_node_df()
#'
#' @export
get_cmty_louvain <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If graph is directed, transform to undirected
  graph <- set_graph_undirected(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the community object using the
  # `cluster_louvain()` function
  cmty_louvain_obj <-
    igraph::cluster_louvain(ig_graph)

  # Create df with node memberships
  data.frame(
    id = igraph::membership(cmty_louvain_obj) |> names() |> as.integer(),
    louvain_group = as.vector(igraph::membership(cmty_louvain_obj)),
    stringsAsFactors = FALSE)
}

#' Get the number of strongly-connected components
#'
#' @description
#'
#' Get the number of strongly-connected components in the graph.
#'
#' @inheritParams render_graph
#'
#' @return A single integer value representing the number of strongly-connected
#'   graph components.
#'
#' @examples
#' # Create a graph and add
#' # several graph islands
#' graph <-
#'   create_graph() |>
#'   add_islands_graph(
#'     n_islands = 4,
#'     island_size = 10,
#'     p = 1/5,
#'     edges_between = 1,
#'     set_seed = 23)
#'
#' # Get a count of strongly-connected
#' # components in the graph
#' graph |> count_s_connected_cmpts()
#'
#' @export
count_s_connected_cmpts <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the number of strongly-connected
  # components in the graph
  igraph::components(ig_graph, mode = "strong")[["no"]]
}

#' Set node attribute values
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, set node attribute values for one
#' or more nodes.
#'
#' @inheritParams render_graph
#' @param node_attr The name of the attribute to set. Examples of attributes can be found in [node_aes()].
#' @param values The values to be set for the chosen attribute for the chosen
#'   nodes.
#' @param nodes An optional vector of node IDs for filtering the list of nodes
#'   present in the graph.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = "basic",
#'     label = TRUE,
#'     value = c(3.5, 2.6, 9.4, 2.7))
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "leading_to")
#'
#' # Create a graph
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Set attribute `color = "green"` for
#' # nodes `1` and `3` using the graph object
#' graph <-
#'   graph %>%
#'   set_node_attrs(
#'     node_attr = color,
#'     values = "green",
#'     nodes = c(1, 3))
#'
#' # View the graph's node data frame
#' graph %>% get_node_df()
#'
#' # Set attribute `color = "blue"` for
#' # all nodes in the graph
#' graph <-
#'   graph %>%
#'   set_node_attrs(
#'     node_attr = color,
#'     values = "blue")
#'
#' # Display the graph's ndf
#' graph %>% get_node_df()
#'
#' @family node creation and removal
#'
#' @export
set_node_attrs <- function(
    graph,
    node_attr,
    values,
    nodes = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Get the requested `node_attr`
  node_attr <-
    rlang::enquo(node_attr) %>% rlang::get_expr() %>% as.character()

  # Stop function if `node_attr` is `id`
  # if (node_attr == "id") {
  #   stop(
  #     "You cannot use the value `id` for `node_attr`.",
  #     call. = FALSE)
  # }

  # Extract the graph's ndf
  ndf <- graph$nodes_df

  if (length(values) != 1 &&
      length(values) != nrow(ndf)) {

    cli::cli_abort(
      "The length of `values` provided must either be 1 or that of the number of rows in the ndf ({nrow(ndf)}).")
  }

  if (length(values) == 1) {
    if (node_attr %in% colnames(ndf)) {
      if (is.null(nodes)) {
        ndf[, which(colnames(ndf) %in%
                           node_attr)] <- values
      } else {
        ndf[which(ndf[, 1] %in% nodes),
                 which(colnames(ndf) %in%
                         node_attr)] <- values
      }
    }

    if (!(node_attr %in% colnames(ndf))) {

      if (is.null(nodes)) {
        # Add a new column and map the node attribute
        # value to every row in `ndf`
        ndf <-
          dplyr::mutate(
            ndf,
            new_attr__ = ifelse(id > 0, values, NA))

        colnames(ndf)[ncol(ndf)] <- node_attr

      } else {
        # Add a new column and map the node attribute
        # value to the correct rows
        ndf <-
          dplyr::mutate(
            ndf,
            new_attr__ = ifelse(id %in% nodes, values, NA))

        colnames(ndf)[ncol(ndf)] <- node_attr
      }
    }
  }

  # If the length of values supplied is the same
  # as the number of rows, insert those values
  # to a new or existing node attribute column
  if (length(values) == nrow(ndf)) {

    if (node_attr %in% colnames(ndf)) {
      ndf[, which(colnames(ndf) %in%
                         node_attr)] <- values
    }

    if (!(node_attr %in% colnames(ndf))) {
      ndf <-
        cbind(ndf,
              rep(NA_character_, nrow(ndf)))

      ndf[, ncol(ndf)] <-
        ndf[, ncol(ndf)]

      colnames(ndf)[ncol(ndf)] <-
        node_attr

      ndf[, ncol(ndf)] <- values
    }
  }

    # Replace the graph's ndf with the
    # revised version
    graph$nodes_df <- ndf

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph$graph_log <-
      add_action_to_log(
        graph_log = graph$graph_log,
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df))

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }

  graph
}

#' Add nodes from distinct values in data frame columns
#'
#' @description
#'
#' Add new nodes to a graph object of class `dgr_graph` using distinct values
#' from one or more columns in a data frame. The values will serve as node
#' labels and the number of nodes added depends on the number of distinct values
#' found in the specified columns.
#'
#' @inheritParams render_graph
#' @param df A data frame from which values will be taken as new nodes for the
#'   graph.
#' @param columns A character vector of column names or a numeric vector of
#'   column numbers for the data frame supplied in `df`. The distinct values in
#'   these columns will serve as labels for the nodes added to the graph.
#' @param type An optional, single-length character vector that provides a group
#'   identifier for the nodes to be added to the graph.
#' @param keep_duplicates An option to exclude incoming nodes where the labels
#'   (i.e., values found in columns of the specified `df`) match label values
#'   available in the graph's nodes. By default, this is set to `FALSE`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph
#' graph <- create_graph()
#'
#' # Create a data frame from
#' # which several columns have
#' # values designated as graph nodes
#' df <-
#'   data.frame(
#'     col_1 = c("f", "p", "q"),
#'     col_2 = c("q", "x", "f"),
#'     col_3 = c(1, 5, 3),
#'     col_4 = c("a", "v", "h"),
#'     stringsAsFactors = FALSE)
#'
#' # Add nodes from columns `col_1`
#' # and `col_2` from the data frame
#' # to the graph object
#' graph <-
#'   graph |>
#'   add_nodes_from_df_cols(
#'     df = df,
#'     columns = c("col_1", "col_2"))
#'
#' # Show the graph's node data
#' # frame; duplicate labels are
#' # prevented with `keep_duplicates =
#' # FALSE`)
#' graph |> get_node_df()
#'
#' # Add new nodes from columns 3 and 4;
#' # We can specify the columns by their
#' # numbers as well
#' graph <-
#'   graph |>
#'   add_nodes_from_df_cols(
#'     df = df,
#'     columns = 3:4)
#'
#' # Show the graph's node data
#' # frame; note that nodes didn't
#' # get made with columns that
#' # are not character class columns
#' graph |> get_node_df()
#'
#' @family node creation and removal
#'
#' @export
add_nodes_from_df_cols <- function(
    graph,
    df,
    columns,
    type = NULL,
    keep_duplicates = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the df column numbers from which nodes
  # will be generated
  if (inherits(columns, "numeric")) {

    # Verify that the none of the values provided
    # are greater than the number of df columns
    if (max(columns) > ncol(df)) {
      cli::cli_abort(c(
        "One or more of the column numbers exceeds the number of columns in `df`."
      ))
    }
  }

  # Get column numbers from the column names
  # provided and verify that at least 1 column
  # number is returned
  if (inherits(columns, "character")) {
    columns <- which(colnames(df) %in% columns)
    check_number_whole(length(columns), min = 1)
  }

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph |> count_nodes()

  # Isolate the relevant columns in the data frame;
  # Exclude any columns that are not character class
  df <-
    dplyr::as_tibble(df) |>
    dplyr::select(dplyr::all_of(columns) & dplyr::where(is.character))

  # Create an empty `nodes` vector
  nodes <- vector(mode = "character")

  # Obtain a vector of values from each column
  # in the tibble object
  for (i in seq_len(ncol(df))) {
    nodes <-
      c(nodes,
        df[, i] |>
          purrr::flatten_chr() |>
          trimws() |>
          stringr::str_split(" ") |>
          purrr::flatten_chr() |>
          tibble::enframe(name = NULL) |>
          tidyr::drop_na() |>
          dplyr::distinct() |>
          purrr::flatten_chr())
  }

  # Get the unique set of nodes
  nodes <- unique(nodes)

  # If `keep_duplicates` is set to FALSE, exclude
  # duplicate labels from being added to the graph
  if (!keep_duplicates) {
    existing_labels <- graph$nodes_df$label
    nodes <- base::setdiff(nodes, existing_labels)
  }

  # Get the number of nodes
  n <- length(nodes)

  # If there are any unique labels, create an ndf
  # of the correct length
  if (n > 0) {

    if (is.null(type)) {
      new_nodes <-
        create_node_df(
          n = n,
          label = nodes)
    } else {
      new_nodes <-
        create_node_df(
          n = n,
          type = type,
          label = nodes)
    }

    # Renumber the node ID values based on the
    # last node in the graph
    new_nodes[, 1] <- new_nodes[, 1] + graph$last_node

    # Add `new_nodes` ndf to the graph
    graph$nodes_df <-
      dplyr::bind_rows(graph$nodes_df, new_nodes)

    # Update the `last_node` counter
    graph$last_node <- graph$last_node + n
  }

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph |> count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get node attribute values
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get node attribute values for one
#' or more nodes.
#'
#' @inheritParams render_graph
#' @param nodes An optional vector of node IDs for filtering list of nodes
#'   present in the graph or node data frame.
#' @param node_attr The name of the attribute for which to get values.
#'
#' @return A named vector of node attribute values for the attribute given by
#'   `node_attr` by node ID.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 4,
#'     m = 4,
#'     set_seed = 23) %>%
#'   set_node_attrs(
#'     node_attr = value,
#'     values = c(2.5, 8.2, 4.2, 2.4))
#'
#' # Get all of the values from
#' # the `value` node attribute
#' # as a named vector
#' graph %>%
#'   get_node_attrs(
#'     node_attr = value)
#'
#' # To only return node attribute
#' # values for specified nodes,
#' # use the `nodes` argument
#' graph %>%
#'   get_node_attrs(
#'     node_attr = value,
#'     nodes = c(1, 3))
#'
#' @export
get_node_attrs <- function(
    graph,
    node_attr,
    nodes = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  node_attr <- rlang::enquo(node_attr)

  if (rlang::enquo(node_attr) %>%
      rlang::get_expr() %>%
      as.character() %in% c("id", "nodes")) {

    cli::cli_abort(
      "This is not a node attribute.")
  }

  # Extract the node data frame (ndf)
  # from the graph
  ndf <- graph$nodes_df

  if (is.null(nodes)) {

    # Extract the node attribute values
    node_attr_vals <- ndf %>% dplyr::pull(!!node_attr)

    # Extract the node names
    node_names <- ndf$id

    # Assign node names
    names(node_attr_vals) <- node_names
  }

  if (!is.null(nodes)) {

    nodes <- sort(nodes)

    # Filter the ndf by the supplied
    # nodes
    ndf <-
      ndf %>%
      dplyr::filter(id %in% nodes)

    # Extract the node attribute values
    node_attr_vals <- ndf %>% dplyr::pull(!!node_attr)

    # Extract the node names
    node_names <- ndf$id

    # Assign node names
    names(node_attr_vals) <- node_names
  }

  node_attr_vals
}

#' Get count of edge definitions where multiple edges occur
#'
#' @description
#'
#' Get a count of the number of edge definitions (e.g, `1` -> `2`) where there
#' are multiple edges (i.e., more than 1 edge of that definition, having
#' distinct edge ID values). So, for example, if there are 2 edge definitions in
#' the graph that involve 6 separate edge IDs (3 such edge IDs for each of the
#' pairs of nodes), the count will be `2`.
#'
#' @inheritParams render_graph
#'
#' @return A vector with a single, numerical value.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 5,
#'     label = TRUE)
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 4, 4, 3, 5, 1, 3, 4),
#'       to = c(4, 1, 1, 2, 2, 2, 2, 1))
#'
#' # Create a graph with the ndf and edf
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Get the total number of edge
#' # definitions (e.g., `4` -> `1`) where
#' # there are multiple edges (i.e.,
#' # distinct edges with separate edge
#' # ID values)
#' graph %>% get_edge_count_w_multiedge()
#'
#' @export
get_edge_count_w_multiedge <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Check for the number of multiple edges
  # regardless of which definitions these
  # edges have
  multiedge_distinct_edge_def_count <-
    graph$edges_df %>%
    dplyr::select(from, to) %>%
    dplyr::mutate(edge_from_to = paste0(from, "_", to), .keep = "none") %>%
    dplyr::group_by(edge_from_to) %>%
    dplyr::summarize(n = dplyr::n(), .groups = "drop") %>%
    dplyr::filter(n > 1) %>%
    nrow()

  multiedge_distinct_edge_def_count
}

#' Subset a graph series object
#'
#' @description
#'
#' Subsetting a graph series by the graphs' index positions in the graph series
#' or through selection via graphs' date-time attributes.
#'
#' @param graph_series A graph series object of type `dgr_graph_1D`.
#' @param by Either `number`, which allows for subsetting of the graph series by
#'   graph indices, or `time` which for graph series objects of type `temporal`
#'   allows for a subsetting of graphs by a date-time or time range.
#' @param values Where the subsetting of the graph series by to occur via graph
#'   indices (where `by = number`), provide a vector of those indices; when
#'   subsetting by time (where `by = time`), a range of times can be provided as
#'   a vector.
#' @param tz The time zone (`tz`) corresponding to dates or date-time string
#'   provided in `values` (if `by = "date"`).
#'
#' @return A graph series object of type `dgr_graph_1D`.
#'
#' @examples
#' # Create three graphs
#' graph_time_1 <-
#'   create_graph(
#'     graph_name = "graph_with_time_1") |>
#'   set_graph_time(
#'     time = "2015-03-25 03:00",
#'     tz = "GMT")
#'
#' graph_time_2 <-
#'   create_graph(
#'     graph_name = "graph_with_time_2") |>
#'   set_graph_time(
#'     time = "2015-03-26 03:00",
#'     tz = "GMT")
#'
#' graph_time_3 <-
#'   create_graph(
#'     graph_name = "graph_with_time_3") |>
#'   set_graph_time(
#'     time = "2015-03-27 15:00",
#'     tz = "GMT")
#'
#' # Create an empty graph series and add
#' # the graphs
#' series_temporal <-
#'   create_graph_series(
#'     series_type = "temporal") |>
#'   add_graph_to_graph_series(
#'     graph = graph_time_1) |>
#'   add_graph_to_graph_series(
#'     graph = graph_time_2) |>
#'   add_graph_to_graph_series(
#'     graph = graph_time_3)
#'
#' # Subset graph series by sequence
#' series_sequence_subset <-
#'   filter_graph_series(
#'     graph_series = series_temporal,
#'     by = "number",
#'     values = 2)
#'
#' # Get a count of graphs in
#' # the series
#' series_sequence_subset |>
#'   count_graphs_in_graph_series()
#'
#' # Subset graph series by date-time
#' series_time_subset <-
#'   filter_graph_series(
#'     graph_series = series_temporal,
#'     by = "time",
#'     values = c("2015-03-25 12:00",
#'                "2015-03-26 12:00"),
#'     tz = "GMT")
#'
#' # Get a count of graphs in
#' # the series
#' series_time_subset |>
#'   count_graphs_in_graph_series()
#'
#' @export
filter_graph_series <- function(
    graph_series,
    by = "number",
    values,
    tz = NULL
) {

  if (count_graphs_in_graph_series(
    graph_series = graph_series) == 0) {

    return(graph_series)
  }

  if (by == "number") {

    # validate the value provided for 'values'
    if (!inherits(values, "numeric")) {

      return(graph_series)
    }

    indices_in_graph_series <-
      1:count_graphs_in_graph_series(graph_series = graph_series)

    indices_in_subset_value <- values

    graphs_to_remove <-
      which(!(indices_in_graph_series %in%
                indices_in_subset_value))

    graphs_to_remove <-
      sort(graphs_to_remove, decreasing = TRUE)

    for (i in graphs_to_remove) {

      graph_series <-
        remove_graph_from_graph_series(
          graph_series = graph_series,
          index = i)
    }

    return(graph_series)
  }

  if (by == "time") {

    # validate the value provided for 'values'
    if (inherits(values, "numeric")) {

      return(graph_series)
    }

    is_tz_in_correct_format <-
      tz %in% OlsonNames()

    if (!is_tz_in_correct_format) {

      return(graph_series)
    }

    for (i in seq_along(values)) {

      is_time_in_correct_format <-
        grepl("^[0-9]{4}-[0-9]{2}-[0-9]{2}$", values[i]) ||
        grepl("^[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}$", values[i]) ||
        grepl("^[0-9]{4}-[0-9]{2}-[0-9]{2} [0-9]{2}:[0-9]{2}:[0-9]{2}$", values[i])

      if (!is_time_in_correct_format) {

        return(graph_series)
      }
    }

    # Create subset based on range
    if (length(values) == 2) {

      for (i in 1:count_graphs_in_graph_series(graph_series = graph_series)) {

        if (i == 1) {
          dates_times_in_series <- vector(mode = "numeric", length = 0)
          tz_in_series <- vector(mode = "character", length = 0)
        }

        dates_times_in_series <-
          c(dates_times_in_series, graph_series$graphs[[i]]$graph_time)

        tz_in_series <-
          c(tz_in_series, graph_series$graphs[[i]]$graph_tz)

      }

      for (i in 1:length(dates_times_in_series)) {

        if (i == 1) {
          dates_times_in_series_with_tz <-
            as.POSIXct(rep("1970-01-01", length(dates_times_in_series)),
                       tz = "GMT")
        }

        dates_times_in_series_with_tz[i] <-
          as.POSIXct(dates_times_in_series[i], tz = tz_in_series[i],
                     origin = "1970-01-01")
      }

      graphs_to_remove <-
        which(!(1:length(dates_times_in_series_with_tz) %in%
                  which(dates_times_in_series_with_tz >=
                          as.POSIXct(values[1], tz = tz) &
                          dates_times_in_series_with_tz <=
                          as.POSIXct(values[2], tz = tz))))

      graphs_to_remove <-
        sort(graphs_to_remove, decreasing = TRUE)

      # Remove selected graphs from the series
      for (i in graphs_to_remove) {

        graph_series <-
          remove_graph_from_graph_series(
            graph_series = graph_series,
            index = i)
      }

      return(graph_series)
    }
  }
}

#' Combine multiple node data frames
#'
#' @description
#'
#' Combine several node data frames into a single node data frame.
#'
#' @param ... Two or more node data frames, which contain node IDs and
#'   associated attributes.
#'
#' @return A combined node data frame.
#'
#' @examples
#' # Create two node data frames
#' node_df_1 <-
#'   create_node_df(
#'     n = 2,
#'     type = c("a", "b"),
#'     label = c("D", "Z"),
#'     value = c(8.4, 3.4))
#'
#' node_df_2 <-
#'   create_node_df(
#'     n = 2,
#'     type = c("b", "c"),
#'     label = c("U", "A"),
#'     value = c(0.4, 3.4))
#'
#' # Combine the ndfs using the
#' # `combine_ndfs()` function
#' node_df_combined <-
#'   combine_ndfs(
#'     node_df_1,
#'     node_df_2)
#'
#' # Inspect the combined ndf
#' node_df_combined
#'
#' @export
combine_ndfs <- function(...) {

  ndfs <- list(...)

  for (l in seq_along(ndfs)) {
    if (l == 1) {
      df1 <- ndfs[l][[1]]
      df2 <- ndfs[l + 1][[1]]
    }

    if (l > 1 && l < length(ndfs)) {
      df1 <- ndf_new
      df2 <- ndfs[l + 1][[1]]
    }

    # Bind rows from df1 and df2
    ndf_new <- dplyr::bind_rows(df1, df2)

    if (l == length(ndfs)) {
      break
    }
  }

  # Create montonically-increasing integer id
  # values for new table based on input order
  ndf_new[, 1] <- seq_len(nrow(ndf_new))

  ndf_new
}

# Creating and combining node data frames

test_that("a correct node data frame is generated", {
  # Create 'nodes_1' node data frame
  nodes_1 <-
    create_node_df(
      n = 4,
      label = NULL,
      type = "lower",
      style = "filled",
      color = "aqua",
      shape = c("circle", "circle",
                "rectangle", "rectangle"),
      data = c(3.5, 2.6, 9.4, 2.7)
    )

  # Create 'nodes_2' node data frame
  nodes_2 <-
    create_node_df(
      n = 4,
      label = NULL,
      type = "upper",
      style = "filled",
      color = "red",
      shape = "triangle",
      data = c(0.5, 3.9, 3.7, 8.2)
    )

  # Expect that each of the node data frames has 4 rows
  expect_equal(nrow(nodes_1), 4)
  expect_equal(nrow(nodes_2), 4)

  # Expect that each of the node data frames has 7 columns
  expect_equal(ncol(nodes_1), 7)
  expect_equal(ncol(nodes_2), 7)

  # Expect that 'label = NULL' produces blank label columns
  expect_equal(nodes_1$label, rep(NA_character_, 4))
  expect_equal(nodes_2$label, rep(NA_character_, 4))

  # Expect that a single value repeats across rows
  expect_equal(nodes_1$type, rep("lower", 4))
  expect_equal(nodes_2$color, rep("red", 4))

  # Expect that the numeric `data` values are numeric
  expect_type(nodes_1$data, "double")
  expect_type(nodes_2$data, "double")

  # Create a node data frame using a vector with length > 1 and
  # length < length(from | to)
  nodes_var_1 <-
    create_node_df(n = 4,
                   color = c("green", "green"))

  # Expect that a data frame is generated
  expect_s3_class(nodes_var_1, "data.frame")

  # Expect that the data frame has 4 rows
  expect_equal(nrow(nodes_var_1), 4)

  # Expect that the 'color' attribute is only written twice and not
  # repeated down
  expect_equal(nodes_var_1$color, c("green", "green", "", ""))

  # Create a node data frame using a vector with
  # length > length(from | to)
  nodes_var_2 <-
    create_node_df(
      n = 4,
      color = c("green", "green",
                "green", "green",
                "green", "green")
    )

  # Expect that a data frame is generated
  expect_s3_class(nodes_var_2, "data.frame")

  # Expect that the data frame has 4 rows
  expect_equal(nrow(nodes_var_2), 4)

  # Expect that the `color` attribute is trimmed to
  # the correct number of rows
  expect_equal(nodes_var_2$color,
               c("green", "green", "green", "green"))

  # Expect an error if not supplying an integer or
  # supplying 2 values for `n`
  expect_snapshot(error = TRUE, {
    create_node_df(n = "a")
    create_node_df(n = c(1, 2))
  })

  # Create a node data frame using a vector with
  # `type` having length > n
  nodes_var_3 <-
    create_node_df(n = 4,
                   type = c("a", "a", "b", "a", "c"))

  # Expect the `type` values to be the first four
  expect_equal(nodes_var_3$type, c("a", "a", "b", "a"))
})

test_that("node data frames can be successfully combined", {
  # Create `ndf_1` node data frame
  ndf_1 <-
    create_node_df(
      n = 4,
      label = NULL,
      type = "lower",
      style = "filled",
      color = "aqua",
      shape = c("circle", "circle",
                "rectangle", "rectangle"),
      data = c(3.5, 2.6, 9.4, 2.7)
    )

  # Create 'nodes_2' node data frame
  ndf_2 <-
    create_node_df(
      n = 4,
      label = NULL,
      type = "upper",
      style = "filled",
      color = "red",
      shape = "triangle",
      data = c(0.5, 3.9, 3.7, 8.2)
    )

  # Combine the 2 node data frames
  all_nodes <- combine_ndfs(ndf_1, ndf_2)

  # Expect that the combined node data frame has 8 rows and 7 columns
  expect_equal(dim(all_nodes), c(8, 7))

  # Expect that the 'label' columns has spaces for labels
  expect_equal(all_nodes$label, rep(NA_character_, 8))

  # Expect that the rows combined correctly
  expect_equal(all_nodes[, 1],
               c(1, 2, 3, 4, 5, 6, 7, 8))
  expect_equal(c(ndf_1[, 2], ndf_2[, 2]),
              all_nodes[, 2])
  expect_equal(c(ndf_1[, 3], ndf_2[, 3]),
               all_nodes[, 3])
  expect_equal(c(ndf_1[, 4], ndf_2[, 4]),
               all_nodes[, 4])
  expect_equal(c(ndf_1[, 5], ndf_2[, 5]),
               all_nodes[, 5])
  expect_equal(c(ndf_1[, 6], ndf_2[, 6]),
               all_nodes[, 6])
  expect_equal(c(ndf_1[, 7], ndf_2[, 7]),
              all_nodes[, 7])
})

#' Delete one or more graph actions stored within a graph object
#'
#' @description
#'
#' Delete one or more graph actions stored within a graph object of class
#' `dgr_graph`).
#'
#' @inheritParams render_graph
#' @param actions Either a vector of integer numbers indicating which actions to
#'   delete (based on `action_index` values), or, a character vector
#'   corresponding to `action_name` values.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 5,
#'     m = 8,
#'     set_seed = 23)
#'
#' # Add three graph actions to the
#' # graph
#' graph <-
#'   graph |>
#'   add_graph_action(
#'     fcn = "set_node_attr_w_fcn",
#'     node_attr_fcn = "get_pagerank",
#'     column_name = "pagerank",
#'     action_name = "get_pagerank") |>
#'   add_graph_action(
#'     fcn = "rescale_node_attrs",
#'     node_attr_from = "pagerank",
#'     node_attr_to = "width",
#'     action_name = "pagerank_to_width") |>
#'   add_graph_action(
#'     fcn = "colorize_node_attrs",
#'     node_attr_from = "width",
#'     node_attr_to = "fillcolor",
#'     action_name = "pagerank_fillcolor")
#'
#' # View the graph actions for the graph
#' # object by using the `get_graph_actions()`
#' # function
#' graph |> get_graph_actions()
#'
#' # Delete the second and third graph
#' # actions using `delete_graph_actions()`
#' graph <-
#'   graph |>
#'   delete_graph_actions(
#'     actions = c(2, 3))
#'
#' # Verify that these last two graph
#' # actions were deleted by again using
#' # the `get_graph_actions()` function
#' graph |> get_graph_actions()
#'
#' @export
delete_graph_actions <- function(
    graph,
    actions
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Determine whether there any
  # available graph actions
  if (nrow(graph$graph_actions) == 0) {

    cli::cli_abort("There are no graph actions to delete")
  }

  if (inherits(actions, "character")) {

    graph_action_names <-
      graph |>
      get_graph_actions() |>
      dplyr::pull("action_name")

    if (!any(actions %in% graph_action_names)) {

      cli::cli_abort(
        "One or more provided `actions` do not exist in the graph.")
    }

    # Get a revised data frame with graph actions
    revised_graph_actions <-
      graph |>
      get_graph_actions() |>
      dplyr::filter(!(action_name %in% actions)) |>
      dplyr::mutate(action_index = dplyr::row_number())
  }

  if (inherits(actions, "numeric")) {

    graph_action_indices <-
      graph |>
      get_graph_actions() |>
      dplyr::pull(action_index)

    if (!any(actions %in% graph_action_indices)) {
      cli::cli_abort("One or more provided `actions` do not exist in the graph.")
    }

    # Get a revised data frame with graph actions
    revised_graph_actions <-
      graph |>
      get_graph_actions() |>
      dplyr::filter(!(action_index %in% actions)) |>
      dplyr::mutate(action_index = dplyr::row_number())
  }

  # Replace `graph$graph_actions` with the
  # revised version
  graph$graph_actions <- revised_graph_actions

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Delete all nodes in a node selection
#'
#' @description
#'
#' In a graph object of class `dgr_graph`, delete all nodes present in a
#' selection of nodes.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 3 nodes
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(n = 3) |>
#'   add_edges_w_string(
#'     edges = "1->3 1->2 2->3")
#'
#' # Select node with ID `1`
#' graph <-
#'   graph |>
#'   select_nodes_by_id(nodes = 1)
#'
#' # Delete node in selection (this
#' # also deletes any attached edges)
#' graph <-
#'   graph |>
#'   delete_nodes_ws()
#'
#' # Get a count of nodes in the graph
#' graph |> count_nodes()
#'
#' @family node creation and removal
#'
#' @export
delete_nodes_ws <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, "So, no node can be deleted.")

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph |> count_nodes()

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Get a vector of the nodes to be deleted
  nodes_to_delete <- graph$node_selection$node

  # Delete all nodes in `nodes_to_delete`
  for (i in seq_along(nodes_to_delete)) {

    graph <-
      delete_node(
        graph = graph,
        node = nodes_to_delete[i])
  }

  # Replace `graph$node_selection` with an empty df
  graph$node_selection <- create_empty_nsdf()

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Remove any `delete_node` records from the graph log
  graph$graph_log <-
    graph$graph_log[-((nrow(graph$graph_log) - (i - 1)):nrow(graph$graph_log)), ]

  # Scavenge any invalid, linked data frames
  graph <-
    graph |>
    remove_linked_dfs()

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph |> count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_deleted <- nodes_graph_2 - nodes_graph_1

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph |> count_edges()

  # Get the number of edges added to
  # the graph
  edges_deleted <- edges_graph_2 - edges_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_deleted,
      d_e = edges_deleted)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get the graph reciprocity
#'
#' @description
#'
#' Get the reciprocity of a directed graph. The reciprocity of a graph is the
#' fraction of reciprocal edges (e.g., `1` -> `2` and `2` -> `1`) over all edges
#' available in the graph. Note that for an undirected graph, all edges are
#' reciprocal. This function does not consider loop edges (e.g., `1` -> `1`).
#'
#' @inheritParams render_graph
#'
#' @return A single, numerical value that is the ratio value of reciprocal edges
#'   over all graph edges.
#'
#' @examples
#' # Define a graph where 2 edge definitions
#' # have pairs of reciprocal edges
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 3) |>
#'   add_node(
#'     from = 1,
#'       to = 1) |>
#'   add_node(
#'     from = 1,
#'       to = 1)
#'
#' # Get the graph reciprocity, which will
#' # be calculated as the ratio 4/7 (where
#' # 4 is the number reciprocating edges
#' # and 7 is the total number of edges
#' # in the graph)
#' graph |>
#'   get_reciprocity()
#'
#' # For an undirected graph, all edges
#' # are reciprocal, so the ratio will
#' # always be 1
#' graph |>
#'   set_graph_undirected() |>
#'   get_reciprocity()
#'
#' # For a graph with no edges, the graph
#' # reciprocity cannot be determined (and
#' # the same NA result is obtained from an
#' # empty graph)
#' create_graph() |>
#'   add_n_nodes(n = 5) |>
#'   get_reciprocity()
#'
#' @export
get_reciprocity <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph contains no edges, it
  # cannot return any valid reciprocity
  # value
  if (nrow(graph$edges_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the reciprocity value for the graph
  igraph::reciprocity(ig_graph)
}

#' Fully disconnect all nodes in a selection of nodes
#'
#' With a selection of nodes in a graph, remove any edges to or from those
#' nodes.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph and
#' # add a path of 6 nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 6)
#'
#' # Select nodes `3` and `4`
#' # and fully disconnect them
#' # from the graph
#' graph <-
#'   graph %>%
#'   select_nodes_by_id(
#'     nodes = 3:4) %>%
#'   fully_disconnect_nodes_ws()
#'
#' # Get the graph's edge data frame
#' graph %>% get_edge_df()
#'
#' @export
fully_disconnect_nodes_ws <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, extra_msg = "So, there are no nodes to disconnect.")

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # Get the graph's edf
  edf <- graph$edges_df

  # Get the number of edges in the graph
  edges_graph_1 <- graph %>% count_edges()

  # Filter edf such that any edges containing
  # nodes in the node selection are removed
  edf_replacement <-
    edf %>%
    dplyr::filter(
      !(from %in% suppressMessages(get_selection(graph)) |
          to %in% suppressMessages(get_selection(graph))))

  # Update the graph's edf
  graph$edges_df <- edf_replacement

  # Scavenge any invalid, linked data frames
  graph <- graph %>% remove_linked_dfs()

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_removed <- edges_graph_2 - edges_graph_1

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_removed)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Select the last set of edges created in a graph
#'
#' @description
#'
#' Select the last edges that were created in a graph object of class
#' `dgr_graph`. This function should ideally be used just after creating the
#' edges to be selected.
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph and add a cycle and then
#' # a tree in 2 separate function calls
#' graph <-
#'   create_graph() %>%
#'   add_cycle(
#'     n = 3,
#'     rel = "a") %>%
#'   add_balanced_tree(
#'     k = 2, h = 2,
#'     rel = "b")
#'
#' # Select the last edges created (all edges
#' # from the tree) and then set their edge
#' # color to be `red`
#' graph <-
#'   graph %>%
#'   select_last_edges_created() %>%
#'   set_edge_attrs_ws(
#'     edge_attr = color,
#'     value = "red") %>%
#'   clear_selection()
#'
#' # Display the graph's internal edge
#' # data frame to verify the change
#' graph %>% get_edge_df()
#'
#' @export
select_last_edges_created <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  graph_transform_steps <-
    graph$graph_log %>%
    dplyr::mutate(
      step_created_edges = as.integer(function_used %in% edge_creation_functions()),
      step_deleted_edges = as.integer(function_used %in% edge_deletion_functions()),
      step_init_with_edges = as.integer(function_used %in% graph_init_functions() &
                                          edges > 0)
    ) %>%
    dplyr::filter(
      dplyr::if_any(
        .cols = c(step_created_edges, step_deleted_edges, step_init_with_edges),
        .fns = function(x) x == 1
      )
    ) %>%
    dplyr::select(-"version_id", -"time_modified", -"duration")

  if (nrow(graph_transform_steps) > 0) {

    if (graph_transform_steps %>%
        utils::tail(1) %>%
        dplyr::pull(step_deleted_edges) == 1) {

      cli::cli_abort(
        "The previous graph transformation function resulted in a removal of edges.")

    } else {
      if (nrow(graph_transform_steps) > 1) {
        number_of_edges_created <-
          (graph_transform_steps %>%
             dplyr::select(edges) %>%
             utils::tail(2) %>%
             dplyr::pull(edges))[2] -
          (graph_transform_steps %>%
             dplyr::select(edges) %>%
             utils::tail(2) %>%
             dplyr::pull(edges))[1]
      } else {
        number_of_edges_created <-
          graph_transform_steps %>%
          dplyr::pull(edges)
      }
    }

    edge_id_values <-
      graph$edges_df %>%
      dplyr::select("id") %>%
      utils::tail(number_of_edges_created) %>%
      dplyr::pull(id)
  } else {
    edge_id_values <- NA
  }

  if (!anyNA(edge_id_values)) {

    # Apply the selection of edges to the graph
    graph <-
      suppressMessages(
        select_edges_by_edge_id(
          graph = graph,
          edges = edge_id_values)
      )

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph$graph_log <-
      graph$graph_log[-nrow(graph$graph_log), ] %>%
      add_action_to_log(
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df))

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }
  }

  graph
}

#' Mutate node attribute values for a selection of nodes
#'
#' @description
#'
#' Within a graph's internal node data frame (ndf), mutate node attribute values
#' only for nodes in a selection by using one or more expressions.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param ... Expressions used for the mutation of node attributes. LHS of each
#'   expression is either an existing or new node attribute name. The RHS can
#'   consist of any valid R code that uses node attributes as variables.
#'   Expressions are evaluated in the order provided, so, node attributes
#'   created or modified are ready to use in subsequent expressions.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 3 nodes
#' # and then select node `1`
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 3) %>%
#'   set_node_attrs(
#'     node_attr = width,
#'     values = c(1.4, 0.3, 1.1)) %>%
#'   select_nodes(nodes = 1)
#'
#' # Get the graph's internal ndf
#' # to show which node attributes
#' # are available
#' graph %>% get_node_df()
#'
#' # Mutate the `width` node
#' # attribute for the nodes
#' # only in the active selection
#' # of nodes (node `1`); here,
#' # we divide each value in the
#' # selection by 2
#' graph <-
#'   graph %>%
#'   mutate_node_attrs_ws(
#'     width = width / 2)
#'
#' # Get the graph's internal
#' # ndf to show that the node
#' # attribute `width` was
#' # mutated only for node `1`
#' graph %>% get_node_df()
#'
#' # Create a new node attribute,
#' # called `length`, that is the
#' # log of values in `width` plus
#' # 2 (and, also, round all values
#' # to 2 decimal places)
#' graph <-
#'   graph %>%
#'   clear_selection() %>%
#'   select_nodes(nodes = 2:3) %>%
#'   mutate_node_attrs_ws(
#'     length = (log(width) + 2) %>%
#'                round(2))
#'
#' # Get the graph's internal ndf
#' # to show that the node attribute
#' # values had been mutated only
#' # for nodes `2` and `3` (since
#' # node `1` is excluded, an NA
#' # value is applied)
#' graph %>% get_node_df()
#'
#' # Create a new node attribute
#' # called `area`, which is the
#' # product of the `width` and
#' # `length` attributes
#' graph <-
#'   graph %>%
#'   mutate_node_attrs_ws(
#'     area = width * length)
#'
#' # Get the graph's internal ndf
#' # to show that the node attribute
#' # values had been multiplied
#' # together (with new attr `area`)
#' # for nodes `2` and `3`
#' graph %>% get_node_df()
#'
#' # We can invert the selection
#' # and mutate node `1` several
#' # times to get an `area` value
#' # for that node
#' graph <-
#'   graph %>%
#'   invert_selection() %>%
#'   mutate_node_attrs_ws(
#'     length = (log(width) + 5) %>%
#'                round(2),
#'     area = width * length)
#'
#' # Get the graph's internal ndf
#' # to show that the 2 mutations
#' # occurred for node `1`, yielding
#' # non-NA values for its node
#' # attributes without changing
#' # those of the other nodes
#' graph %>% get_node_df()
#'
#' @family node creation and removal
#'
#' @export
mutate_node_attrs_ws <- function(
    graph,
    ...
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph object has valid node selection
  check_graph_contains_node_selection(graph)

  # Collect expressions
  exprs <- rlang::exprs(...)

  # Extract the graph's ndf
  ndf <- get_node_df(graph)

  # Stop function if any supplied
  # expressions mutate columns that
  # should not be changed
  if ("id" %in% names(exprs)) {

    cli::cli_abort("The variable `id` cannot undergo mutation.")
  }

  # Determine which nodes are not
  # in the active selection
  unselected_nodes <-
    base::setdiff(
      get_node_ids(graph),
      suppressMessages(get_selection(graph)))

  s <- ndf$id %in% unselected_nodes
  order <- c(which(!s), which(s))

  ndf <-
    dplyr::bind_rows(
      ndf %>% dplyr::filter(!(!!enquo(s))) %>% dplyr::mutate(!!!enquos(...)),
      ndf %>% dplyr::filter(!!enquo(s))
    ) %>%
    dplyr::arrange(!!enquo(order))

  graph$nodes_df <- ndf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Set graph date-time and timezone
#'
#' @description
#'
#' Set the time and timezone for a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#' @param time The date-time to set for the graph.
#' @param tz The timezone to set for the graph.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph
#' graph <- create_graph()
#'
#' # Provide the new graph with a timestamp (if `tz`
#' # is not supplied, `GMT` is used as the time zone)
#' graph_1 <-
#'   graph %>%
#'   set_graph_time(time = "2015-10-25 15:23:00")
#'
#' # Provide the new graph with a timestamp that is
#' # the current time; the time zone is inferred from
#' # the user's locale
#' graph_2 <-
#'   graph %>%
#'   set_graph_time()
#'
#' # The time zone can be updated when a timestamp
#' # is present
#' graph_2 <-
#'   graph_2 %>%
#'   set_graph_time(tz = "America/Los_Angeles")
#'
#' @export
set_graph_time <- function(
    graph,
    time = NULL,
    tz = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  if (is.null(time) && is.null(tz)) {
    time <- format(Sys.time(), "%Y-%m-%d %H:%M:%S")
    tz <- Sys.timezone()
  }

  if (inherits(time[1], "POSIXct")) {

    if (!is.null(attr(time, "tzone"))) {
      tz <- attr(time, "tzone")
    }

    time <- format(time, "%Y-%m-%d %H:%M:%S")
  }

  if (!is.null(tz) && !(tz %in% OlsonNames())) {
      cli::cli_abort(
        "The time zone provided must be available in `OlsonNames()`."
      )
  }

  if (is.null(tz) && is.null(graph$graph_tz)) {
    tz <- "GMT"
  }

  # Modify the graph's time and time zone attributes
  graph$graph_info$graph_time[1] <- time
  graph$graph_info$graph_tz[1] <- tz


  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Apply colors based on edge attribute values
#'
#' @description
#'
#' Within a graph's internal edge data frame (edf), use a categorical edge
#' attribute to generate a new edge attribute with color values.
#'
#' @inheritParams render_graph
#' @param edge_attr_from The name of the edge attribute column from which color
#'   values will be based.
#' @param edge_attr_to The name of the new edge attribute to which the color
#'   values will be applied.
#' @param cut_points An optional vector of numerical breaks for bucketizing
#'   continuous numerical values available in a edge attribute column.
#' @param palette Can either be: (1) a palette name from the RColorBrewer
#'   package (e.g., `Greens`, `OrRd`, `RdYlGn`), (2) `viridis`, which indicates
#'   use of the `viridis` color scale from the package of the same name, or (3)
#'   a vector of hexadecimal color names.
#' @param alpha An optional alpha transparency value to apply to the generated
#'   colors. Should be in the range of `0` (completely transparent) to `100`
#'   (completely opaque).
#' @param reverse_palette An option to reverse the order of colors in the chosen
#'   palette. The default is `FALSE`.
#' @param default_color A hexadecimal color value to use for instances when the
#'   values do not fall into the bucket ranges specified in the `cut_points`
#'   vector.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 5
#' # nodes and 4 edges
#' graph <-
#'   create_graph() |>
#'   add_path(n = 5) |>
#'   set_edge_attrs(
#'     edge_attr = weight,
#'     values = c(3.7, 6.3, 9.2, 1.6))
#'
#' # We can bucketize values in
#' # the edge `weight` attribute using
#' # `cut_points` and, by doing so,
#' # assign colors to each of the
#' # bucketed ranges (for values not
#' # part of any bucket, a gray color
#' # is assigned by default)
#' graph <-
#'   graph |>
#'   colorize_edge_attrs(
#'     edge_attr_from = weight,
#'     edge_attr_to = color,
#'     cut_points = c(0, 2, 4, 6, 8, 10),
#'     palette = "RdYlGn")
#'
#' # Now there will be a `color`
#' # edge attribute with distinct
#' # colors (from the RColorBrewer
#' # Red-Yellow-Green palette)
#' graph |> get_edge_df()
#'
#' @export
colorize_edge_attrs <- function(
    graph,
    edge_attr_from,
    edge_attr_to,
    cut_points = NULL,
    palette = "Spectral",
    alpha = NULL,
    reverse_palette = FALSE,
    default_color = "#D9D9D9"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the requested `edge_attr_from`
  edge_attr_from <-
    rlang::enquo(edge_attr_from) |> rlang::get_expr() |> as.character()

  # Get the requested `edge_attr_to`
  edge_attr_to <-
    rlang::enquo(edge_attr_to) |> rlang::get_expr() |> as.character()

  # Extract edf from graph
  edges_df <- graph$edges_df

  # Get the column number in the edf from which to
  # recode values
  col_to_recode_no <-
    which(colnames(edges_df) %in% edge_attr_from)

  # Get the number of recoded values
  if (is.null(cut_points)) {
    num_recodings <-
      nrow(unique(edges_df[col_to_recode_no]))
  } else {
    num_recodings <- length(cut_points) - 1
  }

  # Handle vector of hexadecimal or named colors
  if (length(palette) > 1) {
    # Verify colors are valid
    is_valid_hex <- grepl(toupper(palette), pattern = "#[0-9A-F]{6}")
    if (!all(is_valid_hex)) {

      cli::cli_abort(
        "The color palette contains invalid hexadecimal values.")
    }

    if (length(palette) < num_recodings) {
      # Revert to viridis if provided color vector is too short
      palette <- "viridis"
    } else {
      color_palette <- toupper(palette)[seq_len(num_recodings)]
    }
  }

  # Handle viridis and ColorBrewer palette name input
  if (length(palette) == 1) {
    # If the number of recodings lower than any Color
    # Brewer palette, shift palette to `viridis`
    if ((num_recodings < 3 || num_recodings > 10) && palette %in%
        c(row.names(RColorBrewer::brewer.pal.info))) {
      palette <- "viridis"
    }

    # or any of the RColorBrewer palettes
    if (!(palette %in% c(row.names(RColorBrewer::brewer.pal.info),
                         "viridis"))) {
      cli::cli_abort(c(
        "The color palette is not an `RColorBrewer` or `viridis` palette.",
        i = "See {.topic RColorBrewer::brewer.pal.info}."
      ))
    }

    # Obtain a color palette
    if (palette %in% row.names(RColorBrewer::brewer.pal.info)) {
      color_palette <- RColorBrewer::brewer.pal(num_recodings, palette)
    } else if (palette == "viridis") {
      color_palette <- viridisLite::viridis(num_recodings)
      color_palette <- gsub("..$", "", color_palette)
    }
  }

  # Reverse color palette if `reverse_palette = TRUE`
  if (reverse_palette) {
    color_palette <- rev(color_palette)
  }

  # Create a data frame with initial values
  new_edge_attr_col <-
    data.frame(
      edge_attr_to = rep(default_color, nrow(edges_df)),
      stringsAsFactors = FALSE)

  # Get the column number for the new edge attribute
  to_edge_attr_colnum <- ncol(edges_df) + 1

  # Bind the current edf with the new column
  edges_df <- cbind(edges_df, new_edge_attr_col)

  # Rename the new column with the target edge attr name
  colnames(edges_df)[to_edge_attr_colnum] <- edge_attr_to

  # Get a data frame of recodings
  if (is.null(cut_points)) {

    recode_df <-
      data.frame(
        to_recode = names(table(edges_df[, col_to_recode_no])),
        colors = color_palette,
        stringsAsFactors = FALSE)

    # Recode rows in the new edge attribute
    for (i in seq_along(names(table(edges_df[, col_to_recode_no])))) {

      recode_rows <-
        which(edges_df[, col_to_recode_no] %in%
                recode_df[i, 1])

      if (is.null(alpha)) {
        edges_df[recode_rows, to_edge_attr_colnum] <-
          color_palette[i]
      } else if (!is.null(alpha)) {
        if (alpha < 100) {
          edges_df[recode_rows, to_edge_attr_colnum] <-
            gsub("$", alpha, color_palette[i])
        } else if (alpha == 100) {
          edges_df[recode_rows, to_edge_attr_colnum] <-
            gsub("$", "", color_palette[i])
        }
      }
    }
  }

  # Recode according to provided cut points
  if (!is.null(cut_points)) {
    for (i in seq_len(length(cut_points) - 1)) {
      recode_rows <-
        which(
          as.numeric(edges_df[, col_to_recode_no]) >=
            cut_points[i] &
            as.numeric(edges_df[, col_to_recode_no]) <
            cut_points[i + 1])

      edges_df[recode_rows, to_edge_attr_colnum] <-
        color_palette[i]
    }

    if (!is.null(alpha)) {
      if (alpha < 100) {
        edges_df[, to_edge_attr_colnum] <-
          gsub("$", alpha, edges_df[, to_edge_attr_colnum])
      } else if (alpha == 100) {
        edges_df[, to_edge_attr_colnum] <-
          gsub("$", "", edges_df[, to_edge_attr_colnum])
      }
    }
  }

  # Get the finalized column of values
  edges_attr_vector_colorized <- edges_df[, ncol(edges_df)]

  edge_attr_to_2 <- rlang::enquo(edge_attr_to)

  # Set the edge attribute values for nodes specified
  # in selection
  graph <-
    set_edge_attrs(
      graph = graph,
      edge_attr = !!edge_attr_to_2,
      values = edges_attr_vector_colorized
    )

  # Remove last action from the `graph_log`
  graph$graph_log <- graph$graph_log[1:(nrow(graph$graph_log) - 1), ]

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Is the edge a multiple edge?
#'
#' @description
#'
#' Determines whether an edge definition has multiple edge IDs associated with
#' the same node pair.
#'
#' @inheritParams render_graph
#' @param edge A numeric edge ID value.
#'
#' @return A logical value.
#'
#' @examples
#' # Create a graph that has multiple
#' # edges across some node pairs
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 4) %>%
#'   add_edge(
#'     from = 1,
#'     to = 2) %>%
#'   add_edge(
#'     from = 3,
#'     to = 4)
#'
#' # Get the graph's internal
#' # edge data frame
#' graph %>% get_edge_df()
#'
#' # Determine if edge `1` is
#' # a multiple edge
#' graph %>%
#'   is_edge_multiple(edge = 1)
#'
#' # Determine if edge `2` is
#' # a multiple edge
#' graph %>%
#'   is_edge_multiple(edge = 2)
#'
#' @export
is_edge_multiple <- function(
    graph,
    edge
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Stop function if `edge` ID is not a single numeric
  # possible number_whole?
  check_number_decimal(edge)

  # Obtain the graph's edf
  edf <- graph$edges_df

  # Stop function if the edge ID provided
  # is not a valid edge ID
  if (!(edge %in% edf$id)) {

    cli::cli_abort(
      "The provided edge ID ({edge}) is not present in the graph.")
  }

  # Obtain the edge definition
  from <-
    edf %>%
    dplyr::filter(id == !!edge) %>%
    dplyr::pull("from")

  to <-
    edf %>%
    dplyr::filter(id == !!edge) %>%
    dplyr::pull("to")

  # Determine if there are mulitple rows
  # where the definition of `from` and `to`
  # is valid
  multiple_edges <-
    edf %>%
    dplyr::filter(from == !!from, to == !!to)

  res <- nrow(multiple_edges) > 1
  res
}

#' Get the graph girth
#'
#' @description
#'
#' Get the girth of a graph, which is the length of the shortest circle in the
#' graph. Loop edges and multiple edges are not considered. If the graph
#' contains no cycles then zero is returned.
#'
#' @inheritParams render_graph
#'
#' @return A single numeric value representing the length of the shortest circle
#'   in the graph.
#'
#' @examples
#' # Create a cycle graph
#' graph <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' # Determine the graph's girth
#' graph %>% get_girth()
#'
#' # Create a full graph and then
#' # get the girth for that
#' create_graph() %>%
#'   add_full_graph(n = 10) %>%
#'   get_girth()
#'
#' @export
get_girth <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the girth of the graph
  igraph::girth(ig_graph, circle = FALSE)$girth
}

#' Get a cached vector from a graph object
#'
#' @description
#'
#' Get the vector cached in a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#' @param name the name of the object to extract from the cache. If none
#'   supplied, the most recent object added to the cache will be returned.
#'
#' @return A vector.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a graph with 5 nodes and 5 edges
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(n = 5) %>%
#'   set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = count_nodes(.),
#'       mean = 8,
#'       sd = 2)) %>%
#'   add_edges_w_string(
#'     edges = "1->2 1->3 2->4 2->5 3->2")
#'
#' # Cache all values from the node attribute `value`
#' # as a numeric vector
#' graph <-
#'   graph %>%
#'   set_cache(
#'     name = "value",
#'     to_cache = get_node_attrs(
#'       graph = .,
#'       node_attr = value))
#'
#' # Return the cached vector
#' graph %>% get_cache()
#'
#' @export
get_cache <- function(
    graph,
    name = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If there is no cached vector available, return NA
  if (length(graph$cache) == 0) {
    return(NA)
  }

  if (is.null(name)) {
    return(graph$cache[[length(graph$cache)]])
  } else {
    return(graph$cache[[name]])
  }
}

#' Get the eigen centrality for all nodes
#'
#' @description
#'
#' Get the eigen centrality values for all nodes in the graph.
#'
#' @inheritParams render_graph
#' @param weights_attr An optional name of the edge attribute to use in the
#'   adjacency matrix. If `NULL` then, if it exists, the `weight` edge attribute
#'   of the graph will be used. If `NA` then no edge weights will be used.
#'
#' @return A data frame with eigen centrality scores for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10, m = 15,
#'     set_seed = 23)
#'
#' # Get the eigen centrality scores
#' # for nodes in the graph
#' graph |> get_eigen_centrality()
#'
#' @export
get_eigen_centrality <- function(
    graph,
    weights_attr = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  if (inherits(weights_attr, "character")) {
    # Stop function if the edge attribute does not exist
    if (!(weights_attr %in% colnames(graph$edges_df))) {

      cli::cli_abort(
        "The edge attribute to be used as weights does not exist in the graph.")
    }

    # Stop function if the edge attribute is not numeric
    if (!is.numeric(graph$edges_df[, which(colnames(graph$edges_df) == weights_attr)])) {

      cli::cli_abort(
        "The edge attribute to be used as weights is not numeric.")
    }

    weights_attr <- graph$edges_df[, which(colnames(graph$edges_df) == weights_attr)]
  }

  # Get the eigen centrality values for each of the
  # graph's nodes
  eigen_centrality_values <-
    igraph::eigen_centrality(
      graph = ig_graph,
      weights = weights_attr)

  # Create df with eigen centrality values
  data.frame(
    id = names(eigen_centrality_values$vector) |> as.integer(),
    eigen_centrality = unname(eigen_centrality_values$vector) |> round(4),
    stringsAsFactors = FALSE)
}

#' Create a node data frame
#'
#' @description
#'
#' Combine several vectors for nodes and their attributes into a data frame,
#' which can be combined with other similarly-generated data frames, or, added
#' to a graph object. A node data frame, or ndf, has at least the following
#' columns:
#'
#' - `id` (of type `integer`)
#'
#' - `type` (of type `character`)
#'
#' - `label` (of type `character`)
#'
#' An arbitrary number of additional columns containing aesthetic or data
#' attributes can be part of the ndf, see [node_aes()] for additional attributes that can be used with ..., so long as they follow the aforementioned
#' columns.
#'
#' @param n The total number of nodes to include in the node data frame.
#' @param type An optional `type` for each node.
#' @param label An optional `label` for each node.
#' @param ... Additional attributes. Some are present in [node_aes()]
#' @return A node data frame (ndf).
#'
#' @examples
#' # Create a node data frame (ndf) where the labels
#' # are equivalent to the node ID values (this is not
#' # recommended); the `label` and `type` node
#' # attributes will always be a `character` class
#' # whereas `id` will always be an `integer`
#' node_df <-
#'   create_node_df(
#'     n = 4,
#'     type = c("a", "a", "b", "b"),
#'     label = TRUE)
#'
#' # Display the node data frame
#' node_df
#'
#' # Create an ndf with distinct labels and
#' # additional node attributes (where their classes
#' # will be inferred from the input vectors)
#' node_df <-
#'   create_node_df(
#'     n = 4,
#'     type = "a",
#'     label = c(2384, 3942, 8362, 2194),
#'     style = "filled",
#'     color = "aqua",
#'     shape = c("circle", "circle",
#'               "rectangle", "rectangle"),
#'     value = c(3.5, 2.6, 9.4, 2.7))
#'
#' # Display the node data frame
#' node_df
#'
#' @family node creation and removal
#'
#' @export
create_node_df <- function(
    n,
    type = NULL,
    label = NULL,
    ...
) {

  check_number_whole(n)
  # FIXME, nothing stopping ae
  type <- type %||% rep(NA_character_, n)

  # Expand vectors with single values to fill to
  # the number of nodes
  if (length(type) == 1) {
    type <- rep(type, n)
  }

  # Expand vectors with `length` > `1` and
  # `length` < `length(nodes)`
  if (length(type) > 1 &&
      length(type) < n) {
    type <- c(type, rep(NA_character_, (n - length(type))))
  }

  # Trim vectors with number of values exceeding the
  # number of nodes
  if (length(type) > n) {
    type <- type[seq_len(n)]
  }

  # Collect extra vectors of data as `extras`
  extras <- list(...)

  if (length(extras) > 0L) {

    for (i in seq_along(extras)) {

      # Expand vectors with single values to fill to
      # the number of nodes
      if (length(extras[[i]]) == 1) {
        extras[[i]] <- rep(extras[[i]], n)
      }

      # Expand vectors with `length` > `1` and
      # `length` < `length(nodes)`
      if (length(extras[[i]]) > 1 &&
          length(extras[[i]]) < n) {
        extras[[i]] <-
          c(extras[[i]],
            rep("", (n - length(extras[[i]]))))
      }

      # Trim vectors with number of values exceeding
      # the number of nodes
      if (length(extras[[i]]) > n) {
        extras[[i]] <- extras[[i]][seq_len(n)]
      }
    }

    # Create a data frame from the `extras` list
    extras <-
        as.data.frame(
          extras, stringsAsFactors = FALSE)
  }

  # Interpret node label values
  if (is.null(label)) {
    label <- rep(NA_character_, n)
  } else if (rlang::inherits_any(label, c("numeric", "character"))) {
    label <- as.character(label)
  } else if (rlang::is_logical(label, n = 1)) {
    if (label) {
      label <- as.character(seq_len(n))
    } else {
      label <- rep(NA_character_, n)
    }
  }

  if (inherits(extras, "data.frame")) {
    nodes_df <-
      dplyr::bind_cols(
        data.frame(
          id = seq_len(n),
          type = type,
          label = label,
          stringsAsFactors = FALSE),
        extras)

  } else {
    nodes_df <-
      data.frame(
        id = seq_len(n),
        type = type,
        label = label,
        stringsAsFactors = FALSE)
  }

  nodes_df
}

#' Get Jaccard similarity coefficient scores
#'
#' @description
#'
#' Get the Jaccard similarity coefficient scores for one or more nodes in a
#' graph.
#'
#' @inheritParams render_graph
#' @param nodes An optional vector of node IDs to consider for Jaccard
#'   similarity scores. If not supplied, then similarity scores will be provided
#'   for every pair of nodes in the graph.
#' @param direction Using `all` (the default), the function will ignore edge
#'   direction when determining scores for neighboring nodes. With `out` and
#'   `in`, edge direction for neighboring nodes will be considered.
#' @param round_to The maximum number of decimal places to retain for the
#'   Jaccard similarity coefficient scores. The default value is `3`.
#'
#' @return A matrix with Jaccard similarity values for each pair of nodes
#'   considered.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the Jaccard similarity
#' # values for nodes `5`, `6`,
#' # and `7`
#' graph |>
#'   get_jaccard_similarity(
#'     nodes = 5:7)
#'
#' @export
get_jaccard_similarity <- function(
    graph,
    nodes = NULL,
    direction = "all",
    round_to = 3
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Ensure that values provided for the
  # `direction` argument are from the
  # valid options
  arg_match0(direction, c("all", "in", "out"))
  # Stop function if nodes provided not in
  # the graph
  if (!is.null(nodes) && !all(nodes %in% get_node_ids(graph))) {

      cli::cli_abort("One or more nodes provided not in graph.")
    }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  ig_nodes <- igraph::V(ig_graph)

  # Get an igraph representation of node ID values
  if (!is.null(nodes)) {
    ig_nodes <- ig_nodes[nodes]
  }

  # Get the Jaccard similarity scores
  # with mode = either "in", "out", or "all". (direction)
  j_sim_values <-
    as.matrix(
      igraph::similarity(
        graph = ig_graph,
        vids  = ig_nodes,
        mode  = direction,
        method = "jaccard"))

  if (is.null(nodes)) {
    row.names(j_sim_values) <- graph$nodes_df$id
    colnames(j_sim_values) <- graph$nodes_df$id
  }

  if (!is.null(nodes)) {
    row.names(j_sim_values) <- graph$nodes_df$id[nodes]
    colnames(j_sim_values) <- graph$nodes_df$id[nodes]
  }

  # Round all values in matrix to set SD
  round(j_sim_values, round_to)
}

#' Cache a vector in the graph
#'
#' @description
#'
#' Place any vector in the cache of a graph object of class `dgr_graph`.
#'
#' @inheritParams render_graph
#' @param to_cache Any vector or data frame. If a data frame is supplied then a
#'   single column for the vector to pull must be provided in the `col`
#'   argument.
#' @param name An optional name for the cached vector.
#' @param col If a data frame is provided in `to_cache` then a column name from
#'   that data frame must provided here.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 22,
#'     set_seed = 23
#'   )
#'
#' # Get the closeness values for
#' # all nodes from `1` to `10` and
#' # store in the graph's cache
#' graph <-
#'   graph %>%
#'   set_cache(
#'     name = "closeness_vector",
#'     to_cache = get_closeness(.),
#'     col = "closeness"
#'   )
#'
#' # Get the graph's cache
#' graph %>%
#'   get_cache(name = "closeness_vector")
#'
#' # Get the difference of betweenness
#' # and closeness values for nodes in
#' # the graph and store the vector in
#' # the graph's cache
#' graph <-
#'   graph %>%
#'   set_cache(
#'     name = "difference",
#'     to_cache =
#'       get_betweenness(.)$betweenness -
#'         get_closeness(.)$closeness
#'   )
#'
#' # Get the graph's cache
#' graph %>%
#'   get_cache(name = "difference")
#'
#' @export
set_cache <- function(
    graph,
    to_cache,
    name = NULL,
    col = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  if (rlang::inherits_any(to_cache, c("numeric", "integer", "character"))) {

    # Cache vector in the graph's `cache` list object
    if (!is.null(name)) {
      graph$cache[[name]] <- to_cache
    } else {
      if (length(graph$cache) == 0) {
        graph$cache[[1]] <- to_cache
        names(graph$cache) <- 1
      } else {
        graph$cache[[(length(graph$cache) + 1)]] <-
          to_cache
      }
    }
  }

  if (inherits(to_cache, "data.frame")) {

    if (is.null(col)) {

      cli::cli_abort(
        "`col` must be a column name that exists in the data frame.")
    }

    if (!is.null(col)) {

      if (!(col %in% colnames(to_cache))) {

        cli::cli_abort(
          "`col` must be a column name that exists in the data frame.")
      }

      # Extract the vector from the data frame and cache
      # vector in the graph's `cache` list object
      if (!is.null(name)) {
        graph$cache[[name]] <- to_cache[, which(colnames(to_cache) == col)]
      } else {
        if (length(graph$cache) == 0) {
          graph$cache[[1]] <- to_cache[, which(colnames(to_cache) == col)]
          names(graph$cache) <- 1
        } else {
          graph$cache[[(length(graph$cache) + 1)]] <-
            to_cache[, which(colnames(to_cache) == col)]
        }
      }
    }
  }

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Transform a tibble to a data frame
#'
#' Simply uses `as.data.frame()` with an input tibble and ensures that
#' `stringsAsFactors` is `FALSE`.
#' @param tbl A tibble.
#' @noRd
asdf <- function(tbl) {
  tbl |> as.data.frame(stringsAsFactors = FALSE)
}

#' The default attribute theme
#'
#' @noRd
attr_theme_default <- function() {

  dplyr::tribble(
    ~attr,                  ~value,                  ~attr_type,
    "layout",               "neato",                 "graph",
    "outputorder",          "edgesfirst",            "graph",
    "bgcolor",              "white",                 "graph",
    "fontname",             "Helvetica",             "node",
    "fontsize",             "10",                    "node",
    "shape",                "circle",                "node",
    "fixedsize",            "true",                  "node",
    "width",                "0.5",                   "node",
    "style",                "filled",                "node",
    "fillcolor",            "aliceblue",             "node",
    "color",                "gray70",                "node",
    "fontcolor",            "gray50",                "node",
    "fontname",             "Helvetica",             "edge",
    "fontsize",             "8",                     "edge",
    "len",                  "1.5",                   "edge",
    "color",                "gray80",                "edge",
    "arrowsize",            "0.5",                   "edge"
  ) |>
    asdf()
}

#' The lr attribute theme
#'
#' @noRd
attr_theme_lr <- function() {

  dplyr::bind_rows(
    dplyr::tribble(
      ~attr,                  ~value,                  ~attr_type,
      "layout",               "dot",                   "graph",
      "rankdir",              "LR",                    "graph"
    ) |>
      asdf(),
    attr_theme_default()[-1, ])
}

#' The tb attribute theme
#'
#' @noRd
attr_theme_tb <- function() {

  dplyr::bind_rows(
    dplyr::tribble(
      ~attr,                  ~value,                  ~attr_type,
      "layout",               "dot",                   "graph",
      "rankdir",              "TB",                    "graph"
    ) |>
      asdf(),
    attr_theme_default()[-1, ])
}

#' The rl attribute theme
#'
#' @noRd
attr_theme_rl <- function() {

  dplyr::bind_rows(
    dplyr::tribble(
      ~attr,                  ~value,                  ~attr_type,
      "layout",               "dot",                   "graph",
      "rankdir",              "RL",                    "graph"
    ) |>
      asdf(),
    attr_theme_default()[-1, ])
}

#' The bt attribute theme
#'
#' @noRd
attr_theme_bt <- function() {

  dplyr::bind_rows(
    dplyr::tribble(
      ~attr,                  ~value,                  ~attr_type,
      "layout",               "dot",                   "graph",
      "rankdir",              "BT",                    "graph"
    ) |>
      asdf(),
    attr_theme_default()[-1, ])
}

#' The fdp attribute theme
#'
#' @noRd
attr_theme_fdp <- function() {

  dplyr::tribble(
    ~attr,                  ~value,                  ~attr_type,
    "layout",               "fdp",                   "graph",
    "outputorder",          "edgesfirst",            "graph",
    "bgcolor",              "white",                 "graph",
    "fontname",             "Helvetica",             "node",
    "fontsize",             "5",                     "node",
    "shape",                "circle",                "node",
    "fixedsize",            "true",                  "node",
    "width",                "0.12",                  "node",
    "style",                "filled",                "node",
    "fillcolor",            "aliceblue",             "node",
    "color",                "gray70",                "node",
    "fontcolor",            "gray70",                "node",
    "fontname",             "Helvetica",             "edge",
    "fontsize",             "5",                     "edge",
    "len",                  "1.5",                   "edge",
    "color",                "gray80",                "edge",
    "arrowsize",            "0.5",                   "edge"
  ) |>
    asdf()
}

#' The kk attribute theme
#'
#' @noRd
attr_theme_kk <- function() {

  dplyr::bind_rows(
    dplyr::tribble(
      ~attr,                  ~value,                  ~attr_type,
      "layout",               "neato",                 "graph",
      "mode",                 "KK",                    "graph"
    ) |>
      asdf(),
    attr_theme_default()[-1, ])
}

#' Add edges from an edge data frame to an existing graph object
#'
#' @description
#'
#' With a graph object of class `dgr_graph`, add edges from an edge data frame
#' to that graph.
#'
#' @inheritParams render_graph
#' @param edge_df An edge data frame that is created using [create_edge_df()].
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 4 nodes
#' # and no edges
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(n = 4)
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'       to = c(4, 3, 1))
#'
#' # Add the edge data frame to
#' # the graph object to create
#' # a graph with both nodes
#' # and edges
#' graph <-
#'   graph |>
#'   add_edge_df(
#'     edge_df = edf)
#'
#' # Get the graph's edges to
#' # verify that the edf had
#' # been added
#' graph |>
#'   get_edges(
#'     return_type = "vector")
#'
#' @family edge creation and removal
#'
#' @export
add_edge_df <- function(
    graph,
    edge_df
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph, extra_msg = "So, edges cannot be added.")

  # Get the number of edges ever created for
  # this graph
  edges_created <- graph$last_edge

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Combine the incoming edge data frame
  # with those in the graph
  combined_edges <-
    combine_edfs(
      graph$edges_df,
      edge_df)

  # Replace the graph's internal edge
  # data frame with the `combined_edges`
  # edge data frame
  graph$edges_df <- combined_edges

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph |> count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Update the `last_edge` counter
  graph$last_edge <- edges_created + nrow(combined_edges)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get metrics for a graph
#'
#' @description
#'
#' Get a data frame with metrics for a graph.
#'
#' @inheritParams render_graph
#'
#' @return A data frame containing metrics pertaining to the graph
#'
#' @examples
#' \dontrun{
#' # Import a GML graph file available
#' # in the DiagrammeR package
#' karate_club <-
#'   system.file(
#'     "extdata", "karate.gml",
#'     package = "DiagrammeR") %>%
#'   import_graph() %>%
#'   set_graph_name("karate")
#'
#' # Display a data frame with
#' # graph information
#' karate_club %>%
#'   get_graph_info()
#' }
#'
#' @export
get_graph_info <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the graph density
  density <-
    round(count_nodes(graph) / ((count_nodes(graph) * (count_nodes(graph) - 1)) / 2), 4)

  # Get a table of node degree values
  degree_table <-
    table(c(graph$edges_df$from,
            graph$edges_df$to))

  # Create a data frame with the graph metrics
  data.frame(
    name = as.character(graph$graph_info$graph_name),
    n = as.integer(count_nodes(graph)),
    e = as.integer(count_edges(graph)),
    dens = as.numeric(density),
    mn_deg = as.integer(min(degree_table)),
    mx_deg = as.integer(max(degree_table)),
    avg_deg = as.numeric(round(mean(degree_table))),
    time = graph$graph_info$graph_time,
    tz = as.character(graph$graph_info$graph_tz),
    stringsAsFactors = FALSE)
}

#' Render graph with visNetwork
#'
#' @description
#'
#' Render a graph object with the visNetwork R package.
#'
#' @inheritParams render_graph
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 6,
#'     label = TRUE,
#'     fillcolor = c("lightgrey", "red", "orange",
#'                   "pink", "aqua", "yellow"),
#'     shape = "dot",
#'     size = c(20, 80, 40, 10, 30, 50),
#'     type = c("1", "1", "1", "2", "2", "2")
#'   )
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3, 4, 6, 5),
#'     to = c(4, 3, 1, 3, 1, 4),
#'     color = c("green", "green", "grey",
#'               "grey", "blue", "blue"),
#'     rel = "leading_to"
#'   )
#'
#' # Create a graph object
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf
#'   )
#'
#' # visnetwork(graph)
#'
#' @export
visnetwork <- function(graph) {

  # Extract node and edge data frames from the graph object
  nodes <- graph |> get_node_df()
  edges <- graph |> get_edge_df()

  # Render an empty graph if no nodes or edges exist
  if (is_graph_empty(graph)) {

    nodes <- create_node_df(n = 1)
    nodes <- nodes[-1, ]

    edges <- create_edge_df(from = 1, to = 1)
    edges <- edges[-1, ]
  }

  # Remove the 'pos' column, if it exists
  if ("pos" %in% colnames(nodes)) {
    nodes <- nodes[, -(which(colnames(nodes) %in% "pos"))]
  }

  # Modify names of columns in `nodes` for compatibility with
  # visNetwork data frames for nodes
  colnames(nodes)[which(colnames(nodes) == "nodes")] <- "id"
  colnames(nodes)[which(colnames(nodes) == "type")] <- "group"
  colnames(nodes)[which(colnames(nodes) == "tooltip")] <- "title"
  colnames(nodes)[which(colnames(nodes) == "fillcolor")] <- "color"

  # Modify names of columns in 'edges' for compatibility with
  # visNetwork data frames for edges
  colnames(edges)[which(colnames(edges) == "rel")] <- "label"
  colnames(edges)[which(colnames(edges) == "tooltip")] <- "title"
  colnames(edges)[which(colnames(edges) == "penwidth")] <- "width"

  # Obtain `fontcolor` values if the column exists in `edges`
  if ("fontcolor" %in% colnames(edges)) {
    fontcolor <- edges[, -(which(colnames(edges) %in% "fontcolor"))]
  }

  # Create the visNetwork object
  if (!all(c("x", "y") %in% colnames(nodes))) {

    if (nrow(graph$edges_df) == 0) {

      vn_obj <- visNetwork(nodes = nodes)
    }

    if (nrow(graph$edges_df) > 0) {

      vn_obj <- visNetwork(nodes = nodes, edges = edges)

      if (is_graph_directed(graph)) {

        vn_obj <-
          visEdges(
            graph = vn_obj,
            arrows = list(
              to = list(
                enabled = TRUE,
                scaleFactor = 1)))
      }

      if (is_graph_undirected(graph)) {

        vn_obj <-
          visEdges(
            graph = vn_obj,
            arrows = list(
              to = list(
                enabled = FALSE,
                scaleFactor = 1)))
      }
    }

    vn_obj <-
      visPhysics(
        graph = vn_obj,
        solver = "barnesHut",
        stabilization = list(
          enabled = TRUE,
          onlyDynamicEdges = FALSE,
          fit = TRUE))

    vn_obj <-
      visLayout(
        graph = vn_obj,
        improvedLayout = TRUE)
  }

  if (all(c("x", "y") %in% colnames(nodes))) {

    # Reverse y values
    nodes$y <- -as.numeric(nodes$y)

    if (is.null(graph$edges_df)) {

      vn_obj <- visNetwork(nodes = nodes)

      vn_obj <-
        visNodes(
          graph = vn_obj,
          physics = FALSE,
          fixed = FALSE)

      vn_obj <-
        visPhysics(
          graph = vn_obj,
          stabilization = list(
            enabled = FALSE,
            onlyDynamicEdges = FALSE,
            fit = TRUE))

      vn_obj <-
        visInteraction(
          graph = vn_obj,
          dragNodes = FALSE)
    }

    if (nrow(graph$edges_df) > 0) {

      if ("arrow" %in% colnames(edges)) {
        # TODO Check if logic is okay.

        if (!all(edges[which(colnames(edges) %in% "arrow")])) {

          arrows_for_edges <- FALSE
        } else {
          arrows_for_edges <- FALSE
        }
      } else {
        arrows_for_edges <- FALSE
      }

      vn_obj <-
        visNetwork(
          nodes = nodes,
          edges = edges)

      vn_obj <-
        visNodes(
          graph = vn_obj,
          physics = FALSE,
          fixed = FALSE)

      vn_obj <-
        visEdges(
          graph = vn_obj,
          arrows = list(
            to =
              list(
                enabled = arrows_for_edges, # TRUE or FALSE
                scaleFactor = 1)),
          smooth = FALSE,
          font = list(
            color = "#343434",
            size = 14,
            face = "arial",
            background = NULL,
            strokeWidth = 2,
            strokeColor = "#ffffff",
            align = "middle"))

      vn_obj <-
        visPhysics(
          graph = vn_obj,
          stabilization = list(
            enabled = TRUE,
            onlyDynamicEdges = FALSE,
            fit = TRUE))

      vn_obj <-
        visInteraction(
          graph = vn_obj,
          dragNodes = FALSE)
    }
  }

  vn_obj
}

#' Get histogram data for a graph's degree frequency
#'
#' @description
#'
#' Get histogram data for a graph's degree frequency. The bin width is set to 1
#' and zero-value degrees are omitted from the output.
#'
#' @inheritParams render_graph
#' @param mode using `total` (the default), degree considered for each node
#'   will be the total degree. With `in` and `out` the degree used
#'   will be the in-degree and out-degree, respectively.
#'
#' @return A data frame with degree counts.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get degree histogram data for
#' # the graph (reporting total degree)
#' graph |>
#'   get_degree_histogram(
#'     mode = "total")
#'
#' @export
get_degree_histogram <- function(
    graph,
    mode = "total"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Convert the graph to an igraph object
  # ig_graph <- to_igraph(graph) # not used?

  # Get the total degree histogram for the graph
  if (mode %in% c("all", "total", "both")) {

    deg_hist_df <-
      get_degree_distribution(graph) |>
      dplyr::mutate(total_degree_hist = total_degree_dist * count_nodes(graph)) |>
      dplyr::select("degree", "total_degree_hist")
  }

  # Get the total in-degree distribution for the graph
  if (mode == "in") {

    deg_hist_df <-
      get_degree_distribution(graph, mode = "in") |>
      dplyr::mutate(indegree_hist = indegree_dist * count_nodes(graph)) |>
      dplyr::select("degree", "indegree_hist")
  }

  # Get the total out-degree distribution for the graph
  if (mode == "out") {

     deg_hist_df <-
      get_degree_distribution(graph, mode = "out") |>
      dplyr::mutate(outdegree_hist = outdegree_dist * count_nodes(graph)) |>
      dplyr::select("degree", "outdegree_hist")
  }

  deg_hist_df
}

#' Get community membership by modularity optimization
#'
#' @description
#'
#' Through the use of greedy optimization of a modularity score, obtain the
#' group membership values for each of the nodes in the graph. Note that this
#' method only works on graphs without multiple edges.
#'
#' @inheritParams render_graph
#'
#' @return a data frame with group membership assignments for each of the nodes.
#'
#' @examples
#' # Create a graph with a
#' # balanced tree
#' graph <-
#'   create_graph() %>%
#'   add_balanced_tree(
#'     k = 2,
#'     h = 2)
#'
#' # Get the group membership
#' # values for all nodes in
#' # the graph through the greedy
#' # optimization of modularity
#' # algorithm
#' graph %>%
#'   get_cmty_fast_greedy()
#'
#' # Add the group membership
#' # values to the graph as a
#' # node attribute
#' graph <-
#'   graph %>%
#'   join_node_attrs(
#'     df = get_cmty_fast_greedy(.))
#'
#' # Display the graph's
#' # node data frame
#' graph %>% get_node_df()
#'
#' @export
get_cmty_fast_greedy <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If graph is directed, transform to undirected
  graph <- set_graph_undirected(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the community object using the
  # `cluster_fast_greedy()` function
  cmty_fast_greedy_obj <-
    igraph::cluster_fast_greedy(ig_graph)

  # Create df with node memberships
  data.frame(
    id = igraph::membership(cmty_fast_greedy_obj) %>% names() %>% as.integer(),
    f_g_group = as.vector(igraph::membership(cmty_fast_greedy_obj)),
    stringsAsFactors = FALSE)
}

#' Get the alpha centrality for all nodes
#'
#' @description
#'
#' Get the alpha centrality values for all nodes in the graph.
#'
#' @inheritParams render_graph
#' @param alpha the parameter that specifies the relative importance of
#'   endogenous versus exogenous factors in the determination of centrality.
#' @param exo the exogenous factors, in most cases this is either a constant
#'   (which applies the same factor to every node), or a vector giving the
#'   factor for every node.
#' @param weights_attr an optional name of the edge attribute to use in the
#'   adjacency matrix. If `NULL` then, if it exists, the `weight` edge attribute
#'   of the graph will be used. Failing that, the standard adjacency matrix will
#'   be used in calculations.
#' @param tol the tolerance for near-singularities during matrix inversion. The
#'   default value is set to `1e-7`.
#'
#' @return A data frame with alpha centrality scores for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 12,
#'     set_seed = 23)
#'
#' # Get the alpha centrality scores
#' # for all nodes
#' graph %>%
#'   get_alpha_centrality()
#'
#' # Add the alpha centrality
#' # scores to the graph as a node
#' # attribute
#' graph <-
#'   graph %>%
#'   join_node_attrs(
#'     df = get_alpha_centrality(.))
#'
#' # Display the graph's node
#' # data frame
#' graph %>% get_node_df()
#'
#' @export
get_alpha_centrality <- function(
    graph,
    alpha = 1,
    exo = 1,
    weights_attr = NULL,
    tol = 1e-7
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the alpha centrality values for each of the
  # graph's nodes
  alpha_centrality_values <-
    igraph::alpha_centrality(
      graph = ig_graph,
      alpha = alpha,
      exo = exo,
      weights = weights_attr,
      tol = tol)

  # Create df with alpha centrality values
  data.frame(
    id = alpha_centrality_values %>%
      names() %>%
      as.integer(),
    alpha_centrality = alpha_centrality_values,
    stringsAsFactors = FALSE)
}

#' Is the graph empty?
#'
#' @description
#'
#' Provides a logical value on whether the graph is empty (i.e., contains no
#' nodes).
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create an empty graph
#' graph <- create_graph()
#'
#' # Determine whether the graph is empty
#' graph |> is_graph_empty()
#'
#' # Create a non-empty graph
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(n = 3)
#'
#' # Determine whether this graph is empty
#' graph |> is_graph_empty()
#'
#' @export
is_graph_empty <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Determine if graph is empty by checking the
  # number of rows in `graph$nodes_df`
  if (nrow(graph$nodes_df) == 0) {
    return(TRUE)
  } else {
    return(FALSE)
  }
}

#' ISO-4217 currency data.
#'
#' A dataset containing currency information from the ISO-4217 standard.
#'
#' @format A data frame with 171 rows and 4 variables:
#' \describe{
#'   \item{iso_4217_code}{the three-letter currency code according to the ISO-4217 standard}
#'   \item{curr_number}{the three-digit code number assigned to each currency under the ISO-4217 standard}
#'   \item{exponent}{the base 10 exponent of the minor currency unit in relation to the major currency unit (it can be assumed also to be number of decimal places that is commonly considered for the currency)}
#'   \item{currency_name}{the English name of the currency}
#' }
#' @source <https://en.wikipedia.org/wiki/ISO_4217>
"currencies"

#' US Dollar exchange rates.
#'
#' A dataset containing exchange rates from USD to all other currencies.
#'
#' @format A data frame with 196 rows and 3 variables:
#' \describe{
#'   \item{from_currency}{the currency from which units will be used to buy units in the alternate currency (this is always USD)}
#'   \item{to_currency}{the currency that is to be bought}
#'   \item{cost_unit}{the cost per unit of the currency to be bought}
#' }
"usd_exchange_rates"

#' Node list - Version 1.
#'
#' A very simple, 2-column data frame that can be used to generate graph nodes.
#'
#' @format A data frame with 10 rows and 2 variables:
#' \describe{
#'   \item{id}{a unique, monotonically increasing integer ID value}
#'   \item{label}{a unique label associated with each ID value}
#' }
"node_list_1"

#' Node list - Version 2.
#'
#' A simple, 5-column data frame that can be used to generate graph nodes.
#'
#' @format A data frame with 10 rows and 5 variables:
#' \describe{
#'   \item{id}{a unique, monotonically increasing integer ID value}
#'   \item{label}{a unique label associated with each ID value}
#'   \item{type}{a grouping variable of either x, y, or z}
#'   \item{value_1}{a randomized set of numeric values between 0 and 10}
#'   \item{value_2}{a randomized set of numeric values between 0 and 10}
#' }
"node_list_2"

#' Edge list - Version 1.
#'
#' A very simple, 2-column data frame that can be used to generate graph edges.
#'
#' @format A data frame with 19 rows and 2 variables:
#' \describe{
#'   \item{from}{integer values that state the node ID values where an edge starts}
#'   \item{to}{integer values that state the node ID values where an edge terminates}
#' }
"edge_list_1"

#' Edge list - Version 2.
#'
#' A simple, 5-column data frame that can be used to generate graph edges.
#'
#' @format A data frame with 19 rows and 5 variables:
#' \describe{
#'   \item{from}{integer values that state the node ID values where an edge starts}
#'   \item{to}{integer values that state the node ID values where an edge terminates}
#'   \item{rel}{a grouping variable of either a, b, or c}
#'   \item{value_1}{a randomized set of numeric values between 0 and 10}
#'   \item{value_2}{a randomized set of numeric values between 0 and 10}
#' }
"edge_list_2"

#' Use the breadth-first search (bfs) algorithm
#'
#' @description
#'
#' With a chosen or random node serving as the starting point, perform a
#' breadth-first search of the whole graph and return the node ID values
#' visited. The bfs algorithm differs from depth-first search (dfs) in that bfs
#' will follow tree branches branches one level at a time until terminating at
#' leaf node (dfs traverses branches as far as possible).
#'
#' @inheritParams render_graph
#' @param node An optional node ID value to specify a single starting point for
#'   the bfs. If not provided, a random node from the graph will be chosen.
#' @param direction Using `all` (the default), the bfs will ignore edge
#'   direction while traversing through the graph. With `out` and `in`,
#'   traversals between adjacent nodes will respect the edge direction.
#'
#' @return A vector containing node ID values for nodes visited during the
#'   breadth-first search. The order of the node IDs corresponds to the order
#'   visited.
#'
#' @examples
#' # Create a graph containing
#' # two balanced trees
#' graph <-
#'   create_graph() %>%
#'   add_balanced_tree(
#'     k = 2, h = 2) %>%
#'   add_balanced_tree(
#'     k = 3, h = 2)
#'
#' # Perform a breadth-first
#' # search of the graph,
#' # beginning at the root node
#' # `1` (the default
#' # `direction = "all"` doesn't
#' # take edge direction into
#' # account)
#' graph %>%
#'   do_bfs(node = 1)
#'
#' # If not specifying a
#' # starting node, the function
#' # will begin the search from
#' # a random node
#' graph %>%
#'   do_bfs()
#'
#' # It's also possible to
#' # perform bfs while taking
#' # into account edge direction;
#' # using `direction = "in"`
#' # causes the bfs routine to
#' # visit nodes along inward edges
#' graph %>%
#'   do_bfs(
#'     node = 1,
#'     direction = "in")
#'
#' # Using `direction = "out"`
#' # results in the bfs moving
#' # along solely outward edges
#' graph %>%
#'   do_bfs(
#'     node = 1,
#'     direction = "out")
#'
#' @export
do_bfs <- function(
    graph,
    node = NULL,
    direction = "all"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # If no node provided, choose a random node
  if (is.null(node)) {
    node <- sample(graph$nodes_df[, 1], 1)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Perform the breadth-first search algorithm in
  # the direction requested
  rlang::arg_match0(direction, c("all", "out", "in"))

  if (direction == "all") {

    bfs_result <-
      igraph::bfs(
        graph = ig_graph,
        root = node,
        mode = "all")

  } else if (direction == "out") {

    bfs_result <-
      igraph::bfs(
        graph = ig_graph,
        root = node,
        mode = "out")

  } else if (direction == "in") {

    bfs_result <-
      igraph::bfs(
        graph = ig_graph,
        root = node,
        mode = "in")

  }

  # Get the nodes visited during the bfs
  as.integer(as.numeric(bfs_result$order))
}

#' Add a graph action for execution at every transform
#'
#' @description
#'
#' Add a graph function along with its arguments to be run at every graph
#' transformation step.
#'
#' @inheritParams render_graph
#' @param fcn The name of the function to use.
#' @param ... Arguments and values to pass to the named function in `fcn`, if
#'   necessary.
#' @param action_name An optional name for labeling the action.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 22,
#'     set_seed = 23)
#'
#' # Add a graph action that sets a node
#' # attr column with a function; the
#' # main function `set_node_attr_w_fcn()`
#' # uses the `get_betweenness()` function
#' # to provide betweenness values in the
#' # `btwns` column; this action will
#' # occur whenever there is a function
#' # called on the graph that modifies it
#' # (e.g., `add_n_nodes()`)
#' graph <-
#'   graph %>%
#'   add_graph_action(
#'     fcn = "set_node_attr_w_fcn",
#'     node_attr_fcn = "get_betweenness",
#'     column_name = "btwns",
#'     action_name = "get_btwns")
#'
#' # To ensure that the action is
#' # available in the graph, use the
#' # `get_graph_actions()` function
#' graph %>% get_graph_actions()
#'
#' @export
add_graph_action <- function(
    graph,
    fcn,
    ...,
    action_name = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Collect any function arguments into the
  # `fcn_args` list object
  fcn_args <- list(...)

  # Create a character expression for the
  # function to evaluate at every graph
  # transformation step
  if (length(fcn_args) == 0L) {
    char_expr <-
      paste0(
        fcn,
        "(graph = graph)")

  } else {

    arg_names <- vector(mode = "character")
    arg_values <- vector(mode = "character")

    for (i in seq_along(fcn_args)) {

      arg_names <- c(arg_names, names(fcn_args)[i])

      arg_value_class <- class(unname(fcn_args)[[i]])

      if (arg_value_class == "character") {

        fcn_arg <- unname(fcn_args)[[i]]

        fcn_arg <- paste0("'", fcn_arg, "'")

        arg_values <-
          c(arg_values, fcn_arg)

      } else {
        arg_values <-
          c(arg_values,
            unname(fcn_args)[[i]])
      }
    }

    # Create a character-based expression
    char_expr <-
      paste0(
        fcn,
        "(graph = graph, ",
        paste(arg_names, "=", arg_values, collapse = ", "),
        ")")
  }

  # Create a data frame row with the new graph action
  new_graph_action <-
    data.frame(
      action_index = ifelse(nrow(graph$graph_actions) == 0L, 1L,
                            max(graph$graph_actions$action_index) + 1L),
      action_name = action_name %||% NA_character_,
      expression = char_expr,
      stringsAsFactors = FALSE)

  # Append `new_graph_action` to `graph$graph_actions`
  graph$graph_actions <-
    dplyr::bind_rows(graph$graph_actions, new_graph_action)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Generate DOT code using a graph object
#'
#' @description
#'
#' Generates Graphviz DOT code as an R character object using DiagrammeR graph
#' object.
#'
#' @inheritParams render_graph
#'
#' @return A character vector of length 1 containing Graphviz DOT code.
#'
#' @export
generate_dot <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Extract objects from the graph object
  nodes_df <- graph$nodes_df
  edges_df <- graph$edges_df
  directed <- graph$directed
  global_attrs <- graph$global_attrs

  if ("graph" %in% global_attrs$attr_type) {
    graph_attrs <-
      global_attrs |>
      dplyr::filter(attr_type == "graph") |>
      dplyr::mutate(string = paste0(attr, " = '", value, "'"))

    graph_attrs <-
      graph_attrs |>
      dplyr::pull("string")

  } else {
    graph_attrs <- NA
  }

  if ("node" %in% global_attrs$attr_type) {
    node_attrs <-
      global_attrs |>
      dplyr::filter(attr_type == "node") |>
      dplyr::mutate(string = paste0(attr, " = '", value, "'"))

    node_attrs <-
      node_attrs |>
      dplyr::pull("string")

    # Fill in NA attribute values with global preset values
    node_global_attrs <- global_attrs |> dplyr::filter(attr_type == "node")
    for (i in seq_len(nrow(node_global_attrs))) {

      node_attr_to_set <- node_global_attrs[i, 1]

      if (node_attr_to_set %in% colnames(nodes_df)) {

        col_num <- which(colnames(nodes_df) == node_attr_to_set)

        nodes_df[which(is.na(nodes_df[, col_num])), col_num] <-
          node_global_attrs[i, 2]
      }
    }

  } else {
    node_attrs <- NA
  }

  if ("edge" %in% global_attrs$attr_type) {
    edge_attrs <-
      global_attrs |>
      dplyr::filter(attr_type == "edge") |>
      dplyr::mutate(string = paste0(attr, " = '", value, "'"))

    edge_attrs <-
      edge_attrs |>
      dplyr::pull("string")

    # Fill in NA attribute values with global preset values
    edge_global_attrs <- global_attrs |> dplyr::filter(attr_type == "edge")
    for (i in seq_len(nrow(edge_global_attrs))) {

      edge_attr_to_set <- edge_global_attrs[i, 1]

      if (edge_attr_to_set %in% colnames(edges_df)) {

        col_num <- which(colnames(edges_df) == edge_attr_to_set)

        edges_df[which(is.na(edges_df[, col_num])), col_num] <-
          edge_global_attrs[i, 2]
      }
    }

  } else {
    edge_attrs <- NA
  }


  # Replace NA values with empty strings in character columns of `nodes_df`
  if (!is.null(nodes_df)) {

    if (ncol(nodes_df) >= 4) {
      vars <- base::setdiff(colnames(nodes_df), c("id", "type", "label"))
      vars <- vars[purrr::map_lgl(nodes_df[vars], is.character)]

      nodes_df <-
        nodes_df |>
        dplyr::mutate(
          dplyr::across(
            .cols = dplyr::all_of(vars),
            .fns = ~ dplyr::coalesce(.x, "")
          ))
    }
  }


  # Replace NA values with empty strings in character columns of `edges_df`
  if (!is.null(edges_df)) {

    if (ncol(edges_df) >= 5) {
      vars <- base::setdiff(colnames(edges_df), c("id", "from", "to", "rel"))
      vars <- vars[purrr::map_lgl(edges_df[vars], is.character)]

      edges_df <-
        edges_df |>
        dplyr::mutate(
          dplyr::across(
            .cols = dplyr::all_of(vars),
            .fns = ~ dplyr::coalesce(.x, "")
            ))
    }
  }

  # If `equation` column in `nodes_df`, ensure the
  # right amount of escaping is present
  if ("equation" %in% colnames(nodes_df)) {
    equation_col <- which(colnames(nodes_df) == "equation")

    for (i in seq_len(nrow(nodes_df))) {
      if (grepl("^\\$.*\\$$", nodes_df[i, equation_col])) {
        nodes_df[i, equation_col] <-
          stringr::str_replace_all(
            nodes_df[i, equation_col], "\\\\", "\\\\\\\\")
      } else {
        nodes_df[i, equation_col] <- ""
      }
    }
  }

  # If `display` column in `nodes_df`, modify label
  # column for this render
  if ("display" %in% colnames(nodes_df)) {

    display_col <- which(colnames(nodes_df) == "display")
    label_col <- which(colnames(nodes_df) == "label")

    for (i in seq_len(nrow(nodes_df))) {

      if (nzchar(nodes_df[i, display_col])) {

        nodes_df[i, label_col] <-
          nodes_df[
            i, which(colnames(nodes_df) == nodes_df[i, display_col])]

      } else {
        nodes_df[i, label_col] <- ""
      }
    }
  }

  # If `display` column in `edges_df`, modify label
  # column for this render
  if ("display" %in% colnames(edges_df)) {

    display_col <- which(colnames(edges_df) == "display")

    if (!("label" %in% colnames(edges_df))) {

      edges_df$label <- NA_character_
    }

    label_col <- which(colnames(edges_df) == "label")

    for (i in seq_len(nrow(edges_df))) {
      if (!is.na(edges_df[i, display_col]) ) {
        if (nzchar(edges_df[i, display_col])) {

          edges_df[i, label_col] <-
            edges_df[
              i, which(colnames(edges_df) == edges_df[i, display_col])]
        }
      } else {
        edges_df[i, label_col] <- ""
      }
    }
  }

  # Create vector of graph attributes
  graph_attributes <- gv_graph_attributes()

  # Create vector of node attributes
  node_attributes <- gv_node_attributes()

  # Create vector of edge attributes
  edge_attributes <- gv_edge_attributes()

  if (nrow(nodes_df) == 0 &&
      nrow(edges_df) == 0) {

    # Create DOT code with nothing in graph
    dot_code <-
      paste0(ifelse(directed,
                    "digraph", "graph"),
             " {\n", "\n}")

  } else {

    #
    # Create the DOT attributes block
    #

    # Create the default attributes statement
    # for graph attributes
    if (anyNA(graph_attrs)) {
      graph_attr_stmt <- ""
    } else {
      graph_attr_stmt <-
        paste0(
          "graph [",
          paste(graph_attrs, collapse = ",\n       "),
          "]\n"
          )
    }

    # Create the default attributes statement
    # for node attributes
    if (anyNA(node_attrs)) {
      node_attr_stmt <- ""

    } else {
      node_attr_stmt <-
        paste0(
          "node [",
          paste(node_attrs, collapse = ",\n      "),
          "]\n"
          )

    }

    # Create the default attributes statement
    # for edge attributes
    if (anyNA(edge_attrs)) {
      edge_attr_stmt <- ""

    } else {
      edge_attr_stmt <-
        paste0(
          "edge [",
          paste(edge_attrs, collapse = ",\n     "),
          "]\n"
          )

    }

    # Combine default attributes into a single block
    combined_attr_stmts <-
      paste(
        graph_attr_stmt,
        node_attr_stmt,
        edge_attr_stmt, sep = "\n")

    #
    # Create the DOT node block
    #

    if (nrow(nodes_df) > 0) {

      # Determine whether positional (x,y) data is included
      column_with_x <- which(colnames(nodes_df) %in% "x")[1]

      column_with_y <- which(colnames(nodes_df) %in% "y")[1]

      if (!is.na(column_with_x) && !is.na(column_with_y)) {
        pos <-
          paste0(
            nodes_df |> dplyr::pull(column_with_x), ",",
            nodes_df |> dplyr::pull(column_with_y), "!"
          )

        nodes_df <-
          nodes_df |>
          dplyr::mutate(pos = !!pos)
      }

      # Determine whether column 'alpha' exists
      if (any(grepl("$alpha^", colnames(nodes_df)))) {
        column_with_alpha_assigned <-
          grep("$alpha^", colnames(nodes_df))
      } else {
        column_with_alpha_assigned <- NA
      }

      if (!is.na(column_with_alpha_assigned)) {

        # Determine the number of color attributes in
        # the node data frame
        number_of_col_attr <-
          length(which(colnames(nodes_df) %in%
                         c("color", "fillcolor",
                           "fontcolor")))

        # If the number of color attrs in df is 1,
        # rename referencing alpha column
        if (number_of_col_attr == 1) {

          name_of_col_attr <-
            colnames(nodes_df)[
              which(colnames(nodes_df) %in%
                      c("color", "fillcolor",
                        "fontcolor"))]

          colnames(nodes_df)[column_with_alpha_assigned] <-
            paste0("alpha:", name_of_col_attr)
        }
      }

      # Determine whether column 'alpha' with
      # color attr exists
      if (any(grepl("alpha:.*", colnames(nodes_df)))) {

        alpha_column_no <- grep("alpha:.*", colnames(nodes_df))

        color_attr_column_name <-
          unlist(strsplit(colnames(nodes_df)[
            (grep("alpha:.*", colnames(nodes_df)))
          ], ":"))[-1]

        color_attr_column_no <-
          which(colnames(nodes_df) %in% color_attr_column_name)

        # Append alpha value only if referenced
        # column is for color
        if (any(c("color", "fillcolor", "fontcolor") %in%
                colnames(nodes_df)[color_attr_column_no])) {

          # Append alpha for color values that are
          # X11 color names
          if (all(grepl("[a-z]*",
                        as.character(nodes_df[, color_attr_column_no]))) &&
              all(as.character(nodes_df[, color_attr_column_no]) %in%
                  x11_hex()[, 1])) {

            for (i in seq_len(nrow(nodes_df))) {
              nodes_df[i, color_attr_column_no] <-
                paste0(x11_hex()[
                  which(x11_hex()[, 1] %in%
                          as.character(nodes_df[i, color_attr_column_no])), 2],
                  formatC(round(as.numeric(nodes_df[i, alpha_column_no]), 0),
                          flag = "0", width = 2))
            }
          }

          # Append alpha for color values that
          # are hex color values
          if (all(grepl("#[0-9a-fA-F]{6}$",
                        as.character(nodes_df[, color_attr_column_no])))) {

            for (i in seq_len(nrow(nodes_df))) {
              nodes_df[, color_attr_column_no] <-
                as.character(nodes_df[, color_attr_column_no])

              nodes_df[i, color_attr_column_no] <-
                paste0(nodes_df[i, color_attr_column_no],
                       round(as.numeric(nodes_df[i, alpha_column_no]), 0))
            }
          }
        }
      }

      # Determine which other columns correspond
      # to node attribute values
      other_columns_with_node_attributes <-
        which(colnames(nodes_df) %in% node_attributes)

      # Construct the 'node_block' character object
      for (i in seq_len(nrow(nodes_df))) {
        if (i == 1) {
          node_block <- vector(mode = "character", length = 0)
        }

        if (length(other_columns_with_node_attributes) > 0) {

          for (j in other_columns_with_node_attributes) {

            if (j == other_columns_with_node_attributes[1]) {
              attr_string <- vector(mode = "character", length = 0)
            }

            # Create the node attributes for labels
            # and tooltips when provided
            if (all(colnames(nodes_df)[j] %in%
                    c("label", "tooltip"),
                    is.na(nodes_df[i, j]))) {
              attribute <- NULL
            } else if (all(colnames(nodes_df)[j] %in%
                           c("label", "tooltip"),
                           !is.na(nodes_df[i, j]))) {
              attribute <-
                paste0(colnames(nodes_df)[j],
                       " = ", "'", nodes_df[i, j], "'")
            } else if (all(!(colnames(nodes_df)[j] %in%
                             c("label", "tooltip")),
                           is.na(nodes_df[i, j]))) {
              attribute <- NULL
            } else if (all(!(colnames(nodes_df)[j] %in%
                             c("label", "tooltip")),
                           !is.na(nodes_df[i, j]))) {
              attribute <-
                paste0(colnames(nodes_df)[j],
                       " = ", "'", nodes_df[i, j], "'")
            }
            attr_string <- c(attr_string, attribute)
          }

          if (j == other_columns_with_node_attributes[
            length(other_columns_with_node_attributes)]) {
            attr_string <- paste(attr_string, collapse = ", ")
          }
        }

        # Generate a line of node objects when an
        # attribute string exists
        if (exists("attr_string")) {
          line <- paste0("  '", nodes_df[i, 1], "'",
                         " [", attr_string, "] ")
        }

        # Generate a line of node objects when an
        # attribute string doesn't exist
        if (!exists("attr_string")) {
          line <-
            paste0("  '",
                   nodes_df[i, 1],
                   "'")
        }
        node_block <- c(node_block, line)
      }

      if ("rank" %in% colnames(nodes_df)) {
        node_block <-
          c(node_block,
            tapply(node_block,
                   nodes_df$rank, FUN = function(x) {
                     if(length(x) > 1) {
                       x <- paste0('subgraph{rank = same\n',
                                   paste0(x, collapse = '\n'),
                                   '}\n')
                     }
                     return(x)
                   })
          )

      } else if ("cluster" %in% colnames(nodes_df)) {

        cluster_vals <- nodes_df$cluster

        clustered_node_block <- character(0L)
        clusters <- split(node_block, cluster_vals)

        for (i in seq_along(clusters)) {
          if (names(clusters)[[i]] == "") {
            # nodes not in clusters
            cluster_block <- clusters[[i]]
          } else {
            cluster_block <-
              paste0(
                "subgraph cluster", i, "{\nlabel='",
                names(clusters)[[i]], "'\n",
                paste0(clusters[[i]], collapse = "\n"), "}\n"
              )
          }

          clustered_node_block <- c(clustered_node_block, cluster_block)
        }

        node_block <- clustered_node_block

        # cleanup variables
        rm(clustered_node_block, clusters, cluster_block)
      }

      # Construct the `node_block` character object
      node_block <- paste(node_block, collapse = "\n")

      # Remove the `attr_string` object if it exists
      if (exists("attr_string")) {
        rm(attr_string)
      }

      # Remove the `attribute` object if it exists
      if (exists("attribute")) {
        rm(attribute)
      }
    }

    #
    # Create the DOT edge block
    #

    if (nrow(edges_df) > 0) {

      # Determine whether `from` or `to` columns are
      # in `edges_df`
      from_to_columns <-
        any(c("from", "to") %in% colnames(edges_df))

      # Determine which columns in `edges_df`
      # contain edge attributes
      other_columns_with_edge_attributes <-
        which(colnames(edges_df) %in% edge_attributes)

      # Determine whether the complementary set of
      # columns is present
      if (from_to_columns) {
        both_from_to_columns <-
          all(c("from", "to") %in% colnames(edges_df))
      }

      # If the complementary set of columns is present,
      # determine the positions
      if (exists("both_from_to_columns")) {
        if (both_from_to_columns) {
          from_column <-
            which(colnames(edges_df) %in% c("from"))[1]
          to_column <-
            which(colnames(edges_df) %in% c("to"))[1]
        }
      }

      # Construct the `edge_block` character object
      if (exists("from_column") &&
          exists("to_column")) {

        if (length(from_column) == 1 &&
            length(from_column) == 1) {

          for (i in 1:nrow(edges_df)) {

            if (i == 1) {
              edge_block <-
                vector(mode = "character", length = 0)
            }

            if (length(other_columns_with_edge_attributes) > 0) {

              for (j in other_columns_with_edge_attributes) {

                if (j == other_columns_with_edge_attributes[1]) {
                  attr_string <- vector(mode = "character", length = 0)
                }

                # Create the edge attributes for labels
                # and tooltips when provided
                if (all(colnames(edges_df)[j] %in%
                        c("edgetooltip", "headtooltip",
                          "label", "labeltooltip",
                          "taillabel", "tailtooltip",
                          "tooltip"),
                        is.na(edges_df[i, j]))) {
                  attribute <- NULL
                } else if (all(colnames(edges_df)[j] %in%
                               c("edgetooltip", "headtooltip",
                                 "label", "labeltooltip",
                                 "taillabel", "tailtooltip",
                                 "tooltip"),
                               edges_df[i, j] != '')) {
                  attribute <-
                    paste0(colnames(edges_df)[j],
                           " = ", "'", edges_df[i, j],
                           "'")
                } else if (all(!(colnames(edges_df)[j] %in%
                                 c("edgetooltip", "headtooltip",
                                   "label", "labeltooltip",
                                   "taillabel", "tailtooltip",
                                   "tooltip")),
                               is.na(edges_df[i, j]))) {

                  attribute <- NULL
                } else if (all(!(colnames(edges_df)[j] %in%
                                 c("edgetooltip", "headtooltip",
                                   "label", "labeltooltip",
                                   "taillabel", "tailtooltip",
                                   "tooltip")),
                               edges_df[i, j] != '')) {
                  attribute <-
                    paste0(colnames(edges_df)[j],
                           " = ", "'", edges_df[i, j], "'")
                }else if(edges_df[i, j] == ''){
                  attribute <- NULL
                }

                attr_string <- c(attr_string, attribute)
              }

              if (j == other_columns_with_edge_attributes[
                length(other_columns_with_edge_attributes)]) {
                attr_string <- paste(attr_string, collapse = ", ")
              }
            }

            # Generate a line of edge objects when an
            # attribute string exists
            if (exists("attr_string")) {
              line <-
                paste0("'", edges_df[i, from_column], "'",
                       ifelse(directed, "->", "--"),
                       "'", edges_df[i, to_column], "'",
                       paste0(" [", attr_string, "] "))
            }

            # Generate a line of edge objects when an
            # attribute string doesn't exist
            if (!exists("attr_string")) {
              line <-
                paste0("  ",
                       "'", edges_df[i, from_column], "'",
                       ifelse(directed, "->", "--"),
                       "'", edges_df[i, to_column], "'",
                       " ")
            }
            edge_block <- c(edge_block, line)
          }
        }
      }

      # Construct the `edge_block` character object
      if (exists("edge_block")) {
        edge_block <- paste(edge_block, collapse = "\n")
      }
    }

    # Create the graph code from the chosen attributes,
    # and the nodes and edges blocks
    if (exists("combined_attr_stmts")) {
      if (exists("edge_block") && exists("node_block")) {
        combined_block <-
          paste(combined_attr_stmts,
                node_block, edge_block,
                sep = "\n")
      }
      if (!exists("edge_block") && exists("node_block")) {
        combined_block <-
          paste(combined_attr_stmts,
                node_block,
                sep = "\n")
      }
    }
    if (!exists("combined_attr_stmts")) {
      if (exists("edge_block")) {
        combined_block <- paste(node_block, edge_block,
                                sep = "\n")
      } else {
        combined_block <- node_block
      }
    }

    # Create DOT code
    dot_code <-
      paste0(ifelse(directed, "digraph", "graph"),
             " {\n", "\n", combined_block, "\n}")

    # Remove empty node or edge attribute statements
    dot_code <- gsub(" \\[\\] ", "", dot_code)
  }

  dot_code
}

#' Get betweenness centrality scores
#'
#' @description
#'
#' Get the betweenness centrality scores for all nodes in a graph.
#'
#' @inheritParams render_graph
#'
#' @return a data frame with betweenness scores for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 12,
#'     set_seed = 23)
#'
#' # Get the betweenness scores
#' # for nodes in the graph
#' graph %>% get_betweenness()
#'
#' # Add the betweenness
#' # values to the graph
#' # as a node attribute
#' graph <-
#'   graph %>%
#'   join_node_attrs(
#'     df = get_betweenness(.))
#'
#' # Display the graph's node
#' # data frame
#' graph %>% get_node_df()
#'
#' @export
get_betweenness <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the betweenness scores for each of the
  # graph's nodes
  betweenness_scores <-
    igraph::betweenness(
      graph = ig_graph,
      v = igraph::V(ig_graph),
      directed = graph$directed)

  # Create df with betweenness scores
  data.frame(
    id = betweenness_scores %>%
      names() %>%
      as.integer(),
    betweenness = betweenness_scores,
    stringsAsFactors = FALSE)
}

#' Get information on any available graph actions
#'
#' @description
#'
#' Get a tibble of the available graph actions, which contains information on
#' function invocations to be called on the graph at every transformation step,
#' or, when manually invoked with the [trigger_graph_actions()] function.
#'
#' @inheritParams render_graph
#'
#' @return A `df_tbl` object.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Add a graph action that sets a node
#' # attr column with a function; the
#' # main function `set_node_attr_w_fcn()`
#' # uses the `get_betweenness()` function
#' # to provide betweenness values in the
#' # `btwns` column
#' graph <-
#'   graph |>
#'   add_graph_action(
#'     fcn = "set_node_attr_w_fcn",
#'     node_attr_fcn = "get_betweenness",
#'     column_name = "btwns",
#'     action_name = "get_btwns")
#'
#' # To ensure that the action is
#' # available in the graph, use the
#' # `get_graph_actions()` function
#' graph |> get_graph_actions()
#'
#' @export
get_graph_actions <- function(graph) {
  check_graph_valid(graph)
  dplyr::as_tibble(graph$graph_actions)
}

#' Drop a node attribute column
#'
#' @description
#'
#' Within a graph's internal node data frame (ndf), remove an existing node
#' attribute.
#'
#' @inheritParams render_graph
#' @param node_attr The name of the node attribute column to drop.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 5,
#'     m = 10,
#'     set_seed = 23) %>%
#'  set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = count_nodes(.),
#'       mean = 5,
#'       sd = 1) %>% round(1))
#'
#' # Get the graph's internal
#' # ndf to show which node
#' # attributes are available
#' graph %>% get_node_df()
#'
#' # Drop the `value` node
#' # attribute
#' graph <-
#'   graph %>%
#'   drop_node_attrs(
#'     node_attr = value)
#'
#' # Get the graph's internal
#' # ndf to show that the node
#' # attribute `value` had been
#' # removed
#' graph %>% get_node_df()
#'
#' @family node creation and removal
#'
#' @export
drop_node_attrs <- function(
    graph,
    node_attr
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the requested `node_attr`
  node_attr <-
    rlang::enquo(node_attr) %>% rlang::get_expr() %>% as.character()

  # Stop function if length of `node_attr` is
  # greater than one
  if (length(node_attr) > 1) {

    cli::cli_abort(
      "You can only provide a single column to `node_attr`.")
  }

  # Stop function if `node_attr` is any of
  # `nodes`, `node`, `type`, or `label`
  if (any(c("nodes", "node", "type", "label") %in%
          node_attr)) {

    cli::cli_abort(
      "You cannot drop the {.code {node_attr}} column.")
  }

  # Extract the graph's ndf
  nodes <- get_node_df(graph)

  # Get column names from the graph's ndf
  column_names_graph <- colnames(nodes)

  # Stop function if `node_attr` is not one
  # of the graph's column
  if (!any(column_names_graph %in% node_attr)) {

    cli::cli_abort(
      "The node attribute to drop is not in the ndf.")
  }

  # Get the column number for the node attr to drop
  col_num_drop <-
    which(colnames(nodes) %in% node_attr)

  # Remove the column
  nodes <- nodes[, -col_num_drop]

  # Create a new graph object
  graph$nodes_df <- nodes

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get the minimum cut between source and sink nodes
#'
#' @description
#'
#' Get the minimum cut between source and sink nodes. This is the minimum total
#' capacity of edges needed for removal in order to eliminate all paths from the
#' source and sink nodes.
#'
#' @inheritParams render_graph
#' @param from The node ID for the source node.
#' @param to The node ID for the sink or target node.
#'
#' @return A single numeric value representing the minimum total edge capacity
#'   removed to disconnect the source and sink nodes.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a cycle graph
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' # Determine the minimum cut
#' # between nodes `1` and `4`
#' graph |>
#'   get_min_cut_between(
#'     from = 1,
#'     to = 2)
#'
#' # Create a cycle graph with
#' # randomized values given to all
#' # edges as the `capacity` attribute
#' graph_capacity <-
#'   create_graph() |>
#'   add_cycle(n = 5) |>
#'   select_edges() |>
#'   set_edge_attrs_ws(
#'     edge_attr = capacity,
#'     value =
#'       rnorm(
#'         n = 5,
#'         mean = 5,
#'         sd = 1)) |>
#'   clear_selection()
#'
#' # Determine the minimum cut
#' # between nodes `1` and `4` for
#' # this graph, where `capacity`is
#' # set as an edge attribute
#' graph_capacity |>
#'   get_min_cut_between(
#'     from = 1,
#'     to = 2)
#'
#' # Create a full graph and then
#' # get the minimum cut requirement
#' # between nodes `2` and `8`
#' create_graph() |>
#'   add_full_graph(n = 10) |>
#'   get_min_cut_between(
#'     from = 2,
#'     to = 8)
#'
#' @export
get_min_cut_between <- function(
    graph,
    from,
    to
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the minimum cut value for removing
  # all paths from the source to the sink
  igraph::min_cut(
    graph = ig_graph,
    source = from,
    target = to)
}

#' Is the graph a property graph?
#'
#' @description
#'
#' Provides a logical value on whether the graph is property graph (i.e., all
#' nodes have an assigned `type` value and all edges have an assigned `rel`
#' value).
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create a graph with 2 nodes
#' # (with `type` values) and a
#' # single edge (with a `rel`)
#' simple_property_graph <-
#'   create_graph() %>%
#'   add_node(
#'     type = "a",
#'     label = "first") %>%
#'   add_node(
#'     type = "b",
#'     label = "second") %>%
#'   add_edge(
#'     from = "first",
#'     to = "second",
#'     rel = "rel_1")
#'
#' # This is indeed a property graph
#' # but to confirm this, use the
#' # `is_property_graph()` function
#' is_property_graph(simple_property_graph)
#'
#' # If a `type` attribute is
#' # removed, then this graph will
#' # no longer be a property graph
#' simple_property_graph %>%
#'   set_node_attrs(
#'     node_attr = type,
#'     values = NA,
#'     nodes = 1) %>%
#'   is_property_graph()
#'
#' # An empty graph will return FALSE
#' create_graph() %>%
#'   is_property_graph()
#'
#' @export
is_property_graph <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  if (is_graph_empty(graph)) {
    return(FALSE)
  }

  if (
    all(
      !anyNA(graph$nodes_df$type),
      all(nzchar(graph$nodes_df$type)),
      !anyNA(graph$edges_df$rel),
      all(nzchar(graph$edges_df$rel)))
    ) {
    return(TRUE)
  }

  FALSE
}

#' Recode a set of node attribute values
#'
#' @description
#'
#' Within a graph's internal node data frame (ndf), recode character or numeric
#' node attribute values. Optionally, one can specify a replacement value for
#' any unmatched mappings.
#'
#' @inheritParams render_graph
#' @param node_attr_from The name of the node attribute column from which values
#'   will be recoded.
#' @param ... Single-length character vectors with the recoding instructions.
#'   The first component should have the value to replace and the second should
#'   have the replacement value (in the form `"[to_replace] -> [replacement]",
#'   ...`).
#' @param otherwise An optional single value for recoding any unmatched values.
#' @param node_attr_to An optional name of a new node attribute to which the
#'   recoded values will be applied. This will retain the original node
#'   attribute and its values.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 5,
#'     m = 10,
#'     set_seed = 23) |>
#'   set_node_attrs(
#'     node_attr = shape,
#'     values =
#'       c("circle", "hexagon",
#'         "rectangle", "rectangle",
#'         "circle"))
#'
#' # Get the graph's internal ndf
#' # to show which node
#' # attributes are available
#' graph |> get_node_df()
#'
#' # Recode the `shape` node
#' # attribute, so that `circle`
#' # is recoded to `square` and that
#' # `rectangle` becomes `triangle`
#' graph <-
#'   graph |>
#'   recode_node_attrs(
#'     node_attr_from = shape,
#'     "circle -> square",
#'     "rectangle -> triangle")
#'
#' # Get the graph's internal
#' # ndf to show that the node
#' # attribute values had been recoded
#' graph |> get_node_df()
#'
#' # Create a new node attribute,
#' # called `color`, that is based
#' # on a recoding of `shape`; here,
#' # map the square shape to a `red`
#' # color and map all other shapes
#' # to a `green` color
#' graph <-
#'   graph |>
#'   recode_node_attrs(
#'     node_attr_from = shape,
#'     "square -> red",
#'     otherwise = "green",
#'     node_attr_to = color)
#'
#' # Get the graph's internal ndf
#' # to see the change
#' graph |> get_node_df()
#'
#' @family node creation and removal
#'
#' @export
recode_node_attrs <- function(
    graph,
    node_attr_from,
    ...,
    otherwise = NULL,
    node_attr_to = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get the requested `node_attr_from`
  node_attr_from <-
    rlang::enquo(node_attr_from) |> rlang::get_expr() |> as.character()

  # Get the requested `node_attr_to`
  node_attr_to <-
    rlang::enquo(node_attr_to) |> rlang::get_expr() |> as.character()

  if (length(node_attr_to) == 0) {
    node_attr_to <- NULL
  }

  # Get list object from named vectors
  replacements <- list(...)

  # Extract the graph's ndf
  nodes <- get_node_df(graph)

  # Get column names from the graph's ndf
  column_names_graph <- colnames(nodes)

  # Stop function if `node_attr_from` is not one
  # of the graph's node attributes
  if (!any(column_names_graph %in% node_attr_from)) {

    cli::cli_abort(
      "The node attribute to recode is not in the ndf.")
  }

  # Get the column number for the node attr to recode
  col_num_recode <-
    which(colnames(nodes) %in% node_attr_from)

  # Get the column class for the node attr to recode
  node_attr_class <- class(nodes[, col_num_recode])

  # Extract the vector to recode from the `nodes` df
  vector_to_recode <- nodes[, col_num_recode]

  # Initialize the `indices_stack` vector
  indices_stack <- vector("numeric")

  # Parse the recoding pairs
  for (i in seq_along(replacements)) {

    pairing <-
      trimws(unlist(stringr::str_split(replacements[[i]], "->")))

    if (node_attr_class == "numeric") {
      pairing <- as.numeric(pairing)
    }

    indices <- which(vector_to_recode %in% pairing[1])

    vector_to_recode[base::setdiff(indices, indices_stack)] <- pairing[2]

    indices_stack <- c(indices_stack, indices)
  }

  # If a value is supplied for `otherwise`, apply
  # that value to all unmatched
  if (!is.null(otherwise)) {

    otherwise_indices <-
      which(!(seq_len(nrow(nodes)) %in% indices_stack))

    if (length(otherwise_indices) > 0) {
      vector_to_recode[otherwise_indices] <-
        otherwise
    }
  }

  # Rename the `vector_to_recode` object
  recoded_vector <- vector_to_recode

  if (!is.null(node_attr_to)) {

    # Stop function if `node_attr_to` is
    # `id` or `nodes`
    if (any(c("id", "nodes") %in% node_attr_to)) {

      cli::cli_abort(
        "You cannot use the names `id` or `nodes`.")
    }

    if (any(column_names_graph %in% node_attr_to)) {

      # Node attribute exists and values will be
      # overwritten in this case
      col_num_write_to <-
        which(column_names_graph %in% node_attr_to)

      nodes[, col_num_write_to] <-
        recoded_vector
    }

    if (!any(column_names_graph %in% node_attr_to)) {

      # Node attribute does not exist and values be
      # part of a new node attribute
      nodes <-
        cbind(nodes,
              data.frame(recoded_vector,
                         stringsAsFactors = FALSE))

      # Set the column name for the copied attr
      colnames(nodes)[ncol(nodes)] <- node_attr_to
    }
  } else {
    # The node attribute values will be overwritten
    # by the recoded value (no new node attrs)
    nodes[, col_num_recode] <- recoded_vector
  }

  # Replace the `nodes_df` object in the graph
  # with the `nodes` object
  graph$nodes_df <- nodes

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get the last set of edges created in a graph
#'
#' @description
#'
#' Get the last edges that were created in a graph object of class `dgr_graph`.
#' This function should ideally be used just after creating the edges.
#'
#' @inheritParams render_graph
#'
#' @return A vector of edge ID values.
#'
#' @examples
#' # Create a graph and add a cycle and then
#' # a tree in 2 separate function calls
#' graph <-
#'   create_graph() %>%
#'   add_cycle(
#'     n = 3,
#'     rel = "a") %>%
#'   add_balanced_tree(
#'     k = 2, h = 2,
#'     rel = "b")
#'
#' # Get the last edges created (all edges
#' # from the tree)
#' graph %>% get_last_edges_created()
#'
#' @export
get_last_edges_created <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  graph_transform_steps <-
    graph$graph_log %>%
    dplyr::mutate(
      step_created_edges = as.integer(function_used %in% edge_creation_functions()),
      step_deleted_edges = as.integer(function_used %in% edge_deletion_functions()),
      step_init_with_edges = as.integer(function_used %in% graph_init_functions() &
                                          edges > 0)
    ) %>%
    dplyr::filter(
      dplyr::if_any(
        .cols = c(step_created_edges, step_deleted_edges, step_init_with_edges),
        .fns = function(x) x == 1
        )
    ) %>%
    dplyr::select(-"version_id", -"time_modified", -"duration")

  if (nrow(graph_transform_steps) > 0) {

    if (graph_transform_steps %>%
        utils::tail(1) %>%
        dplyr::pull(step_deleted_edges) == 1) {

      abort("The previous graph transformation function resulted in a removal of edges.")

    } else {
      if (nrow(graph_transform_steps) > 1) {
        number_of_edges_created <-
          (graph_transform_steps %>%
             dplyr::select("edges") %>%
             utils::tail(2) %>%
             dplyr::pull("edges"))[2] -
          (graph_transform_steps %>%
             dplyr::select("edges") %>%
             utils::tail(2) %>%
             dplyr::pull("edges"))[1]
      } else {
        number_of_edges_created <-
          graph_transform_steps %>%
          dplyr::pull("edges")
      }
    }

    edge_id_values <-
      graph$edges_df %>%
      dplyr::select("id") %>%
      utils::tail(number_of_edges_created) %>%
      dplyr::pull("id")
  } else {
    edge_id_values <- NA
  }

  edge_id_values
}

#' Select nodes based on a walk distance from a specified node
#'
#' @description
#'
#' Select those nodes in the neighborhood of nodes connected a specified
#' distance from an initial node.
#'
#' @inheritParams render_graph
#' @param node The node from which the traversal will originate.
#' @param distance The maximum number of steps from the `node` for inclusion in
#'   the selection.
#' @param set_op The set operation to perform upon consecutive selections of
#'   graph nodes. This can either be as a `union` (the default), as an
#'   intersection of selections with `intersect`, or, as a `difference` on the
#'   previous selection, if it exists.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph containing
#' # a balanced tree
#' graph <-
#'   create_graph() |>
#'   add_balanced_tree(
#'     k = 2, h = 2)
#'
#' # Create a graph selection by
#' # selecting nodes in the
#' # neighborhood of node `1`, where
#' # the neighborhood is limited by
#' # nodes that are 1 connection
#' # away from node `1`
#' graph <-
#'   graph |>
#'   select_nodes_in_neighborhood(
#'     node = 1,
#'     distance = 1)
#'
#' # Get the selection of nodes
#' graph |> get_selection()
#'
#' # Perform another selection
#' # of nodes, this time with a
#' # neighborhood spanning 2 nodes
#' # from node `1`
#' graph <-
#'   graph |>
#'   clear_selection() |>
#'   select_nodes_in_neighborhood(
#'     node = 1,
#'     distance = 2)
#'
#' # Get the selection of nodes
#' graph |> get_selection()
#'
#' @export
select_nodes_in_neighborhood <- function(
    graph,
    node,
    distance,
    set_op = "union"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Obtain the input graph's node and edge
  # selection properties
  n_e_select_properties_in <-
    node_edge_selection_properties(graph = graph)

  # Create an empty list object
  nodes <- list()

  # Find nodes belonging to the neighborhood
  for (i in seq_len(distance)) {
    if (i == 1) {

      nodes[[i]] <- vector(mode = "integer")

      nodes[[i]] <-
        c(node,
          get_edges(
            graph,
            return_type = "df")[
              which(
                get_edges(
                  graph,
                  return_type = "df")[, 1] ==
                  node), 2],
          get_edges(
            graph,
            return_type = "df")[
              which(
                get_edges(
                  graph,
                  return_type = "df")[, 2] ==
                  node), 1])
    }

    if (i > 1) {
      for (j in seq_along(nodes[[i - 1]])) {
        if (j == 1) {
          nodes[[i]] <- vector(mode = "integer")
        }

        nodes[[i]] <-
          c(nodes[[i]],
            get_edges(
              graph,
              return_type = "df")[
                which(
                  get_edges(
                    graph,
                    return_type = "df")[, 1] ==
                    nodes[[i - 1]][j]), 2],
            get_edges(
              graph,
              return_type = "df")[
                which(
                  get_edges(
                    graph,
                    return_type = "df")[, 2] ==
                    nodes[[i - 1]][j]), 1])
      }
    }
  }

  # From list of nodes, obtain vector of unique
  # nodes as neighbors
  nodes_selected <- unique(unlist(nodes))

  # If no node ID values in `nodes_selected` return
  # the graph without a changed node selection
  if (length(nodes_selected) == 0) {
    return(graph)
  }

  # Obtain vector with node ID selection of nodes
  # already present
  nodes_prev_selection <- graph$node_selection$node

  # Incorporate selected nodes into graph's selection
  if (set_op == "union") {
    nodes_combined <-
      union(nodes_prev_selection, nodes_selected)
  } else if (set_op == "intersect") {
    nodes_combined <-
      intersect(nodes_prev_selection, nodes_selected)
  } else if (set_op == "difference") {
    nodes_combined <-
      base::setdiff(nodes_prev_selection, nodes_selected)
  }

  # Add the node ID values to the active selection
  # of nodes in `graph$node_selection`
  graph$node_selection <-
    replace_graph_node_selection(
      graph = graph,
      replacement = nodes_combined)

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Obtain the output graph's node and edge
  # selection properties
  n_e_select_properties_out <-
    node_edge_selection_properties(graph = graph)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  # Emit a message about the modification of a selection
  # if that option is set
  if (!is.null(graph$graph_info$display_msgs) &&
      graph$graph_info$display_msgs) {

    # Construct message body
    if (!n_e_select_properties_in[["node_selection_available"]] &&
        !n_e_select_properties_in[["edge_selection_available"]]) {

      msg_body <-
        glue::glue(
          "created a new selection of \\
        {n_e_select_properties_out[['selection_count_str']]}")

    } else if (n_e_select_properties_in[["node_selection_available"]] ||
               n_e_select_properties_in[["edge_selection_available"]]) {

      if (n_e_select_properties_in[["edge_selection_available"]]) {
        msg_body <-
          glue::glue(
            "modified an existing selection of\\
           {n_e_select_properties_in[['selection_count_str']]}:
           * {n_e_select_properties_out[['selection_count_str']]}\\
           are now in the active selection
           * used the `{set_op}` set operation")
      }

      if (n_e_select_properties_in[["node_selection_available"]]) {
        msg_body <-
          glue::glue(
            "created a new selection of\\
           {n_e_select_properties_out[['selection_count_str']]}:
           * this replaces\\
           {n_e_select_properties_in[['selection_count_str']]}\\
           in the prior selection")
      }
    }

    # Issue a message to the user
    emit_message(
      fcn_name = fcn_name,
      message_body = msg_body)
  }

  graph
}

#' Get a vector of edge ID values
#'
#' @description
#'
#' Obtain a vector of edge ID values from a graph object. An optional filter by
#' edge attribute can limit the set of edge ID values returned.
#'
#' @inheritParams render_graph
#' @param conditions an option to use filtering conditions for the retrieval of
#'   edges.
#'
#' @return A vector of edge ID values.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 4,
#'     type = "letter",
#'     color = c("red", "green", "grey", "blue"),
#'     value = c(3.5, 2.6, 9.4, 2.7))
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 3),
#'     to = c(4, 3, 1),
#'     rel = "leading_to",
#'     color = c("pink", "blue", "blue"),
#'     value = c(3.9, 2.5, 7.3))
#'
#' # Create a graph
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Get a vector of all edges in a graph
#' graph %>% get_edge_ids()
#'
#' # Get a vector of edge ID values using a
#' # numeric comparison (i.e., all edges with
#' # `value` attribute greater than 3)
#' get_edge_ids(
#'   graph,
#'   conditions = value > 3)
#'
#' # Get a vector of edge ID values using
#' # a match pattern (i.e., all edges with
#' # `color` attribute of `pink`)
#' get_edge_ids(
#'   graph,
#'   conditions = color == "pink")
#'
#' # Use multiple conditions to return edges
#' # with the desired attribute values
#' get_edge_ids(
#'   graph,
#'   conditions =
#'     color == "blue" &
#'     value > 5)
#'
#' @export
get_edge_ids <- function(
    graph,
    conditions = NULL
) {


  # If the graph contains no edges, return NA
  if (nrow(graph$edges_df) == 0) {
    return(NA)
  }

  # Extract edge data frame from the graph
  edges_df <- graph$edges_df

  # If conditions are provided then
  # pass in those conditions and filter the
  # data frame of `edges_df`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    edges_df <- dplyr::filter(.data = edges_df, {{ conditions }})
  }

  # If no edges remain then return NA
  if (nrow(edges_df) == 0) {
    return(NA)
  }

  edges_df %>% dplyr::pull("id")
}

#' Rename an edge attribute
#'
#' @description
#'
#' Within a graph's internal edge data frame (edf), rename an existing edge
#' attribute.
#'
#' @inheritParams render_graph
#' @param edge_attr_from The name of the edge attribute that will be renamed.
#' @param edge_attr_to The new name of the edge attribute column identified in
#'   `edge_attr_from`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 5,
#'     m = 8,
#'     set_seed = 23) |>
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "green")
#'
#' # Get the graph's internal edf
#' # to show which edge attributes
#' # are available
#' graph |> get_edge_df()
#'
#' # Rename the `color` node
#' # attribute as `weight`
#' graph <-
#'   graph |>
#'   rename_edge_attrs(
#'     edge_attr_from = color,
#'     edge_attr_to = labelfontcolor)
#'
#' # Get the graph's internal
#' # edf to show that the edge
#' # attribute had been renamed
#' graph |> get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
rename_edge_attrs <- function(
    graph,
    edge_attr_from,
    edge_attr_to
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Get the requested `edge_attr_from`
  edge_attr_from <-
    rlang::enquo(edge_attr_from) |> rlang::get_expr() |> as.character()

  # Get the requested `edge_attr_to`
  edge_attr_to <-
    rlang::enquo(edge_attr_to) |> rlang::get_expr() |> as.character()

  # Stop function if `edge_attr_from` and
  # `edge_attr_to` are identical
  if (edge_attr_from == edge_attr_to) {

    cli::cli_abort(
      "`edge_attr_from` must be different than `edge_attr_to`.")
  }

  # Stop function if `edge_attr_to` is `from`, `to`,
  # or any other column name in the graph's edge
  # data frame
  if (any(colnames(get_edge_df(graph)) %in%
          edge_attr_to)) {

    cli::cli_abort(
      "You cannot use that name for `edge_attr_to`.")
  }

  # Stop function if `edge_attr_from` is `from`, `to`
  # or `rel`
  forbidden_vals <- c("from", "to", "rel")
  if (any(forbidden_vals %in% edge_attr_from)) {

    cli::cli_abort(
      "`edge_attr_from` cannot be {.or {.val {forbidden_vals}}}.")
  }

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  # Extract the graph's edf
  edges <- get_edge_df(graph)

  # Get column names from the graph's ndf
  column_names_graph <- colnames(edges)

  # Stop function if `edge_attr_from` is not one
  # of the graph's columns
  if (!any(column_names_graph %in% edge_attr_from)) {

    cli::cli_abort(
      "The edge attribute to rename is not in the edf.")
  }

  # Set the column name for the renamed attr
  colnames(edges)[
    which(colnames(edges) %in%
            edge_attr_from)] <- edge_attr_to

  # Modify the graph's edf
  graph$edges_df <- edges

  # Update the `last_node` counter
  graph$last_node <- nodes_created

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get all nodes associated with connected components
#'
#' @description
#'
#' Determine which nodes in a graph belong to different weakly connected
#' components (i.e., distinct sets of nodes with traversable paths to and from
#' each node in the set).
#'
#' @inheritParams render_graph
#'
#' @return A data frame with nodes and their membership in different weakly
#'   connected components.
#'
#' @examples
#' # Create a graph with 2 cycles
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 4) |>
#'   add_cycle(n = 3)
#'
#' # Check if the graph is connected
#' graph |>
#'   is_graph_connected()
#'
#' # Get the graph's weakly-connected
#' # components
#' graph |> get_w_connected_cmpts()
#'
#' @export
get_w_connected_cmpts <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Create a graph where only mandatory
  # node and edge attributes are retained;
  # transform to an igraph object
  ig_graph <-
    create_graph(
      nodes_df = graph |>
        get_node_df() |>
        dplyr::select(id, type, label),
      edges_df = graph |>
        get_edge_df() |>
        dplyr::select(id, from, to, rel),
      directed = is_graph_directed(graph)) |>
    to_igraph()

  # Get the component list from the graph
  components <-
    ig_graph |>
    igraph::components(mode = "weak")

  # Create the output data frame
  data.frame(
    id = as.integer(names(components$membership)),
    wc_component = components$membership,
    stringsAsFactors = FALSE)
}

#' Reverse the direction of selected edges in a graph using an edge selection
#'
#' @description
#'
#' Using a directed graph with a selection of edges as input, reverse the
#' direction of those selected edges in input graph.
#'
#' This function makes use of an active selection of edges (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with a
#' # directed tree
#' graph <-
#'   create_graph() |>
#'   add_balanced_tree(
#'     k = 2, h = 2)
#'
#' # Inspect the graph's edges
#' graph |> get_edges()
#'
#' # Select all edges associated
#' # with nodes `1` and `2`
#' graph <-
#'   graph |>
#'   select_edges_by_node_id(
#'     nodes = 1:2)
#'
#' # Reverse the edge directions
#' # of the edges associated with
#' # nodes `1` and `2`
#' graph <-
#'   graph |>
#'   rev_edge_dir_ws()
#'
#' # Inspect the graph's edges
#' # after their reversal
#' graph |> get_edges()
#'
#' @family edge creation and removal
#'
#' @export
rev_edge_dir_ws <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid edge selection
  check_graph_contains_edge_selection(graph)

  # If graph is undirected, stop function
  if (!graph$directed) {
    cli::cli_abort("The input graph must be a directed graph.")
  }

  # Get the graph nodes in the `from` and `to` columns
  # of the edf
  from <- get_edges(graph, return_type = "df")[, 1]
  to <- get_edges(graph, return_type = "df")[, 2]

  # Extract the graph's edge data frame
  edges <- get_edge_df(graph)

  # Get edge ID values in edge selection
  edge_ids <- suppressMessages(get_selection(graph))

  # Selectively modify the edge direction and create
  # a new edf
  edges_new <-
    edges |>
    dplyr::filter(id %in% edge_ids) |>
    dplyr::filter(from != to) |>
    dplyr::rename(from = "to", to = "from") |>
    dplyr::relocate("id", "from", "to") |>
    dplyr::bind_rows(edges |> dplyr::filter(!(id %in% edge_ids)))

  # Modify the graph object
  graph$edges_df <- edges_new

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Create a random growing graph with m edges added per step
#'
#' @description
#'
#' To an existing graph object, add a graph built by adding `m` new edges at
#' each time step (where a node is added).
#'
#' @inheritParams node_edge_aes_data
#' @inheritParams render_graph
#' @param n The number of nodes comprising the generated graph.
#' @param m The number of edges added per time step.
#' @param citation A logical value (default is `FALSE`) that governs whether a
#'   citation graph is to be created. This is where new edges specifically
#'   originate from the newly added node in the most recent time step.
#' @param type An optional string that describes the entity type for all the
#'   nodes to be added.
#' @param label A logical value where setting to `TRUE` ascribes node IDs to the
#'   label and `FALSE` yields a blank label.
#' @param rel An optional string for providing a relationship label to all edges
#'   to be added.
#' @param set_seed Supplying a value sets a random seed of the
#'   `Mersenne-Twister` implementation.
#' @examples
#' # Create a random, growing
#' # citation graph with 100
#' # nodes, adding an edge after
#' # each node addition
#' growing_graph <-
#'   create_graph() |>
#'   add_growing_graph(
#'     n = 100,
#'     m = 1,
#'     citation = TRUE,
#'     set_seed = 23)
#'
#' # Get a count of nodes
#' growing_graph |> count_nodes()
#'
#' # Get a count of edges
#' growing_graph |> count_edges()
#'
#' @export
add_growing_graph <- function(
    graph,
    n,
    m = 1,
    citation = FALSE,
    type = NULL,
    label = TRUE,
    rel = NULL,
    node_aes = NULL,
    edge_aes = NULL,
    node_data = NULL,
    edge_data = NULL,
    set_seed = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If a seed value is supplied, set a seed
  if (!is.null(set_seed)) {
    set.seed(set_seed, kind = "Mersenne-Twister")
  }

  # Stop if n is too small
  check_number_whole(n, min = 1)

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  # Get the number of edges ever created for
  # this graph
  edges_created <- graph$last_edge

  # Get the graph's global attributes
  global_attrs <- graph$global_attrs

  # Get the graph's log
  graph_log <- graph$graph_log

  # Get the graph's info
  graph_info <- graph$graph_info

  # Get the graph's state of being directed
  # or undirected
  graph_directed <- graph$directed

  # Create the graph to be added
  sample_growing_igraph <-
    igraph::sample_growing(
      n = n,
      m = m,
      directed = graph_directed,
      citation = citation)

  sample_growing_graph <- from_igraph(sample_growing_igraph)

  # Add in a static `type` value for all new nodes
  if (!is.null(type)) {
    sample_growing_graph$nodes_df$type <- as.character(type[1])
  }

  # Add in a static `rel` value for all new nodes
  if (!is.null(rel)) {
    sample_growing_graph$edges_df$rel <- as.character(rel[1])
  }

  # If `label` is requested, use the node ID to
  # create a unique label for all new nodes
  if (label) {
    sample_growing_graph$nodes_df$label <-
      sample_growing_graph$nodes_df$id |> as.character()
  }

  n_nodes <- nrow(sample_growing_graph$nodes_df)

  n_edges <- nrow(sample_growing_graph$edges_df)

  # Collect node aesthetic attributes
  if (!is.null(node_aes)) {

    node_aes_tbl <- dplyr::as_tibble(node_aes)

    if (nrow(node_aes_tbl) < nrow(sample_growing_graph$nodes_df)) {

      node_aes$index__ <- seq_len(nrow(sample_growing_graph$nodes_df))

      node_aes_tbl <-
        dplyr::as_tibble(node_aes) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_aes_tbl)) {
      node_aes_tbl$id <- NULL
    }
  }

  # Collect node data attributes
  if (!is.null(node_data)) {

    node_data_tbl <- dplyr::as_tibble(node_data)

    if (nrow(node_data_tbl) < nrow(sample_growing_graph$nodes_df)) {

      node_data$index__ <- seq_len(nrow(sample_growing_graph$nodes_df))

      node_data_tbl <-
        dplyr::as_tibble(node_data) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(node_data_tbl)) {
      node_data_tbl$id <- NULL
    }
  }

  # Collect edge aesthetic attributes
  if (!is.null(edge_aes)) {

    edge_aes_tbl <- dplyr::as_tibble(edge_aes)

    if (nrow(edge_aes_tbl) < nrow(sample_growing_graph$edges_df)) {

      edge_aes$index__ <- seq_len(nrow(sample_growing_graph$edges_df))

      edge_aes_tbl <-
        dplyr::as_tibble(edge_aes) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_aes_tbl)) {
      edge_aes_tbl$id <- NULL
    }
  }

  # Collect edge data attributes
  if (!is.null(edge_data)) {

    edge_data_tbl <- dplyr::as_tibble(edge_data)

    if (nrow(edge_data_tbl) < nrow(sample_growing_graph$edges_df)) {

      edge_data$index__ <- seq_len(nrow(sample_growing_graph$edges_df))

      edge_data_tbl <-
        dplyr::as_tibble(edge_data) |>
        dplyr::select(-"index__")
    }

    if ("id" %in% colnames(edge_data_tbl)) {
      edge_data_tbl$id <- NULL
    }
  }

  # Add node aesthetics if available
  if (exists("node_aes_tbl")) {

    sample_growing_graph$nodes_df <-
      sample_growing_graph$nodes_df |>
      dplyr::bind_cols(node_aes_tbl)
  }

  # Add node data if available
  if (exists("node_data_tbl")) {

    sample_growing_graph$nodes_df <-
      sample_growing_graph$nodes_df |>
      dplyr::bind_cols(node_data_tbl)
  }

  # Add edge aesthetics if available
  if (exists("edge_aes_tbl")) {

    sample_growing_graph$edges_df <-
      sample_growing_graph$edges_df |>
      dplyr::bind_cols(edge_aes_tbl)
  }

  # Add edge data if available
  if (exists("edge_data_tbl")) {

    sample_growing_graph$edges_df <-
      sample_growing_graph$edges_df |>
      dplyr::bind_cols(edge_data_tbl)
  }

  # If the input graph is not empty, combine graphs
  # using the `combine_graphs()` function
  if (is_graph_empty(graph)) {
    graph <- sample_growing_graph
  } else {
    graph <- combine_graphs(graph, sample_growing_graph)
  }

  # Update the `last_node` counter
  graph$last_node <- nodes_created + n_nodes

  # Update the `last_edge` counter
  graph$last_edge <- edges_created + n_edges

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph_log <-
    add_action_to_log(
      graph_log = graph_log,
      version_id = nrow(graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = n_nodes,
      d_e = n_edges)

  graph$global_attrs <- global_attrs
  graph$graph_log <- graph_log
  graph$graph_info <- graph_info

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Combine two graphs into a single graph
#'
#' @description
#'
#' Combine two graphs in order to make a new graph.
#'
#' @param x A `DiagrammeR` graph object to which another graph will be unioned.
#'   This graph should be considered the graph from which global graph
#'   attributes will be inherited in the resulting graph.
#' @param y A `DiagrammeR` graph object that is to be unioned with the graph
#'   supplied as `x`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with a cycle
#' # containing 6 nodes
#' graph_cycle <-
#'  create_graph() %>%
#'    add_cycle(n = 6)
#'
#' # Create a random graph with
#' # 8 nodes and 15 edges using the
#' # `add_gnm_graph()` function
#' graph_random <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 8,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Combine the two graphs in a
#' # union operation
#' combined_graph <-
#'   combine_graphs(
#'     graph_cycle,
#'     graph_random)
#'
#' # Get the number of nodes in
#' # the combined graph
#' combined_graph %>% count_nodes()
#'
#' # The `combine_graphs()`
#' # function will renumber
#' # node ID values in graph `y`
#' # during the union; this ensures
#' # that node ID values are unique
#' combined_graph %>% get_node_ids()
#'
#' @export
combine_graphs <- function(
    x,
    y
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object `x` is valid
  if (!graph_object_valid(x)) {
    cli::cli_abort("The graph object supplied to `x` is not valid.")
  }

  # Validation: Graph object `y` is valid
  if (!graph_object_valid(y)) {
    cli::cli_abort("The graph object supplied to `y` is not valid.")
  }

  # Get the number of nodes ever created for
  # graph `x`
  nodes_created <- x$last_node

  # Get the number of nodes in the graph
  nodes_graph_1 <- x %>% count_nodes()

  # Get the number of edges ever created for
  # graph `x`
  edges_created <- x$last_edge

  # Get the number of edges in the graph
  edges_graph_1 <- x %>% count_edges()

  # Get the node data frame for graph `x`
  x_nodes_df <- get_node_df(x)

  # Get the node data frame for graph `y`
  y_nodes_df <- get_node_df(y)

  # Is label a copy of node IDs in graph `y`?
  y_label_node <-
    all(as.character(y_nodes_df[, 1]) == y_nodes_df[, 3]) &&
    !anyNA(y_nodes_df[, 3])

  # Add a new node attribute `new_node_id`
  y_nodes_df$new_node_id <-
    seq(nodes_created + 1L,
        nodes_created + nrow(y_nodes_df))

  # Get the edge data frame for graph `x`
  x_edges_df <- get_edge_df(x)

  # Get the edge data frame for graph `y`
  y_edges_df <- get_edge_df(y)

  y_edges_df <-
    dplyr::inner_join(
      y_edges_df,
      y_nodes_df,
      by = c("from" = "id")) %>%
    dplyr::rename(from_new = "new_node_id") %>%
    dplyr::select(-"type", -"label")

  # Rename `id` if it has a `.x` suffix
  if ("id.x" %in% colnames(y_edges_df)) {

    y_edges_df <-
      y_edges_df %>%
      dplyr::rename(id = "id.x")
  }

  y_edges_df <-
    dplyr::inner_join(
      y_edges_df,
      y_nodes_df,
      by = c("to" = "id")) %>%
    dplyr::rename(to_new = "new_node_id") %>%
    dplyr::select(-"type", -"label")

  # Rename `id` if it has a `.x` suffix
  if ("id.x" %in% colnames(y_edges_df)) {

    y_edges_df <-
      y_edges_df %>%
      dplyr::rename(id = "id.x")
  }

  # Copy new node IDs to `from` and `to` edge attrs
  y_edges_df$from <- y_edges_df$from_new
  y_edges_df$to <- y_edges_df$to_new

  # Remove columns ending with `.x` or `_new`
  y_edges_df <-
    y_edges_df %>%
    dplyr::select(
      !dplyr::ends_with(c(".x", "_new")))

  # Rename column names with `.y` suffixes
  colnames(y_edges_df) <-
    gsub(".y$", "", colnames(y_edges_df))

  y_edges_df$id <- y_edges_df$id + edges_created

  # Copy new node IDs to `nodes` node attr
  y_nodes_df$id <- y_nodes_df$new_node_id

  # Remove the last column from `y_nodes_df`
  y_nodes_df <-
    y_nodes_df[, -ncol(y_nodes_df)]

  # If label is a copy of node ID in graph `y`,
  # rewrite labels to match new node ID values
  if (y_label_node) {
    y_nodes_df[, 3] <- as.character(y_nodes_df[, 1])
  }

  # Combine the node data frames for both graphs
  combined_nodes <-
    dplyr::bind_rows(x_nodes_df, y_nodes_df)

  # Combine the edge data frames for both graphs
  combined_edges <-
    dplyr::bind_rows(x_edges_df, y_edges_df)

  # Modify the graph object and inherit attributes
  # from the first graph provided (`x`)
  x$nodes_df <- combined_nodes
  x$edges_df <- combined_edges
  x$directed <- is_graph_directed(x) && is_graph_directed(y)
  x$last_node <- nrow(combined_nodes)
  x$last_edge <- nrow(combined_edges)

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- x %>% count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Get the updated number of edges in the graph
  edges_graph_2 <- x %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  x$graph_log <-
    add_action_to_log(
      graph_log = x$graph_log,
      version_id = nrow(x$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(x$nodes_df),
      edges = nrow(x$edges_df),
      d_n = nodes_added,
      d_e = edges_added)

  # Write graph backup if the option is set
  if (x$graph_info$write_backups) {
    save_graph_as_rds(graph = x)
  }

  x
}

#' Traverse from one or more selected nodes onto adjacent, inward nodes
#'
#' @description
#'
#' From a graph object of class `dgr_graph` move along inward edges from one or
#' more nodes present in a selection to other connected nodes, replacing the
#' current nodes in the selection with those nodes traversed to. An optional
#' filter by node attribute can limit the set of nodes traversed to.
#'
#' This traversal function makes use of an active selection of nodes. After the
#' traversal, depending on the traversal conditions, there will either be a
#' selection of nodes or no selection at all.
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#' @param conditions An option to use filtering conditions for the traversal.
#' @param copy_attrs_from providing a node attribute name will copy those node
#'   attribute values to the traversed nodes. Any values extant on the nodes
#'   traversed to will be replaced.
#' @param copy_attrs_as If a node attribute name is provided in
#'   `copy_attrs_from`, this option will allow the copied attribute values to be
#'   written under a different attribute name. If the attribute name provided in
#'   `copy_attrs_as` does not exist in the graph's ndf, the new node attribute
#'   will be created with the chosen name.
#' @param agg If a node attribute is provided to `copy_attrs_from`, then an
#'   aggregation function is required since there may be cases where multiple
#'   edge attribute values will be passed onto the traversed node(s). To pass
#'   only a single value, the following aggregation functions can be used:
#'   `sum`, `min`, `max`, `mean`, or `median`.
#' @param add_to_selection An option to either add the traversed to nodes to the
#'   active selection of nodes (`TRUE`) or switch the active selection entirely
#'   to those traversed to nodes (`FALSE`, the default case).
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(
#'     n = 2,
#'     type = "a",
#'     label = c("asd", "iekd")) |>
#'   add_n_nodes(
#'     n = 3,
#'     type = "b",
#'     label = c("idj", "edl", "ohd")) |>
#'   add_edges_w_string(
#'     edges = "1->2 1->3 2->4 2->5 3->5",
#'     rel = c(NA, "A", "B", "C", "D"))
#'
#' # Create a data frame with node ID values
#' # representing the graph edges (with `from`
#' # and `to` columns), and, a set of numeric values
#' df_edges <-
#'   data.frame(
#'     from = c(1, 1, 2, 2, 3),
#'     to = c(2, 3, 4, 5, 5),
#'     values = round(rnorm(5, 5), 2))
#'
#' # Create a data frame with node ID values
#' # representing the graph nodes (with the `id`
#' # columns), and, a set of numeric values
#' df_nodes <-
#'   data.frame(
#'     id = 1:5,
#'     values = round(rnorm(5, 7), 2))
#'
#' # Join the data frame to the graph's internal
#' # edge data frame (edf)
#' graph <-
#'   graph |>
#'   join_edge_attrs(df = df_edges) |>
#'   join_node_attrs(df = df_nodes)
#'
#' # Show the graph's internal node data frame
#' graph |> get_node_df()
#'
#' # Show the graph's internal edge data frame
#' graph |> get_edge_df()
#'
#' # Perform a simple traversal from node `4` to
#' # inward adjacent edges with no conditions
#' # on the nodes traversed to
#' graph |>
#'   select_nodes_by_id(nodes = 4) |>
#'   trav_in() |>
#'   get_selection()
#'
#' # Traverse from node `5` to inbound-facing
#' # nodes, filtering to those nodes that have
#' # numeric values greater than `5.0` for
#' # the `values` node attribute
#' graph |>
#'   select_nodes_by_id(nodes = 4) |>
#'   trav_in(
#'     conditions = values > 5.0) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any inbound
#' # nodes, filtering to those nodes that
#' # have a `type` attribute of `b`
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in(
#'     conditions = type == "b") |>
#'   get_selection()
#'
#' # Traverse from node `5` to any inbound
#' # nodes, filtering to those nodes that
#' # have a degree of `2`
#' node_degrees <- graph |>
#'   get_node_info() |>
#'   dplyr::select(id, deg)
#'
#' graph <- graph |>
#'   join_node_attrs(df = node_degrees)
#'
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in(
#'     conditions = deg == 2) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any inbound
#' # nodes, and use multiple conditions for the
#' # traversal
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in(
#'     conditions =
#'       type == "a" &
#'       values > 6.0) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any inbound
#' # nodes, and use multiple conditions with
#' # a single-length vector
#' graph |>
#'   select_nodes_by_id(nodes = 5) |>
#'   trav_in(
#'     conditions =
#'       type == "b" | values > 6.0) |>
#'   get_selection()
#'
#' # Traverse from node `5` to any inbound
#' # nodes, and use a regular expression as
#' # a filtering condition
#' graph |>
#'   select_nodes_by_id(nodes = 2) |>
#'   trav_in(
#'     conditions = grepl("^i.*", label)) |>
#'   get_selection()
#'
#' # Create another simple graph to demonstrate
#' # copying of node attribute values to traversed
#' # nodes
#' graph <-
#'   create_graph() |>
#'   add_node() |>
#'   select_nodes() |>
#'   add_n_nodes_ws(
#'     n = 2,
#'     direction = "from") |>
#'   clear_selection() |>
#'   select_nodes_by_id(nodes = 2:3) |>
#'   set_node_attrs_ws(
#'     node_attr = value,
#'     value = 5)
#'
#' # Show the graph's internal node data frame
#' graph |> get_node_df()
#'
#' # Show the graph's internal edge data frame
#' graph |> get_edge_df()
#'
#' # Perform a traversal from the outer nodes
#' # (`2` and `3`) to the central node (`1`) while
#' # also applying the node attribute `value` to
#' # node `1` (summing the `value` of 5 from
#' # both nodes before applying the value to the
#' # target node)
#' graph <-
#'   graph |>
#'   trav_in(
#'     copy_attrs_from = value,
#'     agg = "sum")
#'
#' # Show the graph's internal node data frame
#' # after this change
#' graph |> get_node_df()
#'
#' @export
trav_in <- function(
    graph,
    conditions = NULL,
    copy_attrs_from = NULL,
    copy_attrs_as = NULL,
    agg = "sum",
    add_to_selection = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has a valid node selection
  check_graph_contains_node_selection(graph, c(
        "any traversal requires an active selection",
        "this type of traversal requires a selection of nodes"))

  # Get the requested `copy_attrs_from`
  copy_attrs_from <-
    rlang::enquo(copy_attrs_from) |> rlang::get_expr() |> as.character()

  # Get the requested `copy_attrs_as`
  copy_attrs_as <-
    rlang::enquo(copy_attrs_as) |> rlang::get_expr() |> as.character()

  if (length(copy_attrs_from) == 0) {
    copy_attrs_from <- NULL
  }

  if (length(copy_attrs_as) == 0) {
    copy_attrs_as <- NULL
  }

  if (!is.null(copy_attrs_as) && !is.null(copy_attrs_from)) {
    if (copy_attrs_as == copy_attrs_from) {
      copy_attrs_as <- NULL
    }
  }

  # Get the selection of nodes as the starting
  # nodes for the traversal
  starting_nodes <- graph$node_selection$node

  # Get the graph's edge data frame
  edf <- graph$edges_df

  # Get the graph's node data frame
  ndf <- graph$nodes_df

  # Find all nodes that are connected to the
  # starting nodes via incoming edges
  valid_nodes <-
    edf |>
    dplyr::filter(to != from) |>
    dplyr::filter(to %in% starting_nodes) |>
    dplyr::distinct(from)

  valid_nodes <-
    dplyr::as_tibble(valid_nodes) |>
    dplyr::rename(id = "from") |>
    dplyr::inner_join(ndf, by = "id")

  # If no rows returned, then there are no
  # valid traversals, so return the same graph
  # object without modifying the selection
  if (nrow(valid_nodes) == 0) {
    return(graph)
  }

  # If traversal conditions are provided then
  # pass in those conditions and filter the
  # data frame of `valid_nodes`
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    valid_nodes <- dplyr::filter(.data = valid_nodes, {{ conditions }})
  }

  # If the option is taken to copy node attribute
  # values to the traversed nodes, perform the join
  # operations
  if (!is.null(copy_attrs_from)) {

    nodes <-
      valid_nodes |>
      dplyr::select(id) |>
      dplyr::inner_join(edf |> dplyr::select("from", "to"), by = c("id" = "from")) |>
      dplyr::inner_join(ndf |> dplyr::select("id", !!enquo(copy_attrs_from)), by = c("to" = "id")) |>
      dplyr::select("id", !!enquo(copy_attrs_from))

    # If the values to be copied are numeric,
    # perform aggregation on the values
    if (nodes[, 2] |> unlist() |> is.numeric()) {
      nodes <-
        nodes |>
        dplyr::group_by(id) |>
        dplyr::summarize(!!copy_attrs_from :=
                           match.fun(!!agg)(!!as.name(copy_attrs_from),
                                             na.rm = TRUE)) |>
        dplyr::ungroup()
    }

    nodes <-
      nodes |>
      dplyr::right_join(ndf, by = "id") |>
      dplyr::relocate("id", "type", "label") |>
      dplyr::arrange(id) |>
      as.data.frame(stringsAsFactors = FALSE)

    # Get column numbers that end with ".x" or ".y"
    split_var_x_col <-
      grep("\\.x$", colnames(nodes))

    split_var_y_col <-
      grep("\\.y$", colnames(nodes))

    if (is.null(copy_attrs_as)) {

      # Selectively merge in values to the existing
      # edge attribute column
      for (i in seq_len(nrow(nodes))) {
        if (!is.na(nodes[i, split_var_x_col])) {
          nodes[i, split_var_y_col] <- nodes[i, split_var_x_col]
        }
      }
    }

    if (!is.null(copy_attrs_as)) {

      # Reorder the columns generated
      nodes <-
        nodes[, c(c(seq_len(ncol(nodes) - 2)), split_var_y_col, split_var_x_col)]
    }

    # Rename the ".y" column
    colnames(nodes)[split_var_y_col] <- copy_attrs_from

    if (is.null(copy_attrs_as)) {

      # Drop the ".x" column
      nodes <- nodes[-split_var_x_col]

    } else {

      # Rename the two columns
      colnames(nodes)[split_var_x_col] <- copy_attrs_from
      colnames(nodes)[split_var_y_col] <- copy_attrs_as
    }

    # Update the graph's internal node data frame
    graph$nodes_df <- nodes
  }

  # If no rows returned, then there are no
  # valid traversals, so return the same graph
  # object without modifying the selection
  if (nrow(valid_nodes) == 0) {
    return(graph)
  }

  # Obtain vector with node ID selection of nodes
  # already present
  nodes_prev_selection <- graph$node_selection$node

  if (add_to_selection) {

    # If TRUE supplied to `add_to_selection` add
    # the nodes to which we traversed to the
    # previous selection
    nodes_combined <- union(nodes_prev_selection, valid_nodes$id)

    graph$node_selection <-
      replace_graph_node_selection(
        graph = graph,
        replacement = nodes_combined)

  } else {

    # Add the node ID values to the active selection
    # of nodes in `graph$node_selection`
    graph$node_selection <-
      replace_graph_node_selection(
        graph = graph,
        replacement = valid_nodes$id)
  }

  # Replace `graph$edge_selection` with an empty df
  graph$edge_selection <- create_empty_esdf()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Is the graph a weighted graph?
#'
#' @description
#'
#' Provides a logical value on whether the graph is weighted. A graph is
#' considered to be weighted when it contains edges that all have a edge
#' `weight` attribute with numerical values assigned for all edges.
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create a graph where the edges have
#' # a `weight` attribute
#' graph <-
#'   create_graph() %>%
#'   add_cycle(n = 5) %>%
#'   select_edges() %>%
#'   set_edge_attrs_ws(
#'     edge_attr = weight,
#'     value = c(3, 5, 2, 9, 6)) %>%
#'   clear_selection()
#'
#' # Determine whether the graph
#' # is a weighted graph
#' graph %>% is_graph_weighted()
#'
#' # Create graph where the edges do
#' # not have a `weight` attribute
#' graph <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' # Determine whether this graph
#' # is weighted
#' graph %>% is_graph_weighted()
#'
#' @export
is_graph_weighted <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty or contains no edges, it cannot be
  # classified as a weighted graph
  if (nrow(graph$nodes_df) == 0 || nrow(graph$edges_df) == 0) {
    return(FALSE)
  }

  # Determine if the `weight` edge
  # attribute exists and contains non-NA
  # numerical values
  if ("weight" %in% colnames(graph$edges_df)) {

    if (inherits(graph$edges_df$weight, c("numeric", "integer")) &&
        !anyNA(graph$edges_df$weight)) {
      return(TRUE)
    } else {
      return(FALSE)
    }
  } else {
    return(FALSE)
  }
}

#' Get graph adhesion
#'
#' @description
#'
#' Get the adhesion of a graph, which is the minimum number of edges needed to
#' remove to obtain a graph which is not strongly connected. This is the same as
#' the edge connectivity of the graph.
#'
#' @inheritParams render_graph
#'
#' @return A single numeric value representing the minimum number of edges to
#'   remove.
#'
#' @examples
#' # Create a cycle graph
#' graph <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' # Determine the graph's adhesion
#' graph |> get_adhesion()
#'
#' # Create a full graph and then
#' # get the adhesion for that
#' create_graph() |>
#'   add_full_graph(n = 8) |>
#'   get_adhesion()
#'
#' @export
get_adhesion <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph is empty, then return NA
  if (nrow(graph$nodes_df) == 0) {
    return(NA_real_)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the number of edges required for
  # removal to transition to state where
  # the graph is no longer strongly connected
  igraph::edge_connectivity(ig_graph)
}

#' Rename a node attribute
#'
#' @description
#'
#' Within a graph's internal node data frame (ndf), rename an existing node
#' attribute.
#'
#' @inheritParams render_graph
#' @param node_attr_from The name of the node attribute that will be renamed.
#' @param node_attr_to The new name of the node attribute column identified in
#'   `node_attr_from`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 5,
#'     m = 8,
#'     set_seed = 23) %>%
#'   set_node_attrs(
#'     node_attr = shape,
#'     values = "circle") %>%
#'   set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = count_nodes(.),
#'       mean = 5,
#'       sd = 1) %>% round(1))
#'
#' # Get the graph's internal ndf
#' # to show which node attributes
#' # are available
#' graph %>% get_node_df()
#'
#' # Rename the `value` node
#' # attribute as `weight`
#' graph <-
#'   graph %>%
#'   rename_node_attrs(
#'     node_attr_from = value,
#'     node_attr_to = weight)
#'
#' # Get the graph's internal
#' # ndf to show that the node
#' # attribute had been renamed
#' graph %>% get_node_df()
#'
#' @family node creation and removal
#'
#' @export
rename_node_attrs <- function(
    graph,
    node_attr_from,
    node_attr_to
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get the requested `node_attr_from`
  node_attr_from <-
    rlang::enquo(node_attr_from) %>% rlang::get_expr() %>% as.character()

  # Get the requested `node_attr_to`
  node_attr_to <-
    rlang::enquo(node_attr_to) %>% rlang::get_expr() %>% as.character()

  # Stop function if `node_attr_from` and
  # `node_attr_to` are identical
  if (node_attr_from == node_attr_to) {

    cli::cli_abort(
      "`node_attr_from` must be different than `node_attr_to` for renaming.")
  }

  # Stop function if `node_attr_to` is `id` or any
  # other column name in the graph's node data frame
  if (any(unique(c("id",
                   colnames(get_node_df(graph)))) %in%
          node_attr_to)) {

    cli::cli_abort(
      "You cannot use {node_attr_to} as a name for `node_attr_to`.")
  }

  # Stop function if `node_attr_from` is `id`, `label`
  # or `type`
  if (any(c("id", "label", "type") %in%
          node_attr_from)) {

    cli::cli_abort(
      "You cannot use {node_attr_from} as a name for `node_attr_from`.")
  }

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  # Extract the graph's ndf
  nodes <- get_node_df(graph)

  # Get column names from the graph's ndf
  column_names_graph <- colnames(nodes)

  # Stop function if `node_attr_from` is not one
  # of the graph's columns
  if (!any(column_names_graph %in% node_attr_from)) {

    cli::cli_abort(
      "The node attribute to rename is not in the ndf.")
  }

  # Set the column name for the renamed attr
  colnames(nodes)[
    which(colnames(nodes) %in%
            node_attr_from)] <- node_attr_to

  # Modify the graph's ndf
  graph$nodes_df <- nodes

  # Update the `last_node` counter
  graph$last_node <- nodes_created

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Convert a DiagrammeR graph to an igraph one
#'
#' @description
#'
#' Convert a DiagrammeR graph to an igraph graph object.
#'
#' @inheritParams render_graph
#'
#' @return An igraph object.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 36,
#'     m = 50,
#'     set_seed = 23)
#'
#' # Confirm that `graph` is a
#' # DiagrammeR graph by getting
#' # the object's class
#' class(graph)
#'
#' # Convert the DiagrammeR graph
#' # to an igraph object
#' ig_graph <- to_igraph(graph)
#'
#' # Get the class of the converted
#' # graph, just to be certain
#' class(ig_graph)
#'
#' # Get a summary of the igraph
#' # graph object
#' summary(ig_graph)
#'
#' @export
to_igraph <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Extract the graph's node data frame
  ndf <- graph$nodes_df

  # Extract the graph's edge data frame and
  # exclude the `id` column
  edf <-
    graph$edges_df %>%
    dplyr::select(-"id")

  igraph::graph_from_data_frame(
    d = edf,
    directed = graph$directed,
    vertices = ndf)
}

#' Add nodes from a node data frame to an existing graph object
#'
#' @description
#'
#' With a graph object of class `dgr_graph` add nodes from a node data frame to
#' that graph.
#'
#' @inheritParams render_graph
#' @param node_df A node data frame that is created using [create_node_df()].
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create an empty graph
#' graph <- create_graph()
#'
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(n = 2)
#'
#' # Add the node data frame to
#' # the graph object to create
#' # a graph with nodes
#' graph <-
#'   graph %>%
#'   add_node_df(
#'     node_df = ndf)
#'
#' # Inspect the graph's ndf
#' graph %>% get_node_df()
#'
#' # Create another ndf
#' ndf_2 <-
#'   create_node_df(n = 3)
#'
#' # Add the second node data
#' # frame to the graph object
#' # to add more nodes with
#' # attributes to the graph
#' graph <-
#'   graph %>%
#'   add_node_df(
#'     node_df = ndf_2)
#'
#' # View the graph's internal
#' # node data frame using the
#' # `get_node_df()` function
#' graph %>% get_node_df()
#'
#' @family node creation and removal
#'
#' @export
add_node_df <- function(
    graph,
    node_df
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the number of nodes ever created for
  # this graph
  nodes_created <- graph$last_node

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph %>% count_nodes()

  # Combine the incoming node data frame with the
  # existing node definitions in the graph object
  node_df[, 1] <-
    as.integer(nodes_created + seq_len(nrow(node_df)))

  node_df[, 2] <- as.character(node_df[, 2])
  node_df[, 3] <- as.character(node_df[, 3])

  graph$nodes_df <-
    dplyr::bind_rows(
      graph$nodes_df, node_df)

  # Update the `last_node` counter
  graph$last_node <-
    nodes_created + nrow(node_df)

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph %>% count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get the graph's ndf filtered by a selection of nodes
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get the graph's internal node data
#' frame that is filtered by the node ID values currently active as a selection.
#'
#' This function makes use of an active selection of nodes (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of nodes can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], or
#' [trav_in_until()].
#'
#' @inheritParams render_graph
#'
#' @return A node data frame.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 4,
#'     m = 4,
#'     set_seed = 23) |>
#'   set_node_attrs(
#'     node_attr = value,
#'     values = c(2.5, 8.2, 4.2, 2.4))
#'
#' # Select nodes with ID values
#' # `1` and `3`
#' graph <-
#'   graph |>
#'   select_nodes_by_id(
#'     nodes = c(1, 3))
#'
#' # Get the node data frame that's
#' # limited to the rows that correspond
#' # to the node selection
#' graph |> get_node_df_ws()
#'
#' @export
get_node_df_ws <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph object has a valid node selection
  check_graph_contains_node_selection(graph)

  # Extract the node data frame (ndf)
  # from the graph and get only those nodes
  # from the node selection
  graph$nodes_df |>
    dplyr::filter(id %in% graph$node_selection$node)
}

#' Get node IDs for successor nodes to the specified node
#'
#' @description
#'
#' Provides a vector of node IDs for all nodes that have a connection from the
#' given node.
#'
#' @inheritParams render_graph
#' @param node A node ID for the selected node.
#'
#' @return A vector of node ID values.
#'
#' @examples
#' # Set a seed
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a node data frame (ndf)
#' ndf <- create_node_df(n = 26)
#'
#' # Create an edge data
#' # frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = sample(
#'       1:26, replace = TRUE),
#'     to = sample(
#'       1:26, replace = TRUE))
#'
#' # From the ndf and edf,
#' # create a graph object
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Get sucessors for node
#' # `4` in the graph
#' graph |>
#'   get_successors(
#'     node = 4)
#'
#' # If there are no successors,
#' # NA is returned
#' graph |>
#'   get_successors(
#'     node = 1)
#'
#' @export
get_successors <- function(
    graph,
    node
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph contains no edges, return NA
  if (is.null(graph$edges_df)) {
    return(NA)
  }

  # Determine whether the graph has any nodes
  graph_is_not_empty <- !is_graph_empty(graph)

  # Determine whether `node` is in the graph
  node_is_in_graph <- node %in% graph$nodes_df$id

  # Obtain the node's successors
  if (graph_is_not_empty &&
      node_is_in_graph &&
      nrow(get_edge_info(graph)) > 0) {

    if (length(graph$edges_df[graph$edges_df$from ==
                              node, ]$to) == 0) {
      successors <- NA
    } else {
      successors <-
        graph$edges_df[graph$edges_df$from == node, ]$to
    }
    return(successors)
  }
}

#' Is the graph a directed acyclic graph?
#'
#' @description
#'
#' Provides a logical value on whether the graph is a directed acyclic graph
#' (DAG). The conditions for a graph that is a DAG are that it should be a
#' directed graph and it should not contain any cycles.
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create a directed graph containing
#' # only a balanced tree
#' graph_tree <-
#'   create_graph() |>
#'   add_balanced_tree(
#'     k = 2, h = 3)
#'
#' # Determine whether this graph
#' # is a DAG
#' graph_tree |>
#'   is_graph_dag()
#'
#' # Create a directed graph containing
#' # a single cycle
#' graph_cycle <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' # Determine whether this graph
#' # is a DAG
#' graph_cycle |>
#'   is_graph_dag()
#'
#' # Create an undirected graph
#' # containing a balanced tree
#' graph_tree_undirected <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_balanced_tree(
#'     k = 2, h = 2)
#'
#' # Determine whether this graph
#' # is a DAG
#' graph_tree_undirected |>
#'   is_graph_dag()
#'
#' @export
is_graph_dag <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If the graph contains no nodes, it
  # cannot be a DAG
  if (nrow(graph$nodes_df) == 0) {
    return(FALSE)
  }

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Determine whether the graph is
  # a simple graph
  igraph::is_dag(ig_graph)
}

#' Get articulation points
#'
#' @description
#'
#' Get the nodes in the graph that are identified as articulation points.
#'
#' @inheritParams render_graph
#'
#' @return a vector of node IDs.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 12,
#'     set_seed = 23) %>%
#'   set_node_attrs(
#'     node_attr = shape,
#'     values = "square")
#'
#' # Get the articulation points
#' # in the graph (i.e., those
#' # nodes that if any were to be
#' # removed, the graph would
#' # become disconnected)
#' graph %>%
#'   get_articulation_points()
#'
#' # For the articulation points,
#' # change the node shape to
#' # a `circle`
#' graph <-
#'   graph %>%
#'   select_nodes_by_id(
#'     nodes = get_articulation_points(.)) %>%
#'   set_node_attrs_ws(
#'     node_attr = shape,
#'     value = "circle")
#'
#' @export
get_articulation_points <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the graph's articulation points
  articulation_points <-
    igraph::articulation_points(ig_graph)

  articulation_points <-
    igraph::as_ids(articulation_points)

  sort(as.integer(articulation_points))
}

#' Get an edge data frame from a graph
#'
#' @description
#'
#' From a graph, obtain an edge data frame with all current edge attributes.
#'
#' @inheritParams render_graph
#'
#' @return An edge data frame.
#'
#' @examples
#' # Create a graph
#' graph <-
#'   create_graph() %>%
#'   add_n_nodes(
#'     n = 1,
#'     type = "a") %>%
#'   select_last_nodes_created() %>%
#'   add_n_nodes_ws(
#'     n = 5,
#'     direction = "from",
#'     type = "b") %>%
#'   select_edges_by_node_id(
#'     nodes = 3:5) %>%
#'   set_edge_attrs_ws(
#'     edge_attr = color,
#'     value = "green") %>%
#'   set_edge_attrs_ws(
#'     edge_attr = rel,
#'     value = "a") %>%
#'   invert_selection %>%
#'   set_edge_attrs_ws(
#'     edge_attr = color,
#'     value = "blue") %>%
#'   set_edge_attrs_ws(
#'     edge_attr = rel,
#'     value = "b") %>%
#'   clear_selection()
#'
#' # Get the graph's internal
#' # edge data frame (edf)
#' graph %>% get_edge_df()
#'
#' @export
get_edge_df <- function(graph) {
  check_graph_valid(graph)
  graph$edges_df
}

#' Export a graph to various image formats
#'
#' @description
#'
#' Export a graph to a variety of image formats such as PNG, PDF, SVG, and
#' PostScript.
#'
#' @inheritParams render_graph
#' @param file_name The name of the exported file (including it's extension).
#' @param file_type The type of file to be exported. Options for graph files
#'   are: `png`, `pdf`, `svg`, and `ps`.
#' @param title An optional title for the output graph.
#' @param width Output width in pixels or `NULL` for default. Only useful for
#'   export to image file formats `png`, `pdf`, `svg`, and `ps`.
#' @param height Output height in pixels or `NULL` for default. Only useful for
#'   export to image file formats `png`, `pdf`, `svg`, and `ps`.
#'
#' @examples
#' # Create a simple graph
#' graph <-
#'   create_graph() %>%
#'     add_path(
#'       n = 5,
#'       edge_aes = edge_aes(
#'         arrowhead = c(
#'           "normal", "vee",
#'           "tee", "dot"
#'         ),
#'         color = c(
#'         "red", "blue",
#'         "orange", "purple"
#'         )
#'       )
#'     )
#'
#' # Create a PDF file for
#' # the graph (`graph.pdf`)
#' # graph %>%
#' #   export_graph(
#' #     file_name = "graph.pdf",
#' #     title = "Simple Graph"
#' #   )
#'
#' # Create a PNG file for
#' # the graph (`mypng.png`)
#' # graph %>%
#' #   export_graph(
#' #     file_name = "mypng.png",
#' #     file_type = "PNG"
#' #   )
#'
#' @family Display and Save
#'
#' @export
export_graph <- function(
    graph,
    file_name = NULL,
    file_type = NULL,
    title = NULL,
    width = NULL,
    height = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If no `file_name` or `file_type` provided, default to
  # writing a PDF with a unique `file_name` value based
  # on user's current date/time
  if (is.null(file_name) && is.null(file_type)) {
    file_name <-
      paste0("graph_", format(Sys.time(), "%Y_%m_%d__%H_%M_%S"), ".pdf")
    file_type <- "pdf"
  }

  # If `file_name` provided but `file_type` is not, infer
  # the output file type based on the extension; if no
  # extension provided, default to PDF export
  if (is.null(file_type) && !is.null(file_name)) {
    if (grepl("\\.", file_name)) {
      file_type <- gsub(".*\\.([A-Za-z])", "\\1", file_name)
    } else {
      file_name <- paste0(file_name, ".pdf")
      file_type <- "pdf"
    }
  }

  if (tolower(file_type) == "png") {

    # Stop function if `DiagrammeRsvg` or `rsvg` package is not available
    rlang::check_installed(c("DiagrammeRsvg", "rsvg"), "to produce a PNG file.")

    if (!is.null(title)) {

      graph <- add_global_graph_attrs(graph, "label", title, "graph")
      graph <- add_global_graph_attrs(graph, "labelloc", "t", "graph")
      graph <- add_global_graph_attrs(graph, "labeljust", "c", "graph")
      graph <- add_global_graph_attrs(graph, "fontname", "Helvetica", "graph")
      graph <- add_global_graph_attrs(graph, "fontcolor", "gray30", "graph")
    }

    # Generate DOT code
    dot_code <- generate_dot(graph)

    # Produce a PNG file in the working directory
    rsvg::rsvg_png(
      charToRaw(DiagrammeRsvg::export_svg(grViz(dot_code))),
      file = file_name,
      width = width,
      height = height
    )
  }

  if (tolower(file_type) == "pdf") {

    # Stop function if `DiagrammeRsvg` or `rsvg` package is not available
    rlang::check_installed(c("DiagrammeRsvg", "rsvg"), "to produce a PDF file.")

    if (!is.null(title)) {
      graph <- add_global_graph_attrs(graph, "label", title, "graph")
      graph <- add_global_graph_attrs(graph, "labelloc", "t", "graph")
      graph <- add_global_graph_attrs(graph, "labeljust", "c", "graph")
      graph <- add_global_graph_attrs(graph, "fontname", "Helvetica", "graph")
      graph <- add_global_graph_attrs(graph, "fontcolor", "gray30", "graph")
    }

    # Generate DOT code
    dot_code <- generate_dot(graph)

    # Produce a PDF file in the working directory
    rsvg::rsvg_pdf(
      charToRaw(DiagrammeRsvg::export_svg(grViz(dot_code))),
      file = file_name,
      width = width,
      height = height
    )
  }

  if (tolower(file_type) == "svg") {

    # Stop function if `DiagrammeRsvg` or `rsvg` package is not available
    rlang::check_installed(c("DiagrammeRsvg", "rsvg"), "to produce a SVG file.")

    if (!is.null(title)) {

      graph <- add_global_graph_attrs(graph, "label", title, "graph")
      graph <- add_global_graph_attrs(graph, "labelloc", "t", "graph")
      graph <- add_global_graph_attrs(graph, "labeljust", "c", "graph")
      graph <- add_global_graph_attrs(graph, "fontname", "Helvetica", "graph")
      graph <- add_global_graph_attrs(graph, "fontcolor", "gray30", "graph")
    }

    # Generate DOT code
    dot_code <- generate_dot(graph)

    # Produce an SVG file in the working directory
    rsvg::rsvg_svg(
      charToRaw(DiagrammeRsvg::export_svg(grViz(dot_code))),
      file = file_name,
      width = width,
      height = height
    )
  }

  if (tolower(file_type) == "ps") {

    rlang::check_installed(c("DiagrammeRsvg", "rsvg"), "to produce a PS file.")

    if (!is.null(title)) {

      graph <- add_global_graph_attrs(graph, "label", title, "graph")
      graph <- add_global_graph_attrs(graph, "labelloc", "t", "graph")
      graph <- add_global_graph_attrs(graph, "labeljust", "c", "graph")
      graph <- add_global_graph_attrs(graph, "fontname", "Helvetica", "graph")
      graph <- add_global_graph_attrs(graph, "fontcolor", "gray30", "graph")
    }

    # Generate DOT code
    dot_code <- generate_dot(graph)

    # Produce a PS file in the working directory
    rsvg::rsvg_ps(
      charToRaw(DiagrammeRsvg::export_svg(grViz(dot_code))),
      file = file_name,
      width = width,
      height = height
    )
  }
}

#' Mutate edge attribute values for a selection of edges
#'
#' @description
#'
#' Within a graph's internal edge data frame (edf), mutate edge attribute values
#' only for edges in a selection by using one or more expressions.
#'
#' This function makes use of an active selection of edges (and the function
#' ending with `_ws` hints at this).
#'
#' Selections of edges can be performed using the following selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of edges can also be performed using the following traversal
#' (`trav_*()`) functions: [trav_out_edge()], [trav_in_edge()],
#' [trav_both_edge()], or [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#' @param ... Expressions used for the mutation of edge attributes. LHS of each
#'   expression is either an existing or new edge attribute name. The RHS can
#'   consist of any valid R code that uses edge attributes as variables.
#'   Expressions are evaluated in the order provided, so, edge attributes
#'   created or modified are ready to use in subsequent expressions.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 3 edges
#' # and then select edge `1`
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 4) %>%
#'   set_edge_attrs(
#'     edge_attr = width,
#'     values = c(3.4, 2.3, 7.2)) %>%
#'   select_edges(edges = 1)
#'
#' # Get the graph's internal edf
#' # to show which edge attributes
#' # are available
#' graph %>% get_edge_df()
#'
#' # Mutate the `width` edge
#' # attribute for the edges
#' # only in the active selection
#' # of edges (edge `1`); here,
#' # we divide each value in the
#' # selection by 2
#' graph <-
#'   graph %>%
#'   mutate_edge_attrs_ws(
#'     width = width / 2)
#'
#' # Get the graph's internal
#' # edf to show that the edge
#' # attribute `width` had its
#' # values changed
#' graph %>% get_edge_df()
#'
#' # Create a new edge attribute,
#' # called `length`, that is the
#' # log of values in `width` plus
#' # 2 (and, also, round all values
#' # to 2 decimal places)
#' graph <-
#'   graph %>%
#'   clear_selection() %>%
#'   select_edges(edges = 2:3) %>%
#'   mutate_edge_attrs_ws(
#'     length = (log(width) + 2) %>%
#'                round(2))
#'
#' # Get the graph's internal edf
#' # to show that the edge attribute
#' # values had been mutated only
#' # for edges `2` and `3` (since
#' # edge `1` is excluded, an NA
#' # value is applied)
#' graph %>% get_edge_df()
#'
#' # Create a new edge attribute
#' # called `area`, which is the
#' # product of the `width` and
#' # `length` attributes
#' graph <-
#'   graph %>%
#'   mutate_edge_attrs_ws(
#'     area = width * length)
#'
#' # Get the graph's internal edf
#' # to show that the edge attribute
#' # values had been multiplied
#' # together (with new attr `area`)
#' # for nodes `2` and `3`
#' graph %>% get_edge_df()
#'
#' # We can invert the selection
#' # and mutate edge `1` several
#' # times to get an `area` value
#' # for that edge
#' graph <-
#'   graph %>%
#'   invert_selection() %>%
#'   mutate_edge_attrs_ws(
#'     length = (log(width) + 5) %>%
#'                round(2),
#'     area = width * length)
#'
#' # Get the graph's internal edf
#' # to show that the 2 mutations
#' # occurred for edge `1`, yielding
#' # non-NA values for its edge
#' # attributes without changing
#' # those of the other edges
#' graph %>% get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
mutate_edge_attrs_ws <- function(
    graph,
    ...
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Validation: Graph object has valid edge selection
  check_graph_contains_edge_selection(graph)

  # Collect expressions
  exprs <- rlang::exprs(...)

  # Extract the graph's edf
  edf <- get_edge_df(graph)

  # Stop function if any supplied
  # expressions mutate columns that
  # should not be changed
  if ("id" %in% names(exprs) ||
      "from" %in% names(exprs) ||
      "to" %in% names(exprs)) {

    cli::cli_abort(
      "The variables `id`, `from`, or `to` cannot undergo mutation.")
  }

  # Determine which edges are not
  # in the active selection
  unselected_edges <-
    base::setdiff(
      get_edge_ids(graph),
      suppressMessages(get_selection(graph)))

  s <- edf$id %in% unselected_edges
  order <- c(which(!s), which(s))

  edf <-
    dplyr::bind_rows(
      edf %>% dplyr::filter(!(!!enquo(s))) %>% dplyr::mutate(!!!enquos(...)),
      edf %>% dplyr::filter(!!enquo(s))
    ) %>%
    dplyr::arrange(!!enquo(order))

  graph$edges_df <- edf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Add one or more global graph attributes
#'
#' @description
#'
#' Add global attributes of a specific type (either `graph_attrs`, `node_attrs`,
#' or `edge_attrs` for a graph object of class `dgr_graph`).
#'
#' @inheritParams render_graph
#' @param attr The name of the attribute to set for the `type` of global
#'   attribute specified.
#' @param value The value to be set for the chosen attribute specified in the
#'   `attr_for_type` argument.
#' @param attr_type The specific type of global graph attribute to set. The type
#'   is specified with `graph`, `node`, or `edge`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a new graph with no
#' # global graph attributes and
#' # add a global graph attribute
#' graph <-
#'   create_graph(
#'     attr_theme = NULL) %>%
#'   add_global_graph_attrs(
#'     attr = "overlap",
#'     value = "true",
#'     attr_type = "graph")
#'
#' # Verify that the attribute
#' # addition has been made
#' graph %>%
#'   get_global_graph_attr_info()
#'
#' # Add another attribute with
#' # `add_global_graph_attrs()`
#' graph <-
#'   graph %>%
#'   add_global_graph_attrs(
#'     attr = "penwidth",
#'     value = 12,
#'     attr_type = "node")
#'
#' # Verify that the attribute
#' # addition has been made
#' graph %>%
#'   get_global_graph_attr_info()
#'
#' # When adding an attribute where
#' # `attr` and `attr_type` already
#' # exists, the value provided will
#' # serve as an update
#' graph %>%
#'   add_global_graph_attrs(
#'     attr = "penwidth",
#'     value = 15,
#'     attr_type = "node") %>%
#'   get_global_graph_attr_info()
#'
#' @export
add_global_graph_attrs <- function(
    graph,
    attr,
    value,
    attr_type
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Coerce any logical value for `value` to a
  # lowercase character value
  # If value is a single true or false
  if (rlang::is_bool(value)) {
    value <- tolower(as.character(value))
  }

  # Create a table for the attributes
  global_attrs_to_add <-
    data.frame(
      attr = as.character(attr),
      value = as.character(value),
      attr_type = as.character(attr_type),
      stringsAsFactors = FALSE)

  # Get the global graph attributes already set
  # in the graph object
  global_attrs_available <- graph$global_attrs

  # Join the new attributes to those available
  # on the `attr` and `attr_type` columns
  global_attrs_joined <-
    global_attrs_available %>%
    dplyr::full_join(
      global_attrs_to_add,
      by = c("attr", "attr_type")) %>%
    dplyr::mutate(
      attr, attr_type,
      value = dplyr::coalesce(value.y, value.x),
      .keep = "none") %>%
    dplyr::select("attr", "value", "attr_type")

  # Replace the graph's global attributes with
  # the revised set
  graph$global_attrs <- global_attrs_joined

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Copy an edge attribute column and set the name
#'
#' @description
#'
#' Within a graph's internal edge data frame (edf), copy the contents an
#' existing edge attribute and create a distinct edge attribute within the edf
#' with a different attribute name.
#'
#' @inheritParams render_graph
#' @param edge_attr_from The name of the edge attribute column from which values
#'   will be copied.
#' @param edge_attr_to The name of the new edge attribute column to which the
#'   copied values will be placed.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 5,
#'     m = 8,
#'     set_seed = 23) |>
#'   set_edge_attrs(
#'     edge_attr = color,
#'     values = "green")
#'
#' # Get the graph's internal
#' # edf to show which edge
#' # attributes are available
#' graph |> get_edge_df()
#'
#' # Make a copy the `color`
#' # edge attribute as the
#' # `color_2` edge attribute
#' graph <-
#'   graph |>
#'   copy_edge_attrs(
#'     edge_attr_from = color,
#'     edge_attr_to = color_2)
#'
#' # Get the graph's internal
#' # edf to show that the edge
#' # attribute had been copied
#' graph |> get_edge_df()
#'
#' @family edge creation and removal
#' @export
copy_edge_attrs <- function(
    graph,
    edge_attr_from,
    edge_attr_to
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the requested `edge_attr_from`
  edge_attr_from <-
    rlang::enquo(edge_attr_from) |> rlang::get_expr() |> as.character()

  # Get the requested `edge_attr_to`
  edge_attr_to <-
    rlang::enquo(edge_attr_to) |> rlang::get_expr() |> as.character()

  # Stop function if `edge_attr_from` and
  # `edge_attr_to` are identical
  if (edge_attr_from == edge_attr_to) {

    cli::cli_abort(
      "You cannot make a copy with the same name.")
  }

  # Stop function if `edge_attr_to` is `from` or `to`
  if (any(c("from", "to") %in% edge_attr_to)) {

    cli::cli_abort(
      "You cannot use `from` or `to` as names.")
  }

  # Extract the graph's edf
  edges <- get_edge_df(graph)

  # Get column names from the graph's edf
  column_names_graph <- colnames(edges)

  # Stop function if `edge_attr_from` is not one
  # of the graph's column
  if (!any(column_names_graph %in% edge_attr_from)) {

    cli::cli_abort(
      "The edge attribute to copy is not in the ndf.")
  }

  # Get the column number for the edge attr to copy
  col_num_copy_from <-
    which(colnames(edges) %in% edge_attr_from)

  # Copy the column using `bind_cols()`
  edges <-
    dplyr::bind_cols(
      edges,
      as.data.frame(
        edges[, col_num_copy_from],
        stringsAsFactors = FALSE))

  # Set the column name for the copied attr
  colnames(edges)[ncol(edges)] <- edge_attr_to

  # Modify the new graph object
  graph$edges_df <- edges

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get total degree distribution data for a graph
#'
#' @description
#'
#' Get degree distribution data for a graph. Graph degree is represented as a
#' frequency of total degree values over all nodes in the graph.
#'
#' @inheritParams render_graph
#' @param mode using `total` (the default), degree considered for each node
#'   will be the total degree. With `in` and `out` the degree used
#'   will be the in-degree and out-degree, respectively.
#'
#' @return A data frame with degree frequencies.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get the total degree
#' # distribution for the graph
#' graph |>
#'   get_degree_distribution(
#'     mode = "total")
#'
#' @export
get_degree_distribution <- function(
    graph,
    mode = "total"
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get the total degree distribution for the graph
  if (mode %in% c("all", "total", "both")) {
    deg_dist <- igraph::degree_distribution(ig_graph, mode = "all")

    # Transform to a data frame
    deg_dist_df <-
      data.frame(
        degree = seq_along(deg_dist) - 1L,
        total_degree_dist = deg_dist,
        stringsAsFactors = FALSE)
  }

  # Get the total in-degree distribution for the graph
  if (mode == "in") {
    deg_dist <- igraph::degree_distribution(ig_graph, mode = "in")

    # Transform to a data frame
    deg_dist_df <-
      data.frame(
        degree = seq_along(deg_dist) - 1L,
        indegree_dist = deg_dist,
        stringsAsFactors = FALSE)
  }

  # Get the total out-degree distribution for the graph
  if (mode == "out") {
    deg_dist <- igraph::degree_distribution(ig_graph, mode = "out")

    # Transform to a data frame
    deg_dist_df <-
      data.frame(
        degree = seq_along(deg_dist) - 1L,
        outdegree_dist = deg_dist,
        stringsAsFactors = FALSE)
  }

  deg_dist_df
}

#' Set the edge attribute values to be rendered
#'
#' @description
#'
#' Set a edge attribute type to display as edge text when calling the
#' [render_graph()] function. This allows for display of different types of edge
#' attribute values on a per-edge basis. Without setting the `display`
#' attribute, rendering a graph will default to not printing any text on edges.
#' Setting the `display` edge attribute with this function for the first time
#' (i.e., the `display` column doesn't exist in the graph's internal edge data
#' frame) will insert the `attr` value for all edges specified in `edges` and a
#' default value (`default`) for all remaining edges.
#'
#' @inheritParams render_graph
#' @param attr The name of the attribute from which label text for the edge will
#'   be obtained. If set to `NULL`, then `NA` values will be assigned to the
#'   `display` column for the chosen edges.
#' @param edges A length vector containing one or several edge ID values (as
#'   integers) for which edge attributes are set for display in the rendered
#'   graph. If `NULL`, all edges from the graph are assigned the `display` value
#'   given as `attr`.
#' @param default The name of an attribute to set for all other graph edges not
#'   included in `edges`. This value only gets used if the `display` edge
#'   attribute is not in the graph's internal edge data frame.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 4,
#'     m = 4,
#'     set_seed = 23) %>%
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = c(2.5, 8.2, 4.2, 2.4))
#'
#' # For edge ID values of `1`,
#' # `2`, and `3`, choose to display
#' # the edge `value` attribute (for
#' # the other edges, display nothing)
#' graph <-
#'   graph %>%
#'   set_edge_attr_to_display(
#'     edges = 1:3,
#'     attr = value,
#'     default = NA)
#'
#' # Show the graph's edge data frame; the
#' # `display` edge attribute will show, for
#' # each row, which edge attribute value to
#' # display when the graph is rendered
#' graph %>% get_edge_df()
#'
#' # This function can be called multiple
#' # times on a graph; after the first time
#' # (i.e., creation of the `display`
#' # attribute), the `default` value won't
#' # be used
#' graph %>%
#'   set_edge_attr_to_display(
#'     edges = 4,
#'     attr = to) %>%
#'   set_edge_attr_to_display(
#'     edges = c(1, 3),
#'     attr = id) %>%
#'   get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
set_edge_attr_to_display <- function(
    graph,
    attr = NULL,
    edges = NULL,
    default = "label"
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Get the requested `attr`
  attr <-
    rlang::enquo(attr) %>% rlang::get_expr() %>% as.character()

  if (attr == "NULL") {
    attr <- NULL
  }

  # Get the graph's edge data frame as an object
  edf <- graph$edges_df

  # If `edges` is NULL, assume that all edges to
  # be assigned a `display` value
  edges <- edges %||% get_edge_ids(graph)

  # Stop function if any of the edge ID values
  # provided in `edges` do not exist in the graph
  if (!any(edges %in% edf$id)) {

    cli::cli_abort(
      "One or more edge ID values in `edges` are not present in the graph.")
  }

  # Stop function if the edge attribute supplied as
  # `attr` does not exist in the edf
  if (!is.null(attr) && !rlang::has_name(edf, attr)) {

    cli::cli_abort(
      "The edge attribute given in `attr` is not in the graph's edf.")
  }

  # If the `display` edge attribute doesn't exist,
  # create that column and fill with the default value
  if (!rlang::has_name(edf, "display")) {

    edf$display <- as.character(default)
  }

  # Create a tibble with the edge ID values and the
  # requested edge attribute to display
  if (is.null(attr)) {

    attr_to_display <-
      dplyr::tibble(
        id = as.integer(edges),
        display = "is_na")

  } else {

    attr_to_display <-
      dplyr::tibble(
        id = as.integer(edges),
        display = as.character(attr))
  }

  # Join the `attr_to_display` table with the `edf`
  edf <-
    edf %>%
    dplyr::left_join(attr_to_display, by = "id")

  # Get the column numbers for the `.x`
  # and `.y` columns
  x_col <- grep("\\.x$", colnames(edf))
  y_col <- grep("\\.y$", colnames(edf))

  # Coalesce the 2 generated columns and create a
  # single-column data frame
  if (!is.null(attr)) {

    display_col <-
      dplyr::coalesce(edf[, y_col], edf[, x_col]) %>%
      as.data.frame(stringsAsFactors = FALSE)

  } else if (is.null(attr)) {

    display_col <-
      dplyr::coalesce(edf[, y_col], edf[, x_col])

    display_col <-
      dplyr::case_when(
        display_col == "is_na" ~ NA_character_,
        .default = display_col) %>%
      as.data.frame(stringsAsFactors = FALSE)
  }

  # Rename the column
  colnames(display_col)[1] <- "display"

  # Remove column numbers that end with ".x" or ".y"
  edf <- edf[-grep("\\.x$", colnames(edf))]
  edf <- edf[-grep("\\.y$", colnames(edf))]

  # Bind the `display_col` df to the `edf` df and
  # modify the ordering of the columns
  edf <-
    dplyr::bind_cols(edf, display_col) %>%
    dplyr::relocate("id", "from", "to", "rel", "display")

  # Replace the graph's edge data frame with `edf`
  graph$edges_df <- edf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

# Get a file's extension
get_file_ext <- function(file) {

  pos <- regexpr("\\.([[:alnum:]]+)$", file)
  ifelse(pos > -1L, substring(file, pos + 1L), "")
}

# Encode a raw string to a Base64 string
encode_base64 <- function(raw) {

  b64 <- c(LETTERS, letters, 0:9, "+", "/")
  s <- as.integer(raw)
  n <- length(s)
  res <- rep(NA, (n + 2) / 3 * 4)
  i <- 0L
  j <- 1L

  while (n > 2L) {
    res[i <- i + 1L] <- b64[s[j] %/% 4L + 1L]
    res[i <- i + 1L] <- b64[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
    res[i <- i + 1L] <- b64[4L * (s[j + 1L] %% 16) + s[j + 2L] %/% 64L + 1L]
    res[i <- i + 1L] <- b64[s[j + 2L] %% 64L + 1L]
    j <- j + 3L
    n <- n - 3L
  }

  if (n) {

    res[i <- i + 1L] <- b64[s[j] %/% 4L + 1L]

    if (n > 1L) {
      res[i <- i + 1L] <- b64[16 * (s[j] %% 4L) + s[j + 1L] %/% 16 + 1L]
      res[i <- i + 1L] <- b64[4L * (s[j + 1L] %% 16) + 1L]
      res[i <- i + 1L] <- "="

    } else {

      res[i <- i + 1L] <- b64[16 * (s[j] %% 4L) + 1L]
      res[i <- i + 1L] <- "="
      res[i <- i + 1L] <- "="
    }
  }
  paste(res[!is.na(res)], collapse = "")
}

# Helper to set the MIME type
get_mime_type <- function(file) {

  extension <-
    file |>
    get_file_ext() |>
    tolower()

  switch(
    extension,
    svg = "image/svg+xml",
    jpg = "image/jpeg",
    file.path("image", extension, fsep = "/")
    )
}

# Get an image URI from an on-disk graphics file
# as a Base64-encoded image string
get_image_uri <- function(file) {

  image_raw <-
    readBin(
      con = file,
      what = "raw",
      n = file.info(file)$size)

  paste0("data:", get_mime_type(file), ";base64,", encode_base64(image_raw))
}

#' Trigger the execution of a series of graph actions
#'
#' @description
#'
#' Execute the graph actions stored in the graph through the use of the
#' [add_graph_action()] function. These actions will be invoked in order and any
#' errors encountered will trigger a warning message and result in no change to
#' the input graph. Normally, graph actions are automatically triggered at every
#' transformation step but this function allows for the manual triggering of
#' graph actions after setting them, for example.
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() |>
#'   add_gnm_graph(
#'     n = 5,
#'     m = 10,
#'     set_seed = 23)
#'
#' # Add a graph action that sets a node
#' # attr column with a function; this
#' # uses the `get_pagerank()` function
#' # to provide PageRank values in the
#' # `pagerank` column
#' graph <-
#'   graph |>
#'   add_graph_action(
#'     fcn = "set_node_attr_w_fcn",
#'     node_attr_fcn = "get_pagerank",
#'     column_name = "pagerank",
#'     action_name = "get_pagerank")
#'
#' # Add a second graph action (to be
#' # executed after the first one) that
#' # rescales values in the `pagerank`
#' # column between 0 and 1, and, puts
#' # these values in the `width` column
#' graph <-
#'   graph |>
#'   add_graph_action(
#'     fcn = "rescale_node_attrs",
#'     node_attr_from = "pagerank",
#'     node_attr_to = "width",
#'     action_name = "pgrnk_to_width")
#'
#' # Add a third and final graph action
#' # (to be executed last) that creates
#' # color values in the `fillcolor` column,
#' # based on the numeric values from the
#' # `width` column
#' graph <-
#'   graph |>
#'   add_graph_action(
#'     fcn = "colorize_node_attrs",
#'     node_attr_from = "width",
#'     node_attr_to = "fillcolor",
#'     action_name = "pgrnk_fillcolor")
#'
#' # View the graph actions for the graph
#' # object by using the `get_graph_actions()`
#' # function
#' graph |> get_graph_actions()
#'
#' # Manually trigger to invocation of
#' # the graph actions using the
#' # `trigger_graph_actions()` function
#' graph <-
#'   graph |>
#'   trigger_graph_actions()
#'
#' # Examine the graph's internal node
#' # data frame (ndf) to verify that
#' # the `pagerank`, `width`, and
#' # `fillcolor` columns are present
#' graph |> get_node_df()
#'
#' @export
trigger_graph_actions <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  if (nrow(graph$graph_actions) == 0) {

    message("There are currently no graph actions.")

  } else {

    # Collect text expressions in a vector
    graph_actions <- graph$graph_actions$expression

    # Copy graph state as the `graph_previous` object
    graph_previous <- graph

    expr_error_at_index <- 0

    for (i in seq_along(graph_actions)) {

      if (class(
        tryCatch(
          eval(
            parse(text = graph$graph_actions$expression[i])),
          error = function(x) x))[1] == "simpleError") {
        expr_error_at_index <- i
        break
      } else {
        graph <-
          eval(parse(text = graph$graph_actions$expression[i]))
      }
    }

    if (expr_error_at_index > 0) {

      # Revert `graph_previous` to be the returned
      # graph (because of an evaluation error)
      graph <- graph_previous

      action_name_at_error <-
        graph$graph_actions |>
        dplyr::filter(action_index == expr_error_at_index) |>
        dplyr::pull("action_name")

      if (!is.na(action_name_at_error)) {
        message(
          paste0(
            "The series of graph actions was not applied to the graph because ",
            "of an error at action index ", expr_error_at_index, "."))
      } else {
        message(
          paste0(
            "The series of graph actions was not applied to the graph because ",
            "of an error at action index ", expr_error_at_index, " (`",
            action_name_at_error, "`)."))
      }
    }

    # Get the name of the function
    fcn_name <- get_calling_fcn()

    # Update the `graph_log` df with an action
    graph$graph_log <-
      add_action_to_log(
        graph_log = graph$graph_log,
        version_id = nrow(graph$graph_log) + 1L,
        function_used = fcn_name,
        time_modified = time_function_start,
        duration = graph_function_duration(time_function_start),
        nodes = nrow(graph$nodes_df),
        edges = nrow(graph$edges_df))

    # Write graph backup if the option is set
    if (graph$graph_info$write_backups) {
      save_graph_as_rds(graph = graph)
    }
  }

  graph
}

#' Add one or more edges using a text string
#'
#' @description
#'
#' With a graph object of class `dgr_graph`, add one or more edges to the graph
#' using a text string.
#'
#' @inheritParams render_graph
#' @param edges A single-length vector with a character string specifying the
#'   edges. For a directed graph, the string object should be formatted as a
#'   series of node ID values as `[node_ID_1]->[node_ID_2]` separated by a one
#'   or more space characters. For undirected graphs, `--` should replace `->`.
#'   Line breaks in the vector won't cause an error.
#' @param rel An optional vector specifying the relationship between the
#'   connected nodes.
#' @param use_labels An option to use node `label` values in the `edges` string
#'   to define node connections. Note that this is only possible if all nodes
#'   have distinct `label` values set and none exist as an empty string.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 4 nodes
#' graph <-
#'   create_graph() |>
#'   add_node(label = "one") |>
#'   add_node(label = "two") |>
#'   add_node(label = "three") |>
#'   add_node(label = "four")
#'
#' # Add edges between nodes using
#' # a character string with node
#' # ID values
#' graph_node_id <-
#'   graph |>
#'   add_edges_w_string(
#'     edges = "1->2 1->3 2->4 2->3")
#'
#' # Show the graph's internal
#' # edge data frame
#' graph_node_id |> get_edge_df()
#'
#' # Add edges between nodes using
#' # a character string with node
#' # label values and setting
#' # `use_labels = TRUE`; note that
#' # all nodes must have unique
#' # `label` values to use this
#' graph_node_label <-
#'   graph |>
#'   add_edges_w_string(
#'     edges =
#'       "one->two one->three
#'        two->four two->three",
#'     use_labels = TRUE)
#'
#' # Show the graph's internal
#' # edge data frame (it's the
#' # same as before)
#' graph_node_label |> get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
add_edges_w_string <- function(
    graph,
    edges,
    rel = NULL,
    use_labels = FALSE
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Get the value for the latest `version_id` for
  # graph (in the `graph_log`)
  current_graph_log_version_id <-
    max(graph$graph_log$version_id)

  # Remove linebreak characters from `edges`
  edges_cleaned <-
    gsub("\n", " ", edges)

  # Remove extra spaces within the string
  edges_cleaned <-
    gsub("(?<=[\\s])\\s*|^\\s+|\\s+$", "",
         edges_cleaned, perl = TRUE)

  # Split by single spaces into separate edge
  # expressions
  edges_split <-
    unlist(strsplit(edges_cleaned, " "))

  # Split the edge expressions in a directed
  # graph into `from` and `to` vectors
  if (graph$directed) {
    from <-
      sapply(strsplit(edges_split, "->"), "[[", 1)

    to <-
      sapply(strsplit(edges_split, "->"), "[[", 2)
  }

  # Split the edge expressions in an undirected
  # graph into `from` and `to` vectors
  if (!graph$directed) {
    from <-
      sapply(strsplit(edges_split, "--"), "[[", 1)

    to <-
      sapply(strsplit(edges_split, "--"), "[[", 2)
  }

  # If `use_label` is set to TRUE, treat values in
  # list as labels; need to map to node ID values
  if (use_labels) {
    from_to_node_id <-
      translate_to_node_id(
        graph = graph,
        from = from,
        to = to)

    from <- from_to_node_id$from
    to <- from_to_node_id$to
  }

  # Create an edge data frame (edf) without
  # associated `rel` values
  if (is.null(rel)) {
    new_edges <-
      create_edge_df(
        from = from,
        to = to)
  }

  # Create an edge data frame (edf) with
  # associated `rel` values
  if (!is.null(rel)) {
    new_edges <-
      create_edge_df(
        from = from,
        to = to,
        rel = rel)
  }

  # Get the number of edges in the graph
  edges_graph_1 <- graph |> count_edges()

  # Add the new edges to the graph
  graph <- add_edge_df(graph, new_edges)

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph |> count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Clear the graph's active selection
  graph <-
    suppressMessages(
      graph |>
        clear_selection())

  # Remove extra items from the `graph_log`
  graph$graph_log <-
    graph$graph_log |>
    dplyr::filter(version_id <= current_graph_log_version_id)

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      trigger_graph_actions(graph)
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Drop an edge attribute column
#'
#' @description
#'
#' Within a graph's internal edge data frame (edf), remove an existing edge
#' attribute.
#'
#' @inheritParams render_graph
#' @param edge_attr The name of the edge attribute column to drop.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 5,
#'     m = 6,
#'     set_seed = 23) %>%
#'   set_edge_attrs(
#'     edge_attr = value,
#'     values = 3) %>%
#'   mutate_edge_attrs(
#'     penwidth = value * 2)
#'
#' # Get the graph's internal
#' # edf to show which edge
#' # attributes are available
#' graph %>% get_edge_df()
#'
#' # Drop the `value` edge
#' # attribute
#' graph <-
#'   graph %>%
#'   drop_edge_attrs(
#'     edge_attr = value)
#'
#' # Get the graph's internal
#' # edf to show that the edge
#' # attribute `value` had been
#' # removed
#' graph %>% get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
drop_edge_attrs <- function(
    graph,
    edge_attr
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Get the requested `edge_attr`
  edge_attr <-
    rlang::enquo(edge_attr) %>% rlang::get_expr() %>% as.character()

  # Stop function if `edge_attr` is any of
  # `from`, `to`, or `rel`
  if (any(c("from", "to", "rel") %in%
          edge_attr)) {
    cli::cli_abort(
      "You cannot drop {.val from}, {.val to} or {.val rel} column.")
  }

  # Extract the graph's edf
  edges <- get_edge_df(graph)

  # Get column names from the graph's edf
  column_names_graph <- colnames(edges)

  # Stop function if `edge_attr` is not one
  # of the graph's column
  if (!any(column_names_graph %in% edge_attr)) {
    cli::cli_abort(
      "The edge attribute to drop is not in the ndf.")
  }

  # Get the column number for the edge attr to drop
  col_num_drop <-
    which(colnames(edges) %in% edge_attr)

  # Remove the column
  edges <- edges[, -col_num_drop]

  # Update the graph object
  graph$edges_df <- edges

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Is the graph a directed graph?
#'
#' @description
#'
#' Determines whether a graph is set to be directed or not and returns a logical
#' value to that effect.
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create an empty graph; by default,
#' # new graphs made by `create_graph()`
#' # are directed
#' graph <- create_graph()
#'
#' # Determine whether the graph
#' # is directed
#' graph %>% is_graph_directed()
#'
#' # Use the `set_graph_undirected()`
#' # function and check again whether
#' # the graph is directed
#' graph %>%
#'   set_graph_undirected() %>%
#'   is_graph_directed()
#'
#' @export
is_graph_directed <- function(graph) {
  check_graph_valid(graph)
  # Determine if graph is directed by getting the
  # value at `graph$directed`
  graph$directed
}

#' Convert an undirected graph to a directed graph
#'
#' @description
#'
#' Take a graph which is undirected and convert it to a directed graph.
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with a
#' # undirected tree
#' graph <-
#'   create_graph(
#'     directed = FALSE) %>%
#'   add_balanced_tree(
#'     k = 2, h = 2)
#'
#' # Convert this graph from
#' # undirected to directed
#' graph <-
#'   graph %>%
#'   set_graph_directed()
#'
#' # Perform a check on whether
#' # graph is directed
#' graph %>% is_graph_directed()
#'
#' @export
set_graph_directed <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Set the `directed` vector to TRUE
  graph$directed <- TRUE

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Render a graph available in a series
#'
#' @description
#'
#' Using a graph series object of type `dgr_graph_1D`, either render graph in
#' the Viewer or output in various formats.
#'
#' @param graph_series A graph series object of type `dgr_graph_1D`.
#' @param graph_no The index of the graph in the graph series.
#' @param output A string specifying the output type; `graph` (the default)
#'   renders the graph using the [grViz()] function and `visNetwork` renders the
#'   graph using the [visnetwork()] function.
#' @param width An optional parameter for specifying the width of the resulting
#'   graphic in pixels.
#' @param height An optional parameter for specifying the height of the
#'   resulting graphic in pixels.
#' @examples
#' \dontrun{
#' # Create three graphs
#' graph_1 <-
#'   create_graph() |>
#'   add_path(n = 4)
#'
#' graph_2 <-
#'   create_graph() |>
#'   add_cycle(n = 5)
#'
#' graph_3 <-
#'   create_graph() |>
#'   add_star(n = 6)
#'
#' # Create an empty graph series
#' # and add the graphs
#' series <-
#'   create_graph_series() |>
#'   add_graph_to_graph_series(
#'     graph = graph_1) |>
#'   add_graph_to_graph_series(
#'     graph = graph_2) |>
#'   add_graph_to_graph_series(
#'     graph = graph_3)
#'
#' # View the second graph in
#' # the series in the Viewer
#' render_graph_from_graph_series(
#'   graph_series = series,
#'   graph_no = 2)
#' }
#'
#' @family Display and Save
#'
#' @export
render_graph_from_graph_series <- function(
    graph_series,
    graph_no,
    output = "graph",
    width = NULL,
    height = NULL
) {

  # Stop function if no graphs are available
  if (is.null(graph_series$graphs)) {

    cli::cli_abort(
      "There are no graphs in this graph series.")
  }

  # Stop function if `graph_no` is out of range
  if (!(graph_no %in% 1:count_graphs_in_graph_series(graph_series))) {

    cli::cli_abort(
      "The index chosen doesn't correspond to that of a graph in the series.")
  }

  # Extract the specified graph from the series
  graph <- graph_series$graphs[[graph_no]]

  render_graph(
    graph = graph,
    output = output,
    width = width,
    height = height)
}

#' Remove a graph from a graph series
#'
#' @description
#'
#' Remove a single graph object from an set of graph objects contained within a
#' graph series object.
#'
#' @param graph_series A graph series object from which the graph object will be
#'   removed.
#' @param index The index of the graph object to be removed from the graph
#'   series object.
#'
#' @return A graph series object of type `dgr_graph_1D`.
#'
#' @examples
#' # Create three graphs
#' graph_1 <-
#'   create_graph() %>%
#'   add_path(n = 4)
#'
#' graph_2 <-
#'   create_graph() %>%
#'   add_cycle(n = 5)
#'
#' graph_3 <-
#'   create_graph() %>%
#'   add_star(n = 6)
#'
#' # Create an empty graph series
#' # and add the graphs
#' series <-
#'   create_graph_series() %>%
#'   add_graph_to_graph_series(
#'     graph = graph_1) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_2) %>%
#'   add_graph_to_graph_series(
#'     graph = graph_3)
#'
#' # Remove the second graph
#' # from the graph series
#' series <-
#'   series %>%
#'   remove_graph_from_graph_series(
#'     index = 2)
#'
#' # With `get_graph_series_info()`,
#' # we can ensure that a graph
#' # was removed
#' series %>%
#'   get_graph_series_info()
#'
#' @export
remove_graph_from_graph_series <- function(
    graph_series,
    index = "last"
) {

  if (index == "last") {
    graph_series$graphs[[length(graph_series$graphs)]] <- NULL
    return(graph_series)
  }

  if (index == "first") {
    graph_series$graphs[[1]] <- NULL
    return(graph_series)
  }

  if (rlang::inherits_any(index, c("numeric", "integer"))) {
    graph_series$graphs[[index]] <- NULL
    return(graph_series)
  }
}

#' Get leverage centrality
#'
#' @description
#'
#' Get the leverage centrality values for all nodes in the graph. Leverage
#' centrality is a measure of the relationship between the degree of a given
#' node and the degree of each of its neighbors, averaged over all neighbors. A
#' node with negative leverage centrality is influenced by its neighbors, as the
#' neighbors connect and interact with far more nodes. A node with positive
#' leverage centrality influences its neighbors since the neighbors tend to have
#' far fewer connections.
#'
#' @inheritParams render_graph
#'
#' @return A data frame with leverage centrality values for each of the nodes.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph(
#'     directed = FALSE) |>
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23)
#'
#' # Get leverage centrality values
#' # for all nodes in the graph
#' graph |>
#'   get_leverage_centrality()
#'
#' # Add the leverage centrality
#' # values to the graph as a
#' # node attribute
#' graph <-
#'   graph |>
#'   join_node_attrs(
#'     df = get_leverage_centrality(graph))
#'
#' # Display the graph's node data frame
#' graph |> get_node_df()
#'
#' @export
get_leverage_centrality <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Convert the graph to an igraph object
  ig_graph <- to_igraph(graph)

  # Get degree values for all graph nodes
  degree_vals <- igraph::degree(ig_graph)

  # Calculate leverage centrality values for
  # all nodes in the graph
  leverage_centrality_values <-
    purrr::map(
      seq_along(degree_vals),
      function(x) {
        mean(
          (degree_vals[x] - degree_vals[igraph::neighbors(ig_graph, x)]) /
            (degree_vals[x] + degree_vals[igraph::neighbors(ig_graph, x)]))
      }) |>
    unlist()

  # Create df with leverage centrality values
  data.frame(
    id = degree_vals |>
      names() |>
      as.integer(),
    leverage_centrality = leverage_centrality_values |> round(4),
    stringsAsFactors = FALSE)
}

#' Get paths from a specified node in a directed graph
#'
#' @description
#'
#' Obtain a list of all possible paths from a given node within a directed
#' graph.
#'
#' @inheritParams render_graph
#' @param from The node from which all paths will be determined.
#' @param to The node to which all paths will be determined.
#' @param shortest_path An option to return paths that are the shortest in the
#'   set of all determined paths.
#' @param longest_path An option to return paths that are the longest in the set
#'   of all determined paths.
#' @param distance A vector of integer values that specify which of the valid
#'   paths to return when filtering by distance.
#'
#' @return A list of paths, sorted by ascending traversal length, comprising
#'   vectors of node IDs in sequence of traversal through the graph.
#'
#' @examples
#' # Create a simple graph
#' graph <-
#'   create_graph() |>
#'   add_n_nodes(n = 8) |>
#'   add_edge(from = 1, to = 2) |>
#'   add_edge(from = 1, to = 3) |>
#'   add_edge(from = 3, to = 4) |>
#'   add_edge(from = 3, to = 5) |>
#'   add_edge(from = 4, to = 6) |>
#'   add_edge(from = 2, to = 7) |>
#'   add_edge(from = 7, to = 5) |>
#'   add_edge(from = 4, to = 8)
#'
#' # Get a list of all paths outward from node `1`
#' graph |>
#'   get_paths(from = 1)
#'
#' # Get a list of all paths leading to node `6`
#' graph |>
#'   get_paths(to = 6)
#'
#' # Get a list of all paths from `1` to `5`
#' graph |>
#'   get_paths(
#'    from = 1,
#'    to = 5)
#'
#' # Get a list of all paths from `1` up to a distance
#' # of 2 node traversals
#' graph |>
#'   get_paths(
#'     from = 1,
#'     distance = 2)
#'
#' # Get a list of the shortest paths from `1` to `5`
#' get_paths(
#'   graph,
#'   from = 1,
#'   to = 5,
#'   shortest_path = TRUE)
#'
#' # Get a list of the longest paths from `1` to `5`
#' get_paths(
#'   graph,
#'   from = 1,
#'   to = 5,
#'   longest_path = TRUE)
#'
#' @export
get_paths <- function(
    graph,
    from = NULL,
    to = NULL,
    shortest_path = FALSE,
    longest_path = FALSE,
    distance = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  reverse_paths <- FALSE

  if (is.null(from) && !is.null(to)) {

    from_switch <- graph$edges_df$from
    to_switch <- graph$edges_df$to

    graph$edges_df$from <- to_switch
    graph$edges_df$to <- from_switch

    from <- to
    to <- NULL

    reverse_paths <- TRUE
  }

  for (m in seq_along(from)) {
    if (m == 1) all_paths <- list()

    # Initialize paths with starting node
    paths <- list(from[m])

    repeat {
      for (i in seq_along(paths)) {
        if (!all(is.na(
          get_successors(
            graph,
            paths[[i]][
              length(paths[[i]])])))) {

          # Get the successors for the last node
          # in the given path
          next_nodes <-
            get_successors(
              graph,
              paths[[i]][
                length(paths[[i]])])

          # Filter next_nodes if cycles are detected
          next_nodes <-
            next_nodes[
              which(!(
                next_nodes %in%
                  paths[[i]][
                    seq_along(length(paths[[i]])) - 1]))]

          # Apply traversed nodes to each of the
          # path vectors in a multiple degree context
          if (length(next_nodes) > 1) {
            for (j in seq_along(next_nodes)) {
              if (j == 1) paths[[i]] <-
                  c(paths[[i]], next_nodes[1])

              if (j > 1) paths[[length(paths) + 1]] <-
                  c(paths[[i]][
                    -length(paths[[i]])],
                    next_nodes[j])
            }
          }

          # Apply traversed nodes to each of the
          # path vectors in a single degree context
          if (length(next_nodes) == 1) {
            paths[[i]] <- c(paths[[i]], next_nodes[1])
          }
        }
      }

      # Check each node visited in the present
      # iteration for whether their traversals
      # should end
      for (k in seq_along(paths)) {
        if (k == 1) check <- vector()

        check <-
          c(check,
            any(is.na(
              get_successors(
                graph,
                paths[[k]][length(paths[[k]])])) |
                all(get_successors(
                  graph,
                  paths[[k]][length(paths[[k]])]) %in%
                    paths[[k]])))

        # Remove nodes from vectors within paths
        # if they were previously traversed
        if (paths[[k]][length(paths[[k]])] %in%
            paths[[k]][seq_along(paths[[k]]) - 1]) {
          paths[[k]] <- paths[[k]][-length(paths[[k]])]
        }
      }

      if (all(check)) break
    }

    if (!(length(paths) == 1 && is.na(paths[1]))) {
      all_paths <- c(all_paths, paths)
    }

    if (m == length(from)) paths <- all_paths
  }

  # Arrange vectors in list in order of
  # increasing length
  order <- sapply(seq_along(paths),
                  function(x) length(paths[[x]]))

  names(order) <- seq_along(paths)
  order <- sort(order)
  order <- as.numeric(names(order))
  paths <- paths[order]

  # If only a single vector returned, return NA
  if (length(paths) == 1 && length(paths[[1]]) == 1) {
    return(NA)
  }

  # Remove paths of single length
  for (i in seq_along(paths)) {
    if (i == 1) {
      single_length_paths <- vector(mode = "numeric")
    }

    if (length(paths[[i]]) == 1) {
      single_length_paths <- c(single_length_paths, i)
    }

    if (i == length(paths)) {
      paths[single_length_paths] <- NULL
    }
  }

  # If only `from` but not the `to` node specified,
  # get paths that consider the filtering criteria
  if (!is.null(from) && is.null(to)) {

    # Filter the `path_lengths` vector
    # by the chosen criteria
    if (shortest_path && !longest_path) {

      # Remove paths not of shortest length
      for (i in seq_along(paths)) {
        if (i == 1) {
          not_shortest_length_paths <-
            vector(mode = "numeric")
          shortest_length <-
            min(sapply(seq_along(paths),
                       function(x) length(paths[[x]])))
        }

        if (length(paths[[i]]) != shortest_length) {
          not_shortest_length_paths <-
            c(not_shortest_length_paths, i)
        }

        if (i == length(paths)) {
          paths[not_shortest_length_paths] <- NULL
        }
      }

    } else if (!shortest_path && longest_path) {

      # Remove paths not of longest length
      for (i in seq_along(paths)) {
        if (i == 1) {
          not_longest_length_paths <-
            vector(mode = "numeric")
          longest_length <-
            max(sapply(seq_along(paths),
                       function(x) length(paths[[x]])))
        }

        if (length(paths[[i]]) != longest_length) {
          not_longest_length_paths <-
            c(not_longest_length_paths, i)
        }

        if (i == length(paths)) {
          paths[not_longest_length_paths] <- NULL
        }
      }

    } else if (!is.null(distance)) {

      # Remove paths not of the specified distances
      for (i in seq_along(paths)) {
        if (i == 1) {
          not_specified_length_paths <-
            vector(mode = "numeric")
          specified_lengths <- distance
        }

        if (length(paths[[i]]) <
            specified_lengths + 1) {
          not_specified_length_paths <-
            c(not_specified_length_paths, i)
        }

        if (i == length(paths)) {
          paths[not_specified_length_paths] <- NULL
        }
      }

      # Trim paths to specified distance
      for (i in seq_along(paths)) {
        paths[[i]] <-
          paths[[i]][seq_len(specified_lengths + 1)]
      }

      # Create a unique list of paths
      paths <- unique(paths)
    }
  }

  # If both a 'from' node and a 'to' node specified,
  # get paths that begin and end with those nodes
  if (!is.null(from) && !is.null(to)) {

    # Determine which paths contain the 2nd node target
    paths_with_end_node <-
      which(sapply(seq_along(paths),
                   function(x) to %in% paths[[x]]))

    if (length(paths_with_end_node) == 0) {
      return(NA)
    }

    # Obtain a vector of path lengths for each of
    # the valid paths
    path_lengths <-
      sapply(paths_with_end_node,
             function(x) which(paths[[x]] %in% to))

    # Apply list indices as names for the
    # `paths_lengths` vector
    names(path_lengths) <- paths_with_end_node

    # Filter the `path_lengths` vector by the
    # chosen criteria
    if (shortest_path && !longest_path) {
      path_lengths <-
        as.numeric(names(path_lengths[
          which(path_lengths == min(path_lengths))]))
    } else if (!shortest_path && longest_path) {
      path_lengths <-
        as.numeric(names(path_lengths[
          which(path_lengths == max(path_lengths))]))
    } else if (!is.null(distance)) {
      path_lengths <-
        as.numeric(names(path_lengths[
          which(path_lengths %in% distance)]))
    } else if (is.null(distance) && !shortest_path && !longest_path) {
      path_lengths <- as.numeric(names(path_lengths))
    }

    # If there are no paths that match the chosen
    # criteria, return NA
    if (length(path_lengths) == 0) {
      return(NA)
    }

    paths <- paths[path_lengths]

    # Modify `paths` list of vectors such that nodes
    # at the beginning and end of each vector are the
    # `from` and `to` nodes
    for (i in seq_along(paths)) {
      paths[[i]] <-
        paths[[i]][1:which(paths[[i]] == to)]
    }
  }

  if (reverse_paths) {

    for (i in seq_along(paths)) {
      paths[[i]] <- rev(paths[[i]])
    }
  }

  paths
}

#' Get nodes within strongly connected components
#'
#' @description
#'
#' Determine which nodes in a graph belong to different strongly connected
#' components.
#'
#' @inheritParams render_graph
#'
#' @return A data frame with nodes and their membership in different strongly
#'   connected components.
#'
#' @examples
#' suppressWarnings(RNGversion("3.5.0"))
#' set.seed(23)
#'
#' # Create a graph with a random
#' # connection between 2 different
#' # node cycles
#' graph <-
#'   create_graph() %>%
#'   add_cycle(
#'     n = 3,
#'     type = "cycle_1") %>%
#'   add_cycle(
#'     n = 4,
#'     type = "cycle_2") %>%
#'   add_edge(
#'     from =
#'       get_node_ids(
#'         graph = .,
#'         conditions =
#'           type == "cycle_1") %>%
#'         sample(size = 1),
#'     to =
#'       get_node_ids(
#'         graph = .,
#'         conditions =
#'           type == "cycle_2") %>%
#'         sample(size = 1))
#'
#' # Get the strongly connected
#' # components as a data frame of
#' # nodes and their groupings
#' graph %>% get_s_connected_cmpts()
#'
#' @export
get_s_connected_cmpts <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Create a graph where only mandatory
  # node and edge attributes are retained;
  # transform to an igraph object
  ig_graph <-
    create_graph(
      nodes_df = graph %>%
        get_node_df() %>%
        dplyr::select("id", "type", "label"),
      edges_df = graph %>%
        get_edge_df() %>%
        dplyr::select("id", "from", "to", "rel"),
      directed = is_graph_directed(graph)) %>%
    to_igraph()

  # Get the component list from the graph
  components <-
    ig_graph %>%
    igraph::components(mode = "strong")

  # Create the output data frame
  data.frame(
    id = as.integer(names(components$membership)),
    sc_component = components$membership,
    stringsAsFactors = FALSE)
}

#' Get the graph log information
#'
#' @description
#'
#' Get a tibble of the graph log, which contains information on the functions
#' called on the graph that resulted in some transformation of the graph.
#'
#' @inheritParams render_graph
#'
#' @return A `df_tbl` object.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function and
#' # delete 2 nodes from the graph
#' graph <-
#'   create_graph(
#'     directed = FALSE) %>%
#'   add_gnm_graph(
#'     n = 10,
#'     m = 15,
#'     set_seed = 23) %>%
#'   delete_node(node = 5) %>%
#'   delete_node(node = 7)
#'
#' # Get the graph log, which is a
#' # record of all graph transformations
#' graph %>% get_graph_log()
#'
#' @export
get_graph_log <- function(graph) {
  check_graph_valid(graph)
  dplyr::as_tibble(graph$graph_log)
}

#' Is the graph an undirected graph?
#'
#' @description
#'
#' Determines whether a graph is set as undirected or not and returns a logical
#' value to that effect.
#'
#' @inheritParams render_graph
#'
#' @return A logical value.
#'
#' @examples
#' # Create an empty graph; by
#' # default, new graphs made
#' # by `create_graph()` are
#' # directed graph, so, use
#' # `directed = FALSE` to create
#' # an undirected graph
#' graph <-
#'   create_graph(
#'     directed = FALSE)
#'
#' # Determine whether the
#' # graph is undirected
#' graph %>% is_graph_undirected()
#'
#' # Use the `set_graph_directed()`
#' # function and check again
#' # as to whether the graph is
#' # undirected
#' graph %>%
#'   set_graph_directed() %>%
#'   is_graph_undirected()
#'
#' @export
is_graph_undirected <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Determine if graph is
  # undirected by using a
  # statement with `graph$directed`
  !graph$directed
}

#' Options for node and edge aesthetics and data
#'
#' @param node_aes An optional list of named vectors comprising node aesthetic
#'   attributes. The helper function [node_aes()] is strongly recommended for
#'   use here as it contains arguments for each of the accepted node aesthetic
#'   attributes (e.g., `shape`, `style`, `color`, `fillcolor`).
#' @param edge_aes An optional list of named vectors comprising edge aesthetic
#'   attributes. The helper function [edge_aes()] is strongly recommended for
#'   use here as it contains arguments for each of the accepted edge aesthetic
#'   attributes (e.g., `shape`, `style`, `penwidth`, `color`).
#' @param node_data An optional list of named vectors comprising node data
#'   attributes. The helper function [node_data()] is strongly recommended for
#'   use here as it helps bind data specifically to the created nodes.
#' @param edge_data An optional list of named vectors comprising edge data
#'   attributes. The helper function [edge_data()] is strongly recommended for
#'   use here as it helps bind data specifically to the created edges.
#' @name node_edge_aes_data
#'
#' @keywords internal
#' @family aesthetics
NULL

#' Create a subgraph using a node or edge selection
#'
#' @description
#'
#' Create a subgraph based on a selection of nodes or edges stored in the graph
#' object.
#'
#' This function makes use of an active selection of nodes or edges (and the
#' function ending with `_ws` hints at this).
#'
#' Selections of nodes can be performed using the following node selection
#' (`select_*()`) functions: [select_nodes()], [select_last_nodes_created()],
#' [select_nodes_by_degree()], [select_nodes_by_id()], or
#' [select_nodes_in_neighborhood()].
#'
#' Selections of edges can be performed using the following edge selection
#' (`select_*()`) functions: [select_edges()], [select_last_edges_created()],
#' [select_edges_by_edge_id()], or [select_edges_by_node_id()].
#'
#' Selections of nodes or edges can also be performed using the following
#' traversal (`trav_*()`) functions: [trav_out()], [trav_in()], [trav_both()],
#' [trav_out_node()], [trav_in_node()], [trav_out_until()], [trav_in_until()],
#' [trav_out_edge()], [trav_in_edge()], [trav_both_edge()], or
#' [trav_reverse_edge()].
#'
#' @inheritParams render_graph
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a node data frame (ndf)
#' ndf <-
#'   create_node_df(
#'     n = 6,
#'     value =
#'       c(3.5, 2.6, 9.4,
#'         2.7, 5.2, 2.1))
#'
#' # Create an edge data frame (edf)
#' edf <-
#'   create_edge_df(
#'     from = c(1, 2, 4, 5, 2, 6, 2),
#'       to = c(2, 4, 1, 3, 5, 5, 4))
#'
#' # Create a graph
#' graph <-
#'   create_graph(
#'     nodes_df = ndf,
#'     edges_df = edf)
#'
#' # Create a selection of nodes, this selects
#' # nodes `1`, `3`, and `5`
#' graph <-
#'   graph %>%
#'   select_nodes(
#'     conditions = value > 3)
#'
#' # Create a subgraph based on the selection
#' subgraph <-
#'   graph %>%
#'   transform_to_subgraph_ws()
#'
#' # Display the graph's node data frame
#' subgraph %>% get_node_df()
#'
#' # Display the graph's edge data frame
#' subgraph %>% get_edge_df()
#'
#' # Create a selection of edges, this selects
#' # edges `1`, `2`
#' graph <-
#'   graph %>%
#'   clear_selection() %>%
#'   select_edges(
#'   edges = c(1,2))
#'
#' # Create a subgraph based on the selection
#' subgraph <-
#'   graph %>%
#'   transform_to_subgraph_ws()
#'
#' # Display the graph's node data frame
#' subgraph %>% get_node_df()
#'
#' # Display the graph's edge data frame
#' subgraph %>% get_edge_df()
#'
#' @export
transform_to_subgraph_ws <- function(graph) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph object has valid selection of
  # nodes or edges
  if (!graph_contains_node_selection(graph) &&
      !graph_contains_edge_selection(graph)) {

    cli::cli_abort(
      "There is no selection of node or edges available.")
  }

  # Get the number of nodes in the graph
  nodes_graph_1 <- graph %>% count_nodes()

  # Get the number of edges in the graph
  edges_graph_1 <- graph %>% count_edges()

  # Filter the nodes in the graph
  if (graph_contains_node_selection(graph)) {

    selection <- graph$node_selection$node

    ndf <-
      graph$nodes_df %>%
      dplyr::filter(id %in% selection)

    edf <-
      graph$edges_df %>%
      dplyr::filter(from %in% selection, to %in% selection)

    # Create a subgraph
    graph$nodes_df <- ndf
    graph$edges_df <- edf
  }

  # Filter the edges in the graph
  if (graph_contains_edge_selection(graph)) {

    selection <- graph$edge_selection$edge

    selection_df <-
      data.frame(id = selection)

    edf <-
      graph$edges_df %>%
      dplyr::semi_join(selection_df, by = "id")

    ndf <-
      graph$nodes_df %>%
      dplyr::filter(id %in% unique(c(edf$from, edf$to)))

    # Create a subgraph
    graph$nodes_df <- ndf
    graph$edges_df <- edf
  }

  # Scavenge any invalid, linked data frames
  graph <-
    remove_linked_dfs(graph)

  # Get the updated number of nodes in the graph
  nodes_graph_2 <- graph %>% count_nodes()

  # Get the number of nodes added to
  # the graph
  nodes_added <- nodes_graph_2 - nodes_graph_1

  # Get the updated number of edges in the graph
  edges_graph_2 <- graph %>% count_edges()

  # Get the number of edges added to
  # the graph
  edges_added <- edges_graph_2 - edges_graph_1

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df),
      d_n = nodes_added,
      d_e = edges_added)

  # Perform graph actions, if any are available
  if (nrow(graph$graph_actions) > 0) {
    graph <-
      graph %>%
      trigger_graph_actions()
  }

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Recode a set of edge attribute values
#'
#' @description
#'
#' Within a graph's internal edge data frame (edf), recode character or numeric
#' edge attribute values. Optionally, one can specify a replacement value for
#' any unmatched mappings.
#'
#' @inheritParams render_graph
#' @param edge_attr_from The name of the edge attribute column from which values
#'   will be recoded.
#' @param ... Single-length character vectors with the recoding instructions.
#'   The first component should have the value to replace and the second should
#'   have the replacement value (in the form `"[to_replace] -> [replacement]",
#'   ...`).
#' @param otherwise An optional single value for recoding any unmatched values.
#' @param edge_attr_to An optional name of a new edge attribute to which the
#'   recoded values will be applied. This will retain the original edge
#'   attribute and its values.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 4,
#'     m = 6,
#'     set_seed = 23) %>%
#'   set_edge_attrs(
#'     edge_attr = rel,
#'     values = c("a", "b", "a",
#'                "c", "b", "d"))
#'
#' # Get the graph's internal edf
#' # to show which edge attributes
#' # are available
#' graph %>% get_edge_df()
#'
#' # Recode the `rel` node
#' # attribute, creating a new edge
#' # attribute called `penwidth`;
#' # here, `a` is recoded to `1.0`,
#' # `b` maps to `1.5`, and all
#' # other values become `0.5`
#' graph <-
#'   graph %>%
#'   recode_edge_attrs(
#'     edge_attr_from = rel,
#'     "a -> 1.0",
#'     "b -> 1.5",
#'     otherwise = 0.5,
#'     edge_attr_to = penwidth)
#'
#' # Get the graph's internal edf
#' # to show that the node
#' # attribute values had been
#' # recoded and copied into a
#' # new node attribute
#' graph %>% get_edge_df()
#'
#' @family edge creation and removal
#'
#' @export
recode_edge_attrs <- function(
    graph,
    edge_attr_from,
    ...,
    otherwise = NULL,
    edge_attr_to = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains edges
  check_graph_contains_edges(graph)

  # Get the requested `edge_attr_from`
  edge_attr_from <-
    rlang::enquo(edge_attr_from) %>% rlang::get_expr() %>% as.character()

  # Get the requested `edge_attr_to`
  edge_attr_to <-
    rlang::enquo(edge_attr_to) %>% rlang::get_expr() %>% as.character()


  if (length(edge_attr_to) == 0) {
    edge_attr_to <- NULL
  }

  # Get list object from named vectors
  replacements <- list(...)

  # Extract the graph's edf
  edges <- get_edge_df(graph)

  # Get column names from the graph's edf
  column_names_graph <- colnames(edges)

  # Stop function if `edge_attr_from` is not one
  # of the graph's edge attributes
  if (!any(column_names_graph %in% edge_attr_from)) {

    cli::cli_abort(
      "The edge attribute to recode is not in the edf.")
  }

  # Get the column number for the edge attr to recode
  col_num_recode <-
    which(colnames(edges) %in% edge_attr_from)

  # Get the column class for the edge attr to recode
  edge_attr_class <- class(edges[, col_num_recode])

  # Extract the vector to recode from the `edges` df
  vector_to_recode <- edges[, col_num_recode]

  # Initialize the `indices_stack` vector
  indices_stack <- vector("numeric")

  # Parse the recoding pairs
  for (i in seq_along(replacements)) {

    pairing <-
      trimws(unlist(stringr::str_split(replacements[[i]], "->")))

    if (edge_attr_class == "numeric") {
      pairing <- as.numeric(pairing)
    }

    indices <- which(vector_to_recode %in% pairing[1])

    vector_to_recode[base::setdiff(indices, indices_stack)] <- pairing[2]

    indices_stack <- c(indices_stack, indices)
  }

  # If a value is supplied for `otherwise`, apply
  # that value to all unmatched
  if (!is.null(otherwise)) {

    otherwise_indices <-
      which(!(seq_len(nrow(edges)) %in% indices_stack))

    if (length(otherwise_indices) > 0) {
      vector_to_recode[otherwise_indices] <-
        otherwise
    }
  }

  # Rename the `vector_to_recode` object
  recoded_vector <- vector_to_recode

  if (!is.null(edge_attr_to)) {

    # Stop function if `edge_attr_to` is
    # `from` or `to`
    if (any(c("from", "to") %in% edge_attr_to)) {

      cli::cli_abort(
        "You cannot use the names `from` or `to`.")
    }

    if (any(column_names_graph %in% edge_attr_to)) {

      # Edge attribute exists and values will be
      # overwritten in this case
      col_num_write_to <-
        which(column_names_graph %in% edge_attr_to)

      edges[, col_num_write_to] <-
        recoded_vector
    }

    if (!any(column_names_graph %in% edge_attr_to)) {

      # Edge attribute does not exist and values be
      # part of a new edge attribute
      edges <-
        cbind(edges,
              data.frame(recoded_vector,
                         stringsAsFactors = FALSE))

      # Set the column name for the copied attr
      colnames(edges)[ncol(edges)] <- edge_attr_to
    }
  } else {
    # The edge attribute values will be overwritten
    # by the recoded value (no new edge attrs)
    edges[, col_num_recode] <- recoded_vector
  }

  # Replace the `edges_df` object in the graph
  # with the `edges` object
  graph$edges_df <- edges

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Delete one of the global graph attributes stored within a graph object
#'
#' @description
#'
#' Delete one of the global attributes stored within a graph object of class
#' `dgr_graph`).
#'
#' @inheritParams render_graph
#' @param attr The name of the attribute to delete for the `type` of global
#'   attribute specified.
#' @param attr_type The specific type of global graph attribute to delete. The
#'   type is specified with `graph`, `node`, or `edge`.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a new graph and add
#' # some extra global graph attrs
#' graph <-
#'   create_graph() %>%
#'   add_global_graph_attrs(
#'     attr = "overlap",
#'     value = "true",
#'     attr_type = "graph") %>%
#'   add_global_graph_attrs(
#'     attr = "penwidth",
#'     value = 3,
#'     attr_type = "node") %>%
#'   add_global_graph_attrs(
#'     attr = "penwidth",
#'     value = 3,
#'     attr_type = "edge")
#'
#' # Inspect the graph's global
#' # attributes
#' graph %>%
#'   get_global_graph_attr_info()
#'
#' # Delete the `penwidth` attribute
#' # for the graph's nodes using the
#' # `delete_global_graph_attrs()` fcn
#' graph <-
#'   graph %>%
#'   delete_global_graph_attrs(
#'     attr = "penwidth",
#'     attr_type = "node")
#'
#' # View the remaining set of global
#' # attributes for the graph
#' graph %>%
#'   get_global_graph_attr_info()
#'
#' @export
delete_global_graph_attrs <- function(
    graph,
    attr = NULL,
    attr_type = NULL
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If no `attr` or `attr_type` provided then
  # all global graph attributes will be removed
  if (is.null(attr) && is.null(attr_type)) {

    # Clear the global graph attributes data frame
    # by removing all rows from it
    graph$global_attrs <-
      graph$global_attrs[-(seq_len(nrow(graph$global_attrs))), ]

    message(
      glue::glue("Deleted all existing global graph attributes."))
  }

  # If an `attr` is provided but not an
  # `attr_type`, then delete all of those
  # `attr`s without regard to their type
  if (is.null(attr_type) && !is.null(attr)) {

    # Capture provided attr
    attr <- rlang::enquo(attr)

    graph$global_attrs <-
      graph$global_attrs %>%
      dplyr::filter(!(attr %in% !!attr))
  }

  if (!is.null(attr_type) && is.null(attr)) {

    # Stop function if `attr_type` is not a valid
    # attribute type
    if (length(attr_type) > 1) {
      cli::cli_abort("Problem. attr_type must be")
    }
    rlang::arg_match(attr_type, c("graph", "node", "edge"), multiple = TRUE)

    # Capture provided `attr_type`
    attr_type <- rlang::enquo(attr_type)

    graph$global_attrs <-
      graph$global_attrs %>%
      dplyr::filter(!(attr_type %in% !!attr_type))
  }

  if (!is.null(attr_type) && !is.null(attr)) {

    # Stop function if `attr_type` is not a valid
    # attribute type
    rlang::arg_match(attr_type, c("graph", "node", "edge"), multiple = TRUE)

    # Get the global graph attributes already set
    # in the graph object
    global_attrs_available <- graph$global_attrs

    # Create a table with a single row for the
    # attribute to remove
    global_attrs_to_remove <-
      dplyr::tibble(
        attr = as.character(attr),
        value = NA_character_,
        attr_type = as.character(attr_type)) %>%
      as.data.frame(stringsAsFactors = FALSE)

    # Use the `anti_join()` to remove global attribute
    # rows from the graph
    global_attrs_joined <-
      global_attrs_available %>%
      dplyr::anti_join(
        global_attrs_to_remove,
        by = c("attr", "attr_type"))

    # Replace the graph's global attributes with
    # the revised set
    graph$global_attrs <- global_attrs_joined
  }

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get a count of all nodes
#'
#' @description
#'
#' From a graph object of class `dgr_graph`, get a count of nodes in the graph.
#'
#' @inheritParams render_graph
#'
#' @return A numeric vector of single length.
#'
#' @examples
#' # Create a graph with a
#' # path of nodes and 3
#' # unconnected nodes
#' graph <-
#'   create_graph() %>%
#'   add_path(n = 3) %>%
#'   add_n_nodes(n = 3)
#'
#' # Get a count of all nodes
#' # in the graph
#' graph %>%
#'   count_nodes()
#'
#' @export
count_nodes <- function(graph) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If graph is empty, return 0
  if (is_graph_empty(graph)) {
    return(0L)
  }

  nrow(graph$nodes_df)
}

#' Insert edge aesthetic attributes during edge creation
#'
#' @description
#'
#' This helper function should be invoked to provide values for the namesake
#' `edge_aes` argument, which is present in any function where edges are
#' created.
#'
#' @param style The edge line style. The `style` types that can be used are
#'   `solid`, `bold`, `dashed`, `dotted`, `tapered`, and `invisible`.
#' @param penwidth The thickness of the stroke line for the edge itself.
#' @param color The color of the edge. Can be an X11 color or a hexadecimal
#'   color code.
#' @param arrowsize A scaling factor for arrowheads. The default value is `1.0`
#'   and the minimum is `0`.
#' @param arrowhead The type of arrowhead to use. The `style` attribute can be
#'   any of these types: `normal`, `vee`, `tee`, `dot`, `diamond`, `box`,
#'   `curve`, `icurve`, `inv`, `crow`, or `none`.
#' @param arrowtail The type of arrowtail to use. The `style` attribute can any
#'   of these types: `normal`, `vee`, `tee`, `dot`, `diamond`, `box`, `curve`,
#'   `icurve`, `inv`, `crow`, or `none`.
#' @param fontname The name of the system font that will be used for any edge
#'   text.
#' @param fontsize The point size of the font used for any edge text.
#' @param fontcolor The color used for any edge text. Can be an X11 color or a
#'   hexadecimal color code.
#' @param len The preferred edge length for an edge, in inches. Default value is
#'   `1.0`.
#' @param tooltip Text for a tooltip that appears when hovering over an edge. If
#'   text is not provided, then the default tooltip text will provide the edge
#'   definition (i.e., `[id]->[id] or [id]--[id]`).
#' @param URL A URL to associate with an edge. Upon rendering the plot, clicking
#'   edges with any associated URLs will open the URL in the default browser.
#' @param label The label text associated with the edge. This text will appear
#'   near the center of the edge.
#' @param labelfontname The name of the system font that will be used for the
#'   `headlabel` and the `taillabel` label text. If not set, the `fontname`
#'   value will instead be used.
#' @param labelfontsize The point size of the font used for the `headlabel` and
#'   the `taillabel` label text. If not set, the `fontsize` value will instead
#'   be used.
#' @param labelfontcolor The color used for the label text of the `headlabel`
#'   and the `taillabel` label text. If not set, the `fontcolor` value will
#'   instead be used. Can be an X11 color or a hexadecimal color code.
#' @param labeltooltip Text for a tooltip that will appear when hovering over
#'   the main label of an edge (if label text provided in the `label` edge
#'   attribute). If text is not provided and an edge label is visible, then the
#'   default tooltip text will provide the edge definition (i.e., `[id]->[id] or
#'   [id]--[id]`).
#' @param labelURL A URL to associate with edge label text. Upon rendering the
#'   plot, clicking edge labels with any associated URLs will open the URL in
#'   the default browser.
#' @param edgetooltip This option provides a means to specify a tooltip with
#'   only the non-label parts of an edge. If this is defined, the value
#'   overrides any `tooltip` defined for the edge. This tooltip text is when
#'   hovering along the edge (even near the head or tail node) unless overridden
#'   by a `headtooltip` or `tailtooltip` value.
#' @param edgeURL This option provides a means to specify a URL with only the
#'   non-label parts of an edge. If this is defined, the value overrides any
#'   `URL` defined for the edge. This URL is used along the edge (even near the
#'   head or tail node) unless overridden by a `headURL` or `tailURL` value.
#' @param dir An optional direction type. Normally, for directed graphs, this is
#'   `forward` and needn't be set. For undirected graphs, this would be `none`
#'   and again no explicit setting is required. However, one can also use the
#'   `back` or `both` options. The `back` option draws an arrowhead in the
#'   reverse direction of an edge. The `both` option draws two arrowheads. When
#'   using any of these options in such an explicit manner, the `head...` and
#'   `tail...` edge attributes allow control over aesthetic edge attributes in
#'   either side of the edge.
#' @param headtooltip This option provides a means to specify a tooltip that can
#'   be displayed by hovering over the part of an edge that is adjacent to
#'   incoming node (see the `tooltip` argument for further details).
#' @param headURL This option provides a means to specify a URL that can be
#'   accessed by clicking the part of an edge that is adjacent to incoming node
#'   (see the `URL` argument for further details).
#' @param headclip If `TRUE` (the default behavior), then the head of the
#'   affected edge is clipped to the node boundary. Using `FALSE` places the
#'   head of the outgoing edge at the center of its node.
#' @param headlabel This option provides a means to display a label near the
#'   part of an edge that is adjacent to incoming node (see the `label` argument
#'   for further details).
#' @param headport Allows one to specify which compass position on the incoming
#'   node the head of the edge will alight. Options are `n`, `ne`, `e`, `se`,
#'   `s`, `sw`, `w`, and `nw`.
#' @param tailtooltip This option provides a means to specify a tooltip that can
#'   be displayed by hovering over the part of an edge that is adjacent to
#'   outgoing node (see the `tooltip` argument for further details).
#' @param tailURL This option provides a means to specify a URL that can be
#'   accessed by clicking the part of an edge that is adjacent to outgoing node
#'   (see the `URL` argument for further details).
#' @param tailclip If `TRUE` (the default behavior), then the tail of the
#'   affected edge is clipped to the node boundary. Using `FALSE` places the
#'   tail of the outgoing edge at the center of its node.
#' @param taillabel This option provides a means to display a label near the
#'   part of an edge that is adjacent to outgoing node (see the `label` argument
#'   for further details).
#' @param tailport Allows one to specify which compass position on the outgoing
#'   node the tail of the edge will be emitted from. Options are `n`, `ne`, `e`,
#'   `se`, `s`, `sw`, `w`, and `nw`.
#' @param decorate If `TRUE` then attach any edge label to the edge line via a
#'   2-segment polyline, underlining the label text and partially overlapping
#'   the edge line.
#'
#' @examples
#' # Create a new graph and add
#' # a path with several edge
#' # aesthetic attributes
#' graph <-
#'   create_graph() |>
#'   add_path(
#'     n = 3,
#'     type = "path",
#'     edge_aes = edge_aes(
#'       style = "dot",
#'       color = c("red", "blue")))
#'
#' # View the graph's internal
#' # node data frame; the node
#' # aesthetic attributes have
#' # been inserted
#' graph |>
#'   get_edge_df()
#'
#' @family aesthetics
#'
#' @export
edge_aes <- function(style = NULL,
                     penwidth = NULL,
                     color = NULL,
                     arrowsize = NULL,
                     arrowhead = NULL,
                     arrowtail = NULL,
                     fontname = NULL,
                     fontsize = NULL,
                     fontcolor = NULL,
                     len = NULL,
                     tooltip = NULL,
                     URL = NULL,
                     label = NULL,
                     labelfontname = NULL,
                     labelfontsize = NULL,
                     labelfontcolor = NULL,
                     labeltooltip = NULL,
                     labelURL = NULL,
                     edgetooltip = NULL,
                     edgeURL = NULL,
                     dir = NULL,
                     headtooltip = NULL,
                     headURL = NULL,
                     headclip = NULL,
                     headlabel = NULL,
                     headport = NULL,
                     tailtooltip = NULL,
                     tailURL = NULL,
                     tailclip = NULL,
                     taillabel = NULL,
                     tailport = NULL,
                     decorate = NULL) {

  # Collect vectors of node aesthetic
  # attribute values into a list object
  attr_values <-
    list(
      style = style,
      penwidth = penwidth,
      color = color,
      arrowsize = arrowsize,
      arrowhead = arrowhead,
      arrowtail = arrowtail,
      fontname = fontname,
      fontsize = fontsize,
      fontcolor = fontcolor,
      len = len,
      tooltip = tooltip,
      URL = URL,
      label = label,
      labelfontname = labelfontname,
      labelfontsize = labelfontsize,
      labelfontcolor = labelfontcolor,
      labeltooltip = labeltooltip,
      labelURL = labelURL,
      edgetooltip = edgetooltip,
      edgeURL = edgeURL,
      dir = dir,
      headtooltip = headtooltip,
      headURL = headURL,
      headclip = headclip,
      headlabel = headlabel,
      headport = headport,
      tailtooltip = tailtooltip,
      tailURL = tailURL,
      tailclip = tailclip,
      taillabel = taillabel,
      tailport = tailport,
      decorate = decorate)

  non_null_attrs <-
    seq_along(attr_values) |> # 1:length(attr_values)
    purrr::map_chr(.f = function(x) {
      if (is.null(attr_values[[x]])) {
        NA_character_
      } else {
        names(attr_values[x])
      }
    })

  non_null_attrs <- non_null_attrs[which(!is.na(non_null_attrs))]

  attr_values[non_null_attrs]
}

#' Mutate a set of node attribute values
#'
#' @description
#'
#' Within a graph's internal node data frame (ndf), mutate numeric node
#' attribute values using one or more expressions.
#'
#' @inheritParams render_graph
#' @param ... Expressions used for the mutation of node attributes. LHS of each
#'   expression is either an existing or new node attribute name. The RHS can
#'   consist of any valid R code that uses node attributes as variables.
#'   Expressions are evaluated in the order provided, so, node attributes
#'   created or modified are ready to use in subsequent expressions.
#'
#' @return A graph object of class `dgr_graph`.
#'
#' @examples
#' # Create a graph with 3 nodes
#' graph <-
#'   create_graph() |>
#'   add_path(n = 3) |>
#'   set_node_attrs(
#'     node_attr = width,
#'     values = c(1.4, 0.3, 1.1))
#'
#' # Get the graph's internal ndf
#' # to show which node attributes
#' # are available
#' graph |> get_node_df()
#'
#' # Mutate the `width` node
#' # attribute, dividing each
#' # value by 2
#' graph <-
#'   graph |>
#'   mutate_node_attrs(
#'     width = width / 2)
#'
#' # Get the graph's internal
#' # ndf to show that the node
#' # attribute `width` had its
#' # values changed
#' graph |> get_node_df()
#'
#' # Create a new node attribute,
#' # called `length`, that is the
#' # log of values in `width` plus
#' # 2 (and, also, round all values
#' # to 2 decimal places)
#' graph <-
#'   graph |>
#'   mutate_node_attrs(
#'     length = (log(width) + 2) |>
#'                round(2))
#'
#' # Get the graph's internal ndf
#' # to show that the node attribute
#' # values had been mutated
#' graph |> get_node_df()
#'
#' # Create a new node attribute
#' # called `area`, which is the
#' # product of the `width` and
#' # `length` attributes
#' graph <-
#'   graph |>
#'   mutate_node_attrs(
#'     area = width * length)
#'
#' # Get the graph's internal ndf
#' # to show that the node attribute
#' # values had been multiplied
#' # together (with new attr `area`)
#' graph |> get_node_df()
#'
#' @family node creation and removal
#'
#' @export
mutate_node_attrs <- function(
    graph,
    ...
) {

  # Get the time of function start
  time_function_start <- Sys.time()

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # Validation: Graph contains nodes
  check_graph_contains_nodes(graph)

  # Collect expressions
  exprs <- rlang::exprs(...)

  # Extract the graph's ndf
  ndf <- get_node_df(graph)

  # Stop function if any supplied
  # expressions mutate columns that
  # should not be changed
  if ("id" %in% names(exprs)) {
    cli::cli_abort("The variable `id` cannot undergo mutation.")
  }

  ndf <- ndf |> dplyr::mutate(!!!enquos(...))

  # Update the graph
  graph$nodes_df <- ndf

  # Get the name of the function
  fcn_name <- get_calling_fcn()

  # Update the `graph_log` df with an action
  graph$graph_log <-
    add_action_to_log(
      graph_log = graph$graph_log,
      version_id = nrow(graph$graph_log) + 1L,
      function_used = fcn_name,
      time_modified = time_function_start,
      duration = graph_function_duration(time_function_start),
      nodes = nrow(graph$nodes_df),
      edges = nrow(graph$edges_df))

  # Write graph backup if the option is set
  if (graph$graph_info$write_backups) {
    save_graph_as_rds(graph = graph)
  }

  graph
}

#' Get an aggregate value from the outdegree of nodes
#'
#' @description
#'
#' Get a single, aggregate value from the outdegree values for all nodes in a
#' graph, or, a subset of graph nodes.
#'
#' @inheritParams render_graph
#' @param agg The aggregation function to use for summarizing outdegree values
#'   from graph nodes. The following aggregation functions can be used: `sum`,
#'   `min`, `max`, `mean`, or `median`.
#' @param conditions An option to use filtering conditions for the nodes to
#'   consider.
#'
#' @return A vector with an aggregate outdegree value.
#'
#' @examples
#' # Create a random graph using the
#' # `add_gnm_graph()` function
#' graph <-
#'   create_graph() %>%
#'   add_gnm_graph(
#'     n = 20,
#'     m = 35,
#'     set_seed = 23) %>%
#'   set_node_attrs(
#'     node_attr = value,
#'     values = rnorm(
#'       n = count_nodes(.),
#'       mean = 5,
#'       sd = 1) %>% round(1))
#'
#' # Get the mean outdegree value from all
#' # nodes in the graph
#' graph %>%
#'   get_agg_degree_out(
#'     agg = "mean")
#'
#' # Other aggregation functions can be used
#' # (`min`, `max`, `median`, `sum`); let's
#' # get the median in this example
#' graph %>%
#'   get_agg_degree_out(
#'     agg = "median")
#'
#' # The aggregation of outdegree can occur
#' # for a subset of the graph nodes and this
#' # is made possible by specifying `conditions`
#' # for the nodes
#' graph %>%
#'   get_agg_degree_out(
#'     agg = "mean",
#'     conditions = value < 5.0)
#'
#' @export
get_agg_degree_out <- function(
    graph,
    agg,
    conditions = NULL
) {

  # Validation: Graph object is valid
  check_graph_valid(graph)

  # If filtering conditions are provided then
  # pass in those conditions and filter the ndf
  if (!rlang::quo_is_null(rlang::enquo(conditions))) {

    # Extract the node data frame from the graph
    ndf <- get_node_df(graph)

    # Apply filtering conditions to the ndf
    ndf <- dplyr::filter(.data = ndf, {{ conditions }})

    # Get a vector of node ID values
    node_ids <-
      ndf %>%
      dplyr::pull("id")
  }

  # Get a data frame with outdegree values for
  # all nodes in the graph
  outdegree_df <- get_degree_out(graph)

  if (exists("node_ids")) {
    outdegree_df <-
      outdegree_df %>%
      dplyr::filter(id %in% node_ids)
  }

  # Verify that the value provided for `agg`
  # is one of the accepted aggregation types
  arg_match0(agg, c("sum", "min", "max", "mean", "median"))

  # Get the aggregate value of total degree based
  # on the aggregate function provided
  fun <- match.fun(agg)

  outdegree_agg <-
    outdegree_df %>%
    dplyr::group_by() %>%
    dplyr::summarize(fun(outdegree, na.rm = TRUE), .groups = "drop") %>%
    purrr::flatten_dbl()

  outdegree_agg
}

# "Getting a table of X11 color names and hexadecimal values"

test_that("a data frame X11 color names and hexadecimal values can be made", {

  # Get a data frame object containing X11 color names and hex values
  x11_hex_df <- x11_hex()

  # Expect an object of class "data.frame"
  expect_s3_class(x11_hex_df, "data.frame")

  # Expect that the "x11_hex_df" data frame has 2 columns
  expect_equal(ncol(x11_hex_df), 2)

  # Expect that the "x11_hex_df" data frame has 655 rows
  expect_equal(nrow(x11_hex_df), 655)

  # Expect specific column names
  expect_in(colnames(x11_hex_df), c("x11_name", "hex"))

  # Expect that both columns are of the "character" class
  expect_type(x11_hex_df[, 1], "character")
  expect_type(x11_hex_df[, 2], "character")

  # Expect that the color names are lowercased with no spaces
  expect_match(x11_hex_df[, 1], "[a-z1-4]")

  # Expect that the hexadecimal color values are well-formed
  expect_match(x11_hex_df[, 2], "^#[a-f0-9]{6}$")
})

# Getting edge attributes is possible

    Code
      get_edge_attrs(graph = graph, edge_attr = from)
    Condition
      Error in `get_edge_attrs()`:
      ! This is not an edge attribute.
    Code
      get_edge_attrs(graph = graph, edge_attr = to)
    Condition
      Error in `get_edge_attrs()`:
      ! This is not an edge attribute.
    Code
      get_edge_attrs(graph = graph, edge_attr = value, from = c(1, 2), to = 2)
    Condition
      Error in `get_edge_attrs()`:
      ! The number of nodes in `from` and `to` must be the same.

# Reversing edges in a selection is possible

    Code
      create_graph(directed = FALSE) %>% add_balanced_tree(k = 2, h = 2) %>%
        select_edges_by_node_id(nodes = 1:2) %>% rev_edge_dir_ws()
    Condition
      Error in `rev_edge_dir_ws()`:
      ! The input graph must be a directed graph.

# Fully disconnecting selected nodes is possible

    Code
      create_graph() %>% add_path(n = 2) %>% fully_disconnect_nodes_ws()
    Condition
      Error in `fully_disconnect_nodes_ws()`:
      ! There is no selection of nodes available.

# getting or selecting the last nodes created is possible

    Code
      get_last_nodes_created(graph_node_deleted)
    Condition
      Error in `get_last_nodes_created()`:
      ! The previous graph transformation function resulted in a removal of nodes.
    Code
      select_last_nodes_created(graph_node_deleted)
    Condition
      Error in `select_last_nodes_created()`:
      ! The previous graph transformation function resulted in a removal of nodes.

# the dice similarity algorithm is functional

    Code
      get_dice_similarity(graph = graph, nodes = 8:12, direction = "all")
    Condition
      Error in `get_dice_similarity()`:
      ! One or more nodes provided not in graph.

# the Jaccard similarity algorithm is functional

    Code
      get_jaccard_similarity(graph = graph, nodes = 5:7, direction = "away")
    Condition
      Error in `get_jaccard_similarity()`:
      ! `direction` must be one of "all", "in", or "out", not "away".

# edges can be deleted from a graph using node label values

    Code
      graph_labeled_nodes %>% delete_edge(from = "zero", to = "two")
    Condition
      Error in `delete_edge()`:
      ! The value provided in `from` does not exist as a node `label` value.
    Code
      graph_labeled_nodes %>% delete_edge(from = "one", to = "three")
    Condition
      Error in `delete_edge()`:
      ! The value provided in `to` does not exist as a node `label` value.

# adding edges from a table to a graph is possible

    Code
      graph %>% add_edges_from_table(edge_table, from_col = from, to_col = to_currency,
        from_to_map = iso_4217_code)
    Condition
      Error in `add_edges_from_table()`:
      ! The value specified in `from_col` is not in the table.
    Code
      graph %>% add_edges_from_table(edge_table, from_col = from_currency, to_col = to,
        from_to_map = iso_4217_code)
    Condition
      Error in `add_edges_from_table()`:
      ! The value specified in `to_col` is not in the table.
    Code
      graph %>% add_edges_from_table(edge_table, from_col = from_currency, to_col = to_currency,
        from_to_map = iso_4217)
    Condition
      Error in `add_edges_from_table()`:
      ! The value specified in `from_to_map` is not in the graph.

# reverse edges can be added given a selection of edges

    Code
      create_graph() %>% add_n_nodes(n = 2, type = "type_a", label = c("a_1", "a_2")) %>%
        add_edge(from = 1, to = 2, rel = "a") %>% add_reverse_edges_ws("b")
    Condition
      Error in `add_reverse_edges_ws()`:
      ! The graph contains no selection of edges.

# a subgraph can be created and such an object is correct

    Code
      transform_to_subgraph_ws(graph)
    Condition
      Error in `transform_to_subgraph_ws()`:
      ! There is no selection of node or edges available.

# selecting a node in a graph is possible

    Code
      select_nodes(graph = graph, nodes = 5)
    Condition
      Error in `select_nodes()`:
      ! `nodes` must correspond to values in the graph.
      i `Graph values IDs include 1, 2, 3, and 4, not 5.
    Code
      select_nodes(graph = graph, nodes = 4:5)
    Output
      DiagrammeR Graph // 4 nodes / 3 edges
        -- directed / connected / DAG / property graph / simple
      
        NODES / type: 1 vals - complete / label: 4 vals - complete & unique
          -- 1 additional node attribute (value)
        EDGES / rel: 1 vals - complete                           info: `get_edge_df()`
          -- 1 additional edge attribute (width)
        SELECTION / 1 node selected                            info: `get_selection()`
        CACHE / <none>
        GLOBAL ATTRS / 17 are set                 info: `get_global_graph_attr_info()`
        GRAPH ACTIONS / <none>
        GRAPH LOG / create_graph() -> select_nodes()

# Setting a cache is possible

    Code
      set_cache(graph = graph, name = "closeness_df_2", to_cache = closeness_df)
    Condition
      Error in `set_cache()`:
      ! `col` must be a column name that exists in the data frame.
    Code
      set_cache(graph = graph, name = "closeness_df_3", to_cache = closeness_df, col = "nonexistent")
    Condition
      Error in `set_cache()`:
      ! `col` must be a column name that exists in the data frame.

# Adding a balanced tree is possible

    Code
      add_balanced_tree(graph = graph, k = 1, h = 3, type = "a", rel = "z")
    Condition
      Error in `add_balanced_tree()`:
      ! `k` must be a whole number larger than or equal to 2, not the number 1.
    Code
      add_balanced_tree(graph = graph, k = 3, h = 1, type = "a", rel = "z")
    Condition
      Error in `add_balanced_tree()`:
      ! `h` must be a whole number larger than or equal to 2, not the number 1.

# Adding a path is possible

    Code
      add_path(graph = graph, n = 1, type = "a", rel = "z")
    Condition
      Error in `add_path()`:
      ! `n` must be a whole number larger than or equal to 2, not the number 1.

# Adding a preferential attachment graph is possible

    Code
      create_graph() %>% add_pa_graph(n = 0, m = 1)
    Condition
      Error in `add_pa_graph()`:
      ! `n` must be a whole number larger than or equal to 1, not the number 0.
    Code
      create_graph() %>% add_pa_graph(n = 0, m = 1, algo = "plumtree")
    Condition
      Error in `add_pa_graph()`:
      ! `n` must be a whole number larger than or equal to 1, not the number 0.

# Getting a graph from a series is possible

    Code
      create_graph_series() %>% get_graph_from_graph_series(graph_no = 1)
    Condition
      Error in `get_graph_from_graph_series()`:
      ! There are no graphs in this graph series.

# Adding color based on edge attributes is possible

    Code
      create_graph() %>% add_gnm_graph(n = 10, m = 10, edge_data = edge_data(weight = rnorm(
        10, 5, 2)), set_seed = 23) %>% set_edge_attrs(edge_attr = "rel", values = c(
        "A", "A", "B", "B", "D", "A", "B", "C", "D", "A")) %>% colorize_edge_attrs(
        edge_attr_from = rel, edge_attr_to = color, palette = c("#458b00l15", "foo",
          "#00eeee", "bar", "orange"))
    Condition
      Error in `colorize_edge_attrs()`:
      ! The color palette contains invalid hexadecimal values.

# simple traversals are possible

    Code
      graph <- graph %>% clear_selection()
      trav_in(graph)
    Condition
      Error in `trav_in()`:
      ! There is no selection of nodes available.
      * any traversal requires an active selection
      * this type of traversal requires a selection of nodes
    Code
      trav_out(graph)
    Condition
      Error in `trav_out()`:
      ! There is no selection of nodes available.
      * Any traversal requires an active selection.
      * This type of traversal requires a selection of nodes.
    Code
      trav_both(graph)
    Condition
      Error in `trav_both()`:
      ! There is no selection of nodes available.
      * Any traversal requires an active selection.
      * This type of traversal requires a selection of nodes.
    Code
      trav_in_node(graph)
    Condition
      Error in `trav_in_node()`:
      ! The graph contains no selection of edges.
      * Any traversal requires an active selection.
      * This type of traversal requires a selection of edges.
    Code
      trav_out_node(graph)
    Condition
      Error in `trav_out_node()`:
      ! The graph contains no selection of edges.
      * Any traversal requires an active selection.
      * This type of traversal requires a selection of edges.
    Code
      trav_in_edge(graph)
    Condition
      Error in `trav_in_edge()`:
      ! There is no selection of nodes available.
      * Any traversal requires an active selection.
      * This type of traversal requires a selection of nodes.
    Code
      trav_out_edge(graph)
    Condition
      Error in `trav_out_edge()`:
      ! There is no selection of nodes available.
      * Any traversal requires an active selection.
      * This type of traversal requires a selection of nodes.

# Using an invalid graph triggers an error.

    Code
      add_balanced_tree(invalid_graph)
    Condition
      Error in `add_balanced_tree()`:
      ! The graph is not valid.
    Code
      add_cycle(invalid_graph)
    Condition
      Error in `add_cycle()`:
      ! The graph is not valid.
    Code
      add_edge(invalid_graph)
    Condition
      Error in `add_edge()`:
      ! The graph is not valid.
    Code
      add_edge_clone(invalid_graph)
    Condition
      Error in `add_edge_clone()`:
      ! The graph is not valid.
    Code
      add_edge_df(invalid_graph)
    Condition
      Error in `add_edge_df()`:
      ! The graph is not valid.
    Code
      add_edges_from_table(invalid_graph)
    Condition
      Error in `add_edges_from_table()`:
      ! The graph is not valid.
    Code
      add_edges_w_string(invalid_graph)
    Condition
      Error in `add_edges_w_string()`:
      ! The graph is not valid.
    Code
      add_forward_edges_ws(invalid_graph)
    Condition
      Error in `add_forward_edges_ws()`:
      ! The graph is not valid.
    Code
      add_full_graph(invalid_graph)
    Condition
      Error in `add_full_graph()`:
      ! The graph is not valid.
    Code
      add_global_graph_attrs(invalid_graph)
    Condition
      Error in `add_global_graph_attrs()`:
      ! The graph is not valid.
    Code
      add_gnm_graph(invalid_graph)
    Condition
      Error in `add_gnm_graph()`:
      ! The graph is not valid.
    Code
      add_gnp_graph(invalid_graph)
    Condition
      Error in `add_gnp_graph()`:
      ! The graph is not valid.
    Code
      add_graph_action(invalid_graph)
    Condition
      Error in `add_graph_action()`:
      ! The graph is not valid.
    Code
      add_graph_to_graph_series(graph_series = invalid_graph)
    Condition
      Error in `add_graph_to_graph_series()`:
      ! `graph` is absent but must be supplied.
    Code
      add_grid_2d(invalid_graph)
    Condition
      Error in `add_grid_2d()`:
      ! The graph is not valid.
    Code
      add_grid_3d(invalid_graph)
    Condition
      Error in `add_grid_3d()`:
      ! The graph is not valid.
    Code
      add_growing_graph(invalid_graph)
    Condition
      Error in `add_growing_graph()`:
      ! The graph is not valid.
    Code
      add_islands_graph(invalid_graph)
    Condition
      Error in `add_islands_graph()`:
      ! The graph is not valid.
    Code
      add_n_node_clones(invalid_graph)
    Condition
      Error in `add_n_node_clones()`:
      ! The graph is not valid.
    Code
      add_n_nodes(invalid_graph)
    Condition
      Error in `add_n_nodes()`:
      ! The graph is not valid.
    Code
      add_n_nodes_ws(invalid_graph)
    Condition
      Error in `add_n_nodes_ws()`:
      ! The graph is not valid.
    Code
      add_node(invalid_graph)
    Condition
      Error in `add_node()`:
      ! The graph is not valid.
    Code
      add_node_clones_ws(invalid_graph)
    Condition
      Error in `add_node_clones_ws()`:
      ! The graph is not valid.
    Code
      add_node_df(invalid_graph)
    Condition
      Error in `add_node_df()`:
      ! The graph is not valid.
    Code
      add_nodes_from_df_cols(invalid_graph)
    Condition
      Error in `add_nodes_from_df_cols()`:
      ! The graph is not valid.
    Code
      add_nodes_from_table(invalid_graph)
    Condition
      Error in `add_nodes_from_table()`:
      ! The graph is not valid.
    Code
      add_pa_graph(invalid_graph)
    Condition
      Error in `add_pa_graph()`:
      ! The graph is not valid.
    Code
      add_path(invalid_graph)
    Condition
      Error in `add_path()`:
      ! The graph is not valid.
    Code
      add_prism(invalid_graph)
    Condition
      Error in `add_prism()`:
      ! The graph is not valid.
    Code
      add_smallworld_graph(invalid_graph)
    Condition
      Error in `add_smallworld_graph()`:
      ! The graph is not valid.
    Code
      add_star(invalid_graph)
    Condition
      Error in `add_star()`:
      ! The graph is not valid.
    Code
      add_reverse_edges_ws(invalid_graph)
    Condition
      Error in `add_reverse_edges_ws()`:
      ! The graph is not valid.
    Code
      clear_selection(invalid_graph)
    Condition
      Error in `clear_selection()`:
      ! The graph is not valid.
    Code
      colorize_edge_attrs(invalid_graph)
    Condition
      Error in `colorize_edge_attrs()`:
      ! The graph is not valid.
    Code
      colorize_node_attrs(invalid_graph)
    Condition
      Error in `colorize_node_attrs()`:
      ! The graph is not valid.
    Code
      copy_edge_attrs(invalid_graph)
    Condition
      Error in `copy_edge_attrs()`:
      ! The graph is not valid.
    Code
      copy_node_attrs(invalid_graph)
    Condition
      Error in `copy_node_attrs()`:
      ! The graph is not valid.
    Code
      count_asymmetric_node_pairs(invalid_graph)
    Condition
      Error in `count_asymmetric_node_pairs()`:
      ! The graph is not valid.
    Code
      count_automorphisms(invalid_graph)
    Condition
      Error in `count_automorphisms()`:
      ! The graph is not valid.
    Code
      count_edges(invalid_graph)
    Condition
      Error in `count_edges()`:
      ! The graph is not valid.
    Code
      count_loop_edges(invalid_graph)
    Condition
      Error in `count_loop_edges()`:
      ! The graph is not valid.
    Code
      count_mutual_node_pairs(invalid_graph)
    Condition
      Error in `count_mutual_node_pairs()`:
      ! The graph is not valid.
    Code
      count_nodes(invalid_graph)
    Condition
      Error in `count_nodes()`:
      ! The graph is not valid.
    Code
      count_s_connected_cmpts(invalid_graph)
    Condition
      Error in `count_s_connected_cmpts()`:
      ! The graph is not valid.
    Code
      count_unconnected_node_pairs(invalid_graph)
    Condition
      Error in `count_unconnected_node_pairs()`:
      ! The graph is not valid.
    Code
      count_unconnected_nodes(invalid_graph)
    Condition
      Error in `count_unconnected_nodes()`:
      ! The graph is not valid.
    Code
      count_w_connected_cmpts(invalid_graph)
    Condition
      Error in `count_w_connected_cmpts()`:
      ! The graph is not valid.
    Code
      create_graph_series(invalid_graph)
    Condition
      Error in `create_graph_series()`:
      ! The graph is not valid.
    Code
      delete_edge(invalid_graph)
    Condition
      Error in `delete_edge()`:
      ! The graph is not valid.
    Code
      delete_edges_ws(invalid_graph)
    Condition
      Error in `delete_edges_ws()`:
      ! The graph is not valid.
    Code
      delete_global_graph_attrs(invalid_graph)
    Condition
      Error in `delete_global_graph_attrs()`:
      ! The graph is not valid.
    Code
      delete_graph_actions(invalid_graph)
    Condition
      Error in `delete_graph_actions()`:
      ! The graph is not valid.
    Code
      delete_loop_edges_ws(invalid_graph)
    Condition
      Error in `delete_loop_edges_ws()`:
      ! The graph is not valid.
    Code
      deselect_edges(invalid_graph)
    Condition
      Error in `deselect_edges()`:
      ! The graph is not valid.
    Code
      deselect_nodes(invalid_graph)
    Condition
      Error in `deselect_nodes()`:
      ! The graph is not valid.
    Code
      delete_node(invalid_graph)
    Condition
      Error in `delete_node()`:
      ! The graph is not valid.
    Code
      delete_nodes_ws(invalid_graph)
    Condition
      Error in `delete_nodes_ws()`:
      ! The graph is not valid.
    Code
      display_metagraph(invalid_graph)
    Condition
      Error in `display_metagraph()`:
      ! The graph is not valid.
    Code
      do_bfs(invalid_graph)
    Condition
      Error in `do_bfs()`:
      ! The graph is not valid.
    Code
      do_dfs(invalid_graph)
    Condition
      Error in `do_dfs()`:
      ! The graph is not valid.
    Code
      drop_edge_attrs(invalid_graph)
    Condition
      Error in `drop_edge_attrs()`:
      ! The graph is not valid.
    Code
      drop_node_attrs(invalid_graph)
    Condition
      Error in `drop_node_attrs()`:
      ! The graph is not valid.
    Code
      export_csv(invalid_graph)
    Condition
      Error in `export_csv()`:
      ! The graph is not valid.
    Code
      export_graph(invalid_graph)
    Condition
      Error in `export_graph()`:
      ! The graph is not valid.
    Code
      fully_connect_nodes_ws(invalid_graph)
    Condition
      Error in `fully_connect_nodes_ws()`:
      ! The graph is not valid.
    Code
      fully_disconnect_nodes_ws(invalid_graph)
    Condition
      Error in `fully_disconnect_nodes_ws()`:
      ! The graph is not valid.
    Code
      generate_dot(invalid_graph)
    Condition
      Error in `generate_dot()`:
      ! The graph is not valid.
    Code
      get_adhesion(invalid_graph)
    Condition
      Error in `get_adhesion()`:
      ! The graph is not valid.
    Code
      get_agg_degree_in(invalid_graph)
    Condition
      Error in `get_agg_degree_in()`:
      ! The graph is not valid.
    Code
      get_agg_degree_out(invalid_graph)
    Condition
      Error in `get_agg_degree_out()`:
      ! The graph is not valid.
    Code
      get_agg_degree_total(invalid_graph)
    Condition
      Error in `get_agg_degree_total()`:
      ! The graph is not valid.
    Code
      get_all_connected_nodes(invalid_graph)
    Condition
      Error in `get_all_connected_nodes()`:
      ! The graph is not valid.
    Code
      get_alpha_centrality(invalid_graph)
    Condition
      Error in `get_alpha_centrality()`:
      ! The graph is not valid.
    Code
      get_articulation_points(invalid_graph)
    Condition
      Error in `get_articulation_points()`:
      ! The graph is not valid.
    Code
      get_authority_centrality(invalid_graph)
    Condition
      Error in `get_authority_centrality()`:
      ! The graph is not valid.
    Code
      get_betweenness(invalid_graph)
    Condition
      Error in `get_betweenness()`:
      ! The graph is not valid.
    Code
      get_cache(invalid_graph)
    Condition
      Error in `get_cache()`:
      ! The graph is not valid.
    Code
      get_closeness(invalid_graph)
    Condition
      Error in `get_closeness()`:
      ! The graph is not valid.
    Code
      get_closeness_vitality(invalid_graph)
    Condition
      Error in `get_closeness_vitality()`:
      ! The graph is not valid.
    Code
      get_cmty_edge_btwns(invalid_graph)
    Condition
      Error in `get_cmty_edge_btwns()`:
      ! The graph is not valid.
    Code
      get_cmty_fast_greedy(invalid_graph)
    Condition
      Error in `get_cmty_fast_greedy()`:
      ! The graph is not valid.
    Code
      get_cmty_l_eigenvec(invalid_graph)
    Condition
      Error in `get_cmty_l_eigenvec()`:
      ! The graph is not valid.
    Code
      get_cmty_louvain(invalid_graph)
    Condition
      Error in `get_cmty_louvain()`:
      ! The graph is not valid.
    Code
      get_cmty_walktrap(invalid_graph)
    Condition
      Error in `get_cmty_walktrap()`:
      ! The graph is not valid.
    Code
      get_common_nbrs(invalid_graph)
    Condition
      Error in `get_common_nbrs()`:
      ! The graph is not valid.
    Code
      get_coreness(invalid_graph)
    Condition
      Error in `get_coreness()`:
      ! The graph is not valid.
    Code
      get_degree_distribution(invalid_graph)
    Condition
      Error in `get_degree_distribution()`:
      ! The graph is not valid.
    Code
      get_degree_histogram(invalid_graph)
    Condition
      Error in `get_degree_histogram()`:
      ! The graph is not valid.
    Code
      get_degree_in(invalid_graph)
    Condition
      Error in `get_degree_in()`:
      ! The graph is not valid.
    Code
      get_degree_out(invalid_graph)
    Condition
      Error in `get_degree_out()`:
      ! The graph is not valid.
    Code
      get_degree_total(invalid_graph)
    Condition
      Error in `get_degree_total()`:
      ! The graph is not valid.
    Code
      get_dice_similarity(invalid_graph)
    Condition
      Error in `get_dice_similarity()`:
      ! The graph is not valid.
    Code
      get_eccentricity(invalid_graph)
    Condition
      Error in `get_eccentricity()`:
      ! The graph is not valid.
    Code
      get_edge_attrs(invalid_graph)
    Condition
      Error in `get_edge_attrs()`:
      ! The graph is not valid.
    Code
      get_edge_attrs_ws(invalid_graph)
    Condition
      Error in `get_edge_attrs_ws()`:
      ! The graph is not valid.
    Code
      get_edge_count_w_multiedge(invalid_graph)
    Condition
      Error in `get_edge_count_w_multiedge()`:
      ! The graph is not valid.
    Code
      get_edge_df(invalid_graph)
    Condition
      Error in `get_edge_df()`:
      ! The graph is not valid.
    Code
      get_edge_df_ws(invalid_graph)
    Condition
      Error in `get_edge_df_ws()`:
      ! The graph is not valid.
    Code
      get_edge_info(invalid_graph)
    Condition
      Error in `get_edge_info()`:
      ! The graph is not valid.
    Code
      get_eigen_centrality(invalid_graph)
    Condition
      Error in `get_eigen_centrality()`:
      ! The graph is not valid.
    Code
      get_girth(invalid_graph)
    Condition
      Error in `get_girth()`:
      ! The graph is not valid.
    Code
      get_global_graph_attr_info(invalid_graph)
    Condition
      Error in `get_global_graph_attr_info()`:
      ! The graph is not valid.
    Code
      get_graph_actions(invalid_graph)
    Condition
      Error in `get_graph_actions()`:
      ! The graph is not valid.
    Code
      get_graph_log(invalid_graph)
    Condition
      Error in `get_graph_log()`:
      ! The graph is not valid.
    Code
      get_graph_name(invalid_graph)
    Condition
      Error in `get_graph_name()`:
      ! The graph is not valid.
    Code
      get_graph_time(invalid_graph)
    Condition
      Error in `get_graph_time()`:
      ! The graph is not valid.
    Code
      get_jaccard_similarity(invalid_graph)
    Condition
      Error in `get_jaccard_similarity()`:
      ! The graph is not valid.
    Code
      get_last_edges_created(invalid_graph)
    Condition
      Error in `get_last_edges_created()`:
      ! The graph is not valid.
    Code
      get_last_nodes_created(invalid_graph)
    Condition
      Error in `get_last_nodes_created()`:
      ! The graph is not valid.
    Code
      get_leverage_centrality(invalid_graph)
    Condition
      Error in `get_leverage_centrality()`:
      ! The graph is not valid.
    Code
      get_max_eccentricity(invalid_graph)
    Condition
      Error in `get_max_eccentricity()`:
      ! The graph is not valid.
    Code
      get_mean_distance(invalid_graph)
    Condition
      Error in `get_mean_distance()`:
      ! The graph is not valid.
    Code
      get_min_cut_between(invalid_graph)
    Condition
      Error in `get_min_cut_between()`:
      ! The graph is not valid.
    Code
      get_min_eccentricity(invalid_graph)
    Condition
      Error in `get_min_eccentricity()`:
      ! The graph is not valid.
    Code
      get_multiedge_count(invalid_graph)
    Condition
      Error in `get_multiedge_count()`:
      ! The graph is not valid.
    Code
      get_nbrs(invalid_graph)
    Condition
      Error in `get_nbrs()`:
      ! The graph is not valid.
    Code
      get_node_attrs(invalid_graph)
    Condition
      Error in `get_node_attrs()`:
      ! The graph is not valid.
    Code
      get_node_attrs_ws(invalid_graph)
    Condition
      Error in `get_node_attrs_ws()`:
      ! The graph is not valid.
    Code
      get_node_df(invalid_graph)
    Condition
      Error in `get_node_df()`:
      ! The graph is not valid.
    Code
      get_node_df_ws(invalid_graph)
    Condition
      Error in `get_node_df_ws()`:
      ! The graph is not valid.
    Code
      get_node_ids(invalid_graph)
    Condition
      Error in `is_graph_empty()`:
      ! The graph is not valid.
    Code
      get_node_info(invalid_graph)
    Condition
      Error in `get_node_info()`:
      ! The graph is not valid.
    Code
      get_non_nbrs(invalid_graph)
    Condition
      Error in `get_non_nbrs()`:
      ! The graph is not valid.
    Code
      get_pagerank(invalid_graph)
    Condition
      Error in `get_pagerank()`:
      ! The graph is not valid.
    Code
      get_paths(invalid_graph)
    Condition
      Error in `get_paths()`:
      ! The graph is not valid.
    Code
      get_periphery(invalid_graph)
    Condition
      Error in `get_periphery()`:
      ! The graph is not valid.
    Code
      get_predecessors(invalid_graph)
    Condition
      Error in `get_predecessors()`:
      ! The graph is not valid.
    Code
      get_radiality(invalid_graph)
    Condition
      Error in `get_radiality()`:
      ! The graph is not valid.
    Code
      get_reciprocity(invalid_graph)
    Condition
      Error in `get_reciprocity()`:
      ! The graph is not valid.
    Code
      get_s_connected_cmpts(invalid_graph)
    Condition
      Error in `get_s_connected_cmpts()`:
      ! The graph is not valid.
    Code
      get_selection(invalid_graph)
    Condition
      Error in `get_selection()`:
      ! The graph is not valid.
    Code
      get_similar_nbrs(invalid_graph)
    Condition
      Error in `get_similar_nbrs()`:
      ! The graph is not valid.
    Code
      get_successors(invalid_graph)
    Condition
      Error in `get_successors()`:
      ! The graph is not valid.
    Code
      get_w_connected_cmpts(invalid_graph)
    Condition
      Error in `get_w_connected_cmpts()`:
      ! The graph is not valid.
    Code
      get_graph_info(invalid_graph)
    Condition
      Error in `get_graph_info()`:
      ! The graph is not valid.
    Code
      invert_selection(invalid_graph)
    Condition
      Error in `invert_selection()`:
      ! The graph is not valid.
    Code
      is_edge_loop(invalid_graph)
    Condition
      Error in `is_edge_loop()`:
      ! The graph is not valid.
    Code
      is_edge_multiple(invalid_graph)
    Condition
      Error in `is_edge_multiple()`:
      ! The graph is not valid.
    Code
      is_edge_mutual(invalid_graph)
    Condition
      Error in `is_edge_mutual()`:
      ! The graph is not valid.
    Code
      is_edge_present(invalid_graph)
    Condition
      Error in `is_edge_present()`:
      ! The graph is not valid.
    Code
      is_graph_connected(invalid_graph)
    Condition
      Error in `is_graph_connected()`:
      ! The graph is not valid.
    Code
      is_graph_dag(invalid_graph)
    Condition
      Error in `is_graph_dag()`:
      ! The graph is not valid.
    Code
      is_graph_directed(invalid_graph)
    Condition
      Error in `is_graph_directed()`:
      ! The graph is not valid.
    Code
      is_graph_empty(invalid_graph)
    Condition
      Error in `is_graph_empty()`:
      ! The graph is not valid.
    Code
      is_graph_simple(invalid_graph)
    Condition
      Error in `is_graph_simple()`:
      ! The graph is not valid.
    Code
      is_graph_undirected(invalid_graph)
    Condition
      Error in `is_graph_undirected()`:
      ! The graph is not valid.
    Code
      is_graph_weighted(invalid_graph)
    Condition
      Error in `is_graph_weighted()`:
      ! The graph is not valid.
    Code
      is_node_present(invalid_graph)
    Condition
      Error in `is_node_present()`:
      ! The graph is not valid.
    Code
      is_property_graph(invalid_graph)
    Condition
      Error in `is_property_graph()`:
      ! The graph is not valid.
    Code
      join_edge_attrs(invalid_graph)
    Condition
      Error in `join_edge_attrs()`:
      ! The graph is not valid.
    Code
      join_node_attrs(invalid_graph)
    Condition
      Error in `join_node_attrs()`:
      ! The graph is not valid.
    Code
      layout_nodes_w_string(invalid_graph)
    Condition
      Error in `layout_nodes_w_string()`:
      ! The graph is not valid.
    Code
      mutate_edge_attrs(invalid_graph)
    Condition
      Error in `mutate_edge_attrs()`:
      ! The graph is not valid.
    Code
      mutate_edge_attrs_ws(invalid_graph)
    Condition
      Error in `mutate_edge_attrs_ws()`:
      ! The graph is not valid.
    Code
      mutate_node_attrs(invalid_graph)
    Condition
      Error in `mutate_node_attrs()`:
      ! The graph is not valid.
    Code
      mutate_node_attrs_ws(invalid_graph)
    Condition
      Error in `mutate_node_attrs_ws()`:
      ! The graph is not valid.
    Code
      nudge_node_positions_ws(invalid_graph)
    Condition
      Error in `nudge_node_positions_ws()`:
      ! The graph is not valid.
    Code
      recode_edge_attrs(invalid_graph)
    Condition
      Error in `recode_edge_attrs()`:
      ! The graph is not valid.
    Code
      recode_node_attrs(invalid_graph)
    Condition
      Error in `recode_node_attrs()`:
      ! The graph is not valid.
    Code
      rename_edge_attrs(invalid_graph)
    Condition
      Error in `rename_edge_attrs()`:
      ! The graph is not valid.
    Code
      rename_node_attrs(invalid_graph)
    Condition
      Error in `rename_node_attrs()`:
      ! The graph is not valid.
    Code
      render_graph(invalid_graph)
    Condition
      Error in `render_graph()`:
      ! The graph is not valid.
    Code
      reorder_graph_actions(invalid_graph)
    Condition
      Error in `reorder_graph_actions()`:
      ! The graph is not valid.
    Code
      rescale_edge_attrs(invalid_graph)
    Condition
      Error in `rescale_edge_attrs()`:
      ! The graph is not valid.
    Code
      rescale_node_attrs(invalid_graph)
    Condition
      Error in `rescale_node_attrs()`:
      ! The graph is not valid.
    Code
      rev_edge_dir(invalid_graph)
    Condition
      Error in `rev_edge_dir()`:
      ! The graph is not valid.
    Code
      rev_edge_dir_ws(invalid_graph)
    Condition
      Error in `rev_edge_dir_ws()`:
      ! The graph is not valid.
    Code
      select_edges(invalid_graph)
    Condition
      Error in `select_edges()`:
      ! The graph is not valid.
    Code
      select_edges_by_edge_id(invalid_graph)
    Condition
      Error in `select_edges_by_edge_id()`:
      ! The graph is not valid.
    Code
      select_edges_by_node_id(invalid_graph)
    Condition
      Error in `select_edges_by_node_id()`:
      ! The graph is not valid.
    Code
      select_last_edges_created(invalid_graph)
    Condition
      Error in `select_last_edges_created()`:
      ! The graph is not valid.
    Code
      select_last_nodes_created(invalid_graph)
    Condition
      Error in `select_last_nodes_created()`:
      ! The graph is not valid.
    Code
      select_nodes(invalid_graph)
    Condition
      Error in `select_nodes()`:
      ! The graph is not valid.
    Code
      select_nodes_by_degree(invalid_graph)
    Condition
      Error in `select_nodes_by_degree()`:
      ! The graph is not valid.
    Code
      select_nodes_by_id(invalid_graph)
    Condition
      Error in `select_nodes_by_id()`:
      ! The graph is not valid.
    Code
      select_nodes_in_neighborhood(invalid_graph)
    Condition
      Error in `select_nodes_in_neighborhood()`:
      ! The graph is not valid.
    Code
      set_cache(invalid_graph)
    Condition
      Error in `set_cache()`:
      ! The graph is not valid.
    Code
      set_edge_attr_to_display(invalid_graph)
    Condition
      Error in `set_edge_attr_to_display()`:
      ! The graph is not valid.
    Code
      set_edge_attrs(invalid_graph)
    Condition
      Error in `set_edge_attrs()`:
      ! argument "values" is missing, with no default
    Code
      set_edge_attrs_ws(invalid_graph)
    Condition
      Error in `set_edge_attrs_ws()`:
      ! The graph is not valid.
    Code
      set_graph_directed(invalid_graph)
    Condition
      Error in `set_graph_directed()`:
      ! The graph is not valid.
    Code
      set_graph_name(invalid_graph)
    Condition
      Error in `set_graph_name()`:
      ! The graph is not valid.
    Code
      set_graph_time(invalid_graph)
    Condition
      Error in `set_graph_time()`:
      ! The graph is not valid.
    Code
      set_graph_undirected(invalid_graph)
    Condition
      Error in `set_graph_undirected()`:
      ! The graph is not valid.
    Code
      set_node_attr_to_display(invalid_graph)
    Condition
      Error in `set_node_attr_to_display()`:
      ! The graph is not valid.
    Code
      set_node_attr_w_fcn(invalid_graph)
    Condition
      Error in `set_node_attr_w_fcn()`:
      ! The graph is not valid.
    Code
      set_node_attrs(invalid_graph)
    Condition
      Error in `set_node_attrs()`:
      ! argument "values" is missing, with no default
    Code
      set_node_attrs_ws(invalid_graph)
    Condition
      Error in `set_node_attrs_ws()`:
      ! The graph is not valid.
    Code
      set_node_position(invalid_graph)
    Condition
      Error in `set_node_position()`:
      ! The graph is not valid.
    Code
      to_igraph(invalid_graph)
    Condition
      Error in `to_igraph()`:
      ! The graph is not valid.
    Code
      transform_to_complement_graph(invalid_graph)
    Condition
      Error in `transform_to_complement_graph()`:
      ! The graph is not valid.
    Code
      transform_to_min_spanning_tree(invalid_graph)
    Condition
      Error in `transform_to_min_spanning_tree()`:
      ! The graph is not valid.
    Code
      transform_to_subgraph_ws(invalid_graph)
    Condition
      Error in `transform_to_subgraph_ws()`:
      ! The graph is not valid.
    Code
      trav_both(invalid_graph)
    Condition
      Error in `trav_both()`:
      ! The graph is not valid.
    Code
      trav_both_edge(invalid_graph)
    Condition
      Error in `trav_both_edge()`:
      ! The graph is not valid.
    Code
      trav_in(invalid_graph)
    Condition
      Error in `trav_in()`:
      ! The graph is not valid.
    Code
      trav_in_edge(invalid_graph)
    Condition
      Error in `trav_in_edge()`:
      ! The graph is not valid.
    Code
      trav_in_node(invalid_graph)
    Condition
      Error in `trav_in_node()`:
      ! The graph is not valid.
    Code
      trav_in_until(invalid_graph)
    Condition
      Error in `trav_in_until()`:
      ! The graph is not valid.
    Code
      trav_reverse_edge(invalid_graph)
    Condition
      Error in `trav_reverse_edge()`:
      ! The graph is not valid.
    Code
      trav_out(invalid_graph)
    Condition
      Error in `trav_out()`:
      ! The graph is not valid.
    Code
      trav_out_edge(invalid_graph)
    Condition
      Error in `trav_out_edge()`:
      ! The graph is not valid.
    Code
      trav_out_node(invalid_graph)
    Condition
      Error in `trav_out_node()`:
      ! The graph is not valid.
    Code
      trav_out_until(invalid_graph)
    Condition
      Error in `trav_out_until()`:
      ! The graph is not valid.
    Code
      trav_reverse_edge(invalid_graph)
    Condition
      Error in `trav_reverse_edge()`:
      ! The graph is not valid.
    Code
      trigger_graph_actions(invalid_graph)
    Condition
      Error in `trigger_graph_actions()`:
      ! The graph is not valid.

# Using an empty graph with trigger an error

    Code
      add_edge(empty_graph)
    Condition
      Error in `add_edge()`:
      ! The graph does not contain nodes.
      * An edge cannot be added.
    Code
      add_edge_df(empty_graph)
    Condition
      Error in `add_edge_df()`:
      ! The graph does not contain nodes.
      * So, edges cannot be added.

# Using a graph with no edges with trigger an error

    Code
      add_edge_clone(graph_no_edges)
    Condition
      Error in `add_edge_clone()`:
      ! The graph contains no edges.
    Code
      add_forward_edges_ws(graph_no_edges)
    Condition
      Error in `add_forward_edges_ws()`:
      ! The graph contains no edges.
    Code
      add_reverse_edges_ws(graph_no_edges)
    Condition
      Error in `add_reverse_edges_ws()`:
      ! The graph contains no edges.
    Code
      delete_edge(graph_no_edges)
    Condition
      Error in `delete_edge()`:
      ! The graph contains no edges.
    Code
      delete_edges_ws(graph_no_edges)
    Condition
      Error in `delete_edges_ws()`:
      ! The graph contains no edges.
    Code
      get_edge_count_w_multiedge(graph_no_edges)
    Condition
      Error in `get_edge_count_w_multiedge()`:
      ! The graph contains no edges.
    Code
      get_last_edges_created(graph_no_edges)
    Condition
      Error in `get_last_edges_created()`:
      ! The graph contains no edges.
    Code
      get_multiedge_count(graph_no_edges)
    Condition
      Error in `get_multiedge_count()`:
      ! The graph contains no edges.
    Code
      is_edge_loop(graph_no_edges)
    Condition
      Error in `is_edge_loop()`:
      ! The graph contains no edges.
    Code
      is_edge_multiple(graph_no_edges)
    Condition
      Error in `is_edge_multiple()`:
      ! The graph contains no edges.
    Code
      is_edge_mutual(graph_no_edges)
    Condition
      Error in `is_edge_mutual()`:
      ! The graph contains no edges.
    Code
      mutate_edge_attrs_ws(graph_no_edges)
    Condition
      Error in `mutate_edge_attrs_ws()`:
      ! The graph contains no edges.
    Code
      recode_edge_attrs(graph_no_edges)
    Condition
      Error in `recode_edge_attrs()`:
      ! The graph contains no edges.
    Code
      rename_edge_attrs(graph_no_edges)
    Condition
      Error in `rename_edge_attrs()`:
      ! The graph contains no edges.
    Code
      rescale_edge_attrs(graph_no_edges)
    Condition
      Error in `rescale_edge_attrs()`:
      ! The graph contains no edges.
    Code
      rev_edge_dir(graph_no_edges)
    Condition
      Error in `rev_edge_dir()`:
      ! The graph contains no edges.
    Code
      rev_edge_dir_ws(graph_no_edges)
    Condition
      Error in `rev_edge_dir_ws()`:
      ! The graph contains no edges.
    Code
      select_edges(graph_no_edges)
    Condition
      Error in `select_edges()`:
      ! The graph contains no edges.
    Code
      select_edges_by_edge_id(graph_no_edges)
    Condition
      Error in `select_edges_by_edge_id()`:
      ! The graph contains no edges.
    Code
      select_edges_by_node_id(graph_no_edges)
    Condition
      Error in `select_edges_by_node_id()`:
      ! The graph contains no edges.
    Code
      select_last_edges_created(graph_no_edges)
    Condition
      Error in `select_last_edges_created()`:
      ! The graph contains no edges.
    Code
      set_edge_attr_to_display(graph_no_edges)
    Condition
      Error in `set_edge_attr_to_display()`:
      ! The graph contains no edges.
    Code
      set_edge_attrs_ws(graph_no_edges)
    Condition
      Error in `set_edge_attrs_ws()`:
      ! The graph contains no edges.
    Code
      trav_both(graph_no_edges)
    Condition
      Error in `trav_both()`:
      ! The graph contains no edges.
    Code
      trav_both_edge(graph_no_edges)
    Condition
      Error in `trav_both_edge()`:
      ! The graph contains no edges.
    Code
      trav_in(graph_no_edges)
    Condition
      Error in `trav_in()`:
      ! The graph contains no edges.
    Code
      trav_in_edge(graph_no_edges)
    Condition
      Error in `trav_in_edge()`:
      ! The graph contains no edges.
    Code
      trav_in_node(graph_no_edges)
    Condition
      Error in `trav_in_node()`:
      ! The graph contains no edges.
    Code
      trav_in_until(graph_no_edges)
    Condition
      Error in `trav_in_until()`:
      ! The graph contains no edges.
    Code
      trav_out(graph_no_edges)
    Condition
      Error in `trav_out()`:
      ! The graph contains no edges.
    Code
      trav_out_edge(graph_no_edges)
    Condition
      Error in `trav_out_edge()`:
      ! The graph contains no edges.
    Code
      trav_out_node(graph_no_edges)
    Condition
      Error in `trav_out_node()`:
      ! The graph contains no edges.
    Code
      trav_out_until(graph_no_edges)
    Condition
      Error in `trav_out_until()`:
      ! The graph contains no edges.
    Code
      trav_reverse_edge(graph_no_edges)
    Condition
      Error in `trav_reverse_edge()`:
      ! The graph contains no edges.

# selective traversals with `trav_in_until()` are possible

    Code
      create_graph() %>% add_path(n = 10, node_data = node_data(value = 1:10)) %>%
        trav_in_until(conditions = value == 1)
    Condition
      Error in `trav_in_until()`:
      ! There is no selection of nodes available.
      * Any traversal requires an active selection.
      * This type of traversal requires a selection of nodes.

# a correct node data frame is generated

    Code
      create_node_df(n = "a")
    Condition
      Error in `create_node_df()`:
      ! `n` must be a whole number, not the string "a".
    Code
      create_node_df(n = c(1, 2))
    Condition
      Error in `create_node_df()`:
      ! `n` must be a whole number, not a double vector.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trav_reverse_edge.R
\name{trav_reverse_edge}
\alias{trav_reverse_edge}
\title{Traverse to any reverse edges}
\usage{
trav_reverse_edge(graph, add_to_selection = FALSE)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{add_to_selection}{An option to either add the reverse edges to the
active selection of edges (\code{TRUE}) or switch the active selection entirely
to those reverse edges (\code{FALSE}, the default case).}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From an active selection of edges in a graph object of class \code{dgr_graph},
traverse to any available reverse edges between the nodes common to the
selected edges. For instance, if an active selection has the edge \code{1->2} but
there is also an (not selected) edge \code{2->1}, then this function can either
switch to the selection of \code{2->1}, or, incorporate both those edges into the
active selection of edges.

This traversal function makes use of an active selection of edges. After the
traversal, depending on the traversal conditions, there will either be a
selection of edges or no selection at all.

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = "basic",
    label = TRUE)

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 4, 2, 3, 3),
    to =   c(4, 1, 3, 2, 1))

# Create a graph with the
# ndf and edf
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Explicitly select the edges
# `1`->`4` and `2`->`3`
graph <-
  graph \%>\%
  select_edges(
    from = 1,
      to = 4) \%>\%
  select_edges(
    from = 2,
      to = 3)

# Get the inital edge selection
graph \%>\% get_selection()

# Traverse to the reverse edges
# (edges `2`: `4`->`1` and
# `4`:`3`->`2`)
graph <-
  graph \%>\%
  trav_reverse_edge()

# Get the current selection of edges
graph \%>\% get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{currencies}
\alias{currencies}
\title{ISO-4217 currency data.}
\format{
A data frame with 171 rows and 4 variables:
\describe{
\item{iso_4217_code}{the three-letter currency code according to the ISO-4217 standard}
\item{curr_number}{the three-digit code number assigned to each currency under the ISO-4217 standard}
\item{exponent}{the base 10 exponent of the minor currency unit in relation to the major currency unit (it can be assumed also to be number of decimal places that is commonly considered for the currency)}
\item{currency_name}{the English name of the currency}
}
}
\source{
\url{https://en.wikipedia.org/wiki/ISO_4217}
}
\usage{
currencies
}
\description{
A dataset containing currency information from the ISO-4217 standard.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_graph_dag.R
\name{is_graph_dag}
\alias{is_graph_dag}
\title{Is the graph a directed acyclic graph?}
\usage{
is_graph_dag(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Provides a logical value on whether the graph is a directed acyclic graph
(DAG). The conditions for a graph that is a DAG are that it should be a
directed graph and it should not contain any cycles.
}
\examples{
# Create a directed graph containing
# only a balanced tree
graph_tree <-
  create_graph() \%>\%
  add_balanced_tree(
    k = 2, h = 3)

# Determine whether this graph
# is a DAG
graph_tree \%>\%
  is_graph_dag()

# Create a directed graph containing
# a single cycle
graph_cycle <-
  create_graph() \%>\%
  add_cycle(n = 5)

# Determine whether this graph
# is a DAG
graph_cycle \%>\%
  is_graph_dag()

# Create an undirected graph
# containing a balanced tree
graph_tree_undirected <-
  create_graph(
    directed = FALSE) \%>\%
  add_balanced_tree(
    k = 2, h = 2)

# Determine whether this graph
# is a DAG
graph_tree_undirected \%>\%
  is_graph_dag()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_min_eccentricity.R
\name{get_min_eccentricity}
\alias{get_min_eccentricity}
\title{Get the minimum graph eccentricity}
\usage{
get_min_eccentricity(graph, direction = "all")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{direction}{Using \code{all} (the default), the search will ignore edge
direction while traversing through the graph. With \code{out}, measurements of
paths will be from a node whereas with \verb{in}, measurements of paths will be
to a node.}
}
\value{
A single numeric value representing the minimum eccentricity of the
graph.
}
\description{
Get the radius of a graph, which is the smallest eccentricity in the graph.
The graph eccentricity of a node is its shortest path from the farthest other
node in the graph.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() \%>\%
  add_cycle(n = 5)

# Determine the graph's minimum
# eccentricity
graph \%>\%
  get_min_eccentricity()

# Create a full graph and then
# get the minimum eccentricity
# value for that
create_graph() \%>\%
  add_full_graph(n = 10) \%>\%
  get_min_eccentricity()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_gnp_graph.R
\name{add_gnp_graph}
\alias{add_gnp_graph}
\title{Add a G(n, p) Erdos-Renyi graph}
\usage{
add_gnp_graph(
  graph,
  n,
  p,
  loops = FALSE,
  type = NULL,
  label = TRUE,
  rel = NULL,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL,
  set_seed = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{n}{The number of nodes comprising the generated graph.}

\item{p}{The probability of creating an edge between two arbitrary nodes.}

\item{loops}{A logical value (default is \code{FALSE}) that governs whether loops
are allowed to be created.}

\item{type}{An optional string that describes the entity type for all the
nodes to be added.}

\item{label}{A boolean value where setting to \code{TRUE} ascribes node IDs to the
label and \code{FALSE} yields a blank label.}

\item{rel}{An optional string for providing a relationship label to all edges
to be added.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}

\item{set_seed}{Supplying a value sets a random seed of the
\code{Mersenne-Twister} implementation.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
To an existing graph object, add a graph built according to the Erdos-Renyi
G(n, p) model, which uses a constant probability when creating edges.
}
\examples{
# Create an undirected GNP
# graph with 100 nodes using
# a probability value of 0.05
gnp_graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnp_graph(
    n = 100,
    p = 0.05)

# Get a count of nodes
gnp_graph \%>\% count_nodes()

# Get a count of edges
gnp_graph \%>\% count_edges()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate_dot.R
\name{generate_dot}
\alias{generate_dot}
\title{Generate DOT code using a graph object}
\usage{
generate_dot(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A character vector of length 1 containing Graphviz DOT code.
}
\description{
Generates Graphviz DOT code as an R character object using DiagrammeR graph
object.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rev_edge_dir.R
\name{rev_edge_dir}
\alias{rev_edge_dir}
\title{Reverse the direction of all edges in a graph}
\usage{
rev_edge_dir(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Using a directed graph as input, reverse the direction of all edges in that
graph.
}
\examples{
# Create a graph with a
# directed tree
graph <-
  create_graph() \%>\%
  add_balanced_tree(
    k = 2, h = 2)

# Inspect the graph's edges
graph \%>\% get_edges()

# Reverse the edge directions
# such that edges are directed
# toward the root of the tree
graph <-
  graph \%>\%
  rev_edge_dir()

# Inspect the graph's edges
# after their reversal
graph \%>\% get_edges()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils-pipe.R
\name{\%>\%}
\alias{\%>\%}
\title{Pipe operator}
\usage{
lhs \%>\% rhs
}
\arguments{
\item{lhs}{A value or the magrittr placeholder.}

\item{rhs}{A function call using the magrittr semantics.}
}
\value{
The result of calling \code{rhs(lhs)}.
}
\description{
See \code{magrittr::\link[magrittr:pipe]{\%>\%}} for details.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export_csv.R
\name{export_csv}
\alias{export_csv}
\title{Export a graph to CSV files}
\usage{
export_csv(
  graph,
  ndf_name = "nodes.csv",
  edf_name = "edges.csv",
  output_path = getwd(),
  colnames_type = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{ndf_name}{The name to provide to the CSV file containing node
information. By default this CSV will be called \code{nodes.csv}.}

\item{edf_name}{The name to provide to the CSV file containing edge
information. By default this CSV will be called \code{edges.csv}.}

\item{output_path}{The path to which the CSV files will be placed. By
default, this is the current working directory.}

\item{colnames_type}{Provides options to modify CSV column names to allow for
easier import into other graph systems. The \code{neo4j} option modifies column
names to allow for direct import of CSVs into Neo4J with the \verb{LOAD CSV}
clause. The \code{graphframes} option modifies column names to match those
required by the Spark GraphFrames package.}
}
\description{
Export a graph to separate CSV files for nodes and edges.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = c("a", "a", "z", "z"),
    label = TRUE,
    value = c(3.5, 2.6, 9.4, 2.7)
  )

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = c("rel_a", "rel_z", "rel_a")
  )

# Create a graph with the ndf and edf
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf
  )

# Create separate `nodes.csv` and
# `edges.csv` files
# graph \%>\% export_csv()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edge_attrs_ws.R
\name{get_edge_attrs_ws}
\alias{get_edge_attrs_ws}
\title{Get edge attribute values from a selection of edges}
\usage{
get_edge_attrs_ws(graph, edge_attr)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr}{the name of the attribute for which to get values.}
}
\value{
A named vector of edge attribute values for the attribute given by
\code{edge_attr} by edge.
}
\description{
From a graph object of class \code{dgr_graph}, get edge attribute values for one
or more edges.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a simple graph where
# edges have an edge attribute
# named `value`
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 4) \%>\%
  {
    edges <-
      create_edge_df(
        from = c(1, 2, 1, 4),
          to = c(2, 3, 4, 3),
         rel = "rel")
    add_edge_df(
      graph = .,
      edge_df = edges)
  } \%>\%
  set_edge_attrs(
    edge_attr = value,
    values = 1.6,
    from = 1,
      to = 2) \%>\%
  set_edge_attrs(
    edge_attr = value,
    values = 4.3,
    from = 1,
      to = 4) \%>\%
  set_edge_attrs(
    edge_attr = value,
    values = 2.9,
    from = 2,
      to = 3) \%>\%
  set_edge_attrs(
    edge_attr = value,
    values = 8.4,
    from = 4,
      to = 3)

# Select the edges defined as
# `1`->`3` and `2`->`3`
graph <-
  graph \%>\%
  select_edges(
    from = c(1, 2),
    to = c(2, 3))

# Get the edge attribute values
# for the `value` attribute, limited
# to the current edge selection
graph \%>\%
  get_edge_attrs_ws(
    edge_attr = value)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_aes.R
\name{edge_aes}
\alias{edge_aes}
\title{Insert edge aesthetic attributes during edge creation}
\usage{
edge_aes(
  style = NULL,
  penwidth = NULL,
  color = NULL,
  arrowsize = NULL,
  arrowhead = NULL,
  arrowtail = NULL,
  fontname = NULL,
  fontsize = NULL,
  fontcolor = NULL,
  len = NULL,
  tooltip = NULL,
  URL = NULL,
  label = NULL,
  labelfontname = NULL,
  labelfontsize = NULL,
  labelfontcolor = NULL,
  labeltooltip = NULL,
  labelURL = NULL,
  edgetooltip = NULL,
  edgeURL = NULL,
  dir = NULL,
  headtooltip = NULL,
  headURL = NULL,
  headclip = NULL,
  headlabel = NULL,
  headport = NULL,
  tailtooltip = NULL,
  tailURL = NULL,
  tailclip = NULL,
  taillabel = NULL,
  tailport = NULL,
  decorate = NULL
)
}
\arguments{
\item{style}{The edge line style. The \code{style} types that can be used are
\code{solid}, \code{bold}, \code{dashed}, \code{dotted}, \code{tapered}, and \code{invisible}.}

\item{penwidth}{The thickness of the stroke line for the edge itself.}

\item{color}{The color of the edge. Can be an X11 color or a hexadecimal
color code.}

\item{arrowsize}{A scaling factor for arrowheads. The default value is \code{1.0}
and the minimum is \code{0}.}

\item{arrowhead}{The type of arrowhead to use. The \code{style} attribute can be
any of these types: \code{normal}, \code{vee}, \code{tee}, \code{dot}, \code{diamond}, \code{box},
\code{curve}, \code{icurve}, \code{inv}, \code{crow}, or \code{none}.}

\item{arrowtail}{The type of arrowtail to use. The \code{style} attribute can any
of these types: \code{normal}, \code{vee}, \code{tee}, \code{dot}, \code{diamond}, \code{box}, \code{curve},
\code{icurve}, \code{inv}, \code{crow}, or \code{none}.}

\item{fontname}{The name of the system font that will be used for any edge
text.}

\item{fontsize}{The point size of the font used for any edge text.}

\item{fontcolor}{The color used for any edge text. Can be an X11 color or a
hexadecimal color code.}

\item{len}{The preferred edge length for an edge, in inches. Default value is
\code{1.0}.}

\item{tooltip}{Text for a tooltip that appears when hovering over an edge. If
text is not provided, then the default tooltip text will provide the edge
definition (i.e., \verb{[id]->[id] or [id]--[id]}).}

\item{URL}{A URL to associate with an edge. Upon rendering the plot, clicking
edges with any associated URLs will open the URL in the default browser.}

\item{label}{The label text associated with the edge. This text will appear
near the center of the edge.}

\item{labelfontname}{The name of the system font that will be used for the
\code{headlabel} and the \code{taillabel} label text. If not set, the \code{fontname}
value will instead be used.}

\item{labelfontsize}{The point size of the font used for the \code{headlabel} and
the \code{taillabel} label text. If not set, the \code{fontsize} value will instead
be used.}

\item{labelfontcolor}{The color used for the label text of the \code{headlabel}
and the \code{taillabel} label text. If not set, the \code{fontcolor} value will
instead be used. Can be an X11 color or a hexadecimal color code.}

\item{labeltooltip}{Text for a tooltip that will appear when hovering over
the main label of an edge (if label text provided in the \code{label} edge
attribute). If text is not provided and an edge label is visible, then the
default tooltip text will provide the edge definition (i.e., \verb{[id]->[id] or [id]--[id]}).}

\item{labelURL}{A URL to associate with edge label text. Upon rendering the
plot, clicking edge labels with any associated URLs will open the URL in
the default browser.}

\item{edgetooltip}{This option provides a means to specify a tooltip with
only the non-label parts of an edge. If this is defined, the value
overrides any \code{tooltip} defined for the edge. This tooltip text is when
hovering along the edge (even near the head or tail node) unless overridden
by a \code{headtooltip} or \code{tailtooltip} value.}

\item{edgeURL}{This option provides a means to specify a URL with only the
non-label parts of an edge. If this is defined, the value overrides any
\code{URL} defined for the edge. This URL is used along the edge (even near the
head or tail node) unless overridden by a \code{headURL} or \code{tailURL} value.}

\item{dir}{An optional direction type. Normally, for directed graphs, this is
\code{forward} and needn't be set. For undirected graphs, this would be \code{none}
and again no explicit setting is required. However, one can also use the
\code{back} or \code{both} options. The \code{back} option draws an arrowhead in the
reverse direction of an edge. The \code{both} option draws two arrowheads. When
using any of these options in such an explicit manner, the \code{head...} and
\code{tail...} edge attributes allow control over aesthetic edge attributes in
either side of the edge.}

\item{headtooltip}{This option provides a means to specify a tooltip that can
be displayed by hovering over the part of an edge that is adjacent to
incoming node (see the \code{tooltip} argument for further details).}

\item{headURL}{This option provides a means to specify a URL that can be
accessed by clicking the part of an edge that is adjacent to incoming node
(see the \code{URL} argument for further details).}

\item{headclip}{If \code{TRUE} (the default behavior), then the head of the
affected edge is clipped to the node boundary. Using \code{FALSE} places the
head of the outgoing edge at the center of its node.}

\item{headlabel}{This option provides a means to display a label near the
part of an edge that is adjacent to incoming node (see the \code{label} argument
for further details).}

\item{headport}{Allows one to specify which compass position on the incoming
node the head of the edge will alight. Options are \code{n}, \code{ne}, \code{e}, \code{se},
\code{s}, \code{sw}, \code{w}, and \code{nw}.}

\item{tailtooltip}{This option provides a means to specify a tooltip that can
be displayed by hovering over the part of an edge that is adjacent to
outgoing node (see the \code{tooltip} argument for further details).}

\item{tailURL}{This option provides a means to specify a URL that can be
accessed by clicking the part of an edge that is adjacent to outgoing node
(see the \code{URL} argument for further details).}

\item{tailclip}{If \code{TRUE} (the default behavior), then the tail of the
affected edge is clipped to the node boundary. Using \code{FALSE} places the
tail of the outgoing edge at the center of its node.}

\item{taillabel}{This option provides a means to display a label near the
part of an edge that is adjacent to outgoing node (see the \code{label} argument
for further details).}

\item{tailport}{Allows one to specify which compass position on the outgoing
node the tail of the edge will be emitted from. Options are \code{n}, \code{ne}, \code{e},
\code{se}, \code{s}, \code{sw}, \code{w}, and \code{nw}.}

\item{decorate}{If \code{TRUE} then attach any edge label to the edge line via a
2-segment polyline, underlining the label text and partially overlapping
the edge line.}
}
\description{
This helper function should be invoked to provide values for the namesake
\code{edge_aes} argument, which is present in any function where edges are
created.
}
\examples{
# Create a new graph and add
# a path with several edge
# aesthetic attributes
graph <-
  create_graph() \%>\%
  add_path(
    n = 3,
    type = "path",
    edge_aes = edge_aes(
      style = "dot",
      color = c("red", "blue")))

# View the graph's internal
# node data frame; the node
# aesthetic attributes have
# been inserted
graph \%>\%
  get_edge_df()

}
\seealso{
Other aesthetics: 
\code{\link{node_aes}()},
\code{\link{node_edge_aes_data}}
}
\concept{aesthetics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_nodes_from_table.R
\name{add_nodes_from_table}
\alias{add_nodes_from_table}
\title{Add nodes and attributes to graph from a table}
\usage{
add_nodes_from_table(
  graph,
  table,
  label_col = NULL,
  type_col = NULL,
  set_type = NULL,
  drop_cols = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{table}{Either a path to a CSV file, or, a data frame object.}

\item{label_col}{An option to apply a column of data in the table as \code{label}
attribute values.}

\item{type_col}{An option to apply a column of data in the table as \code{type}
attribute values.}

\item{set_type}{An optional string to apply a \code{type} attribute to all nodes
created from data in the external table.}

\item{drop_cols}{An optional column selection statement for dropping columns
from the external table before inclusion as attributes in the graph's
internal node data frame. Several columns can be dropped by name using the
syntax \code{col_1 & col_2 & ...}. Columns can also be dropped using a numeric
column range with \code{:} (e.g., \code{5:8}), or, by using the \code{:} between column
names to specify the range (e.g., \code{col_5_name:col_8_name}).}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add nodes and their attributes to an existing graph object from data in a CSV
file or a data frame.
}
\examples{
# To add nodes from the dataset called
# `currencies` (available as a dataset
# in the package), call the
# `add_nodes_from_table()` function
# after creating an empty graph; new
# node ID values will be created as
# monotonically-increasing values
graph_1 <-
  create_graph() |>
  add_nodes_from_table(
    table = currencies)

# View part of the graph's internal
# node data frame (ndf)
ndf_1 <- graph_1 |> get_node_df()

ndf_1[, 1:5] |> head()

# If you would like to assign
# any of the table's columns as
# `type` or `label` attributes,
# this can be done with the `type_col`
# and `label_col` arguments; to set
# a static `type` attribute for all
# of the table records, use `set_type`
graph_2 <-
  create_graph() |>
  add_nodes_from_table(
    table = currencies,
    label_col = iso_4217_code,
    set_type = currency)

# View part of the graph's internal ndf
ndf_2 <- graph_2 |> get_node_df()

ndf_2[, 1:5] |> head()

# Suppose we would like to not
# include certain columns from the
# external table in the resulting
# graph; we can use the `drop_cols`
# argument to choose which columns
# to not include as attributes
graph_3 <-
  create_graph() |>
  add_nodes_from_table(
    table = currencies,
    label_col = iso_4217_code,
    set_type = currency,
    drop_cols = exponent & currency_name)

# Show the node attribute names
# for the graph; note that the
# `exponent` and `currency_name`
# columns are not attributes in the
# graph's internal node data frame
graph_3 |>
  get_node_df() |>
  colnames()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trav_both_edge.R
\name{trav_both_edge}
\alias{trav_both_edge}
\title{Traverse from one or more selected nodes onto adjacent edges}
\usage{
trav_both_edge(
  graph,
  conditions = NULL,
  copy_attrs_from = NULL,
  copy_attrs_as = NULL,
  agg = "sum"
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the traversal.}

\item{copy_attrs_from}{Providing a node attribute name will copy those node
attribute values to the traversed edges. If the edge attribute already
exists, the values will be merged to the traversed edges; otherwise, a new
edge attribute will be created.}

\item{copy_attrs_as}{If a node attribute name is provided in
\code{copy_attrs_from}, this option will allow the copied attribute values
to be written under a different edge attribute name. If the attribute name
provided in \code{copy_attrs_as} does not exist in the graph's edf, the new
edge attribute will be created with the chosen name.}

\item{agg}{If a node attribute is provided to \code{copy_attrs_from}, then an
aggregation function is required since there may be cases where multiple
node attribute values will be passed onto the traversed edge(s). To pass
only a single value, the following aggregation functions can be used:
\code{sum}, \code{min}, \code{max}, \code{mean}, or \code{median}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} move to adjacent edges from a
selection of one or more selected nodes, thereby creating a selection of
edges. An optional filter by edge attribute can limit the set of edges
traversed to.

This traversal function makes use of an active selection of nodes. After the
traversal, depending on the traversal conditions, there will either be a
selection of edges or no selection at all.

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a simple graph
graph <-
  create_graph() \%>\%
  add_n_nodes(
    n = 2,
    type = "a",
    label = c("asd", "iekd")) \%>\%
  add_n_nodes(
    n = 3,
    type = "b",
    label = c("idj", "edl", "ohd")) \%>\%
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->5",
    rel = c(NA, "A", "B", "C", "D"))

# Create a data frame with node ID values
# representing the graph edges (with `from`
# and `to` columns), and, a set of numeric values
df <-
  data.frame(
    from = c(1, 1, 2, 2, 3),
    to = c(2, 3, 4, 5, 5),
    values = round(rnorm(5, 5), 2))

# Join the data frame to the graph's internal
# edge data frame (edf)
graph <-
  graph \%>\%
  join_edge_attrs(df = df)

# Show the graph's internal edge data frame
graph \%>\% get_edge_df()

# Perform a simple traversal from nodes to
# adjacent edges with no conditions on the
# nodes traversed to
graph \%>\%
  select_nodes_by_id(nodes = 3) \%>\%
  trav_both_edge() \%>\%
  get_selection()

# Traverse from node `2` to any adjacent
# edges, filtering to those edges that have
# NA values for the `rel` edge attribute
graph \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  trav_both_edge(
    conditions = is.na(rel)) \%>\%
  get_selection()

# Traverse from node `2` to any adjacent
# edges, filtering to those edges that have
# numeric values greater than `6.5` for
# the `rel` edge attribute
graph \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  trav_both_edge(
    conditions = values > 6.5) \%>\%
  get_selection()

# Traverse from node `5` to any adjacent
# edges, filtering to those edges that
# have values equal to `C` for the `rel`
# edge attribute
graph \%>\%
  select_nodes_by_id(nodes = 5) \%>\%
  trav_both_edge(
    conditions = rel == "C") \%>\%
  get_selection()

# Traverse from node `2` to any adjacent
# edges, filtering to those edges that
# have values in the set `B` and `C` for
# the `rel` edge attribute
graph \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  trav_both_edge(
    conditions = rel \%in\% c("B", "C")) \%>\%
  get_selection()

# Traverse from node `2` to any adjacent
# edges, and use multiple conditions for the
# traversal
graph \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  trav_both_edge(
    conditions =
      rel \%in\% c("B", "C") &
      values > 4.0) \%>\%
  get_selection()

# Traverse from node `2` to any adjacent
# edges, and use multiple conditions with
# a single-length vector
graph \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  trav_both_edge(
    conditions =
      rel \%in\% c("B", "C") |
      values > 4.0) \%>\%
  get_selection()

# Traverse from node `2` to any adjacent
# edges, and use a regular expression as
# a filtering condition
graph \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  trav_both_edge(
    conditions = grepl("B|C", rel)) \%>\%
  get_selection()

# Create another simple graph to demonstrate
# copying of node attribute values to traversed
# edges
graph <-
  create_graph() \%>\%
  add_path(n = 4) \%>\%
  select_nodes_by_id(nodes = 2:3) \%>\%
  set_node_attrs_ws(
    node_attr = value,
    value = 5)

# Show the graph's internal edge data frame
graph \%>\%get_edge_df()

# Show the graph's internal node data frame
graph \%>\% get_node_df()

# Perform a traversal from the nodes to
# the adjacent edges while also applying
# the node attribute `value` to the edges (in
# this case summing the `value` of 5 from
# all contributing nodes adding as an edge
# attribute)
graph <-
  graph \%>\%
  trav_both_edge(
    copy_attrs_from = value,
    agg = "sum")

# Show the graph's internal edge data frame
# after this change
graph \%>\% get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_closeness_vitality.R
\name{get_closeness_vitality}
\alias{get_closeness_vitality}
\title{Get closeness vitality}
\usage{
get_closeness_vitality(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame with closeness vitality values for each of the nodes.
}
\description{
Get the closeness vitality values for all nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 10,
    m = 12,
    set_seed = 23)

# Get closeness vitality values
# for all nodes in the graph
graph \%>\% get_closeness_vitality()

# Add the closeness vitality
# values to the graph as a
# node attribute
graph <-
  graph \%>\%
  join_node_attrs(
    df = get_closeness_vitality(.))

# Display the graph's
# node data frame
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fully_connect_nodes_ws.R
\name{fully_connect_nodes_ws}
\alias{fully_connect_nodes_ws}
\title{Fully connect all nodes in a selection of nodes}
\usage{
fully_connect_nodes_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a selection of nodes in a graph, add any remaining edges required to
fully connect this group of edges to each other.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create an empty graph and
# then add a path of 3 nodes
# and two isolated nodes
graph <-
  create_graph() |>
  add_path(n = 3) |>
  add_n_nodes(n = 2)

# Select a node in the path
# of nodes (node `3`) and
# the two isolated nodes (`4`
# and `5`); then, and fully
# connect these nodes together
graph <-
  graph |>
  select_nodes_by_id(
    nodes = 3:5) |>
  fully_connect_nodes_ws()

# Get the graph's edge data frame
graph |> get_edge_df()

# Create an undirected, empty
# graph; add a path of 3 nodes
# and two isolated nodes
graph <-
  create_graph(
    directed = FALSE) |>
  add_path(n = 3) |>
  add_n_nodes(n = 2)

# Select a node in the path
# of nodes (node `3`) and
# the two isolated nodes (`4`
# and `5`); then, and fully
# connect these nodes together
graph <-
  graph |>
  select_nodes_by_id(
    nodes = 3:5) |>
  fully_connect_nodes_ws()

# Get the graph's edge data
# frame; in the undirected
# case, reverse edges aren't
# added
graph |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_property_graph.R
\name{is_property_graph}
\alias{is_property_graph}
\title{Is the graph a property graph?}
\usage{
is_property_graph(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Provides a logical value on whether the graph is property graph (i.e., all
nodes have an assigned \code{type} value and all edges have an assigned \code{rel}
value).
}
\examples{
# Create a graph with 2 nodes
# (with `type` values) and a
# single edge (with a `rel`)
simple_property_graph <-
  create_graph() |>
  add_node(
    type = "a",
    label = "first") |>
  add_node(
    type = "b",
    label = "second") |>
  add_edge(
    from = "first",
    to = "second",
    rel = "rel_1")

# This is indeed a property graph
# but to confirm this, use the
# `is_property_graph()` function
is_property_graph(simple_property_graph)

# If a `type` attribute is
# removed, then this graph will
# no longer be a property graph
simple_property_graph |>
  set_node_attrs(
    node_attr = type,
    values = NA,
    nodes = 1) |>
  is_property_graph()

# An empty graph will return FALSE
create_graph() |>
  is_property_graph()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trav_in_edge.R
\name{trav_in_edge}
\alias{trav_in_edge}
\title{Traverse from one or more selected nodes onto adjacent, inward edges}
\usage{
trav_in_edge(
  graph,
  conditions = NULL,
  copy_attrs_from = NULL,
  copy_attrs_as = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the traversal.}

\item{copy_attrs_from}{Providing a node attribute name will copy those node
attribute values to the traversed edges. If the edge attribute already
exists, the values will be merged to the traversed edges; otherwise, a new
edge attribute will be created.}

\item{copy_attrs_as}{If a node attribute name is provided in
\code{copy_attrs_from}, this option will allow the copied attribute values to be
written under a different edge attribute name. If the attribute name
provided in \code{copy_attrs_as} does not exist in the graph's edf, the new edge
attribute will be created with the chosen name.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} move to incoming edges from a
selection of one or more selected nodes, thereby creating a selection of
edges. An optional filter by edge attribute can limit the set of edges
traversed to.

This traversal function makes use of an active selection of nodes. After the
traversal, depending on the traversal conditions, there will either be a
selection of edges or no selection at all.

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a simple graph
graph <-
  create_graph() |>
  add_n_nodes(
    n = 2,
    type = "a",
    label = c("asd", "iekd")) |>
  add_n_nodes(
    n = 3,
    type = "b",
    label = c("idj", "edl", "ohd")) |>
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->5",
    rel = c(NA, "A", "B", "C", "D"))

# Create a data frame with node ID
# values representing the graph edges
# (with `from` and `to` columns), and,
# a set of numeric values
df <-
  data.frame(
    from = c(1, 1, 2, 2, 3),
    to = c(2, 3, 4, 5, 5),
    values = round(rnorm(5, 5), 2))

# Join the data frame to the graph's
# internal edge data frame (edf)
graph <-
  graph |>
  join_edge_attrs(df = df)

# Show the graph's internal edge data frame
graph |> get_edge_df()

# Perform a simple traversal from
# nodes to inbound edges with no
# conditions on the nodes
# traversed to
graph |>
  select_nodes_by_id(nodes = 2) |>
  trav_in_edge() |>
  get_selection()

# Traverse from node `2` to any
# inbound edges, filtering to
# those edges that have NA values
# for the `rel` edge attribute
graph |>
  select_nodes_by_id(nodes = 2) |>
  trav_in_edge(
    conditions = is.na(rel)) |>
  get_selection()

# Traverse from node `2` to any
# inbound edges, filtering to those
# edges that do not have NA values
# for the `rel` edge attribute
# (since there are no allowed
# traversals, the selection of node
# `2` is retained)
graph |>
  select_nodes_by_id(nodes = 2) |>
  trav_in_edge(
    conditions = !is.na(rel)) |>
  get_selection()

# Traverse from node `5` to any
# inbound edges, filtering to those
# edges that have numeric values
# greater than `5.5` for the `rel`
# edge attribute
graph |>
  select_nodes_by_id(nodes = 5) |>
  trav_in_edge(
    conditions = values > 5.5) |>
  get_selection()

# Traverse from node `5` to any
# inbound edges, filtering to those
# edges that have values equal to
# `D` for the `rel` edge attribute
graph |>
  select_nodes_by_id(nodes = 5) |>
  trav_in_edge(
    conditions = rel == "D") |>
  get_selection()

# Traverse from node `5` to any
# inbound edges, filtering to those
# edges that have values in the set
# `C` and `D` for the `rel` edge
# attribute
graph |>
  select_nodes_by_id(nodes = 5) |>
  trav_in_edge(
    conditions = rel \%in\% c("C", "D")) |>
  get_selection()

# Traverse from node `5` to any
# inbound edges, and use multiple
# conditions for the traversal
graph |>
  select_nodes_by_id(nodes = 5) |>
  trav_in_edge(
    conditions =
      rel \%in\% c("C", "D") &
      values > 5.5) |>
  get_selection()

# Traverse from node `5` to any
# inbound edges, and use multiple
# conditions with a single-length
# vector
graph |>
  select_nodes_by_id(nodes = 5) |>
  trav_in_edge(
    conditions =
      rel \%in\% c("D", "E") |
      values > 5.5) |>
  get_selection()

# Traverse from node `5` to any
# inbound edges, and use a regular
# expression as a filtering condition
graph |>
  select_nodes_by_id(nodes = 5) |>
  trav_in_edge(
    conditions = grepl("C|D", rel)) |>
  get_selection()

# Show the graph's internal ndf
graph |> get_node_df()

# Show the graph's internal edf
graph |> get_edge_df()

# Perform a traversal from all
# nodes to their incoming edges and,
# while doing so, copy the `label`
# node attribute to any of the nodes'
# incoming edges
graph <-
  graph |>
  select_nodes() |>
  trav_in_edge(
    copy_attrs_from = label)

# Show the graph's internal edge
# data frame after this change
graph |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_pagerank.R
\name{get_pagerank}
\alias{get_pagerank}
\title{Get the PageRank values for all nodes}
\usage{
get_pagerank(graph, directed = TRUE, damping = 0.85)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{directed}{If \code{TRUE} (the default) then directed paths will be
considered for directed graphs. This is ignored for undirected graphs.}

\item{damping}{The damping factor. The default value is set to \code{0.85}.}
}
\value{
A data frame with PageRank values for each of the nodes.
}
\description{
Get the PageRank values for all nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the PageRank scores
# for all nodes in the graph
graph |>
  get_pagerank()

# Colorize nodes according to their
# PageRank scores
graph <-
  graph |>
  join_node_attrs(
    df = get_pagerank(graph)) |>
  colorize_node_attrs(
    node_attr_from = pagerank,
    node_attr_to = fillcolor,
    palette = "RdYlGn")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/import_graph.R
\name{import_graph}
\alias{import_graph}
\title{Import a graph from various graph formats}
\usage{
import_graph(
  graph_file,
  file_type = NULL,
  edges_extra_attr_names = NULL,
  edges_extra_attr_coltypes = NULL,
  graph_name = NULL,
  attr_theme = "default",
  write_backups = FALSE,
  display_msgs = FALSE
)
}
\arguments{
\item{graph_file}{A connection to a graph file. When provided as a path to a
file, it will read the file from disk. Files starting with \verb{http://},
\verb{https://}, \verb{ftp://}, or \verb{ftps://} will be automatically downloaded.}

\item{file_type}{The type of file to be imported. Options are: \code{gml} (GML),
\code{sif} (SIF), \code{edges} (a .edges file), and \code{mtx} (MatrixMarket format). If
not supplied, the type of graph file will be inferred by its file
extension.}

\item{edges_extra_attr_names}{For \code{edges} files, a vector of attribute names
beyond the \code{from} and \code{to} data columns can be provided in the order they
appear in the input data file.}

\item{edges_extra_attr_coltypes}{For \code{edges} files, this is a string of
column types for any attribute columns provided for
\code{edges_extra_attr_names}. This string representation is where each
character represents each of the extra columns of data and the mappings
are: \code{c} -> character, \code{i} -> integer, \code{n} -> number, \code{d} -> double, \code{l} ->
logical, \code{D} -> date, \code{T} -> date time, \code{t} -> time, \verb{?} -> guess, or
\verb{_/-}, which skips the column.}

\item{graph_name}{An optional string for labeling the graph object.}

\item{attr_theme}{The theme (i.e., collection of \code{graph}, \code{node}, and \code{edge}
global graph attributes) to use for this graph. The default theme is called
\code{default}; there are hierarchical layout themes called \code{lr}, \code{tb}, \code{rl},
and \code{bt} (these operate from left-to-right, top-to-bottom, right-to-left,
and bottom-to-top); and, for larger graphs, the \code{fdp} theme provides a
force directed layout. If this is set to \code{NULL} then no global graph
attributes will be applied to the graph upon creation.}

\item{write_backups}{An option to write incremental backups of changing graph
states to disk. If \code{TRUE}, a subdirectory within the working directory will
be created and used to store \code{RDS} files. The default value is \code{FALSE} so
one has to opt in to use this functionality.}

\item{display_msgs}{An option to display messages primarily concerned with
changes in graph selections. By default, this is \code{FALSE}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Import a variety of graphs from different graph formats and create a graph
object.
}
\examples{
\dontrun{
# Import a GML graph file
gml_graph <-
  import_graph(
    system.file(
      "extdata/karate.gml",
      package = "DiagrammeR"))

# Get a count of the graph's nodes
gml_graph \%>\%
  count_nodes()

# Get a count of the graph's edges
gml_graph \%>\%
  count_edges()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/to_igraph.R
\name{to_igraph}
\alias{to_igraph}
\title{Convert a DiagrammeR graph to an igraph one}
\usage{
to_igraph(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
An igraph object.
}
\description{
Convert a DiagrammeR graph to an igraph graph object.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 36,
    m = 50,
    set_seed = 23)

# Confirm that `graph` is a
# DiagrammeR graph by getting
# the object's class
class(graph)

# Convert the DiagrammeR graph
# to an igraph object
ig_graph <- to_igraph(graph)

# Get the class of the converted
# graph, just to be certain
class(ig_graph)

# Get a summary of the igraph
# graph object
summary(ig_graph)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/invert_selection.R
\name{invert_selection}
\alias{invert_selection}
\title{Invert selection of nodes or edges in a graph}
\usage{
invert_selection(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Modify the selection of nodes or edges within a graph object such that all
nodes or edges previously not selected will now be selected and vice versa.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = "standard")

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "leading_to")

# Create a graph
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Select nodes with ID
# values `1` and `3`
graph <-
  graph |>
  select_nodes(
    nodes = c(1, 3))

# Verify that a node
# selection has been made
graph |> get_selection()

# Invert the selection
graph <-
  graph |>
  invert_selection()

# Verify that the node
# selection has been changed
graph |> get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clear_selection.R
\name{clear_selection}
\alias{clear_selection}
\title{Clear an active selection of nodes or edges}
\usage{
clear_selection(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Clear the selection of nodes or edges within a graph object.
}
\examples{
# Create a graph with
# a single path
graph <-
  create_graph() |>
  add_path(n = 5)

# Select nodes with IDs `1`
# and `3`
graph <-
  graph |>
  select_nodes(
    nodes = c(1, 3))

# Verify that a node selection
# has been made
graph |> get_selection()

# Clear the selection with
# `clear_selection()`
graph <-
  graph |>
  clear_selection()

# Verify that the node
# selection has been cleared
graph |> get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_edge_attrs_ws.R
\name{set_edge_attrs_ws}
\alias{set_edge_attrs_ws}
\title{Set edge attributes with an edge selection}
\usage{
set_edge_attrs_ws(graph, edge_attr, value)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr}{The name of the attribute to set.}

\item{value}{The value to be set for the chosen attribute for the edges in
the current selection.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} or an edge data frame, set edge
attribute properties for one or more edges.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a simple graph
graph <-
  create_graph() |>
  add_path(n = 6)

# Select specific edges from
# the graph and apply the edge
# attribute `color = blue` to
# those selected edges
graph <-
  graph |>
  select_nodes_by_id(nodes = 2:4) |>
  trav_out_edge() |>
  set_edge_attrs_ws(
    edge_attr = color,
    value = "blue")

# Show the internal edge data
# frame to verify that the
# edge attribute has been set
# for specific edges
graph |> get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_adhesion.R
\name{get_adhesion}
\alias{get_adhesion}
\title{Get graph adhesion}
\usage{
get_adhesion(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single numeric value representing the minimum number of edges to
remove.
}
\description{
Get the adhesion of a graph, which is the minimum number of edges needed to
remove to obtain a graph which is not strongly connected. This is the same as
the edge connectivity of the graph.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() \%>\%
  add_cycle(n = 5)

# Determine the graph's adhesion
graph \%>\% get_adhesion()

# Create a full graph and then
# get the adhesion for that
create_graph() \%>\%
  add_full_graph(n = 8) \%>\%
  get_adhesion()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_paths.R
\name{get_paths}
\alias{get_paths}
\title{Get paths from a specified node in a directed graph}
\usage{
get_paths(
  graph,
  from = NULL,
  to = NULL,
  shortest_path = FALSE,
  longest_path = FALSE,
  distance = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{from}{The node from which all paths will be determined.}

\item{to}{The node to which all paths will be determined.}

\item{shortest_path}{An option to return paths that are the shortest in the
set of all determined paths.}

\item{longest_path}{An option to return paths that are the longest in the set
of all determined paths.}

\item{distance}{A vector of integer values that specify which of the valid
paths to return when filtering by distance.}
}
\value{
A list of paths, sorted by ascending traversal length, comprising
vectors of node IDs in sequence of traversal through the graph.
}
\description{
Obtain a list of all possible paths from a given node within a directed
graph.
}
\examples{
# Create a simple graph
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 8) \%>\%
  add_edge(from = 1, to = 2) \%>\%
  add_edge(from = 1, to = 3) \%>\%
  add_edge(from = 3, to = 4) \%>\%
  add_edge(from = 3, to = 5) \%>\%
  add_edge(from = 4, to = 6) \%>\%
  add_edge(from = 2, to = 7) \%>\%
  add_edge(from = 7, to = 5) \%>\%
  add_edge(from = 4, to = 8)

# Get a list of all paths outward from node `1`
graph \%>\%
  get_paths(from = 1)

# Get a list of all paths leading to node `6`
graph \%>\%
  get_paths(to = 6)

# Get a list of all paths from `1` to `5`
graph \%>\%
  get_paths(
   from = 1,
   to = 5)

# Get a list of all paths from `1` up to a distance
# of 2 node traversals
graph \%>\%
  get_paths(
    from = 1,
    distance = 2)

# Get a list of the shortest paths from `1` to `5`
get_paths(
  graph,
  from = 1,
  to = 5,
  shortest_path = TRUE)

# Get a list of the longest paths from `1` to `5`
get_paths(
  graph,
  from = 1,
  to = 5,
  longest_path = TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_balanced_tree.R
\name{add_balanced_tree}
\alias{add_balanced_tree}
\title{Add a balanced tree to the graph}
\usage{
add_balanced_tree(
  graph,
  k,
  h,
  type = NULL,
  label = TRUE,
  rel = NULL,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{k}{The branching factor for the tree.}

\item{h}{The height of the tree.}

\item{type}{An optional string that describes the entity type for the nodes
to be added.}

\item{label}{Either a vector object of length \code{n} that provides optional
labels for the new nodes, or, a boolean value where setting to \code{TRUE}
ascribes node IDs to the label and \code{FALSE} yields a blank label.}

\item{rel}{An optional string for providing a relationship label to all new
edges created in the node tree.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add a balanced tree to the graph.
}
\examples{
# Create a new graph and
# add 2 different types of
# balanced trees of height
# 2 (branching twice) and
# different branching ratios
graph <-
  create_graph() \%>\%
  add_balanced_tree(
    k = 2,
    h = 2,
    type = "binary") \%>\%
  add_balanced_tree(
    k = 3,
    h = 2,
    type = "tertiary")

# Get some node information
# from this graph
graph \%>\%
  get_node_info() \%>\%
  head(5)

# Node and edge aesthetic and data
# attributes can be specified in
# the `node_aes`, `edge_aes`,
# `node_data`, and `edge_data`
# arguments
graph_w_attrs <-
  create_graph() \%>\%
  add_balanced_tree(
    k = 2,
    h = 2,
    label = c(
      "one", "two",
      "three", "four",
      "five", "six", "seven"),
    type = c(
      "a", "b", "b", "c",
      "c", "c", "c"),
    rel = "A",
    node_aes = node_aes(
      fillcolor = "steelblue"),
    node_data = node_data(
      value = c(
        1.6, 2.8, 3.4, 8.3,
        3.8, 5.2, 3.2)),
    edge_aes = edge_aes(
      color = "red",
      penwidth = 1.2))

# Get the first three rows of
# the graph's node data frame
graph_w_attrs \%>\%
  get_node_df() \%>\%
  head(3)

# Get the first three rows of
# the graph's edge data frame
graph_w_attrs \%>\%
  get_edge_df() \%>\%
  head(3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DiagrammeR.R
\name{renderDiagrammeR}
\alias{renderDiagrammeR}
\title{Widget render function for use in Shiny}
\usage{
renderDiagrammeR(expr, env = parent.frame(), quoted = FALSE)
}
\arguments{
\item{expr}{An expression that generates a DiagrammeR graph}

\item{env}{The environment in which to evaluate expr.}

\item{quoted}{Is expr a quoted expression (with quote())? This is useful if
you want to save an expression in a variable.}
}
\description{
Widget render function for use in Shiny
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_periphery.R
\name{get_periphery}
\alias{get_periphery}
\title{Get nodes that form the graph periphery}
\usage{
get_periphery(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A vector of node IDs.
}
\description{
Get those nodes that are part of the graph periphery (i.e., have the maximum
eccentricity in the graph).
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function and
# get the nodes in the graph periphery
create_graph() \%>\%
  add_gnm_graph(
    n = 28,
    m = 35,
    set_seed = 23) \%>\%
  get_periphery()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_edges_w_string.R
\name{add_edges_w_string}
\alias{add_edges_w_string}
\title{Add one or more edges using a text string}
\usage{
add_edges_w_string(graph, edges, rel = NULL, use_labels = FALSE)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edges}{A single-length vector with a character string specifying the
edges. For a directed graph, the string object should be formatted as a
series of node ID values as \verb{[node_ID_1]->[node_ID_2]} separated by a one
or more space characters. For undirected graphs, \verb{--} should replace \verb{->}.
Line breaks in the vector won't cause an error.}

\item{rel}{An optional vector specifying the relationship between the
connected nodes.}

\item{use_labels}{An option to use node \code{label} values in the \code{edges} string
to define node connections. Note that this is only possible if all nodes
have distinct \code{label} values set and none exist as an empty string.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add one or more edges to the graph
using a text string.
}
\examples{
# Create a graph with 4 nodes
graph <-
  create_graph() \%>\%
  add_node(label = "one") \%>\%
  add_node(label = "two") \%>\%
  add_node(label = "three") \%>\%
  add_node(label = "four")

# Add edges between nodes using
# a character string with node
# ID values
graph_node_id <-
  graph \%>\%
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->3")

# Show the graph's internal
# edge data frame
graph_node_id \%>\% get_edge_df()

# Add edges between nodes using
# a character string with node
# label values and setting
# `use_labels = TRUE`; note that
# all nodes must have unique
# `label` values to use this
graph_node_label <-
  graph \%>\%
  add_edges_w_string(
    edges =
      "one->two one->three
       two->four two->three",
    use_labels = TRUE)

# Show the graph's internal
# edge data frame (it's the
# same as before)
graph_node_label \%>\% get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_graph_undirected.R
\name{set_graph_undirected}
\alias{set_graph_undirected}
\title{Convert a directed graph to an undirected graph}
\usage{
set_graph_undirected(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Take a graph which is directed and convert it to an undirected graph.
}
\examples{
# Create a graph with a
# directed tree
graph <-
  create_graph() |>
  add_balanced_tree(
    k = 2, h = 2)

# Convert this graph from
# directed to undirected
graph <-
  graph |>
  set_graph_undirected()

# Perform a check on whether
# graph is directed
graph |> is_graph_directed()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_degree_histogram.R
\name{get_degree_histogram}
\alias{get_degree_histogram}
\title{Get histogram data for a graph's degree frequency}
\usage{
get_degree_histogram(graph, mode = "total")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{mode}{using \code{total} (the default), degree considered for each node
will be the total degree. With \verb{in} and \code{out} the degree used
will be the in-degree and out-degree, respectively.}
}
\value{
A data frame with degree counts.
}
\description{
Get histogram data for a graph's degree frequency. The bin width is set to 1
and zero-value degrees are omitted from the output.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get degree histogram data for
# the graph (reporting total degree)
graph \%>\%
  get_degree_histogram(
    mode = "total")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{usd_exchange_rates}
\alias{usd_exchange_rates}
\title{US Dollar exchange rates.}
\format{
A data frame with 196 rows and 3 variables:
\describe{
\item{from_currency}{the currency from which units will be used to buy units in the alternate currency (this is always USD)}
\item{to_currency}{the currency that is to be bought}
\item{cost_unit}{the cost per unit of the currency to be bought}
}
}
\usage{
usd_exchange_rates
}
\description{
A dataset containing exchange rates from USD to all other currencies.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_graph.R
\name{create_graph}
\alias{create_graph}
\title{Create a graph object}
\usage{
create_graph(
  nodes_df = NULL,
  edges_df = NULL,
  directed = TRUE,
  graph_name = NULL,
  attr_theme = "default",
  write_backups = FALSE,
  display_msgs = FALSE
)
}
\arguments{
\item{nodes_df}{An optional data frame containing, at minimum, a column
(called \code{id}) which contains node IDs for the graph. Additional columns
(node attributes) can be included with values for the named node attribute.}

\item{edges_df}{An optional data frame containing, at minimum, two columns
(called \code{from} and \code{to}) where node IDs are provided. Additional columns
(edge attributes) can be included with values for the named edge attribute.}

\item{directed}{With \code{TRUE} (the default) or \code{FALSE}, either directed or
undirected edge operations will be generated, respectively.}

\item{graph_name}{An optional string for labeling the graph object.}

\item{attr_theme}{The theme (i.e., collection of \code{graph}, \code{node}, and \code{edge}
global graph attributes) to use for this graph. The default theme is called
\code{default}; there are hierarchical layout themes called \code{lr}, \code{tb}, \code{rl},
and \code{bt} (these operate from left-to-right, top-to-bottom, right-to-left,
and bottom-to-top); and, for larger graphs, the \code{fdp} theme provides a
force directed layout. If this is set to \code{NULL} then no global graph
attributes will be applied to the graph upon creation.}

\item{write_backups}{An option to write incremental backups of changing graph
states to disk. If \code{TRUE}, a subdirectory within the working directory will
be created and used to store \code{RDS} files. The default value is \code{FALSE} so
one has to opt in to use this functionality.}

\item{display_msgs}{An option to display messages primarily concerned with
changes in graph selections. By default, this is \code{FALSE}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Generates a graph object with the option to use node data frames (ndfs)
and/or edge data frames (edfs) to populate the initial graph.
}
\examples{
# With `create_graph()` we can
# simply create an empty graph (and
# add in nodes and edges later
# with other functions)
graph <- create_graph()

# A graph can be created with
# nodes and without having any edges;
# this can be done in 2 steps:
# 1. create a node data frame (ndf)
#    using `create_node_df()`
ndf <-
  create_node_df(n = 4)

# 2. create a new graph object with
#    `create_graph()` and then pass
#    in the ndf to `nodes_df`
graph <-
  create_graph(
    nodes_df = ndf)

# Get information on the graph's nodes
graph \%>\%
  get_node_info()

# You can create a similar graph with
# just nodes but also providing a
# range of attributes for the nodes
# (e.g., types, labels, or arbitrary
# 'values')
ndf <-
  create_node_df(
    n = 4,
    label = TRUE,
    type = c("type_1", "type_1",
             "type_5", "type_2"),
    shape = c("circle", "circle",
              "rectangle", "rectangle"),
    values = c(3.5, 2.6, 9.4, 2.7))

graph <-
  create_graph(nodes_df = ndf)

# Get information on the graph's
# internal node data frame (ndf)
graph \%>\%
  get_node_df()

# A graph can also be created by
# specifying both the nodes and
# edges; create an edge data frame
# (edf) using the `create_edge_df()`
# function:
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "leading_to",
    values = c(7.3, 2.6, 8.3))

# Create the graph object with
# `create_graph()` and pass in the
# ndf and edf objects
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Get information on the graph's
# internal edge data frame (edf)
graph \%>\% get_edge_df()

# Get information on the graph's
# internal node data frame (ndf)
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_max_eccentricity.R
\name{get_max_eccentricity}
\alias{get_max_eccentricity}
\title{Get the maximum graph eccentricity}
\usage{
get_max_eccentricity(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single numeric value representing the maximum eccentricity of the
graph.
}
\description{
Get the diameter of a graph, which is the largest eccentricity in the graph.
The graph eccentricity of a node is its shortest path from the farthest other
node in the graph.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() |>
  add_cycle(n = 5)

# Determine the graph's maximum
# eccentricity
graph |>
  get_max_eccentricity()

# Create a full graph and then
# get the maximum eccentricity
# value for that
create_graph() |>
  add_full_graph(n = 10) |>
  get_max_eccentricity()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remove_graph_from_graph_series.R
\name{remove_graph_from_graph_series}
\alias{remove_graph_from_graph_series}
\title{Remove a graph from a graph series}
\usage{
remove_graph_from_graph_series(graph_series, index = "last")
}
\arguments{
\item{graph_series}{A graph series object from which the graph object will be
removed.}

\item{index}{The index of the graph object to be removed from the graph
series object.}
}
\value{
A graph series object of type \code{dgr_graph_1D}.
}
\description{
Remove a single graph object from an set of graph objects contained within a
graph series object.
}
\examples{
# Create three graphs
graph_1 <-
  create_graph() |>
  add_path(n = 4)

graph_2 <-
  create_graph() |>
  add_cycle(n = 5)

graph_3 <-
  create_graph() |>
  add_star(n = 6)

# Create an empty graph series
# and add the graphs
series <-
  create_graph_series() |>
  add_graph_to_graph_series(
    graph = graph_1) |>
  add_graph_to_graph_series(
    graph = graph_2) |>
  add_graph_to_graph_series(
    graph = graph_3)

# Remove the second graph
# from the graph series
series <-
  series |>
  remove_graph_from_graph_series(
    index = 2)

# With `get_graph_series_info()`,
# we can ensure that a graph
# was removed
series |>
  get_graph_series_info()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_all_connected_nodes.R
\name{get_all_connected_nodes}
\alias{get_all_connected_nodes}
\title{Get all nodes connected to a specified node}
\usage{
get_all_connected_nodes(graph, node)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{a single-length vector containing a node ID value.}
}
\value{
A vector of node ID values.
}
\description{
With a single node serving as the starting point get all nodes connected
(i.e., reachable with a traversable path) to that node.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function; it
# has an unconnected node (`6`)
graph_1 <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 20,
    m = 32,
    set_seed = 23)

# There won't be any connected
# nodes to `6` so when specifying
# this node with `get_all_connected_nodes()`
# we get NA back
graph_1 \%>\%
  get_all_connected_nodes(
    node = 6)

# Any other node in `graph_1` will
# provide a vector of all the nodes
# other than `6`
graph_1 \%>\%
  get_all_connected_nodes(
    node = 1)

# The following graph has two
# clusters of nodes (i.e., the
# graph has two connected components)
graph_2 <-
  create_graph() \%>\%
  add_path(n = 6) \%>\%
  add_path(n = 4)

# In `graph_2`, node `1` is in
# the larger of the two
# connected components
graph_2 \%>\%
  get_all_connected_nodes(
    node = 1)

# Also in `graph_2`, node `8`
# is in the smaller of the two
# connected components
graph_2 \%>\%
  get_all_connected_nodes(
    node = 8)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_full_graph.R
\name{add_full_graph}
\alias{add_full_graph}
\title{Add a fully connected graph}
\usage{
add_full_graph(
  graph,
  n,
  type = NULL,
  label = TRUE,
  rel = NULL,
  edge_wt_matrix = NULL,
  keep_loops = FALSE,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{n}{The number of nodes comprising the fully connected graph.}

\item{type}{An optional string that describes the entity type for the nodes
to be added.}

\item{label}{Either a vector object of length \code{n} that provides optional
labels for the new nodes, or, a boolean value where setting to \code{TRUE}
ascribes node IDs to the label and \code{FALSE} or \code{NULL} yields a blank label.}

\item{rel}{An optional string for providing a relationship label to all new
edges created in the connected graph.}

\item{edge_wt_matrix}{An optional matrix of \code{n} by \code{n} dimensions containing
values to apply as edge weights. If the matrix has row names or column
names and \code{label = TRUE}, those row or column names will be used as node
label values.}

\item{keep_loops}{An option to simplify the fully connected graph by removing
loops (edges from and to the same node). The default value is \code{FALSE}.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add a fully connected graph either
with or without loops. If the graph object set as directed, the added graph
will have edges to and from each pair of nodes. In the undirected case, a
single edge will link each pair of nodes.
}
\examples{
# Create a new graph object
# and add a directed and fully
# connected graph with 3 nodes
# and edges to and from all
# pairs of nodes; with the option
# `keep_loops = TRUE` nodes
# will also have edges from
# and to themselves
graph <-
  create_graph() \%>\%
  add_full_graph(
    n = 3, keep_loops = TRUE
  )

# Get node information
# from this graph
graph \%>\% get_node_info()

# Using `keep_loops = FALSE`
# (the default) will remove
# the loops
create_graph() \%>\%
  add_full_graph(n = 3) \%>\%
  get_node_info()

# Values can be set for
# the node `label`, node
# `type`, and edge `rel`
graph <-
  create_graph() \%>\%
  add_full_graph(
    n = 3,
    type = "connected",
    label = c("1st", "2nd", "3rd"),
    rel = "connected_to"
  )

# Show the graph's node
# data frame (ndf)
graph \%>\% get_node_df()

# Show the graph's edge
# data frame (edf)
graph \%>\% get_edge_df()

# Create a fully-connected and
# directed graph with 3 nodes,
# and, where a matrix provides
# edge weights; first, create the
# matrix (with row names to be
# used as node labels)
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

edge_wt_matrix <-
  rnorm(100, 5, 2) \%>\%
  sample(9, FALSE) \%>\%
  round(2) \%>\%
  matrix(
    ncol = 3,
    nrow = 3,
    dimnames = list(c("a", "b", "c"))
  )

# Create the fully-connected
# graph (without loops however)
graph <-
  create_graph() \%>\%
  add_full_graph(
    n = 3,
    type = "weighted",
    label = TRUE,
    rel = "related_to",
    edge_wt_matrix = edge_wt_matrix,
    keep_loops = FALSE
  )

# Show the graph's node
# data frame (ndf)
graph \%>\% get_node_df()

# Show the graph's edge
# data frame (edf)
graph \%>\% get_edge_df()

# An undirected graph can
# also use a matrix with
# edge weights, but only
# the lower triangle of
# that matrix will be used
create_graph(directed = FALSE) \%>\%
  add_full_graph(
    n = 3,
    type = "weighted",
    label = TRUE,
    rel = "related_to",
    edge_wt_matrix = edge_wt_matrix,
    keep_loops = FALSE
  ) \%>\%
  get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_non_nbrs.R
\name{get_non_nbrs}
\alias{get_non_nbrs}
\title{Get non-neighbors of a node in a graph}
\usage{
get_non_nbrs(graph, node)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{A single-length vector containing a node ID value.}
}
\value{
A vector of node ID values.
}
\description{
Get the set of all nodes not neighboring a single graph node.
}
\examples{
# Create a simple, directed graph with 5
# nodes and 4 edges
graph <-
  create_graph() |>
  add_path(n = 5)

# Find all non-neighbors of node `2`
graph |> get_non_nbrs(node = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recode_node_attrs.R
\name{recode_node_attrs}
\alias{recode_node_attrs}
\title{Recode a set of node attribute values}
\usage{
recode_node_attrs(
  graph,
  node_attr_from,
  ...,
  otherwise = NULL,
  node_attr_to = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr_from}{The name of the node attribute column from which values
will be recoded.}

\item{...}{Single-length character vectors with the recoding instructions.
The first component should have the value to replace and the second should
have the replacement value (in the form \verb{"[to_replace] -> [replacement]", ...}).}

\item{otherwise}{An optional single value for recoding any unmatched values.}

\item{node_attr_to}{An optional name of a new node attribute to which the
recoded values will be applied. This will retain the original node
attribute and its values.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal node data frame (ndf), recode character or numeric
node attribute values. Optionally, one can specify a replacement value for
any unmatched mappings.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 5,
    m = 10,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = shape,
    values =
      c("circle", "hexagon",
        "rectangle", "rectangle",
        "circle"))

# Get the graph's internal ndf
# to show which node
# attributes are available
graph |> get_node_df()

# Recode the `shape` node
# attribute, so that `circle`
# is recoded to `square` and that
# `rectangle` becomes `triangle`
graph <-
  graph |>
  recode_node_attrs(
    node_attr_from = shape,
    "circle -> square",
    "rectangle -> triangle")

# Get the graph's internal
# ndf to show that the node
# attribute values had been recoded
graph |> get_node_df()

# Create a new node attribute,
# called `color`, that is based
# on a recoding of `shape`; here,
# map the square shape to a `red`
# color and map all other shapes
# to a `green` color
graph <-
  graph |>
  recode_node_attrs(
    node_attr_from = shape,
    "square -> red",
    otherwise = "green",
    node_attr_to = color)

# Get the graph's internal ndf
# to see the change
graph |> get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trav_in.R
\name{trav_in}
\alias{trav_in}
\title{Traverse from one or more selected nodes onto adjacent, inward nodes}
\usage{
trav_in(
  graph,
  conditions = NULL,
  copy_attrs_from = NULL,
  copy_attrs_as = NULL,
  agg = "sum",
  add_to_selection = FALSE
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the traversal.}

\item{copy_attrs_from}{providing a node attribute name will copy those node
attribute values to the traversed nodes. Any values extant on the nodes
traversed to will be replaced.}

\item{copy_attrs_as}{If a node attribute name is provided in
\code{copy_attrs_from}, this option will allow the copied attribute values to be
written under a different attribute name. If the attribute name provided in
\code{copy_attrs_as} does not exist in the graph's ndf, the new node attribute
will be created with the chosen name.}

\item{agg}{If a node attribute is provided to \code{copy_attrs_from}, then an
aggregation function is required since there may be cases where multiple
edge attribute values will be passed onto the traversed node(s). To pass
only a single value, the following aggregation functions can be used:
\code{sum}, \code{min}, \code{max}, \code{mean}, or \code{median}.}

\item{add_to_selection}{An option to either add the traversed to nodes to the
active selection of nodes (\code{TRUE}) or switch the active selection entirely
to those traversed to nodes (\code{FALSE}, the default case).}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} move along inward edges from one or
more nodes present in a selection to other connected nodes, replacing the
current nodes in the selection with those nodes traversed to. An optional
filter by node attribute can limit the set of nodes traversed to.

This traversal function makes use of an active selection of nodes. After the
traversal, depending on the traversal conditions, there will either be a
selection of nodes or no selection at all.

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a simple graph
graph <-
  create_graph() \%>\%
  add_n_nodes(
    n = 2,
    type = "a",
    label = c("asd", "iekd")) \%>\%
  add_n_nodes(
    n = 3,
    type = "b",
    label = c("idj", "edl", "ohd")) \%>\%
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->5",
    rel = c(NA, "A", "B", "C", "D"))

# Create a data frame with node ID values
# representing the graph edges (with `from`
# and `to` columns), and, a set of numeric values
df_edges <-
  data.frame(
    from = c(1, 1, 2, 2, 3),
    to = c(2, 3, 4, 5, 5),
    values = round(rnorm(5, 5), 2))

# Create a data frame with node ID values
# representing the graph nodes (with the `id`
# columns), and, a set of numeric values
df_nodes <-
  data.frame(
    id = 1:5,
    values = round(rnorm(5, 7), 2))

# Join the data frame to the graph's internal
# edge data frame (edf)
graph <-
  graph \%>\%
  join_edge_attrs(df = df_edges) \%>\%
  join_node_attrs(df = df_nodes)

# Show the graph's internal node data frame
graph \%>\% get_node_df()

# Show the graph's internal edge data frame
graph \%>\% get_edge_df()

# Perform a simple traversal from node `4` to
# inward adjacent edges with no conditions
# on the nodes traversed to
graph \%>\%
  select_nodes_by_id(nodes = 4) \%>\%
  trav_in() \%>\%
  get_selection()

# Traverse from node `5` to inbound-facing
# nodes, filtering to those nodes that have
# numeric values greater than `5.0` for
# the `values` node attribute
graph \%>\%
  select_nodes_by_id(nodes = 4) \%>\%
  trav_in(
    conditions = values > 5.0) \%>\%
  get_selection()

# Traverse from node `5` to any inbound
# nodes, filtering to those nodes that
# have a `type` attribute of `b`
graph \%>\%
  select_nodes_by_id(nodes = 5) \%>\%
  trav_in(
    conditions = type == "b") \%>\%
  get_selection()

# Traverse from node `5` to any inbound
# nodes, filtering to those nodes that
# have a degree of `2`
graph \%>\%
  {
  node_degrees <-
    get_node_info(.) \%>\%
    dplyr::select(id, deg)
  join_node_attrs(., node_degrees)
  } \%>\%
  select_nodes_by_id(nodes = 5) \%>\%
  trav_in(
    conditions = deg == 2) \%>\%
  get_selection()

# Traverse from node `5` to any inbound
# nodes, and use multiple conditions for the
# traversal
graph \%>\%
  select_nodes_by_id(nodes = 5) \%>\%
  trav_in(
    conditions =
      type == "a" &
      values > 6.0) \%>\%
  get_selection()

# Traverse from node `5` to any inbound
# nodes, and use multiple conditions with
# a single-length vector
graph \%>\%
  select_nodes_by_id(nodes = 5) \%>\%
  trav_in(
    conditions =
      type == "b" | values > 6.0) \%>\%
  get_selection()

# Traverse from node `5` to any inbound
# nodes, and use a regular expression as
# a filtering condition
graph \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  trav_in(
    conditions = grepl("^i.*", label)) \%>\%
  get_selection()

# Create another simple graph to demonstrate
# copying of node attribute values to traversed
# nodes
graph <-
  create_graph() \%>\%
  add_node() \%>\%
  select_nodes() \%>\%
  add_n_nodes_ws(
    n = 2,
    direction = "from") \%>\%
  clear_selection() \%>\%
  select_nodes_by_id(nodes = 2:3) \%>\%
  set_node_attrs_ws(
    node_attr = value,
    value = 5)

# Show the graph's internal node data frame
graph \%>\% get_node_df()

# Show the graph's internal edge data frame
graph \%>\% get_edge_df()

# Perform a traversal from the outer nodes
# (`2` and `3`) to the central node (`1`) while
# also applying the node attribute `value` to
# node `1` (summing the `value` of 5 from
# both nodes before applying the value to the
# target node)
graph <-
  graph \%>\%
  trav_in(
    copy_attrs_from = value,
    agg = "sum")

# Show the graph's internal node data frame
# after this change
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/layout_nodes_w_string.R
\name{layout_nodes_w_string}
\alias{layout_nodes_w_string}
\title{Layout nodes using a text-based schematic}
\usage{
layout_nodes_w_string(
  graph,
  layout,
  nodes,
  sort = NULL,
  width = 8,
  height = 8,
  ll = c(0, 0)
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{layout}{A layout character string that provides a schematic for the
layout. This consists of a rectangular collection of \code{-} characters (for no
node placement), and numbers from \code{1} to \code{9} (representing different
groupings of nodes, further described in the \code{nodes} argument).}

\item{nodes}{A named vector of the form: \code{c("1" = "[node_attr]:[value]", ...)}. The LHS corresponds to the numbers used in the \code{layout} schematic.
The RHS provides a shorthand for the node attribute and a value for
grouping together nodes (separated by a colon). For instance, with
\code{"type:a"} in the RHS (and \code{"1"} in the LHS) we would target all nodes with
a \code{type} attribute equal to \code{a} for positioning in the graph as described
by the \code{1}s in the \code{layout}.}

\item{sort}{An optional sorting method to apply to the collection of nodes
before assigning positional information. Like \code{nodes}, this is a named
vector of the form: \code{c("1" = "[node_attr]:asc|desc", ...)}. The \code{node_attr}
in this case should be different than that used in \code{nodes}. Ideally, this
node attribute should have unique values. Choose either \code{asc} or \code{desc}
right of the colon for ascending or descending sorts.}

\item{width}{The width of the \code{layout} diagram.}

\item{height}{The height of the \code{layout} diagram.}

\item{ll}{A vector describing the the lower-left coordinates of the \code{layout}}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Layout one or several groups of nodes using a text-based schematic. The
option is available to apply sorting to each of the groups.
}
\examples{
# Create a graph with unique labels and
# several node `type` groupings
graph <-
  create_graph() |>
  add_node(type = "a", label = "a") |>
  add_node(type = "a", label = "b") |>
  add_node(type = "b", label = "c") |>
  add_node(type = "b", label = "d") |>
  add_node(type = "b", label = "e") |>
  add_node(type = "c", label = "f") |>
  add_node(type = "c", label = "g")

# Define a 'layout' for groups of nodes
# using a text string (dashes are empty
# grid cells, numbers--representing
# ad-hoc groupings--correspond to
# individual nodes); here, define a layout
# with 3 groups of nodes
layout <-
"
1--------
1--------
---222---
--------3
--------3
"

# Use the `layout` along with what nodes
# the numbers correspond to in the graph
# with the `nodes` named vectors; the
# optional `sort` vector describes how
# we should sort the collection of node
# before adding position information
graph <-
  graph |>
  layout_nodes_w_string(
    layout = layout,
    nodes = c("1" = "type:a",
              "2" = "type:b",
              "3" = "type:c"),
    sort = c("1" = "label:asc",
             "2" = "label:desc",
             "3" = "label:desc"))

# Show the graph's node data frame
# to confirm that `x` and `y` values
# were added to each of the nodes
graph |> get_node_df()
}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_node_present.R
\name{is_node_present}
\alias{is_node_present}
\title{Determine whether a specified node is present}
\usage{
is_node_present(graph, node)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{Either a node ID value or a node label to test for presence in
the graph.}
}
\value{
A logical value.
}
\description{
From a graph object of class \code{dgr_graph}, determine whether a specified node
is present.
}
\examples{
# Create a simple graph with
# a path of four nodes
graph <-
  create_graph() \%>\%
  add_path(
    n = 4,
    type = "path",
    label = c(
      "one", "two",
      "three", "four"))

# Determine if there is a node
# with ID `1` in the graph
graph \%>\%
  is_node_present(node = 1)

# Determine if there is a node
# with ID `5` in the graph
graph \%>\%
  is_node_present(node = 5)

# Determine if there is a node
# with label `two` in the graph
graph \%>\%
  is_node_present(node = "two")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_edge_present.R
\name{is_edge_present}
\alias{is_edge_present}
\title{Determine whether a specified edge is present}
\usage{
is_edge_present(graph, edge = NULL, from = NULL, to = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge}{An edge ID value to test for presence in the graph. If a single,
numeric value is provided then values for \code{from} or \code{to} needn't be
supplied.}

\item{from}{A node ID from which the edge is outgoing, or, the label
associated with the node. For an undirected graph, the value in \code{from} can
be interchangeable with that in \code{to}.}

\item{to}{A node ID to which the edge is incoming, or, the label associated
with the node. For an undirected graph, the value in \code{to} can be
interchangeable with that in \code{from}.}
}
\value{
A logical value.
}
\description{
From a graph object of class \code{dgr_graph}, determine whether an edge (defined
by a pair of node IDs or node label values) is present.
}
\examples{
# Create a simple graph with
# a path of four nodes
graph <-
  create_graph() |>
  add_path(
    n = 4,
    type = "path",
    label = c("one", "two",
              "three", "four"))

# Find out if edge ID `3`
# is present in the graph
graph |>
  is_edge_present(edge = 3)

# Determine if there are any edges
# with the definition `1` -> `2`
graph |>
  is_edge_present(
    from = 1,
    to = 2)

# Determine if there are any edges
# with the definition `4` -> `5`
graph |>
  is_edge_present(
    from = 4,
    to = 5)

# Determine whether an edge,
# defined by its labels as
# `two` -> `three`, exists in
# the graph
graph |>
  is_edge_present(
    from = "two",
    to = "three")

# Set the graph as undirected
# and determine whether an
# edge between nodes with labels
# `three` and `two` exists
graph |>
  set_graph_undirected() |>
  is_edge_present(
    from = "three",
    to = "two")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_s_connected_cmpts.R
\name{get_s_connected_cmpts}
\alias{get_s_connected_cmpts}
\title{Get nodes within strongly connected components}
\usage{
get_s_connected_cmpts(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame with nodes and their membership in different strongly
connected components.
}
\description{
Determine which nodes in a graph belong to different strongly connected
components.
}
\examples{
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a graph with a random
# connection between 2 different
# node cycles
graph <-
  create_graph() \%>\%
  add_cycle(
    n = 3,
    type = "cycle_1") \%>\%
  add_cycle(
    n = 4,
    type = "cycle_2") \%>\%
  add_edge(
    from =
      get_node_ids(
        graph = .,
        conditions =
          type == "cycle_1") \%>\%
        sample(size = 1),
    to =
      get_node_ids(
        graph = .,
        conditions =
          type == "cycle_2") \%>\%
        sample(size = 1))

# Get the strongly connected
# components as a data frame of
# nodes and their groupings
graph \%>\% get_s_connected_cmpts()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_global_graph_attrs.R
\name{delete_global_graph_attrs}
\alias{delete_global_graph_attrs}
\title{Delete one of the global graph attributes stored within a graph object}
\usage{
delete_global_graph_attrs(graph, attr = NULL, attr_type = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{attr}{The name of the attribute to delete for the \code{type} of global
attribute specified.}

\item{attr_type}{The specific type of global graph attribute to delete. The
type is specified with \code{graph}, \code{node}, or \code{edge}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Delete one of the global attributes stored within a graph object of class
\code{dgr_graph}).
}
\examples{
# Create a new graph and add
# some extra global graph attrs
graph <-
  create_graph() \%>\%
  add_global_graph_attrs(
    attr = "overlap",
    value = "true",
    attr_type = "graph") \%>\%
  add_global_graph_attrs(
    attr = "penwidth",
    value = 3,
    attr_type = "node") \%>\%
  add_global_graph_attrs(
    attr = "penwidth",
    value = 3,
    attr_type = "edge")

# Inspect the graph's global
# attributes
graph \%>\%
  get_global_graph_attr_info()

# Delete the `penwidth` attribute
# for the graph's nodes using the
# `delete_global_graph_attrs()` fcn
graph <-
  graph \%>\%
  delete_global_graph_attrs(
    attr = "penwidth",
    attr_type = "node")

# View the remaining set of global
# attributes for the graph
graph \%>\%
  get_global_graph_attr_info()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reorder_graph_actions.R
\name{reorder_graph_actions}
\alias{reorder_graph_actions}
\title{Reorder the stored series of graph actions}
\usage{
reorder_graph_actions(graph, indices)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{indices}{A numeric vector that provides the new ordering of graph
actions. This vector can be the same length as the number of graph actions,
or, of shorter length. In the latter case, the ordering places the given
items first and the remaining actions will follow.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Reorder the graph actions stored in the graph through the use of the
\code{\link[=add_graph_action]{add_graph_action()}} function. These actions are be invoked in a specified
order via the \code{\link[=trigger_graph_actions]{trigger_graph_actions()}} function.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 4,
    m = 4,
    set_seed = 23)

# Add three graph actions to the
# graph
graph <-
  graph \%>\%
  add_graph_action(
    fcn = "rescale_node_attrs",
    node_attr_from = "pagerank",
    node_attr_to = "width",
    action_name = "pgrnk_to_width") \%>\%
  add_graph_action(
    fcn = "set_node_attr_w_fcn",
    node_attr_fcn = "get_pagerank",
    column_name = "pagerank",
    action_name = "get_pagerank") \%>\%
  add_graph_action(
    fcn = "colorize_node_attrs",
    node_attr_from = "width",
    node_attr_to = "fillcolor",
    action_name = "pgrnk_fillcolor")

# View the graph actions for the graph
# object by using the function called
# `get_graph_actions()`
graph \%>\% get_graph_actions()

# We note that the order isn't
# correct and that the `get_pagerank`
# action should be the 1st action
# and `pgrnk_to_width` should go
# in 2nd place; to fix this, use the
# function `reorder_graph_actions()`
# and specify the reordering with a
# numeric vector
graph <-
  graph \%>\%
  reorder_graph_actions(
    indices = c(2, 1, 3))

# View the graph actions for the graph
# object once again to verify that
# we have the desired order of actions
graph \%>\% get_graph_actions()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_graph_weighted.R
\name{is_graph_weighted}
\alias{is_graph_weighted}
\title{Is the graph a weighted graph?}
\usage{
is_graph_weighted(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Provides a logical value on whether the graph is weighted. A graph is
considered to be weighted when it contains edges that all have a edge
\code{weight} attribute with numerical values assigned for all edges.
}
\examples{
# Create a graph where the edges have
# a `weight` attribute
graph <-
  create_graph() \%>\%
  add_cycle(n = 5) \%>\%
  select_edges() \%>\%
  set_edge_attrs_ws(
    edge_attr = weight,
    value = c(3, 5, 2, 9, 6)) \%>\%
  clear_selection()

# Determine whether the graph
# is a weighted graph
graph \%>\% is_graph_weighted()

# Create graph where the edges do
# not have a `weight` attribute
graph <-
  create_graph() \%>\%
  add_cycle(n = 5)

# Determine whether this graph
# is weighted
graph \%>\% is_graph_weighted()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_reverse_edges_ws.R
\name{add_reverse_edges_ws}
\alias{add_reverse_edges_ws}
\title{Add new edges in the opposite directions of a selection of edges}
\usage{
add_reverse_edges_ws(graph, rel = NULL, edge_aes = NULL, edge_data = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{rel}{An optional string to apply a \code{rel} attribute to all newly created
edges.}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add edges in the opposite direction of one or more edges available as an edge
selection in a graph object of class \code{dgr_graph}. New graph edges have the
opposite edge definitions as those in the selection. For example, a graph
with the edge \code{1->2} in its active selection will gain a new \code{2->1} edge.
There is also the option to assign a common \code{rel} grouping to the newly
created edges. Upon addition of the edges, the edge selection will be
retained for further selection or traversal operations.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create an empty graph, add 2 nodes to it,
# and create the edge `1->2`
graph <-
  create_graph() |>
  add_n_nodes(
    n = 2,
    type = "type_a",
    label = c("a_1", "a_2")) |>
  add_edge(
    from = 1,
    to = 2,
    rel = "a")

# Get the graph's edges
graph |> get_edge_ids()

# Select the edge and create 2 additional edges
# with the opposite definition of `1->2`, which
# is `2->1`; also, apply, different `rel` values
# (`b` and `c`)
graph <-
  graph |>
  select_edges() |>
  add_reverse_edges_ws(rel = "b") |>
  add_reverse_edges_ws(rel = "c") |>
  clear_selection()

# Get the graph's edge data frame
graph |> get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_graph_info.R
\name{get_graph_info}
\alias{get_graph_info}
\title{Get metrics for a graph}
\usage{
get_graph_info(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame containing metrics pertaining to the graph
}
\description{
Get a data frame with metrics for a graph.
}
\examples{
\dontrun{
# Import a GML graph file available
# in the DiagrammeR package
karate_club <-
  system.file(
    "extdata", "karate.gml",
    package = "DiagrammeR") \%>\%
  import_graph() \%>\%
  set_graph_name("karate")

# Display a data frame with
# graph information
karate_club \%>\%
  get_graph_info()
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{edge_list_1}
\alias{edge_list_1}
\title{Edge list - Version 1.}
\format{
A data frame with 19 rows and 2 variables:
\describe{
\item{from}{integer values that state the node ID values where an edge starts}
\item{to}{integer values that state the node ID values where an edge terminates}
}
}
\usage{
edge_list_1
}
\description{
A very simple, 2-column data frame that can be used to generate graph edges.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_cache.R
\name{delete_cache}
\alias{delete_cache}
\title{Delete vectors cached in a graph object}
\usage{
delete_cache(graph, name = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{name}{One or more name of vector objects to delete from the cache. If
none supplied, all cached vectors available in the graph will be deleted.}
}
\value{
A vector.
}
\description{
Delete vectors cached in a graph object of class \code{dgr_graph}.
}
\examples{
# Create an empty graph
graph <-
  create_graph()

# Cache 3 different vectors inside
# the graph object
graph <-
  graph \%>\%
  set_cache(
    name = "a",
    to_cache = 1:4) \%>\%
  set_cache(
    name = "b",
    to_cache = 5:9) \%>\%
  set_cache(
    name = "c",
    to_cache = 10:14)

# Delete cache `b`
graph <-
  graph \%>\%
  delete_cache(name = "b")

# Delete remaining cached vectors
graph <-
  graph \%>\%
  delete_cache()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_edges_from_table.R
\name{add_edges_from_table}
\alias{add_edges_from_table}
\title{Add edges and attributes to graph from a table}
\usage{
add_edges_from_table(
  graph,
  table,
  from_col,
  to_col,
  from_to_map,
  rel_col = NULL,
  set_rel = NULL,
  drop_cols = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{table}{Either a path to a CSV file, or, a data frame object.}

\item{from_col}{The name of the table column from which edges originate.}

\item{to_col}{The name of the table column to which edges terminate.}

\item{from_to_map}{A single character value for the mapping of the \code{from} and
\code{to} columns in the external table (supplied as \code{from_col} and \code{to_col},
respectively) to a column in the graph's internal node data frame (ndf).}

\item{rel_col}{An option to apply a column of data in the table as \code{rel}
attribute values.}

\item{set_rel}{an optional string to apply a \code{rel} attribute to all edges
created from the table records.}

\item{drop_cols}{An optional column selection statement for dropping columns
from the external table before inclusion as attributes in the graph's
internal edge data frame. Several columns can be dropped by name using the
syntax \code{col_1 & col_2 & ...}. Columns can also be dropped using a numeric
column range with \code{:} (e.g., \code{5:8}), or, by using the \code{:} between column
names to specify the range (e.g., \code{col_5_name:col_8_name}).}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add edges and their attributes to an existing graph object from data in a CSV
file or a data frame.
}
\examples{
# Create an empty graph and then
# add nodes to it from the
# `currencies` dataset available
# in the package
graph <-
  create_graph() \%>\%
  add_nodes_from_table(
    table = currencies)

# Now we want to add edges to the
# graph using an included dataset,
# `usd_exchange_rates`, which has
# exchange rates between USD and
# many other currencies; the key
# here is that the data in the
# `from` and `to` columns in the
# external table maps to graph
# node data available in the
# `iso_4217_code` column of the
# graph's internal node data frame
graph_1 <-
  graph \%>\%
    add_edges_from_table(
      table = usd_exchange_rates,
      from_col = from_currency,
      to_col = to_currency,
      from_to_map = iso_4217_code)

# View part of the graph's
# internal edge data frame
graph_1 \%>\%
  get_edge_df() \%>\%
  head()

# If you would like to assign
# any of the table's columns as the
# `rel` attribute, this can done
# with the `rel_col` argument; to
# set a static `rel` attribute for
# all edges created, use `set_rel`
graph_2 <-
  graph \%>\%
    add_edges_from_table(
      table = usd_exchange_rates,
      from_col = from_currency,
      to_col = to_currency,
      from_to_map = iso_4217_code,
      set_rel = "from_usd")

# View part of the graph's internal
# edge data frame (edf)
graph_2 \%>\%
  get_edge_df() \%>\%
  head()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_graph_time.R
\name{set_graph_time}
\alias{set_graph_time}
\title{Set graph date-time and timezone}
\usage{
set_graph_time(graph, time = NULL, tz = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{time}{The date-time to set for the graph.}

\item{tz}{The timezone to set for the graph.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Set the time and timezone for a graph object of class \code{dgr_graph}.
}
\examples{
# Create an empty graph
graph <- create_graph()

# Provide the new graph with a timestamp (if `tz`
# is not supplied, `GMT` is used as the time zone)
graph_1 <-
  graph |>
  set_graph_time(time = "2015-10-25 15:23:00")

# Provide the new graph with a timestamp that is
# the current time; the time zone is inferred from
# the user's locale
graph_2 <-
  graph |>
  set_graph_time()

# The time zone can be updated when a timestamp
# is present
graph_2 <-
  graph_2 |>
  set_graph_time(tz = "America/Los_Angeles")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_similar_nbrs.R
\name{get_similar_nbrs}
\alias{get_similar_nbrs}
\title{Get neighboring nodes based on node attribute similarity}
\usage{
get_similar_nbrs(graph, node, node_attr, tol_abs = NULL, tol_pct = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{A single-length vector containing a node ID value.}

\item{node_attr}{The name of the node attribute to use to compare with
adjacent nodes.}

\item{tol_abs}{If the values contained in the node attribute \code{node_attr} are
numeric, one can optionally supply a numeric vector of length 2 that
provides a lower and upper numeric bound as criteria for neighboring node
similarity to the starting node.}

\item{tol_pct}{If the values contained in the node attribute \code{node_attr} are
numeric, one can optionally supply a numeric vector of length 2 that
specifies lower and upper bounds as negative and positive percentage
changes to the value of the starting node. These bounds serve as criteria
for neighboring node similarity to the starting node.}
}
\value{
A vector of node ID values.
}
\description{
With a graph a single node serving as the starting point, get those nodes in
a potential neighborhood of nodes (adjacent to the starting node) that have a
common or similar (within threshold values) node attribute to the starting
node.
}
\examples{
# Getting similar neighbors can
# be done through numerical comparisons;
# start by creating a random, directed
# graph with 18 nodes and 22 edges
# using the `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 18,
    m = 25,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = value,
    values = rnorm(
      n = 18,
      mean = 5,
      sd = 1) |> round(0))

# Starting with node `10`, we
# can test whether any nodes
# adjacent and beyond are
# numerically equivalent in `value`
graph |>
  get_similar_nbrs(
    node = 10,
    node_attr = value)

# We can also set a tolerance
# for ascribing similarly by using
# either the `tol_abs` or `tol_pct`
# arguments (the first applies absolute
# lower and upper bounds from the
# value in the starting node and the
# latter uses a percentage difference
# to do the same); try setting `tol_abs`
# with a fairly large range to
# determine if several nodes can be
# selected
graph |>
  get_similar_nbrs(
    node = 10,
    node_attr = value,
    tol_abs = c(1, 1))

# That resulted in a fairly large
# set of 4 neigboring nodes; for
# sake of example, setting the range
# to be very large will effectively
# return all nodes in the graph
# except for the starting node
graph |>
  get_similar_nbrs(
    node = 10,
    node_attr = value,
    tol_abs = c(10, 10)) |>
    length()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_radiality.R
\name{get_radiality}
\alias{get_radiality}
\title{Get radiality centrality scores}
\usage{
get_radiality(graph, direction = "all")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{direction}{Using \code{all} (the default), the search will ignore edge
direction while traversing through the graph. With \code{out}, measurements of
paths will be from a node whereas with \verb{in}, measurements of paths will be
to a node.}
}
\value{
A data frame with radiality centrality scores for each of the nodes.
}
\description{
Get the radiality centrality for all nodes in a graph. These scores describe
the ease to which nodes can reach other nodes.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the radiality scores for nodes in the graph
graph |>
  get_radiality()

# Add the radiality values
# to the graph as a node
# attribute
graph <-
  graph |>
  join_node_attrs(
    df = get_radiality(graph))

# Display the graph's node data frame
graph |>
  get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_to_min_spanning_tree.R
\name{transform_to_min_spanning_tree}
\alias{transform_to_min_spanning_tree}
\title{Get a minimum spanning tree subgraph}
\usage{
transform_to_min_spanning_tree(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Get a minimum spanning tree subgraph for a connected graph of class
\code{dgr_graph}.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Obtain Jaccard similarity
# values for each pair of
# nodes as a square matrix
j_sim_matrix <-
  graph \%>\%
    get_jaccard_similarity()

# Create a weighted, undirected
# graph from the resultant matrix
# (effectively treating that
# matrix as an adjacency matrix)
graph <-
  j_sim_matrix \%>\%
  from_adj_matrix(weighted = TRUE)

# The graph in this case is a fully connected graph
# with loops, where jaccard similarity values are
# assigned as edge weights (edge attribute `weight`);
# The minimum spanning tree for this graph is the
# connected subgraph where the edges retained have
# the lowest similarity values possible
min_spanning_tree_graph <-
  graph \%>\%
  transform_to_min_spanning_tree() \%>\%
  copy_edge_attrs(
    edge_attr_from = weight,
    edge_attr_to = label) \%>\%
  set_edge_attrs(
    edge_attr = fontname,
    values = "Helvetica") \%>\%
  set_edge_attrs(
    edge_attr = color,
    values = "gray85") \%>\%
  rescale_edge_attrs(
    edge_attr_from = weight,
    to_lower_bound = 0.5,
    to_upper_bound = 4.0,
      edge_attr_to = penwidth)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_betweenness.R
\name{get_betweenness}
\alias{get_betweenness}
\title{Get betweenness centrality scores}
\usage{
get_betweenness(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
a data frame with betweenness scores for each of the nodes.
}
\description{
Get the betweenness centrality scores for all nodes in a graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 12,
    set_seed = 23)

# Get the betweenness scores
# for nodes in the graph
graph |> get_betweenness()

# Add the betweenness
# values to the graph
# as a node attribute
graph <-
  graph |>
  join_node_attrs(
    df = get_betweenness(graph))

# Display the graph's node
# data frame
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/node_data.R
\name{node_data}
\alias{node_data}
\title{Insert node data attributes during node creation}
\usage{
node_data(...)
}
\arguments{
\item{...}{Node data attributes provided as one or more named vectors.}
}
\description{
This helper function should be invoked to provide values for the namesake
\code{node_data} argument, which is present in any function where nodes are
created.
}
\examples{
# Create a new graph and add
# a path with several node
# data attributes
graph <-
  create_graph() \%>\%
  add_path(
    n = 3,
    type = "path",
    node_data = node_data(
      hour = 5,
      index = c(1, 3, 2)))

# View the graph's internal
# node data frame; the node
# data attributes have been
# inserted
graph \%>\% get_node_df()
}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_node.R
\name{delete_node}
\alias{delete_node}
\title{Delete a node from an existing graph object}
\usage{
delete_node(graph, node)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{A node ID for the node to be deleted from the graph.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph}, delete an existing node by
specifying its node ID.
}
\examples{
# Create a graph with 5 nodes and
# edges between each in a path
graph <-
  create_graph() \%>\%
  add_path(n = 5)

# Delete node with ID `3`
graph <- delete_node(graph, node = 3)

# Verify that the node with ID `3`
# is no longer in the graph
graph \%>\% get_node_ids()

# Also note that edges are removed
# since there were edges between the
# removed node to and from other nodes
graph \%>\% get_edges()
}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_w_connected_cmpts.R
\name{count_w_connected_cmpts}
\alias{count_w_connected_cmpts}
\title{Get the number of weakly-connected components}
\usage{
count_w_connected_cmpts(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single integer value representing the number of weakly-connected
graph components.
}
\description{
Get the number of weakly-connected components in the graph.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() |>
  add_cycle(n = 5) |>
  add_cycle(n = 5)

# Get a count of weakly-connected
# components in the graph
graph |> count_w_connected_cmpts()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_graph.R
\name{render_graph}
\alias{render_graph}
\title{Render the graph in various formats}
\usage{
render_graph(
  graph,
  layout = NULL,
  output = NULL,
  as_svg = FALSE,
  title = NULL,
  width = NULL,
  height = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{layout}{A string specifying a layout type to use for node placement in
this rendering. Possible layouts include: \code{nicely}, \code{circle}, \code{tree}, \code{kk},
and \code{fr}.}

\item{output}{A string specifying the output type; \code{graph} (the default)
renders the graph using the \code{\link[=grViz]{grViz()}} function and \code{visNetwork} renders the
graph using the \code{\link[=visnetwork]{visnetwork()}} function.}

\item{as_svg}{An option to render the graph as an SVG document.}

\item{title}{An optional title for a graph when using \code{output = "graph"}.}

\item{width}{An optional parameter for specifying the width of the resulting
graphic in pixels.}

\item{height}{An optional parameter for specifying the height of the
resulting graphic in pixels.}
}
\description{
Using a \code{dgr_graph} object, render the graph in the RStudio Viewer.
}
\examples{
if (interactive()) {

  # Render a graph that's a
  # balanced tree
  create_graph() \%>\%
    add_balanced_tree(
      k = 2, h = 3
    ) \%>\%
    render_graph()

  # Use the `tree` layout for
  # better node placement in this
  # hierarchical graph
  create_graph() \%>\%
    add_balanced_tree(
      k = 2, h = 3
    ) \%>\%
    render_graph(layout = "tree")

  # Plot the same tree graph but
  # don't show the node ID values
  create_graph() \%>\%
    add_balanced_tree(
      k = 2, h = 3
    ) \%>\%
    set_node_attr_to_display() \%>\%
    render_graph(layout = "tree")

  # Create a circle graph
  create_graph() \%>\%
    add_gnm_graph(
      n = 55,
      m = 75,
      set_seed = 23
    ) \%>\%
    render_graph(
      layout = "circle"
    )

  # Render the graph using the
  # `visNetwork` output option
  create_graph() \%>\%
    add_balanced_tree(
      k = 2, h = 3
    ) \%>\%
    render_graph(
      output = "visNetwork"
    )
}

}
\seealso{
Other Display and Save: 
\code{\link{export_graph}()},
\code{\link{render_graph_from_graph_series}()},
\code{\link{save_graph}()}
}
\concept{Display and Save}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_node_ids.R
\name{get_node_ids}
\alias{get_node_ids}
\title{Get a vector of node ID values}
\usage{
get_node_ids(graph, conditions = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the retrieval of
nodes.}
}
\value{
A vector of node ID values.
}
\description{
Obtain a vector of node ID values from a graph object. An optional filter by
node attribute can limit the set of node ID values returned.
}
\examples{
# Create a node data
# frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = "letter",
    color = c(
      "red", "green",
      "blue", "blue"),
    value = c(
      3.5, 2.6, 9.4, 2.7))

# Create a graph using
# the ndf
graph <-
  create_graph(
    nodes_df = ndf)

# Get a vector of all nodes in a graph
graph |> get_node_ids()

# Get a vector of node ID values using a
# numeric comparison (i.e., all nodes with
# `value` attribute greater than 3)
graph |>
  get_node_ids(
    conditions = value > 3)

# Get a vector of node ID values using
# a match pattern (i.e., all nodes with
# `color` attribute of `green`)
graph |>
  get_node_ids(
    conditions = color == "green")

# Use multiple conditions to return nodes
# with the desired attribute values
graph |>
  get_node_ids(
    conditions =
      color == "blue" &
      value > 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_adj_matrix.R
\name{from_adj_matrix}
\alias{from_adj_matrix}
\title{Create a graph using an adjacency matrix}
\usage{
from_adj_matrix(
  x,
  mode = "undirected",
  weighted = FALSE,
  use_diag = TRUE,
  graph_name = NULL,
  write_backups = FALSE,
  display_msgs = FALSE
)
}
\arguments{
\item{x}{A square \code{matrix} object serving as the adjacency matrix.}

\item{mode}{The method in which to interpret the input adjacency matrix.
Options include: \code{undirected}, \code{directed}, \code{upper}, \code{lower}, \code{max}, \code{min},
and \code{plus}.}

\item{weighted}{Whether to create a weighted graph from the adjacency matrix.}

\item{use_diag}{Whether to use the diagonal of the adjacency matrix in
calculations. If \code{TRUE} then the diagonal values will be included as is. If
\code{FALSE} then the diagonal values will be replaced with zero values before
inclusion in any calculations.}

\item{graph_name}{An optional string for labeling the graph object.}

\item{write_backups}{An option to write incremental backups of changing graph
states to disk. If \code{TRUE}, a subdirectory within the working directory will
be created and used to store \code{RDS} files. The default value is \code{FALSE} so
one has to opt in to use this functionality.}

\item{display_msgs}{An option to display messages primarily concerned with
changes in graph selections. By default, this is \code{FALSE}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Using an adjacency matrix object, generate a graph of class \code{dgr_graph}.
}
\examples{
# Create an adjacency matrix
adj_matrix <-
  sample(
    0:1, 100,
    replace = TRUE,
    prob = c(0.9,0.1)
  ) |>
  matrix(ncol = 10)

# Create a graph from the adjacency matrix
graph <- from_adj_matrix(adj_matrix)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename_node_attrs.R
\name{rename_node_attrs}
\alias{rename_node_attrs}
\title{Rename a node attribute}
\usage{
rename_node_attrs(graph, node_attr_from, node_attr_to)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr_from}{The name of the node attribute that will be renamed.}

\item{node_attr_to}{The new name of the node attribute column identified in
\code{node_attr_from}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal node data frame (ndf), rename an existing node
attribute.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 5,
    m = 8,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = shape,
    values = "circle") |>
  set_node_attrs(
    node_attr = value,
    values = rnorm(
      n = 5,
      mean = 5,
      sd = 1) |> round(1))

# Get the graph's internal ndf
# to show which node attributes
# are available
graph |> get_node_df()

# Rename the `value` node
# attribute as `weight`
graph <-
  graph |>
  rename_node_attrs(
    node_attr_from = value,
    node_attr_to = weight)

# Get the graph's internal
# ndf to show that the node
# attribute had been renamed
graph |> get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_automorphisms.R
\name{count_automorphisms}
\alias{count_automorphisms}
\title{Get the number of automorphisms}
\usage{
count_automorphisms(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single numeric value representing the number of automorphisms the
graph contains.
}
\description{
Get the number of automorphisms the graph contains. An automorphism of a
graph is a form of symmetry in which the graph is mapped onto itself while
preserving edge-node connectivity.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() \%>\%
  add_cycle(n = 5)

# Get a count of automorphisms
graph \%>\%
  count_automorphisms()

# Create a full graph and then
# count the automorphisms
create_graph() \%>\%
  add_full_graph(n = 10) \%>\%
  count_automorphisms()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_nodes_ws.R
\name{delete_nodes_ws}
\alias{delete_nodes_ws}
\title{Delete all nodes in a node selection}
\usage{
delete_nodes_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
In a graph object of class \code{dgr_graph}, delete all nodes present in a
selection of nodes.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create a graph with 3 nodes
graph <-
  create_graph() |>
  add_n_nodes(n = 3) |>
  add_edges_w_string(
    edges = "1->3 1->2 2->3")

# Select node with ID `1`
graph <-
  graph |>
  select_nodes_by_id(nodes = 1)

# Delete node in selection (this
# also deletes any attached edges)
graph <-
  graph |>
  delete_nodes_ws()

# Get a count of nodes in the graph
graph |> count_nodes()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_girth.R
\name{get_girth}
\alias{get_girth}
\title{Get the graph girth}
\usage{
get_girth(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single numeric value representing the length of the shortest circle
in the graph.
}
\description{
Get the girth of a graph, which is the length of the shortest circle in the
graph. Loop edges and multiple edges are not considered. If the graph
contains no cycles then zero is returned.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() |>
  add_cycle(n = 5)

# Determine the graph's girth
graph |> get_girth()

# Create a full graph and then
# get the girth for that
create_graph() |>
  add_full_graph(n = 10) |>
  get_girth()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edge_info.R
\name{get_edge_info}
\alias{get_edge_info}
\title{Get detailed information on edges}
\usage{
get_edge_info(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame containing information specific to each edge within the
graph.
}
\description{
Obtain a data frame with detailed information on edges and their
interrelationships within the graph.
}
\examples{
# Create a simple graph
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 5, m = 10,
    set_seed = 23)

# Get information on the
# graph's edges
graph |> get_edge_info()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_edge_df.R
\name{create_edge_df}
\alias{create_edge_df}
\title{Create an edge data frame}
\usage{
create_edge_df(from, to, rel = NULL, ...)
}
\arguments{
\item{from}{A vector of node ID values from which edges are outbound. The
vector length must equal that of the \code{to} vector.}

\item{to}{A vector of node ID values to which edges are incoming. The vector
length must equal that of the \code{from} vector.}

\item{rel}{An optional \code{rel} label for each edge.}

\item{...}{One or more vectors for associated edge attributes. Can be some of \code{\link[=edge_aes]{edge_aes()}}}
}
\value{
An edge data frame (edf).
}
\description{
Combine several vectors for edges and their attributes into a data frame,
which can be combined with other similarly-generated data frames, or, added
to a graph object. An edge data frame, or edf, has at least the following
columns:
\itemize{
\item \code{id} (of type \code{integer})
\item \code{from} (of type \code{integer})
\item \code{to} (of type \code{integer})
\item \code{rel} (of type \code{character})
}

An arbitrary number of additional columns containing aesthetic or data
attributes can be part of the edf, so long as they follow the aforementioned
columns. Some examples are included in \code{\link[=edge_aes]{edge_aes()}}
}
\examples{
# Create a simple edge data frame (edf) and
# view the results
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "a")

# Display the edge data frame
edf

# Create an edf with additional edge
# attributes (where their classes will
# be inferred from the input vectors)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "a",
    length = c(50, 100, 250),
    color = "green",
    width = c(1, 5, 2))

# Display the edge data frame
edf

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_node_attrs_ws.R
\name{set_node_attrs_ws}
\alias{set_node_attrs_ws}
\title{Set node attributes with a node selection}
\usage{
set_node_attrs_ws(graph, node_attr, value)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr}{The name of the attribute to set.}

\item{value}{The value to be set for the chosen attribute for the nodes in
the current selection.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} or a node data frame, set node
attribute properties for nodes present in a node selection.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create a simple graph
graph <-
  create_graph() \%>\%
  add_path(n = 6)

# Select specific nodes from the graph and
# apply the node attribute `color = blue` to
# those selected nodes
graph <-
  graph \%>\%
  select_nodes_by_id(
    nodes = 1:4) \%>\%
  trav_out() \%>\%
  set_node_attrs_ws(
    node_attr = color,
    value = "blue")

# Show the internal node data frame to verify
# that the node attribute has been set for
# specific node
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_node_attrs.R
\name{mutate_node_attrs}
\alias{mutate_node_attrs}
\title{Mutate a set of node attribute values}
\usage{
mutate_node_attrs(graph, ...)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{...}{Expressions used for the mutation of node attributes. LHS of each
expression is either an existing or new node attribute name. The RHS can
consist of any valid R code that uses node attributes as variables.
Expressions are evaluated in the order provided, so, node attributes
created or modified are ready to use in subsequent expressions.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal node data frame (ndf), mutate numeric node
attribute values using one or more expressions.
}
\examples{
# Create a graph with 3 nodes
graph <-
  create_graph() |>
  add_path(n = 3) |>
  set_node_attrs(
    node_attr = width,
    values = c(1.4, 0.3, 1.1))

# Get the graph's internal ndf
# to show which node attributes
# are available
graph |> get_node_df()

# Mutate the `width` node
# attribute, dividing each
# value by 2
graph <-
  graph |>
  mutate_node_attrs(
    width = width / 2)

# Get the graph's internal
# ndf to show that the node
# attribute `width` had its
# values changed
graph |> get_node_df()

# Create a new node attribute,
# called `length`, that is the
# log of values in `width` plus
# 2 (and, also, round all values
# to 2 decimal places)
graph <-
  graph |>
  mutate_node_attrs(
    length = (log(width) + 2) |>
               round(2))

# Get the graph's internal ndf
# to show that the node attribute
# values had been mutated
graph |> get_node_df()

# Create a new node attribute
# called `area`, which is the
# product of the `width` and
# `length` attributes
graph <-
  graph |>
  mutate_node_attrs(
    area = width * length)

# Get the graph's internal ndf
# to show that the node attribute
# values had been multiplied
# together (with new attr `area`)
graph |> get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_global_graph_attr_info.R
\name{get_global_graph_attr_info}
\alias{get_global_graph_attr_info}
\title{Get global graph attributes}
\usage{
get_global_graph_attr_info(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame containing global attributes for the graph.
}
\description{
Get the available global attributes for a graph object of class \code{dgr_graph}.
}
\examples{
# Create a new, empty graph
graph <- create_graph()

# View the graph's set of
# global attributes
graph \%>\%
  get_global_graph_attr_info()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_edge_attrs.R
\name{mutate_edge_attrs}
\alias{mutate_edge_attrs}
\title{Mutate a set of edge attribute values}
\usage{
mutate_edge_attrs(graph, ...)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{...}{Expressions used for the mutation of edge attributes. LHS of each
expression is either an existing or new edge attribute name. The RHS can
consist of any valid R code that uses edge attributes as variables.
Expressions are evaluated in the order provided, so, edge attributes
created or modified are ready to use in subsequent expressions.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal edge data frame (edf), mutate numeric edge
attribute values using one or more expressions.
}
\examples{
# Create a graph with 3 edges
graph <-
  create_graph() \%>\%
  add_path(n = 4) \%>\%
  set_edge_attrs(
    edge_attr = width,
    values = c(3.4, 2.3, 7.2))

# Get the graph's internal edf
# to show which edge attributes
# are available
graph \%>\% get_edge_df()

# Mutate the `width` edge
# attribute, dividing each
# value by 2
graph <-
  graph \%>\%
  mutate_edge_attrs(
    width = width / 2)

# Get the graph's internal
# edf to show that the edge
# attribute `width` had its
# values changed
graph \%>\% get_edge_df()

# Create a new edge attribute,
# called `length`, that is the
# log of values in `width` plus
# 2 (and, also, round all values
# to 2 decimal places)
graph <-
  graph \%>\%
  mutate_edge_attrs(
    length = (log(width) + 2) \%>\%
               round(2))

# Get the graph's internal edf
# to show that the edge attribute
# values had been mutated
graph \%>\% get_edge_df()

# Create a new edge attribute
# called `area`, which is the
# product of the `width` and
# `length` attributes
graph <-
  graph \%>\%
  mutate_edge_attrs(
    area = width * length)

# Get the graph's internal edf
# to show that the edge attribute
# values had been multiplied
# together (with new attr `area`)
graph \%>\% get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_dice_similarity.R
\name{get_dice_similarity}
\alias{get_dice_similarity}
\title{Get Dice similarity coefficient scores}
\usage{
get_dice_similarity(graph, nodes = NULL, direction = "all", round_to = 3)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{nodes}{an optional vector of node IDs to consider for Dice similarity
scores. If not supplied, then similarity scores will be provided for every
pair of nodes in the graph.}

\item{direction}{using \code{all} (the default), the function will ignore
edge direction when determining scores for neighboring nodes. With
\code{out} and \verb{in}, edge direction for neighboring nodes will be
considered.}

\item{round_to}{the maximum number of decimal places to retain for the Dice
similarity coefficient scores. The default value is \code{3}.}
}
\value{
A matrix with Dice similarity values for each pair of nodes
considered.
}
\description{
Get the Dice similarity coefficient scores for one or more nodes in a graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the Dice similarity
# values for nodes `5`, `6`,
# and `7`
graph \%>\%
  get_dice_similarity(
    nodes = 5:7)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_cmty_walktrap.R
\name{get_cmty_walktrap}
\alias{get_cmty_walktrap}
\title{Get community membership using the Walktrap method}
\usage{
get_cmty_walktrap(graph, steps = 4)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{steps}{the number of steps to take for each of the random walks.}
}
\value{
A data frame with group membership assignments for each of the nodes.
}
\description{
With the Walktrap community finding algorithm, obtain the group membership
values for each of the nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the group membership
# values for all nodes in the
# graph through the Walktrap
# community finding algorithm
graph \%>\%
  get_cmty_walktrap()

# Add the group membership
# values to the graph as a
# node attribute
graph <-
  graph \%>\%
  join_node_attrs(
    df = get_cmty_walktrap(.))

# Display the graph's
# node data frame
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_eccentricity.R
\name{get_eccentricity}
\alias{get_eccentricity}
\title{Get node eccentricities}
\usage{
get_eccentricity(graph, mode = "out")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{mode}{the mode with which the shortest paths to or from the given
vertices should be calculated for directed graphs. If \code{out} (the
default) then the shortest paths from the node, if \verb{in} then only
shortest paths to each node are considered. If \code{all} is used, then the
corresponding undirected graph will be used and edge directions will be
ignored. For undirected graphs, this argument is ignored.}
}
\value{
A data frame containing eccentricity values by node ID value.
}
\description{
Get a data frame with node eccentricity values.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the eccentricity values for
# all nodes in the graph
graph |> get_eccentricity()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_graph_from_graph_series.R
\name{get_graph_from_graph_series}
\alias{get_graph_from_graph_series}
\title{Get a graph available in a series}
\usage{
get_graph_from_graph_series(graph_series, graph_no)
}
\arguments{
\item{graph_series}{A graph series object of type \code{dgr_graph_1D}.}

\item{graph_no}{The index of the graph in the graph series.}
}
\description{
Using a graph series object of type \code{dgr_graph_1D}, get a graph object.
}
\examples{
# Create three graphs
graph_1 <-
  create_graph() \%>\%
  add_path(n = 4)

graph_2 <-
  create_graph() \%>\%
  add_cycle(n = 5)

graph_3 <-
  create_graph() \%>\%
  add_star(n = 6)

# Create an empty graph series
# and add the graphs
series <-
  create_graph_series() \%>\%
  add_graph_to_graph_series(
    graph = graph_1) \%>\%
  add_graph_to_graph_series(
    graph = graph_2) \%>\%
  add_graph_to_graph_series(
    graph = graph_3)

# Get the second graph in the series
extracted_graph <-
  series \%>\%
  get_graph_from_graph_series(
    graph_no = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/copy_edge_attrs.R
\name{copy_edge_attrs}
\alias{copy_edge_attrs}
\title{Copy an edge attribute column and set the name}
\usage{
copy_edge_attrs(graph, edge_attr_from, edge_attr_to)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr_from}{The name of the edge attribute column from which values
will be copied.}

\item{edge_attr_to}{The name of the new edge attribute column to which the
copied values will be placed.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal edge data frame (edf), copy the contents an
existing edge attribute and create a distinct edge attribute within the edf
with a different attribute name.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 5,
    m = 8,
    set_seed = 23) |>
  set_edge_attrs(
    edge_attr = color,
    values = "green")

# Get the graph's internal
# edf to show which edge
# attributes are available
graph |> get_edge_df()

# Make a copy the `color`
# edge attribute as the
# `color_2` edge attribute
graph <-
  graph |>
  copy_edge_attrs(
    edge_attr_from = color,
    edge_attr_to = color_2)

# Get the graph's internal
# edf to show that the edge
# attribute had been copied
graph |> get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_graph_simple.R
\name{is_graph_simple}
\alias{is_graph_simple}
\title{Is the graph a simple graph?}
\usage{
is_graph_simple(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Determine whether the graph is a simple graph. A simple graph is one that
does not contain any loops nor any multiple edges.
}
\examples{
# Create a graph with 2 cycles
graph <-
  create_graph() |>
  add_cycle(n = 4) |>
  add_cycle(n = 3)

# Check if the graph is simple
graph |> is_graph_simple()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/join_edge_attrs.R
\name{join_edge_attrs}
\alias{join_edge_attrs}
\title{Join new edge attribute values using a data frame}
\usage{
join_edge_attrs(graph, df, by_graph = NULL, by_df = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{df}{The data frame to use for joining.}

\item{by_graph}{Optional specification of the column in the graph's internal
edge data frame for the left join. If both \code{by_graph} and \code{by_df} are not
provided, then a natural join will occur if there are columns in the
graph's edf and in \code{df} with identical names.}

\item{by_df}{Optional specification of the column in \code{df} for the left join.
If both \code{by_graph} and \code{by_df} are not provided, then a natural join will
occur if there are columns in the graph's edf and in \code{df} with identical
names.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Join new edge attribute values in a left join using a data frame. The use of
a left join in this function allows for no possibility that edges in the
graph might be removed after the join.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a simple graph
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 5) \%>\%
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->5")

# Create a data frame with node ID values
# representing the graph edges (with `from` and `to`
# columns), and, a set of numeric values
df <-
  data.frame(
    from = c(1, 1, 2, 2, 3),
    to = c(2, 3, 4, 5, 5),
    values = rnorm(5, 5))

# Join the values in the data frame to the
# graph's edges; this works as a left join using
# identically-named columns in the graph and the df
# (in this case `from` and `to` are common to both)
graph <-
  graph \%>\%
  join_edge_attrs(
    df = df)

# Get the graph's internal edf to show that the
# join has been made
graph \%>\% get_edge_df()
}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_graph_action.R
\name{add_graph_action}
\alias{add_graph_action}
\title{Add a graph action for execution at every transform}
\usage{
add_graph_action(graph, fcn, ..., action_name = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{fcn}{The name of the function to use.}

\item{...}{Arguments and values to pass to the named function in \code{fcn}, if
necessary.}

\item{action_name}{An optional name for labeling the action.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add a graph function along with its arguments to be run at every graph
transformation step.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 22,
    set_seed = 23)

# Add a graph action that sets a node
# attr column with a function; the
# main function `set_node_attr_w_fcn()`
# uses the `get_betweenness()` function
# to provide betweenness values in the
# `btwns` column; this action will
# occur whenever there is a function
# called on the graph that modifies it
# (e.g., `add_n_nodes()`)
graph <-
  graph |>
  add_graph_action(
    fcn = "set_node_attr_w_fcn",
    node_attr_fcn = "get_betweenness",
    column_name = "btwns",
    action_name = "get_btwns")

# To ensure that the action is
# available in the graph, use the
# `get_graph_actions()` function
graph |> get_graph_actions()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_last_edges_created.R
\name{get_last_edges_created}
\alias{get_last_edges_created}
\title{Get the last set of edges created in a graph}
\usage{
get_last_edges_created(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A vector of edge ID values.
}
\description{
Get the last edges that were created in a graph object of class \code{dgr_graph}.
This function should ideally be used just after creating the edges.
}
\examples{
# Create a graph and add a cycle and then
# a tree in 2 separate function calls
graph <-
  create_graph() \%>\%
  add_cycle(
    n = 3,
    rel = "a") \%>\%
  add_balanced_tree(
    k = 2, h = 2,
    rel = "b")

# Get the last edges created (all edges
# from the tree)
graph \%>\% get_last_edges_created()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_graph_name.R
\name{set_graph_name}
\alias{set_graph_name}
\title{Set graph name}
\usage{
set_graph_name(graph, name)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{name}{The name to set for the graph.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Set a name for a graph object of class \code{dgr_graph}.
}
\examples{
# Create an empty graph
graph <- create_graph()

# Provide the new graph with a name
graph <-
  graph \%>\%
  set_graph_name(
    name = "example_name")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fully_disconnect_nodes_ws.R
\name{fully_disconnect_nodes_ws}
\alias{fully_disconnect_nodes_ws}
\title{Fully disconnect all nodes in a selection of nodes}
\usage{
fully_disconnect_nodes_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a selection of nodes in a graph, remove any edges to or from those
nodes.
}
\details{
This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create an empty graph and
# add a path of 6 nodes
graph <-
  create_graph() |>
  add_path(n = 6)

# Select nodes `3` and `4`
# and fully disconnect them
# from the graph
graph <-
  graph |>
  select_nodes_by_id(
    nodes = 3:4) |>
  fully_disconnect_nodes_ws()

# Get the graph's edge data frame
graph |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_s_connected_cmpts.R
\name{count_s_connected_cmpts}
\alias{count_s_connected_cmpts}
\title{Get the number of strongly-connected components}
\usage{
count_s_connected_cmpts(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single integer value representing the number of strongly-connected
graph components.
}
\description{
Get the number of strongly-connected components in the graph.
}
\examples{
# Create a graph and add
# several graph islands
graph <-
  create_graph() |>
  add_islands_graph(
    n_islands = 4,
    island_size = 10,
    p = 1/5,
    edges_between = 1,
    set_seed = 23)

# Get a count of strongly-connected
# components in the graph
graph |> count_s_connected_cmpts()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_degree_distribution.R
\name{get_degree_distribution}
\alias{get_degree_distribution}
\title{Get total degree distribution data for a graph}
\usage{
get_degree_distribution(graph, mode = "total")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{mode}{using \code{total} (the default), degree considered for each node
will be the total degree. With \verb{in} and \code{out} the degree used
will be the in-degree and out-degree, respectively.}
}
\value{
A data frame with degree frequencies.
}
\description{
Get degree distribution data for a graph. Graph degree is represented as a
frequency of total degree values over all nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the total degree
# distribution for the graph
graph |>
  get_degree_distribution(
    mode = "total")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/export_graph.R
\name{export_graph}
\alias{export_graph}
\title{Export a graph to various image formats}
\usage{
export_graph(
  graph,
  file_name = NULL,
  file_type = NULL,
  title = NULL,
  width = NULL,
  height = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{file_name}{The name of the exported file (including it's extension).}

\item{file_type}{The type of file to be exported. Options for graph files
are: \code{png}, \code{pdf}, \code{svg}, and \code{ps}.}

\item{title}{An optional title for the output graph.}

\item{width}{Output width in pixels or \code{NULL} for default. Only useful for
export to image file formats \code{png}, \code{pdf}, \code{svg}, and \code{ps}.}

\item{height}{Output height in pixels or \code{NULL} for default. Only useful for
export to image file formats \code{png}, \code{pdf}, \code{svg}, and \code{ps}.}
}
\description{
Export a graph to a variety of image formats such as PNG, PDF, SVG, and
PostScript.
}
\examples{
# Create a simple graph
graph <-
  create_graph() |>
    add_path(
      n = 5,
      edge_aes = edge_aes(
        arrowhead = c(
          "normal", "vee",
          "tee", "dot"
        ),
        color = c(
        "red", "blue",
        "orange", "purple"
        )
      )
    )

# Create a PDF file for
# the graph (`graph.pdf`)
# graph |>
#   export_graph(
#     file_name = "graph.pdf",
#     title = "Simple Graph"
#   )

# Create a PNG file for
# the graph (`mypng.png`)
# graph |>
#   export_graph(
#     file_name = "mypng.png",
#     file_type = "PNG"
#   )

}
\seealso{
Other Display and Save: 
\code{\link{render_graph}()},
\code{\link{render_graph_from_graph_series}()},
\code{\link{save_graph}()}
}
\concept{Display and Save}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_edges_by_node_id.R
\name{select_edges_by_node_id}
\alias{select_edges_by_node_id}
\title{Select edges in a graph using node ID values}
\usage{
select_edges_by_node_id(graph, nodes, set_op = "union")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{nodes}{A vector of node IDs for the selection of edges present in the
graph.}

\item{set_op}{The set operation to perform upon consecutive selections of
graph edges This can either be as a \code{union} (the default), as an
intersection of selections with \code{intersect}, or, as a \code{difference} on the
previous selection, if it exists.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Select edges in a graph object of class \code{dgr_graph} using node ID values. All
edges associated with the provided nodes will be included in the selection.
}
\examples{
# Create a graph with 5 nodes
graph <-
  create_graph() |>
  add_path(n = 5)

# Create a graph selection by selecting edges
# associated with nodes `1` and `2`
graph <-
  graph |>
  select_edges_by_node_id(
    nodes = 1:2)

# Get the selection of edges
graph |> get_selection()

# Perform another selection of edges, with nodes
# `1`, `2`, and `4`
graph <-
  graph |>
  clear_selection() |>
  select_edges_by_node_id(
    nodes = c(1, 2, 4))

# Get the selection of edges
graph |> get_selection()

# Get a fraction of the edges selected over all
# the edges in the graph
l <- graph |>
  get_selection() |>
  length()

e <- graph |> count_edges()

l/e

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_graph_series_info.R
\name{get_graph_series_info}
\alias{get_graph_series_info}
\title{Get information on a graph series}
\usage{
get_graph_series_info(graph_series)
}
\arguments{
\item{graph_series}{A graph series object of type \code{dgr_graph_1D}.}
}
\value{
A data frame containing information on the graphs within the supplied
graph series.
}
\description{
Obtain a data frame with information on the graphs within a graph series.
}
\examples{
# Create three graphs
graph_1 <-
  create_graph() \%>\%
  add_path(n = 4)

graph_2 <-
  create_graph() \%>\%
  add_cycle(n = 5)

graph_3 <-
  create_graph() \%>\%
  add_star(n = 6)

# Create an empty graph series
# and add the graphs
series <-
  create_graph_series() \%>\%
  add_graph_to_graph_series(
    graph = graph_1) \%>\%
  add_graph_to_graph_series(
    graph = graph_2) \%>\%
  add_graph_to_graph_series(
    graph = graph_3)

# Get information on the graphs in the series
series \%>\% get_graph_series_info()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_successors.R
\name{get_successors}
\alias{get_successors}
\title{Get node IDs for successor nodes to the specified node}
\usage{
get_successors(graph, node)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{A node ID for the selected node.}
}
\value{
A vector of node ID values.
}
\description{
Provides a vector of node IDs for all nodes that have a connection from the
given node.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a node data frame (ndf)
ndf <- create_node_df(n = 26)

# Create an edge data
# frame (edf)
edf <-
  create_edge_df(
    from = sample(
      1:26, replace = TRUE),
    to = sample(
      1:26, replace = TRUE))

# From the ndf and edf,
# create a graph object
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Get sucessors for node
# `4` in the graph
graph |>
  get_successors(
    node = 4)

# If there are no successors,
# NA is returned
graph |>
  get_successors(
    node = 1)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_agg_degree_in.R
\name{get_agg_degree_in}
\alias{get_agg_degree_in}
\title{Get an aggregate value from the indegree of nodes}
\usage{
get_agg_degree_in(graph, agg, conditions = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{agg}{The aggregation function to use for summarizing indegree values
from graph nodes. The following aggregation functions can be used: \code{sum},
\code{min}, \code{max}, \code{mean}, or \code{median}.}

\item{conditions}{An option to use filtering conditions for the nodes to
consider.}
}
\value{
A vector with an aggregate indegree value.
}
\description{
Get a single, aggregate value from the indegree values for all nodes in a
graph, or, a subset of graph nodes.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 20,
    m = 35,
    set_seed = 23) \%>\%
  set_node_attrs(
    node_attr = value,
    values = rnorm(
      n = count_nodes(.),
      mean = 5,
      sd = 1) \%>\% round(1))

# Get the mean indegree value
# from all nodes in the graph
graph \%>\%
  get_agg_degree_in(
    agg = "mean")

# Other aggregation functions
# can be used (`min`, `max`,
# `median`, `sum`); let's get
# the median in this example
graph \%>\%
  get_agg_degree_in(
    agg = "median")

# The aggregation of indegree
# can occur for a subset of the
# graph nodes and this is made
# possible by specifying
# `conditions` for the nodes
graph \%>\%
  get_agg_degree_in(
    agg = "mean",
    conditions = value > 5.0)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_nodes_in_neighborhood.R
\name{select_nodes_in_neighborhood}
\alias{select_nodes_in_neighborhood}
\title{Select nodes based on a walk distance from a specified node}
\usage{
select_nodes_in_neighborhood(graph, node, distance, set_op = "union")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{The node from which the traversal will originate.}

\item{distance}{The maximum number of steps from the \code{node} for inclusion in
the selection.}

\item{set_op}{The set operation to perform upon consecutive selections of
graph nodes. This can either be as a \code{union} (the default), as an
intersection of selections with \code{intersect}, or, as a \code{difference} on the
previous selection, if it exists.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Select those nodes in the neighborhood of nodes connected a specified
distance from an initial node.
}
\examples{
# Create a graph containing
# a balanced tree
graph <-
  create_graph() \%>\%
  add_balanced_tree(
    k = 2, h = 2)

# Create a graph selection by
# selecting nodes in the
# neighborhood of node `1`, where
# the neighborhood is limited by
# nodes that are 1 connection
# away from node `1`
graph <-
  graph \%>\%
  select_nodes_in_neighborhood(
    node = 1,
    distance = 1)

# Get the selection of nodes
graph \%>\% get_selection()

# Perform another selection
# of nodes, this time with a
# neighborhood spanning 2 nodes
# from node `1`
graph <-
  graph \%>\%
  clear_selection() \%>\%
  select_nodes_in_neighborhood(
    node = 1,
    distance = 2)

# Get the selection of nodes
graph \%>\% get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_mutual_node_pairs.R
\name{count_mutual_node_pairs}
\alias{count_mutual_node_pairs}
\title{Get the number of mutually-connected node pairs}
\usage{
count_mutual_node_pairs(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single numeric value representing the number of mutually-connected
node pairs.
}
\description{
Get the number of mutually-connected node pairs. This works for directed
graphs.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() \%>\%
  add_cycle(n = 5)

# Get a count of mutually-connected
# node pairs
graph \%>\% count_mutual_node_pairs()

# Create a full graph and then
# count the mutually-connected
# node pairs
create_graph() \%>\%
  add_full_graph(n = 10) \%>\%
  count_mutual_node_pairs()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_graph_directed.R
\name{is_graph_directed}
\alias{is_graph_directed}
\title{Is the graph a directed graph?}
\usage{
is_graph_directed(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Determines whether a graph is set to be directed or not and returns a logical
value to that effect.
}
\examples{
# Create an empty graph; by default,
# new graphs made by `create_graph()`
# are directed
graph <- create_graph()

# Determine whether the graph
# is directed
graph |> is_graph_directed()

# Use the `set_graph_undirected()`
# function and check again whether
# the graph is directed
graph |>
  set_graph_undirected() |>
  is_graph_directed()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_min_cut_between.R
\name{get_min_cut_between}
\alias{get_min_cut_between}
\title{Get the minimum cut between source and sink nodes}
\usage{
get_min_cut_between(graph, from, to)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{from}{The node ID for the source node.}

\item{to}{The node ID for the sink or target node.}
}
\value{
A single numeric value representing the minimum total edge capacity
removed to disconnect the source and sink nodes.
}
\description{
Get the minimum cut between source and sink nodes. This is the minimum total
capacity of edges needed for removal in order to eliminate all paths from the
source and sink nodes.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a cycle graph
graph <-
  create_graph() |>
  add_cycle(n = 5)

# Determine the minimum cut
# between nodes `1` and `4`
graph |>
  get_min_cut_between(
    from = 1,
    to = 2)

# Create a cycle graph with
# randomized values given to all
# edges as the `capacity` attribute
graph_capacity <-
  create_graph() |>
  add_cycle(n = 5) |>
  select_edges() |>
  set_edge_attrs_ws(
    edge_attr = capacity,
    value =
      rnorm(
        n = 5,
        mean = 5,
        sd = 1)) |>
  clear_selection()

# Determine the minimum cut
# between nodes `1` and `4` for
# this graph, where `capacity`is
# set as an edge attribute
graph_capacity |>
  get_min_cut_between(
    from = 1,
    to = 2)

# Create a full graph and then
# get the minimum cut requirement
# between nodes `2` and `8`
create_graph() |>
  add_full_graph(n = 10) |>
  get_min_cut_between(
    from = 2,
    to = 8)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grViz.R
\name{grViz}
\alias{grViz}
\title{R + viz.js}
\usage{
grViz(
  diagram = "",
  engine = "dot",
  allow_subst = TRUE,
  options = NULL,
  width = NULL,
  height = NULL,
  envir = parent.frame()
)
}
\arguments{
\item{diagram}{Spec for a diagram as either text, filename string, or file
connection.}

\item{engine}{String for the Graphviz layout engine; can be \code{dot} (default),
\code{neato}, \code{circo}, or \code{twopi}.}

\item{allow_subst}{A boolean that enables/disables substitution
functionality.}

\item{options}{Parameters supplied to the htmlwidgets framework.}

\item{width}{An optional parameter for specifying the width of the resulting
graphic in pixels.}

\item{height}{An optional parameter for specifying the height of the
resulting graphic in pixels.}

\item{envir}{The environment in which substitution functionality takes place.}
}
\value{
An object of class \code{htmlwidget} that will intelligently print itself
into HTML in a variety of contexts including the R console, within R
Markdown documents, and within Shiny output bindings.
}
\description{
Make diagrams in R using \code{viz.js} with infrastructure provided by
\pkg{htmlwidgets}.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/do_dfs.R
\name{do_dfs}
\alias{do_dfs}
\title{Use the depth-first search (dfs) algorithm}
\usage{
do_dfs(graph, node = NULL, direction = "all")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{An optional node ID value to specify a single starting point for
the dfs. If not provided, a random node from the graph will be chosen.}

\item{direction}{Using \code{all} (the default), the bfs will ignore edge
direction while traversing through the graph. With \code{out} and \verb{in},
traversals between adjacent nodes will respect the edge direction.}
}
\value{
A vector containing node ID values for nodes visited during the
depth-first search. The order of the node IDs corresponds to the order
visited.
}
\description{
With a chosen or random node serving as the starting point, perform a
depth-first search of the whole graph and return the node ID values visited.
The dfs algorithm differs from breadth-first search (bfs) in that dfs will
follow tree branches as far as possible until terminating at leaf node (bfs
traverses branches one level at a time).
}
\examples{
# Create a graph containing
# two balanced trees
graph <-
  create_graph() |>
  add_balanced_tree(
    k = 2, h = 2) |>
  add_balanced_tree(
  k = 3, h = 2)

# Perform a depth-first
# search of the graph,
# beginning at the root
# node `1` (the default
# `direction = "all"`
# doesn't take edge
# direction into account)
graph |>
  do_dfs(node = 1)

# If not specifying a
# starting node, the function
# will begin the search
# from a random node
graph |>
  do_dfs()

# It's also possible to
# perform dfs while taking
# into account edge direction;
# using `direction = "in"`
# causes the dfs routine to
# visit nodes along inward edges
graph |>
  do_dfs(
    node = 1,
    direction = "in")

# Using `direction = "out"`
# results in the dfs moving
# along solely outward edges
graph |>
  do_dfs(
    node = 1,
    direction = "out")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_nodes_by_degree.R
\name{select_nodes_by_degree}
\alias{select_nodes_by_degree}
\title{Select nodes in the graph based on their degree values}
\usage{
select_nodes_by_degree(graph, expressions, set_op = "union")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{expressions}{One or more expressions for filtering nodes by degree
values. Use a combination of degree type (\code{deg} for total degree, \code{indeg}
for in-degree, and \code{outdeg} for out-degree) with a comparison operator and
values for comparison (e.g., use \code{"deg >= 2"} to select nodes with a degree
greater than or equal to 2).}

\item{set_op}{The set operation to perform upon consecutive selections of
graph nodes. This can either be as a \code{union} (the default), as an
intersection of selections with \code{intersect}, or, as a \code{difference} on the
previous selection, if it exists.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Using a graph object of class \code{dgr_graph}, create a selection of nodes that
have certain degree values.
}
\examples{
# Create a random graph using
# the `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 35, m = 125,
    set_seed = 23)

# Report which nodes have a
# total degree (in-degree +
# out-degree) of exactly 9
graph |>
  select_nodes_by_degree(
    expressions = "deg == 9") |>
  get_selection()

# Report which nodes have a
# total degree greater than or
# equal to 9
graph |>
  select_nodes_by_degree(
    expressions = "deg >= 9") |>
  get_selection()

# Combine two calls of
# `select_nodes_by_degree()` to
# get those nodes with total
# degree less than 3 and total
# degree greater than 10 (by
# default, those `select...()`
# functions will `union` the
# sets of nodes selected)
graph |>
  select_nodes_by_degree(
    expressions = "deg < 3") |>
  select_nodes_by_degree(
    expressions = "deg > 10") |>
  get_selection()

# Combine two calls of
# `select_nodes_by_degree()` to
# get those nodes with total
# degree greater than or equal
# to 3 and less than or equal
# to 10 (the key here is to
# `intersect` the sets of nodes
# selected in the second call)
graph |>
  select_nodes_by_degree(
    expressions = "deg >= 3") |>
  select_nodes_by_degree(
    expressions = "deg <= 10",
    set_op = "intersect") |>
  get_selection()

# Select all nodes with an
# in-degree greater than 5, then,
# apply a node attribute to those
# selected nodes (coloring the
# selected nodes red)
graph_2 <-
  graph |>
  select_nodes_by_degree(
    expressions = "indeg > 5") |>
  set_node_attrs_ws(
    node_attr = color,
    value = "red")

# Get the selection of nodes
graph_2 |> get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_common_nbrs.R
\name{get_common_nbrs}
\alias{get_common_nbrs}
\title{Get all common neighbors between two or more nodes}
\usage{
get_common_nbrs(graph, nodes)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{nodes}{a vector of node ID values of length at least 2.}
}
\value{
a vector of node ID values.
}
\description{
With two or more nodes, get the set of common neighboring nodes.
}
\examples{
# Create a directed graph with 5 nodes
graph <-
  create_graph() \%>\%
  add_path(n = 5)

# Find all common neighbor nodes
# for nodes `1` and `2` (there are no
# common neighbors amongst them)
graph \%>\%
  get_common_nbrs(
    nodes = c(1, 2))

# Find all common neighbor nodes for
# nodes `1` and `3`
graph \%>\%
  get_common_nbrs(
    nodes = c(1, 3))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/open_graph.R
\name{open_graph}
\alias{open_graph}
\title{Read a graph or graph series from disk}
\usage{
open_graph(file)
}
\arguments{
\item{file}{The filename for the graph or graph series. Optionally, this may
contain a path to the file.}
}
\description{
Load a graph or a graph series object from disk.
}
\examples{
# Create an undirected GNP
# graph with 100 nodes using
# a probability value of 0.05
gnp_graph <-
  create_graph(
    directed = FALSE) |>
  add_gnp_graph(
    n = 100,
    p = 0.05
  )

# Save the graph to disk; use
# the file name `gnp_graph.dgr`
# save_graph(
#   x = gnp_graph,
#   file = "gnp_graph"
# )

# To read the graph file from
# disk, use `open_graph()`
# gnp_graph_2 <-
#   open_graph(
#     file = "gnp_graph.dgr"
#   )

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_edge.R
\name{delete_edge}
\alias{delete_edge}
\title{Delete an edge from an existing graph object}
\usage{
delete_edge(graph, from = NULL, to = NULL, id = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{from}{a node ID from which the edge to be removed is outgoing. If an
edge ID is provided to \code{id}, then this argument is ignored. There is
the option to use a node \code{label} value here (and this must
correspondingly also be done for the \code{to} argument) for defining node
connections. Note that this is only possible if all nodes have distinct
\code{label} values set and none exist as an empty string.}

\item{to}{a node ID to which the edge to be removed is incoming. If an edge
ID is provided to \code{id}, then this argument is ignored. There is the
option to use a node \code{label} value here (and this must correspondingly
also be for the \code{from} argument) for defining node connections. Note
that this is only possible if all nodes have distinct \code{label} values
set and none exist as an empty string.}

\item{id}{an edge ID of the edge to be removed.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph}, delete an existing edge by
specifying either: (1) a pair of node IDs corresponding to the edge (keeping
into consideration the direction of the edge in a directed graph), or (2) an
edge ID.
}
\examples{
# Create a graph with 2 nodes
graph <-
  create_graph() |>
  add_n_nodes(n = 2)

# Add an edge
graph <-
  graph |>
  add_edge(
    from = 1,
    to = 2)

# Delete the edge
graph <-
  graph |>
  delete_edge(
    from = 1,
    to = 2)

# Get the count of edges in the graph
graph |> count_edges()

# Create an undirected graph with
# 2 nodes and an edge
graph_undirected <-
  create_graph(directed = FALSE) |>
  add_n_nodes(n = 2) |>
  add_edge(
    from = 1,
    to = 2)

# Delete the edge; the order of node ID
# values provided in `from` and `to`
# don't matter for the undirected case
graph_undirected |>
  delete_edge(
    from = 2,
    to = 1) |>
  count_edges()

# The undirected graph has a single
# edge with ID `1`; it can be
# deleted by specifying `id`
graph_undirected |>
  delete_edge(id = 1) |>
  count_edges()

# Create a directed graph with 2
# labeled nodes and an edge
graph_labeled_nodes <-
  create_graph() |>
  add_n_nodes(
    n = 2,
    label = c("one", "two")) |>
  add_edge(
    from = "one",
    to = "two")

# Delete the edge using the node
# labels in `from` and `to`; this
# is analogous to creating the
# edge using node labels
graph_labeled_nodes |>
  delete_edge(
    from = "one",
    to = "two") |>
  count_edges()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deselect_edges.R
\name{deselect_edges}
\alias{deselect_edges}
\title{Deselect any selected edges in a graph}
\usage{
deselect_edges(graph, edges)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edges}{A vector of edge IDs that should be deselected.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Deselect edges in a graph object of class \code{dgr_graph}.
}
\examples{
# Create a graph with
# a single path
graph <-
  create_graph() |>
  add_path(n = 5)

# Select edges with IDs `1`
# and `3`
graph <-
  graph |>
  select_edges_by_edge_id(
    edges = c(1, 3))

# Verify that an edge selection
# has been made
graph |> get_selection()

# Deselect edge `1`
graph <-
  graph |>
  select_edges_by_edge_id(
    edges = c(1, 3)) |>
  deselect_edges(edges = 1)

# Verify that the edge selection
# has been made for edges `1` and
# `3` and that edge `1` has been
# deselected (leaving only `3`)
graph |> get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trav_in_until.R
\name{trav_in_until}
\alias{trav_in_until}
\title{Traverse inward node-by-node until stopping conditions are met}
\usage{
trav_in_until(
  graph,
  conditions,
  max_steps = 30,
  exclude_unmatched = TRUE,
  add_to_selection = FALSE
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use a stopping condition for the traversal. If
the condition is met during the traversal (i.e., the node(s) traversed to
match the condition), then those traversals will terminate at those nodes.
Otherwise, traversals with continue and terminate when the number of steps
provided in \code{max_steps} is reached.}

\item{max_steps}{The maximum number of \code{trav_in()} steps (i.e., node-to-node
traversals in the inward direction) to allow before stopping.}

\item{exclude_unmatched}{If \code{TRUE} (the default value) then any nodes not
satisfying the conditions provided in \code{conditions} that are in the ending
selection are excluded.}

\item{add_to_selection}{If \code{TRUE} then every node traversed will be part of
the final selection of nodes. If \code{FALSE} (the default value) then only the
nodes finally traversed to will be part of the final node selection.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph}, move along inward edges from one or
more nodes present in a selection to other connected nodes, replacing the
current nodes in the selection with those nodes traversed to until reaching
nodes that satisfy one or more conditions.

This traversal function makes use of an active selection of nodes. After the
traversal, depending on the traversal conditions, there will either be a
selection of nodes or no selection at all.

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create a path graph and add
# values of 1 to 10 across the
# nodes from beginning to end;
# select the last path node
graph <-
  create_graph() \%>\%
  add_path(
    n = 10,
    node_data = node_data(
      value = 1:10)) \%>\%
  select_nodes_by_id(
    nodes = 10)

# Traverse inward, node-by-node
# until stopping at a node where
# the `value` attribute is 1
graph <-
  graph \%>\%
  trav_in_until(
    conditions =
      value == 1)

# Get the graph's node selection
graph \%>\% get_selection()

# Create two cycles in a graph and
# add values of 1 to 6 to the
# first cycle, and values 7 to
# 12 in the second; select nodes
# `6` and `12`
graph <-
  create_graph() \%>\%
  add_cycle(
    n = 6,
    node_data = node_data(
      value = 1:6)) \%>\%
  add_cycle(
    n = 6,
    node_data = node_data(
      value = 7:12)) \%>\%
  select_nodes_by_id(
    nodes = c(6, 12))

# Traverse inward, node-by-node
# from `6` and `12` until stopping
# at the first nodes where the
# `value` attribute is 1, 2, or 10;
# specify that we should only
# keep the finally traversed to
# nodes that satisfy the conditions
graph <-
  graph \%>\%
  trav_in_until(
    conditions =
      value \%in\% c(1, 2, 10),
    exclude_unmatched = TRUE)

# Get the graph's node selection
graph \%>\% get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filter_graph_series.R
\name{filter_graph_series}
\alias{filter_graph_series}
\title{Subset a graph series object}
\usage{
filter_graph_series(graph_series, by = "number", values, tz = NULL)
}
\arguments{
\item{graph_series}{A graph series object of type \code{dgr_graph_1D}.}

\item{by}{Either \code{number}, which allows for subsetting of the graph series by
graph indices, or \code{time} which for graph series objects of type \code{temporal}
allows for a subsetting of graphs by a date-time or time range.}

\item{values}{Where the subsetting of the graph series by to occur via graph
indices (where \code{by = number}), provide a vector of those indices; when
subsetting by time (where \code{by = time}), a range of times can be provided as
a vector.}

\item{tz}{The time zone (\code{tz}) corresponding to dates or date-time string
provided in \code{values} (if \code{by = "date"}).}
}
\value{
A graph series object of type \code{dgr_graph_1D}.
}
\description{
Subsetting a graph series by the graphs' index positions in the graph series
or through selection via graphs' date-time attributes.
}
\examples{
# Create three graphs
graph_time_1 <-
  create_graph(
    graph_name = "graph_with_time_1") \%>\%
  set_graph_time(
    time = "2015-03-25 03:00",
    tz = "GMT")

graph_time_2 <-
  create_graph(
    graph_name = "graph_with_time_2") \%>\%
  set_graph_time(
    time = "2015-03-26 03:00",
    tz = "GMT")

graph_time_3 <-
  create_graph(
    graph_name = "graph_with_time_3") \%>\%
  set_graph_time(
    time = "2015-03-27 15:00",
    tz = "GMT")

# Create an empty graph series and add
# the graphs
series_temporal <-
  create_graph_series(
    series_type = "temporal") \%>\%
  add_graph_to_graph_series(
    graph = graph_time_1) \%>\%
  add_graph_to_graph_series(
    graph = graph_time_2) \%>\%
  add_graph_to_graph_series(
    graph = graph_time_3)

# Subset graph series by sequence
series_sequence_subset <-
  filter_graph_series(
    graph_series = series_temporal,
    by = "number",
    values = 2)

# Get a count of graphs in
# the series
series_sequence_subset \%>\%
  count_graphs_in_graph_series()

# Subset graph series by date-time
series_time_subset <-
  filter_graph_series(
    graph_series = series_temporal,
    by = "time",
    values = c("2015-03-25 12:00",
               "2015-03-26 12:00"),
    tz = "GMT")

# Get a count of graphs in
# the series
series_time_subset \%>\%
  count_graphs_in_graph_series()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_graph_actions.R
\name{delete_graph_actions}
\alias{delete_graph_actions}
\title{Delete one or more graph actions stored within a graph object}
\usage{
delete_graph_actions(graph, actions)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{actions}{Either a vector of integer numbers indicating which actions to
delete (based on \code{action_index} values), or, a character vector
corresponding to \code{action_name} values.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Delete one or more graph actions stored within a graph object of class
\code{dgr_graph}).
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 5,
    m = 8,
    set_seed = 23)

# Add three graph actions to the
# graph
graph <-
  graph \%>\%
  add_graph_action(
    fcn = "set_node_attr_w_fcn",
    node_attr_fcn = "get_pagerank",
    column_name = "pagerank",
    action_name = "get_pagerank") \%>\%
  add_graph_action(
    fcn = "rescale_node_attrs",
    node_attr_from = "pagerank",
    node_attr_to = "width",
    action_name = "pagerank_to_width") \%>\%
  add_graph_action(
    fcn = "colorize_node_attrs",
    node_attr_from = "width",
    node_attr_to = "fillcolor",
    action_name = "pagerank_fillcolor")

# View the graph actions for the graph
# object by using the `get_graph_actions()`
# function
graph \%>\% get_graph_actions()

# Delete the second and third graph
# actions using `delete_graph_actions()`
graph <-
  graph \%>\%
  delete_graph_actions(
    actions = c(2, 3))

# Verify that these last two graph
# actions were deleted by again using
# the `get_graph_actions()` function
graph \%>\% get_graph_actions()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_edge_df.R
\name{add_edge_df}
\alias{add_edge_df}
\title{Add edges from an edge data frame to an existing graph object}
\usage{
add_edge_df(graph, edge_df)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_df}{An edge data frame that is created using \code{\link[=create_edge_df]{create_edge_df()}}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add edges from an edge data frame
to that graph.
}
\examples{
# Create a graph with 4 nodes
# and no edges
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 4)

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
      to = c(4, 3, 1))

# Add the edge data frame to
# the graph object to create
# a graph with both nodes
# and edges
graph <-
  graph \%>\%
  add_edge_df(
    edge_df = edf)

# Get the graph's edges to
# verify that the edf had
# been added
graph \%>\%
  get_edges(
    return_type = "vector")

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rescale_edge_attrs.R
\name{rescale_edge_attrs}
\alias{rescale_edge_attrs}
\title{Rescale numeric edge attribute values}
\usage{
rescale_edge_attrs(
  graph,
  edge_attr_from,
  to_lower_bound = 0,
  to_upper_bound = 1,
  edge_attr_to = NULL,
  from_lower_bound = NULL,
  from_upper_bound = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr_from}{The edge attribute containing numeric data that is to
be rescaled to new numeric or color values.}

\item{to_lower_bound}{The lower bound value for the set of rescaled values.
This can be a numeric value or an X11 color name.}

\item{to_upper_bound}{The upper bound value for the set of rescaled values.
This can be a numeric value or an X11 color name.}

\item{edge_attr_to}{An optional name of a new edge attribute to which the
recoded values will be applied. This will retain the original edge
attribute and its values.}

\item{from_lower_bound}{An optional, manually set lower bound value for the
rescaled values. If not set, the minimum value from the set will be used.}

\item{from_upper_bound}{An optional, manually set upper bound value for the
rescaled values. If not set, the minimum value from the set will be used.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph}, take a set of numeric values for an
edge attribute, rescale to a new numeric or color range, then write to the
same edge attribute or to a new edge attribute column.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 7,
    set_seed = 23) |>
  set_edge_attrs(
    edge_attr = weight,
    values = rnorm(
      n = 7,
      mean = 5,
      sd = 1))

# Get the graph's internal edf
# to show which edge attributes
# are available
graph |> get_edge_df()

# Rescale the `weight` edge
# attribute, so that its values
# are rescaled between 0 and 1
graph <-
  graph |>
  rescale_edge_attrs(
    edge_attr_from = weight,
    to_lower_bound = 0,
    to_upper_bound = 1)

# Get the graph's internal edf
# to show that the edge attribute
# values had been rescaled
graph |> get_edge_df()

# Scale the values in the `weight`
# edge attribute to different
# shades of gray for the `color`
# edge attribute and different
# numerical values for the
# `penwidth` attribute
graph <-
  graph |>
  rescale_edge_attrs(
    edge_attr_from = weight,
    to_lower_bound = "gray80",
    to_upper_bound = "gray20",
    edge_attr_to = color) |>
  rescale_edge_attrs(
    edge_attr_from = weight,
    to_lower_bound = 0.5,
    to_upper_bound = 3,
    edge_attr_to = penwidth)

# Get the graph's internal edf
# once more to show that scaled
# grayscale colors are now available
# in `color` and scaled numerical
# values are in the `penwidth`
# edge attribute
graph |> get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/deselect_nodes.R
\name{deselect_nodes}
\alias{deselect_nodes}
\title{Deselect any selected nodes in a graph}
\usage{
deselect_nodes(graph, nodes)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{nodes}{A vector of node IDs that should be deselected.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Deselect nodes in a graph object of class \code{dgr_graph}.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = c("a", "a", "z", "z"),
    label = TRUE,
    value = c(3.5, 2.6, 9.4, 2.7))

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = c("a", "z", "a"))

# Create a graph with the ndf and edf
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Explicitly select nodes `1` and `3`
graph <-
  graph |>
  select_nodes(nodes = c(1, 3)) |>
  deselect_nodes(nodes = 1)

# Verify that the node selection
# has been made for nodes `1` and
# `3` and that node `1` has been
# deselected (leaving only `3`)
graph |> get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_cmty_edge_btwns.R
\name{get_cmty_edge_btwns}
\alias{get_cmty_edge_btwns}
\title{Get community membership by edge betweenness}
\usage{
get_cmty_edge_btwns(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame with group membership assignments for each of the nodes.
}
\description{
Using edge betweenness, obtain the group membership values for each of the
nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the group membership
# values for all nodes in the
# graph through calculation of
# the leading non-negative
# eigenvector of the modularity
# matrix of the graph
graph |>
  get_cmty_edge_btwns()

# Add the group membership
# values to the graph
# as a node attribute
graph <-
  graph |>
  join_node_attrs(
     df = get_cmty_edge_btwns(graph))

# Display the graph's
# node data frame
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/display_metagraph.R
\name{display_metagraph}
\alias{display_metagraph}
\title{Display a property graph's underlying model}
\usage{
display_metagraph(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}. This graph must fulfill the
condition of being a property graph, otherwise the function yields an
error.}
}
\description{
With a graph object of class \code{dgr_graph} that is also a property graph (i.e.,
all nodes have an assigned \code{type} value and all edges have an assigned \code{rel}
value), display its metagraph in the RStudio Viewer. This representation
provides all combinations of edges of different \code{rel} values to all nodes
with distinct \code{type} values, including any edges to nodes of the same \code{type}
(shown as loops). The precondition of the graph being a property graph can be
verified by using the \code{\link[=is_property_graph]{is_property_graph()}} function.
}
\examples{
# Create a randomized property
# graph with 1000 nodes and 1350 edges
property_graph <-
  create_graph() |>
  add_gnm_graph(
    n = 1000,
    m = 1350,
    set_seed = 23) |>
  select_nodes_by_degree(
    expressions = "deg >= 3") |>
  set_node_attrs_ws(
    node_attr = type,
    value = "a") |>
  clear_selection() |>
  select_nodes_by_degree(
    expressions = "deg < 3") |>
  set_node_attrs_ws(
    node_attr = type,
    value = "b") |>
  clear_selection() |>
  select_nodes_by_degree(
    expressions = "deg == 0") |>
  set_node_attrs_ws(
    node_attr = type,
    value = "c") |>
  set_node_attr_to_display(
    attr = type)

# Select a random subset of edges
# for setting edge attributes
node_ids <- get_node_ids(property_graph)
sampled_nodes <- sample(
  node_ids,
  size = floor(0.15 * length(node_ids)))

property_graph <-
  property_graph |>
  select_edges_by_node_id(
    nodes = sampled_nodes) |>
  set_edge_attrs_ws(
    edge_attr = rel,
    value = "r_1") |>
  invert_selection() |>
  set_edge_attrs_ws(
    edge_attr = rel,
    value = "r_2") |>
  clear_selection() |>
  copy_edge_attrs(
    edge_attr_from = rel,
    edge_attr_to = label) |>
  add_global_graph_attrs(
    attr = "fontname",
    value = "Helvetica",
    attr_type = "edge") |>
  add_global_graph_attrs(
    attr = "fontcolor",
    value = "gray50",
    attr_type = "edge") |>
  add_global_graph_attrs(
    attr = "fontsize",
    value = 10,
    attr_type = "edge")

# Display this graph's
# metagraph, or, the underlying
# graph model for a property graph
# display_metagraph(property_graph)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/do_bfs.R
\name{do_bfs}
\alias{do_bfs}
\title{Use the breadth-first search (bfs) algorithm}
\usage{
do_bfs(graph, node = NULL, direction = "all")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{An optional node ID value to specify a single starting point for
the bfs. If not provided, a random node from the graph will be chosen.}

\item{direction}{Using \code{all} (the default), the bfs will ignore edge
direction while traversing through the graph. With \code{out} and \verb{in},
traversals between adjacent nodes will respect the edge direction.}
}
\value{
A vector containing node ID values for nodes visited during the
breadth-first search. The order of the node IDs corresponds to the order
visited.
}
\description{
With a chosen or random node serving as the starting point, perform a
breadth-first search of the whole graph and return the node ID values
visited. The bfs algorithm differs from depth-first search (dfs) in that bfs
will follow tree branches branches one level at a time until terminating at
leaf node (dfs traverses branches as far as possible).
}
\examples{
# Create a graph containing
# two balanced trees
graph <-
  create_graph() \%>\%
  add_balanced_tree(
    k = 2, h = 2) \%>\%
  add_balanced_tree(
    k = 3, h = 2)

# Perform a breadth-first
# search of the graph,
# beginning at the root node
# `1` (the default
# `direction = "all"` doesn't
# take edge direction into
# account)
graph \%>\%
  do_bfs(node = 1)

# If not specifying a
# starting node, the function
# will begin the search from
# a random node
graph \%>\%
  do_bfs()

# It's also possible to
# perform bfs while taking
# into account edge direction;
# using `direction = "in"`
# causes the bfs routine to
# visit nodes along inward edges
graph \%>\%
  do_bfs(
    node = 1,
    direction = "in")

# Using `direction = "out"`
# results in the bfs moving
# along solely outward edges
graph \%>\%
  do_bfs(
    node = 1,
    direction = "out")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_node_attr_w_fcn.R
\name{set_node_attr_w_fcn}
\alias{set_node_attr_w_fcn}
\title{Set node attribute values with a graph function}
\usage{
set_node_attr_w_fcn(graph, node_attr_fcn, ..., column_name = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr_fcn}{The name of the function to use for creating a column of
node attribute values. Valid functions are: \code{\link[=get_alpha_centrality]{get_alpha_centrality()}},
\code{\link[=get_authority_centrality]{get_authority_centrality()}}, \code{\link[=get_betweenness]{get_betweenness()}}, \code{\link[=get_closeness]{get_closeness()}},
\code{\link[=get_cmty_edge_btwns]{get_cmty_edge_btwns()}}, \code{\link[=get_cmty_fast_greedy]{get_cmty_fast_greedy()}}, \code{\link[=get_cmty_l_eigenvec]{get_cmty_l_eigenvec()}},
\code{\link[=get_cmty_louvain]{get_cmty_louvain()}}, \code{\link[=get_cmty_walktrap]{get_cmty_walktrap()}}, \code{\link[=get_degree_distribution]{get_degree_distribution()}},
\code{\link[=get_degree_histogram]{get_degree_histogram()}}, \code{\link[=get_degree_in]{get_degree_in()}}, \code{\link[=get_degree_out]{get_degree_out()}},
\code{\link[=get_degree_total]{get_degree_total()}}, \code{\link[=get_eccentricity]{get_eccentricity()}}, \code{\link[=get_eigen_centrality]{get_eigen_centrality()}},
\code{\link[=get_pagerank]{get_pagerank()}}, \code{\link[=get_s_connected_cmpts]{get_s_connected_cmpts()}}, and \code{\link[=get_w_connected_cmpts]{get_w_connected_cmpts()}}.}

\item{...}{Arguments and values to pass to the named function in
\code{node_attr_fcn}, if necessary.}

\item{column_name}{An option to supply a column name for the new node
attribute column. If \code{NULL} then the column name supplied by the function
will used along with a \verb{__A} suffix.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} or a node data frame, set node
attribute properties for all nodes in the graph using one of several
whole-graph functions.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 10,
    m = 22,
    set_seed = 23) \%>\%
  set_node_attrs(
    node_attr = value,
    values = rnorm(
      n = count_nodes(.),
      mean = 5,
      sd = 1) \%>\% round(1))

# Get the betweenness values for
# each of the graph's nodes as a
# node attribute
graph_1 <-
  graph \%>\%
  set_node_attr_w_fcn(
    node_attr_fcn = "get_betweenness")

# Inspect the graph's internal
# node data frame
graph_1 \%>\% get_node_df()

# If a specified function takes argument
# values, these can be supplied as well
graph_2 <-
  graph \%>\%
  set_node_attr_w_fcn(
    node_attr_fcn = "get_alpha_centrality",
    alpha = 2,
    exo = 2)

# Inspect the graph's internal
# node data frame
graph_2 \%>\% get_node_df()

# The new column name can be provided
graph_3 <-
  graph \%>\%
  set_node_attr_w_fcn(
    node_attr_fcn = "get_pagerank",
    column_name = "pagerank")

# Inspect the graph's internal
# node data frame
graph_3 \%>\% get_node_df()

# If `graph_3` is modified by
# adding a new node then the column
# `pagerank` will have stale data; we
# can run the function again and re-use
# the existing column name to provide
# updated values
graph_3 <-
  graph_3 \%>\%
  add_node(
    from = 1,
    to = 3) \%>\%
  set_node_attr_w_fcn(
    node_attr_fcn = "get_pagerank",
    column_name = "pagerank")

# Inspect the graph's internal
# node data frame
graph_3 \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_graph_actions.R
\name{get_graph_actions}
\alias{get_graph_actions}
\title{Get information on any available graph actions}
\usage{
get_graph_actions(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A \code{df_tbl} object.
}
\description{
Get a tibble of the available graph actions, which contains information on
function invocations to be called on the graph at every transformation step,
or, when manually invoked with the \code{\link[=trigger_graph_actions]{trigger_graph_actions()}} function.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Add a graph action that sets a node
# attr column with a function; the
# main function `set_node_attr_w_fcn()`
# uses the `get_betweenness()` function
# to provide betweenness values in the
# `btwns` column
graph <-
  graph \%>\%
  add_graph_action(
    fcn = "set_node_attr_w_fcn",
    node_attr_fcn = "get_betweenness",
    column_name = "btwns",
    action_name = "get_btwns")

# To ensure that the action is
# available in the graph, use the
# `get_graph_actions()` function
graph \%>\% get_graph_actions()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DiagrammeR.R
\name{DiagrammeR}
\alias{DiagrammeR}
\title{R + mermaid.js}
\usage{
DiagrammeR(diagram = "", type = "mermaid", ...)
}
\arguments{
\item{diagram}{The diagram in \code{graphviz} or \code{mermaid} format, or, a file (as
a connection or file name) containing a diagram specification. The
recommended filename extensions are \code{.gv} and \code{.mmd} for the Graphviz and
the mermaid diagram specifications, respectively. If no diagram is provided
(\code{diagram = ""}) then the function will assume that a diagram will be
provided by \code{\link[htmltools:builder]{htmltools::tags()}} and \code{DiagrammeR} is just being used for
dependency injection.}

\item{type}{A string, either \code{mermaid} (default) or \code{grViz} indicating the
type of diagram spec and the desired parser/renderer.}

\item{...}{Any other parameters to pass to \code{grViz} or \code{mermaid}}
}
\value{
An object of class \code{htmlwidget} that will intelligently print itself
into HTML in a variety of contexts including the R console, within R
Markdown documents, and within Shiny output bindings.
}
\description{
Make diagrams in R using \href{https://github.com/mdaines/viz-js}{viz.js} or
\href{https://github.com/mermaid-js/mermaid}{mermaid.js} with infrastructure
provided by \href{http://www.htmlwidgets.org/}{htmlwidgets}.
}
\examples{
\dontrun{
# note the whitespace is not important
DiagrammeR("
  graph LR
    A-->B
    A-->C
    C-->E
    B-->D
    C-->D
    D-->F
    E-->F
")

DiagrammeR("
   graph TB
   A-->B
   A-->C
   C-->E
   B-->D
   C-->D
   D-->F
   E-->F
")

DiagrammeR("graph LR;A(Rounded)-->B[Squared];B-->C{A Decision};
 C-->D[Square One];C-->E[Square Two];
 style A fill:#E5E25F;  style B fill:#87AB51; style C fill:#3C8937;
 style D fill:#23772C;  style E fill:#B6E6E6;"
)

# Load in the 'mtcars' dataset
mtcars
connections <- sapply(
 1:ncol(mtcars)
  ,function(i) {
     paste0(
        i
      ,"(",colnames(mtcars)[i],")---"
      ,i,"-stats("
      ,paste0(
        names(summary(mtcars[,i]))
        ,": "
        ,unname(summary(mtcars[,i]))
        ,collapse="<br/>"
      )
      ,")"
   )
 }
)

DiagrammeR(
   paste0(
     "graph TD;", "\n",
     paste(connections, collapse = "\n"),"\n",
     "classDef column fill:#0001CC, stroke:#0D3FF3, stroke-width:1px;" ,"\n",
     "class ", paste0(1:length(connections), collapse = ","), " column;"
   )
 )

# also with DiagrammeR() you can use tags from htmltools
# just make sure to use class = "mermaid"
library(htmltools)
diagramSpec = "
graph LR;
  id1(Start)-->id2(Stop);
  style id1 fill:#f9f,stroke:#333,stroke-width:4px;
  style id2 fill:#ccf,stroke:#f66,stroke-width:2px,stroke-dasharray: 5, 5;
"
html_print(tagList(
  tags$h1("R + mermaid.js = Something Special")
  ,tags$pre(diagramSpec)
  ,tags$div(class="mermaid",diagramSpec)
  ,DiagrammeR()
))

# sequence diagrams
# Using this "How to Draw a Sequence Diagram"
#   http://www.cs.uku.fi/research/publications/reports/A-2003-1/page91.pdf
# draw some sequence diagrams with DiagrammeR

DiagrammeR("
sequenceDiagram;
   customer->>ticket seller: ask for ticket;
   ticket seller->>database: seats;
   alt tickets available
     database->>ticket seller: ok;
     ticket seller->>customer: confirm;
     customer->>ticket seller: ok;
     ticket seller->>database: book a seat;
     ticket seller->>printer: print ticket;
   else sold out
     database->>ticket seller: none left;
     ticket seller->>customer:  sorry;
   end
")
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/from_igraph.R
\name{from_igraph}
\alias{from_igraph}
\title{Convert an igraph graph to a DiagrammeR one}
\usage{
from_igraph(
  igraph,
  graph_name = NULL,
  write_backups = FALSE,
  display_msgs = FALSE
)
}
\arguments{
\item{igraph}{An \pkg{igraph} graph object.}

\item{graph_name}{An optional string for labeling the graph object.}

\item{write_backups}{An option to write incremental backups of changing graph
states to disk. If \code{TRUE}, a subdirectory within the working directory will
be created and used to store \code{RDS} files. The default value is \code{FALSE} so
one has to opt in to use this functionality.}

\item{display_msgs}{An option to display messages primarily concerned with
changes in graph selections. By default, this is \code{FALSE}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Convert an igraph graph to a DiagrammeR graph object.
}
\examples{
# Create a DiagrammeR graph object
dgr_graph_orig <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 36,
    m = 50,
    set_seed = 23)

# Convert the DiagrammeR
# graph to an igraph object
ig_graph <-
  dgr_graph_orig \%>\%
  to_igraph()

# Convert the igraph graph
# back to a DiagrammeR graph
dgr_graph_new <-
  ig_graph \%>\%
  from_igraph()

# Get some graph information
(dgr_graph_new \%>\%
  get_graph_info())[, 1:6]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_edfs.R
\name{combine_edfs}
\alias{combine_edfs}
\title{Combine multiple edge data frames into a single edge data frame}
\usage{
combine_edfs(...)
}
\arguments{
\item{...}{Two or more edge data frames, which contain edge IDs and
associated attributes.}
}
\value{
A combined edge data frame.
}
\description{
Combine several edge data frames in the style of \code{rbind()}, except, it works
regardless of the number and ordering of the columns.
}
\examples{
# Create an edge data frame (edf)
edf_1 <-
  create_edge_df(
    from = c(1, 1, 2, 3),
    to = c(2, 4, 4, 1),
    rel = "requires",
    color = "green",
    data = c(2.7, 8.9, 2.6, 0.6))

# Create a second edge data frame
edf_2 <-
  create_edge_df(
    from = c(5, 7, 8, 8),
    to = c(7, 8, 6, 5),
    rel = "receives",
    arrowhead = "dot",
    color = "red")

# Combine the two edge data frames
all_edges <- combine_edfs(edf_1, edf_2)

# View the combined edge data frame
all_edges

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_graph_empty.R
\name{is_graph_empty}
\alias{is_graph_empty}
\title{Is the graph empty?}
\usage{
is_graph_empty(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Provides a logical value on whether the graph is empty (i.e., contains no
nodes).
}
\examples{
# Create an empty graph
graph <- create_graph()

# Determine whether the graph is empty
graph \%>\% is_graph_empty()

# Create a non-empty graph
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 3)

# Determine whether this graph is empty
graph \%>\% is_graph_empty()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_node.R
\name{add_node}
\alias{add_node}
\title{Add a node to an existing graph object}
\usage{
add_node(
  graph,
  type = NULL,
  label = NULL,
  from = NULL,
  to = NULL,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{type}{An optional character object that acts as a group identifier for
the node to be added.}

\item{label}{An optional character object that describes the node.}

\item{from}{An optional vector containing node IDs from which edges will be
directed to the new node.}

\item{to}{An optional vector containing node IDs to which edges will be
directed from the new node.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add a new node to the graph. One
can optionally provide node attributes for the created node. There is also
the option to create edges to and from existing nodes in the graph. Because
new edges can also be created through this function, there is the possibility
to set edge attributes for any new graph edges.
}
\examples{
# Create an empty graph and add 2 nodes by using
# the `add_node()` function twice
graph <-
  create_graph() \%>\%
  add_node() \%>\%
  add_node()

# Get a count of all nodes
# in the graph
graph \%>\% count_nodes()

# The nodes added were given
# ID values `1` and `2`; obtain
# the graph's node IDs
graph \%>\% get_node_ids()

# Add a node with a `type`
# value defined
graph <-
  graph \%>\%
  add_node(type = "person")

# View the graph's internal
# node data frame (ndf)
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trav_out_edge.R
\name{trav_out_edge}
\alias{trav_out_edge}
\title{Traverse from one or more selected nodes onto adjacent, outward edges}
\usage{
trav_out_edge(
  graph,
  conditions = NULL,
  copy_attrs_from = NULL,
  copy_attrs_as = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the traversal.}

\item{copy_attrs_from}{Providing a node attribute name will copy those node
attribute values to the traversed edges. If the edge attribute already
exists, the values will be merged to the traversed edges; otherwise, a new
edge attribute will be created.}

\item{copy_attrs_as}{If a node attribute name is provided in
\code{copy_attrs_from}, this option will allow the copied attribute values to be
written under a different edge attribute name. If the attribute name
provided in \code{copy_attrs_as} does not exist in the graph's edf, the new edge
attribute will be created with the chosen name.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} move to outgoing edges from a
selection of one or more selected nodes, thereby creating a selection of
edges. An optional filter by edge attribute can limit the set of edges
traversed to.

This traversal function makes use of an active selection of nodes. After the
traversal, depending on the traversal conditions, there will either be a
selection of edges or no selection at all.

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a simple graph
graph <-
  create_graph() |>
  add_n_nodes(
    n = 2,
    type = "a",
    label = c("asd", "iekd")) |>
  add_n_nodes(
    n = 3,
    type = "b",
    label = c("idj", "edl", "ohd")) |>
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->5",
    rel = c(NA, "A", "B", "C", "D")) |>
  set_node_attrs(
    node_attr = values,
    values = c(2.3, 4.7, 9.4,
               8.3, 6.3))

# Create a data frame with node ID values
# representing the graph edges (with `from`
# and `to` columns), and, a set of numeric values
df <-
  data.frame(
    from = c(1, 1, 2, 2, 3),
    to = c(2, 3, 4, 5, 5),
    values = round(rnorm(5, 5), 2))

# Join the data frame to the graph's internal
# edge data frame (edf)
graph <-
  graph |>
  join_edge_attrs(
    df = df)

# Show the graph's internal node data frame
graph |> get_node_df()

# Show the graph's internal edge data frame
graph |> get_edge_df()

# Perform a simple traversal from nodes to
# outbound edges with no conditions on the
# nodes traversed to
graph |>
  select_nodes_by_id(nodes = 1) |>
  trav_out_edge() |>
  get_selection()

# Traverse from node `1` to any outbound
# edges, filtering to those edges that have
# NA values for the `rel` edge attribute
graph |>
  select_nodes_by_id(nodes = 1) |>
  trav_out_edge(
    conditions = is.na(rel)) |>
  get_selection()

# Traverse from node `3` to any outbound
# edges, filtering to those edges that have
# numeric values greater than `5.0` for
# the `rel` edge attribute
graph |>
  select_nodes_by_id(nodes = 3) |>
  trav_out_edge(
    conditions = values > 5.0) |>
  get_selection()

# Traverse from node `1` to any outbound
# edges, filtering to those edges that
# have values equal to `A` for the `rel`
# edge attribute
graph |>
  select_nodes_by_id(nodes = 1) |>
  trav_out_edge(
    conditions = rel == "A") |>
  get_selection()

# Traverse from node `2` to any outbound
# edges, filtering to those edges that
# have values in the set `B` and `C` for
# the `rel` edge attribute
graph |>
  select_nodes_by_id(nodes = 2) |>
  trav_out_edge(
    conditions = rel \%in\% c("B", "C")) |>
  get_selection()

# Traverse from node `2` to any
# outbound edges, and use multiple
# conditions for the traversal
graph |>
  select_nodes_by_id(nodes = 2) |>
  trav_out_edge(
    conditions =
      rel \%in\% c("B", "C") &
      values >= 5.0) |>
  get_selection()

# Traverse from node `2` to any
# outbound edges, and use multiple
# conditions
graph |>
  select_nodes_by_id(nodes = 2) |>
  trav_out_edge(
    conditions =
      rel \%in\% c("B", "C") |
      values > 6.0) |>
  get_selection()

# Traverse from node `2` to any outbound
# edges, and use a regular expression as
# a filtering condition
graph |>
  select_nodes_by_id(nodes = 2) |>
  trav_out_edge(
    conditions = grepl("B|C", rel)) |>
  get_selection()

# Perform a traversal from all nodes to
# their outgoing edges and, while doing
# so, copy the `label` node attribute
# to any of the nodes' incoming edges
graph <-
  graph |>
  select_nodes() |>
  trav_out_edge(
    copy_attrs_from = label)

# Show the graph's internal edge
# data frame after this change
graph |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_graph_name.R
\name{get_graph_name}
\alias{get_graph_name}
\title{Get graph name}
\usage{
get_graph_name(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single-length character vector with the assigned graph name. If a
graph name has not been set, NA is returned.
}
\description{
Get the name of a graph object of class \code{dgr_graph}.
}
\examples{
# Create an empty graph
graph <- create_graph()

# Provide the new graph with a name
graph <-
  set_graph_name(
    graph,
    name = "the_name")

# Get the graph's name
graph \%>\% get_graph_name()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{node_list_1}
\alias{node_list_1}
\title{Node list - Version 1.}
\format{
A data frame with 10 rows and 2 variables:
\describe{
\item{id}{a unique, monotonically increasing integer ID value}
\item{label}{a unique label associated with each ID value}
}
}
\usage{
node_list_1
}
\description{
A very simple, 2-column data frame that can be used to generate graph nodes.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorize_node_attrs.R
\name{colorize_node_attrs}
\alias{colorize_node_attrs}
\title{Apply colors based on node attribute values}
\usage{
colorize_node_attrs(
  graph,
  node_attr_from,
  node_attr_to,
  cut_points = NULL,
  palette = "Spectral",
  alpha = NULL,
  reverse_palette = FALSE,
  default_color = "#D9D9D9"
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr_from}{The name of the node attribute column from which color
values will be based.}

\item{node_attr_to}{The name of the new node attribute to which the color
values will be applied.}

\item{cut_points}{An optional vector of numerical breaks for bucketizing
continuous numerical values available in a edge attribute column.}

\item{palette}{Can either be: (1) a palette name from the RColorBrewer
package (e.g., \code{Greens}, \code{OrRd}, \code{RdYlGn}), (2) \code{viridis}, which indicates
use of the \code{viridis} color scale from the package of the same name, or (3)
a vector of hexadecimal color names.}

\item{alpha}{An optional alpha transparency value to apply to the generated
colors. Should be in the range of \code{0} (completely transparent) to \code{100}
(completely opaque).}

\item{reverse_palette}{An option to reverse the order of colors in the chosen
palette. The default is \code{FALSE}.}

\item{default_color}{A hexadecimal color value to use for instances when the
values do not fall into the bucket ranges specified in the \code{cut_points}
vector.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal node data frame (ndf), use a categorical node
attribute to generate a new node attribute with color values.
}
\examples{
# Create a graph with 8
# nodes and 7 edges
graph <-
  create_graph() \%>\%
  add_path(n = 8) \%>\%
  set_node_attrs(
    node_attr = weight,
    values = c(
      8.2, 3.7, 6.3, 9.2,
      1.6, 2.5, 7.2, 5.4))

# Find group membership values for all nodes
# in the graph through the Walktrap community
# finding algorithm and join those group values
# to the graph's internal node data frame (ndf)
# with the `join_node_attrs()` function
graph <-
  graph \%>\%
  join_node_attrs(
    df = get_cmty_walktrap(.))

# Inspect the number of distinct communities
graph \%>\%
  get_node_attrs(
    node_attr = walktrap_group) \%>\%
  unique() \%>\%
  sort()

# Visually distinguish the nodes in the different
# communities by applying colors using the
# `colorize_node_attrs()` function; specifically,
# set different `fillcolor` values with an alpha
# value of 90 and apply opaque colors to the node
# border (with the `color` node attribute)
graph <-
  graph \%>\%
  colorize_node_attrs(
    node_attr_from = walktrap_group,
    node_attr_to = fillcolor,
    palette = "Greens",
    alpha = 90) \%>\%
  colorize_node_attrs(
    node_attr_from = walktrap_group,
    node_attr_to = color,
    palette = "viridis",
    alpha = 80)

# Show the graph's internal node data frame
graph \%>\% get_node_df()

# Create a graph with 8 nodes and 7 edges
graph <-
  create_graph() \%>\%
  add_path(n = 8) \%>\%
  set_node_attrs(
    node_attr = weight,
    values = c(
      8.2, 3.7, 6.3, 9.2,
      1.6, 2.5, 7.2, 5.4))

# We can bucketize values in `weight` using
# `cut_points` and assign colors to each of the
# bucketed ranges (for values not part of any
# bucket, a gray color is assigned by default)
graph <-
  graph \%>\%
  colorize_node_attrs(
    node_attr_from = weight,
    node_attr_to = fillcolor,
    cut_points = c(1, 3, 5, 7, 9))

# Now there will be a `fillcolor` node attribute
# with distinct colors (the `#D9D9D9` color is
# the default `gray85` color)
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_graph_undirected.R
\name{is_graph_undirected}
\alias{is_graph_undirected}
\title{Is the graph an undirected graph?}
\usage{
is_graph_undirected(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Determines whether a graph is set as undirected or not and returns a logical
value to that effect.
}
\examples{
# Create an empty graph; by
# default, new graphs made
# by `create_graph()` are
# directed graph, so, use
# `directed = FALSE` to create
# an undirected graph
graph <-
  create_graph(
    directed = FALSE)

# Determine whether the
# graph is undirected
graph \%>\% is_graph_undirected()

# Use the `set_graph_directed()`
# function and check again
# as to whether the graph is
# undirected
graph \%>\%
  set_graph_directed() \%>\%
  is_graph_undirected()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/save_graph.R
\name{save_graph}
\alias{save_graph}
\title{Save a graph or graph series to disk}
\usage{
save_graph(x, file)
}
\arguments{
\item{x}{A graph object of class \code{dgr_graph} or a graph series object of type
\code{dgr_graph_1D}.}

\item{file}{A file name for the graph or graph series. Provide a character
string and the \code{.dgr} extension will be applied to it.}
}
\description{
Save a graph or a graph series object to disk.
}
\examples{
# Create an undirected GNP
# graph with 100 nodes using
# a probability value of 0.05
gnp_graph <-
  create_graph(
    directed = FALSE) |>
  add_gnp_graph(
    n = 100,
    p = 0.05)

# Save the graph to disk; use
# the file name `gnp_graph.dgr`
# save_graph(
#   x = gnp_graph,
#   file = "gnp_graph"
# )

# To read the graph file from
# disk, use `open_graph()`
# gnp_graph_2 <-
#   open_graph(
#     file = "gnp_graph.dgr"
# )

}
\seealso{
Other Display and Save: 
\code{\link{export_graph}()},
\code{\link{render_graph}()},
\code{\link{render_graph_from_graph_series}()}
}
\concept{Display and Save}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_graph_time.R
\name{get_graph_time}
\alias{get_graph_time}
\title{Get the graph date-time or timezone}
\usage{
get_graph_time(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single-length \code{POSIXct} vector with the assigned graph time.
}
\description{
Get the time and timezone for a graph object of class \code{dgr_graph}.
}
\examples{
# Create an empty graph and
# set the graph's time; if nothing
# is supplied for the `tz` argument,
# `GMT` is used as the time zone
graph <-
  create_graph() \%>\%
    set_graph_time(
      time = "2015-10-25 15:23:00")

# Get the graph's time as a POSIXct
# object using `get_graph_time()`
graph \%>\% get_graph_time()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_authority_centrality.R
\name{get_authority_centrality}
\alias{get_authority_centrality}
\title{Get the authority scores for all nodes}
\usage{
get_authority_centrality(graph, weights_attr = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{weights_attr}{an optional name of the edge attribute to use in the
adjacency matrix. If \code{NULL} then, if it exists, the \code{weight} edge
attribute of the graph will be used.}
}
\value{
a data frame with authority scores for each of the nodes.
}
\description{
Get the Kleinberg authority centrality scores for all nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the authority centrality scores
# for all nodes in the graph
graph \%>\%
  get_authority_centrality()

# Add the authority centrality
# scores to the graph as a node
# attribute
graph <-
  graph \%>\%
  join_node_attrs(
    df = get_authority_centrality(.))

# Display the graph's node data frame
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/grViz.R
\name{add_mathjax}
\alias{add_mathjax}
\title{Add MathJax-formatted equation text}
\usage{
add_mathjax(gv = NULL, include_mathjax = TRUE)
}
\arguments{
\item{gv}{A \code{grViz} htmlwidget.}

\item{include_mathjax}{A \code{logical} to add mathjax JS. Change to \code{FALSE} if
using with \pkg{rmarkdown} since MathJax will likely already be added.}
}
\value{
A \code{grViz} htmlwidget
}
\description{
Add MathJax-formatted equation text
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_n_node_clones.R
\name{add_n_node_clones}
\alias{add_n_node_clones}
\title{Add one or several clones of an existing node to the graph}
\usage{
add_n_node_clones(graph, n, node, label = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{n}{The number of node clones to add to the graph.}

\item{node}{A node ID corresponding to the graph node to be cloned.}

\item{label}{An optional vector of node label values. The vector length
should correspond to the value set for \code{n}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add \code{n} new nodes to a graph object of class \code{dgr_graph} which are clones of
a node already in the graph. All node attributes are preserved except for the
node \code{label} attribute (to maintain the uniqueness of non-\code{NA} node label
values). A vector of node \code{label} can be provided to bind new labels to the
cloned nodes.
}
\examples{
# Create a graph with a path of
# nodes; supply `label`, `type`,
# and `value` node attributes
graph <-
  create_graph() \%>\%
  add_path(
    n = 3,
    label = c("d", "g", "r"),
    type = c("a", "b", "c"))

# Display the graph's internal
# node data frame
graph \%>\% get_node_df()

# Create 3 clones of node `1`
# but assign new node label
# values (leaving `label` as
# NULL yields NA values)
graph <-
  graph \%>\%
  add_n_node_clones(
    n = 3,
    node = 1,
    label = c("x", "y", "z"))

# Display the graph's internal
# node data frame: nodes `4`,
# `5`, and `6` are clones of `1`
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_node_position.R
\name{set_node_position}
\alias{set_node_position}
\title{Apply a layout position to a single node}
\usage{
set_node_position(graph, node, x, y, use_labels = FALSE)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node}{A single-length vector containing either a node ID value
(integer) or a node label (character) for which position information should
be applied.}

\item{x}{The x coordinate to set for the node.}

\item{y}{The y coordinate to set for the node.}

\item{use_labels}{An option to use a node \code{label} value in \code{node}. Note that
this is only possible if all nodes have distinct \code{label} values set and
none exist as an NA value.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Apply position information for a single node. This is done by setting the \code{x}
and \code{y} attrs for a node \code{id} or node \code{label} supplied in \code{node}. When
rendering the graph, nodes with attribute values set for \code{x} and \code{y} will be
fixed to those positions on the graph canvas.
}
\examples{
# Create a simple graph with 4 nodes
graph <-
  create_graph() \%>\%
  add_node(label = "one") \%>\%
  add_node(label = "two") \%>\%
  add_node(label = "three") \%>\%
  add_node(label = "four")

# Add position information to each of
# the graph's nodes
graph <-
  graph \%>\%
  set_node_position(
    node = 1,
    x = 1, y = 1) \%>\%
  set_node_position(
    node = 2,
    x = 2, y = 2) \%>\%
  set_node_position(
    node = 3,
    x = 3, y = 3) \%>\%
  set_node_position(
    node = 4,
    x = 4, y = 4)

# View the graph's node data frame to
# verify that the `x` and `y` node
# attributes are available and set to
# the values provided
graph \%>\% get_node_df()

# The same function can modify the data
# in the `x` and `y` attributes
graph <-
  graph \%>\%
  set_node_position(
    node = 1,
    x = 1, y = 4) \%>\%
  set_node_position(
    node = 2,
    x = 3, y = 3) \%>\%
  set_node_position(
    node = 3,
    x = 3, y = 2) \%>\%
  set_node_position(
    node = 4,
    x = 4, y = 1)

# View the graph's node data frame
graph \%>\% get_node_df()

# Position changes can also be made by
# supplying a node `label` value (and setting
# `use_labels` to TRUE). For this to work,
# all `label` values in the graph's ndf must
# be unique and non-NA
graph <-
  graph \%>\%
  set_node_position(
    node = "one",
    x = 1, y = 1,
    use_labels = TRUE) \%>\%
  set_node_position(
    node = "two",
    x = 2, y = 2,
    use_labels = TRUE)

# View the graph's node data frame
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/x11_hex.R
\name{x11_hex}
\alias{x11_hex}
\title{X11 colors and hexadecimal color values}
\usage{
x11_hex()
}
\description{
Create a data frame containing information on X11 colors and their
corresponding hexadecimal color values.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/visnetwork.R
\name{visnetwork}
\alias{visnetwork}
\title{Render graph with visNetwork}
\usage{
visnetwork(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\description{
Render a graph object with the visNetwork R package.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 6,
    label = TRUE,
    fillcolor = c("lightgrey", "red", "orange",
                  "pink", "aqua", "yellow"),
    shape = "dot",
    size = c(20, 80, 40, 10, 30, 50),
    type = c("1", "1", "1", "2", "2", "2")
  )

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3, 4, 6, 5),
    to = c(4, 3, 1, 3, 1, 4),
    color = c("green", "green", "grey",
              "grey", "blue", "blue"),
    rel = "leading_to"
  )

# Create a graph object
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf
  )

# visnetwork(graph)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_cache.R
\name{set_cache}
\alias{set_cache}
\title{Cache a vector in the graph}
\usage{
set_cache(graph, to_cache, name = NULL, col = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{to_cache}{Any vector or data frame. If a data frame is supplied then a
single column for the vector to pull must be provided in the \code{col}
argument.}

\item{name}{An optional name for the cached vector.}

\item{col}{If a data frame is provided in \code{to_cache} then a column name from
that data frame must provided here.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Place any vector in the cache of a graph object of class \code{dgr_graph}.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 10,
    m = 22,
    set_seed = 23
  )

# Get the closeness values for
# all nodes from `1` to `10` and
# store in the graph's cache
graph <-
  graph \%>\%
  set_cache(
    name = "closeness_vector",
    to_cache = get_closeness(.),
    col = "closeness"
  )

# Get the graph's cache
graph \%>\%
  get_cache(name = "closeness_vector")

# Get the difference of betweenness
# and closeness values for nodes in
# the graph and store the vector in
# the graph's cache
graph <-
  graph \%>\%
  set_cache(
    name = "difference",
    to_cache =
      get_betweenness(.)$betweenness -
        get_closeness(.)$closeness
  )

# Get the graph's cache
graph \%>\%
  get_cache(name = "difference")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/render_graph_from_graph_series.R
\name{render_graph_from_graph_series}
\alias{render_graph_from_graph_series}
\title{Render a graph available in a series}
\usage{
render_graph_from_graph_series(
  graph_series,
  graph_no,
  output = "graph",
  width = NULL,
  height = NULL
)
}
\arguments{
\item{graph_series}{A graph series object of type \code{dgr_graph_1D}.}

\item{graph_no}{The index of the graph in the graph series.}

\item{output}{A string specifying the output type; \code{graph} (the default)
renders the graph using the \code{\link[=grViz]{grViz()}} function and \code{visNetwork} renders the
graph using the \code{\link[=visnetwork]{visnetwork()}} function.}

\item{width}{An optional parameter for specifying the width of the resulting
graphic in pixels.}

\item{height}{An optional parameter for specifying the height of the
resulting graphic in pixels.}
}
\description{
Using a graph series object of type \code{dgr_graph_1D}, either render graph in
the Viewer or output in various formats.
}
\examples{
\dontrun{
# Create three graphs
graph_1 <-
  create_graph() \%>\%
  add_path(n = 4)

graph_2 <-
  create_graph() \%>\%
  add_cycle(n = 5)

graph_3 <-
  create_graph() \%>\%
  add_star(n = 6)

# Create an empty graph series
# and add the graphs
series <-
  create_graph_series() \%>\%
  add_graph_to_graph_series(
    graph = graph_1) \%>\%
  add_graph_to_graph_series(
    graph = graph_2) \%>\%
  add_graph_to_graph_series(
    graph = graph_3)

# View the second graph in
# the series in the Viewer
render_graph_from_graph_series(
  graph_series = series,
  graph_no = 2)
}

}
\seealso{
Other Display and Save: 
\code{\link{export_graph}()},
\code{\link{render_graph}()},
\code{\link{save_graph}()}
}
\concept{Display and Save}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_graph_connected.R
\name{is_graph_connected}
\alias{is_graph_connected}
\title{Is the graph a connected graph?}
\usage{
is_graph_connected(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A logical value.
}
\description{
Determines whether a graph is a connected graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function; this
# graph is not connected
create_graph() |>
  add_gnm_graph(
    n = 15,
    m = 10,
    set_seed = 23) |>
  is_graph_connected()

# Create another random graph;
# this graph is connected
create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23) |>
  is_graph_connected()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{node_list_2}
\alias{node_list_2}
\title{Node list - Version 2.}
\format{
A data frame with 10 rows and 5 variables:
\describe{
\item{id}{a unique, monotonically increasing integer ID value}
\item{label}{a unique label associated with each ID value}
\item{type}{a grouping variable of either x, y, or z}
\item{value_1}{a randomized set of numeric values between 0 and 10}
\item{value_2}{a randomized set of numeric values between 0 and 10}
}
}
\usage{
node_list_2
}
\description{
A simple, 5-column data frame that can be used to generate graph nodes.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/node_aes.R
\name{node_aes}
\alias{node_aes}
\title{Insert node aesthetic attributes during node creation}
\usage{
node_aes(
  shape = NULL,
  style = NULL,
  penwidth = NULL,
  color = NULL,
  fillcolor = NULL,
  image = NULL,
  fontname = NULL,
  fontsize = NULL,
  fontcolor = NULL,
  peripheries = NULL,
  height = NULL,
  width = NULL,
  x = NULL,
  y = NULL,
  group = NULL,
  tooltip = NULL,
  xlabel = NULL,
  URL = NULL,
  sides = NULL,
  orientation = NULL,
  skew = NULL,
  distortion = NULL,
  gradientangle = NULL,
  fixedsize = NULL,
  labelloc = NULL,
  margin = NULL
)
}
\arguments{
\item{shape}{The shape to use for the node. Some possible \code{shape} types
include: \code{circle}, \code{rectangle}, \code{triangle}, \code{plaintext}, \code{square}, and
\code{polygon}.}

\item{style}{The node line style. The \code{style} types that can be used are:
\code{filled}, \code{invisible}, \code{diagonals}, \code{rounded}, \code{dashed}, \code{dotted}, \code{solid},
and \code{bold}.}

\item{penwidth}{The thickness of the stroke line (in pt units) for the node
shape. The default value is \code{1.0}.}

\item{color}{The color of the node's outline. Can be any of the named colors
that R knows about (obtained using the \code{colors()} function), or, a
hexadecimal color code.}

\item{fillcolor}{The color with which to fill the shape of the node. Can be
any of the named colors that R knows about (obtained using the \code{colors()}
function), or, a hexadecimal color code.}

\item{image}{A reference to an image location.}

\item{fontname}{The name of the system font that will be used for any node
text.}

\item{fontsize}{The point size of the font used for any node text.}

\item{fontcolor}{The color used for any node text. Can be any of the named
colors that R knows about (obtained using the \code{colors()} function), or, a
hexadecimal color code.}

\item{peripheries}{The repeated number of node shapes (of increasing size) to
draw at the node periphery.}

\item{height}{The height of the node shape, in inches. The default value is
\code{0.5} whereas the minimum value is \code{0.02}. This is understood as the
initial, minimum height of the node. If \code{fixedsize} is set to \code{TRUE}, this
will be the final height of the node. Otherwise, if the node label requires
more height to fit, the node's height will be increased to contain the
label.}

\item{width}{The width of the node shape, in inches. The default value is
\code{0.5} whereas the minimum value is \code{0.02}. This is understood as the
initial, minimum width of the node. If \code{fixedsize} is set to \code{TRUE}, this
will be the final width of the node. Otherwise, if the node label requires
more width to fit, the node's width will be increased to contain the label.}

\item{x}{The fixed position of the node in the x direction. Any integer-based
or floating point value will be accepted.}

\item{y}{The fixed position of the node in the y direction. Any integer-based
or floating point value will be accepted.}

\item{group}{The node group.}

\item{tooltip}{Text for a node tooltip.}

\item{xlabel}{External label for a node. The label will be placed outside of
the node but near it. These labels are added after all nodes and edges have
been placed. The labels will be placed so that they do not overlap any node
or label. This means it may not be possible to place all of them.}

\item{URL}{A URL to associate with a node. Upon rendering the plot, clicking
nodes with any associated URLs will open the URL in the default browser.}

\item{sides}{When using the shape \code{polygon}, this value will provide the
number of sides for that polygon.}

\item{orientation}{This is the angle, in degrees, that is used to rotate
nodes that have a \code{shape} of \code{polygon}. Not that for any of the polygon
shapes (set by the \code{sides} node attribute), a value for \code{orientation} that
is \code{0} results in a flat base.}

\item{skew}{A \code{0-1} value that will result in the node shape being skewed to
the right (from bottom to top). A value in the range \code{0} to \code{-1} will skew
the shape to the left.}

\item{distortion}{A distortion factor that is used only when a \code{shape} of
\code{polygon} is used. A \code{0-1} value will increasingly result in the top part
of the node polygon shape to be larger than the bottom. Moving from \code{0}
toward \code{-1} will result in the opposite distortion effect.}

\item{gradientangle}{The path angle for the node color fill gradient.}

\item{fixedsize}{If set to \code{FALSE}, the size of a node is determined by
smallest width and height needed to contain its label, if any, with a
margin specified by the \code{margin} node attribute. The width and height must
also be at least as large as the sizes specified by the \code{width} and
\code{height} node attributes, which specify the minimum values. If set to
\code{TRUE}, the node size is entirely specified by the values of the \code{width}
and \code{height} node attributes (i.e., the node is not expanded in size to
contain the text label).}

\item{labelloc}{Sets the vertical placement of labels for nodes and clusters.
This attribute is used only when the height of the node is larger than the
height of its label. The \code{labelloc} node attribute can be set to either \code{t}
(top), \code{c} (center), or \code{b} (bottom). By default, the label is vertically
centered.}

\item{margin}{Sets the amount of space around the node's label. By default,
the value is \verb{0.11,0.055}.}
}
\description{
This helper function should be invoked to provide values for the namesake
\code{node_aes} argument, which is present in any function where nodes are
created.
}
\examples{
# Create a new graph and add
# a path with several node
# aesthetic attributes
graph <-
  create_graph() |>
  add_path(
    n = 3,
    type = "path",
    node_aes = node_aes(
      shape = "circle",
      x = c(1, 3, 2),
      y = c(4, -1, 3)
    )
  )

# View the graph's internal
# node data frame; the node
# aesthetic attributes have
# been inserted
graph |> get_node_df()

# Create a new graph which is
# fully connected
graph <-
  create_graph() |>
  add_full_graph(
    n = 4,
    node_data = node_data(value = 1:4),
    node_aes = node_aes(
      x = c(2, 1, 3, 2),
      y = c(3, 2, 2, 1)
    ),
    edge_aes = edge_aes(color = "blue")
  )
}
\seealso{
Other aesthetics: 
\code{\link{edge_aes}()},
\code{\link{node_edge_aes_data}}
}
\concept{aesthetics}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_closeness.R
\name{get_closeness}
\alias{get_closeness}
\title{Get closeness centrality values}
\usage{
get_closeness(graph, direction = "all")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{direction}{using \code{all} (the default), the search will ignore edge
direction while traversing through the graph. With \code{out}, measurements
of paths will be from a node whereas with \verb{in}, measurements of paths
will be to a node.}
}
\value{
A data frame with closeness values for each of the nodes.
}
\description{
Get the closeness centrality values for all nodes in a graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 12,
    set_seed = 23)

# Get closeness values for all nodes
# in the graph
graph |> get_closeness()

# Add the closeness values to
# the graph as a node attribute
graph <-
  graph |>
  join_node_attrs(
    df = get_closeness(graph))

# Display the graph's node data frame
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.dgr_graph}
\alias{print.dgr_graph}
\title{Print the graph to the terminal}
\usage{
\method{print}{dgr_graph}(x, ...)
}
\arguments{
\item{x}{A graph object of class \code{dgr_graph}.}
}
\description{
This function will provide a summary of the graph.
}
\examples{
\dontrun{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get a summary of the graph
graph
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_nodes.R
\name{select_nodes}
\alias{select_nodes}
\title{Select nodes in a graph}
\usage{
select_nodes(graph, conditions = NULL, set_op = "union", nodes = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the retrieval of
nodes.}

\item{set_op}{The set operation to perform upon consecutive selections of
graph nodes. This can either be as a \code{union} (the default), as an
intersection of selections with \code{intersect}, or, as a \code{difference} on the
previous selection, if it exists.}

\item{nodes}{An optional vector of node IDs for filtering the list of nodes
present in the graph.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Select nodes from a graph object of class \code{dgr_graph}.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = c("a", "a", "z", "z"),
    label = TRUE,
    value = c(3.5, 2.6, 9.4, 2.7))

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = c("a", "z", "a"))

# Create a graph with the ndf and edf
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Explicitly select nodes `1` and `3`
graph <-
  graph \%>\%
  select_nodes(nodes = c(1, 3))

# Verify that the node selection has been made
# using the `get_selection()` function
graph \%>\% get_selection()

# Select nodes based on the node `type`
# being `z`
graph <-
  graph \%>\%
  clear_selection() \%>\%
  select_nodes(
    conditions = type == "z")

# Verify that an node selection has been made, and
# recall that the `3` and `4` nodes are of the
# `z` type
graph \%>\% get_selection()

# Select edges based on the node value attribute
# being greater than 3.0 (first clearing the current
# selection of nodes)
graph <-
  graph \%>\%
  clear_selection() \%>\%
  select_nodes(
    conditions = value > 3.0)

# Verify that the correct node selection has been
# made; in this case, nodes `1` and `3` have values
# for `value` greater than 3.0
graph \%>\% get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_edge_attrs_ws.R
\name{mutate_edge_attrs_ws}
\alias{mutate_edge_attrs_ws}
\title{Mutate edge attribute values for a selection of edges}
\usage{
mutate_edge_attrs_ws(graph, ...)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{...}{Expressions used for the mutation of edge attributes. LHS of each
expression is either an existing or new edge attribute name. The RHS can
consist of any valid R code that uses edge attributes as variables.
Expressions are evaluated in the order provided, so, edge attributes
created or modified are ready to use in subsequent expressions.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal edge data frame (edf), mutate edge attribute values
only for edges in a selection by using one or more expressions.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a graph with 3 edges
# and then select edge `1`
graph <-
  create_graph() \%>\%
  add_path(n = 4) \%>\%
  set_edge_attrs(
    edge_attr = width,
    values = c(3.4, 2.3, 7.2)) \%>\%
  select_edges(edges = 1)

# Get the graph's internal edf
# to show which edge attributes
# are available
graph \%>\% get_edge_df()

# Mutate the `width` edge
# attribute for the edges
# only in the active selection
# of edges (edge `1`); here,
# we divide each value in the
# selection by 2
graph <-
  graph \%>\%
  mutate_edge_attrs_ws(
    width = width / 2)

# Get the graph's internal
# edf to show that the edge
# attribute `width` had its
# values changed
graph \%>\% get_edge_df()

# Create a new edge attribute,
# called `length`, that is the
# log of values in `width` plus
# 2 (and, also, round all values
# to 2 decimal places)
graph <-
  graph \%>\%
  clear_selection() \%>\%
  select_edges(edges = 2:3) \%>\%
  mutate_edge_attrs_ws(
    length = (log(width) + 2) \%>\%
               round(2))

# Get the graph's internal edf
# to show that the edge attribute
# values had been mutated only
# for edges `2` and `3` (since
# edge `1` is excluded, an NA
# value is applied)
graph \%>\% get_edge_df()

# Create a new edge attribute
# called `area`, which is the
# product of the `width` and
# `length` attributes
graph <-
  graph \%>\%
  mutate_edge_attrs_ws(
    area = width * length)

# Get the graph's internal edf
# to show that the edge attribute
# values had been multiplied
# together (with new attr `area`)
# for nodes `2` and `3`
graph \%>\% get_edge_df()

# We can invert the selection
# and mutate edge `1` several
# times to get an `area` value
# for that edge
graph <-
  graph \%>\%
  invert_selection() \%>\%
  mutate_edge_attrs_ws(
    length = (log(width) + 5) \%>\%
               round(2),
    area = width * length)

# Get the graph's internal edf
# to show that the 2 mutations
# occurred for edge `1`, yielding
# non-NA values for its edge
# attributes without changing
# those of the other edges
graph \%>\% get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DiagrammeR-package.R
\docType{package}
\name{DiagrammeR-package}
\alias{DiagrammeR-package}
\title{DiagrammeR: Graph/Network Visualization}
\description{
Build graph/network structures using functions for stepwise addition and deletion of nodes and edges. Work with data available in tables for bulk addition of nodes, edges, and associated metadata. Use graph selections and traversals to apply changes to specific nodes or edges. A wide selection of graph algorithms allow for the analysis of graphs. Visualize the graphs and take advantage of any aesthetic properties assigned to nodes and edges.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://rich-iannone.github.io/DiagrammeR/}
  \item \url{https://github.com/rich-iannone/DiagrammeR}
  \item Report bugs at \url{https://github.com/rich-iannone/DiagrammeR/issues}
}

}
\author{
\strong{Maintainer}: Richard Iannone \email{riannone@me.com} (\href{https://orcid.org/0000-0003-3925-190X}{ORCID})

Authors:
\itemize{
  \item Olivier Roy \email{olivierroy71@hotmail.com}
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_last_edges_created.R
\name{select_last_edges_created}
\alias{select_last_edges_created}
\title{Select the last set of edges created in a graph}
\usage{
select_last_edges_created(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Select the last edges that were created in a graph object of class
\code{dgr_graph}. This function should ideally be used just after creating the
edges to be selected.
}
\examples{
# Create a graph and add a cycle and then
# a tree in 2 separate function calls
graph <-
  create_graph() |>
  add_cycle(
    n = 3,
    rel = "a") |>
  add_balanced_tree(
    k = 2, h = 2,
    rel = "b")

# Select the last edges created (all edges
# from the tree) and then set their edge
# color to be `red`
graph <-
  graph |>
  select_last_edges_created() |>
  set_edge_attrs_ws(
    edge_attr = color,
    value = "red") |>
  clear_selection()

# Display the graph's internal edge
# data frame to verify the change
graph |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nudge_node_positions_ws.R
\name{nudge_node_positions_ws}
\alias{nudge_node_positions_ws}
\title{Move layout positions of a selection of nodes}
\usage{
nudge_node_positions_ws(graph, dx, dy)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{dx}{A single numeric value specifying the amount that selected nodes
(with non-\code{NA} values for the \code{x} and \code{y} attributes) will be moved in the
x direction. A positive value will move nodes right, negative left.}

\item{dy}{A single numeric value specifying the amount that selected nodes
(with non-\code{NA} values for the \code{x} and \code{y} attributes) will be moved in the
y direction. A positive value will move nodes up, negative down.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With an active selection of nodes, move the position in either the \code{x} or \code{y}
directions, or both. Nodes in the selection that do not have position
information (i.e., \code{NA} values for the \code{x} or \code{y} node attributes) will be
ignored.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create a simple graph with 4 nodes
graph <-
  create_graph() |>
  add_node(
    type = "a",
    label = "one") |>
  add_node(
    type = "a",
    label = "two") |>
  add_node(
    type = "b",
    label = "three") |>
  add_node(
    type = "b",
    label = "four")

# Add position information to each of
# the graph's nodes
graph <-
  graph |>
  set_node_position(
    node = 1, x = 1, y = 1) |>
  set_node_position(
    node = 2, x = 2, y = 2) |>
  set_node_position(
    node = 3, x = 3, y = 3) |>
  set_node_position(
    node = 4, x = 4, y = 4)

# Select all of the graph's nodes using the
# `select_nodes()` function (and only
# specifying the graph object)
graph <- select_nodes(graph)

# Move the selected nodes (all the nodes,
# in this case) 5 units to the right
graph <-
  graph |>
  nudge_node_positions_ws(
    dx = 5, dy = 0)

# View the graph's node data frame
graph |> get_node_df()

# Now select nodes that have `type == "b"`
# and move them in the `y` direction 2 units
# (the graph still has an active selection
# and so it must be cleared first)
graph <-
  graph |>
  clear_selection() |>
  select_nodes(
    conditions = type == "b") |>
  nudge_node_positions_ws(
    dx = 0, dy = 2)

# View the graph's node data frame
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rename_edge_attrs.R
\name{rename_edge_attrs}
\alias{rename_edge_attrs}
\title{Rename an edge attribute}
\usage{
rename_edge_attrs(graph, edge_attr_from, edge_attr_to)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr_from}{The name of the edge attribute that will be renamed.}

\item{edge_attr_to}{The new name of the edge attribute column identified in
\code{edge_attr_from}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal edge data frame (edf), rename an existing edge
attribute.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 5,
    m = 8,
    set_seed = 23) \%>\%
  set_edge_attrs(
    edge_attr = color,
    values = "green")

# Get the graph's internal edf
# to show which edge attributes
# are available
graph \%>\% get_edge_df()

# Rename the `color` node
# attribute as `weight`
graph <-
  graph \%>\%
  rename_edge_attrs(
    edge_attr_from = color,
    edge_attr_to = labelfontcolor)

# Get the graph's internal
# edf to show that the edge
# attribute had been renamed
graph \%>\% get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_pa_graph.R
\name{add_pa_graph}
\alias{add_pa_graph}
\title{Add a preferential attachment graph}
\usage{
add_pa_graph(
  graph,
  n,
  m = NULL,
  power = 1,
  out_dist = NULL,
  use_total_degree = FALSE,
  zero_appeal = 1,
  algo = "psumtree",
  type = NULL,
  label = TRUE,
  rel = NULL,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL,
  set_seed = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{n}{The number of nodes comprising the preferential attachment graph.}

\item{m}{The number of edges to add in each time step.}

\item{power}{The power of the preferential attachment. The default value of
\code{1} indicates a linear preferential attachment.}

\item{out_dist}{A numeric vector that provides the distribution of the number
of edges to add in each time step.}

\item{use_total_degree}{A logical value (default is \code{TRUE}) that governs
whether the total degree should be used for calculating the citation
probability. If \code{FALSE}, the indegree is used.}

\item{zero_appeal}{A measure of the attractiveness of the nodes with no
adjacent edges.}

\item{algo}{The algorithm to use to generate the graph. The available options
are \code{psumtree}, \code{psumtree-multiple}, and \code{bag}. With the \code{psumtree}
algorithm, a partial prefix-sum tree is used to to create the graph. Any
values for \code{power} and \code{zero_appeal} can be provided and this algorithm
never generates multiple edges. The \code{psumtree-multiple} algorithm also uses
a partial prefix-sum tree but the difference here is that multiple edges
are allowed. The \code{bag} algorithm places the node IDs into a bag as many
times as their in-degree (plus once more). The required number of cited
nodes are drawn from the bag with replacement. Multiple edges may be
produced using this method (it is not disallowed).}

\item{type}{An optional string that describes the entity type for all the
nodes to be added.}

\item{label}{A logical value where setting to \code{TRUE} ascribes node IDs to the
label and \code{FALSE} yields a blank label.}

\item{rel}{An optional string for providing a relationship label to all edges
to be added.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}

\item{set_seed}{Supplying a value sets a random seed of the
\code{Mersenne-Twister} implementation.}
}
\description{
To an existing graph object, add a graph built according to the
Barabasi-Albert model, which uses preferential attachment in its stochastic
algorithm.
}
\examples{
# Create an undirected PA
# graph with 100 nodes, adding
# 2 edges at every time step
pa_graph <-
  create_graph(
    directed = FALSE) |>
  add_pa_graph(
    n = 100,
    m = 1)

# Get a count of nodes
pa_graph |> count_nodes()

# Get a count of edges
pa_graph |> count_edges()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_edges.R
\name{select_edges}
\alias{select_edges}
\title{Select edges in a graph}
\usage{
select_edges(
  graph,
  conditions = NULL,
  set_op = "union",
  from = NULL,
  to = NULL,
  edges = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the retrieval of
edges.}

\item{set_op}{The set operation to perform upon consecutive selections of
graph nodes. This can either be as a \code{union} (the default), as an
intersection of selections with \code{intersect}, or, as a \code{difference} on the
previous selection, if it exists.}

\item{from}{An optional vector of node IDs from which the edge is outgoing
for filtering the list of edges present in the graph.}

\item{to}{An optional vector of node IDs to which the edge is incoming for
filtering the list of edges present in the graph.}

\item{edges}{An optional vector of edge IDs for filtering the list of edges
present in the graph.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Select edges from a graph object of class \code{dgr_graph}.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = "basic",
    label = TRUE,
    value = c(3.5, 2.6, 9.4, 2.7))

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = c("a", "z", "a"),
    value = c(6.4, 2.9, 5.0))

# Create a graph with the ndf and edf
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Explicitly select the edge `1`->`4`
graph <-
  graph |>
  select_edges(
    from = 1,
    to = 4)

# Verify that an edge selection has been made
# using the `get_selection()` function
graph |> get_selection()

# Select edges based on the relationship label
# being `z`
graph <-
  graph |>
  clear_selection() |>
  select_edges(
    conditions = rel == "z")

# Verify that an edge selection has been made, and
# recall that the `2`->`3` edge uniquely has the
# `z` relationship label
graph |> get_selection()

# Select edges based on the edge value attribute
# being greater than 3.0 (first clearing the current
# selection of edges)
graph <-
  graph |>
  clear_selection() |>
  select_edges(
    conditions = value > 3.0)

# Verify that the correct edge selection has been
# made; in this case, edges `1`->`4` and
# `3`->`1` have values for `value` > 3.0
graph |> get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_to_subgraph_ws.R
\name{transform_to_subgraph_ws}
\alias{transform_to_subgraph_ws}
\title{Create a subgraph using a node or edge selection}
\usage{
transform_to_subgraph_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Create a subgraph based on a selection of nodes or edges stored in the graph
object.

This function makes use of an active selection of nodes or edges (and the
function ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of edges can be performed using the following edge selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of nodes or edges can also be performed using the following
traversal (\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, \code{\link[=trav_in_until]{trav_in_until()}},
\code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}}, \code{\link[=trav_both_edge]{trav_both_edge()}}, or
\code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 6,
    value =
      c(3.5, 2.6, 9.4,
        2.7, 5.2, 2.1))

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 4, 5, 2, 6, 2),
      to = c(2, 4, 1, 3, 5, 5, 4))

# Create a graph
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Create a selection of nodes, this selects
# nodes `1`, `3`, and `5`
graph <-
  graph \%>\%
  select_nodes(
    conditions = value > 3)

# Create a subgraph based on the selection
subgraph <-
  graph \%>\%
  transform_to_subgraph_ws()

# Display the graph's node data frame
subgraph \%>\% get_node_df()

# Display the graph's edge data frame
subgraph \%>\% get_edge_df()

# Create a selection of edges, this selects
# edges `1`, `2`
graph <-
  graph \%>\%
  clear_selection() \%>\%
  select_edges(
  edges = c(1,2))

# Create a subgraph based on the selection
subgraph <-
  graph \%>\%
  transform_to_subgraph_ws()

# Display the graph's node data frame
subgraph \%>\% get_node_df()

# Display the graph's edge data frame
subgraph \%>\% get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_edges_ws.R
\name{delete_edges_ws}
\alias{delete_edges_ws}
\title{Delete all selected edges in an edge selection}
\usage{
delete_edges_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
In a graph object of class \code{dgr_graph}, delete all edges present in a
selection.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a graph
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 3) \%>\%
  add_edges_w_string(
    edges = "1->3 1->2 2->3")

# Select edges attached to
# node with ID `3` (these are
# `1`->`3` and `2`->`3`)
graph <-
  graph \%>\%
  select_edges_by_node_id(nodes = 3)

# Delete edges in selection
graph <-
  graph \%>\%
  delete_edges_ws()

# Get a count of edges in the graph
graph \%>\% count_edges()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_n_nodes_ws.R
\name{add_n_nodes_ws}
\alias{add_n_nodes_ws}
\title{Add a multiple of new nodes with edges to or from one or more selected nodes}
\usage{
add_n_nodes_ws(
  graph,
  n,
  direction = NULL,
  type = NULL,
  label = NULL,
  rel = NULL,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{n}{The number of new nodes to attach as successor nodes to the nodes in
the selection.}

\item{direction}{Using \code{from} will create new edges from existing nodes to
the new nodes. The \code{to} option will create new edges directed toward the
existing nodes.}

\item{type}{An optional character vector that provides group identifiers for
the nodes to be added.}

\item{label}{An optional character object that describes the nodes to be
added.}

\item{rel}{An optional string to apply a \code{rel} attribute to all newly created
edges.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add \code{n} new nodes to or from one or more nodes available as a selection in a
graph object of class \code{dgr_graph}. New graph edges will all move either from
the nodes in the selection toward the newly created nodes (with the option
\code{direction = "from"}), or to the selected nodes already in the graph (using
\code{direction = "to"}). Optionally, set node \code{type} and edge \code{rel} values for
all the new nodes and edges created, respectively.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create an empty graph, add a node to it, select
# that node, and then add 5 more nodes to the graph
# with edges from the original node to all of the
# new nodes
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 1) \%>\%
  select_last_nodes_created() \%>\%
  add_n_nodes_ws(
    n = 5,
    direction = "from")

# Get the graph's nodes
graph \%>\% get_node_ids()

# Get the graph's edges
graph \%>\% get_edges()

# Create an empty graph, add a node to it, select
# that node, and then add 5 more nodes to the graph
# with edges toward the original node from all of
# the new nodes
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 1) \%>\%
  select_last_nodes_created() \%>\%
  add_n_nodes_ws(
    n = 5,
    direction = "to")

# Get the graph's nodes
graph \%>\% get_node_ids()

# Get the graph's edges
graph \%>\% get_edges()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_node_df.R
\name{get_node_df}
\alias{get_node_df}
\title{Get a node data frame from a graph}
\usage{
get_node_df(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A node data frame.
}
\description{
From a graph, obtain a node data frame with all current node attributes.
}
\examples{
# Create a graph
graph <-
  create_graph() |>
  add_n_nodes(
    n = 1,
    type = "a") |>
  select_last_nodes_created() |>
  add_n_nodes_ws(
    n = 5,
    direction = "from",
    type = "b") |>
  select_nodes_by_id(
    nodes = 1) |>
  set_node_attrs_ws(
    node_attr = value,
    value = 25.3) |>
  clear_selection() |>
  select_nodes_by_id(
    nodes = 2:4) |>
  set_node_attrs_ws(
    node_attr = color,
    value = "grey70") |>
  invert_selection() |>
  set_node_attrs_ws(
    node_attr = color,
    value = "grey80") |>
  clear_selection()

# Get the graph's internal node
# data frame (ndf)
graph |>
  get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_articulation_points.R
\name{get_articulation_points}
\alias{get_articulation_points}
\title{Get articulation points}
\usage{
get_articulation_points(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
a vector of node IDs.
}
\description{
Get the nodes in the graph that are identified as articulation points.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 10,
    m = 12,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = shape,
    values = "square")

# Get the articulation points
# in the graph (i.e., those
# nodes that if any were to be
# removed, the graph would
# become disconnected)
graph |>
  get_articulation_points()

# For the articulation points,
# change the node shape to
# a `circle`
graph <-
  graph |>
  select_nodes_by_id(
    nodes = get_articulation_points(graph)) |>
  set_node_attrs_ws(
    node_attr = shape,
    value = "circle")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_node_df.R
\name{create_node_df}
\alias{create_node_df}
\title{Create a node data frame}
\usage{
create_node_df(n, type = NULL, label = NULL, ...)
}
\arguments{
\item{n}{The total number of nodes to include in the node data frame.}

\item{type}{An optional \code{type} for each node.}

\item{label}{An optional \code{label} for each node.}

\item{...}{Additional attributes. Some are present in \code{\link[=node_aes]{node_aes()}}}
}
\value{
A node data frame (ndf).
}
\description{
Combine several vectors for nodes and their attributes into a data frame,
which can be combined with other similarly-generated data frames, or, added
to a graph object. A node data frame, or ndf, has at least the following
columns:
\itemize{
\item \code{id} (of type \code{integer})
\item \code{type} (of type \code{character})
\item \code{label} (of type \code{character})
}

An arbitrary number of additional columns containing aesthetic or data
attributes can be part of the ndf, see \code{\link[=node_aes]{node_aes()}} for additional attributes that can be used with ..., so long as they follow the aforementioned
columns.
}
\examples{
# Create a node data frame (ndf) where the labels
# are equivalent to the node ID values (this is not
# recommended); the `label` and `type` node
# attributes will always be a `character` class
# whereas `id` will always be an `integer`
node_df <-
  create_node_df(
    n = 4,
    type = c("a", "a", "b", "b"),
    label = TRUE)

# Display the node data frame
node_df

# Create an ndf with distinct labels and
# additional node attributes (where their classes
# will be inferred from the input vectors)
node_df <-
  create_node_df(
    n = 4,
    type = "a",
    label = c(2384, 3942, 8362, 2194),
    style = "filled",
    color = "aqua",
    shape = c("circle", "circle",
              "rectangle", "rectangle"),
    value = c(3.5, 2.6, 9.4, 2.7))

# Display the node data frame
node_df

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/colorize_edge_attrs.R
\name{colorize_edge_attrs}
\alias{colorize_edge_attrs}
\title{Apply colors based on edge attribute values}
\usage{
colorize_edge_attrs(
  graph,
  edge_attr_from,
  edge_attr_to,
  cut_points = NULL,
  palette = "Spectral",
  alpha = NULL,
  reverse_palette = FALSE,
  default_color = "#D9D9D9"
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr_from}{The name of the edge attribute column from which color
values will be based.}

\item{edge_attr_to}{The name of the new edge attribute to which the color
values will be applied.}

\item{cut_points}{An optional vector of numerical breaks for bucketizing
continuous numerical values available in a edge attribute column.}

\item{palette}{Can either be: (1) a palette name from the RColorBrewer
package (e.g., \code{Greens}, \code{OrRd}, \code{RdYlGn}), (2) \code{viridis}, which indicates
use of the \code{viridis} color scale from the package of the same name, or (3)
a vector of hexadecimal color names.}

\item{alpha}{An optional alpha transparency value to apply to the generated
colors. Should be in the range of \code{0} (completely transparent) to \code{100}
(completely opaque).}

\item{reverse_palette}{An option to reverse the order of colors in the chosen
palette. The default is \code{FALSE}.}

\item{default_color}{A hexadecimal color value to use for instances when the
values do not fall into the bucket ranges specified in the \code{cut_points}
vector.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal edge data frame (edf), use a categorical edge
attribute to generate a new edge attribute with color values.
}
\examples{
# Create a graph with 5
# nodes and 4 edges
graph <-
  create_graph() |>
  add_path(n = 5) |>
  set_edge_attrs(
    edge_attr = weight,
    values = c(3.7, 6.3, 9.2, 1.6))

# We can bucketize values in
# the edge `weight` attribute using
# `cut_points` and, by doing so,
# assign colors to each of the
# bucketed ranges (for values not
# part of any bucket, a gray color
# is assigned by default)
graph <-
  graph |>
  colorize_edge_attrs(
    edge_attr_from = weight,
    edge_attr_to = color,
    cut_points = c(0, 2, 4, 6, 8, 10),
    palette = "RdYlGn")

# Now there will be a `color`
# edge attribute with distinct
# colors (from the RColorBrewer
# Red-Yellow-Green palette)
graph |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/create_graph_series.R
\name{create_graph_series}
\alias{create_graph_series}
\title{Create a graph series object}
\usage{
create_graph_series(
  graph = NULL,
  series_name = NULL,
  series_type = "sequential"
)
}
\arguments{
\item{graph}{A graph object to add to the new graph series object.}

\item{series_name}{An optional name to ascribe to the series.}

\item{series_type}{Either a \code{sequential} type (the default) or a \code{temporal}
type (which requires date-time strings and time zone codes to be supplied).}
}
\value{
A graph series object of type \code{dgr_graph_1D}.
}
\description{
Create a graph series object for the storage of multiple graphs across a
sequential or temporal one-dimensional array.
}
\examples{
# Create three graphs
graph_1 <-
  create_graph() \%>\%
  add_path(n = 4)

graph_2 <-
  create_graph() \%>\%
  add_cycle(n = 5)

graph_3 <-
  create_graph() \%>\%
  add_star(n = 6)

# Create an empty graph series
# and add the graphs
series <-
  create_graph_series() \%>\%
  add_graph_to_graph_series(
    graph = graph_1) \%>\%
  add_graph_to_graph_series(
    graph = graph_2) \%>\%
  add_graph_to_graph_series(
    graph = graph_3)

# Count the number of graphs
# in the graph series
series \%>\%
  count_graphs_in_graph_series()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/delete_loop_edges_ws.R
\name{delete_loop_edges_ws}
\alias{delete_loop_edges_ws}
\title{Delete all loop edges associated with a selection of nodes}
\usage{
delete_loop_edges_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a selection of nodes in a graph, remove any associated loop edges.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create an undirected, full graph
# of 5 nodes with loops retained
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_full_graph(
    n = 5,
    keep_loops = TRUE)

# Select nodes `3` and `4`
# and remove the loop edges
# associated with those nodes
graph <-
  graph \%>\%
  select_nodes_by_id(
    nodes = 3:4) \%>\%
  delete_loop_edges_ws()

# Count the number of loop
# edges remaining in the graph
graph \%>\% count_loop_edges()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_degree_out.R
\name{get_degree_out}
\alias{get_degree_out}
\title{Get outdegree values for all nodes}
\usage{
get_degree_out(graph, normalized = FALSE)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{normalized}{set as \code{FALSE} (the default), the outdegree will be
provided for each of the nodes (as a count of edges outgoing from each
node). When set as \code{TRUE}, then the result for each node will be
divided by the total number of nodes in the graph minus 1.}
}
\value{
A data frame with outdegree values for each of the nodes.
}
\description{
Get the outdegree values for all nodes in a graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the outdegree values
# for all nodes in the graph
graph |>
  get_degree_out()

# Add the outdegree values
# to the graph as a node
# attribute
graph <-
  graph |>
  join_node_attrs(
    df = get_degree_out(graph))

# Display the graph's
# node data frame
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_alpha_centrality.R
\name{get_alpha_centrality}
\alias{get_alpha_centrality}
\title{Get the alpha centrality for all nodes}
\usage{
get_alpha_centrality(
  graph,
  alpha = 1,
  exo = 1,
  weights_attr = NULL,
  tol = 1e-07
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{alpha}{the parameter that specifies the relative importance of
endogenous versus exogenous factors in the determination of centrality.}

\item{exo}{the exogenous factors, in most cases this is either a constant
(which applies the same factor to every node), or a vector giving the
factor for every node.}

\item{weights_attr}{an optional name of the edge attribute to use in the
adjacency matrix. If \code{NULL} then, if it exists, the \code{weight} edge attribute
of the graph will be used. Failing that, the standard adjacency matrix will
be used in calculations.}

\item{tol}{the tolerance for near-singularities during matrix inversion. The
default value is set to \code{1e-7}.}
}
\value{
A data frame with alpha centrality scores for each of the nodes.
}
\description{
Get the alpha centrality values for all nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 10,
    m = 12,
    set_seed = 23)

# Get the alpha centrality scores
# for all nodes
graph \%>\%
  get_alpha_centrality()

# Add the alpha centrality
# scores to the graph as a node
# attribute
graph <-
  graph \%>\%
  join_node_attrs(
    df = get_alpha_centrality(.))

# Display the graph's node
# data frame
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edge_ids.R
\name{get_edge_ids}
\alias{get_edge_ids}
\title{Get a vector of edge ID values}
\usage{
get_edge_ids(graph, conditions = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{an option to use filtering conditions for the retrieval of
edges.}
}
\value{
A vector of edge ID values.
}
\description{
Obtain a vector of edge ID values from a graph object. An optional filter by
edge attribute can limit the set of edge ID values returned.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = "letter",
    color = c("red", "green", "grey", "blue"),
    value = c(3.5, 2.6, 9.4, 2.7))

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "leading_to",
    color = c("pink", "blue", "blue"),
    value = c(3.9, 2.5, 7.3))

# Create a graph
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Get a vector of all edges in a graph
graph |> get_edge_ids()

# Get a vector of edge ID values using a
# numeric comparison (i.e., all edges with
# `value` attribute greater than 3)
get_edge_ids(
  graph,
  conditions = value > 3)

# Get a vector of edge ID values using
# a match pattern (i.e., all edges with
# `color` attribute of `pink`)
get_edge_ids(
  graph,
  conditions = color == "pink")

# Use multiple conditions to return edges
# with the desired attribute values
get_edge_ids(
  graph,
  conditions =
    color == "blue" &
    value > 5)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_graph_to_graph_series.R
\name{add_graph_to_graph_series}
\alias{add_graph_to_graph_series}
\title{Add graph object to a graph series object}
\usage{
add_graph_to_graph_series(graph_series, graph)
}
\arguments{
\item{graph_series}{A graph series object to which the graph object will be
added.}

\item{graph}{A graph object to add to the graph series object.}
}
\value{
A graph series object of type \code{dgr_graph_1D}.
}
\description{
Add a graph object to an extant graph series object for storage of multiple
graphs across a sequential or temporal one-dimensional array.
}
\examples{
# Create three graphs
graph_1 <-
  create_graph() \%>\%
  add_path(n = 4)

graph_2 <-
  create_graph() \%>\%
  add_cycle(n = 5)

graph_3 <-
  create_graph() \%>\%
  add_star(n = 6)

# Create an empty graph series
# and add the graphs
series <-
  create_graph_series() \%>\%
  add_graph_to_graph_series(
    graph = graph_1) \%>\%
  add_graph_to_graph_series(
    graph = graph_2) \%>\%
  add_graph_to_graph_series(
    graph = graph_3)

# Count the number of graphs
# in the graph series
series \%>\%
  count_graphs_in_graph_series()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/DiagrammeR.R
\name{DiagrammeROutput}
\alias{DiagrammeROutput}
\title{Widget output function for use in Shiny}
\usage{
DiagrammeROutput(outputId, width = "100\%", height = "auto")
}
\arguments{
\item{outputId}{Output variable to read from}

\item{width}{A valid CSS unit for the width or a number, which will be
coerced to a string and have \code{px} appended.}

\item{height}{A valid CSS unit for the height or a number, which will be
coerced to a string and have \code{px} appended.}
}
\description{
Widget output function for use in Shiny
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_node_df.R
\name{add_node_df}
\alias{add_node_df}
\title{Add nodes from a node data frame to an existing graph object}
\usage{
add_node_df(graph, node_df)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_df}{A node data frame that is created using \code{\link[=create_node_df]{create_node_df()}}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph} add nodes from a node data frame to
that graph.
}
\examples{
# Create an empty graph
graph <- create_graph()

# Create a node data frame (ndf)
ndf <-
  create_node_df(n = 2)

# Add the node data frame to
# the graph object to create
# a graph with nodes
graph <-
  graph \%>\%
  add_node_df(
    node_df = ndf)

# Inspect the graph's ndf
graph \%>\% get_node_df()

# Create another ndf
ndf_2 <-
  create_node_df(n = 3)

# Add the second node data
# frame to the graph object
# to add more nodes with
# attributes to the graph
graph <-
  graph \%>\%
  add_node_df(
    node_df = ndf_2)

# View the graph's internal
# node data frame using the
# `get_node_df()` function
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edge_df.R
\name{get_edge_df}
\alias{get_edge_df}
\title{Get an edge data frame from a graph}
\usage{
get_edge_df(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
An edge data frame.
}
\description{
From a graph, obtain an edge data frame with all current edge attributes.
}
\examples{
# Create a graph
graph <-
  create_graph() \%>\%
  add_n_nodes(
    n = 1,
    type = "a") \%>\%
  select_last_nodes_created() \%>\%
  add_n_nodes_ws(
    n = 5,
    direction = "from",
    type = "b") \%>\%
  select_edges_by_node_id(
    nodes = 3:5) \%>\%
  set_edge_attrs_ws(
    edge_attr = color,
    value = "green") \%>\%
  set_edge_attrs_ws(
    edge_attr = rel,
    value = "a") \%>\%
  invert_selection \%>\%
  set_edge_attrs_ws(
    edge_attr = color,
    value = "blue") \%>\%
  set_edge_attrs_ws(
    edge_attr = rel,
    value = "b") \%>\%
  clear_selection()

# Get the graph's internal
# edge data frame (edf)
graph \%>\% get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_loop_edges.R
\name{count_loop_edges}
\alias{count_loop_edges}
\title{Get count of all loop edges}
\usage{
count_loop_edges(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A numeric vector of single length.
}
\description{
From a graph object of class \code{dgr_graph}, get a count of all loop edges in
the graph.
}
\examples{
# Create an undirected, full graph
# with 3 nodes and all possible
# edges, including loop edges
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_full_graph(
    n = 3,
    keep_loops = TRUE)

# Get a count of all loop edges
# in the graph
graph \%>\% count_loop_edges()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edge_df_ws.R
\name{get_edge_df_ws}
\alias{get_edge_df_ws}
\title{Get the graph's edf filtered by a selection of edges}
\usage{
get_edge_df_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
an edge data frame.
}
\description{
From a graph object of class \code{dgr_graph}, get the graph's internal edge data
frame that is filtered by the edge ID values currently active as a selection.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 4,
    m = 4,
    set_seed = 23) \%>\%
  set_edge_attrs(
    edge_attr = value,
    values = c(2.5, 8.2, 4.2, 2.4))

# Select edges with ID values
# `1` and `3`
graph <-
  graph \%>\%
  select_edges_by_edge_id(
    edges = c(1, 3))

# Get the edge data frame that's
# limited to the rows that correspond
# to the edge selection
graph \%>\% get_edge_df_ws()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mermaid.R
\name{mermaid}
\alias{mermaid}
\title{R + mermaid.js}
\usage{
mermaid(diagram = "", ..., width = NULL, height = NULL)
}
\arguments{
\item{diagram}{Diagram in mermaid markdown-like language or file (as a
connection or file name) containing a diagram specification. If no diagram
is provided \code{diagram = ""} then the function will assume that a diagram
will be provided by \code{\link[htmltools:builder]{htmltools::tags()}} and \code{DiagrammeR} is just being used
for dependency injection.}

\item{...}{Other arguments and parameters you would like to send to
JavaScript.}

\item{width}{The width of the resulting graphic in pixels.}

\item{height}{The height of the resulting graphic in pixels.}
}
\value{
An object of class \code{htmlwidget} that will intelligently print itself
into HTML in a variety of contexts including the R console, within R
Markdown documents, and within Shiny output bindings.
}
\description{
Make diagrams in R using
\href{https://github.com/mermaid-js/mermaid/wiki}{mermaid.js} with
infrastructure provided by \href{http://www.htmlwidgets.org/}{htmlwidgets}.
}
\examples{
# Create a simple graph running left to right (note
# that the whitespace is not important)
# DiagrammeR("
#   graph LR
#     A-->B
#     A-->C
#     C-->E
#     B-->D
#     C-->D
#     D-->F
#     E-->F
# ")

# Create the equivalent graph but have it running
# from top to bottom
# DiagrammeR("
#    graph TB
#    A-->B
#    A-->C
#    C-->E
#    B-->D
#    C-->D
#    D-->F
#    E-->F
# ")

# Create a graph with different node shapes and
# provide fill styles for each node
# DiagrammeR("graph LR;A(Rounded)-->B[Squared];B-->C{A Decision};
#  C-->D[Square One];C-->E[Square Two];
#  style A fill:#E5E25F;  style B fill:#87AB51; style C fill:#3C8937;
#  style D fill:#23772C;  style E fill:#B6E6E6;"
# )

# Load in the 'mtcars' dataset
# data(mtcars)
# connections <- sapply(
#  1:ncol(mtcars)
#   ,function(i) {
#      paste0(
#         i
#       ,"(",colnames(mtcars)[i],")---"
#       ,i,"-stats("
#       ,paste0(
#         names(summary(mtcars[,i]))
#         ,": "
#         ,unname(summary(mtcars[,i]))
#         ,collapse="<br/>"
#       )
#       ,")"
#    )
#  }
# )

# Create a diagram using the 'connections' object
# DiagrammeR(
#    paste0(
#      "graph TD;", "\n",
#      paste(connections, collapse = "\n"),"\n",
#      "classDef column fill:#0001CC, stroke:#0D3FF3, stroke-width:1px;" ,"\n",
#      "class ", paste0(1:length(connections), collapse = ","), " column;"
#    )
#  )

# Also with `DiagrammeR()`, you can use tags
# from `htmltools` (just make sure to use
# `class = "mermaid"`)
library(htmltools)
# diagramSpec = "
# graph LR;
#   id1(Start)-->id2(Stop);
#   style id1 fill:#f9f,stroke:#333,stroke-width:4px;
#   style id2 fill:#ccf,stroke:#f66,stroke-width:2px,stroke-dasharray: 5, 5;
# "
# html_print(tagList(
#   tags$h1("R + mermaid.js = Something Special")
#   ,tags$pre(diagramSpec)
#   ,tags$div(class="mermaid", diagramSpec)
#   ,DiagrammeR()
# ))

# Create a sequence diagram
# DiagrammeR("
# sequenceDiagram;
#    customer->>ticket seller: ask for a ticket;
#    ticket seller->>database: seats;
#    alt tickets available
#      database->>ticket seller: ok;
#      ticket seller->>customer: confirm;
#      customer->>ticket seller: ok;
#      ticket seller->>database: book a seat;
#      ticket seller->>printer: print a ticket;
#    else sold out
#      database->>ticket seller: none left;
#      ticket seller->>customer: sorry;
#    end
# ")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_forward_edges_ws.R
\name{add_forward_edges_ws}
\alias{add_forward_edges_ws}
\title{Add new edges with identical definitions as with a selection of edges}
\usage{
add_forward_edges_ws(graph, rel = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{rel}{An optional string to apply a \code{rel} attribute to all newly created
edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add edges in the same direction of one or more edges available as an edge
selection in a graph object of class \code{dgr_graph}. New graph edges have the
same edge definitions as those in the selection except with new edge ID
values. There is also the option to assign a common \code{rel} grouping to the
newly created edges. Upon addition of the edges, the edge selection will be
retained for further selection or traversal operations.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create an empty graph, add 2 nodes
# to it, and create the edge `1->2`
graph <-
  create_graph() |>
  add_n_nodes(
    n = 2,
    type = "type_a",
    label = c("a_1", "a_2")) |>
  add_edge(
    from = 1, to = 2, rel = "a")

# Get the graph's edges
graph |> get_edge_ids()

# Select the edge and create 2
# additional edges with the same
# definition (`1->2`) but with
# different `rel` values (`b` and `c`)
graph <-
  graph |>
  select_edges() |>
  add_forward_edges_ws(rel = "b") |>
  add_forward_edges_ws(rel = "c") |>
  clear_selection()

# Get the graph's edge data frame
graph |> get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_node_attrs.R
\name{set_node_attrs}
\alias{set_node_attrs}
\title{Set node attribute values}
\usage{
set_node_attrs(graph, node_attr, values, nodes = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr}{The name of the attribute to set. Examples of attributes can be found in \code{\link[=node_aes]{node_aes()}}.}

\item{values}{The values to be set for the chosen attribute for the chosen
nodes.}

\item{nodes}{An optional vector of node IDs for filtering the list of nodes
present in the graph.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph}, set node attribute values for one
or more nodes.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    type = "basic",
    label = TRUE,
    value = c(3.5, 2.6, 9.4, 2.7))

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "leading_to")

# Create a graph
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Set attribute `color = "green"` for
# nodes `1` and `3` using the graph object
graph <-
  graph \%>\%
  set_node_attrs(
    node_attr = color,
    values = "green",
    nodes = c(1, 3))

# View the graph's node data frame
graph \%>\% get_node_df()

# Set attribute `color = "blue"` for
# all nodes in the graph
graph <-
  graph \%>\%
  set_node_attrs(
    node_attr = color,
    values = "blue")

# Display the graph's ndf
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_cache.R
\name{get_cache}
\alias{get_cache}
\title{Get a cached vector from a graph object}
\usage{
get_cache(graph, name = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{name}{the name of the object to extract from the cache. If none
supplied, the most recent object added to the cache will be returned.}
}
\value{
A vector.
}
\description{
Get the vector cached in a graph object of class \code{dgr_graph}.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a graph with 5 nodes and 5 edges
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 5) \%>\%
  set_node_attrs(
    node_attr = value,
    values = rnorm(
      n = count_nodes(.),
      mean = 8,
      sd = 2)) \%>\%
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->2")

# Cache all values from the node attribute `value`
# as a numeric vector
graph <-
  graph \%>\%
  set_cache(
    name = "value",
    to_cache = get_node_attrs(
      graph = .,
      node_attr = value))

# Return the cached vector
graph \%>\% get_cache()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_multiedge_count.R
\name{get_multiedge_count}
\alias{get_multiedge_count}
\title{Get the count of multiple edges}
\usage{
get_multiedge_count(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A vector with a single, numerical value.
}
\description{
Get a count of the number of multiple edges in the graph. Included in the
count is the number of separate edges that share the same edge definition
(i.e., same pair of nodes) across the entire graph. So, for example, if there
are 2 edge definitions in the graph that involve 6 separate edge IDs, the
count will be \code{4}.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 5,
    label = TRUE)

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 4, 4, 3, 5, 1, 3, 4),
      to = c(4, 1, 1, 2, 2, 2, 2, 1))

# Create a graph with the ndf and edf
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Get the total number of multiple
# edges (those edges that share an
# edge definition) in the graph
graph \%>\% get_multiedge_count()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_eigen_centrality.R
\name{get_eigen_centrality}
\alias{get_eigen_centrality}
\title{Get the eigen centrality for all nodes}
\usage{
get_eigen_centrality(graph, weights_attr = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{weights_attr}{An optional name of the edge attribute to use in the
adjacency matrix. If \code{NULL} then, if it exists, the \code{weight} edge attribute
of the graph will be used. If \code{NA} then no edge weights will be used.}
}
\value{
A data frame with eigen centrality scores for each of the nodes.
}
\description{
Get the eigen centrality values for all nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10, m = 15,
    set_seed = 23)

# Get the eigen centrality scores
# for nodes in the graph
graph \%>\% get_eigen_centrality()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edge_attrs.R
\name{get_edge_attrs}
\alias{get_edge_attrs}
\title{Get edge attribute values}
\usage{
get_edge_attrs(graph, edge_attr, from = NULL, to = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr}{the name of the attribute for which to get values.}

\item{from}{an optional vector of node IDs from which the edge is outgoing
for filtering the list of edges.}

\item{to}{an optional vector of node IDs from which the edge is incoming for
filtering the list of edges.}
}
\value{
A named vector of edge attribute values for the attribute given by
\code{edge_attr} by edge.
}
\description{
From a graph object of class \code{dgr_graph}, get edge attribute values for one
or more edges.
}
\examples{
# Create a simple graph where
# edges have an edge attribute
# named `value`
graph <-
  create_graph() |>
  add_n_nodes(n = 4)

edges <-
  create_edge_df(
    from = c(1, 2, 1, 4),
      to = c(2, 3, 4, 3),
     rel = "rel")

graph <-
  add_edge_df(
    graph = graph,
    edge_df = edges) |>
  set_edge_attrs(
    edge_attr = value,
    values = 1.6,
    from = 1,
      to = 2) |>
  set_edge_attrs(
    edge_attr = value,
    values = 4.3,
    from = 1,
      to = 4) |>
  set_edge_attrs(
    edge_attr = value,
    values = 2.9,
    from = 2,
      to = 3) |>
  set_edge_attrs(
    edge_attr = value,
    values = 8.4,
    from = 4,
      to = 3)

# Get the values for the
# `value` edge attribute
graph |>
  get_edge_attrs(
    edge_attr = value)

# To only return edge attribute
# values for specified edges, use
# the `from` and `to` arguments
graph |>
  get_edge_attrs(
    edge_attr = value,
    from = c(1, 2),
      to = c(2, 3))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/is_edge_multiple.R
\name{is_edge_multiple}
\alias{is_edge_multiple}
\title{Is the edge a multiple edge?}
\usage{
is_edge_multiple(graph, edge)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge}{A numeric edge ID value.}
}
\value{
A logical value.
}
\description{
Determines whether an edge definition has multiple edge IDs associated with
the same node pair.
}
\examples{
# Create a graph that has multiple
# edges across some node pairs
graph <-
  create_graph() |>
  add_path(n = 4) |>
  add_edge(
    from = 1,
    to = 2) |>
  add_edge(
    from = 3,
    to = 4)

# Get the graph's internal
# edge data frame
graph |> get_edge_df()

# Determine if edge `1` is
# a multiple edge
graph |>
  is_edge_multiple(edge = 1)

# Determine if edge `2` is
# a multiple edge
graph |>
  is_edge_multiple(edge = 2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_node_clones_ws.R
\name{add_node_clones_ws}
\alias{add_node_clones_ws}
\title{Add clones of a selection of nodes}
\usage{
add_node_clones_ws(graph, add_edges = FALSE, direction = NULL, label = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{add_edges}{An option for whether to add edges from the selected nodes
to each of their clones, or, in the opposite direction.}

\item{direction}{Using \code{from} will create new edges from existing nodes to
the new, cloned nodes. The \code{to} option will create new edges directed
toward the existing nodes.}

\item{label}{An optional vector of node label values. The vector length
should correspond to the number of nodes in the active selection of nodes.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add new nodes to a graph object of class \code{dgr_graph} which are clones of
nodes in an active selection of nodes. All node attributes are preserved
except for the node \code{label} attribute (to maintain the uniqueness of non-\code{NA}
node label values). A vector of node \code{label} can be provided to bind new
labels to the cloned nodes.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create a graph with a path of
# nodes; supply `label`, `type`,
# and `value` node attributes,
# and select the created nodes
graph <-
  create_graph() |>
  add_path(
    n = 3,
    label = c("d", "g", "r"),
    type = c("a", "b", "c")) |>
  select_last_nodes_created()

# Display the graph's internal
# node data frame
graph |> get_node_df()

# Create clones of all nodes
# in the selection but assign
# new node label values
# (leaving `label` as NULL
# yields NA values)
graph <-
  graph |>
  add_node_clones_ws(
    label = c("a", "b", "v"))

# Display the graph's internal
# node data frame: nodes `4`,
# `5`, and `6` are clones of
# `1`, `2`, and `3`
graph |> get_node_df()

# Select the last nodes
# created (`4`, `5`, and `6`)
# and clone those nodes and
# their attributes while
# creating new edges between
# the new and existing nodes
graph <-
  graph |>
  select_last_nodes_created() |>
  add_node_clones_ws(
    add_edges = TRUE,
    direction = "to",
    label = c("t", "z", "s"))

# Display the graph's internal
# edge data frame; there are
# edges between the selected
# nodes and their clones
graph |> get_edge_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_ndfs.R
\name{combine_ndfs}
\alias{combine_ndfs}
\title{Combine multiple node data frames}
\usage{
combine_ndfs(...)
}
\arguments{
\item{...}{Two or more node data frames, which contain node IDs and
associated attributes.}
}
\value{
A combined node data frame.
}
\description{
Combine several node data frames into a single node data frame.
}
\examples{
# Create two node data frames
node_df_1 <-
  create_node_df(
    n = 2,
    type = c("a", "b"),
    label = c("D", "Z"),
    value = c(8.4, 3.4))

node_df_2 <-
  create_node_df(
    n = 2,
    type = c("b", "c"),
    label = c("U", "A"),
    value = c(0.4, 3.4))

# Combine the ndfs using the
# `combine_ndfs()` function
node_df_combined <-
  combine_ndfs(
    node_df_1,
    node_df_2)

# Inspect the combined ndf
node_df_combined

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_graph_log.R
\name{get_graph_log}
\alias{get_graph_log}
\title{Get the graph log information}
\usage{
get_graph_log(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A \code{df_tbl} object.
}
\description{
Get a tibble of the graph log, which contains information on the functions
called on the graph that resulted in some transformation of the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function and
# delete 2 nodes from the graph
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23) \%>\%
  delete_node(node = 5) \%>\%
  delete_node(node = 7)

# Get the graph log, which is a
# record of all graph transformations
graph \%>\% get_graph_log()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_edge_clone.R
\name{add_edge_clone}
\alias{add_edge_clone}
\title{Add a clone of an existing edge to the graph}
\usage{
add_edge_clone(graph, edge, from, to)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge}{An edge ID corresponding to the graph edge to be cloned.}

\item{from}{The outgoing node from which the edge is connected.}

\item{to}{The incoming nodes to which each edge is connected.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add a new edge to a graph object of class \code{dgr_graph} which is a clone of an
edge already in the graph. All edge attributes are preserved.
}
\examples{
# Create a graph with a path of
# 2 nodes; supply a common `rel`
# edge attribute for all edges
# in this path and then add a
# `color` edge attribute
graph <-
  create_graph() \%>\%
  add_path(
    n = 2,
    rel = "a") \%>\%
  select_last_edges_created() \%>\%
  set_edge_attrs(
    edge_attr = color,
    values = "steelblue") \%>\%
  clear_selection()

# Display the graph's internal
# edge data frame
graph \%>\% get_edge_df()

# Create a new node (will have
# node ID of `3`) and then
# create an edge between it and
# node `1` while reusing the edge
# attributes of edge `1` -> `2`
# (edge ID `1`)
graph_2 <-
  graph \%>\%
  add_node() \%>\%
  add_edge_clone(
    edge = 1,
    from = 3,
      to = 1)

# Display the graph's internal
# edge data frame
graph_2 \%>\% get_edge_df()

# The same change can be performed
# with some helper functions in the
# `add_edge_clone()` function call
graph_3 <-
  graph \%>\%
    add_node() \%>\%
    add_edge_clone(
      edge = get_last_edges_created(.),
      from = get_last_nodes_created(.),
      to = 1)

# Display the graph's internal
# edge data frame
graph_3 \%>\% get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_cmty_louvain.R
\name{get_cmty_louvain}
\alias{get_cmty_louvain}
\title{Get community membership by Louvain optimization}
\usage{
get_cmty_louvain(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame with group membership assignments for each of the nodes.
}
\description{
Through the use of multi-level optimization of a modularity score, obtain the
group membership values for each of the nodes in the graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the group membership values
# for all nodes in the graph
# through the multi-level
# optimization of modularity
# algorithm
graph \%>\%
  get_cmty_louvain()

# Add the group membership
# values to the graph as a
# node attribute
graph <-
  graph \%>\%
  join_node_attrs(
    df = get_cmty_louvain(.))

# Display the graph's
# node data frame
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_node_attr_to_display.R
\name{set_node_attr_to_display}
\alias{set_node_attr_to_display}
\title{Set the node attribute values to be rendered}
\usage{
set_node_attr_to_display(graph, attr = NULL, nodes = NULL, default = "label")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{attr}{The name of the attribute from which label text for the node will
be obtained. If set to \code{NULL}, then \code{NA} values will be assigned to the
\code{display} column for the chosen nodes.}

\item{nodes}{A length vector containing one or several node ID values (as
integers) for which node attributes are set for display in the rendered
graph. If \code{NULL}, all nodes from the graph are assigned the \code{display} value
given as \code{attr}.}

\item{default}{The name of an attribute to set for all other graph nodes not
included in \code{nodes}. This value only gets used if the \code{display} node
attribute is not in the graph's internal node data frame.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Set a node attribute type to display as node text when calling the
\code{\link[=render_graph]{render_graph()}} function. This allows for display of different types of node
attribute values on a per-node basis. Without setting the \code{display}
attribute, rendering a graph will default to printing text from the \code{label}
attribute on nodes. Setting the \code{display} node attribute with this function
for the first time (i.e., the \code{display} column doesn't exist in the graph's
internal node data frame) will insert the \code{attr} value for all nodes
specified in \code{nodes} and a default value (\code{default}) for all remaining nodes.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 4,
    m = 4,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = value,
    values = c(2.5, 8.2, 4.2, 2.4))

# For node ID values of `1`,
# `2`, and `3`, choose to display
# the node `value` attribute (for
# the other nodes, display nothing)
graph <-
  graph |>
  set_node_attr_to_display(
    nodes = 1:3,
    attr = value,
    default = NA)

# Show the graph's node data frame; the
# `display` node attribute will show for
# each row, which node attribute value to
# display when the graph is rendered
graph |> get_node_df()

# This function can be called multiple
# times on a graph; after the first time
# (i.e., creation of the `display`
# attribute), the `default` value won't
# be used
graph |>
  set_node_attr_to_display(
    nodes = 4,
    attr = label) |>
  set_node_attr_to_display(
    nodes = c(1, 3),
    attr = id) |>
  get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_reciprocity.R
\name{get_reciprocity}
\alias{get_reciprocity}
\title{Get the graph reciprocity}
\usage{
get_reciprocity(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single, numerical value that is the ratio value of reciprocal edges
over all graph edges.
}
\description{
Get the reciprocity of a directed graph. The reciprocity of a graph is the
fraction of reciprocal edges (e.g., \code{1} -> \code{2} and \code{2} -> \code{1}) over all edges
available in the graph. Note that for an undirected graph, all edges are
reciprocal. This function does not consider loop edges (e.g., \code{1} -> \code{1}).
}
\examples{
# Define a graph where 2 edge definitions
# have pairs of reciprocal edges
graph <-
  create_graph() |>
  add_cycle(n = 3) |>
  add_node(
    from = 1,
      to = 1) |>
  add_node(
    from = 1,
      to = 1)

# Get the graph reciprocity, which will
# be calculated as the ratio 4/7 (where
# 4 is the number reciprocating edges
# and 7 is the total number of edges
# in the graph)
graph |>
  get_reciprocity()

# For an undirected graph, all edges
# are reciprocal, so the ratio will
# always be 1
graph |>
  set_graph_undirected() |>
  get_reciprocity()

# For a graph with no edges, the graph
# reciprocity cannot be determined (and
# the same NA result is obtained from an
# empty graph)
create_graph() |>
  add_n_nodes(n = 5) |>
  get_reciprocity()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_jaccard_similarity.R
\name{get_jaccard_similarity}
\alias{get_jaccard_similarity}
\title{Get Jaccard similarity coefficient scores}
\usage{
get_jaccard_similarity(graph, nodes = NULL, direction = "all", round_to = 3)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{nodes}{An optional vector of node IDs to consider for Jaccard
similarity scores. If not supplied, then similarity scores will be provided
for every pair of nodes in the graph.}

\item{direction}{Using \code{all} (the default), the function will ignore edge
direction when determining scores for neighboring nodes. With \code{out} and
\verb{in}, edge direction for neighboring nodes will be considered.}

\item{round_to}{The maximum number of decimal places to retain for the
Jaccard similarity coefficient scores. The default value is \code{3}.}
}
\value{
A matrix with Jaccard similarity values for each pair of nodes
considered.
}
\description{
Get the Jaccard similarity coefficient scores for one or more nodes in a
graph.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) \%>\%
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get the Jaccard similarity
# values for nodes `5`, `6`,
# and `7`
graph \%>\%
  get_jaccard_similarity(
    nodes = 5:7)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mutate_node_attrs_ws.R
\name{mutate_node_attrs_ws}
\alias{mutate_node_attrs_ws}
\title{Mutate node attribute values for a selection of nodes}
\usage{
mutate_node_attrs_ws(graph, ...)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{...}{Expressions used for the mutation of node attributes. LHS of each
expression is either an existing or new node attribute name. The RHS can
consist of any valid R code that uses node attributes as variables.
Expressions are evaluated in the order provided, so, node attributes
created or modified are ready to use in subsequent expressions.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal node data frame (ndf), mutate node attribute values
only for nodes in a selection by using one or more expressions.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create a graph with 3 nodes
# and then select node `1`
graph <-
  create_graph() \%>\%
  add_path(n = 3) \%>\%
  set_node_attrs(
    node_attr = width,
    values = c(1.4, 0.3, 1.1)) \%>\%
  select_nodes(nodes = 1)

# Get the graph's internal ndf
# to show which node attributes
# are available
graph \%>\% get_node_df()

# Mutate the `width` node
# attribute for the nodes
# only in the active selection
# of nodes (node `1`); here,
# we divide each value in the
# selection by 2
graph <-
  graph \%>\%
  mutate_node_attrs_ws(
    width = width / 2)

# Get the graph's internal
# ndf to show that the node
# attribute `width` was
# mutated only for node `1`
graph \%>\% get_node_df()

# Create a new node attribute,
# called `length`, that is the
# log of values in `width` plus
# 2 (and, also, round all values
# to 2 decimal places)
graph <-
  graph \%>\%
  clear_selection() \%>\%
  select_nodes(nodes = 2:3) \%>\%
  mutate_node_attrs_ws(
    length = (log(width) + 2) \%>\%
               round(2))

# Get the graph's internal ndf
# to show that the node attribute
# values had been mutated only
# for nodes `2` and `3` (since
# node `1` is excluded, an NA
# value is applied)
graph \%>\% get_node_df()

# Create a new node attribute
# called `area`, which is the
# product of the `width` and
# `length` attributes
graph <-
  graph \%>\%
  mutate_node_attrs_ws(
    area = width * length)

# Get the graph's internal ndf
# to show that the node attribute
# values had been multiplied
# together (with new attr `area`)
# for nodes `2` and `3`
graph \%>\% get_node_df()

# We can invert the selection
# and mutate node `1` several
# times to get an `area` value
# for that node
graph <-
  graph \%>\%
  invert_selection() \%>\%
  mutate_node_attrs_ws(
    length = (log(width) + 5) \%>\%
               round(2),
    area = width * length)

# Get the graph's internal ndf
# to show that the 2 mutations
# occurred for node `1`, yielding
# non-NA values for its node
# attributes without changing
# those of the other nodes
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/docs_node_edge_aes_data.R
\name{node_edge_aes_data}
\alias{node_edge_aes_data}
\title{Options for node and edge aesthetics and data}
\arguments{
\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\description{
Options for node and edge aesthetics and data
}
\seealso{
Other aesthetics: 
\code{\link{edge_aes}()},
\code{\link{node_aes}()}
}
\concept{aesthetics}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_edges.R
\name{count_edges}
\alias{count_edges}
\title{Get a count of all edges}
\usage{
count_edges(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single-length numeric vector.
}
\description{
From a graph object of class \code{dgr_graph}, get a count of edges in the graph.
}
\examples{
# Create a graph with a
# path of nodes and 3
# unconnected nodes
graph <-
  create_graph() \%>\%
  add_path(n = 3) \%>\%
  add_n_nodes(n = 3)

# Get a count of all edges
# in the graph
graph \%>\%
  count_edges()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_nbrs.R
\name{get_nbrs}
\alias{get_nbrs}
\title{Get all neighbors of one or more nodes}
\usage{
get_nbrs(graph, nodes)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{nodes}{A vector of node ID values.}
}
\value{
A vector of node ID values.
}
\description{
With one or more nodes, get the set of all neighboring nodes.
}
\examples{
# Create a simple, directed graph with 5
# nodes and 4 edges
graph <-
  create_graph() \%>\%
  add_path(n = 5)

# Find all neighbor nodes for node `2`
graph \%>\% get_nbrs(nodes = 2)

# Find all neighbor nodes for nodes `1`
# and `5`
graph \%>\% get_nbrs(nodes = c(1, 5))

# Color node `3` with purple, get its
# neighbors and color those nodes green
graph <-
  graph \%>\%
  select_nodes_by_id(nodes = 3) \%>\%
  set_node_attrs_ws(
    node_attr = color,
    value = "purple") \%>\%
  clear_selection() \%>\%
  select_nodes_by_id(
    nodes = get_nbrs(
      graph = .,
      nodes = 3)) \%>\%
  set_node_attrs_ws(
    node_attr = color,
    value = "green")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_last_nodes_created.R
\name{select_last_nodes_created}
\alias{select_last_nodes_created}
\title{Select the last set of nodes created in a graph}
\usage{
select_last_nodes_created(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Select the last nodes that were created in a graph object of class
\code{dgr_graph}. This function should ideally be used just after creating the
nodes to be selected.
}
\examples{
# Create a graph and add 4 nodes
# in 2 separate function calls
graph <-
  create_graph() |>
  add_n_nodes(
    n = 2,
    type = "a",
    label = c("a_1", "a_2")) |>
  add_n_nodes(
    n = 2,
    type = "b",
    label = c("b_1", "b_2"))

# Select the last nodes created (2 nodes
# from the last function call) and then
# set their color to be `red`
graph <-
  graph |>
  select_last_nodes_created() |>
  set_node_attrs_ws(
    node_attr = color,
    value = "red") |>
  clear_selection()

# Display the graph's internal node
# data frame to verify the change
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_n_nodes.R
\name{add_n_nodes}
\alias{add_n_nodes}
\title{Add one or several unconnected nodes to the graph}
\usage{
add_n_nodes(
  graph,
  n,
  type = NULL,
  label = NULL,
  node_aes = NULL,
  node_data = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{n}{The number of new nodes to add to the graph.}

\item{type}{An optional character vector that provides group identifiers for
the nodes to be added.}

\item{label}{An optional character object that describes the nodes to be
added.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add \code{n} new nodes to a graph object of class \code{dgr_graph}. Optionally, set
node \code{type} values for the new nodes.
}
\examples{
# Create an empty graph and
# add 5 nodes; these nodes
# will be assigned ID values
# from `1` to `5`
graph <-
  create_graph() \%>\%
  add_n_nodes(n = 5)

# Get the graph's node IDs
graph \%>\% get_node_ids()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_nodes_from_df_cols.R
\name{add_nodes_from_df_cols}
\alias{add_nodes_from_df_cols}
\title{Add nodes from distinct values in data frame columns}
\usage{
add_nodes_from_df_cols(
  graph,
  df,
  columns,
  type = NULL,
  keep_duplicates = FALSE
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{df}{A data frame from which values will be taken as new nodes for the
graph.}

\item{columns}{A character vector of column names or a numeric vector of
column numbers for the data frame supplied in \code{df}. The distinct values in
these columns will serve as labels for the nodes added to the graph.}

\item{type}{An optional, single-length character vector that provides a group
identifier for the nodes to be added to the graph.}

\item{keep_duplicates}{An option to exclude incoming nodes where the labels
(i.e., values found in columns of the specified \code{df}) match label values
available in the graph's nodes. By default, this is set to \code{FALSE}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add new nodes to a graph object of class \code{dgr_graph} using distinct values
from one or more columns in a data frame. The values will serve as node
labels and the number of nodes added depends on the number of distinct values
found in the specified columns.
}
\examples{
# Create an empty graph
graph <- create_graph()

# Create a data frame from
# which several columns have
# values designated as graph nodes
df <-
  data.frame(
    col_1 = c("f", "p", "q"),
    col_2 = c("q", "x", "f"),
    col_3 = c(1, 5, 3),
    col_4 = c("a", "v", "h"),
    stringsAsFactors = FALSE)

# Add nodes from columns `col_1`
# and `col_2` from the data frame
# to the graph object
graph <-
  graph \%>\%
  add_nodes_from_df_cols(
    df = df,
    columns = c("col_1", "col_2"))

# Show the graph's node data
# frame; duplicate labels are
# prevented with `keep_duplicates =
# FALSE`)
graph \%>\% get_node_df()

# Add new nodes from columns 3 and 4;
# We can specify the columns by their
# numbers as well
graph <-
  graph \%>\%
  add_nodes_from_df_cols(
    df = df,
    columns = 3:4)

# Show the graph's node data
# frame; note that nodes didn't
# get made with columns that
# are not character class columns
graph \%>\% get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_cmty_fast_greedy.R
\name{get_cmty_fast_greedy}
\alias{get_cmty_fast_greedy}
\title{Get community membership by modularity optimization}
\usage{
get_cmty_fast_greedy(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
a data frame with group membership assignments for each of the nodes.
}
\description{
Through the use of greedy optimization of a modularity score, obtain the
group membership values for each of the nodes in the graph. Note that this
method only works on graphs without multiple edges.
}
\examples{
# Create a graph with a
# balanced tree
graph <-
  create_graph() |>
  add_balanced_tree(
    k = 2,
    h = 2)

# Get the group membership
# values for all nodes in
# the graph through the greedy
# optimization of modularity
# algorithm
graph |>
  get_cmty_fast_greedy()

# Add the group membership
# values to the graph as a
# node attribute
graph <-
  graph |>
  join_node_attrs(
    df = get_cmty_fast_greedy(graph))

# Display the graph's
# node data frame
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trigger_graph_actions.R
\name{trigger_graph_actions}
\alias{trigger_graph_actions}
\title{Trigger the execution of a series of graph actions}
\usage{
trigger_graph_actions(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Execute the graph actions stored in the graph through the use of the
\code{\link[=add_graph_action]{add_graph_action()}} function. These actions will be invoked in order and any
errors encountered will trigger a warning message and result in no change to
the input graph. Normally, graph actions are automatically triggered at every
transformation step but this function allows for the manual triggering of
graph actions after setting them, for example.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 5,
    m = 10,
    set_seed = 23)

# Add a graph action that sets a node
# attr column with a function; this
# uses the `get_pagerank()` function
# to provide PageRank values in the
# `pagerank` column
graph <-
  graph |>
  add_graph_action(
    fcn = "set_node_attr_w_fcn",
    node_attr_fcn = "get_pagerank",
    column_name = "pagerank",
    action_name = "get_pagerank")

# Add a second graph action (to be
# executed after the first one) that
# rescales values in the `pagerank`
# column between 0 and 1, and, puts
# these values in the `width` column
graph <-
  graph |>
  add_graph_action(
    fcn = "rescale_node_attrs",
    node_attr_from = "pagerank",
    node_attr_to = "width",
    action_name = "pgrnk_to_width")

# Add a third and final graph action
# (to be executed last) that creates
# color values in the `fillcolor` column,
# based on the numeric values from the
# `width` column
graph <-
  graph |>
  add_graph_action(
    fcn = "colorize_node_attrs",
    node_attr_from = "width",
    node_attr_to = "fillcolor",
    action_name = "pgrnk_fillcolor")

# View the graph actions for the graph
# object by using the `get_graph_actions()`
# function
graph |> get_graph_actions()

# Manually trigger to invocation of
# the graph actions using the
# `trigger_graph_actions()` function
graph <-
  graph |>
  trigger_graph_actions()

# Examine the graph's internal node
# data frame (ndf) to verify that
# the `pagerank`, `width`, and
# `fillcolor` columns are present
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/count_asymmetric_node_pairs.R
\name{count_asymmetric_node_pairs}
\alias{count_asymmetric_node_pairs}
\title{Get the number of asymmetrically-connected node pairs}
\usage{
count_asymmetric_node_pairs(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single numeric value representing the number of
asymmetrically-connected node pairs.
}
\description{
Get the number of asymmetrically-connected node pairs. This works for
directed graphs.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() |>
  add_cycle(n = 5)

# Get a count of asymmetrically-
# connected node pairs
graph |>
  count_asymmetric_node_pairs()

# Create a full graph and then
# count the asymmetrically-
# connected node pairs
create_graph() |>
  add_full_graph(n = 10) |>
  count_asymmetric_node_pairs()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/spectools.R
\name{replace_in_spec}
\alias{replace_in_spec}
\title{Razor-like template for diagram specification}
\usage{
replace_in_spec(spec, envir = parent.frame())
}
\arguments{
\item{spec}{String spec to be parsed and evaluated.}

\item{envir}{The environment in which substitution functionality takes place.}
}
\description{
Use Razor-like syntax to define a template for use in a \code{grViz} diagram.
}
\examples{
\dontrun{
# a simple example to use a LETTER as a node label
spec <- "
  digraph { '@1' }

 [1]: LETTERS[1]
"
grViz(replace_in_spec(spec))


spec <- "
digraph a_nice_graph {
node [fontname = Helvetica]
a [label = '@1']
b [label = '@2-1']
c [label = '@2-2']
d [label = '@2-3']
e [label = '@2-4']
f [label = '@2-5']
g [label = '@2-6']
h [label = '@2-7']
i [label = '@2-8']
j [label = '@2-9']
a -> { b c d e f g h i j}
}

[1]: 'top'
[2]: 10:20
"
grViz(replace_in_spec(spec))
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_selection.R
\name{get_selection}
\alias{get_selection}
\title{Get the current selection available in a graph object}
\usage{
get_selection(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A vector with the current selection of node or edge ID values.
}
\description{
Get the current selection of node IDs or edge IDs from a graph object of
class \code{dgr_graph}.
}
\examples{
# Create a simple graph
graph <-
  create_graph() \%>\%
  add_path(n = 6)

# Select node `4`, then select
# all nodes a distance of 1 away
# from node `4`, and finally
# return the selection of nodes as
# a vector object
graph \%>\%
  select_nodes(nodes = 4) \%>\%
  select_nodes_in_neighborhood(
    node = 4,
    distance = 1) \%>\%
  get_selection()

# Select edges associated with
# node `4` and return the
# selection of edges
graph \%>\%
  select_edges_by_node_id(
    nodes = 4) \%>\%
  get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/select_nodes_by_id.R
\name{select_nodes_by_id}
\alias{select_nodes_by_id}
\title{Select nodes in a graph by their ID values}
\usage{
select_nodes_by_id(graph, nodes, set_op = "union")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{nodes}{A vector of node IDs for the selection of nodes present in the
graph.}

\item{set_op}{The set operation to perform upon consecutive selections of
graph nodes. This can either be as a \code{union} (the default), as an
intersection of selections with \code{intersect}, or, as a \code{difference} on the
previous selection, if it exists.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Select nodes in a graph object of class \code{dgr_graph} by their node ID values.
If nodes have IDs that are monotonically increasing integer values, then
numeric ranges can be used for the selection.
}
\examples{
# Create a node data frame (ndf)
ndf <- create_node_df(n = 10)

# Create a graph
graph <-
  create_graph(
    nodes_df = ndf)

# Select nodes `1` to `5` and show that
# selection of nodes with `get_selection()`
graph |>
  select_nodes_by_id(nodes = 1:5) |>
  get_selection()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/recode_edge_attrs.R
\name{recode_edge_attrs}
\alias{recode_edge_attrs}
\title{Recode a set of edge attribute values}
\usage{
recode_edge_attrs(
  graph,
  edge_attr_from,
  ...,
  otherwise = NULL,
  edge_attr_to = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr_from}{The name of the edge attribute column from which values
will be recoded.}

\item{...}{Single-length character vectors with the recoding instructions.
The first component should have the value to replace and the second should
have the replacement value (in the form \verb{"[to_replace] -> [replacement]", ...}).}

\item{otherwise}{An optional single value for recoding any unmatched values.}

\item{edge_attr_to}{An optional name of a new edge attribute to which the
recoded values will be applied. This will retain the original edge
attribute and its values.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal edge data frame (edf), recode character or numeric
edge attribute values. Optionally, one can specify a replacement value for
any unmatched mappings.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 4,
    m = 6,
    set_seed = 23) |>
  set_edge_attrs(
    edge_attr = rel,
    values = c("a", "b", "a",
               "c", "b", "d"))

# Get the graph's internal edf
# to show which edge attributes
# are available
graph |> get_edge_df()

# Recode the `rel` node
# attribute, creating a new edge
# attribute called `penwidth`;
# here, `a` is recoded to `1.0`,
# `b` maps to `1.5`, and all
# other values become `0.5`
graph <-
  graph |>
  recode_edge_attrs(
    edge_attr_from = rel,
    "a -> 1.0",
    "b -> 1.5",
    otherwise = 0.5,
    edge_attr_to = penwidth)

# Get the graph's internal edf
# to show that the node
# attribute values had been
# recoded and copied into a
# new node attribute
graph |> get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_w_connected_cmpts.R
\name{get_w_connected_cmpts}
\alias{get_w_connected_cmpts}
\title{Get all nodes associated with connected components}
\usage{
get_w_connected_cmpts(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame with nodes and their membership in different weakly
connected components.
}
\description{
Determine which nodes in a graph belong to different weakly connected
components (i.e., distinct sets of nodes with traversable paths to and from
each node in the set).
}
\examples{
# Create a graph with 2 cycles
graph <-
  create_graph() |>
  add_cycle(n = 4) |>
  add_cycle(n = 3)

# Check if the graph is connected
graph |>
  is_graph_connected()

# Get the graph's weakly-connected
# components
graph |> get_w_connected_cmpts()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rev_edge_dir_ws.R
\name{rev_edge_dir_ws}
\alias{rev_edge_dir_ws}
\title{Reverse the direction of selected edges in a graph using an edge selection}
\usage{
rev_edge_dir_ws(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Using a directed graph with a selection of edges as input, reverse the
direction of those selected edges in input graph.

This function makes use of an active selection of edges (and the function
ending with \verb{_ws} hints at this).

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Create a graph with a
# directed tree
graph <-
  create_graph() |>
  add_balanced_tree(
    k = 2, h = 2)

# Inspect the graph's edges
graph |> get_edges()

# Select all edges associated
# with nodes `1` and `2`
graph <-
  graph |>
  select_edges_by_node_id(
    nodes = 1:2)

# Reverse the edge directions
# of the edges associated with
# nodes `1` and `2`
graph <-
  graph |>
  rev_edge_dir_ws()

# Inspect the graph's edges
# after their reversal
graph |> get_edges()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rescale_node_attrs.R
\name{rescale_node_attrs}
\alias{rescale_node_attrs}
\title{Rescale numeric node attribute values}
\usage{
rescale_node_attrs(
  graph,
  node_attr_from,
  to_lower_bound = 0,
  to_upper_bound = 1,
  node_attr_to = NULL,
  from_lower_bound = NULL,
  from_upper_bound = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr_from}{The node attribute containing numeric data that is to
be rescaled to new numeric or color values.}

\item{to_lower_bound}{The lower bound value for the set of rescaled values.
This can be a numeric value or an X11 color name.}

\item{to_upper_bound}{The upper bound value for the set of rescaled values.
This can be a numeric value or an X11 color name.}

\item{node_attr_to}{An optional name of a new node attribute to which the
recoded values will be applied. This will retain the original node
attribute and its values.}

\item{from_lower_bound}{An optional, manually set lower bound value for the
rescaled values. If not set, the minimum value from the set will be used.}

\item{from_upper_bound}{An optional, manually set upper bound value for the
rescaled values. If not set, the minimum value from the set will be used.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph}, take a set of numeric values for a
node attribute, rescale to a new numeric or color range, then write to the
same node attribute or to a new node attribute column.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 5,
    m = 10,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = value,
    values = rnorm(
      n = 5,
      mean = 5,
      sd = 1) |> round(1))

# Get the graph's internal ndf
# to show which node attributes
# are available
graph |> get_node_df()

# Rescale the `value` node
# attribute, so that its values
# are rescaled between 0 and 1
graph <-
  graph |>
  rescale_node_attrs(
    node_attr_from = value,
    to_lower_bound = 0,
    to_upper_bound = 1)

# Get the graph's internal ndf
# to show that the node attribute
# values had been rescaled
graph |> get_node_df()

# Scale the values in the `value`
# node attribute to different
# shades of gray for the `fillcolor`
# and `fontcolor` node attributes
graph <-
  graph |>
  rescale_node_attrs(
    node_attr_from = value,
    to_lower_bound = "gray80",
    to_upper_bound = "gray20",
    node_attr_to = fillcolor) |>
  rescale_node_attrs(
    node_attr_from = value,
    to_lower_bound = "gray5",
    to_upper_bound = "gray95",
    node_attr_to = fontcolor)

# Get the graph's internal ndf
# once more to show that scaled
# grayscale colors are now available
# in the `fillcolor` and `fontcolor`
# node attributes
graph |> get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_node_attrs_ws.R
\name{get_node_attrs_ws}
\alias{get_node_attrs_ws}
\title{Get node attribute values from a selection of nodes}
\usage{
get_node_attrs_ws(graph, node_attr)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr}{The name of the attribute for which to get values.}
}
\value{
A named vector of node attribute values for the attribute given by
\code{node_attr} by node ID.
}
\description{
From a graph object of class \code{dgr_graph}, get node attribute values from
nodes currently active as a selection.

This function makes use of an active selection of nodes (and the function
ending with \verb{_ws} hints at this).

Selections of nodes can be performed using the following node selection
(\verb{select_*()}) functions: \code{\link[=select_nodes]{select_nodes()}}, \code{\link[=select_last_nodes_created]{select_last_nodes_created()}},
\code{\link[=select_nodes_by_degree]{select_nodes_by_degree()}}, \code{\link[=select_nodes_by_id]{select_nodes_by_id()}}, or
\code{\link[=select_nodes_in_neighborhood]{select_nodes_in_neighborhood()}}.

Selections of nodes can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out]{trav_out()}}, \code{\link[=trav_in]{trav_in()}}, \code{\link[=trav_both]{trav_both()}},
\code{\link[=trav_out_node]{trav_out_node()}}, \code{\link[=trav_in_node]{trav_in_node()}}, \code{\link[=trav_out_until]{trav_out_until()}}, or
\code{\link[=trav_in_until]{trav_in_until()}}.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 4,
    m = 4,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = value,
    values = c(2.5, 8.2, 4.2, 2.4))

# Select nodes with ID values
# `1` and `3`
graph <-
  graph |>
  select_nodes_by_id(
    nodes = c(1, 3))

# Get the node attribute values
# for the `value` attribute, limited
# to the current node selection
graph |>
  get_node_attrs_ws(
    node_attr = value)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_mean_distance.R
\name{get_mean_distance}
\alias{get_mean_distance}
\title{Get the mean distance}
\usage{
get_mean_distance(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A single numeric value representing the mean distance of the graph.
}
\description{
Get the mean distance of a graph, which is the average path length in the
graph. This operates through calculation of the shortest paths between all
pairs of nodes.
}
\examples{
# Create a cycle graph
graph <-
  create_graph() |>
  add_cycle(n = 5)

# Determine the mean distance
graph |>
  get_mean_distance()

# Create a full graph and then
# get the mean distance value
create_graph() |>
  add_full_graph(n = 10) |>
  get_mean_distance()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/trav_in_node.R
\name{trav_in_node}
\alias{trav_in_node}
\title{Traverse from one or more selected edges onto adjacent, inward nodes}
\usage{
trav_in_node(
  graph,
  conditions = NULL,
  copy_attrs_from = NULL,
  copy_attrs_as = NULL,
  agg = "sum"
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{An option to use filtering conditions for the traversal.}

\item{copy_attrs_from}{Providing an edge attribute name will copy those edge
attribute values to the traversed nodes. If the edge attribute already
exists, the values will be merged to the traversed nodes; otherwise, a new
node attribute will be created.}

\item{copy_attrs_as}{If an edge attribute name is provided in
\code{copy_attrs_from}, this option will allow the copied attribute values to be
written under a different node attribute name. If the attribute name
provided in \code{copy_attrs_as} does not exist in the graph's ndf, the new node
attribute will be created with the chosen name.}

\item{agg}{If an edge attribute is provided to \code{copy_attrs_from}, then an
aggregation function is required since there may be cases where multiple
edge attribute values will be passed onto the traversed node(s). To pass
only a single value, the following aggregation functions can be used:
\code{sum}, \code{min}, \code{max}, \code{mean}, or \code{median}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph} with an active selection of edges
move with the edge direction to connected nodes, replacing the current edges
in the selection with those nodes traversed to. An optional filter by node
attribute can limit the set of nodes traversed to.

This traversal function makes use of an active selection of edges. After the
traversal, depending on the traversal conditions, there will either be a
selection of nodes or no selection at all.

Selections of edges can be performed using the following selection
(\verb{select_*()}) functions: \code{\link[=select_edges]{select_edges()}}, \code{\link[=select_last_edges_created]{select_last_edges_created()}},
\code{\link[=select_edges_by_edge_id]{select_edges_by_edge_id()}}, or \code{\link[=select_edges_by_node_id]{select_edges_by_node_id()}}.

Selections of edges can also be performed using the following traversal
(\verb{trav_*()}) functions: \code{\link[=trav_out_edge]{trav_out_edge()}}, \code{\link[=trav_in_edge]{trav_in_edge()}},
\code{\link[=trav_both_edge]{trav_both_edge()}}, or \code{\link[=trav_reverse_edge]{trav_reverse_edge()}}.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a simple graph
graph <-
  create_graph() \%>\%
  add_n_nodes(
    n = 2,
    type = "a",
    label = c("asd", "iekd")) \%>\%
  add_n_nodes(
    n = 3,
    type = "b",
    label = c("idj", "edl", "ohd")) \%>\%
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->5",
    rel = c(NA, "A", "B", "C", "D"))

# Create a data frame with node ID values
# representing the graph edges (with `from`
# and `to` columns), and, a set of numeric values
df_edges <-
  data.frame(
    from = c(1, 1, 2, 2, 3),
    to = c(2, 3, 4, 5, 5),
    values = round(rnorm(5, 5), 2))

# Create a data frame with node ID values
# representing the graph nodes (with the `id`
# columns), and, a set of numeric values
df_nodes <-
  data.frame(
    id = 1:5,
    values = round(rnorm(5, 7), 2))

# Join the data frame to the graph's internal
# edge data frame (edf)
graph <-
  graph \%>\%
  join_edge_attrs(df = df_edges) \%>\%
  join_node_attrs(df = df_nodes)

# Show the graph's internal node data frame
graph \%>\% get_node_df()

# Show the graph's internal edge data frame
graph \%>\% get_edge_df()

# Perform a simple traversal from the
# edge `1`->`3` to the attached node
# in the direction of the edge; here, no
# conditions are placed on the nodes
# traversed to
graph \%>\%
  select_edges(
    from = 1,
    to = 3) \%>\%
  trav_in_node() \%>\%
  get_selection()

# Traverse from edges `2`->`5` and
# `3`->`5` to the attached node along
# the direction of the edge; both
# traversals lead to the same node
graph \%>\%
  select_edges(
    from = 2,
    to = 5) \%>\%
  select_edges(
    from = 3,
    to = 5) \%>\%
  trav_in_node() \%>\%
  get_selection()

# Traverse from the edge `1`->`3`
# to the attached node where the edge
# is incoming, this time filtering
# numeric values greater than `5.0` for
# the `values` node attribute
graph \%>\%
  select_edges(
    from = 1,
    to = 3) \%>\%
  trav_in_node(
    conditions = values > 5.0) \%>\%
  get_selection()

# Traverse from the edge `1`->`3`
# to the attached node where the edge
# is incoming, this time filtering
# numeric values less than `5.0` for
# the `values` node attribute (the
# condition is not met so the original
# selection of edge `1` -> `3` remains)
graph \%>\%
  select_edges(
    from = 1,
    to = 3) \%>\%
  trav_in_node(
    conditions = values < 5.0) \%>\%
  get_selection()

# Traverse from the edge `1`->`2` to
# the node `2` using multiple conditions
# with a single-length vector
graph \%>\%
  select_edges(
    from = 1,
    to = 2) \%>\%
  trav_in_node(
    conditions =
      grepl(".*d$", label) |
      values < 6.0) \%>\%
  get_selection()

# Create another simple graph to demonstrate
# copying of edge attribute values to traversed
# nodes
graph <-
  create_graph() \%>\%
  add_node() \%>\%
  select_nodes() \%>\%
  add_n_nodes_ws(
    n = 2,
    direction = "to") \%>\%
  clear_selection() \%>\%
  select_nodes_by_id(nodes = 2) \%>\%
  set_node_attrs_ws(
    node_attr = value,
    value = 8) \%>\%
  clear_selection() \%>\%
  select_edges_by_edge_id(edges = 1) \%>\%
  set_edge_attrs_ws(
    edge_attr = value,
    value = 5) \%>\%
  clear_selection() \%>\%
  select_edges_by_edge_id(edges = 2) \%>\%
  set_edge_attrs_ws(
    edge_attr = value,
    value = 5) \%>\%
  clear_selection() \%>\%
  select_edges()

# Show the graph's internal edge data frame
graph \%>\% get_edge_df()

# Show the graph's internal node data frame
graph \%>\% get_node_df()

# Perform a traversal from the edges to
# the central node (`1`) while also applying
# the edge attribute `value` to the node (in
# this case summing the `value` of 5 from
# both edges before adding as a node attribute)
graph <-
  graph \%>\%
  trav_in_node(
    copy_attrs_from = value,
    agg = "sum")

# Show the graph's internal node data frame
# after this change
graph \%>\% get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/edge_data.R
\name{edge_data}
\alias{edge_data}
\title{Insert edge data attributes during edge creation}
\usage{
edge_data(...)
}
\arguments{
\item{...}{Edge data attributes provided as one or more named vectors.}
}
\description{
This helper function should be invoked to provide values for the namesake
\code{edge_data} argument, which is present in any function where edges are
created.
}
\examples{
\dontrun{
# Create a new graph and add
# a path with several edge
# data attributes
graph <-
  create_graph() |>
  add_path(
    n = 3,
    type = "path",
    edge_data = edge_data(
      hour = 5,
      index = c(1, 2)))

# View the graph's internal
# edge data frame; the edge
# data attributes have
# been inserted
graph |> get_edge_df()
}

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edges.R
\name{get_edges}
\alias{get_edges}
\title{Get node IDs associated with edges}
\usage{
get_edges(
  graph,
  conditions = NULL,
  return_type = "vector",
  return_values = "id"
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{conditions}{an option to use filtering conditions for the retrieval of
edges.}

\item{return_type}{using \code{vector} (the default), a vector of character
objects representing the edges is provided. With \code{list} a list object will
be provided that contains vectors of outgoing and incoming node IDs
associated with edges. With \code{df}, a data frame containing outgoing and
incoming node IDs associated with edges.}

\item{return_values}{using \code{id} (the default) results in node ID values
returned in the edge definitions. With \code{label}, the node labels will
instead be used to define edges.}
}
\value{
A list, data frame, or a vector object, depending on the value given
to \code{return_type}.
}
\description{
Obtain a vector, data frame, or list of node IDs associated with edges in a
graph object. An optional filter by edge attribute can limit the set of edges
returned.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 4,
    label = c("one", "two", "three", "four"),
    type = "letter",
    color = c("red", "green", "grey", "blue"),
    value = c(3.5, 2.6, 9.4, 2.7))

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "leading_to",
    color = c("pink", "blue", "blue"),
    value = c(3.9, 2.5, 7.3))

# Create a graph
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Get all edges within a graph, returned as a list
graph \%>\%
  get_edges(
    return_type = "vector")

# Get all edges within a graph, returned as a
# data frame
graph \%>\%
  get_edges(
    return_type = "df")

# Get all edges returned as a list
graph \%>\%
  get_edges(
    return_type = "list")

# Get a vector of edges using
# a numeric comparison (i.e.,
# all edges with a `value`
# attribute greater than 3)
graph \%>\%
  get_edges(
    conditions = value > 3,
    return_type = "vector")

# Get a vector of edges using
# a matching condition
graph \%>\%
  get_edges(
    conditions = color == "pink",
    return_type = "vector")

# Use multiple conditions to
# return edges with the
# desired attribute values
graph \%>\%
  get_edges(
    conditions =
      color == "blue" &
      value > 3,
    return_type = "vector")

# Use `return_values = "label"`
# to return the labels of the
# connected nodes
graph \%>\%
  get_edges(
    conditions =
      color == "blue" &
      value > 3,
    return_type = "vector",
    return_values = "label")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/transform_to_complement_graph.R
\name{transform_to_complement_graph}
\alias{transform_to_complement_graph}
\title{Create a complement of a graph}
\usage{
transform_to_complement_graph(graph, loops = FALSE)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{loops}{An option for whether loops should be generated in the
complement graph.}
}
\value{
a graph object of class \code{dgr_graph}.
}
\description{
Create a complement graph which contains only edges not present in the input
graph. It's important to nodes that any edge attributes in the input graph's
edges will be lost. Node attributes will be retained, since they are not
affected by this transformation.
}
\examples{
# Create a simple graph
# with a single cycle
graph <-
  create_graph() \%>\%
  add_cycle(n = 4)

# Get the graph's edge
# data frame
graph \%>\% get_edge_df()

# Create the complement
# of the graph
graph_c <-
  graph \%>\%
    transform_to_complement_graph()

# Get the edge data frame
# for the complement graph
graph_c \%>\% get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_cycle.R
\name{add_cycle}
\alias{add_cycle}
\title{Add a cycle of nodes to the graph}
\usage{
add_cycle(
  graph,
  n,
  type = NULL,
  label = TRUE,
  rel = NULL,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{n}{The number of nodes comprising the cycle.}

\item{type}{An optional string that describes the entity type for the nodes
to be added.}

\item{label}{Either a vector object of length \code{n} that provides optional
labels for the new nodes, or, a boolean value where setting to \code{TRUE}
ascribes node IDs to the label and \code{FALSE} yields a blank label.}

\item{rel}{An optional string for providing a relationship label to all new
edges created in the node cycle.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add a node cycle to the graph.
}
\examples{
# Create a new graph and
# add a cycle of nodes to it
graph <-
  create_graph() |>
  add_cycle(n = 6)

# Get node information
# from this graph
graph |>
  get_node_info()

# Node and edge aesthetic and data
# attributes can be specified in
# the `node_aes`, `edge_aes`,
# `node_data`, and `edge_data`
# arguments

suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

graph_w_attrs <-
  create_graph() |>
  add_cycle(
    n = 3,
    label = c(
      "one", "two", "three"),
    type = c(
      "a", "a", "b"),
    rel = "A",
    node_aes = node_aes(
      fillcolor = "steelblue"),
    edge_aes = edge_aes(
      color = "red",
      penwidth = 1.2),
    node_data = node_data(
      value = c(
        1.6, 2.8, 3.4)),
    edge_data = edge_data(
      value =
        rnorm(
          n = 3,
          mean = 5.0,
          sd = 1.0)))

# Get the graph's node data frame
graph_w_attrs |> get_node_df()

# Get the graph's edge data frame
graph_w_attrs |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/combine_graphs.R
\name{combine_graphs}
\alias{combine_graphs}
\title{Combine two graphs into a single graph}
\usage{
combine_graphs(x, y)
}
\arguments{
\item{x}{A \code{DiagrammeR} graph object to which another graph will be unioned.
This graph should be considered the graph from which global graph
attributes will be inherited in the resulting graph.}

\item{y}{A \code{DiagrammeR} graph object that is to be unioned with the graph
supplied as \code{x}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Combine two graphs in order to make a new graph.
}
\examples{
# Create a graph with a cycle
# containing 6 nodes
graph_cycle <-
 create_graph() \%>\%
   add_cycle(n = 6)

# Create a random graph with
# 8 nodes and 15 edges using the
# `add_gnm_graph()` function
graph_random <-
  create_graph() \%>\%
  add_gnm_graph(
    n = 8,
    m = 15,
    set_seed = 23)

# Combine the two graphs in a
# union operation
combined_graph <-
  combine_graphs(
    graph_cycle,
    graph_random)

# Get the number of nodes in
# the combined graph
combined_graph \%>\% count_nodes()

# The `combine_graphs()`
# function will renumber
# node ID values in graph `y`
# during the union; this ensures
# that node ID values are unique
combined_graph \%>\% get_node_ids()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/join_node_attrs.R
\name{join_node_attrs}
\alias{join_node_attrs}
\title{Join new node attribute values using a data frame}
\usage{
join_node_attrs(graph, df, by_graph = NULL, by_df = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{df}{The data frame to use for joining.}

\item{by_graph}{Optional specification of the column in the graph's internal
node data frame for the left join. If both \code{by_graph} and \code{by_df} are not
provided, then a natural join will occur if there are columns in the
graph's ndf and in \code{df} with identical names.}

\item{by_df}{Optional specification of the column in \code{df} for the left join.
If both \code{by_graph} and \code{by_df} are not provided, then a natural join will
occur if there are columns in the graph's ndf and in \code{df} with identical
names.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Join new node attribute values in a left join using a data frame. The use of
a left join in this function allows for no possibility that nodes in the
graph might be removed after the join.
}
\examples{
# Set a seed
suppressWarnings(RNGversion("3.5.0"))
set.seed(23)

# Create a simple graph
graph <-
  create_graph() |>
  add_n_nodes(n = 5) |>
  add_edges_w_string(
    edges = "1->2 1->3 2->4 2->5 3->5")

# Create a data frame with node ID values and a
# set of numeric values
df <-
  data.frame(
    values = round(rnorm(6, 5), 2),
    id = 1:6)

# Join the values in the data frame to the
# graph's nodes; this works as a left join using
# identically-named columns in the graph and the df
# (in this case the `id` column is common to both)
graph <-
  graph |>
  join_node_attrs(
    df = df)

# Get the graph's internal ndf to show that the
# join has been made
graph |> get_node_df()

# Get betweenness values for each node and
# add them as a node attribute (Note the
# common column name `id` in the different
# tables results in a natural join)
graph <-
  graph |>
  join_node_attrs(
    df = get_betweenness(graph))

# Get the graph's internal ndf to show that
# this join has been made
graph |> get_node_df()
}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{drop_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_edge_attr_to_display.R
\name{set_edge_attr_to_display}
\alias{set_edge_attr_to_display}
\title{Set the edge attribute values to be rendered}
\usage{
set_edge_attr_to_display(graph, attr = NULL, edges = NULL, default = "label")
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{attr}{The name of the attribute from which label text for the edge will
be obtained. If set to \code{NULL}, then \code{NA} values will be assigned to the
\code{display} column for the chosen edges.}

\item{edges}{A length vector containing one or several edge ID values (as
integers) for which edge attributes are set for display in the rendered
graph. If \code{NULL}, all edges from the graph are assigned the \code{display} value
given as \code{attr}.}

\item{default}{The name of an attribute to set for all other graph edges not
included in \code{edges}. This value only gets used if the \code{display} edge
attribute is not in the graph's internal edge data frame.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Set a edge attribute type to display as edge text when calling the
\code{\link[=render_graph]{render_graph()}} function. This allows for display of different types of edge
attribute values on a per-edge basis. Without setting the \code{display}
attribute, rendering a graph will default to not printing any text on edges.
Setting the \code{display} edge attribute with this function for the first time
(i.e., the \code{display} column doesn't exist in the graph's internal edge data
frame) will insert the \code{attr} value for all edges specified in \code{edges} and a
default value (\code{default}) for all remaining edges.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 4,
    m = 4,
    set_seed = 23) |>
  set_edge_attrs(
    edge_attr = value,
    values = c(2.5, 8.2, 4.2, 2.4))

# For edge ID values of `1`,
# `2`, and `3`, choose to display
# the edge `value` attribute (for
# the other edges, display nothing)
graph <-
  graph |>
  set_edge_attr_to_display(
    edges = 1:3,
    attr = value,
    default = NA)

# Show the graph's edge data frame; the
# `display` edge attribute will show, for
# each row, which edge attribute value to
# display when the graph is rendered
graph |> get_edge_df()

# This function can be called multiple
# times on a graph; after the first time
# (i.e., creation of the `display`
# attribute), the `default` value won't
# be used
graph |>
  set_edge_attr_to_display(
    edges = 4,
    attr = to) |>
  set_edge_attr_to_display(
    edges = c(1, 3),
    attr = id) |>
  get_edge_df()

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_node_attrs.R
\name{get_node_attrs}
\alias{get_node_attrs}
\title{Get node attribute values}
\usage{
get_node_attrs(graph, node_attr, nodes = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr}{The name of the attribute for which to get values.}

\item{nodes}{An optional vector of node IDs for filtering list of nodes
present in the graph or node data frame.}
}
\value{
A named vector of node attribute values for the attribute given by
\code{node_attr} by node ID.
}
\description{
From a graph object of class \code{dgr_graph}, get node attribute values for one
or more nodes.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 4,
    m = 4,
    set_seed = 23) |>
  set_node_attrs(
    node_attr = value,
    values = c(2.5, 8.2, 4.2, 2.4))

# Get all of the values from
# the `value` node attribute
# as a named vector
graph |>
  get_node_attrs(
    node_attr = value)

# To only return node attribute
# values for specified nodes,
# use the `nodes` argument
graph |>
  get_node_attrs(
    node_attr = value,
    nodes = c(1, 3))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/drop_node_attrs.R
\name{drop_node_attrs}
\alias{drop_node_attrs}
\title{Drop a node attribute column}
\usage{
drop_node_attrs(graph, node_attr)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{node_attr}{The name of the node attribute column to drop.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Within a graph's internal node data frame (ndf), remove an existing node
attribute.
}
\examples{
graph <-
  create_graph() |>
  add_gnm_graph(
    n = 5,
    m = 10,
    set_seed = 23)

graph <-
  graph |>
 set_node_attrs(
    node_attr = value,
    values = rnorm(
      n = 5,
      mean = 5,
      sd = 1) |> round(1))

# Get the graph's internal
# ndf to show which node
# attributes are available
graph |> get_node_df()

# Drop the `value` node
# attribute
graph <-
  graph |>
  drop_node_attrs(
    node_attr = value)

# Get the graph's internal
# ndf to show that the node
# attribute `value` had been
# removed
graph |> get_node_df()

}
\seealso{
Other node creation and removal: 
\code{\link{add_n_node_clones}()},
\code{\link{add_n_nodes}()},
\code{\link{add_n_nodes_ws}()},
\code{\link{add_node}()},
\code{\link{add_node_clones_ws}()},
\code{\link{add_node_df}()},
\code{\link{add_nodes_from_df_cols}()},
\code{\link{add_nodes_from_table}()},
\code{\link{colorize_node_attrs}()},
\code{\link{copy_node_attrs}()},
\code{\link{create_node_df}()},
\code{\link{delete_node}()},
\code{\link{delete_nodes_ws}()},
\code{\link{join_node_attrs}()},
\code{\link{layout_nodes_w_string}()},
\code{\link{mutate_node_attrs}()},
\code{\link{mutate_node_attrs_ws}()},
\code{\link{node_data}()},
\code{\link{recode_node_attrs}()},
\code{\link{rename_node_attrs}()},
\code{\link{rescale_node_attrs}()},
\code{\link{set_node_attr_to_display}()},
\code{\link{set_node_attr_w_fcn}()},
\code{\link{set_node_attrs}()},
\code{\link{set_node_attrs_ws}()},
\code{\link{set_node_position}()}
}
\concept{node creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_global_graph_attrs.R
\name{add_global_graph_attrs}
\alias{add_global_graph_attrs}
\title{Add one or more global graph attributes}
\usage{
add_global_graph_attrs(graph, attr, value, attr_type)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{attr}{The name of the attribute to set for the \code{type} of global
attribute specified.}

\item{value}{The value to be set for the chosen attribute specified in the
\code{attr_for_type} argument.}

\item{attr_type}{The specific type of global graph attribute to set. The type
is specified with \code{graph}, \code{node}, or \code{edge}.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
Add global attributes of a specific type (either \code{graph_attrs}, \code{node_attrs},
or \code{edge_attrs} for a graph object of class \code{dgr_graph}).
}
\examples{
# Create a new graph with no
# global graph attributes and
# add a global graph attribute
graph <-
  create_graph(
    attr_theme = NULL) |>
  add_global_graph_attrs(
    attr = "overlap",
    value = "true",
    attr_type = "graph")

# Verify that the attribute
# addition has been made
graph |>
  get_global_graph_attr_info()

# Add another attribute with
# `add_global_graph_attrs()`
graph <-
  graph |>
  add_global_graph_attrs(
    attr = "penwidth",
    value = 12,
    attr_type = "node")

# Verify that the attribute
# addition has been made
graph |>
  get_global_graph_attr_info()

# When adding an attribute where
# `attr` and `attr_type` already
# exists, the value provided will
# serve as an update
graph |>
  add_global_graph_attrs(
    attr = "penwidth",
    value = 15,
    attr_type = "node") |>
  get_global_graph_attr_info()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/set_edge_attrs.R
\name{set_edge_attrs}
\alias{set_edge_attrs}
\title{Set edge attribute values}
\usage{
set_edge_attrs(graph, edge_attr, values, from = NULL, to = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{edge_attr}{The name of the attribute to set. Some examples are located in \code{\link[=edge_aes]{edge_aes()}}}

\item{values}{The values to be set for the chosen attribute for the chosen
edges.}

\item{from}{An optional vector of node IDs from which the edge is outgoing
for filtering list of nodes with outgoing edges in the graph.}

\item{to}{An optional vector of node IDs from which the edge is incoming for
filtering list of nodes with incoming edges in the graph.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
From a graph object of class \code{dgr_graph}, set edge attribute values for one
or more edges.
}
\examples{
# Create a simple graph
ndf <-
  create_node_df(
    n = 4,
    type = "basic",
    label = TRUE,
    value = c(3.5, 2.6, 9.4, 2.7))

edf <-
  create_edge_df(
    from = c(1, 2, 3),
    to = c(4, 3, 1),
    rel = "leading_to")

graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Set attribute `color = "green"`
# for edges `1`->`4` and `3`->`1`
# in the graph
graph <-
  graph \%>\%
  set_edge_attrs(
    edge_attr = color,
    values = "green",
    from = c(1, 3),
    to = c(4, 1))

# Set attribute `color = "blue"`
# for all edges in the graph
graph <-
  graph \%>\%
  set_edge_attrs(
    edge_attr = color,
    values = "blue")

# Set attribute `color = "pink"`
# for all edges in graph outbound
# from node with ID value `1`
graph <-
  graph \%>\%
  set_edge_attrs(
    edge_attr = color,
    values = "pink",
    from = 1)

# Set attribute `color = "black"`
# for all edges in graph inbound
# to node with ID `1`
graph <-
  graph \%>\%
  set_edge_attrs(
    edge_attr = color,
    values = "black",
    to = 1)

}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge}()},
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_edge_count_w_multiedge.R
\name{get_edge_count_w_multiedge}
\alias{get_edge_count_w_multiedge}
\title{Get count of edge definitions where multiple edges occur}
\usage{
get_edge_count_w_multiedge(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A vector with a single, numerical value.
}
\description{
Get a count of the number of edge definitions (e.g, \code{1} -> \code{2}) where there
are multiple edges (i.e., more than 1 edge of that definition, having
distinct edge ID values). So, for example, if there are 2 edge definitions in
the graph that involve 6 separate edge IDs (3 such edge IDs for each of the
pairs of nodes), the count will be \code{2}.
}
\examples{
# Create a node data frame (ndf)
ndf <-
  create_node_df(
    n = 5,
    label = TRUE)

# Create an edge data frame (edf)
edf <-
  create_edge_df(
    from = c(1, 4, 4, 3, 5, 1, 3, 4),
      to = c(4, 1, 1, 2, 2, 2, 2, 1))

# Create a graph with the ndf and edf
graph <-
  create_graph(
    nodes_df = ndf,
    edges_df = edf)

# Get the total number of edge
# definitions (e.g., `4` -> `1`) where
# there are multiple edges (i.e.,
# distinct edges with separate edge
# ID values)
graph |> get_edge_count_w_multiedge()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_leverage_centrality.R
\name{get_leverage_centrality}
\alias{get_leverage_centrality}
\title{Get leverage centrality}
\usage{
get_leverage_centrality(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A data frame with leverage centrality values for each of the nodes.
}
\description{
Get the leverage centrality values for all nodes in the graph. Leverage
centrality is a measure of the relationship between the degree of a given
node and the degree of each of its neighbors, averaged over all neighbors. A
node with negative leverage centrality is influenced by its neighbors, as the
neighbors connect and interact with far more nodes. A node with positive
leverage centrality influences its neighbors since the neighbors tend to have
far fewer connections.
}
\examples{
# Create a random graph using the
# `add_gnm_graph()` function
graph <-
  create_graph(
    directed = FALSE) |>
  add_gnm_graph(
    n = 10,
    m = 15,
    set_seed = 23)

# Get leverage centrality values
# for all nodes in the graph
graph |>
  get_leverage_centrality()

# Add the leverage centrality
# values to the graph as a
# node attribute
graph <-
  graph |>
  join_node_attrs(
    df = get_leverage_centrality(graph))

# Display the graph's node data frame
graph |> get_node_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_last_nodes_created.R
\name{get_last_nodes_created}
\alias{get_last_nodes_created}
\title{Get the last set of nodes created in a graph}
\usage{
get_last_nodes_created(graph)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}
}
\value{
A vector of node ID values.
}
\description{
Get the last nodes that were created in a graph object of class \code{dgr_graph}.
Provides a vector of node ID values. This function should ideally be used
just after creating the nodes.
}
\examples{
# Create a graph and add 4 nodes
# in 2 separate function calls
graph <-
  create_graph() \%>\%
  add_n_nodes(
    n = 2,
    type = "a",
    label = c("a_1", "a_2")) \%>\%
  add_n_nodes(
    n = 2,
    type = "b",
    label = c("b_1", "b_2"))

# Get the last nodes created (2 nodes
# from the last function call)
graph \%>\% get_last_nodes_created()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_grid_2d.R
\name{add_grid_2d}
\alias{add_grid_2d}
\title{Add a 2D grid of nodes to the graph}
\usage{
add_grid_2d(
  graph,
  x,
  y,
  type = NULL,
  label = TRUE,
  rel = NULL,
  node_aes = NULL,
  edge_aes = NULL,
  node_data = NULL,
  edge_data = NULL
)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{x}{The number of nodes in the x direction.}

\item{y}{The number of nodes in the y direction.}

\item{type}{An optional string that describes the entity type for the nodes
to be added.}

\item{label}{Either a vector object of length \code{x * y} that provides optional
labels for the new nodes, or, a logical value where setting to \code{TRUE}
ascribes node IDs to the label and \code{FALSE} yields a blank label.}

\item{rel}{An optional string for providing a relationship label to all new
edges created in the grid.}

\item{node_aes}{An optional list of named vectors comprising node aesthetic
attributes. The helper function \code{\link[=node_aes]{node_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted node aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{color}, \code{fillcolor}).}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{node_data}{An optional list of named vectors comprising node data
attributes. The helper function \code{\link[=node_data]{node_data()}} is strongly recommended for
use here as it helps bind data specifically to the created nodes.}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add a two-dimensional grid to the
graph.
}
\examples{
# Create a new graph and add
# a 3 x 3 grid
graph <-
  create_graph() |>
  add_grid_2d(
    x = 3, y = 3,
    type = "grid")

# Get node information
# from this graph
graph |>
  get_node_info()

# Attributes can be specified
# in extra arguments and these
# are applied in order; Usually
# these attributes are applied
# to nodes (e.g., `type` is a
# node attribute) but the `rel`
# attribute will apply to the
# edges
graph_w_attrs <-
  create_graph() |>
  add_grid_2d(
    x = 3, y = 2,
    label = c("one", "two",
              "three", "four",
              "five", "six"),
    type = c("a", "a",
             "b", "b",
             "c", "c"),
    rel = "grid",
    node_data = node_data(
      value = c(
        1.2, 8.4, 3.4,
        5.2, 6.1, 2.6)))

# Get the graph's node data frame
graph_w_attrs |> get_node_df()

# Get the graph's edge data frame
graph_w_attrs |> get_edge_df()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/add_edge.R
\name{add_edge}
\alias{add_edge}
\title{Add an edge between nodes in a graph object}
\usage{
add_edge(graph, from, to, rel = NULL, edge_aes = NULL, edge_data = NULL)
}
\arguments{
\item{graph}{A graph object of class \code{dgr_graph}.}

\item{from}{The outgoing node from which the edge is connected. There is the
option to use a node \code{label} value here (and this must correspondingly also
be done for the \code{to} argument) for defining node connections. Note that
this is only possible if all nodes have distinct \code{label} values set and
none exist as an empty string.}

\item{to}{The incoming nodes to which each edge is connected. There is the
option to use a node \code{label} value here (and this must correspondingly also
be done for the \code{from} argument) for defining node connections. Note that
this is only possible if all nodes have distinct \code{label} values set and
none exist as an empty string.}

\item{rel}{An optional string specifying the relationship between the
connected nodes.}

\item{edge_aes}{An optional list of named vectors comprising edge aesthetic
attributes. The helper function \code{\link[=edge_aes]{edge_aes()}} is strongly recommended for
use here as it contains arguments for each of the accepted edge aesthetic
attributes (e.g., \code{shape}, \code{style}, \code{penwidth}, \code{color}).}

\item{edge_data}{An optional list of named vectors comprising edge data
attributes. The helper function \code{\link[=edge_data]{edge_data()}} is strongly recommended for
use here as it helps bind data specifically to the created edges.}
}
\value{
A graph object of class \code{dgr_graph}.
}
\description{
With a graph object of class \code{dgr_graph}, add an edge to nodes within the
graph.
}
\examples{
# Create a graph with 4 nodes
graph <-
  create_graph() |>
  add_node(label = "one") |>
  add_node(label = "two") |>
  add_node(label = "three") |>
  add_node(label = "four")

# Add an edge between those
# nodes and attach a
# relationship to the edge
graph <-
 add_edge(
   graph,
   from = 1,
   to = 2,
   rel = "A")

# Use the `get_edge_info()`
# function to verify that
# the edge has been created
graph |>
  get_edge_info()

# Add another node and
# edge to the graph
graph <-
  graph |>
  add_edge(
    from = 3,
    to = 2,
    rel = "A")

# Verify that the edge
# has been created by
# counting graph edges
graph |> count_edges()

# Add edges by specifying
# node `label` values; note
# that all nodes must have
# unique `label` values to
# use this option
graph <-
  graph |>
  add_edge(
    from = "three",
    to = "four",
    rel = "L") |>
  add_edge(
    from = "four",
    to = "one",
    rel = "L")

# Use `get_edges()` to verify
# that the edges were added
graph |> get_edges()

# Add edge aesthetic and data
# attributes during edge creation
graph_2 <-
  create_graph() |>
  add_n_nodes(n = 2) |>
  add_edge(
    from = 1,
    to = 2,
    rel = "M",
    edge_aes = edge_aes(
      penwidth = 1.5,
      color = "blue"),
    edge_data = edge_data(
      value = 4.3))

# Use the `get_edges()` function
# to verify that the attribute
# values were bound to the
# newly created edge
graph_2 |> get_edge_df()


}
\seealso{
Other edge creation and removal: 
\code{\link{add_edge_clone}()},
\code{\link{add_edge_df}()},
\code{\link{add_edges_from_table}()},
\code{\link{add_edges_w_string}()},
\code{\link{add_forward_edges_ws}()},
\code{\link{add_reverse_edges_ws}()},
\code{\link{copy_edge_attrs}()},
\code{\link{create_edge_df}()},
\code{\link{delete_edge}()},
\code{\link{delete_edges_ws}()},
\code{\link{delete_loop_edges_ws}()},
\code{\link{drop_edge_attrs}()},
\code{\link{edge_data}()},
\code{\link{join_edge_attrs}()},
\code{\link{mutate_edge_attrs}()},
\code{\link{mutate_edge_attrs_ws}()},
\code{\link{recode_edge_attrs}()},
\code{\link{rename_edge_attrs}()},
\code{\link{rescale_edge_attrs}()},
\code{\link{rev_edge_dir}()},
\code{\link{rev_edge_dir_ws}()},
\code{\link{set_edge_attr_to_display}()},
\code{\link{set_edge_attrs}()},
\code{\link{set_edge_attrs_ws}()}
}
\concept{edge creation and removal}
